package com.netapp.acquisition.aquisitionserverside.interfaces.data;

import javax.persistence.*;
import java.io.Serializable;
import java.util.*;

/**
 * Created by weis.
 * Date: Jun 8, 2010
 * Time: 11:46:44 AM
 */
@NamedQueries({
        @NamedQuery(
                name = "findAcqUnitWithAttributesByName",
                query = "select DISTINCT  a from AcquisitionUnit a left join fetch a.attributes where a.name=:name"
        ),
        @NamedQuery(
                name = "findAcqUnitByName",
                query = "select a from AcquisitionUnit a where a.name=:name"
        ),
        @NamedQuery(
                name = "findAllAcqUnits",
                query = "select a from AcquisitionUnit a"
        )
})
@Entity
@Table(name = "au")
public class AcquisitionUnit implements Serializable {
    @Transient
    private static final long serialVersionUID = 1L;

    @Id
    @GeneratedValue
    private long id;

    @Column(name = "name", nullable = false)
    private String name;

    @Column(name = "ip", nullable = false)
    private String ip;

    @Column(name = "status", nullable = false)
    private String statusString;

//    @Transient
//    private Status status;

    @Column(name = "leaseperiod", nullable = false)
    private long leasePeriod;

    @Column(name = "nextleaserenewal", nullable = false)
    private long nextLeaseRenewal;

    @OneToMany(cascade = CascadeType.ALL, mappedBy = "acquisitionUnit")
    private Set<AcquisitionUnitAttribute> attributes = new HashSet<AcquisitionUnitAttribute>();

    @Version
    @Column(name = "version")
    private long version;

    @Transient
    Map<AcquisitionUnitAttribute.AcquisitionUnitAttributeBusinessKey, AcquisitionUnitAttribute> attributesByBusinessKey = null;


    public AcquisitionUnit() {
    }

    public AcquisitionUnit(String name, String ip, Status status, long leasePeriod, long nextLeaseRenewal) {
        this.name = name;
        this.ip = ip;
//        this.status = status;
        this.leasePeriod = leasePeriod;
        this.nextLeaseRenewal = nextLeaseRenewal;
        this.statusString = status.getStatus();
    }

    public long getId() {
        return id;
    }

    public void setId(long id) {
        this.id = id;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getIp() {
        return ip;
    }

    public void setIp(String ip) {
        this.ip = ip;
    }

    public Status getStatus() {
        return Status.parseString(statusString);
    }

    public void setStatus(Status status) {
        this.statusString = status.getStatus();
    }

    public long getLeasePeriod() {
        return leasePeriod;
    }

    public void setLeasePeriod(long leasePeriod) {
        this.leasePeriod = leasePeriod;
    }

    public long getNextLeaseRenewal() {
        return nextLeaseRenewal;
    }

    public void setNextLeaseRenewal(long nextLeaseRenewal) {
        this.nextLeaseRenewal = nextLeaseRenewal;
    }

    public boolean isLeaseExpired () {
        return nextLeaseRenewal < System.currentTimeMillis();
    }


    public boolean isLocal() {
        return "local".equals(name);
    }

    /**
     * @return the AU attributes (non-modifiable) - use {@link #addAttribute(AcquisitionUnitAttribute)} to change
     */
    public Set<AcquisitionUnitAttribute> getAttributes() {
        return Collections.unmodifiableSet(attributes);
    }

//    public void setAttributes(Set<AcquisitionUnitAttribute> attributes) {
//        this.attributes = attributes;
//    }

    public void removeAttribute(AcquisitionUnitAttribute attribute) {
        populateAttributesByBusinessKey();
        attribute.setAcquisitionUnit(null);
        attributes.remove(attribute);
        attributesByBusinessKey.remove(attribute.createBusinessKey());
    }

    public void addAttribute(AcquisitionUnitAttribute attribute) {
        populateAttributesByBusinessKey();
        attribute.setAcquisitionUnit(this);
        AcquisitionUnitAttribute existingAttribute = attributesByBusinessKey.get(attribute.createBusinessKey());
        if (existingAttribute != null) {
            // Just override the value of the existing attribute
            existingAttribute.setValue(attribute.getValue());
        } else {
            // Make up a new attribute
            attributes.add(attribute);
            attributesByBusinessKey.put(attribute.createBusinessKey(), attribute);
        }
    }

    private void populateAttributesByBusinessKey() {
        if (attributesByBusinessKey == null) {
            attributesByBusinessKey = new HashMap<AcquisitionUnitAttribute.AcquisitionUnitAttributeBusinessKey, AcquisitionUnitAttribute>();
            for (AcquisitionUnitAttribute attribute : attributes) {
                attributesByBusinessKey.put(attribute.createBusinessKey(), attribute);
            }
        }
    }

    public long getVersion() {
        return version;
    }

    public void setVersion(long version) {
        this.version = version;
    }

    public static enum Status implements Serializable {
        NOT_CONNECTED("Not connected"),
        CONNECTED("ok"),
        WARN("WARNING");

        private final static long serialVersionUID = 1L;
        private String status;

        Status(final String status) {
            this.status = status;
        }

        public String getStatus() {
            return status;
        }

        public String toString() {
            return status;
        }

        public int toInt() {
            switch (this) {
                case NOT_CONNECTED:
                    return -1;
                case CONNECTED:
                    return 1;
                case WARN:
                    return 2;
                default:
                    throw new IllegalStateException("Unknown AcquisitionUnit status : " + status);
            }
        }

        public static Status parseString(final String str) {
            for (Status s : Status.values()) {
                if (s.status.equals(str))
                    return s;
            }

            throw new IllegalArgumentException("Invalid AcquisitionUnitStatus type '" + str + "'");
        }

        public static Status fromInt(final int status) {
            switch (status) {
                case -1:
                    return NOT_CONNECTED;
                case 1:
                    return CONNECTED;
                default:
                    return WARN;
            }
        }
    }
}
