package com.netapp.acquisition.aquisitionserverside.interfaces.data;

import java.io.Serializable;

/**
 * Container class for disk model attributes.
 */
public class DiskModelDescription implements Serializable {

    // Some default attributes
    public static final int DEFAULT_INTERFACE_SPEED = 2000;
    public static final float DEFAULT_SEEK_TIME = 5.0f;
    public static final float DEFAULT_POTENTIAL_TRANSFER_RATE = 80.0f;

    private static final long serialVersionUID = 1L;

    private final String vendor;
    private final String model;
    private final int rpm;
    private final int sizeGib;
    private final Float seekTime;
    private final Float potentialTransferRate;
    private final Float interfaceSpeed;
    private String extraInfo;
    private final String interfaceType;

    public static final int DISK_MAX_SIZE = 10240; //10TB 10*1024 GB
    public static final int DISK_MAX_RPM = 50000;
    private String key;

    /**
     * @param key
     * @param vendor
     * @param model                 Model code as given by vendor
     * @param rpm                   Disk speed in Revolutions Per Minute
     * @param sizeGib               Size of disk in GB (1000 based)
     * @param interfaceType         Disk interface type (FC / SATA / FATA / ...)
     * @param seekTime              Disk seek time in ms
     * @param potentialTransferRate Disk's potential transfer rate in MB/sec
     * @param interfaceSpeed        Disk inteface speec (FC / SCSI, etc...) in MB/sec
     * @param extraInfo             Additional information (text) about this disk
     */
    public DiskModelDescription(final String key, final String vendor, final String model, int rpm, final int sizeGib, final String interfaceType, final Float seekTime, final Float potentialTransferRate, Float interfaceSpeed, final String extraInfo) {
        this.key = key;

        // TODO: validation?   interface type validation?
        this.vendor = vendor;
        this.model = model;
        this.rpm = rpm;
        this.sizeGib = sizeGib;
        this.seekTime = seekTime;
        this.potentialTransferRate = potentialTransferRate;
        this.interfaceType = interfaceType;
        this.interfaceSpeed = interfaceSpeed;
        this.extraInfo = extraInfo;
    }

    public String getVendor() {
        return vendor;
    }

    /**
     * @return Model of the disk
     */
    public final String getModel() {
        return model;
    }

    /**
     * @return disk speed (e.g. 10000, 15000, 7200, etc...)
     */
    public final int getRpm() {
        return rpm;
    }

    /**
     * @return Disk size in GB (1000 based)
     */
    public final long getSizeGiB() {
        return sizeGib;
    }

    //    /**
    //     *
    //     * @return Capacity of the disk
    //     */
    //    public final Capacity getCapacity() {
    //        return Capacity.create(sizeGib, Un);
    //    }

    /**
     * @return Disk seek time in ms
     */
    public final Float getSeekTime() {
        return seekTime;
    }

    public final Float getSeekTimeOrDefault() {
        return (seekTime == null)
               ? DEFAULT_SEEK_TIME
               : seekTime;
    }

    /**
     * @return Disk's potential transfer rate in MB/sec
     */
    public final Float getPotentialTransferRate() {
        return potentialTransferRate;
    }

    public final Float getPotentialTransferRateOrDefault() {
        return (potentialTransferRate == null)
               ? DEFAULT_POTENTIAL_TRANSFER_RATE
               : potentialTransferRate;
    }

    /**
     * @return disk inteface speec (FC / SCSI, etc...) in MB/sec
     */
    public final Float getInterfaceSpeedInternal() {
        return interfaceSpeed;
    }

    public final Float getInterfaceSpeedOrDefault() {
        return (interfaceSpeed == null)
               ? DEFAULT_INTERFACE_SPEED
               : interfaceSpeed;
    }

    /**
     * @return Additional information (text) about this disk
     */
    public final String getExtraInfo() {
        return extraInfo;
    }

    /**
     * @return Disk interface type (FC / SATA / FATA / ...)
     */
    public final String getInterfaceType() {
        return interfaceType;
    }

    public String getKey() {
        return key;
    }

    /**
     * @return
     */
    public String toConfigCSV() {

        // For configuration files, does NOT include default values

        // Fields should be comma separated, so no commas should be used within the fields:
        //   1)  Key(use the model number)
        //   2)  Vendor
        //   3)  Model number
        //   4)  RPM of the disk, e.g. 10000, 15000, etc.
        //   5)  Size(capacity) in GB
        //   6)  Interface type: ATA, SATA, FC, SCSI, SAS, FATA
        //   7)  (optional) Seek Time in ms
        //   8)  (optional) Potential transfer rate in MB/sec
        //   9)  (optional) Interface transfer rate in MB/sec. I believe this is the rate to/from host.
        //       In this case, it's listed as 2GB/sec, so 2000 is the MB/sec
        //   10) (optional) Link to the vendor page where the specs are found (for future reference)
        final StringBuilder csvBuffer = new StringBuilder(255);
        csvBuffer.append(key).append(",");
        csvBuffer.append(vendor).append(",");
        csvBuffer.append(model).append(",");
        csvBuffer.append(rpm).append(",");
        csvBuffer.append(sizeGib).append(",");
        csvBuffer.append(interfaceType).append(",");

        // Do not persist default values, use an empty string in the config
        // to indicate that the values are not known.
        csvBuffer.append(null == seekTime ? "" : seekTime).append(",");
        csvBuffer.append(null == potentialTransferRate ? "" : potentialTransferRate).append(",");
        csvBuffer.append(null == interfaceSpeed ? "" : interfaceSpeed).append(",");
        csvBuffer.append(null == extraInfo ? "" : extraInfo);
        csvBuffer.append("\n");
        return csvBuffer.toString();
    }

    public String toRecordingCSV() {

        // For recording, simulation, DOES capture default values

        // Fields should be comma separated, so no commas should be used within the fields:
        //   1)  Key(use the model number)
        //   2)  Vendor
        //   3)  Model number
        //   4)  RPM of the disk, e.g. 10000, 15000, etc.
        //   5)  Size(capacity) in GB
        //   6)  Interface type: ATA, SATA, FC, SCSI, SAS, FATA
        //   7)  (optional) Seek Time in ms
        //   8)  (optional) Potential transfer rate in MB/sec
        //   9)  (optional) Interface transfer rate in MB/sec. I believe this is the rate to/from host.
        //       In this case, it's listed as 2GB/sec, so 2000 is the MB/sec
        //   10) (optional) Link to the vendor page where the specs are found (for future reference)
        final StringBuffer csvBuffer = new StringBuffer(255);
        csvBuffer.append(key).append(",");
        csvBuffer.append(vendor).append(",");
        csvBuffer.append(model).append(",");
        csvBuffer.append(rpm).append(",");
        csvBuffer.append(sizeGib).append(",");
        csvBuffer.append(interfaceType).append(",");

        // Optional parameters get defaults if undefined
        csvBuffer.append(getSeekTimeOrDefault()).append(",");
        csvBuffer.append(getPotentialTransferRateOrDefault()).append(",");
        csvBuffer.append(getInterfaceSpeedOrDefault()).append(",");
        csvBuffer.append(null == extraInfo ? "" : extraInfo);
        csvBuffer.append("\n");
        return csvBuffer.toString();
    }

    @Override
    public int hashCode() {
        int result = 19;
        result = result + model.hashCode();
        result = result + vendor.hashCode();
        result = result + rpm;
        result = result + sizeGib;
        result = result + interfaceType.hashCode();
        if (interfaceSpeed != null) {
            result = result + interfaceSpeed.hashCode();
        }
        if (seekTime != null) {
            result = result + Float.floatToIntBits(seekTime);
        }
        if (potentialTransferRate != null) {
            result = result + Float.floatToIntBits(potentialTransferRate);
        }
        if (extraInfo != null) {
            result = result + extraInfo.hashCode();
        }
        return result;
    }

    @Override
    public boolean equals(Object o) {
        if (o == this) {
            return true;
        }
        if (!(o instanceof DiskModelDescription)) {
            return false;
        }
        DiskModelDescription dmd = (DiskModelDescription) o;
        if (!model.equals(dmd.model)) {
            return false;
        }
        if (!vendor.equals(dmd.vendor)) {
            return false;
        }
        if (rpm != dmd.rpm) {
            return false;
        }
        if (sizeGib != dmd.sizeGib) {
            return false;
        }
        if (!interfaceType.equals(dmd.interfaceType)) {
            return false;
        }
        if (interfaceSpeed == null) {
            if (dmd.interfaceSpeed != null) {
                return false;
            }
        }
        else {
            if (dmd.interfaceSpeed == null) {
                return false;
            }
            if (!interfaceSpeed.equals(dmd.interfaceSpeed)) {
                return false;
            }
        }
        if (seekTime == null) {
            if (dmd.seekTime != null) {
                return false;
            }
        }
        else {
            if (dmd.seekTime == null) {
                return false;
            }
            if (!seekTime.equals(dmd.seekTime)) {
                return false;
            }
        }
        if (potentialTransferRate == null) {
            if (dmd.potentialTransferRate != null) {
                return false;
            }
        }
        else {
            if (dmd.potentialTransferRate == null) {
                return false;
            }
            if (!potentialTransferRate.equals(dmd.potentialTransferRate)) {
                return false;
            }
        }
        if (extraInfo == null) {
            if (dmd.extraInfo != null) {
                return false;
            }
        }
        else {
            if (dmd.extraInfo == null) {
                return false;
            }
            if (!extraInfo.equals(dmd.extraInfo)) {
                return false;
            }
        }
        return true;
    }

}
