package com.netapp.acquisition.aquisitionserverside.interfaces.data;

import java.io.*;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * Manages master/SANscreen defined and custom/user defined <code>DiskModelDescription</code>  containers
 * of disk model attributes
 */
public class DiskModelUtil {

    /**
     * The version number of the format of the file
     *
     * if the format of the file changes, increase the version by one.
     * going forward the new file version will be used and this keeps support in the field for existing releases
     */
    public static final int DISK_MODEL_FORMAT_CURRENT_VERSION = 1;

    /**
     * The default amount of time (in hours) to keep the current disk model cache
     *
     * Only used if the value can not be retrieved from the param table in the database
     */
    public static final long DEFAULT_DISK_MODEL_LEASE_TIME_HOURS = 24L;

    /**
     * The name of the stream containing SANscreen defined default disk model descriptions
     */
    public static final String DEFAULT_DISK_DESCRIPTIONS_NAME = "disk_models";

    /**
     * The name of the stream containing user defined custom disk model descriptions
     */
    public static final String CUSTOM_DISK_DESCRIPTIONS_NAME = "new_disk_models";

    public static final String TXT = "txt";
    public static final String DISK_DESCR_DEFAULT_STORE = DEFAULT_DISK_DESCRIPTIONS_NAME + "." + TXT;
    public static final String DISK_DESCR_CUSTOM_STORE = CUSTOM_DISK_DESCRIPTIONS_NAME + "." + TXT;

    private static final Map<Integer, String> BANNER_VERSIONS = new HashMap<Integer, String>(DISK_MODEL_FORMAT_CURRENT_VERSION);

    public static final String BANNER_VERSION_1 =
            "#--------------------------------------------------------------------------------------------------------------------------------------------------------------------\n" +
            "# Disk information                                                                                                                                                   \n" +
            "# ================                                                                                                                                                   \n" +
            "#                                                                                                                                                                    \n" +
            "# interface_type should be one of: ATA, SATA, FC, SAS, FATA, SSD, OTHER                                                                                                    \n" +
            "# All fields are required except for interface speed and additional_info. If available, please include the vendor/model website link in additional_info field.       \n" +
            "#                                                                                                                                                                    \n" +
            "# Example (regular disk):                                                                                                                                            \n" +
            "#    ST373405,Seagate,ST373405,10000,73,FC,5.3,64,160,http://www.seagate.com/staticfiles/support/disc/manuals/enterprise/cheetah/73(LP)/100109943e.pdf               \n" +
            "#                                                                                                                                                                    \n" +
            "# Example (SSD disk):                                                                                                                                                \n" +
            "#    ZIV2A420,STEC,ZIV2A420,1000000,400,SSD,,500,,http://www.stec-inc.com/product/zeusiops.php                                                                      \n" +
            "#                                                                                                                                                                    \n" +
            "# Fields should be comma separated, so no commas should be used within the fields:                                                                                   \n" +
            "#    1)  Key(use the model number)                                                                                                                                   \n" +
            "#    2)  Vendor                                                                                                                                                      \n" +
            "#    3)  Model number                                                                                                                                                \n" +
            "#    4)  RPM of the disk, e.g. 10000, 15000, etc.                                                                                                                    \n" +
            "#    5)  Size(capacity) in GB                                                                                                                                        \n" +
            "#    6)  Interface type: ATA, SATA, SATA2, SATA3, FC, SAS, FATA, SSD, OTHER                                                                                          \n" +
            "#    7)  (optional) Seek Time in ms                                                                                                                                  \n" +
            "#    8)  (optional) Potential transfer rate in MB/sec. Maximum Transfer Rate of the Disk itself.                                                                     \n" +
            "#    9)  (optional) Interface transfer rate in MB/sec. Interface Transfer Rate                                                                                       \n" +
            "#    10) (optional) Link to the vendor page where the specs are found (for future reference)                                                                         \n" +
            "#                                                                                                                                                                    \n" +
            "# Typical Interface Speeds                                                                                                                                           \n" +
            "#                                                                                                                                                                    \n" +
            "# SATA                                                                                                                                                               \n" +
            "# ----                                                                                                                                                               \n" +
            "# SATA-150                1,500 Mbit/s   187.5  MB/s                                                                                                                 \n" +
            "# SATA-300                3,000 Mbit/s   375    MB/s                                                                                                                 \n" +
            "# SATA-600                6,000 Mbit/s   750    MB/s                                                                                                                 \n" +
            "#                                                                                                                                                                    \n" +
            "# Fibre Channel                                                                                                                                                      \n" +
            "# -------------                                                                                                                                                      \n" +
            "# 1GFC (1.0625 GHz)         850 Mbit/s   106.25 MB/s                                                                                                                 \n" +
            "# 2GFC (2.125 GHz)        1,700 Mbit/s   212.5  MB/s                                                                                                                 \n" +
            "# 4GFC (4.25 GHz)         3,400 Mbit/s   425    MB/s                                                                                                                 \n" +
            "# 8GFC (8.50 GHz)         6,800 Mbit/s   850    MB/s                                                                                                                 \n" +
            "#                                                                                                                                                                    \n" +
            "# SAS (Serial Attached SCSI)                                                                                                                                         \n" +
            "# ---                                                                                                                                                                \n" +
            "# SAS                     2,400 Mbit/s   300    MB/s                                                                                                                 \n" +
            "# SAS2                    4,800 Mbit/s   600    MB/s                                                                                                                 \n" +
            "# SAS?                    6,000 Mbit/s   750    MB/s                                                                                                                 \n" +
            "#                                                                                                                                                                    \n" +
            "# Note: for SSDs, set RPM=1000000, Interface Type=SSD and Potential Transfer Rate=500                                                                                \n" +
            "#                                                                                                                                                                    \n" +
            "# Some tips on searching for specs. First determine the vendor. Usually a google search of the model number will work.                                               \n" +
            "# However, the vendor spec pages often do not list out the full model number, so searching for different parts of the                                                \n" +
            "# model number string on the vendor site is recommended. For example: Fujitsu has a page for the                                                                     \n" +
            "# MAW3xxx models where xxx varies depending on the size and interface type, so searching for MAW on fujitsu finds the page.                                          \n" +
            "#                                                                                                                                                                    \n" +
            "#--------------------------------------------------------------------------------------------------------------------------------------------------------------------\n\n";

    static {
        BANNER_VERSIONS.put(1, BANNER_VERSION_1);
    }

    /**
     * Private constructor, no initialization required.
     */
    private DiskModelUtil() {
    }

    /**
     * gets the banner for the current disk model schema
     *
     * @return A String
     */
    @SuppressWarnings("UnusedDeclaration")
    static public String getCurrentBanner() {
        return getAllBanners().get(DISK_MODEL_FORMAT_CURRENT_VERSION);
    }

    /**
     * gets all banners for all formats
     * @return a map of form version numbers and their banners
     */
    @SuppressWarnings("UnusedDeclaration")
    static public Map<Integer, String> getAllBanners() {
        return BANNER_VERSIONS;
    }

    /**
     * Parse disk model attributes from the given reader.   Disk model attributes are expected to
     * be defined lines of CSV text.   The specific format is described in the banner below.
     * Returns a Map of disk model ID to <code>DiskModelDescription</code> for the parsed disk models.
     * </p>
     * This package private API is intended to be used on the Server side, to back a simple GUI,
     * as well as within Acquisition, to record and sanscreen and custom disk model descriptions.
     * No other use is intended or tested.
     *
     * @param inputStream the stream from which to parse disk model descriptions
     * @param streamName  the name of the stream being parsed, for debug/exception reporting
     *
     * @return a map of <code>String</code> disk model IDs to the corresponding disk model descrption parsed
     *         from the given input stream.
     *
     * @throws java.io.IOException      when there is a problem reading from the reader
     * @throws IllegalArgumentException when the disk model description read from the reader is invalid
     */
    public static SortedMap<String, DiskModelDescription> parseDiskModelDescriptionStream(final InputStream inputStream,
                                                                                          final String streamName, final boolean fixIllegalTypes)
            throws IOException, IllegalArgumentException {

        final SortedMap<String, DiskModelDescription> diskModelDescriptionMap = new TreeMap<String, DiskModelDescription>();
        BufferedReader modelDescriptionReader = null;
        if (null != inputStream) {
            try {
                modelDescriptionReader = new BufferedReader(new InputStreamReader(inputStream, "utf-8"));
                parseDiskModelDescriptionStream(modelDescriptionReader, streamName, diskModelDescriptionMap, fixIllegalTypes);
            }
            finally {
                if (null != modelDescriptionReader) {
                    try {
                        modelDescriptionReader.close();
                    }
                    catch (IOException ignore) {
                    }
                }
            }
        }
        return diskModelDescriptionMap;
    }

    /**
     * Parse disk model attributes from the given reader.   Disk model attributes are expected to
     * be defined lines of CSV text.   The specific format is described in the banner below.
     * Returns a Map of disk model ID to <code>DiskModelDescription</code> for the parsed disk models.
     * </p>
     * This package private API is intended to be used on the Server side, to back a simple GUI,
     * as well as within Acquisition, to record and sanscreen and custom disk model descriptions.
     * No other use is intended or tested.
     *
     * @param content the String from which to parse disk model descriptions
     *
     * @return a map of <code>String</code> disk model IDs to the corresponding disk model descrption parsed
     *         from the given input stream.
     *
     * @throws java.io.IOException      when there is a problem reading from the reader
     * @throws IllegalArgumentException when the disk model description read from the reader is invalid
     */
    public static SortedMap<String, DiskModelDescription> parseDiskModelDescription(final String content)
            throws IOException, IllegalArgumentException {

        final SortedMap<String, DiskModelDescription> diskModelDescriptionMap = new TreeMap<String, DiskModelDescription>();

        if (null != content) {
            BufferedReader modelDescriptionReader = null;
            try {
                modelDescriptionReader = new BufferedReader(new StringReader(content));
                parseDiskModelDescriptionStream(modelDescriptionReader, "String contents", diskModelDescriptionMap);
            }
            finally {
                if (null != modelDescriptionReader) {
                    try {
                        modelDescriptionReader.close();
                    }
                    catch (IOException ignore) {
                    }
                }
            }
        }
        return diskModelDescriptionMap;
    }

    public static void parseDiskModelDescriptionStream(final BufferedReader modelDescriptionReader, final String streamName, final SortedMap<String, DiskModelDescription> diskModelDescriptionMap) throws IOException {
        parseDiskModelDescriptionStream(modelDescriptionReader, streamName, diskModelDescriptionMap, false);

    }

    public static void parseDiskModelDescriptionStream(final BufferedReader modelDescriptionReader, final String streamName, final SortedMap<String, DiskModelDescription> diskModelDescriptionMap, final boolean fixIllegalTypes) throws IOException {
        int lineNumber = 0;
        String modelDescriptionLine;
        while ((modelDescriptionLine = modelDescriptionReader.readLine()) != null) {
            lineNumber++;
            // skip comment lines (start with #) and blank lines
            if (modelDescriptionLine.startsWith("#") || modelDescriptionLine.trim().equals("")) {
                continue;
            }
            final DiskModelDescription modelDescription =
                    parseDiskModelCSVAttributes(streamName, lineNumber, modelDescriptionLine, fixIllegalTypes);
            diskModelDescriptionMap.put(modelDescription.getKey().toUpperCase(), modelDescription);
        }
    }

    /**
     * Parse a single line of CSV disk model attributes
     *
     * @param streamName              the name of the stream being parsed for debug/exception reporting
     * @param lineNumber              the line number being parsed for debug/exception reporting
     * @param csvDiskModelDescription a single line of CSV disk model attributes
     *
     * @return a <code>DiskModelDescription</code> instance containing the values parsed from the CSV attribute string
     *
     * @throws IllegalArgumentException when there given model attributes string is not valid CSV disk model attributes
     */
    private static DiskModelDescription parseDiskModelCSVAttributes(final String streamName, final int lineNumber,
                                                                    final String csvDiskModelDescription, final boolean fixIllegalTypes)
            throws IllegalArgumentException {

        final int EXPECTED_ATTRIBUTE_COUNT = 10;       // Need ten fields
        final int MANDATORY_ATTRIBUTE_COUNT = 6;       // The first 6 MUST NOT be empty
        final String[] diskParameters = csvDiskModelDescription.split(",", EXPECTED_ATTRIBUTE_COUNT);
        if (diskParameters.length < EXPECTED_ATTRIBUTE_COUNT) {
            final String errorMessage = "Failed to parse " + streamName
                                        + " line " + lineNumber + ", expected " + EXPECTED_ATTRIBUTE_COUNT
                                        + " attributes.";
            throw new IllegalArgumentException(errorMessage);
        }

        for (int i = 0; i < MANDATORY_ATTRIBUTE_COUNT; i++) {
            if (diskParameters[i].trim().length() == 0) {
                final String errorMessage = "Failed to parse " + streamName
                                            + " line " + lineNumber + ", expected " + MANDATORY_ATTRIBUTE_COUNT
                                            + " non-empty attributes.";
                throw new IllegalArgumentException(errorMessage);
            }
        }

        try {
            final String modelKey = diskParameters[0].trim();
            final String vendor = diskParameters[1].trim();
            final String model = diskParameters[2].trim();
            final int speed = Integer.parseInt(diskParameters[3].trim());
            final int size = Integer.parseInt(diskParameters[4].trim());
            //final Capacity capacity = Capacity.create(size, Capacity.Unit.GB_1000);
            String interfaceType = diskParameters[5].trim();
            // When simulating old recordings with illegal disk types, map
            // bad types to acceptable types 
            if (fixIllegalTypes) {
                if ("PATA".equalsIgnoreCase(interfaceType)) {
                    interfaceType = "ATA";
                }
                else if ("SCSI".equalsIgnoreCase(interfaceType)) {
                    interfaceType = "OTHER";
                }
                else if ("FC".equalsIgnoreCase(interfaceType)) {
                    // Some recordings have "Fc"
                    interfaceType = "FC";
                }
            }

            // We don't want to replace 'null' with DEFAULT_SEEK_TIME here.
            // Instead, the acquisition-side should make the translation.
            // Otherwise, if we populate the default seek time here,
            // it may get persisted if the Portal GUI modifies the disk model entry.

            // We don't want to indicate that we know what the actual seek time is
            // by persisting a default value
            final String seekTimeStr = diskParameters[6];
            final Float seekTime = (seekTimeStr.trim().length() > 0)
                                   ? Float.parseFloat(seekTimeStr.trim())
                                   : null;
            final String potentialTransferRateStr = diskParameters[7];
            final Float potentialTransferRate = (potentialTransferRateStr.trim().length() > 0)
                                                ? Float.parseFloat(potentialTransferRateStr.trim())
                                                : null;
            final String interfaceSpeedStr = diskParameters[8];
            final Float interfaceSpeed = (interfaceSpeedStr.trim().length() > 0)
                                         ? Float.parseFloat(interfaceSpeedStr.trim())
                                         : null;
            final String extraInfo = diskParameters[9];
            return new DiskModelDescription(modelKey,
                                            vendor,
                                            model,
                                            speed,
                                            size,
                                            interfaceType,
                                            seekTime,
                                            potentialTransferRate,
                                            interfaceSpeed,
                                            extraInfo);

        }
        catch (NumberFormatException nfe) {
            final String errorMessage = "Failed to parse " + streamName
                                        + " line " + lineNumber + ": " + nfe.getLocalizedMessage();
            throw new IllegalArgumentException(errorMessage, nfe);
        }
    }

    /**
     * Persist the given map of user (custom) disk model attributes to the well known text file.
     * Used in acquisition to persist the disk model into the recording
     *
     * @param diskDescriptionWriter the writer to use to persist the disk model descriptions
     * @param userDiskModels        the Map of disk model descriptions to persist
     * @param isRecording           whether this write is for a recording, in which case default values
     *                              will be persisted, otherwise we are writing for configuration, in which case unknown values
     *                              are left blank.
     *
     * @throws IOException when there is a problem persisting the disk model descriptions
     */
    @SuppressWarnings("UnusedDeclaration")
    static public void writeUserDiskModelAttributes(final BufferedWriter diskDescriptionWriter, final String banner,
                                                    final Map<String, DiskModelDescription> userDiskModels, final boolean isRecording)
            throws IOException {

        try {
            if (banner != null) {
                diskDescriptionWriter.write(banner);
            }
            for (final com.netapp.acquisition.aquisitionserverside.interfaces.data.DiskModelDescription modelDescr : userDiskModels.values()) {
                if (isRecording) {
                    diskDescriptionWriter.write(modelDescr.toRecordingCSV());
                }
                else {
                    diskDescriptionWriter.write(modelDescr.toConfigCSV());
                }
            }
        }
        finally {
            if (null != diskDescriptionWriter) {
                try {
                    diskDescriptionWriter.close();
                }
                catch (IOException ignore) {
                }
            }
        }
    }
}
