/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.acquisition.aquisitionserverside.interfaces.impl;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.annotation.Resource;
import javax.annotation.security.PermitAll;
import javax.annotation.security.RolesAllowed;
import javax.ejb.SessionContext;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.AroundInvoke;
import javax.interceptor.InvocationContext;
import javax.jms.ConnectionFactory;
import javax.jms.Topic;
import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.PersistenceContext;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUnit;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUnitAudit;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUpdateType;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSource;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceAttribute;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceAuditDetail;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceDoesNotExistException;
import com.netapp.acquisition.aquisitionserverside.interfaces.local.AcquisitionServerSessionLocal;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession;

/**
 * Created by weis.
 * Date: Jun 8, 2010
 * Time: 4:09:05 PM
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@SuppressWarnings("unused")
@RolesAllowed({"acquisition", "admin"})
public class AcquisitionServerSessionBean implements AcquisitionServerSession, AcquisitionServerSessionLocal {
    private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger("com.netapp.acquisition.aquisitionserverside.interfaces.impl.AcquisitionServerSessionBean");

    private static final Object locker = new Object();

    private static final AtomicBoolean deployed = new AtomicBoolean(false);

    private static volatile DataSourceAuditManager auditManager = null;

    @PersistenceContext(unitName = "acq-server")
    EntityManager entityManager;

    @Resource
    private SessionContext currentContext;

    @Resource(mappedName = "java:/ConnectionFactory")
    private ConnectionFactory connectionFactory;

    @Resource(mappedName = "java:/topic/AcquisitionUpdateTopic")
    private Topic updateTopic;

    private static final String OCIE_PASSWORD = "netappfoundation.password";

    //this method cannot be called in PostConstruct because transaction is not available
    //todo in the future separate data source audit into another ejb

    /**
     * Add an interceptor around each method call to make sure that deploy was called and audit manager was initialized
     *
     * @param ctx Context
     *
     * @return the return value of the method
     *
     * @throws Exception                On any error
     * @throws IllegalArgumentException When deploy was not called yet. The IllegalArgumentException type is important to
     *                                  enable retries by the client (see SessionCache)
     */
    @AroundInvoke
    public Object init(final InvocationContext ctx) throws Exception {

        final String methodName = ctx.getMethod().getName();

        // Make sure deploy was called (except for the 'deploy' method or method to check if deployed)
        if (!"deploy".equals(methodName) && !"isAcquisitionDeployed".equals(methodName)) {
            if (!deployed.get()) {
                // The IllegalArgumentException type is important to enable retries by the client (see SessionCache)
                throw new IllegalArgumentException("Deploy must be called prior to any service call");
            }
        }

        if (auditManager == null) {
            synchronized (locker) {
                //noinspection ConstantConditions
                if (auditManager == null) {
                    DataSourceAuditManager auditManagerTemp = new DataSourceAuditManager();
                    auditManagerTemp.initialize(entityManager);
                    auditManager = auditManagerTemp;
                }
            }
        }

        try {
            return ctx.proceed();
        }
        catch (Exception e) {
            logger.error("Caught error in " + ctx.getMethod().getName() + ": " + e.getMessage(), e);

            throw e;
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AcquisitionUnit register(AcquisitionUnit acquisitionUnit) throws AcquisitionServerException {
        try {

            acquisitionUnit.setNextLeaseRenewal(System.currentTimeMillis() + acquisitionUnit.getLeasePeriod());
            acquisitionUnit.setStatus(AcquisitionUnit.Status.CONNECTED);

            acquisitionUnit = entityManager.merge(acquisitionUnit);

            AcquisitionUnitNotifier.getAuNotifier().registerAu(acquisitionUnit);
            publishUpdate(AcquisitionUpdateType.ACQUISITION_UNIT_REGISTER, System.currentTimeMillis(), acquisitionUnit.getId(), null, null);
            return acquisitionUnit;
        }
        catch (Exception e) {
            throw new AcquisitionServerException("Error in register", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @RolesAllowed({"guest", "admin", "acquisition"})
    public void removeAcquisitionUnitByName(final String acquisitionUnitName) throws AcquisitionServerException {
        try {
            final AcquisitionUnit au = (AcquisitionUnit) entityManager.createNamedQuery("findAcqUnitWithAttributesByName").
                    setParameter("name", acquisitionUnitName).getSingleResult();

            if (au.getStatus() != AcquisitionUnit.Status.NOT_CONNECTED && !au.isLeaseExpired()) {
                throw new AcquisitionServerException("Could not delete active Acquisition Unit: " + acquisitionUnitName);
            }

            @SuppressWarnings({"unchecked"})
            final List<DataSource> dataSources = entityManager.createNamedQuery("findDataSourcesByAuId").setParameter("auId", au.getId()).getResultList();
            for (final DataSource dataSource : dataSources) {
                if (dataSource.isActive()  && dataSource.getManual() != 2) { //'2' indicates a manual stop from the UI
                    throw new AcquisitionServerException("Can not remove Acquisition Unit " + acquisitionUnitName +
                                                         " since it has at least one active DataSource: " + dataSource.getName());
                }
                notifyDsRemoval(au.getId(), dataSource.getName());
            }
            entityManager.remove(entityManager.getReference(AcquisitionUnit.class, au.getId()));   //also gets DS and DS audits, etc. due to FKs
        }
        catch (NoResultException nre) {
            //no-op, nothing there.
        }

    }

    /**
     * {@inheritDoc}
     */

    @Override
    public AcquisitionUnit getAcquisitionUnit(final String acquisitionUnitName) throws AcquisitionServerException {
        try {
            return (AcquisitionUnit) entityManager.createNamedQuery("findAcqUnitWithAttributesByName").
                    setParameter("name", acquisitionUnitName).getSingleResult();
        }
        catch (NoResultException nre) {
            //contract states to return null when not present.   Happens on first call of new AU for instance.
            return null;
        }
        catch (Exception e) {
            throw new AcquisitionServerException("Error in getAcquisitionUnit", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void unregister(final long acquisitionUnitId) throws AcquisitionServerException {
        try {
            final AcquisitionUnit acqUnit = entityManager.find(AcquisitionUnit.class, acquisitionUnitId);
            acqUnit.setStatus(AcquisitionUnit.Status.NOT_CONNECTED);
            //set all ds active = 0 where au id = pass in value

            @SuppressWarnings({"unchecked"})
            final List<DataSource> dataSources = entityManager.createNamedQuery("findDataSourcesByAuId").setParameter("auId", acqUnit.getId()).getResultList();
            for (DataSource dataSource : dataSources)
                dataSource.setActive(false);

            publishUpdate(AcquisitionUpdateType.ACQUISITION_UNIT_UNREGISTER, System.currentTimeMillis(), acqUnit.getId(), null, null);
            AcquisitionUnitNotifier.getAuNotifier().unregisterAu(acqUnit);
        }
        catch (Exception e) {
            throw new AcquisitionServerException("Error in getAcquisitionUnit", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void renewLease(final long acquisitionUnitId) {
        final AcquisitionUnit acqUnit = entityManager.find(AcquisitionUnit.class, acquisitionUnitId);
        acqUnit.setStatus(AcquisitionUnit.Status.CONNECTED);
        final long now = System.currentTimeMillis();
        acqUnit.setNextLeaseRenewal(acqUnit.getLeasePeriod() + now);
        publishUpdate(AcquisitionUpdateType.ACQUISITION_UNIT_LEASE_RENEW, now, acqUnit.getId(), null, null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void reportAcquisitionUnitAudit(final long acquisitionUnitId, String status) {
        final AcquisitionUnitAudit auAudit = new AcquisitionUnitAudit(System.currentTimeMillis(), acquisitionUnitId, status);
        entityManager.persist(auAudit);
        publishUpdate(AcquisitionUpdateType.ACQUISITION_UNIT_UPDATE, System.currentTimeMillis(), acquisitionUnitId, null, null);
        // TODO: Need to figure out what is best to do here (should we update / insert...)
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @PermitAll
    public void deploy(final boolean force) throws AcquisitionServerException {
        synchronized (locker) {
            if (!force && deployed.get()) {
                return;
            }

            deployed.set(false);

            DataSourceTypeDeployHelper.getInstance().deploy(entityManager);

            deployed.set(true);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @PermitAll
    public boolean isAcquisitionDeployed() {
        return deployed.get();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<DataSource> getDataSourcesToRun(final long acquisitionUnitId) throws AcquisitionServerException {

        // OCIE needs the server to decrypt the netappfoundation.password value before returning values to DataSourceManager.
        // We must make a clone of each DS, with a decrypted password attribute, to return to DataSourceManager.
        // We must not decrypt the password in the original ResultList because we don't want hibernate to persist
        // or otherwise expose the decrypted password.
        final List<DataSource> rawResultsList = entityManager.createNamedQuery("findDataSourcesToRunWithAttributesByAuId").setParameter("auId", acquisitionUnitId).getResultList();
        final List<DataSource> returnDataSourceList = new ArrayList<DataSource>(rawResultsList.size());
        for (final DataSource origDS : rawResultsList) {
            final DataSource decryptedDS = new DataSource(origDS.getDsTypeId(), origDS.getAuId(), origDS.getName(), origDS.getVendor(), origDS.getModel(),
                                                          origDS.getStatus(), origDS.getStatusExt(), origDS.getTime(), origDS.getManual(), origDS.isActive());
            decryptedDS.setId(origDS.getId());
            // final LIst
            final Set<DataSourceAttribute> origAttrs = origDS.getAttributes();
            final Set<DataSourceAttribute> decryptedAttrs = new HashSet<DataSourceAttribute>(origAttrs.size());
            for (final DataSourceAttribute dsa : origAttrs) {
                if (OCIE_PASSWORD.equals(dsa.getName())) {
                    final String encryptedPassword = dsa.getValue();
                    String decryptedPassword;

                    // Null is a flag to indicate upgrade case
                    if (encryptedPassword != null && !encryptedPassword.isEmpty()) {
                        try {
                            decryptedPassword = AcquisitionServerUtil.getCipherFacade().decrypt(encryptedPassword);
                        } catch (com.netapp.crypto.CryptoException ce) {
                            // Expected on upgrade case where previous password was encrypted by the old encryption method.
                            throw new AcquisitionServerException("Failed to decrypt password: " + ce.getLocalizedMessage(), ce);
                        }
                    } else {
                        decryptedPassword = encryptedPassword;
                    }
                    final DataSourceAttribute decryptedPasswordAttribute = new DataSourceAttribute(OCIE_PASSWORD, decryptedPassword);
                    decryptedPasswordAttribute.setDataSource(decryptedDS);
                    decryptedAttrs.add(decryptedPasswordAttribute);

                } else {
                    decryptedAttrs.add(dsa);
                }
            }
            decryptedDS.setAttributes(decryptedAttrs);
            returnDataSourceList.add(decryptedDS);
        }
        return returnDataSourceList;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void updateIsActive(final long dataSourceId, final boolean isActive) throws AcquisitionServerException {
        DataSource ds = entityManager.find(DataSource.class, dataSourceId);
        if (ds == null) {
            return; //no necc. an error - nothing to update.  Treat as a no-op.
        }
        ds.setActive(isActive);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_UPDATE, System.currentTimeMillis(), ds.getAuId(), ds.getId(), ds.getName());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public long createDsAudit(final long dataSourceId, final String packageName, final String status) {
        final long id = auditManager.createDsAudit(dataSourceId, packageName, status, entityManager);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_AUDIT_CREATE, System.currentTimeMillis(), null, null, null);
        return id;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void closeDsAudit(final long auditSessionId, final AuditType type, final String status, final List<DataSourceAuditDetail> messages, final long crc) {
        auditManager.closeDsAudit(auditSessionId, type, status, messages, crc, entityManager);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_AUDIT_CLOSE, System.currentTimeMillis(), null, null, null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void updateDsAudit(final long auditSessionId, final AuditType type, final String status) {
        auditManager.updateAudit(auditSessionId, type, status, entityManager);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_AUDIT_UPDATE, System.currentTimeMillis(), null, null, null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
    public void clearDsAudit(final long dataSourceId) {
        auditManager.clearDsAudit(dataSourceId, entityManager);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_AUDIT_UPDATE, System.currentTimeMillis(), null, null, null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
    public void clearDsAuditForAu(final String auName) {
        auditManager.clearDsAuditForAu(auName,entityManager);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_AUDIT_UPDATE, System.currentTimeMillis(), null, null, null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void reportStandbyForPackage(final long dataSourceId, final String packageName) {
        auditManager.reportStandbyForPackage(dataSourceId, packageName, entityManager);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_AUDIT_UPDATE, System.currentTimeMillis(), null, null, null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void stopReportingAuditForPackage(final long dataSourceId, final String packageName) {
        auditManager.stopReportingAuditForPacakge(dataSourceId, packageName, entityManager);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_AUDIT_UPDATE, System.currentTimeMillis(), null, null, null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
    public void forceCrawl(final long dataSourceId, final String packageName) throws AcquisitionServerException {
        final DataSource dataSource = getDataSource(dataSourceId);
        notifyAu(dataSource.getAuId(), Command.crawl, dataSource.getName(), packageName);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_CRAWL, System.currentTimeMillis(), dataSource.getAuId(), dataSource.getId(), dataSource.getName());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void forceAcquisition(final long dataSourceId, final String packageName, final Properties properties) throws AcquisitionServerException {
        final DataSource dataSource = getDataSource(dataSourceId);
        notifyAu(dataSource.getAuId(), Command.force, dataSource.getName(), packageName, properties);
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_FORCE, System.currentTimeMillis(), dataSource.getAuId(), dataSource.getId(), dataSource.getName(), properties);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
    public void notifyDsConfigurationChange(final long dataSourceId) throws AcquisitionServerException {
        final DataSource dataSource = getDataSource(dataSourceId);
        notifyAu(dataSource.getAuId(), Command.update, dataSource.getName(), "");
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_UPDATE, System.currentTimeMillis(), dataSource.getAuId(), dataSource.getId(), dataSource.getName());
    }

    @Override
    @RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
    public void notifyDsRemoval(final long acquisitionUnitId, final String dsName) throws AcquisitionServerException {
        //don't lookup DataSource here, it has already been deleted.  Simply tell AU to update and resync.
        notifyAu(acquisitionUnitId, Command.remove, dsName, "");
        publishUpdate(AcquisitionUpdateType.DATA_SOURCE_REMOVE, System.currentTimeMillis(), acquisitionUnitId, null, dsName);
    }

    @Override
    @RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
    public void notifyLicensingUpdate() throws AcquisitionServerException {
        @SuppressWarnings({"unchecked"})
        final List<AcquisitionUnit> aus = entityManager.createNamedQuery("findAllAcqUnits").getResultList();
        for (final AcquisitionUnit au : aus) {
            notifyAu(au.getId(), Command.licensingChange, null, null);
        }
    }

    private DataSource getDataSource(final long dataSourceId)  throws DataSourceDoesNotExistException {
        final DataSource dataSource = entityManager.find(DataSource.class, dataSourceId);
        if (dataSource == null) {
            throw new DataSourceDoesNotExistException(dataSourceId);
        } else {
            return dataSource;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Properties listen(final long acquisitionUnitId) throws AcquisitionServerException {

        // Wait for a message with a timeout...
        final AcquisitionUnitNotifier.Message message = AcquisitionUnitNotifier.getAuNotifier().consumeNotification(acquisitionUnitId, 2);

        // Timed out - tell the caller that there is nothing for him. Let him call us again.
        if (message == null) {
            return null;
        }

        return message.asProperties();

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public long getServerTime() {
        return System.currentTimeMillis();
    }

  private void notifyAu(long acquisitionUnitId, final Command command, final String dsName, final String thePackage) throws AcquisitionServerException {
    notifyAu( acquisitionUnitId, command, dsName, thePackage, null );
  }

  /**
   *
   * @param acquisitionUnitId
   * @param command
   * @param dsName
   * @param thePackage
   * @param properties
   * @throws AcquisitionServerException
   */
    private void notifyAu(long acquisitionUnitId, final Command command, final String dsName, final String thePackage, final Properties properties) throws AcquisitionServerException {
        final AcquisitionUnitNotifier.Message message =
                new AcquisitionUnitNotifier.Message(acquisitionUnitId, command)
                        .dataSourceName(dsName)
                        .packageName(thePackage);

        message.setProperties( properties );
        AcquisitionUnitNotifier.getAuNotifier().postNotification(acquisitionUnitId, message);
    }


    /**
     * Publish an update to the JMS acquisition update topic.   Listeners may take actions as appropriate (e.g. update UI).
     *
     * @param updateType the type of update as defined by enum AcquisitionUpdateType.
     * @param updateTime the time of the update.
     * @param auId       The acquisition unit involved.  Optional - may be null.
     * @param dsId       The DataSource Id involved.   Optional - may be null.
     * @param dsName     The name of the DataSource involved.   Optional - may be null.
     */
    private void publishUpdate(final AcquisitionUpdateType updateType,
                               final long updateTime,
                               final Long auId,
                               final Long dsId,
                               final String dsName) {
      publishUpdate( updateType, updateTime, auId, dsId, dsName, null );
    }

  private void publishUpdate(final AcquisitionUpdateType updateType,
                             final long updateTime,
                             final Long auId,
                             final Long dsId,
                             final String dsName,
                             final Properties properties) {
    AcquisitionJms.INSTANCE.send(connectionFactory, updateTopic, updateType, updateTime, auId, dsId, dsName, properties);
  }


    @Override
    public String encryptPassword(final String clearTextPassword) throws AcquisitionServerException {
        return AcquisitionServerUtil.getCipherFacade().encrypt(clearTextPassword);
    }

    @Override
    public String decryptPassword(final String encryptedPassword) throws AcquisitionServerException  {
        if (encryptedPassword == null || encryptedPassword.isEmpty()) {
            return encryptedPassword;
        }
        return AcquisitionServerUtil.getCipherFacade().decrypt(encryptedPassword);
    }

}
