package com.netapp.acquisition.aquisitionserverside.interfaces.impl;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUnit;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession.Command;

import java.util.Map;
import java.util.Properties;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import org.slf4j.LoggerFactory;

/**
 * Self-contained encapsulation of notification pattern to Acquisition Unit.  Offers method both to publish
 * a notification and methods for AU to consume notifications.
 *
 * Singleton model.
 */
final class AcquisitionUnitNotifier {

    private static org.slf4j.Logger logger = LoggerFactory.getLogger(AcquisitionUnitNotifier.class);

    /** Singleton for the notifier */
    private static AcquisitionUnitNotifier aun = new AcquisitionUnitNotifier();

    /**
     * Holds a map of blocking queues, one per registered acquisition unit.  Key is the ID of each AU.
     */
    private final Map<Long, BlockingQueue<Message>> auQueues;

    /**
     * Private constructor - singleton.
     */
    private AcquisitionUnitNotifier() {
        auQueues = new ConcurrentHashMap<Long, BlockingQueue<Message>>();
    }

    /**
     * Get the acquisition unit notifier singleton instance.
     *
     * @return   singleton instance of the notifier.
     */
    public static AcquisitionUnitNotifier getAuNotifier() {
        return aun;
    }

    /**
     * Register a new acquisition unit with the notification system.  Assumes an ID has been populated for
     * newly created/registered units.      If an existing queue exists for this AU, it will be deleted first.
     *
     * @param au the Acquisition Unit to register.
     * @throws IllegalStateException if an acquisition unit with this ID is already registered.
     */
    public synchronized void registerAu (final AcquisitionUnit au) {
        if (auQueues.containsKey(au.getId())) {
            unregisterAu(au);
        }
        final LinkedBlockingQueue<Message> queue = new LinkedBlockingQueue<Message>();
        auQueues.put(au.getId(), queue);
    }


    /**
     * Un-register the given acquisition unit.
     *
     */
    public synchronized void unregisterAu (final AcquisitionUnit au) {
        auQueues.remove(au.getId());
    }

    /**
     * Post a notification to be consumed by the given Acquisition unit.  This is an asynchronous call.   Messages
     * are queued for consumption at a later time by the AU.   Method is thread safe.
     *
     * @param auId  the ID of the acquisition unit that will consume the message.
     * @param message  the message to publish.
     * @return  true if message was en-queued.   False otherwise.
     * @throws IllegalStateException if the given acquisition unit is not registered OR if the ID given in the
     *  message does not match that of the message.
     */
    public boolean postNotification (final long auId, final Message message) {
        final BlockingQueue<Message> queue = getQueueForAu (auId);
        if (queue == null) {
            return false;
        }

        if (message.getAuId() != auId) {
            throw new IllegalStateException ("Trying to post a notification for AU: " + message.getAuId() + " to queue for AU: " + auId);
        }

        return queue.offer(message);
    }

    /**
     * Blocking (for given timeout period) call-back for Acquisition Units to consume published notifications.
     * Returns next message in queue if available.
     *
     * <p>If a message is available on calling this method, it returns immediately with next message.   Otherwise,
     * method blocks for given timeout period until a message is published to this AU.   If nothing is published
     * before the timeout, method will return NULL.
     *
     * <p> Method is thread safe.
     *
     * @param auId the acquisition unit for which messages are to be consumed.
     * @param timeoutSeconds  blocking timeout to wait for a message.
     * @return the message, or NULL if no message available.
     * @throws IllegalStateException if the given AU ID is not registered.
     */
    public Message consumeNotification(final long auId, final int timeoutSeconds) {
        final BlockingQueue<Message> queue = getQueueForAu (auId);
        if (queue == null) {
             throw new IllegalStateException ("Trying to receive a notification on an unregistered Acquisition Unit: " + auId);
        }

        try {
            return queue.poll(timeoutSeconds, TimeUnit.SECONDS);
        } catch (final InterruptedException ie) {
            //treat as empty reply.
            if (logger.isDebugEnabled()) {
                logger.debug("Notify method for AU: " + auId + " was interrupted.");
            }
            return null;
        }
    }

    /**
     * Get the blocking queue for the given AU ID.
     *
     * @param id the ID of the AU.
     * @return  the appropriate blocking queue.
     */
    private BlockingQueue<Message> getQueueForAu (final long id) {
        return auQueues.get(id);
    }

    /**
     * General message to use when notifying acquisition units.
     */
    public static final class Message {

        private final long auId;
        private final Command command;
        private String dsName = null;
        private String packageName = null;
        private Properties properties = null;

        /**
         * Construct a new notification message.
         *
         * @param auId the ID of the acquisition unit.
         * @param command the AcquisitionServerSession.Command of the message.
         */
        public Message (final long auId, final Command command) {
            this.auId = auId;
            this.command = command;
        }

        /**
         * Builder pattern - set the optional data source name for this message.
         *
         * @param dsName the datasource name.
         * @return  the current instance of the Message for builder chaining.
         */
        public Message dataSourceName (final String dsName) {
            this.dsName = dsName;
            return this;
        }

        /**
         * Builder pattern - set the optional package name for this message.
         *
         * @param packageName the package name.
         * @return  the current instance of the Message for builder chaining.
         */
        public Message packageName (final String packageName) {
            this.packageName = packageName;
            return this;
        }

      public Properties getProperties()
      {
        return properties;
      }

      public void setProperties(Properties properties)
      {
        this.properties = properties;
      }

      public String getDsName() {
            return dsName;
        }

        public String getPackageName() {
            return packageName;
        }

        public long getAuId() {
            return auId;
        }

        public Command getCommand() {
            return command;
        }

        /**
         * Return the message data as a property object, keyed for consumption by Acquisition Unit in separate JVM.
         *
         * @return  message data as a property object.
         */
        public Properties asProperties() {
            final Properties p = new Properties();
            p.setProperty("auId", Long.valueOf(auId).toString());
            p.setProperty("command", command.name());
            if (dsName != null) {
                p.setProperty("ds", dsName);
            }
            if (packageName != null) {
                p.setProperty("package", packageName);
            }

            if( properties != null )
              p.putAll( properties );
            return p;
        }

    }


}
