package com.netapp.acquisition.aquisitionserverside.interfaces.impl;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUnit;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSource;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceAudit;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceAuditDetail;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession;
import org.slf4j.Logger;

import javax.persistence.EntityManager;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by weis.
 * Date: Jun 11, 2010
 * Time: 3:23:35 PM
 */
public class DataSourceAuditManager {
    private static transient final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger("com.netapp.acquisition.aquisitionserverside.interfaces.impl.DataSourceAuditManager");

    private static final int ONE_DAY = 24 * 60 * 60 * 1000;

    private static ConcurrentHashMap<Long, ConcurrentHashMap<String, PackageAuditHistory>> auditHistoryCache = new ConcurrentHashMap<Long, ConcurrentHashMap<String, PackageAuditHistory>>();
    private static ConcurrentHashMap<Long, DataSourceAudit> openAuditSessionsCache = new ConcurrentHashMap<Long, DataSourceAudit>();


    public DataSourceAuditManager() {}

    public synchronized void initialize(final EntityManager entityManager) {
        auditHistoryCache.clear();
        //get all datasource audit
        final List<DataSourceAudit> dsAudits = entityManager.createQuery("select a from DataSourceAudit a order by a.dsId, a.startTime asc").getResultList();
        for (final DataSourceAudit audit : dsAudits) {
            final PackageAuditHistory packageAuditHistory = findPackageAuditHistory(audit);
            // Add this to the audit cache. Don't update the auditSessionCache because this audit entry
            // does not reflect an open session
            packageAuditHistory.push(audit);
        }
    }

    /**
     * Create a new audit session
     *
     * <p/>Note: We assume there will be only 1 thread updating a package audit entry at any single time
     *
     * @see com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession#createDsAudit(long, String, String)
     */
    @SuppressWarnings({"JavaDoc"})
    public synchronized long createDsAudit(final long dsID, final String packageName,
                                           final String status, final EntityManager entityManager) {
        if (logger.isDebugEnabled())
            logger.debug("Creating audit for datasource's originId: [" + dsID + "], with status: [" + status + "]");

        final DataSourceAudit audit = new DataSourceAudit(packageName,
                                                          status,
                                                          DataSourceAudit.Type.START,
                                                          dsID,
                                                          new Date().getTime(),
                                                          DataSourceAudit.END_DATE,
                                                          1);

        entityManager.persist(audit);

        if (logger.isDebugEnabled())
            logger.debug("data source audit record was created successfully, returning auditId: [" + audit.getId() + "]");

        updateAuditCache(audit);

        // Now update the aggregated status of the datasource
        final DataSource dataSource = entityManager.find(DataSource.class, dsID);
        if (dataSource == null)
            throw new IllegalStateException("Could not find data source " + dsID);

        final String newDataSourceStatus = figureOutDataSourceStatus(dataSource);
        dataSource.setStatus(newDataSourceStatus);

        return audit.getId();
    }

    /**
     * This method closes the audit session.
     *
     * @see com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession#closeDsAudit
     */
    @SuppressWarnings({"JavaDoc"})
    public synchronized void closeDsAudit(final long auditSessionId, final AcquisitionServerSession.AuditType type,
                                          final String status, final List<DataSourceAuditDetail> messages,
                                          final long crc, final EntityManager entityManager) {
        DataSourceAudit audit = getAuditFromCache(auditSessionId, entityManager);
        if (audit == null)
            return;

        if (!audit.isOpen()) {
            throw new IllegalStateException("Trying to close an already closed audit: " + auditSessionId);
        }

        // Get the history of this session from the history cache.
        final PackageAuditHistory packageAuditHistory = findPackageAuditHistory(audit);
        if (packageAuditHistory == null)
            throw new IllegalStateException("Could not find audit in cache");

        // Now we have several options.
        // 1. This is a different message than the previous one. In this case we want to store everything (including messages)
        // 2. This is the same message as the previous one. In this case we want to throw away this one and just update the previous
        // 3. There is no history to examine - and in this case we want to save everything

        if (!packageAuditHistory.hasHistory() || isDifferentMessage(packageAuditHistory, type, status)) {
            // Simply add this audit and its messages
            audit.setType(DataSourceAudit.Type.valueOf(type.toString()));
            audit.setStatus(status);
            audit.setEndTime(new Date().getTime());

            // Only here we actually want to update the messages
            if (messages != null) {
                for (DataSourceAuditDetail message : messages)
                    audit.addAcquisitionAuditMessage(message);
            }

            // Update DB (no need to update caches since we were working on the object at the top of stack)
            entityManager.merge(audit);

            // Now update the aggregated status of the datasource.  Null check to deal with timing window on
            // DataSource deletion.
            final DataSource dataSource = entityManager.find(DataSource.class, audit.getDsId());
            if (dataSource != null) {
                dataSource.setStatus(figureOutDataSourceStatus(dataSource));
            }
        }
        else {
            assert (packageAuditHistory.hasHistory());

            // Get rid of the open line - this is the last update we got from acq
            packageAuditHistory.pop();

            // Now update the aggregated status of the datasource
            final DataSource dataSource = entityManager.find(DataSource.class, audit.getDsId());
            if (dataSource != null) {
                dataSource.setStatus(figureOutDataSourceStatus(dataSource));
            }

            // Now get audit from previous poll - we already know that the result is the same as current poll
            DataSourceAudit previousAuditRecord = packageAuditHistory.pop();

            // Delete the current record (we will modify the previous one with updated count and end time)
            entityManager.remove(entityManager.getReference(DataSourceAudit.class, audit.getId()));

            // The latest record is now the previous one with an incremented 'number of times' (we just deleted the session)
            previousAuditRecord = entityManager.find(DataSourceAudit.class, previousAuditRecord.getId());

            //BURT 953149 - previousAuditRecord is null. We saw this issue once, unable to root cause.
            if(previousAuditRecord == null) {
                return;
            }
            previousAuditRecord.setNumberOfTimes(previousAuditRecord.getNumberOfTimes() + 1);
            previousAuditRecord.setEndTime(new Date().getTime());
            audit = previousAuditRecord;
        }

        // Update caches
        packageAuditHistory.push(audit);
        openAuditSessionsCache.remove(auditSessionId);
    }

    /**
     * Clear audit log for this datasource.   Also sets up internal caches correctly.
     *
     * @param dataSourceId  the ID of the datasource audit to clear.
     */
    public synchronized void clearDsAudit(final long dataSourceId, EntityManager entityManager) {

        if (logger.isDebugEnabled())
            logger.debug("Removing audit for datasource " + dataSourceId);
        
        entityManager.createQuery("DELETE FROM DataSourceAudit a WHERE a.dsId=:dsId").setParameter("dsId", dataSourceId).executeUpdate();
        auditHistoryCache.remove(dataSourceId);
        final Iterator<DataSourceAudit> it = openAuditSessionsCache.values().iterator();
        while (it.hasNext()) {
            final DataSourceAudit audit = it.next();
            if (audit.getDsId() == dataSourceId)
                it.remove();
        }

    }

    /**
     * Clear audit log for this acquistion unit.   Also sets up internal caches correctly.
     *
     * @param auName  the name of the acquisition unit to clear.
     */
    public synchronized void clearDsAuditForAu(final String auName, final EntityManager entityManager) {
        if (logger.isDebugEnabled()) {
            logger.debug("Removing audit for acq unit '" + auName + '\'');
        }
        final AcquisitionUnit au = (AcquisitionUnit) entityManager.createNamedQuery("findAcqUnitByName").
                setParameter("name", auName).getSingleResult();
        if (au == null) {
            return;
        }

        final long auId = au.getId();

        @SuppressWarnings({"unchecked"})
        final List<DataSource> dataSources = entityManager.createNamedQuery("findDataSourcesByAuId")
                .setParameter("auId", auId).getResultList();

        for (final DataSource ds : dataSources) {
            clearDsAudit(ds.getId(), entityManager);
        }

    }

    private boolean isDifferentMessage(final PackageAuditHistory packageAuditHistory, final AcquisitionServerSession.AuditType type,
                                       final String status) {
        if (packageAuditHistory.peekHistory() == null
            || !packageAuditHistory.peekHistory().getStatus().equals(status)
            || !packageAuditHistory.peekHistory().getType().toString().equals(type.toString())) {
            // TODO: Check CRC to make sure we don't recognize messages with different extended errors as identical!!!
            return true;
        }

        return false;
    }

    private DataSourceAudit getAuditFromCache(Long auditSessionID, final EntityManager entityManager) {
        // Get the audit record from the cache...
        DataSourceAudit audit = openAuditSessionsCache.get(auditSessionID);
        if (audit == null) {
            if (logger.isDebugEnabled())
                logger.debug("closeDsAudit() - Failed to find audit id " + auditSessionID);

            // Try from the database...
            audit = entityManager.find(DataSourceAudit.class, auditSessionID);
            if (audit == null) {
                logger.warn("closeDsAudit() - Failed to find audit id " + auditSessionID);
                return null;
            }

            openAuditSessionsCache.put(auditSessionID, audit);
        }

        if (logger.isDebugEnabled())
            logger.debug("Audit record was found using audit's ID: [" + auditSessionID + "], belongs to data source's OringId: [" + audit.getDsId() + "], current status: [" + audit.getStatus() + "]");
        return audit;
    }

    public synchronized void updateAudit(final Long auditSessionID, final AcquisitionServerSession.AuditType type,
                                         final String newStatus, final EntityManager entityManager) {
        DataSourceAudit audit = getAuditFromCache(auditSessionID, entityManager);
        if (audit == null) return;

        if (!audit.isOpen()) {
            throw new IllegalStateException("Trying to update a closed audit: " + auditSessionID);
        }

        audit.setType(DataSourceAudit.Type.valueOf(type.toString()));
        audit.setStatus(newStatus);

        if (logger.isDebugEnabled())
            logger.debug("Audit record was found using audit's ID: [" + auditSessionID + "], belongs to data source's OringId: [" + audit.getDsId() + "], current status: [" + audit.getStatus() + "]");

        // Now update the aggregated status of the datasource
        final DataSource dataSource = entityManager.find(DataSource.class, audit.getDsId());
        dataSource.setStatus(figureOutDataSourceStatus(dataSource));

        if (logger.isDebugEnabled())
            logger.debug("Audit record was updated using audit's ID: [" + audit.getId() + "], belongs to data source's OringId: [" + audit.getDsId() + "], current status: [" + audit.getStatus() + "]");
    }

    public synchronized void reportStandbyForPackage(final long dataSourceId, final String packageName,
                                                     final EntityManager entityManager) {
        // First check if this package is not marked as disabled already...
        final PackageAuditHistory packageAuditHistory = findPackageAuditHistory(dataSourceId, packageName);

        final DataSourceAudit lastAudit = packageAuditHistory.peek();
        if (lastAudit != null && lastAudit.getStatus() != null && lastAudit.getStatus().equals("Stand-by"))
            return;

        // This package is not marked as disabled. Let's do it now!
        final Long tempSessionID = createDsAudit(dataSourceId, packageName, "Stand-by", entityManager);
        closeDsAudit(tempSessionID, AcquisitionServerSession.AuditType.STANDBY, "Stand-by", null, -1, entityManager);
    }

    public synchronized void stopReportingAuditForPacakge(final long dataSourceId, final String packageName,
                                                          final EntityManager entityManager) {
        // First check if this package is not marked as disabled already...
        final PackageAuditHistory packageAuditHistory = findPackageAuditHistory(dataSourceId, packageName);

        final DataSourceAudit lastAudit = packageAuditHistory.peek();

        // If there is no history for this package, don't add any audit. This might indicate that the history is cleaned
        // or the package never ran (not licensed).
        if (lastAudit == null)
            return;

        // Avoid repetitions of the "Disabled" report...
        if (lastAudit.getStatus() != null && lastAudit.getStatus().equals("Disabled"))
            return;

        // This package is not marked as disabled. Let's do it now!
        final Long tempSessionID = createDsAudit(dataSourceId, packageName, "Disabled", entityManager);
        closeDsAudit(tempSessionID, AcquisitionServerSession.AuditType.DISABLED, "Disabled", null, -1, entityManager);
    }

    private void updateAuditCache(final DataSourceAudit audit) {
        final PackageAuditHistory packageAudits = findPackageAuditHistory(audit);

        packageAudits.push(audit);
        openAuditSessionsCache.put(audit.getId(), audit);
    }

    private PackageAuditHistory findPackageAuditHistory(final DataSourceAudit audit) {
        return findPackageAuditHistory(audit.getDsId(), audit.getPackageName());
    }

    /**
     * Given a specific audit, find the right audit stack in the cache for this audit.
     */
    public PackageAuditHistory findPackageAuditHistory(final long dataSourceId, final String packageName) {
        // Keep this audit in the 'lastAuditEntries' cache
        ConcurrentHashMap<String, PackageAuditHistory> auditHistoryForPackage = auditHistoryCache.get(dataSourceId);
        if (auditHistoryForPackage == null) {
            auditHistoryForPackage = new ConcurrentHashMap<String, PackageAuditHistory>();
            ConcurrentHashMap<String, PackageAuditHistory> temp = auditHistoryCache.putIfAbsent(dataSourceId, auditHistoryForPackage);
            if (temp != null)
                auditHistoryForPackage = temp;
        }

        PackageAuditHistory packageAudits = auditHistoryForPackage.get(packageName);
        if (packageAudits == null) {
            packageAudits = new PackageAuditHistory();
            PackageAuditHistory temp = auditHistoryForPackage.putIfAbsent(packageName, packageAudits);
            if (temp != null)
                packageAudits = temp;
        }

        return packageAudits;
    }

    private String figureOutDataSourceStatus(final DataSource dataSourceData) {
        if (dataSourceData.getManual() == 2)
            return "Stopped";

        // Build a map of package statuses. The map is a sorted map to make sure that the packages are always reported
        // in the same order.
        final SortedMap<String, String> packageStatuses = new TreeMap<String, String>();

        final long now = new Date().getTime();

        // Lookup the history of this data source...
        final Map<String, PackageAuditHistory> auditsForDsPerPackage = auditHistoryCache.get(dataSourceData.getId());
        if (auditsForDsPerPackage == null)
            return "";

        // For each pacakge, get the last audit reported for. This is how we determine the status of the package
        for (final Map.Entry<String, PackageAuditHistory> entryForPacakge : auditsForDsPerPackage.entrySet()) {
            final PackageAuditHistory auditStackForPackage = entryForPacakge.getValue();

            final DataSourceAudit lastAudit = auditStackForPackage.peek();

            // If no history is available, don't report anything about this package
            if (lastAudit == null)
                continue;

            // Disabled packages are not considered when reporting status
            if (isDisabled(lastAudit))
                continue;

            if (isStandby(lastAudit)) {
                packageStatuses.put(lastAudit.getPackageName(), "stand-by");
                continue;
            }

            if (isRunning(lastAudit)) {
                packageStatuses.put(lastAudit.getPackageName(), "acquiring");
                continue;
            }

            if (!isStatusOK(lastAudit)) {
                packageStatuses.put(lastAudit.getPackageName(), "failed");
                continue;
            }

            // Check for package timeout (did not succeed for 24hours)
            final String auditPackageName = lastAudit.getPackageName();
            String packageName = "";
            // TODO: Need to follow the use of the feature name as the package name
            // For now we must string match on the feature name;
            if (("Inventory").equals(auditPackageName)) {
                packageName = "foundation";
            }
            else
            if (("Inventory: VM").equals(auditPackageName)) {
                packageName = "foundation";
            }
            else
            if (("Perform: FC Switch").equals(auditPackageName)) {
                packageName = "performance";
            }
            else
            if (("Perform: Storage").equals(auditPackageName)) {
                packageName = "storageperformance";
            }
            else
            if (("Perform: VM").equals(auditPackageName)) {
                packageName = "hostvirtualization";
            }

            final String pollingIntervalStr = dataSourceData.getAttributesAsProperties().getProperty(packageName + ".poll");
            // Polling value is expressed in seconds in the the server. Convert to ms.
            final Integer pollingIntervalMs = (null != pollingIntervalStr) ? Integer.valueOf(pollingIntervalStr) * 1000 : 0;

            if (!didRunRecently(auditStackForPackage, now, pollingIntervalMs)) {
                packageStatuses.put(lastAudit.getPackageName(), "not running");
                continue;
            }
            
            packageStatuses.put(lastAudit.getPackageName(), "success");
        }

        final StringBuilder sb = new StringBuilder(100);

        for (final Map.Entry<String, String> packageStatus : packageStatuses.entrySet())
            sb.append(packageStatus.getKey()).append(" ").append(packageStatus.getValue()).append(", ");

        if (sb.length() > 0) {
            sb.deleteCharAt(sb.length() - 1);
            sb.deleteCharAt(sb.length() - 1);
        }

        return sb.toString();
    }

    private boolean isDisabled(final DataSourceAudit audit) {
        if (audit == null || audit.getStatus() == null)
            return false;

        return audit.getStatus().equals("Disabled");
    }

    private boolean isRunning(final DataSourceAudit audit) {
        if (audit == null || audit.getStatus() == null)
            return false;

        return audit.getStatus().startsWith("Acquiring") || audit.getStatus().startsWith("Sent");
    }

    private boolean isStandby(final DataSourceAudit audit) {
        if (audit == null || audit.getStatus() == null)
            return false;

        return audit.getStatus().equals("Stand-by");
    }

    private boolean isStatusOK(final DataSourceAudit audit) {
        if (audit == null || audit.getStatus() == null)
            return false;

        return audit.getStatus().equals("Success");
    }

    /**
     * Figures out if this package is stuck.
     * <p/>
     * HEADSUP: Since we don't have an easy way to retrieve the package's poll interval from the server side, we will default to 1 day.
     * HEADSUP2: Max of   1.5 * polling interval or 1 day
     * @param auditStackForPackage
     * @param now
     *
     * @return <code>true</code> if package is running properly, <code>false</code> otherwise
     */
    private boolean didRunRecently(final PackageAuditHistory auditStackForPackage, final long now, final Integer pollingIntervalMs) {
        final long lastPackageSuccessTime = auditStackForPackage.getLastPackageSuccessTime();
        if (lastPackageSuccessTime == 0)
            return false;

        return now - lastPackageSuccessTime < Math.max(ONE_DAY, 1.5*pollingIntervalMs);
    }

    /**
     * Class representing the audit history for a package.
     * This data will be used for:
     * 1. Consolidating audit recordings (increasing count for same status)
     * 2. Identifying packages that are stuck (by recording last success time)
     */
    final static class PackageAuditHistory {
        private DataSourceAudit lastAudit = null;
        private DataSourceAudit previousAudit = null;

        PackageAuditHistory() {
        }

        void push(final DataSourceAudit audit) {
            previousAudit = lastAudit;
            lastAudit = audit;
        }

        public boolean hasHistory() {
            return lastAudit != null && previousAudit != null && lastAudit.getStatus() != null && previousAudit.getStatus() != null;
        }

        public DataSourceAudit pop() {
            final DataSourceAudit audit = lastAudit;
            lastAudit = previousAudit;
            previousAudit = null;

            return audit;
        }

        public DataSourceAudit peek() {
            return lastAudit;
        }

        public DataSourceAudit peekHistory() {
            return previousAudit;
        }

        public long getLastPackageSuccessTime() {
            long lastPackageSuccessTime = 0;
            if (lastAudit.getStatus() != null && lastAudit.getStatus().equalsIgnoreCase("Success"))
                lastPackageSuccessTime = lastAudit.getEndTime();

            return lastPackageSuccessTime;
        }
    }
}
