package com.netapp.acquisition.aquisitionserverside.interfaces.impl;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceType;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceTypeAttribute;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceTypeVendorModel;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.slf4j.LoggerFactory;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceException;
import java.io.*;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * Created by weis.
 * Date: Jun 28, 2010
 * Time: 4:26:32 PM
 */
public class DataSourceTypeDeployHelper {
    private static transient final org.slf4j.Logger logger =
            LoggerFactory
            .getLogger("com.netapp.acquisition.aquisitionserverside.interfaces.impl.DataSourceTypeDeployHelper");

    final private String deploymentsDir = System.getProperty("jboss.server.base.dir", "C:/Program Files/SANscreen/jboss/server/onaro/") + "/deployments/" ;
    final private File datasourcesFolder = new File(deploymentsDir, "datasources.war");
    final private static String UTF_8 = "utf-8";

    private final transient static DataSourceTypeDeployHelper instance = new DataSourceTypeDeployHelper();

    public static DataSourceTypeDeployHelper getInstance() {
        return instance;
    }

    private DataSourceTypeDeployHelper() {
    }

    public List<String> deploy(final EntityManager entityManager) throws AcquisitionServerException{
        try {
            // Remove old deployment (clean all ds types).  This cascade deletes down through attrs and vendormodel tables.
            entityManager.createNativeQuery("delete from ds_type").executeUpdate();

            // Parse the deployfile.dat to understand what DS need to be deployed
            final List<String> xmlFiles = parseDeployFile();

            // Now deploy them
            for (final String xmlFile : xmlFiles) {
                final DataSourceType dst = buildDataSourceType(xmlFile);
                entityManager.persist(dst);
                logger.debug("DataSourceTypeDeployHelper deploy() - built and stored type :" + dst.getName());
            }

            return xmlFiles;
        }
        catch (PersistenceException e) {
            logger.error("Failed to deploy: " + e.getMessage(), e);
            throw e;
        }
    }

    /**
     * @return list of deploy files
     */
    private List<String> parseDeployFile() throws AcquisitionServerException {
        try {
            /*
            The deploy file states what DataSoures to load.   There are currently two of these, one for
            "NER" (OnSight Reporting) and one for standard SANscreen.  We determine which one to load based on
            licensing.  TODO:  Revisit this for a more systematic approach in the future.
             */
            final String activeDeployFile = "deployfile.dat";
            final File deployFile = new File(datasourcesFolder, activeDeployFile);
            final List<String> result = new LinkedList<String>();

            // Now read and parse the file
            final BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(deployFile), UTF_8));
            String line;
            while ((line = br.readLine()) != null) {
                // remove remarks
                final int index = line.indexOf("#");
                if (index != -1)
                    line = line.substring(0, index);

                line = line.trim();

                // Skip empty lines
                if (line.length() == 0)
                    continue;

                // The line is basically a name of a file that is located in the data sources deploy directory (the same
                // directory as the deployfile.dat file we're reading now
                result.add(line);
            }

            return result;
        }
        catch (IOException e) {
            throw new AcquisitionServerException("deploy() - failed to find deployfile.dat", e);
        }
    }

    private DataSourceType buildDataSourceType(final String xmlFileName) throws AcquisitionServerException {
        final File xmlFile = new File(datasourcesFolder, xmlFileName);

        final Element root = parseXmlFile(xmlFile.getAbsolutePath());
        final DataSourceType dst = parseHeader(root);
        parseVendorModels(xmlFileName, root, dst);
        parsePackages(xmlFileName, root, dst);

        return dst;
    }

    private void parsePackages(final String xmlFileName, final Element root, final DataSourceType dst) throws AcquisitionServerException {
        // parse packages
        final List<Element> packageDescriptors = root.getChildren("package");
        if (packageDescriptors.size() == 0)
            throw new AcquisitionServerException("No packages defined in deploy descriptor " + xmlFileName);

        int guiOrder = 0;
        // Iterating over packages
        for (final Element packageRootElement : packageDescriptors) {
            final String packageName = packageRootElement.getAttributeValue("name");

            // DEPENDENCIES
            // if there are package dependencies, we need to create a property for them
            // they are stored in the database so we can send them to the client
            final String dependencyValue = packageRootElement.getAttributeValue("dependency");
            if ((dependencyValue != null) && !dependencyValue.trim().equals("")) {
                final Map<String, String> dependencyAttributeMap = new TreeMap<String, String>();
                dependencyAttributeMap.put("name", packageName + "." + "REQUIRES");
                dependencyAttributeMap.put("value", dependencyValue);
                dependencyAttributeMap.put("role", "hidden");
                dependencyAttributeMap.put("guiorder", Integer.toString(guiOrder++));

                final DataSourceTypeAttribute dependencyAttr = new DataSourceTypeAttribute(dst.getId(), dependencyAttributeMap);
                dst.addAttribute(dependencyAttr.getName(), dependencyAttr);
            }

            // Get the list of properties for this package
            final Element properties = packageRootElement.getChild("properties");

            // Iterating over all properties of the package
            for (final Element property : (List<Element>) properties.getChildren("property")) {
                // This map will hold all the attributes of the property
                final Map<String, String> attributesMap = new TreeMap<String, String>();
                for (final org.jdom.Attribute attr : (List<org.jdom.Attribute>) property.getAttributes()) {
                    final String key = attr.getName();
                    final String value = attr.getValue();
                    attributesMap.put(key, value);
                }

                attributesMap.put("guiorder", Integer.toString(guiOrder++));

                // We process attributes differently based on their type.
                // If type is a "list" we encode child XML nodes as members of the list
                // otherwise, attribute goes directly to the property
                String propertyType = attributesMap.get("type");

                // TODO: fix xml files !!!
                if (propertyType == null)
                    propertyType = "string";

                if (propertyType.equals("list")) {
                    processListProperty(dst, packageName, property, attributesMap);
                }
                else {
                    processRegularProperty(dst, packageName, attributesMap);
                }
            }
        }
    }

    private void parseVendorModels(final String xmlFileName, final Element root, final DataSourceType dst) throws AcquisitionServerException {
        final Element vendorModels = root.getChild("vendorsModels");
        if (vendorModels == null)
            throw new AcquisitionServerException("Missing vendor-model section in file " + xmlFileName);

        // Iterate over all the vendors
        for (final Element vendorElement : (List<Element>) vendorModels.getChildren("vendor")) {
            final String vendorName = vendorElement.getAttributeValue("name").toUpperCase();
            final String vendorDescription = vendorElement.getAttributeValue("description");

            // Iterate over all the models defined for this vendor
            for (final Element modelElement : (List<Element>) vendorElement.getChildren("model")) {
                final String modelName = modelElement.getAttributeValue("name").toUpperCase();
                final String modelDescription = modelElement.getAttributeValue("description");

                dst.addVendor(new DataSourceTypeVendorModel(dst.getId(), vendorName, vendorDescription, modelName, modelDescription));
            }
        }
    }

    private void processRegularProperty(final DataSourceType dst, final String packageName, final Map<String, String> attributesMap) {
        if (attributesMap.get("name").equals("poll") && (packageName.equals("foundation") || packageName.equals("nas"))) {
            // reset the type and value: convert from minutes as a float to seconds as an integer
            final Float pollInMinutes = Float.parseFloat(attributesMap.get("value"));
            final Integer pollInSeconds = (int) (pollInMinutes * 60);
            final String pollStringInSeconds = pollInSeconds.toString();
            attributesMap.put("value", pollStringInSeconds);
            attributesMap.put("type", "integer");
        }

        // non-list attributes are added as properties as usual
        attributesMap.put("name", packageName + "." + attributesMap.get("name"));
        final DataSourceTypeAttribute attr = new DataSourceTypeAttribute(dst.getId(), attributesMap);
        dst.addAttribute(attr.getName(), attr);
    }

    private void processListProperty(final DataSourceType dst, final String packageName, final Element property, final Map<String, String> attributesMap) {
        // process the list type
        final String attrNameForAList = attributesMap.get("name");
        // encode the list name as a property
        attributesMap.put("name", packageName + "." + "LIST" + "." + attrNameForAList);
        final DataSourceTypeAttribute attr = new DataSourceTypeAttribute(dst.getId(), attributesMap);
        dst.addAttribute(attr.getName(), attr);

        // guiorder is restarted for each list member
        int guiOrder = 0;
        // process each column as a new attribute
        for (final Element columnProperty : (List<Element>) property.getChild("columns").getChildren("property")) {
            final Map<String, String> columnMap = new TreeMap<String, String>();

            for (final org.jdom.Attribute col : (List<org.jdom.Attribute>) columnProperty.getAttributes()) {
                final String key = col.getName();
                final String value = col.getValue();
                columnMap.put(key, value);
            }
            // keep the column order for this list to store as "guiorder"
            columnMap.put("guiorder", Integer.toString(guiOrder++));

            // encode the list column as a property
            columnMap.put("name", packageName + "." + "LIST" + "." + attrNameForAList + ".0." + columnMap.get("name"));
            final DataSourceTypeAttribute colAttr = new DataSourceTypeAttribute(dst.getId(), columnMap);
            dst.addAttribute(colAttr.getName(), colAttr);
        }
    }

    private DataSourceType parseHeader(final Element root) throws AcquisitionServerException {
        final Element datasourceType = root.getChild("dataSourceType");
        if (datasourceType == null)
            throw new AcquisitionServerException("Missing 'dataSourceType' element in deployee xml ");

        final long id = Long.parseLong(datasourceType.getTextNormalize());

        final Element name = root.getChild("name");
        if (name == null)
            throw new AcquisitionServerException("Missing 'name' element in deployee xml ");

        final String dstName = name.getTextNormalize();

        final Element description = root.getChild("description");
        if (description == null)
            throw new AcquisitionServerException("Missing 'description' element in deployee xml ");

        final String dstDescription = description.getTextNormalize();

        return new DataSourceType(id, dstName, dstDescription);
    }

    private Element parseXmlFile(final String xmlFileName) throws AcquisitionServerException {
        FileInputStream inputStream = null;
        try {
            inputStream = new FileInputStream(xmlFileName);

            // Parse the deployee XML
            final SAXBuilder builder = new SAXBuilder(com.sun.org.apache.xerces.internal.parsers.SAXParser.class.getName());
            final Document doc = builder.build(inputStream);

            return doc.getRootElement();
        }
        catch (JDOMException e) {
            throw new AcquisitionServerException("Failed to parse Deployee XML file at " + xmlFileName, e);
        }
        catch (IOException e) {
            throw new AcquisitionServerException("Failed to read Deployee XML file at " + xmlFileName, e);
        }
        finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                }
                catch (IOException e) {
                    // nothing to do
                }
            }
        }
    }
}
