package com.netapp.acquisition.aquisitionserverside.interfaces.remote;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUnit;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSource;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSourceAuditDetail;

import javax.ejb.Remote;
import java.util.List;
import java.util.Properties;

/**
 * Created by weis.
 * Date: Jun 8, 2010
 * Time: 10:13:08 AM
 */

@SuppressWarnings({"UnusedDeclaration"})
@Remote
public interface AcquisitionServerSession {
    String JNDI_NAME = "ejb:mega-app/acqServer//AcquisitionServerSessionBean!com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession";

    /**
     * Deploy data sources from the deploy directory.
     * <p/>
     * <p/>This parses the deploy descriptor files (XMLs) and populates the data source type (etc..) tables in the DB.
     * <p/>
     * <p/><u>This method MUST be called prior to any other service call!.</u>
     *
     * @param force          Force redeploy (even if previously deployed).
     */
    void deploy(boolean force) throws AcquisitionServerException;

    /**
     * Has acquisition as a "service" been fully deployed?   If false, no AU has checked in with the server since
     * server start.
     *
     * @return   true if acquisition session is deployed fully - false if no AU has checked in.
     */
    boolean isAcquisitionDeployed();

    /**
     * Retrieve acquisition unit with its attribute from database.        If the AU does not exist with this
     * name, returns NULL.
     *
     *
     * @param acquisitionUnitName Acquisition unit name
     * @return acquisition unit, never null
     * @throws AcquisitionServerException for various connection and lookup errors.
     */
    AcquisitionUnit getAcquisitionUnit(final String acquisitionUnitName) throws AcquisitionServerException;

    /**
     * Register the AU into the server.   Returns the merged AcquisitionUnit with any updated values from DB.
     *
     * @param acquisitionUnit Acquisition Unit data (synced with the DB) - that was previously obtained by calling {@link #getAcquisitionUnit(String)}
     * @return the Acquisition unit with merged database values.
     * @throws AcquisitionServerException for various update errors.
     */
    AcquisitionUnit register(final AcquisitionUnit acquisitionUnit) throws AcquisitionServerException;

    /**
     * Remove an acquisition unit permanently given it's name.  This will also delete all associated Datasources.
     *
     * <p>Fails if acquisition unit is active (active lease) or if any Datasource on it is not in a stopped state.
     *
     * @param acquisitionUnitName   the name of the acquisition unit to delete.
     * @throws AcquisitionServerException if delete fails or cannot be executed.
     */
    void removeAcquisitionUnitByName (final String acquisitionUnitName) throws AcquisitionServerException;

    /**
     * This method will be called by the acquisition in a normal shutdown of the acquisition.
     *
     * @param acquisitionUnitId Acquisition unit ID
     */
    void unregister(final long acquisitionUnitId) throws AcquisitionServerException;

    /**
     * This method will be called by the acquisition units to renew their lease. The lease is the mechanism that allows
     * the server to detect if there is something wrong with the acquisition unit (and perhaps to alert about it).
     *
     * @param acquisitionUnitId Acquisition unit ID
     */
    void renewLease(final long acquisitionUnitId) throws AcquisitionServerException;

    /**
     * Reports / updates the status of the acquisition unit
     *
     * @param acquisitionUnitId Acquisition unit ID
     * @param status            New status for the acquisition unit
     */
    void reportAcquisitionUnitAudit(final long acquisitionUnitId, final String status) throws AcquisitionServerException;

    /**
     * This method is called from the acquisition when synchronizing its state with the server (get the list
     * of data sources to start, stop and update). This returns the data sources that given acquisition unit is required
     * to run.
     *
     * @param acquisitionUnitId AU to get data sources for
     * @return Only active data sources that need to be running for the given AU
     */
    List<DataSource> getDataSourcesToRun(long acquisitionUnitId) throws AcquisitionServerException;

    /**
     * Update the status (active / not active) of a data source. This allows the server to track the status of the
     * deployment of the data source.
     *
     * @param dataSourceId Data source ID
     * @param isActive     true if the data source is active, false otherwise.
     */
    void updateIsActive(final long dataSourceId, final boolean isActive) throws AcquisitionServerException;

    /**
     * Creates a new audit session to track the progress of a single poll of a data source package.
     *
     * @param dataSourceId Data source ID
     * @param packageName  Package polling
     * @param status       The initial status of this package
     * @return The unique ID created for this audit session
     */
    long createDsAudit(final long dataSourceId, final String packageName, final String status) throws AcquisitionServerException;

    /**
     * This method closes the audit session.
     * <p/>
     * <p/>The current record is an opened one. Based on the last error message, we have to look at the record
     * before that and decide if the current record should be merged into the last one. If so, we delete the current
     * record and increment the count in the last record. Otherwise, we should close the current record with the new
     * status. This way when a data source is continuously successful, only one audit record will be written with an
     * incrementing count.
     *
     * @param auditSessionId audit session ID
     * @param type Audit type
     * @param status         Final status to set for the audit session
     * @param messages       Extended messages associated with this session (<code>null</code> if non existent)
     * @param crc            The CRC calculated from the exception that generated this message (a hint to optimize DB saving) or -1 is no messages
     */
    void closeDsAudit(final long auditSessionId, final AuditType type, final String status, final List<DataSourceAuditDetail> messages, final long crc) throws AcquisitionServerException;

    /**
     * Updates the status of an open audit session
     *
     * @param auditSessionId audit session ID
     * @param type Audit type
     * @param status         New status
     */
    void updateDsAudit(final long auditSessionId, final AuditType type, final String status) throws AcquisitionServerException;

    /**
     * Clears the audit log history for a particular DataSource.
     *
     * @param dataSourceId the ID of the datasource.
     */
    void clearDsAudit (final long dataSourceId);

    /**
     * Clear the audit log history for all DataSources on this acquisition unit.              \
     *
     * @param auName the acquisition unit name.
     */
    void clearDsAuditForAu (final String auName);

    void reportStandbyForPackage(final long dataSourceId, final String packageName) throws AcquisitionServerException;

    void stopReportingAuditForPackage(final long dataSourceId, final String packageName) throws AcquisitionServerException;

    /**
     * Send a notification to the acquisition unit that something in the configuration has changed. This will trigger the
     * acq unit to synchronize its state with the server and figure out what changed.
     *
     * @param datasourceId the ID of the DataSource being deleted.
     * @throws AcquisitionServerException in the event of lookup or other system errors.
     */
    void notifyDsConfigurationChange(final long datasourceId) throws AcquisitionServerException;

    /**
     * Notify the given acquisition unit of the deletion of the given DS name.   This will trigger the
     * acq unit to synchronize its state with the server and figure out what changed.
     *
     * @param acquisitionUnitId  the ID of the acquisition unit handling the poll of the former DataSource.
     * @param dsName  the name of the former DataSource.
     * @throws AcquisitionServerException in the event of lookup or other system errors.
     */
    void notifyDsRemoval(final long acquisitionUnitId, final String dsName) throws AcquisitionServerException;

    /**
     * Notify all AUs and DataSources of an update to the server licensing structure.  Will trigger restarts
     * and re-deployments as needed for AUs to re-sync licensing information.
     *
     * @throws AcquisitionServerException in the event of lookup or other system errors.
     */
    void notifyLicensingUpdate () throws AcquisitionServerException;

    /**
     * Send a notification to the acquisition unit to 'crawl' a specific data source (and a package)
     *
     * @param dataSourceId Data source ID
     * @param packageName  Package to activate
     */
    void forceCrawl(final long dataSourceId, final String packageName) throws AcquisitionServerException;

    /**
     * Send a notification to the acquisition unit to 'force' a specific data source (and a package)
     *
     * @param dataSourceId Data source ID
     * @param packageName  Package to activate ('' for all packages)
     * @param properties  Properties needed for polling execution options
     */
    void forceAcquisition(final long dataSourceId, final String packageName, final Properties properties ) throws AcquisitionServerException;

    /**
     * Listens to events for this acquisition unit.
     *
     * @param acquisitionUnitId Acquisition unit ID
     * @return The event details or <code>null</code> if nothing was received
     */
    Properties listen(final long acquisitionUnitId) throws AcquisitionServerException;

    /**
     * Used to allow the acquisition side to synchronize clocks with the server. This is especially important when
     * collecting performance data from multiple acquisition units that might have clock synchronization issues.
     *
     * @return Server time (System.currentTimeMillis() performed on the server side).
     */
    long getServerTime() throws AcquisitionServerException;

    /**
     * Types of commands that can be sent from the server via the notification mechanism between the server and acquisition.
     *
     * @see #listen
     */
    public static enum Command {
        /**
         * Some DS was updated. Synchronize with the server.
         */
        update,
        /**
         * Trigger a specific data source (or package) to 'crawl'.
         * <p/>
         * <p/>Additional parameters:<br>
         * 1. ds - Data source name <br>
         * 2. package - Package name. Use '' (empty string) to 'force' all packages<br>
         */
        force,
        /**
         * Trigger a specific data source package to 'crawl'.
         * <p/>
         * <p/>Additional parameters:<br>
         * 1. ds - Data source name <br>
         * 2. package - Package name<br>
         */
        crawl,
        /**
         * DataSource was removed.
         * Parameters:
         * 1.  Datasource name that was removed.
         * 2.  Package will be empty.
         */
        remove,
        /**
         * Licensing was changed.
         * <p>Additional parameters are empty.
         */
        licensingChange
    }

    /**
     * Types of Audit states.
     *
     */
    public static enum AuditType {
        START,
        ACQUIRING,
        STANDBY,
        SUCCESS,
        ERROR,
        WARNING,
        DISABLED
    }

}
