package com.netapp.ext_algorithms.pbe;

/**
 * PBE.java
 *
 * This example uses SHA-1 and Twofish to encrypt or decrypt sometext
 * using a password. The iteration count is set to 1000 and the
 * salt is randomly chose non encryption. The salt is then written to the
 * first 8 bytes of the output text.
 *
 * When decrypting, this program will read the first 8 bytes of the
 * cipher text and use that as the salt.
 *
 * It requires a JCE-compliant PBE With SHA And Twofish-CBC engine, like the
 * Bouncy Castle Provider (http://www.bouncycastle.org).
 *
 * NETAPP: This code comes from a WROX Press book. 
 */

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.GeneralSecurityException;
import java.util.Random;

import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.PBEParameterSpec;

import sun.misc.BASE64Decoder;
import sun.misc.BASE64Encoder;


public class PBE {
    private static int ITERATIONS = 1000;


    public static void main(String[] args)
            throws Exception {

        if( args.length != 3 )
            usage();

        String result = ""; //$NON-NLS-1$
        String password = args[1];
        String text = args[2];
        if( args[0].equals("-e")  ) { //$NON-NLS-1$
            result=encrypt(password.toCharArray(),text);
        } else if ( args[0].equals("-d") ) { //$NON-NLS-1$
            result=decrypt(password.toCharArray(),text);
        } else {
            usage();
        }

        System.out.println(result);
    }

    private static void usage() {
        System.err.println("Usage: java com.onaro.util.crypt.PBE -e | -d keyPassword textToEncryptDecrypt"); //$NON-NLS-1$
        System.exit(1);
    }

    /* NETAPP: This code comes from a WROX Press book, NetApp has a license. */
    public static String encrypt(char[] password, String plaintext) throws GeneralSecurityException {
        try {
            // Begin by creating a random salt of 64 bits(8 bytes)
            byte[] salt = new byte[8];
            Random random = new Random();
            random.nextBytes(salt);
            //Create the PBE Key Spec with the given password
            PBEKeySpec keySpec = new PBEKeySpec(password);
            // Get a SecretKeyFactory for PBE With MD5 And DES
            SecretKeyFactory keyFactory =
                    SecretKeyFactory.getInstance("PBEWithMD5AndDES"); //$NON-NLS-1$
            //Create our key
            SecretKey key = keyFactory.generateSecret(keySpec);
            //Now create a parameter spec for our salt and iterations
            PBEParameterSpec paramSpec
                    = new PBEParameterSpec(salt, ITERATIONS);
            //Create a cipher and initialize it for encrypting
            Cipher cipher = Cipher.getInstance("PBEWithMD5AndDES"); //$NON-NLS-1$
            cipher.init(Cipher.ENCRYPT_MODE, key, paramSpec);
            byte[] ciphertext = cipher.doFinal(plaintext.getBytes("utf-8")); //$NON-NLS-1$
            BASE64Encoder encoder = new BASE64Encoder();
            String saltString = encoder.encode(salt);
            String ciphertextString = encoder.encode(ciphertext);
            return saltString + ciphertextString;
        } catch (UnsupportedEncodingException e) {
            throw new GeneralSecurityException("Failed to get bytes from string",e); //$NON-NLS-1$
        }
    }

    /* NETAPP: This code comes from a WROX Press book, NetApp has a license. */
    public static String decrypt(char[] password, String text) throws GeneralSecurityException {
        try {
            //Begin by splitting the text into salt and text Strings
            //salt is first 12 chars, BASE64 encoded from 8bytes.
            String salt = text.substring(0, 12);
            String ciphertext = text.substring(12, text.length());

            //BASE64Decode the bytes for the salt and the cipher text
            BASE64Decoder decoder = new BASE64Decoder();
            byte[] saltArray = decoder.decodeBuffer(salt);
            byte[] ciphertextArray = decoder.decodeBuffer(ciphertext);

            //Create the PBEKeySpec with the given password
            PBEKeySpec keySpec = new PBEKeySpec(password);

            //Get a SecretKeyFactory for PBE With MD5 And DES
            SecretKeyFactory keyFactory
                    = SecretKeyFactory.getInstance("PBEWithMD5AndDES"); //$NON-NLS-1$
            //Create our key
            SecretKey key = keyFactory.generateSecret(keySpec);

            //Now create a parameter spec for our salt and iterations
            PBEParameterSpec paramSpec = new PBEParameterSpec(saltArray, ITERATIONS);

            //Create a cipher and initialize it for encrypting
            Cipher cipher = Cipher.getInstance("PBEWithMD5AndDES"); //$NON-NLS-1$
            cipher.init(Cipher.DECRYPT_MODE, key, paramSpec);

            //Perform the actual decryption
            byte[] plaintextArray = cipher.doFinal(ciphertextArray);

            return new String(plaintextArray,"utf-8"); //$NON-NLS-1$
        } catch (IOException e) {
            throw new GeneralSecurityException("Can't decipher",e); //$NON-NLS-1$
        }
    }

    // These use the default password to encrypt/decrypt the string. The default password is decrypted before use.
  //  public static String encryptWithDefaultKey(String plaintext) throws GeneralSecurityException {
        //return encrypt(DEFAULT_PASSWORD, plaintext);
    //}
//
//    public static String decryptWithDefaultKey(String cipherText) throws GeneralSecurityException {
//        return decrypt(DEFAULT_PASSWORD, cipherText);
//    }
}
