package com.netapp.oci.netapp.client.interfaces.impl;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;
import com.netapp.acquisition.aquisitionserverside.interfaces.local.AcquisitionServerSessionLocal;
import com.netapp.oci.netapp.client.interfaces.data.ZAPIManagerAggregateException;
import com.netapp.oci.netapp.client.interfaces.data.ZAPIManagerException;
import com.netapp.oci.platform.common.table.*;
import com.netapp.oci.netapp.db.NetAppModel;
import netapp.manage.*;

import javax.sql.DataSource;
import java.io.IOException;
import java.net.UnknownHostException;
import java.sql.Connection;
import java.util.*;


public class ZAPIManagerImpl {

    public static final ZAPIManagerImpl INSTANCE = new ZAPIManagerImpl();

    private ZAPIManagerImpl() {}


    /**
     * Assign the default QOS policy group to the volumes identified by the given object IDs.
     * All target volumes must exist in teh same cluster.
     * @param clusterObjId the object ID of the cluster hosting the volumes to be configured
     * @param targetVolumeObjIDs the set of Volume object IDs to assign to the default QOS policy group
     * @param qosPolicyGroupId the objid of the policy group to assign (probably objid of "_Performance_Monitor_volumes")
     * @throws com.netapp.oci.netapp.client.interfaces.data.ZAPIManagerException if configuring ANY volume fails
     */
    public void assignDefaultPolicyToVolumes(final DataSource dataSource, final AcquisitionServerSessionLocal acquisitionServerSessionLocal,
                                             final long clusterObjId, final Set<Long> targetVolumeObjIDs, final long qosPolicyGroupId) throws ZAPIManagerException {
        assignDefaultPolicyToVolumes(dataSource, acquisitionServerSessionLocal, clusterObjId, targetVolumeObjIDs, qosPolicyGroupId, null, null);
    }


    /**
     * Assign the default QOS policy group to the volumes identified by the given object IDs.
     * All target volumes must exist in teh same cluster.
     * @param clusterObjId the object ID of the cluster hosting the volumes to be configured
     * @param targetVolumeObjIds the set of Volume object IDs to assign to the default QOS policy group
     * @param qosPolicyGroupId the objid of the policy group to assign (probably objid of "_Performance_Monitor_volumes")
     * @param userName ZAPI username used to assign policy group
     * @param password ZAPI password used to assign policy group
     * @throws com.netapp.oci.netapp.client.interfaces.data.ZAPIManagerException if configuring ANY volume fails
     */
    public void assignDefaultPolicyToVolumes(final DataSource dataSource, final AcquisitionServerSessionLocal acquisitionServerSessionLocal,
                                             final long clusterObjId, final Set<Long> targetVolumeObjIds, final long qosPolicyGroupId,
                                             final String userName, final String password) throws ZAPIManagerException {

        if (targetVolumeObjIds == null || targetVolumeObjIds.size() == 0) {
            return;
        }

        // Just in case any target volume fails
        final ZAPIManagerAggregateException volumeManageFailedException =
                new ZAPIManagerAggregateException("Failed to configure 1 or more target volumes.");

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            final Properties dataSourceProperties = getDataSourceProperties(connection, clusterObjId);
            final String targetIP = dataSourceProperties.getProperty("netappfoundation.ip");
            final String userFromProps = dataSourceProperties.getProperty("netappfoundation.user");
            final String passwordFromProps = dataSourceProperties.getProperty("netappfoundation.password");
            final String decrypedPasswordFromProps = decryptPassword(acquisitionServerSessionLocal, passwordFromProps);
            final String targetPortString = dataSourceProperties.getProperty("netappfoundation.port", "80");
            final int targetPort = parsePortFromString(targetPortString);
            final String protocol = dataSourceProperties.getProperty("netappfoundation.connectionType", "HTTP");
            // Clients may override OCIE's configured user and password for this cluster for the case
            // where OCIE uses read-only use, and configuring volume policy group requires write access
            final String finalUserName = (userName != null) ? userName : userFromProps;
            final String finalPassword = (password != null) ? password : decrypedPasswordFromProps;

            final Map<Long, String[]> volumeNamesByObjID = getVolumeObjIDToNameMap(connection, targetVolumeObjIds);
            assertAllVolumesAreKnown(volumeManageFailedException, volumeNamesByObjID, targetVolumeObjIds);

            // Now configure each target volume
            NaServer zapiConnection = null;
            try {
                zapiConnection = getZAPIConnection(targetIP, finalUserName, finalPassword, targetPort, protocol);

                // First assert that the volume still has no qos policy group assigned.
                // Don't do this, it consumes too much cluster CPU (DE4860)
//                final Map<Long, String> filteredTargetVolumeUUIDsByID =
//                        filterVolumesWithPolicyGroup(volumeManageFailedException, zapiConnection, volumeUUIDsByObjID);

                assignQosPolicyGroupToVolumes(volumeManageFailedException, connection, zapiConnection, volumeNamesByObjID, qosPolicyGroupId);
            } catch (UnknownHostException uhe) {
                throw new ZAPIManagerException("Unknown target cluster IP: " + targetIP, uhe);
            } finally {
                if (zapiConnection != null) {
                    zapiConnection.close();
                }
            }
        } catch (TableException te) {
            throw new ZAPIManagerException("Failed to connect to database", te);
        }
        finally {
            ConnectionHelper.close(connection);
        }
        if (volumeManageFailedException.hasErrors()) {
            throw volumeManageFailedException;
        }
    }

    private  NaElement getVolumeQosPolicySetRequest(final String vServerName, final String volumeName, final String qosPolicyGroupName) {
		final NaElement zapiRequestElt = new NaElement("volume-modify-iter");

        // Identify the target volume by vserver name and volume name
		final NaElement queryElem = new NaElement("query");
		zapiRequestElt.addChildElem(queryElem);
		final NaElement volumeAttrsQueryElt = new NaElement("volume-attributes");
		queryElem.addChildElem(volumeAttrsQueryElt);
		final NaElement volumeIdAttrsElt = new NaElement("volume-id-attributes");
		volumeAttrsQueryElt.addChildElem(volumeIdAttrsElt);
        final NaElement vServerNameElem = new NaElement("owning-vserver-name", vServerName);
        volumeIdAttrsElt.addChildElem(vServerNameElem);
		final NaElement volumeNameElem = new NaElement("name", volumeName);
		volumeIdAttrsElt.addChildElem(volumeNameElem);

        // Update the qos policy group
		final NaElement attributesElt = new NaElement("attributes");
		zapiRequestElt.addChildElem(attributesElt);
		final NaElement volumeAttrsElt = new NaElement("volume-attributes");
		attributesElt.addChildElem(volumeAttrsElt);
		final NaElement qosAttrsElt = new NaElement("volume-qos-attributes");
        volumeAttrsElt.addChildElem(qosAttrsElt);
		final NaElement policyGroupElem = new NaElement("policy-group-name", qosPolicyGroupName);
		qosAttrsElt.addChildElem(policyGroupElem);

		return zapiRequestElt;
    }

    /**
     * For each volume object ID, get a String array consisting of the parent VServer Name and the Volume Name
     * @param connection the database connection
     * @param targetVolumeObjIDs thet set of target volume objids to query
     * @return For each volume object ID, get a String array consisting of the parent VServer Name and the Volume Name
     * @throws TableException for database connection issues
     */
    private Map<Long, String[]> getVolumeObjIDToNameMap(final Connection connection, final Set<Long> targetVolumeObjIDs) throws TableException {
        final NetAppModel netAppModel = NetAppModel.INSTANCE;
        return netAppModel.getVolumeObjIDToNameMap(connection, targetVolumeObjIDs);
    }


    private Properties getDataSourceProperties(final Connection connection, final Long targetClusterObjId) throws TableException {
        final NetAppModel netAppModel = NetAppModel.INSTANCE;
        return netAppModel.getClusterDatasourceProperties(connection, targetClusterObjId);
    }

    private String getQosPolicyGroupNameByObjId(final Connection connection, final long qosPolicyGroupObjId) throws TableException {
        final NetAppModel netAppModel = NetAppModel.INSTANCE;
        return netAppModel.getQosPolicyGroupNameByObjId(connection, qosPolicyGroupObjId);
    }

    private String decryptPassword(final AcquisitionServerSessionLocal acquisitionServerSessionLocal,
                                   final String encryptedPassword) throws ZAPIManagerException {
        try {
            return acquisitionServerSessionLocal.decryptPassword(encryptedPassword);
        } catch (AcquisitionServerException ase) {
            throw new ZAPIManagerException("Failed to decrypt cluster password.");
        }
    }

    private int parsePortFromString(final String portString) {
        try {
            return Integer.parseInt(portString);
        } catch (NumberFormatException nfe) {
            return 80;
        }
    }

    private void assertAllVolumesAreKnown(final ZAPIManagerAggregateException volumeManageFailedException,
                                          final Map<Long, String[]> volumeNamesByObjID,
                                          final Set<Long> targetVolumeObjIds) {
        final Set<Long> missingVolumeUUIDs = new HashSet<Long>(targetVolumeObjIds);
        missingVolumeUUIDs.removeAll(volumeNamesByObjID.keySet());
        if (missingVolumeUUIDs.size() > 0) {
            for (final Long missingVolumeID : missingVolumeUUIDs) {
                volumeManageFailedException.addErrorMessageByVolumeID(missingVolumeID,
                        String.format("Failed to lookup UUID for volume %d",
                                missingVolumeID));
            }
        }
    }

    private NaServer getZAPIConnection(final String targetIP, final String userName, final String password,
                                       final int port, final String protocol) throws UnknownHostException {
        final NaServer zapiConnection = new NaServer(targetIP);
        zapiConnection.setAdminUser(userName, password);
        zapiConnection.setPort(port);
        zapiConnection.setServerType(NaServer.SERVER_TYPE_FILER);
        if ("HTTPS".equalsIgnoreCase(protocol)) {
            zapiConnection.setTransportType(NaServer.TRANSPORT_TYPE_HTTPS);
            zapiConnection.disableSSLv3();
        } else {
            zapiConnection.setTransportType(NaServer.TRANSPORT_TYPE_HTTP);
        }
        return zapiConnection;
    }


    private void assignQosPolicyGroupToVolumes(final ZAPIManagerAggregateException volumeManageFailedException,
                                              final Connection connection, final NaServer zapiConnection,
                                              final Map<Long, String[]> volumeNamesByObjID,
                                              final long qosPolicyGroupId) throws TableException {
        // The ZAPI does not give great feedback for bulk operations.
        // For example, unknown UUIDs are ignored.
        // Successful volumes are returned by vserver and volume name, rather than
        // given UUID.
        // Set them one at a time so that if setting fails, we know specifically which
        // target volume failed.
        // Lookup the name of the given QOS Policy Group
        String qosPolicyGroupName;
        if(qosPolicyGroupId == Long.MIN_VALUE){
            qosPolicyGroupName = "none";
        }else{
            qosPolicyGroupName = getQosPolicyGroupNameByObjId(connection, qosPolicyGroupId);
        }
        for (final Map.Entry<Long, String[]> volumeEntry : volumeNamesByObjID.entrySet()) {
            final Long volObjID = volumeEntry.getKey();
            final String[] volumeNames = volumeEntry.getValue();
            final String vServerName = volumeNames[0];
            final String volumeName = volumeNames[1];
            final NaElement setRequestElem = getVolumeQosPolicySetRequest(vServerName, volumeName, qosPolicyGroupName);
            try {
                final NaElement responseElem = zapiConnection.invokeElem(setRequestElem);
                final String responseContent = responseElem.getContent();
                final int successEndIndex = responseContent.indexOf("</num-succeeded>");
                if (successEndIndex < 0) {
                    volumeManageFailedException.addErrorMessageByVolumeID(volObjID, "Failed to configure default qos policy: " + responseContent);
                    continue;
                }
                final String numSuccessString = responseContent.substring(successEndIndex-1, successEndIndex);
                if (!"1".equals(numSuccessString)) {
                    volumeManageFailedException.addErrorMessageByVolumeID(volObjID, "Failed to configure default qos policy: " + responseContent);
                }
            } catch (IOException ioe) {
                volumeManageFailedException.addErrorMessageByVolumeID(volObjID, ioe.getMessage());
            } catch (NaProtocolException nape) {
                volumeManageFailedException.addErrorMessageByVolumeID(volObjID, nape.getMessage());
            } catch (NaAuthenticationException naae) {
                volumeManageFailedException.addErrorMessageByVolumeID(volObjID, naae.getMessage());
            } catch (NaAPIFailedException naapife) {
                volumeManageFailedException.addErrorMessageByVolumeID(volObjID, naapife.getMessage());
            } catch (Exception e) {
                volumeManageFailedException.addErrorMessageByVolumeID(volObjID, e.getMessage());
            }
        }
    }

}
