/*
 * $Id: //depot/prod/acquisition/main/tools/config/eclipse/.settings/org.eclipse.jdt.ui.prefs#1 $
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.oci.netapp.ems;

import java.io.Serializable;
import java.sql.Connection;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import com.onaro.sanscreen.acquisition.sessions.AcquisitionFacadeUtil;
import com.onaro.sanscreen.server.tables.acquisition.DataSourceAttrsTable;
import netapp.manage.NaServer;

import com.netapp.oci.netapp.client.interfaces.data.EmsZapiManagerException;
import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetAppEmsConfig;
import com.netapp.oci.platform.originator.AttributeChangeDetail;

/**
 * This class handles EMS registration during a cluster generation or major update
 */
public class ClusterUpgradeHandlerForEms  implements Serializable {
    private static final long serialVersionUID = 1L;
    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(ClusterUpgradeHandlerForEms.class);
    private static final String VERSION_MAJOR = "versionmajor";
    private static final String VERSION_GENERATION = "versiongeneration";
    public static final int LONG_BOARD_GENERATION = 9;

    private static final String FILTER_RULE_INCLUDE = "include";

    public static final ClusterUpgradeHandlerForEms INSTANCE = new ClusterUpgradeHandlerForEms();

    private ClusterUpgradeHandlerForEms() { }

    /**
     * Handles cluster upgrade - generation or major update
     * Registers for EMS if the cluster is upgraded to a LB generation from a prior to LB generation
     * else add new filter rules based on the majorVersion changes if cluster was already a LB cluster
     * @param connection
     * @param attributeChangeDetails
     * @param clusterId
     */
    public void handleClusterUpgrade(Connection connection, Set<AttributeChangeDetail> attributeChangeDetails, long clusterId) {
        // Either a generation change or a major version change is important
        String oldGeneration = null;
        String newGeneration = null;
        String oldMajor = null;
        String newMajor = null;
        NaServer zapiConnection = null;

        try {
            for (AttributeChangeDetail attributeChangeDetail : attributeChangeDetails) {
                if (attributeChangeDetail.getAttribute().equals(VERSION_GENERATION)) {
                    oldGeneration = toString(attributeChangeDetail.getOldValue());
                    newGeneration = toString(attributeChangeDetail.getNewValue());
                } else if (attributeChangeDetail.getAttribute().equals(VERSION_MAJOR)) {
                    oldMajor = toString(attributeChangeDetail.getOldValue());
                    newMajor = toString(attributeChangeDetail.getNewValue());
                }
            }
            if (oldGeneration != null || oldMajor != null) { //handle this update event
                long dsId = NetAppModel.INSTANCE.getDsIdByClusterId(connection, clusterId);
                final DataSourceAttrsTable dataSourceAttrsTable = new DataSourceAttrsTable(connection);
                Properties props = dataSourceAttrsTable.findByDsId(dsId);

                // If only major version has changed, Set<AttributeChangeDetail> will not contain generation details, so get these details from database
                String dbGeneration = NetAppModel.INSTANCE.getClusterGenerationByDsId(connection, dsId);
                String dbMajor = NetAppModel.INSTANCE.getClusterMajorByDsId(connection, dsId);

                if (dbGeneration == null || dbMajor == null) {
                    logger.error("Error retrieving cluster version from database. Not handling cluster upgrade for EMS events for cluster id " + clusterId);
                    return;
                }
                String fromGeneration = oldGeneration != null ? oldGeneration : dbGeneration;
                String toGeneration = newGeneration != null ? newGeneration : dbGeneration;

                if (Integer.valueOf(fromGeneration) < LONG_BOARD_GENERATION && Integer.valueOf(toGeneration) >= LONG_BOARD_GENERATION) { // register this cluster for EMS and all filter rules applicable for that cluster
                    try {
                        List<NetAppEmsConfig> emsConfigAll = NetAppModel.INSTANCE.getEmsConfigs(connection);
                        if (!emsConfigAll.isEmpty()) {
                            String hostname = AcquisitionFacadeUtil.getHostName();
                            AcquisitionFacadeUtil.registerEms(hostname, props, emsConfigAll);
                        }
                    } catch (EmsZapiManagerException e) {
                        logger.error("EMS registartion failed for clusterID {}", clusterId, e);
                    }
                } else { // Register new rules beginning from old majorVersion + 1 to current majorVersion
                    // +1 since EMS filters for the current major version are already registered with the cluster
                    String fromMajor = oldMajor != null ? String.valueOf(Integer.parseInt(oldMajor) + 1) : String.valueOf(Integer.parseInt(dbMajor) + 1);
                    String toMajor = newMajor != null ? newMajor : dbMajor;
                    List<String> filterRules = NetAppModel.INSTANCE.getEmsRulesByStateAndVersions(connection, NetAppEmsConfig.EmsMessageState.ADDED,
                        fromGeneration.concat(".").concat(fromMajor), toGeneration.concat(".").concat(toMajor));
                    if (filterRules == null || filterRules.isEmpty()) {
                        logger.info("No new EMS filters to register during cluster upgrade of cluster " + clusterId);
                        return;
                    }

                    final String filterName = EmsZapiManager.INSTANCE.getFilterName(AcquisitionFacadeUtil.getHostName());
                    zapiConnection = EmsZapiManager.INSTANCE.createZAPIConnection(props);

                    // Make a zapi call to register them we do have filter rules to add to this DS
                    for (String filterRule : filterRules) {
                        try {
                            //TODO - Remove this after one round of testing
                            logger.error("EMS Cluster Upgrade - Registering rule {} for clusterId {}", filterRule, clusterId);
                            EmsZapiManager.INSTANCE.addFilterRule(filterName, filterRule, FILTER_RULE_INCLUDE, zapiConnection);
                        } catch (EmsZapiManagerException ex) {
                            if (ex.getErrno() != EmsZapiManager.EDUPLICATEENTRY && ex.getErrno() != EmsZapiManager.E_EMS_NO_EVENT_MATCHED_THE_RULE) {
                                logger.error("EMS registartion failed for clusterId {} for filter rule {} with error no {}", clusterId, filterRule, ex.getErrno());
                            }
                        }
                    }
                }
            }
        } catch (Exception e) {
            logger.error("Cluster upgrade handler for EMS failed with exception " + e);
        } finally {
            if (zapiConnection != null) {
                zapiConnection.close();
            }
        }
    }

    private String toString(Object obj) {
        return obj == null ? null : obj.toString();
    }
}
