/*
 * $Id: //depot/prod/acquisition/main/tools/config/eclipse/.settings/org.eclipse.jdt.ui.prefs#1 $
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.oci.netapp.ems;

import java.util.HashMap;
/*
 * $Id$
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import netapp.manage.NaAPIFailedException;
import netapp.manage.NaElement;
import netapp.manage.NaServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;
import com.netapp.acquisition.aquisitionserverside.interfaces.impl.AcquisitionServerUtil;
import com.netapp.oci.netapp.client.interfaces.data.EmsZapiManagerException;

public final class EmsZapiManager {
    private static final Logger logger = LoggerFactory.getLogger(EmsZapiManager.class);

    public static final String CLIENT_TYPE_KEY = "client";

    private static final String QUERY_TAG = "query";
    private static final String ATTRIBUTES_TAG = "attributes";
    private static final String DESIRED_ATTRIBUTES_TAG = "desired-attributes";
    private static final String EVENT_NOTIFICATION_DESTINATION_TAG = "event-notification-destination";
    private static final String EVENT_NOTIFICATION_DESTINATION_INFO_TAG = "event-notification-destination-info";
    private static final String ELEMENT_MAX_RECORDS = "max-records";
    private static final String MAX_RECORDS = "5000";

    public static final String OCIE_PASSWORD = "netappfoundation.password";
    public static final String DS_IP = "netappfoundation.ip";
    public static final String OCIE_USER = "netappfoundation.user";
    public static final String OCIE_PORT = "netappfoundation.port";
    public static final String DS_PROTOCAL = "netappfoundation.connectionType";

    //zapi call names
    private static final String SECURITY_CERTIFICATE_INSTALL = "security-certificate-install";
    private static final String SECURITY_CERTIFICATE_GET_ITER = "security-certificate-get-iter";
    private static final String SECURITY_CERTIFICATE_DELETE = "security-certificate-delete";
    private static final String EMS_EVENT_FILTER_CREATE = "ems-event-filter-create";
    private static final String EMS_EVENT_FILTER_DESTROY_ITER = "ems-event-filter-destroy-iter";
    private static final String EMS_EVENT_FILTER_RULE_ADD = "ems-event-filter-rule-add";
    private static final String EMS_EVENT_FILTER_RULE_DELETE = "ems-event-filter-rule-delete";
    private static final String EMS_EVENT_NOTIFICATION_DESTINATION_CREATE = "ems-event-notification-destination-create";
    private static final String EMS_EVENT_NOTIFICATION_DESTINATION_MODIFY = "ems-event-notification-destination-modify-iter";
    private static final String EMS_EVENT_NOTIFICATION_DESTINATION_SHOW = "ems-event-notification-destination-get-iter";
    private static final String EMS_EVENT_NOTIFICATION_DESTINATION_DESTROY_ITER = "ems-event-notification-destination-destroy-iter";
    private static final String EMS_EVENT_NOTIFICATION_CREATE = "ems-event-notification-create";
    private static final String EMS_EVENT_NOTIFICATION_DESTINATION_CHECK = "ems-event-notification-destination-check";
    private static final String EMS_EVENT_NOTIFICATION_DESTROY_ITER = "ems-event-notification-destroy-iter";
    private static final String EMS_EVENT_FILTER_GET_ITER = "ems-event-filter-get-iter";
    private static final String CLUSTER_NODE_GET_ITER = "cluster-node-get-iter";
    private static final String SYSTEM_GET_VERTION = "system-get-version";
    private static final String EMS_EVENT_CATALOG_GET_ITER = "ems-event-catalog-get-iter";
    private static final String EMS_MSG_GET_ITER = "ems-message-get-iter";
    private static final String EMS_MSG_INFO = "ems-message-info";
    private static final String EMS_POST_END_POINT = "/acq/ontap/ems";
    private static final String EMS_POST_PROTOCOL = "https";
    private static final String EMS_PORT = "9443";
    private static final int CLUSTER_HTTPS_PORT = 443;

    /**
     * EMS ZAPI response error numbers defined at:
     * http://web.eng.netapp.com/engineering/rlse/DOT/devN/final/bedrock/export/common/zapiprod/web_ontap_cluster/ems/index.html
     */
    public static final int EDUPLICATEENTRY = 13130;
    public static final int E_EMS_NO_EVENT_MATCHED_THE_RULE = 27107;
    public static final int ZAPI_INFRASTRUCTURE_ERROR = 14513;

    public static class ClientKeyInformation{
        private String clientCertificateKeySerial;
        private String clientCertificateKeyAuthority;

        public ClientKeyInformation() {

        }

        public ClientKeyInformation(String clientCertificateKeySerial, String clientCertificateKeyAuthority) {
            this.clientCertificateKeySerial = clientCertificateKeySerial;
            this.clientCertificateKeyAuthority = clientCertificateKeyAuthority;
        }
        public String getClientCertificateKeySerial() {
            return clientCertificateKeySerial;
        }
        public String getClientCertificateKeyAuthority() {
            return clientCertificateKeyAuthority;
        }
    }

    public static final EmsZapiManager INSTANCE = new EmsZapiManager();

    private EmsZapiManager() { }

    /**
     * security-certificate-install
     *
     * @param type
     * @param certificate
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement installSecurityCertificate(String commonName, String type, List<String> certificates, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse = null;
        for (String certificate : certificates) {
            final NaElement zapiRequestElt = new NaElement(EmsZapiManager.SECURITY_CERTIFICATE_INSTALL);
            zapiRequestElt.addNewChild("type", type);
            zapiRequestElt.addNewChild("certificate", certificate);
            zapiRequestElt.addNewChild("common-name", commonName);
            try {
                zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
            } catch (NaAPIFailedException ne) {
                if (ne.getErrno() == EmsZapiManager.EDUPLICATEENTRY) {
                    logger.warn("Certificate has already been installed to cluster vserver '{}' for '{}'", zapiConnection.getVserver(), commonName);
                    continue;
                } else {
                    throw new EmsZapiManagerException(ne);
                }
            } catch (Exception ex) {
                logger.error("Failed to call security-certificate-install -- ", ex);
                throw new EmsZapiManagerException(ex.getMessage());
            }
        }
        return zapiResponse;
    }


    public NaElement installClientKey(String privateKey, String certificate, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse = null;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.SECURITY_CERTIFICATE_INSTALL);
        zapiRequestElt.addNewChild("type", CLIENT_TYPE_KEY);
        zapiRequestElt.addNewChild("private-key", privateKey);
        zapiRequestElt.addNewChild("certificate", certificate);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            if (ne.getErrno() == EmsZapiManager.EDUPLICATEENTRY) {
                logger.warn("Client key has already been installed to cluster vserver '{}'!", zapiConnection.getVserver());
            } else {
                throw new EmsZapiManagerException(ne);
            }
        } catch (Exception ex) {
            logger.error("Failed to call security-certificate-install -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * Check the EMS notification working fine.
     *
     * @param destName - the notification destination name
     * @param nodeName - node name of the cluster
     * @param zapiConnection
     * @return true - check passed
     *         false - check failed
     * @throws EmsZapiManagerException - the notification destination doesn't exist
     */
    public boolean emsNotificationDestinationCheck(String destName, String nodeName, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse = null;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_NOTIFICATION_DESTINATION_CHECK);
        zapiRequestElt.addNewChild("node", nodeName);
        zapiRequestElt.addNewChild("destination-name", destName);

        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException nafe){
            if (nafe.getErrno() == EmsZapiManager.ZAPI_INFRASTRUCTURE_ERROR) {
                logger.warn("Ignore: this error showed that there are missing elements from the response. ", nafe);
                return false;
            } else {
                //Throwing EmsZapiManagerException with error code.
                throw new EmsZapiManagerException(nafe);
            }
        } catch (Exception e) {
            logger.warn("Destination entry doesn't exist!", e);
            throw new EmsZapiManagerException(e.getMessage());
        }

        return "false".equals(zapiResponse.getChildContent("check-passed")) ? false : true;
    }

    @SuppressWarnings("unchecked")
    public String getClusterNodeName(NaServer zapiConnection) throws EmsZapiManagerException {
        final NaElement zapiRequestElt = new NaElement(CLUSTER_NODE_GET_ITER);
        String nodeName = null;

        try {
            NaElement zapiResponse = zapiConnection.invokeElem(zapiRequestElt);

            if (zapiResponse != null) {
                NaElement attributesList = zapiResponse.getChildByName("attributes-list");
                if (attributesList != null) {
                    List<NaElement>  nodeInfoList = attributesList.getChildren();
                    if (nodeInfoList != null) {
                        for (NaElement nodeInfo : nodeInfoList) {
                            // find the first one then return
                            nodeName = nodeInfo.getChildByName("node-name").getContent();
                            break;
                        }
                    }
                }
            }
        } catch (Exception ex) {
            logger.error("Unable invoke cluster-node-get-iter; ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }

        return nodeName;
    }

    private ClientKeyInformation getClientKeyInformation(String destinationName, NaServer zapiConnection) throws EmsZapiManagerException {
        final NaElement zapiQueryRequestElt = new NaElement(EMS_EVENT_NOTIFICATION_DESTINATION_SHOW);

        NaElement query = new NaElement(QUERY_TAG);
        NaElement eventNotificationDestinationQuery = new NaElement(EVENT_NOTIFICATION_DESTINATION_INFO_TAG);
        eventNotificationDestinationQuery.addNewChild("name", destinationName);
        query.addChildElem(eventNotificationDestinationQuery);
        zapiQueryRequestElt.addChildElem(query);

        NaElement desiredAttributes = new NaElement(DESIRED_ATTRIBUTES_TAG);
        NaElement eventNotificationDestinationAttr = new NaElement(EVENT_NOTIFICATION_DESTINATION_INFO_TAG);
        eventNotificationDestinationAttr.addNewChild("certificate-authority", "");
        eventNotificationDestinationAttr.addNewChild("certificate-serial", "");
        desiredAttributes.addChildElem(eventNotificationDestinationAttr);
        zapiQueryRequestElt.addChildElem(desiredAttributes);

        // Execute query for event notification destination
        NaElement zapiQueryResponse;
        try {
            zapiQueryResponse = zapiConnection.invokeElem(zapiQueryRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call " + EMS_EVENT_NOTIFICATION_DESTINATION_SHOW, ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }

        NaElement ca = zapiQueryResponse.getChildByName("certificate-authority");
        NaElement serial = zapiQueryResponse.getChildByName("certificate-serial");
        if(ca != null) {
            return new ClientKeyInformation(serial.getContent(), ca.getContent());
        } else {
            return null;
        }
    }

    public boolean isClientAuthenticationEnabled(String destinationName, NaServer zapiConnection){
        try {
            return getClientKeyInformation(destinationName, zapiConnection) != null;
        } catch (EmsZapiManagerException e) {
            logger.error("Failed to get client key information: " + e.getLocalizedMessage(),e);
            return false;
        }
    }


    public NaElement deleteClientKey(String destinationName, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse = null;

        ClientKeyInformation keyInfo = getClientKeyInformation(destinationName, zapiConnection);

        // if ca is not null then there is an associate client key to delete otherwise do nothing and return null
        if (keyInfo != null) {
            final NaElement zapiRequestElt = new NaElement(EmsZapiManager.SECURITY_CERTIFICATE_DELETE);
            zapiRequestElt.addNewChild("common-name", keyInfo.getClientCertificateKeyAuthority());
            zapiRequestElt.addNewChild("type", CLIENT_TYPE_KEY);
            zapiRequestElt.addNewChild("certificate-authority", keyInfo.getClientCertificateKeyAuthority());
            zapiRequestElt.addNewChild("serial-number", keyInfo.getClientCertificateKeySerial());

            try {
                zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
            } catch (NaAPIFailedException ne) {
                throw new EmsZapiManagerException(ne);
            } catch (Exception ex) {
                logger.error("Failed to call security-certificate-delete -- ", ex);
                throw new EmsZapiManagerException(ex.getMessage());
            }
        }
        return zapiResponse;
    }

    /**
     * security-certificate-delete
     *
     * @param vserver
     * @param serialNumber
     * @param commonName
     * @param type
     * @param certificateAuthority
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement deleteSecurityCertificate(String vserver, String serialNumber, String commonName, String type, String certificateAuthority, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.SECURITY_CERTIFICATE_DELETE);
        zapiRequestElt.addNewChild("common-name", commonName);
        zapiRequestElt.addNewChild("type", type);
        zapiRequestElt.addNewChild("certificate-authority", certificateAuthority);
        zapiRequestElt.addNewChild("vserver", vserver);
        zapiRequestElt.addNewChild("serial-number", serialNumber);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call security-certificate-delete -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * ems-event-filter-create
     *
     * @param filterName
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement createEventFilter(String filterName, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_FILTER_CREATE);
        zapiRequestElt.addNewChild("filter-name", filterName);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call ems-event-filter-create -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * ems-event-filter-destroy-iter
     *
     * @param filterName
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement deleteEventFilter(String filterName, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_FILTER_DESTROY_ITER);
        NaElement queryEle = new NaElement(QUERY_TAG);
        zapiRequestElt.addChildElem(queryEle);
        NaElement eventFilterEle = new NaElement("event-filter");
        queryEle.addChildElem(eventFilterEle);
        eventFilterEle.addNewChild("filter-name", filterName);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call ems-event-filter-destroy-iter -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * ems-event-filter-rule-add
     *
     * @param filterName
     * @param messageName
     * @param eventRuleType  possible values are include or exclude.
     * @param severity
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement addFilterRule(String filterName, String messageName, String type, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_FILTER_RULE_ADD);
        zapiRequestElt.addNewChild("filter-name", filterName);
        zapiRequestElt.addNewChild("message-name", messageName);
        zapiRequestElt.addNewChild("type", type);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        }  catch (Exception ex) {
            logger.error("Failed to call ems-event-filter-rule-add -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
    * Delete a event filter rule.
    *
    * @param filterName
    * @param messageName
    * @param type
    * @param zapiConnection
    * @return
    * @throws EmsZapiManagerException
    */
    public  NaElement deleteEventFilterRule(String filterName, String messageName, String type, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_FILTER_RULE_DELETE);
        zapiRequestElt.addNewChild("filter-name", filterName);
        String position = getFilterRulePosition(filterName, messageName, type, zapiConnection);
        //if position=0, the rule does not exist and so skip zapi call
        if (Integer.parseInt(position) == 0) {
            return new NaElement("empty-element");
        }

        zapiRequestElt.addNewChild("position", position);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call ems-event-filter-rule-delete -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * security-certificate-get-iter
     *
     * This method is used to retrieve certificate-authority for deleting the security certificate
     *
     * @param commonName
     * @param type
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement showNotificationDestinations(NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_NOTIFICATION_DESTINATION_SHOW);
        zapiRequestElt.addNewChild(ELEMENT_MAX_RECORDS, MAX_RECORDS);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call security-certificate-get-iter -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * ems-event-notification-destination-create
     *
     * @param name
     * @param restApiURL  example value: https://10.97.149.20/acq/ontap/ems
     * @param certificateAuthority
     * @param certificateSerial
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public  NaElement createEventNotificationDestination(String name, String restApiURL, String certificateAuthority, String certificateSerial, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_NOTIFICATION_DESTINATION_CREATE);
        zapiRequestElt.addNewChild("name", name);
        zapiRequestElt.addNewChild("rest-api-url", restApiURL);
        if (certificateAuthority != null) {
            zapiRequestElt.addNewChild("certificate-authority", certificateAuthority);
        }
        if (certificateSerial != null) {
            zapiRequestElt.addNewChild("certificate-serial", certificateSerial);
        }
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call ems-event-notification-destination-create -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }


    /**
     * Update already
     * @param name
     * @param restApiURL
     * @param certificateAuthority
     * @param certificateSerial
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement modifyEventNotificationDestination(String name, String restApiURL, String certificateAuthority, String certificateSerial, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EMS_EVENT_NOTIFICATION_DESTINATION_MODIFY);

        NaElement attributes = new NaElement(ATTRIBUTES_TAG);
        NaElement eventNotificationDestinationAttr = new NaElement(EVENT_NOTIFICATION_DESTINATION_TAG);
        eventNotificationDestinationAttr.addNewChild("rest-api-url", restApiURL);
        if (certificateAuthority != null) {
            eventNotificationDestinationAttr.addNewChild("certificate-authority", certificateAuthority);
        }
        if (certificateSerial != null) {
        eventNotificationDestinationAttr.addNewChild("certificate-serial", certificateSerial);
        }
        attributes.addChildElem(eventNotificationDestinationAttr);
        zapiRequestElt.addChildElem(attributes);

        NaElement query = new NaElement(QUERY_TAG);
        NaElement eventNotificationDestinationQuery = new NaElement(EVENT_NOTIFICATION_DESTINATION_TAG);
        eventNotificationDestinationQuery.addNewChild("name", name);
        query.addChildElem(eventNotificationDestinationQuery);
        zapiRequestElt.addChildElem(query);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call ems-event-notification-destination-create -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * ems-event-notification-destination-destroy-iter
     *
     * @param name
     * @param restApiURL
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement deleteEventNotificationDestination(String name, String restApiURL, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_NOTIFICATION_DESTINATION_DESTROY_ITER);
        NaElement queryEle = new NaElement(QUERY_TAG);
        zapiRequestElt.addChildElem(queryEle);
        NaElement eventNotificationDestinationEle = new NaElement("event-notification-destination");
        queryEle.addChildElem(eventNotificationDestinationEle);
        eventNotificationDestinationEle.addNewChild("name", name);
        eventNotificationDestinationEle.addNewChild("rest-api-url", restApiURL);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call ems-event-notification-destination-destroy-iter -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * ems-event-notification-create
     *
     * @param filterName
     * @param destinations
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement createEventNotification(String filterName, List<String> destinations, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_NOTIFICATION_CREATE);
        zapiRequestElt.addNewChild("filter-name", filterName);

        NaElement destinationsElement = new NaElement("destinations");
        zapiRequestElt.addChildElem(destinationsElement);
        for (String destination : destinations) {
            destinationsElement.addNewChild("string", destination);
        }
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call ems-event-notification-create -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * ems-event-notification-destroy-iter
     *
     * @param filterName
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement deleteEventNotification(String filterName, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_NOTIFICATION_DESTROY_ITER);
        NaElement queryEle = new NaElement(QUERY_TAG);
        zapiRequestElt.addChildElem(queryEle);
        NaElement eventNotificationEle = new NaElement("event-notification");
        queryEle.addChildElem(eventNotificationEle);
        eventNotificationEle.addNewChild("filter-name", filterName);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call ems-event-notification-destroy-iter -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * security-certificate-get-iter
     *
     * This method is used to retrieve certificate-authority for deleting the security certificate
     *
     * @param commonName
     * @param type
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement showSecurityCertificates(String commonName, String type, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.SECURITY_CERTIFICATE_GET_ITER);
        NaElement queryEle = new NaElement(QUERY_TAG);
        zapiRequestElt.addChildElem(queryEle);
        NaElement securityCertificateEle = new NaElement("certificate-info");
        queryEle.addChildElem(securityCertificateEle);
        securityCertificateEle.addNewChild("common-name", commonName);
        securityCertificateEle.addNewChild("type", type);
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call security-certificate-get-iter -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }

    /**
     * Uninstall security certificate identified by the EMS destination name.
     *
     * @param emsDestName
     * @param type
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public NaElement uninstallSecurityCertificate(String emsDestName, String type, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse = null;
        try {
            NaElement responseCertShow = showSecurityCertificates(emsDestName, type, zapiConnection);
            NaElement attributesList = responseCertShow.getChildByName("attributes-list");
            if (attributesList != null) {
                NaElement certInfo = attributesList.getChildByName("certificate-info");
                if (certInfo != null) {
                    String serialNumber = certInfo.getChildByName("serial-number").getContent();
                    String vserver = certInfo.getChildByName("vserver").getContent();
                    String certificateAuthority = certInfo.getChildByName("certificate-authority").getContent();
                    zapiResponse = deleteSecurityCertificate(vserver, serialNumber, emsDestName, type, certificateAuthority, zapiConnection);
                }
            }
        } catch (Exception ex) {
            logger.error("Failed to call security-certificate-delete -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }

        return zapiResponse;
    }

    /**
     * This method is calling EMS_EVENT_FILTER_GET_ITER and return the position for the filter.
     *
     * @param filterName
     * @param messageName
     * @param type
     * @param zapiConnection
     * @return
     */
    private String getFilterRulePosition(String filterName, String messageName, String type, NaServer zapiConnection) {
        String posisition = String.valueOf(0);

        NaElement zapiResponse = null;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_FILTER_GET_ITER);
        NaElement queryEle = new NaElement(QUERY_TAG);
        zapiRequestElt.addChildElem(queryEle);
        NaElement filterEle = new NaElement("event-filter");
        queryEle.addChildElem(filterEle);
        filterEle.addNewChild("filter-name", filterName);
        NaElement eventfilterruleEle = new NaElement("event-filter-rule");
        filterEle.addChildElem(eventfilterruleEle);
        NaElement eventfilterruleinfoEle = new NaElement("event-filter-rule-info");
        eventfilterruleEle.addChildElem(eventfilterruleinfoEle);
        eventfilterruleinfoEle.addNewChild("message-name", messageName);
        eventfilterruleinfoEle.addNewChild("type", type);

        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (Exception ex) {
            logger.error("Failed to call ems-event-filter-get-iter -- ", ex);
        }

        if (zapiResponse != null) {
            NaElement attributesList = zapiResponse.getChildByName("attributes-list");
            if (attributesList != null) {
                posisition = attributesList.getChildByName("event-filter").getChildByName("event-filter-rule").getChildByName("event-filter-rule-info").getChildByName("position").getContent();
            }
        }

        return posisition;
    }

    /**
     * Get the cluster version with generation.major.
     *
     * @see com.netapp.oci.netapp.ems.zapi.interfaces.local.EmsZapiManager#getDSClusterVersion(netapp.manage.NaServer)
     */
    public String getDSClusterVersion(NaServer zapiConnection) throws EmsZapiManagerException {
        String version = "0.0";
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.SYSTEM_GET_VERTION);
        try {
            NaElement zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
            if (zapiResponse != null) {
                NaElement versionTuple = zapiResponse.getChildByName("version-tuple").getChildByName("system-version-tuple");
                String generation = versionTuple.getChildContent("generation");
                String major = versionTuple.getChildContent("major");
                version = generation + "." + major;
            }
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed to call system-get-version -- ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return version;
    }

    /**
     * Sample request:
     *
     * <ems-event-catalog-get-iter>
     *   <query>
     *     <event-catalog-info>
     *        <message-name>cloud.aws.iamCredsInvalid, monitor.chassisPowerSupply.degraded </message-name>
     *     </event-catalog-info>
     *   </query>
     * </ems-event-catalog-get-iter>
     *
     * @see com.netapp.oci.netapp.ems.zapi.interfaces.local.EmsZapiManager#getEmsEventCatalogByMsgName(java.lang.String, netapp.manage.NaServer)
     */
    public NaElement getEmsEventCatalogByMsgName(List<String> msgNames, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement zapiResponse = null;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_EVENT_CATALOG_GET_ITER);
        NaElement queryEle = new NaElement(QUERY_TAG);
        zapiRequestElt.addChildElem(queryEle);
        NaElement catalogInfoEle = new NaElement("event-catalog-info");
        queryEle.addChildElem(catalogInfoEle);
        StringBuilder messageNameList = new StringBuilder();
        if (msgNames != null && !msgNames.isEmpty()) {
            String delimiter = "";
            for (String msgName : msgNames) {
                messageNameList.append(delimiter).append(msgName);
                delimiter = ",";
            }
        }
        catalogInfoEle.addNewChild("message-name", messageNameList.toString());
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed ZAPI call " + EmsZapiManager.EMS_EVENT_CATALOG_GET_ITER, ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }


    /**
     * Sample request:
     *
     * <ems-message-get-iter>
     *   <max-records>12345</max-records>
     *   <query>
     *     <ems-message-info>
     *       <time>1494530220..1494514200</time>
     *       <message-name>acp.common.*, monitor.chassisPowerSupply.degraded</message-name>
     *     </ems-message-info>
     *   </query>
     * </ems-message-get-iter>
     *
     * @see com.netapp.oci.netapp.ems.zapi.interfaces.local.EmsZapiManager#getEmsEvents(java.lang.String, netapp.manage.NaServer)
     */
    public NaElement getEmsEvents(NaServer zapiConnection, Set<String> emsMessageRules, long fromTimestamp) throws EmsZapiManagerException {
        NaElement zapiResponse = null;
        final NaElement zapiRequestElt = new NaElement(EmsZapiManager.EMS_MSG_GET_ITER);

        zapiRequestElt.addNewChild(ELEMENT_MAX_RECORDS, MAX_RECORDS);

        NaElement queryEle = new NaElement(QUERY_TAG);
        zapiRequestElt.addChildElem(queryEle);
        NaElement msgInfo = new NaElement(EmsZapiManager.EMS_MSG_INFO);
        queryEle.addChildElem(msgInfo);

        StringBuffer sbTimeStamp = new StringBuffer();
        long toTimestamp = System.currentTimeMillis() / 1000;
        sbTimeStamp.append(toTimestamp).append("..").append(fromTimestamp);
        msgInfo.addNewChild("time", sbTimeStamp.toString());

        StringBuffer sbEmsRules = new StringBuffer();
        if (emsMessageRules != null && !emsMessageRules.isEmpty()) {
            String delimiter = "";
            for (String rule : emsMessageRules) {
                sbEmsRules.append(delimiter).append(rule);
                delimiter = ",";
            }
        }
        msgInfo.addNewChild("message-name", sbEmsRules.toString());
        try {
            zapiResponse = zapiConnection.invokeElem(zapiRequestElt);
        } catch (NaAPIFailedException ne) {
            throw new EmsZapiManagerException(ne);
        } catch (Exception ex) {
            logger.error("Failed ZAPI call " + EmsZapiManager.EMS_MSG_GET_ITER, ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }
        return zapiResponse;
    }


    /**
     * create zapiConnection
     *
     * @param dsProperties
     * @return
     * @throws EmsZapiManagerException
     */
    public NaServer createZAPIConnection(Properties dsProperties) throws EmsZapiManagerException {
        NaServer zapiConnection;
        String targetIP = dsProperties.getProperty(DS_IP);
        String userName = dsProperties.getProperty(OCIE_USER);
        String password = dsProperties.getProperty(OCIE_PASSWORD);
        String protocal = dsProperties.getProperty(DS_PROTOCAL);
        String portString = dsProperties.getProperty(OCIE_PORT);
        int port = 80;
        if (portString != null) {
            port = Integer.valueOf(portString);
        }

        try {
            zapiConnection = new NaServer(targetIP);
            String decryptedPassword = decryptPassword(password);
            zapiConnection.setAdminUser(userName, decryptedPassword);
            zapiConnection.setPort(port);
            zapiConnection.setServerType(NaServer.SERVER_TYPE_FILER);
            if ("HTTPS".equalsIgnoreCase(protocal)) {
                zapiConnection.setTransportType(NaServer.TRANSPORT_TYPE_HTTPS);
                zapiConnection.disableSSLv3();
            } else {
                zapiConnection.setTransportType(NaServer.TRANSPORT_TYPE_HTTP);
            }
        } catch (Exception ex) {
            logger.error("Failed to create zapiConnection: ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }

        return zapiConnection;
    }


    /**
     * Retrieves suggested actions for a list of message names provided as input
     * @param clusterUUID
     * @param msgName
     * @Connection dbConnection
     * @return
     * @throws EmsZapiManagerException
     */
    public Map<String, String> getEmsRemediation(final List<String> msgNames, NaServer zapiConnection) throws EmsZapiManagerException {
        NaElement catalog = EmsZapiManager.INSTANCE.getEmsEventCatalogByMsgName(msgNames, zapiConnection);
        NaElement attributesList = catalog.getChildByName("attributes-list");
        Map<String, String> msgNameActionmap = new HashMap<>();
        if (attributesList != null) {
            @SuppressWarnings("unchecked")
            List<NaElement> children = attributesList.getChildren();
            if (children == null)
                return null;
            for (NaElement element : children) {
                if (element.hasChildren()) {
                    String msgName = element.getChildContent("message-name");
                    if (msgName != null) {
                        msgNameActionmap.put(msgName, element.getChildContent("action"));
                    }
                }
            }
        }
        return msgNameActionmap;
    }

    /**
     * create zapiConnection
     *
     * @param dsProperties
     * @return
     * @throws EmsZapiManagerException
     */
    public NaServer createZAPIConnection(String clusterIp, String username, String pswd) throws EmsZapiManagerException {
        NaServer zapiConnection;

        try {
            zapiConnection = new NaServer(clusterIp);
            zapiConnection.setAdminUser(username, pswd);
            zapiConnection.setPort(CLUSTER_HTTPS_PORT);
            zapiConnection.setServerType(NaServer.SERVER_TYPE_FILER);
            zapiConnection.setTransportType(NaServer.TRANSPORT_TYPE_HTTPS);
            zapiConnection.disableSSLv3();
        } catch (Exception ex) {
            logger.error("Failed to create zapiConnection: ", ex);
            throw new EmsZapiManagerException(ex.getMessage());
        }

        return zapiConnection;
    }

    /**
     * Construct destinationName out of the hostname
     *
     * @param hostname
     * @return
     */
    public String getDestinationName(String hostname) {
        //replace "." with "-" if hostname contains ".", for example, srcr.gdl.englab.netapp.com, since ZAPI call param cannot contain "."
        return hostname.replace('.', '-') + "_server";
    }

    /**
     * Construct destinationURL out of the hostname
     *
     * @param hostname
     * @return
     */
    public String getDestinationURL(String hostname) {
        return  EMS_POST_PROTOCOL + "://" + hostname + ":" + EMS_PORT  + EMS_POST_END_POINT;
    }

    /**
     * Construct filter name out of the hostname
     *
     * @param hostname
     * @return
     */
    public String getFilterName(String hostname) {
        //replace "." with "-" if hostname contains ".", for example, srcr.gdl.englab.netapp.com, since ZAPI call only accepts "-" or ".".
        return hostname.replace('.', '-') + "_filter";
    }


    private String decryptPassword(final String encryptedPassword) throws AcquisitionServerException  {
        if (encryptedPassword == null || encryptedPassword.isEmpty()) {
            return encryptedPassword;
        }
        return AcquisitionServerUtil.getCipherFacade().decrypt(encryptedPassword);
    }

}
