/*
 * $Id: //depot/prod/acquisition/main/tools/config/eclipse/.settings/org.eclipse.jdt.ui.prefs#1 $
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.oci.netapp.ems.shared.interfaces.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import netapp.manage.NaServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.client.interfaces.data.EmsZapiManagerException;
import com.netapp.oci.netapp.ems.EmsZapiManager;
import com.netapp.oci.netapp.ems.shared.interfaces.data.NetAppEmsEventException;
import com.netapp.oci.netapp.ems.shared.interfaces.data.NetappEmsEvent;
import com.netapp.oci.netapp.ems.shared.interfaces.data.NetappEmsEventParameter;
import com.netapp.oci.netapp.interfaces.data.NetappEms;
import com.netapp.oci.netapp.interfaces.data.NetappEms.EmsMessageInfo;
import com.netapp.oci.netapp.interfaces.data.NetappEmsEventDto;
import com.netapp.oci.netapp.interfaces.data.NetappEmsZapi;
import com.netapp.oci.netapp.interfaces.data.NetappEmsZapi.AttributesList.EmsMessageInfo.Parameters.Parameter;

public final class NetappEmsEventUtility {

    private static final Logger logger = LoggerFactory.getLogger(NetappEmsEventUtility.class);
    private static final String EMS_CORRECTIVE_ACTION_KEY = "action";
    private static final String EMS_CORRECTIVE_ACTION_NONE = "None";

    private NetappEmsEventUtility() {
        // not called
    }

    /**
     * Converts a single EMS message received from POST request to entity object -  NetappEmsEvent
     * @param emsMsg
     * @return ems entity object
     * @throws NetAppEmsEventException
     */
    public static NetappEmsEvent convertNetAppEms(NetappEms emsMsg) throws NetAppEmsEventException {

        EmsMessageInfo emsMsgInfo = emsMsg.getEmsMessageInfo();
        String clusterUuid = null;
        String nodeUuid = null;
        if (emsMsg.getEmsMessageOrigin() != null) {
            clusterUuid = emsMsg.getEmsMessageOrigin().getClusterUuid();
            nodeUuid = emsMsg.getEmsMessageOrigin().getNodeUuid();
        }
        if (clusterUuid == null) {
            throw new NetAppEmsEventException("Cluster uuid is NULL");
        }
        NetappEmsEvent event = new NetappEmsEvent.Builder().withMessageName(emsMsgInfo.getMessageName()).withSeverity(emsMsgInfo.getSeverity())
                .withTime(emsMsgInfo.getTime()).withEvent(emsMsgInfo.getEvent()).withEmsSeverity(emsMsgInfo.getEmsSeverity())
                .withSource(emsMsgInfo.getSource()).withVersion(emsMsgInfo.getVersion()).withNodeUuid(nodeUuid).withClusterUuid(clusterUuid).build();

        List<EmsMessageInfo.Parameters.Parameter> params = emsMsgInfo.getParameters().getParameter();
        for (EmsMessageInfo.Parameters.Parameter param : params) {
            NetappEmsEventParameter eventParameter = new NetappEmsEventParameter();
            eventParameter.setParameterName(param.getName());
            eventParameter.setParameterValue(param.getValue());
            event.addEventParameter(eventParameter);
        }
        return event;
    }

    /**
     * Converts a list of EMS messages received from PULL request to a list of entity objects -  NetappEmsEvent
     * @param zapiEmsEvents
     * @param clusterUuid
     * @param zapiConnection
     * @return list of ems entity objects
     * @throws NetAppEmsEventException
     */
    public static List<NetappEmsEvent> convertNetAppEmsZapi(NetappEmsZapi zapiEmsEvents, String clusterUuid, NaServer zapiConnection) throws NetAppEmsEventException {

        if (zapiEmsEvents.getAttributesList() == null || zapiEmsEvents.getAttributesList().getEmsMessageInfo() == null) {
            return null;
        }
        if (clusterUuid == null) {
            throw new NetAppEmsEventException("Cluster uuid is NULL");
        }
        List<NetappEmsEvent> events = new ArrayList<>();
        List<com.netapp.oci.netapp.interfaces.data.NetappEmsZapi.AttributesList.EmsMessageInfo> emsMsgInfoList = zapiEmsEvents.getAttributesList().getEmsMessageInfo();
        // Fetch remediation action for all pulled ems events by first retrieving message names of all message in a list
        // And providing the message list to EmsZapiManager to get a map of <message_Name, suggested_Action>
        List<String> emsMsgNames = new ArrayList<>();
        Map<String, String> msgNameActionMap = new HashMap<>();
        for (com.netapp.oci.netapp.interfaces.data.NetappEmsZapi.AttributesList.EmsMessageInfo emsMsgInfo : emsMsgInfoList) {
            emsMsgNames.add(emsMsgInfo.getMessageName());
        }
        try {
            msgNameActionMap = EmsZapiManager.INSTANCE.getEmsRemediation(emsMsgNames, zapiConnection);
        } catch (EmsZapiManagerException e) { // catch exception and still continue processing the events
            logger.error("Error retrieving suggested actions " + e);
        }

        for (com.netapp.oci.netapp.interfaces.data.NetappEmsZapi.AttributesList.EmsMessageInfo emsMsgInfo : emsMsgInfoList) {
            // Events retrieved from ZAPI calls will have nodeUuid = null since that information is not passed by ONTAP
            NetappEmsEvent event = new NetappEmsEvent.Builder().withMessageName(emsMsgInfo.getMessageName()).withSeverity(emsMsgInfo.getSeverity())
                    .withTime(emsMsgInfo.getTime()).withEvent(emsMsgInfo.getEvent()).withEmsSeverity(emsMsgInfo.getEmsSeverity())
                    .withSource(emsMsgInfo.getSource()).withVersion(emsMsgInfo.getVersion()).withNodeUuid(emsMsgInfo.getNode()).withClusterUuid(clusterUuid).build();
            if (emsMsgInfo.getParameters() != null) {
                List<Parameter> params = emsMsgInfo.getParameters().getParameter();
                for (Parameter param : params) {
                    NetappEmsEventParameter eventParameter = new NetappEmsEventParameter();
                    eventParameter.setParameterName(param.getName());
                    eventParameter.setParameterValue(param.getValue());
                    event.addEventParameter(eventParameter);
                }
            }
            // Add suggested action as a event parameter
            NetappEmsEventParameter eventActionParameter = new NetappEmsEventParameter();
            eventActionParameter.setParameterName(EMS_CORRECTIVE_ACTION_KEY);
            eventActionParameter.setParameterValue(msgNameActionMap.containsKey(emsMsgInfo.getMessageName()) ?  msgNameActionMap.get(emsMsgInfo.getMessageName()): EMS_CORRECTIVE_ACTION_NONE);
            event.addEventParameter(eventActionParameter);

            events.add(event);
        }
        return events;
    }

    /**
     * Converts NetappEmsEvent object to a dto object
     * @param event
     * @return NetappEmsEventDto
     */
    public static NetappEmsEventDto convertToDto(NetappEmsEvent event) {
        if (event == null) {
            return null;
        }
        return new NetappEmsEventDto.Builder().withId(event.getId()).withMessageName(event.getMessageName()).withSeverity(event.getSeverity())
                .withTime(event.getTime()).withEvent(event.getEvent()).withEmsSeverity(event.getEmsSeverity())
                .withSource(event.getSource()).withVersion(event.getVersion()).withNodeUuid(event.getNodeUuid()).withClusterUuid(event.getClusterUuid())
                .withParameters(event.getParameters()).build();
    }
}
