/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.originator;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.platform.acq.interfaces.data.PackagePollCompleteStatus;
import com.netapp.oci.platform.acq.interfaces.data.PackagePollSkippedStatus;
import com.netapp.oci.platform.acq.interfaces.data.PackagePollStartStatus;
import com.netapp.oci.platform.acq.interfaces.data.PackagePollStatus;
import com.netapp.oci.platform.common.jms.JmsUtils;
import com.netapp.oci.platform.common.table.TableException;
import com.onaro.commons.util.CollectionPrinter;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;

import javax.jms.*;
import javax.naming.Context;
import javax.naming.InitialContext;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * @author Iti Behari
 *         Date: 2/4/13
 */
public class JmsSender implements Serializable {
    private static final long serialVersionUID = 1L;

    // JNDI names.
    private static final String TOPIC_CONNECTION_FACTORY = "java:/ConnectionFactory";
    private static final String NOTIFICATION_TOPIC = "java:/topic/NotificationTopic";
    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(JmsSender.class);

    private javax.jms.Connection jmsConnection;
    private Session session;
    private MessageProducer sender;

    public static final JmsSender INSTANCE = new JmsSender();

    private static abstract class JmsMessagePart {
        protected String key;
        protected Object data;

        private JmsMessagePart(String key, Object data) {
            this.key = key;
            this.data = data;
        }

        public static JmsMessagePart addString(String key, String data) {
            return new JmsMessagePart(key, data){
                @Override
                public void add(MapMessage message) throws JMSException {
                    message.setString(key, (String) data);
                }
            };
        }

        public static JmsMessagePart addLong(String key, long data) {
            return new JmsMessagePart(key, data ){
                @Override
                public void add(MapMessage message) throws JMSException {
                    message.setLong(key, (Long) data);
                }
            };
        }

        public static JmsMessagePart addInt(String key, int data) {
            return new JmsMessagePart(key, data ){
                @Override
                public void add(MapMessage message) throws JMSException {
                    message.setInt(key, (Integer) data);
                }
            };
        }

        public static JmsMessagePart addBoolean(String key, boolean data) {
            return new JmsMessagePart(key, data ){
                @Override
                public void add(MapMessage message) throws JMSException {
                    message.setBoolean(key, (Boolean) data);
                }
            };
        }

        @Override
        public String toString() {
            return key + "=" + data;
        }

        public abstract void add(MapMessage message) throws JMSException;
    }

    private static abstract class NotificationMessage {
        public abstract MapMessage message(Session session) throws JMSException;

        public enum MessageType {
            CHANGE,
            PACKAGE_NOTIFICATION,
            PERIODIC_PURGE,
            SUMMARY_COMPLETE
        }

        public enum NotificationType {
            PACKAGE_COMPLETED,
            PACKAGE_STARTED,
            PACKAGE_SKIPPED,
            PACKAGE_HISTORICAL_STARTED,
            PACKAGE_HISTORICAL_COMPLETED
        }

        public enum Key {
            // Header keys
            HEADER_TYPE("MessageType"),
            HEADER_DATASOURCE_ID("DatasourceID"),
            HEADER_DATASOURCE_NAME("DatasourceName"),

            // Change message keys
            CHANGE_CLUSTER_ID("ClusterId"),
            CHANGE_UPDATE_TIME("UpdateTime"),
            CHANGE_CHANGE_TYPE("ChangeType"),

            //
            SUMMARY_START_TIME("StartTime"),
            SUMMARY_SUCCESS( "Success" ),
            SUMMARY_DURATION( "DurationTime" ),
            SUMMARY_TIME( "SummaryTime" ),
            SUMMARY_ERROR( "Error" ),

            // Package Notification message keys
            PACKAGE_NOTIFICATION_TYPE("NotificationType"),
            PACKAGE_PACKAGE_NAME("PackageName"),
            PACKAGE_SUCCESS("Success"),
            PACKAGE_ERROR_COUNT("ErrorCount"),
            PACKAGE_POLL_START_TIME("PollStartTime"),
            PACKAGE_POLL_SAMPLE_TIME("SampleTime"),
            PACKAGE_DURATION_TIME("DurationTime"),
            PACKAGE_UPDATE_TIME("UpdateTime"),
            PACKAGE_POLL_INTERVAL("PollInterval"),
            PACKAGE_TOTAL_ZAPI_TIME("TotalZAPITime"),
            PACKAGE_TOTAL_REPORT_TIME("TotalReportTime");

            private String key;

            private Key(String key) {
                this.key = key;
            }

            public String key() {
                return key;
            }
        }
    }

    private JmsSender() {
        this.jmsConnection = null;
        this.session = null;
        this.sender = null;
    }

    public void sendChangeNotification(final long datasourceId, final String datasourceName, final long clusterId,
                                       final long updateTime, final ChangeType changeType) throws ServerInterfaceException {
        // notify the OCIE client about these changes through JMS.
        List<JmsMessagePart> messageParts = new ArrayList<JmsMessagePart>();
        // header part
        messageParts.add(JmsMessagePart.addString(NotificationMessage.Key.HEADER_TYPE.key(), NotificationMessage.MessageType.CHANGE.name()));
        messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.HEADER_DATASOURCE_ID.key(), datasourceId));
        messageParts.add(JmsMessagePart.addString(NotificationMessage.Key.HEADER_DATASOURCE_NAME.key(), datasourceName));

        // message specific part
        messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.CHANGE_CLUSTER_ID.key(), clusterId));
        messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.CHANGE_UPDATE_TIME.key(), updateTime));
        messageParts.add(JmsMessagePart.addString(NotificationMessage.Key.CHANGE_CHANGE_TYPE.key(), changeType.name()));

        performNotification("JMS Change", messageParts);
    }

    public void sendPurgeJobNotification(final long updateTime) throws ServerInterfaceException {
        // notify the OCIE client about these changes through JMS.
        List<JmsMessagePart> messageParts = new ArrayList<JmsMessagePart>();
        // header part
        messageParts.add(JmsMessagePart.addString( NotificationMessage.Key.HEADER_TYPE.key(), NotificationMessage.MessageType.PERIODIC_PURGE.name() ));
        messageParts.add(JmsMessagePart.addLong( NotificationMessage.Key.CHANGE_UPDATE_TIME.key(), updateTime ));

        performNotification("JMS Periodic Purge", messageParts);
    }

    private synchronized void performNotification(String description, List<JmsMessagePart> messageParts) throws ServerInterfaceException {
        try {
            // setup queue if not already there
            if (jmsConnection == null) {
                Context context = new InitialContext();
                TopicConnectionFactory connectionFactory = (TopicConnectionFactory) context.lookup(TOPIC_CONNECTION_FACTORY);
                Topic notificationTopic = (Topic) context.lookup(NOTIFICATION_TOPIC);

                jmsConnection = connectionFactory.createConnection();
                session = jmsConnection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                sender = session.createProducer(notificationTopic);
            }

            // build and send message
            MapMessage message = session.createMapMessage();
            for (JmsMessagePart messagePart : messageParts) {
                messagePart.add(message);
            }
            sender.send(message);
            if (logger.isDebugEnabled()) {
                logger.debug(dumpMessage(description, messageParts));
            }
        } catch (Exception e) {
            // destroy session
            JmsUtils.close(jmsConnection);
            jmsConnection = null;
            session = null;
            sender = null;
            throw new ServerInterfaceException("Error while sending notification" + dumpMessage(description, messageParts));
        }
    }

    public void sendPackageNotification(java.sql.Connection connection, final PackagePollStatus pollStatus) throws ServerInterfaceException {
        final long updateTime = getUpdateTime(connection, pollStatus.getDatasourceId(), pollStatus.getPollStartTime(), pollStatus.getPackageName());

        List<JmsMessagePart> messageParts = new ArrayList<JmsMessagePart>();
        // header part
        messageParts.add(JmsMessagePart.addString(NotificationMessage.Key.HEADER_TYPE.key(), NotificationMessage.MessageType.PACKAGE_NOTIFICATION.name()));
        messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.HEADER_DATASOURCE_ID.key(), pollStatus.getDatasourceId()));
        messageParts.add(JmsMessagePart.addString(NotificationMessage.Key.HEADER_DATASOURCE_NAME.key(), pollStatus.getDatasourceName()));

        NotificationMessage.NotificationType notificationType;
        if (pollStatus instanceof PackagePollCompleteStatus && !pollStatus.getIsHistorical()) {
            notificationType = NotificationMessage.NotificationType.PACKAGE_COMPLETED;
        } else if (pollStatus instanceof PackagePollStartStatus && !pollStatus.getIsHistorical()) {
            notificationType = NotificationMessage.NotificationType.PACKAGE_STARTED;
        } else if (pollStatus instanceof PackagePollSkippedStatus) {
            notificationType = NotificationMessage.NotificationType.PACKAGE_SKIPPED;
        } else if (pollStatus instanceof PackagePollStartStatus && pollStatus.getIsHistorical()) {
            notificationType = NotificationMessage.NotificationType.PACKAGE_HISTORICAL_STARTED;
        } else if (pollStatus instanceof PackagePollCompleteStatus && pollStatus.getIsHistorical())  {
            notificationType = NotificationMessage.NotificationType.PACKAGE_HISTORICAL_COMPLETED;
        } else {
            throw new RuntimeException("Could not determine package notification type from " + pollStatus.getClass());
        }

        //Common message specific parts.
        messageParts.add(JmsMessagePart.addString(NotificationMessage.Key.PACKAGE_NOTIFICATION_TYPE.key(), notificationType.name()));
        messageParts.add(JmsMessagePart.addString(NotificationMessage.Key.PACKAGE_PACKAGE_NAME.key(), pollStatus.getPackageName()));
        messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.PACKAGE_POLL_START_TIME.key(), pollStatus.getPollStartTime()));

        // We add the pollInterval to the performance package only
        if(PackagePollStatus.NETAPP_STORAGEPERFORMANCE.equals(pollStatus.getPackageName())) {
            messageParts.add(JmsMessagePart.addInt(NotificationMessage.Key.PACKAGE_POLL_INTERVAL.key(), pollStatus.getPollInterval()));
        }


        // SampleTime is the timestamp obtained from the Archiver and is set only for PACKAGE_COMPLETED or PACKAGE_HISTORICAL_COMPLETED
        // and for performance.bdis
        if((NotificationMessage.NotificationType.PACKAGE_COMPLETED.equals(notificationType) ||
                NotificationMessage.NotificationType.PACKAGE_HISTORICAL_COMPLETED.equals(notificationType)) &&
                PackagePollStatus.NETAPP_STORAGEPERFORMANCE.equals(pollStatus.getPackageName()))
            messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.PACKAGE_POLL_SAMPLE_TIME.key(), pollStatus.getSampleTime()));

        // burt1086211 add zapi and db report times for foundation polls for PACKAGE_COMPLETED
        if(NotificationMessage.NotificationType.PACKAGE_COMPLETED.equals(notificationType) &&
                PackagePollStatus.NETAPP_FOUNDATION.equals(pollStatus.getPackageName())) {
            PackagePollCompleteStatus completeStatus = (PackagePollCompleteStatus) pollStatus;
            messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.PACKAGE_TOTAL_ZAPI_TIME.key(), completeStatus.getZapiTime()));
            messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.PACKAGE_TOTAL_REPORT_TIME.key(), completeStatus.getReportTime()));
        }


        //Specific information for package complete
        if (NotificationMessage.NotificationType.PACKAGE_COMPLETED.equals(notificationType) || NotificationMessage.NotificationType.PACKAGE_HISTORICAL_COMPLETED.equals(notificationType)) {
            PackagePollCompleteStatus completeStatus = (PackagePollCompleteStatus) pollStatus;
            List<Map<String, String>> errors = completeStatus.getErrors();
            messageParts.add(JmsMessagePart.addBoolean(NotificationMessage.Key.PACKAGE_SUCCESS.key(), completeStatus.isSuccess()));
            messageParts.add(JmsMessagePart.addInt(NotificationMessage.Key.PACKAGE_ERROR_COUNT.key(), errors.size()));
            messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.PACKAGE_DURATION_TIME.key(), completeStatus.getDurationTime()));
            messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.PACKAGE_UPDATE_TIME.key(), updateTime));

            // encode errors: ErrorX_key
            for (int i = 0; i < errors.size(); i++) {
                Map<String, String> keyToValue = errors.get(i);
                String prefix = "Error" + i + '_';
                for (Map.Entry<String, String> entry : keyToValue.entrySet()) {
                    messageParts.add(JmsMessagePart.addString(prefix + entry.getKey(), entry.getValue()));
                }
            }
        }

        performNotification("JMS PackageNotification", messageParts);
    }

  public void sendSummaryNotification( final SummaryInfo summaryInfo ) throws ServerInterfaceException {
    // notify the OCIE client about these changes through JMS.
    List<JmsMessagePart> messageParts = new ArrayList<>();
    // header part
    messageParts.add(JmsMessagePart.addString( NotificationMessage.Key.HEADER_TYPE.key(), NotificationMessage.MessageType.SUMMARY_COMPLETE.name() ));
    messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.HEADER_DATASOURCE_ID.key(), summaryInfo.getDataSourceId()));
    messageParts.add(JmsMessagePart.addString(NotificationMessage.Key.HEADER_DATASOURCE_NAME.key(), summaryInfo.getDataSourceName()));

    // message specific part
    messageParts.add(JmsMessagePart.addBoolean( NotificationMessage.Key.SUMMARY_SUCCESS.key(), summaryInfo.isSuccess() ));
    messageParts.add(JmsMessagePart.addLong( NotificationMessage.Key.SUMMARY_START_TIME.key(), summaryInfo.getStartTime() ));
    messageParts.add(JmsMessagePart.addLong(NotificationMessage.Key.SUMMARY_DURATION.key(), summaryInfo.getDuration()));
    messageParts.add(JmsMessagePart.addLong( NotificationMessage.Key.SUMMARY_TIME.key(), summaryInfo.getSummaryTime() ));
    if( !summaryInfo.isSuccess() )
      messageParts.add(JmsMessagePart.addString(NotificationMessage.Key.SUMMARY_ERROR.key(), summaryInfo.getErrorMessage()));

    performNotification("JMS Change", messageParts);
  }

  private long getUpdateTime(java.sql.Connection connection, Long datasourceId, long pollStartTime, String packageName) throws ServerInterfaceException {
        try {
            long updateTime = -1;
            if (packageName.equals(PackagePollStatus.NETAPP_FOUNDATION)) {
                updateTime = NetAppModel.INSTANCE.getLastDatasourceUpdateTime(connection, datasourceId.longValue(), pollStartTime);
            } else if (packageName.equals(PackagePollStatus.NETAPP_STORAGEPERFORMANCE)) {
                updateTime = pollStartTime;
            }

            if (updateTime == -1) {
                updateTime = System.currentTimeMillis();
            }
            return updateTime;
        } catch(TableException e) {
            throw new ServerInterfaceException("Error sending JMS notification", e);
        }
    }

    private String dumpMessage(String description, List<JmsMessagePart> messageParts) {
        return description + ": {" + new CollectionPrinter<JmsMessagePart>(", ").toStringNoSort(messageParts) + "}";
    }
}
