/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.originator;

import java.io.Serializable;
import java.sql.Connection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;

import com.netapp.oci.netapp.Param;
import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.ems.ClusterUpgradeHandlerForEms;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.originator.AttributeChangeDetail;
import com.netapp.oci.platform.originator.OriginatorChanges;

/**
 * Helper class for <code>NetAppPlugin</code> to perform actions specific to this plugin whenever there are changes
 * to configuration objects because of originator updates.  This plugin handles those updates by performing a JMS
 * notification as well updates database with information about the changes to configuration objects.
 *
 * @author Shan Ponnusamy
 */
public class OriginatorChangesHandler implements Serializable {

    private static final long serialVersionUID = 1L;

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(OriginatorChangesHandler.class);
    private static final String CLUSTER_XML_TAG_NAME = "Cluster";
    private static final String VERSION = "version";


    public OriginatorChangesHandler() {

    }

    public void logOriginatorChangesInfo(Connection connection, final OriginatorChanges originatorChanges) throws ServerInterfaceException {
        try {
            final ChangeType clusterChangeType = getChangeType(originatorChanges);
            final long clusterId = getClusterId(connection, originatorChanges, clusterChangeType);

            // persist the change and change detail (if any) information in the database.
            persistChanges(connection, originatorChanges, clusterId, clusterChangeType);
            if (isClusterVersionChange(clusterChangeType, originatorChanges)) {
                //TODO - Remove this later after one round of testing
                logger.error("Cluster upgrade event caught for EMS handler");
                ClusterUpgradeHandlerForEms.INSTANCE.handleClusterUpgrade(connection, originatorChanges.getAttributeChangeDetails(), clusterId);
            }

        } catch (TableException e) {
            throw new ServerInterfaceException("Error while persisting originator changes", e);
        }
    }

    public void sendChangeNotification(Connection connection, final long datasourceId, final String datasourceName, final OriginatorChanges originatorChanges) throws ServerInterfaceException {
        try {
            final ChangeType clusterChangeType = getChangeType(originatorChanges);
            final long clusterId = getClusterId(connection, originatorChanges, clusterChangeType);
            // notify the OCIE client about these changes through JMS.
            JmsSender.INSTANCE.sendChangeNotification(datasourceId, datasourceName, clusterId, originatorChanges.updateTime(), clusterChangeType);
        } catch (TableException e) {
            throw new ServerInterfaceException("Error while sending change notification", e);
        }
    }


    private long getClusterId(Connection connection, OriginatorChanges changes, ChangeType clusterChangeType) throws TableException {
        // Try to get the cluster id from the changes object, if it is available.  If not, get it from DB. There are two reasons for this.
        // 1. cluster id will not be available in DB if the cluster was already deleted.
        // 2. It will be a minor optimization of not hitting the database if the cluster id is already available in the changes obj.
        if (clusterChangeType == ChangeType.UPDATED) {
            return NetAppModel.INSTANCE.getClusterIdByOriginatorId(connection, changes.originId());
        } else {
            return getClusterId(changes.addedObjs(), changes.updatedObjs(), changes.deletedObjs(), changes.deletedHistoryObjs());
        }
    }

    private long getClusterId(Map<String, Set<Long>>... changeMaps) {
        for (Map<String, Set<Long>> changeMap : changeMaps) {
            Set<Long> clusterIds = changeMap.get(CLUSTER_XML_TAG_NAME);
            if (clusterIds != null) {
                return clusterIds.iterator().next();
            }
        }
        return -1;
    }


    private ChangeType getChangeType(OriginatorChanges originatorChanges) {
        Set<Long> addedClusterIds = originatorChanges.addedObjs().get(CLUSTER_XML_TAG_NAME);
        if (addedClusterIds != null) {
            return ChangeType.ADDED;
        } else {
            Set<Long> deletedClusterIds = originatorChanges.deletedObjs().get(CLUSTER_XML_TAG_NAME);
            if (deletedClusterIds != null) {
                return ChangeType.DELETED;
            } else {
                return ChangeType.UPDATED;
            }
        }
    }

    private void persistChanges(Connection connection, OriginatorChanges originatorChanges, long clusterId,
                                ChangeType clusterChangeType) throws TableException {
        final OriginatorChangesTable originatorChangesTable = new OriginatorChangesTable();
        updateChangeTable(originatorChangesTable, originatorChanges, clusterId, clusterChangeType, connection);

        // if there are attribute change details, update 'changed_attribute' table.
        Set<AttributeChangeDetail> attributeChangeDetails = originatorChanges.getAttributeChangeDetails();
        if (attributeChangeDetails.size() > 0) {
            originatorChangesTable.insertChangedAttributeBatch(connection, originatorChanges.getAttributeChangeDetails(), originatorChanges.updateTime());
        }
    }

    private boolean isClusterVersionChange(ChangeType clusterChangeType, OriginatorChanges originatorChanges) {
        if (clusterChangeType == ChangeType.UPDATED) {
            Set<AttributeChangeDetail> attributeList = originatorChanges.getAttributeChangeDetails();
            for (AttributeChangeDetail attribute : attributeList) {
                if (attribute.getAttribute().contains(VERSION)) {
                    return true;
                }
            }
        }
        return false;
    }

    private void updateChangeTable(final OriginatorChangesTable originatorChangesTable, OriginatorChanges originatorChanges, long clusterId,
                                   ChangeType clusterChangeType, Connection connection)
            throws TableException {
        boolean recordClusterChange = Param.get(Param.Key.RECORD_FULL_CLUSTER_CHANGE, boolean.class, false);

        // This is an optimization. In case of a cluster add/delete, we will only insert cluster add/delete info.
        //  All other changes will be discarded.  The recordClusterChange variable defaults to false.   If manually
        // set to true all cluster add/delete chanegs are recorded.
        if ((clusterChangeType == ChangeType.ADDED || clusterChangeType == ChangeType.DELETED) && !recordClusterChange) {
            originatorChangesTable.insertChangedRecord(connection, clusterChangeType, clusterId, originatorChanges.updateTime(), CLUSTER_XML_TAG_NAME, clusterId);
        } else {   // no cluster add or delete
            Map<ChangeType, Map<String, Set<Long>>> changedObjs = new HashMap<ChangeType, Map<String, Set<Long>>>();
            changedObjs.put(ChangeType.ADDED, originatorChanges.addedObjs());
            changedObjs.put(ChangeType.UPDATED, originatorChanges.updatedObjs());
            changedObjs.get(ChangeType.UPDATED).putAll(originatorChanges.deletedHistoryObjs());
            changedObjs.put(ChangeType.DELETED, originatorChanges.deletedObjs());

            originatorChangesTable.insertChangedRecordBatch(connection, changedObjs, clusterId, originatorChanges.updateTime());
        }

    }

    public void logOriginatorRemove(final Connection connection, long originId, long clusterId, long updateTime) throws ServerInterfaceException {
        try {
            final OriginatorChangesTable originatorChangesTable = new OriginatorChangesTable();
            originatorChangesTable.insertChangedRecord(connection, ChangeType.DELETED, clusterId, updateTime, CLUSTER_XML_TAG_NAME, clusterId);
            originatorChangesTable.insertChangedRecord(connection, ChangeType.PURGED, clusterId, updateTime, CLUSTER_XML_TAG_NAME, clusterId);
        } catch (TableException e) {
            throw new ServerInterfaceException("Error while logging cluster delete change", e);
        }

    }

}
