/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.originator;

import com.netapp.oci.netapp.startup.NetAppPlugin;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.originator.AttributeChangeDetail;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Map;
import java.util.Set;

/**
 * @author Iti Behari
 *         Date: 2/1/13
 */
public class OriginatorChangesTable {

    private static final String CHANGE_TABLE_INSERT_SQL = "INSERT INTO " + NetAppPlugin.CHANGE_RECORD_TABLE +
            "( updateTime, clusterId, objId, objectType, changeType) VALUES (?,?,?,?,?)";


    public OriginatorChangesTable() {
    }

    public void insertChangedRecord(Connection connection, ChangeType changeType, long clusterId, long updateTime, String xmlTagName, Long objId) throws TableException {
        PreparedStatementWrapper preparedStatement =
                PreparedStatementWrapper.createForUpdate(connection, CHANGE_TABLE_INSERT_SQL);
        prepareChangeRow(preparedStatement, changeType, updateTime, clusterId, xmlTagName, objId);
        preparedStatement.executeUpdateEx();
    }

    public void insertChangedRecordBatch(Connection connection, Map<ChangeType, Map<String, Set<Long>>> changedObjs, long clusterId, long updateTime)
            throws TableException {
        PreparedStatementWrapper preparedStatement = PreparedStatementWrapper.createForUpdate(connection,
                CHANGE_TABLE_INSERT_SQL);
        try {
            for (Map.Entry<ChangeType, Map<String, Set<Long>>> entry : changedObjs.entrySet()) {
                addToChangeTableBatch(preparedStatement, entry.getKey(), entry.getValue(), clusterId, updateTime);
            }
        }
        catch (SQLException e) {
            throw new TableException("Failed to executeUpdate: " + this, e);
        }
        finally {
            preparedStatement.closePreparedStatement();
        }
    }

    public void insertChangedAttributeBatch(Connection connection, Set<AttributeChangeDetail> attributeChangeDetails, long updateTime) throws TableException {
        String changedAttributeTableInsertSql = "INSERT INTO " + NetAppPlugin.CHANGED_ATTRIBUTE_TABLE +
                " (updateTime, objId, attribute, oldValue, newValue) VALUES (?,?,?,?,?)";
        PreparedStatementWrapper preparedStatement = PreparedStatementWrapper.createForUpdate(connection,
                changedAttributeTableInsertSql);

        try {
            for (AttributeChangeDetail attributeChangeDetail : attributeChangeDetails) {
                prepareChangedAttributeRow(preparedStatement, attributeChangeDetail, updateTime);
                preparedStatement.executeUpdate();
            }
        }
        catch (SQLException e) {
            throw new TableException("Failed to executeUpdate: " + this, e);
        }
        finally {
            preparedStatement.closePreparedStatement();
        }
    }

    private void prepareChangedAttributeRow(PreparedStatementWrapper preparedStatement, AttributeChangeDetail attributeChangeDetail,
                                            long updateTime) throws TableException {
        preparedStatement.setNext(updateTime);
        preparedStatement.setNext(attributeChangeDetail.getObjId());
        preparedStatement.setNext(attributeChangeDetail.getAttribute());
        preparedStatement.setNext(toString(attributeChangeDetail.getOldValue()));
        preparedStatement.setNext(toString(attributeChangeDetail.getNewValue()));
    }

    private String toString(Object obj) {
        return obj == null ? null : obj.toString();
    }

    private void addToChangeTableBatch(PreparedStatementWrapper preparedStatement, ChangeType changeType,
                                       Map<String, Set<Long>> changedObjs, long clusterId, long updateTime) throws TableException, SQLException {
            for (String xmlTagName : changedObjs.keySet()) {
                for (Long objId : changedObjs.get(xmlTagName)) {
                    prepareChangeRow(preparedStatement, changeType, updateTime, clusterId, xmlTagName, objId);
                    preparedStatement.executeUpdate();
                }
            }

    }

    private void prepareChangeRow(PreparedStatementWrapper preparedStatement, ChangeType changeType, long updateTime,
                                  long clusterId, String xmlTagName, Long objId) throws TableException {
        preparedStatement.setNext(updateTime);
        preparedStatement.setNext(clusterId);
        preparedStatement.setNext(objId);
        preparedStatement.setNext(getResourceTypeFromTagname(xmlTagName));
        preparedStatement.setNext(changeType.name());
    }

    /**
     * Returns the value that should be used in the changes.objType column
     * for the given xmlTagName.
     * <p/>
     * <p/>
     * For example, getResourceTypeFromTagname("Cluster") will return "cluster".
     */
    private String getResourceTypeFromTagname(String xmlTagName) {
        String resourceType = Registry.INSTANCE.getResourceType(xmlTagName);
        return resourceType == null ? xmlTagName.toLowerCase() : resourceType;
    }
}
