/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.originator;

import com.netapp.oci.netapp.startup.NetAppPlugin;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.netapp.oci.server.TransactionContext;
import com.netapp.oci.server.UpdateTask;
import com.netapp.oci.server.UpdateTaskException;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import org.slf4j.Logger;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

/**
 * Prunes the history and change related tables in "netapp" schema
 *
 * @author Shan Ponnusamy
 */
public class PruneTask extends UpdateTask {

    private static final long serialVersionUID = 1L;

    /**
     * SQL statement template to delete expired 'history' records.
     */
    private static final String[] CHANGE_TABLES = new String[]{NetAppPlugin.CHANGE_RECORD_TABLE, NetAppPlugin.CHANGED_ATTRIBUTE_TABLE};

    private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger(PruneTask.class);
    public static final int PRUNE_TASK_ORIGIN_ID = -1000;


    /**
     * Life time for a history record in milliseconds
     */
    private final long historyRecordLifeTime;

    /**
     * Life time for a change record in milliseconds
     */
    private final long changesRecordLifeTime;
    private int allChangesCount;


    public PruneTask(long historyRecordLifeTime, long changesRecordLifeTime) throws UpdateTaskException {
        super(":prune:", -42L);
        this.historyRecordLifeTime = historyRecordLifeTime;
        this.changesRecordLifeTime = changesRecordLifeTime;
    }

    @Override
    public boolean run(TransactionContext transactionContext) throws UpdateTaskException {
        long updateTime = transactionContext.getTime();
        Connection connection = transactionContext.getConnection();
        try {
            prune(connection, "DELETE FROM %s WHERE updateTime < ?", Arrays.asList(CHANGE_TABLES), getCutOffTime(updateTime, changesRecordLifeTime));
            prune(connection, "DELETE FROM %s WHERE deleteTime < ?", Collections.singleton(NetAppPlugin.CHANGE_DELETE_DETAIL_TABLE), getCutOffTime(updateTime, changesRecordLifeTime));
            allChangesCount = pruneHistoryRecords(connection, updateTime);
            return true;
        } catch (Exception e) {
            throw new UpdateTaskException("Failed to run pruneTask", e);
        }
    }


    private void prune(Connection connection, String deleteSqlTemplate, Iterable<String> tables, long cutOffTime) throws TableException {
        for (String table : tables) {
            prune(connection, table, deleteSqlTemplate, cutOffTime);
        }
    }

    private void prune(Connection connection, String table, String deleteSqlTemplate, long cutOffTime) throws TableException {
        String sql = String.format(deleteSqlTemplate, table);
        PreparedStatementWrapper preparedStatement = PreparedStatementWrapper.createForUpdate(connection, sql);
        preparedStatement.setNext(cutOffTime);
        int noOfDeletes = preparedStatement.executeUpdateEx();
        if (logger.isDebugEnabled()) {
            logger.debug("Deleted " + noOfDeletes + " rows in '" + table + "' table");
        }
    }

    public int pruneHistoryRecords(Connection connection, long updateTime) throws TableException, ServerInterfaceException {
        final Map<String, String> tagNameToHistoryTables = Registry.INSTANCE.getHistoryTables();
        final long cutOffTime = getCutOffTime(updateTime, historyRecordLifeTime);

        Map<Long, List<HistoryRecord>> historyObjectsToBeDeleted = findHistoryRecordsToBeDeleted(connection, tagNameToHistoryTables, cutOffTime);

        final int updatedClusterIdsCount = historyObjectsToBeDeleted.size();
        if (updatedClusterIdsCount > 0) {
            // prune history records
            final String deleteSqlTemplate = "DELETE FROM %s WHERE inactive=true AND updateTime < ?";
            for (Map.Entry<String, String> entry : tagNameToHistoryTables.entrySet()) {
                prune(connection, entry.getValue(), deleteSqlTemplate, cutOffTime);
            }

            // log changeRecords for each deleted history record
            final OriginatorChangesTable originatorChangesTable = new OriginatorChangesTable();
            insertChangedRecord(connection, updateTime, originatorChangesTable, historyObjectsToBeDeleted);

            // send JMS for each updated cluster
            JmsSender.INSTANCE.sendPurgeJobNotification(updateTime);
        }
        return updatedClusterIdsCount;
    }



    /**
     * @param connection
     * @param tagNameToHistoryTables
     * @param cutOffTime
     * @return Map<ClusterId, List<HistoryRecord>>
     * @throws TableException
     */
    private Map<Long, List<HistoryRecord>> findHistoryRecordsToBeDeleted(Connection connection, Map<String, String> tagNameToHistoryTables, long cutOffTime) throws TableException {
        final String findSqlTemplate = "SELECT clusterId, originId, objId FROM %s WHERE inactive=true AND updateTime < ?";
        Map<Long, List<HistoryRecord>> historyObjectsByClusterId = new HashMap<Long, List<HistoryRecord>>();
        for (Map.Entry<String, String> entry : tagNameToHistoryTables.entrySet()) {
            String tableName = entry.getValue();
            List<HistoryRecord> historyObjectsToBeDeleted = findHistoryObjectsToBeDeletedByTable(connection, cutOffTime, findSqlTemplate, tableName, entry.getKey());
            if (historyObjectsToBeDeleted.size() > 0) {
                for (HistoryRecord historyObject : historyObjectsToBeDeleted) {
                    if (historyObjectsByClusterId.containsKey(historyObject.clusterId)) {
                        historyObjectsByClusterId.get(historyObject.clusterId).add(historyObject);
                    } else {
                        List<HistoryRecord> list = new ArrayList<HistoryRecord>();
                        list.add(historyObject);
                        historyObjectsByClusterId.put(historyObject.clusterId, list);
                    }
                }
            }
        }
        return historyObjectsByClusterId;
    }

    private void insertChangedRecord(Connection connection, long currentTime, OriginatorChangesTable originatorChangesTable, Map<Long, List<HistoryRecord>> historyObjects) throws TableException {

        for (Map.Entry<Long, List<HistoryRecord>> historyObject : historyObjects.entrySet()) {
            Map<String, Set<Long>> recordsToLog = new HashMap<String, Set<Long>>();
            Long clusterId = historyObject.getKey();
            List<HistoryRecord> historyObjectsByClusterId = historyObject.getValue();
            for (HistoryRecord historyObjectByClusterId : historyObjectsByClusterId) {
                if (recordsToLog.containsKey(historyObjectByClusterId.tagName)) {
                    recordsToLog.get(historyObjectByClusterId.tagName).add(historyObjectByClusterId.objId);
                } else {
                    Set<Long> list = new HashSet<Long>();
                    list.add(historyObjectByClusterId.objId);
                    recordsToLog.put(historyObjectByClusterId.tagName, list);

                }
            }
            Map<ChangeType, Map<String, Set<Long>>> recordsToLogByChangeLog = new HashMap<ChangeType, Map<String, Set<Long>>>();
            recordsToLogByChangeLog.put(ChangeType.DELETED, recordsToLog);
            originatorChangesTable.insertChangedRecordBatch(connection, recordsToLogByChangeLog, clusterId, currentTime);

        }
    }

    private List<HistoryRecord> findHistoryObjectsToBeDeletedByTable(Connection connection, long cutOffTime, String findSqlTemplate, String tableName, final String tagName) throws TableException {
        TableResultParserDefault<HistoryRecord> parser = new TableResultParserDefault<HistoryRecord>() {
            @Override
            public HistoryRecord parseObject(ResultSet resultSet) throws SQLException {
                return new HistoryRecord(tagName, resultSet.getLong(1), resultSet.getLong(2), resultSet.getLong(3));
            }
        };
        return PreparedStatementWrapper.create(connection, String.format(findSqlTemplate, tableName), parser).setNext(cutOffTime).executeQueryEx();
    }


    private long getCutOffTime(long currentTime, long recordLifeTime) {
        return currentTime - recordLifeTime;
    }

    public int getAllChangesCount() {
        return allChangesCount;
    }

    private class HistoryRecord {
        String tagName;
        Long originId;
        Long objId;
        Long clusterId;

        private HistoryRecord(String tagName, Long clusterId, Long originId, Long objId) {
            this.tagName = tagName;
            this.originId = originId;
            this.objId = objId;
            this.clusterId = clusterId;
        }

    }

}
