package com.netapp.oci.netapp.originator.handler;

import java.util.EnumSet;
import java.util.Map;

import com.netapp.oci.platform.originator.ObjState;
import com.netapp.oci.platform.originator.OriginatorUpdate;
import com.netapp.oci.platform.originator.OriginatorUpdateException;
import com.netapp.oci.platform.originator.handler.ColumnAttr;
import com.netapp.oci.platform.originator.handler.HandlerResult;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.key.KeyCacheException;
import com.netapp.oci.platform.originator.key.UuidKey;

/**
 * @author Iti Behari
 *         Date: 8/12/13
 */
public class LifHandler extends XmlHandler {

    public static enum SubTag {

        NETWORK_LIF("NetworkLIF", "netapp_model.network_lif"),
        FCP_LIF("FCPLIF", "netapp_model.fcp_lif"),
        NVMF_FC_LIF("NvmfFcLif", "netapp_model.nvmf_fc_lif");

        private final String tagName;
        private String tableName;

        SubTag(final String tagName, final String tableName) {
            this.tagName = tagName;
            this.tableName = tableName;
        }

        public String getTagName() {
            return tagName;
        }

        public String getTableName() {
            return tableName;
        }
    }

    public static final String BASE_TABLE_NAME = "netapp_model.lif";
    private SubTag subTag;

    public LifHandler(final SubTag tag) {
        super(tag.getTagName(), tag.getTableName(), BASE_TABLE_NAME);
        this.subTag = tag;
    }

    @Override
    public HandlerResult parseProperties(OriginatorUpdate originatorUpdate, Map<String, String> properties,
                                                  Map<UuidKey, ObjState> keyToObjState) throws OriginatorUpdateException, KeyCacheException {
        Long clusterId = originatorUpdate.peekObjectId(1);
        Long vserverId = originatorUpdate.peekObjectId();
        Long currentNodeId = originatorUpdate.getObjectIdByTagname("Node", properties.get("CurrentNodeUUID"));
        Long homeNodeId = originatorUpdate.getObjectIdByTagname("Node", properties.get("HomeNodeUUID"));

        properties.put("LifType", subTag.name());
        HandlerResult result = new HandlerResult(originatorUpdate, this, properties, "UUID", keyToObjState, true)

                .baseColumn("Name", String.class, EnumSet.of(ColumnAttr.REQUIRED))
                .baseColumn("ClusterId", clusterId)
                .baseColumn("VserverId", vserverId)
                .baseColumn("CurrentNodeId", currentNodeId)
                .baseColumn("LifType", String.class)
                .baseColumn("HomeNodeId", homeNodeId)
                .baseColumn("IsAutoRevert", boolean.class)
                .baseColumn("IsHome", boolean.class)
                .baseColumn("OperationalStatus", Enum.class)
                .baseColumn("OperationalStatusRaw", String.class)
                .baseColumn("AdministrativeStatus", Enum.class)
                .baseColumn("AdministrativeStatusRaw", String.class)
                .baseColumn("Role", Enum.class)
                .baseColumn("RoleRaw", String.class)
                .baseColumn("DataProtocols", String.class)
                .baseColumn("FailoverGroup", String.class);

        if (SubTag.FCP_LIF == subTag) {
            parseFcpLifProperties(originatorUpdate, properties, clusterId, vserverId,  result);
        } else if (SubTag.NVMF_FC_LIF == subTag) {
            parseNvmfFcLifProperties(originatorUpdate, properties, clusterId, vserverId,  result);
        } else {
            parseNetworkLifProperties(originatorUpdate, properties, clusterId, vserverId, result);
        }

        return result;
    }

    private void parseNetworkLifProperties(OriginatorUpdate originatorUpdate, Map<String, String> properties, Long clusterId, Long vserverId, HandlerResult result) throws OriginatorUpdateException {
        Long currentPortId = originatorUpdate.getObjectIdByTagname(NetworkPortHandler.TAG, properties.get("CurrentPortUUID"));
        Long homePortId = originatorUpdate.getObjectIdByTagname(NetworkPortHandler.TAG, properties.get("HomePortUUID"));
        Long routingGroupId = originatorUpdate.getObjectIdByTagname("RoutingGroup", properties.get("RoutingGroupUUID"));
        Long firewallPolicyId = originatorUpdate.getObjectIdByTagname("NetworkFirewallPolicy", properties.get("FirewallPolicyUUID"));
        result.subColumn("ResourceKey", String.class, EnumSet.of(ColumnAttr.REQUIRED, ColumnAttr.STRING_512))
                .subColumn("ClusterId", clusterId)
                .subColumn("VserverId", vserverId)
                .subColumn("CurrentPortId", currentPortId)
                .subColumn("HomePortId", homePortId)
                .subColumn("RoutingGroupId", routingGroupId)
                .subColumn("FirewallPolicyId", firewallPolicyId)
                .subColumn("Address", String.class)
                .subColumn("Netmask", String.class)
                .subColumn("NetmaskLength", int.class)
                .subColumn("DNSDomainName", String.class)
                .subColumn("FailoverPolicy", Enum.class)
                .subColumn("FailoverPolicyRaw", String.class)
                .subColumn("IsDNSUpdateEnabled", boolean.class)
                .subColumn("ListenForDNSQuery", boolean.class);
    }

    private void parseFcpLifProperties(OriginatorUpdate originatorUpdate, Map<String, String> properties, Long clusterId, Long vserverId, HandlerResult result) throws OriginatorUpdateException {
        Long currentPortId = originatorUpdate.getObjectIdByTagname(FcpPortHandler.TAG, properties.get("CurrentPortUUID"));
        Long homePortId = originatorUpdate.getObjectIdByTagname(FcpPortHandler.TAG, properties.get("HomePortUUID"));

        result.subColumn("ResourceKey", String.class, EnumSet.of(ColumnAttr.REQUIRED, ColumnAttr.STRING_512))
                .subColumn("ClusterId", clusterId)
                .subColumn("VserverId", vserverId)
                .subColumn("CurrentPortId", currentPortId)
                .subColumn("HomePortId", homePortId)
                .subColumn("NodeName", String.class)
                .subColumn("PortName", String.class);
    }

    private void parseNvmfFcLifProperties(OriginatorUpdate originatorUpdate, Map<String, String> properties, Long clusterId, Long vserverId, HandlerResult result) throws OriginatorUpdateException {
        Long currentPortId = originatorUpdate.getObjectIdByTagname(FcpPortHandler.TAG, properties.get("CurrentPortUUID"));
        Long homePortId = originatorUpdate.getObjectIdByTagname(FcpPortHandler.TAG, properties.get("HomePortUUID"));

        result.subColumn("ResourceKey", String.class, EnumSet.of(ColumnAttr.REQUIRED, ColumnAttr.STRING_512))
                .subColumn("ClusterId", clusterId)
                .subColumn("VserverId", vserverId)
                .subColumn("CurrentPortId", currentPortId)
                .subColumn("HomePortId", homePortId)
                .subColumn("PortName", String.class);
    }
}
