/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.originator.handler;

import java.sql.Timestamp;
import java.util.EnumSet;
import java.util.Map;

import com.netapp.oci.platform.common.util.Assert;
import com.netapp.oci.platform.originator.ObjState;
import com.netapp.oci.platform.originator.OriginatorUpdate;
import com.netapp.oci.platform.originator.OriginatorUpdateException;
import com.netapp.oci.platform.originator.handler.ColumnAttr;
import com.netapp.oci.platform.originator.handler.HandlerResult;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.key.KeyCacheException;
import com.netapp.oci.platform.originator.key.UuidKey;

/**
 * @author Nancy Walia
 */
public class SnapMirrorHandler extends XmlHandler {
    public static final String TAG = "SnapMirror";
    public static final String TABLE_NAME = "netapp_model.snap_mirror";

    public SnapMirrorHandler() {
        super(TAG, TABLE_NAME);
    }

    @Override
    public HandlerResult parseProperties(OriginatorUpdate originatorUpdate, Map<String, String> properties,
                                         Map<UuidKey, ObjState> keyToObjState) throws OriginatorUpdateException, KeyCacheException {
        final boolean isVServerSnapMirror = ("VSERVER".equals(properties.get("RelationshipGroupType")) && properties.get("DestinationVolumeUUID") == null && properties.get("SourceVolume") == null );

        Long destinationClusterId = originatorUpdate.peekObjectId();
        Long destinationVserverId = originatorUpdate.getRequiredObjectIdByTagname("Vserver", properties.get("DestinationVserverUUID"), getTagName());
        Long destinationVolumeId = null;
        if(!isVServerSnapMirror) {
            destinationVolumeId = originatorUpdate.getRequiredObjectIdByTagname("Volume", properties.get("DestinationVolumeUUID"), getTagName());
        }

        // optional UUIDs
        Long sourceClusterId = null;
        Long sourceVserverId = originatorUpdate.getObjectIdByTagname("Vserver", properties.get("SourceVserverUUID"));
        Long destinationNodeId = originatorUpdate.getObjectIdByTagname("Node", properties.get("DestinationNodeUUID"));
        Long sourceVolumeId = originatorUpdate.getObjectIdByTagname("Volume", properties.get("SourceVolumeUUID"));
        Long exportedSnapshotId = originatorUpdate.getObjectIdByTagname("Snapshot", properties.get("ExportedSnapshotUUID"));
        Long newestSnapshotId = originatorUpdate.getObjectIdByTagname("Snapshot", properties.get("NewestSnapshotUUID"));
        Long snapMirrorPolicyId = originatorUpdate.getObjectIdByTagname("SnapMirrorPolicy", properties.get("SnapMirrorPolicyUUID"));
        Long jobScheduleId = originatorUpdate.getObjectIdByTagname("JobSchedule", properties.get("JobScheduleUUID"));

        // if either sourceVserverUUID & sourceVolumeUUID is set then both must be set
        if (sourceVserverId != null || sourceVolumeId != null) {
            Assert.isLegal(sourceVserverId != null && sourceVolumeId != null, "If either sourceVserverUUID & sourceVolumeUUID is set then both must be set");
        }

        // if we have a source VserverId and VolumeID, then this is a local snap mirror and the source cluster is
        // the same as the destination cluster
        if (sourceVserverId != null && sourceVolumeId != null) {
            sourceClusterId = destinationClusterId;
        }

        return new HandlerResult(originatorUpdate, this, properties, "UUID", keyToObjState)
                .column("ResourceKey", String.class, EnumSet.of(ColumnAttr.REQUIRED, ColumnAttr.STRING_512))
                .column("clusterId", destinationClusterId)
                .column("vserverId", destinationVserverId)
                .column("volumeId", destinationVolumeId)
                .column("nodeId", destinationNodeId)
                .column("DestinationLocation", String.class)
                .column("SourceClusterId", sourceClusterId)
                .column("SourceVserverId", sourceVserverId)
                .column("SourceVolumeId", sourceVolumeId)
                .column("SourceCluster", String.class)
                .column("SourceVolume", String.class)
                .column("SourceVserver", String.class)
                .column("SourceLocation", String.class)
                .column("MirrorState", Enum.class)
                .column("MirrorStateRaw", String.class)
                .column("ExportedSnapshotId", exportedSnapshotId)
                .column("NewestSnapshotId", newestSnapshotId)
                .column("NetworkCompressionRatio", String.class)
                .column("OperationId", String.class)
                .column("TransferSnapshot", String.class)
                .column("SnapMirrorPolicyId", snapMirrorPolicyId)
                .column("IsConstituent", boolean.class)
                .column("IsHealthy", boolean.class)
                .column("MaxTransferRate", long.class)
                .column("RelationshipProgress", long.class)
                .column("RelationshipStatus", Enum.class)
                .column("RelationshipStatusRaw", String.class)
                .column("RelationshipType", Enum.class)
                .column("RelationshipTypeRaw", String.class)
                .column("Tries", String.class)
                .column("UnhealthyReason", String.class)
                .column("SnapshotCheckpoint", long.class)
                .column("SnapshotProgress", long.class)
                .column("CurrentMaxTransferRate", long.class)
                .column("CurrentTransferError", String.class,EnumSet.of(ColumnAttr.STRING_1024))
                .column("CurrentTransferPriority", Enum.class)
                .column("CurrentTransferPriorityRaw", String.class)
                .column("CurrentTransferType", Enum.class)
                .column("CurrentTransferTypeRaw", String.class)
                .column("LagTime", long.class)
                .column("LastTransferDuration", long.class)
                .column("LastTransferEndTimestamp", Timestamp.class)
                .column("LastTransferError", String.class, EnumSet.of(ColumnAttr.STRING_1024))
                .column("LastTransferErrorCodes", String.class)
                .column("LastTransferNetworkCompressionRatio", String.class)
                .column("LastTransferType", Enum.class)
                .column("LastTransferTypeRaw", String.class)
                .column("LastTransferSize", long.class)
                .column("ProgressLastUpdated", Timestamp.class)
                .column("RelationshipControlPlane", String.class)
                .column("RelationshipGroupType", Enum.class)
                .column("RelationshipGroupTypeRaw", String.class)
                .column("RelationshipIdentifier", String.class)
                .column("JobScheduleId", jobScheduleId)
                .column("DestVolDerivedStyle", Enum.class)
                .column("PolicyType", Enum.class)
                .column("PolicyTypeRaw", String.class)
                .column("DerivedRelationshipType", String.class);
    }
}
