package com.netapp.oci.netapp.performance.aggregate.interfaces.impl;

import java.sql.Connection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import org.slf4j.Logger;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.aggregate.interfaces.data.AggregateMeasurements;
import com.netapp.oci.netapp.performance.aggregate.interfaces.local.AggregatePerformanceQueryLocal;
import com.netapp.oci.netapp.performance.aggregate.interfaces.remote.AggregatePerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class AggregatePerformanceQueryBean implements AggregatePerformanceQueryRemote, AggregatePerformanceQueryLocal{
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

	@Override
	public Map<Long, List<SampleInfo<AggregateMeasurements>>> getSamples(
			Set<Long> aggrIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + aggrIds.size() + "aggregate samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<AggregateMeasurements>>> containerIdToSamples = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(aggrIds);

            Map<Long, Set<Long>> clusterIdToAggrIds = NetAppModel.INSTANCE.getClusterIdByAggregateIds(connection, aggrIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToAggrIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find aggegate with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get aggegate samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query aggegate samples: " + (System.currentTimeMillis() - startTime));
        }

	}

	@Override
	public Map<Long, List<SampleInfo<AggregateMeasurements>>> getSamples(Long containerId, Set<Long> aggrIds, 
			Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + aggrIds.size() + "aggregate samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, aggrIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get aggregate samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query aggregate samples: " + (System.currentTimeMillis() - startTime));
        }
	}

	@Override
	public List<SampleInfo<AggregateMeasurements>> getSamples(Long containerId,
			Long aggrId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(aggrId), fromTime, toTime).get(aggrId);
	}

	@Override
	public Map<Long, SampleInfo<AggregateMeasurements>> getLatestSample(
			Long clusterId, Set<Long> aggrIds) throws NetappInterfaceException {

		logger.info("Query " + aggrIds.size() + " aggregate samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getAggregatePerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(aggrIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get aggregate samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
	}

	@Override
	public Map<Long, List<Summary<AggregateMeasurements>>> getSummaries(Set<Long> aggrIds,
                                                                        Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + aggrIds.size() + " aggregate summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<AggregateMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(aggrIds);

            Map<Long, Set<Long>> clusterIdToAggrIds = NetAppModel.INSTANCE.getClusterIdByAggregateIds(connection, aggrIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToAggrIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<AggregateMeasurements>> summaries = ComponentsFactoryRepository.getAggregatePerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find aggregate with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get aggregate summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
	}

	@Override
	public Map<Long, List<Summary<AggregateMeasurements>>> getSummaries(Long clusterId, Set<Long> aggrIds,
                                                                        Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + aggrIds.size() + " aggregate summaries from " + fromTime + " to " + toTime
                + "for cluster " + clusterId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<AggregateMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<AggregateMeasurements>> summaries = ComponentsFactoryRepository.getAggregatePerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), aggrIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get aggregate summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
	}

	@Override
	public Boolean areStatisticsAvailable(Long containerId, Long objId,
			Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "aggregate has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getAggregatePerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if aggregate has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
	}

    private Map<Long, List<SampleInfo<AggregateMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> aggrIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getAggregatePerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(aggrIds, new TimeSpan(fromTime, toTime)));
    }
	
}
