package com.netapp.oci.netapp.performance.cifsvserver.interfaces.impl;

import java.sql.Connection;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.db.NetAppPerformance;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.originator.handler.ClusterHandler;
import com.netapp.oci.netapp.originator.handler.VserverHandler;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.cifsvserver.interfaces.data.CifsVserverMeasurements;
import com.netapp.oci.netapp.performance.cifsvserver.interfaces.data.CifsVserverRawMeasurements;
import com.netapp.oci.netapp.performance.cifsvserver.interfaces.remote.CifsVserverPerformanceRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.shared.SimpleRawSampleConverter;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceDatasourceRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry.SampleType;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.SimpleUuidKey;
import com.onaro.sanscreen.util.Constant;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class CifsVserverPerformanceBean implements CifsVserverPerformanceRemote {
    public static final String PACKAGE_NAME = "cifsvserver";
    public static final String PACKAGE_NAME_RAW = PACKAGE_NAME + PerformanceConstants.RAW_PACKAGE_SUFFIX;
    
    private static final String CIFSVSERVER_PACKAGE = PerformanceConstants.PACKAGE_PREFIX + PACKAGE_NAME;
    private static final Logger LOG = LoggerFactory.getLogger(CIFSVSERVER_PACKAGE);

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * use this bean to register reporting datasource
     */
    @EJB
    private NetappPerformanceDatasourceRegistry datasourceRegistry;

    @EJB
    private NetappPerformanceSampleListenerRegistry sampleListenerRegistry;
    
	@Override
	public void reportCifsVserverRawData(Long datasourceId, String clusterKey, long time, Collection<RawSample<CifsVserverRawMeasurements>> rawSamples, long leaseTime) throws NetappInterfaceException {
		Connection connection = null;
        long startTime = System.currentTimeMillis();

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(CIFSVSERVER_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(CIFSVSERVER_PACKAGE, clusterKey, datasourceId, leaseTime)) {
            	LOG.info("Cluster key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(CIFSVSERVER_PACKAGE, clusterKey, leaseTime) + " (clusterKey:" + clusterKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            Long clusterId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (clusterId == null) {
            	LOG.warn("Failed to find cluster id from clusterKey:" + clusterKey + " datasourceId:" + datasourceId);
            } else {
            	// Convert sample uuids to object ids using the key cache
            	// VserverHandler.TAG is used to identify the object to look up in the key cache 
                Collection<Sample<CifsVserverRawMeasurements>> samples
                        = new SimpleRawSampleConverter<CifsVserverRawMeasurements>(clusterId, keyCache, VserverHandler.TAG, originId).toSamples(rawSamples, time, firstReport);
                if(samples.isEmpty()){
                	LOG.warn("Failed to convert CifsVserver raw-samples retrieved from the acquisiton into server-samples. clusterKey: {} datasourceId: {}", clusterKey, datasourceId);
                	return;
                }
            	ComponentsFactoryRepository.getCifsVserverPerformanceRawComponentsFactory().createPersistenceFacade(connection, clusterKey, clusterId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME_RAW, datasourceId, startTime, System.currentTimeMillis() - startTime,
                        clusterKey, rawSamples.size(), samples.size(), leaseTime);
            }            
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert raw cifs:vserver performance, clusterKey:" + clusterKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }		
	}
	
	@Override
    public void reportCifsVserverData(Long datasourceId, String clusterKey, long time, Collection<RawSample<CifsVserverMeasurements>> rawSamples, long leaseTime, boolean isHistorical) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();
        Long containerId = null;
        Collection<Sample<CifsVserverMeasurements>> samples = null;

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(CIFSVSERVER_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(CIFSVSERVER_PACKAGE, clusterKey, datasourceId, leaseTime)) {
                LOG.info("Cluster key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(CIFSVSERVER_PACKAGE, clusterKey, leaseTime) + " (clusterKey:" + clusterKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            containerId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (containerId == null) {
                LOG.warn("Failed to find containerId from clusterKey:" + clusterKey + " datasourceId:" + datasourceId);
            } else {
                // Convert sample uuids to object ids using the key cache
                // VserverHandler.TAG is used to identify the object to look up in the key cache 
                samples = new SimpleRawSampleConverter<CifsVserverMeasurements>(containerId, keyCache, VserverHandler.TAG, originId).toSamples(rawSamples, time, firstReport);
                if(samples.isEmpty()){
                	LOG.warn("Failed to convert CifsVserver raw-samples retrieved from the acquisiton into server-samples. clusterKey: {} datasourceId: {}", clusterKey, datasourceId);
                	return;
                }
                ComponentsFactoryRepository.getCifsVserverPerformanceComponentsFactory().createPersistenceFacade(connection, clusterKey, containerId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME, datasourceId, startTime, System.currentTimeMillis() - startTime,
                        clusterKey, rawSamples.size(), samples.size(), leaseTime);
                
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert cifs:vserver performance, clusterKey:" + clusterKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }

        // invoke all the registered listeners for this sample
        try {
        	if (samples != null) {       
        		Map<String, PerformanceSampleListener> listeners = sampleListenerRegistry.getSampleListeners(SampleType.CIFS_VSERVER);                
        		if (listeners != null) {
        			LOG.info("There are " + listeners.size() + " listeners for cifsVserver stats.");
        			Map<Long, Collection<Sample<CifsVserverMeasurements>>> rslt = new HashMap<>();
        			rslt.put(containerId, samples);

        			long begin = System.currentTimeMillis();                	
        			for (String listenerName : listeners.keySet()) {
        				long iBegin = System.currentTimeMillis();
        				listeners.get(listenerName).processCifsVserver(rslt, isHistorical);
        				long iDelta = System.currentTimeMillis() - iBegin;
        				LOG.debug(listenerName + " took " + iDelta + " milliseconds to process cifsVserver samples.");        				
        			}                	
        			long delta = System.currentTimeMillis() - begin;

        			if (delta > PerformanceConstants.SAMPLE_LISTENER_MAX_ELAPSED_TIME) {
        				LOG.warn("It took " + delta + " milliseconds for the cifsVserver sample listeners to complete their process.");
        			}
        		}                
        	}
        } catch(Exception e) {
        	LOG.error("Unable to invoke registered listeners for cifsVserver; ", e);
        }
    }
}
