/* Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.cifsvserver.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.cifsvserver.interfaces.data.CifsVserverMeasurements;
import com.netapp.oci.netapp.performance.cifsvserver.interfaces.local.CifsVserverPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.cifsvserver.interfaces.remote.CifsVserverPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class CifsVserverPerformanceQueryBean implements CifsVserverPerformanceQueryRemote, CifsVserverPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<CifsVserverMeasurements>>> getSamples(Set<Long> cifsVserverIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + cifsVserverIds.size() + "cifsVserver samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            
            Map<Long, List<SampleInfo<CifsVserverMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<CifsVserverMeasurements>>>();
              
            Set<Long> remainingIds = new HashSet<Long>(cifsVserverIds);
            Map<Long, Set<Long>> clusterIdToCifsVserverIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, cifsVserverIds);
            
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToCifsVserverIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find cifsVserver with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            
            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get cifsVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query cifsVserver samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public Map<Long, List<SampleInfo<CifsVserverMeasurements>>> getSamples(Long containerId, Set<Long> cifsVserverIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + cifsVserverIds.size() + "cifsVserver samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, cifsVserverIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get cifsVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query cifsVserver samples: " + (System.currentTimeMillis() - startTime));
        }
    }
    

    @Override
    public List<SampleInfo<CifsVserverMeasurements>> getSamples(Long containerId, Long cifsVserverId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(cifsVserverId), fromTime, toTime).get(cifsVserverId);

    }
    
    private Map<Long, List<SampleInfo<CifsVserverMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> cifsVserverIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getCifsVserverPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(cifsVserverIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<CifsVserverMeasurements>> getLatestSample(Long vserverId, Set<Long> cifsVserverIds) throws NetappInterfaceException {
        logger.info("Query " + cifsVserverIds.size() + "cifsVserver samples for vserver:" + vserverId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getCifsVserverPerformanceComponentsFactory().createPersistenceFacade(connection, vserverId)
                            .getLatestSample(cifsVserverIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get cifsVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<CifsVserverMeasurements>>> getSummaries(Set<Long> vServerIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vServerIds.size() + " CIFS vServer summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<CifsVserverMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(vServerIds);

            Map<Long, Set<Long>> clusterIdTovServerIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, vServerIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdTovServerIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<CifsVserverMeasurements>> summaries = ComponentsFactoryRepository
                        .getCifsVserverPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSummariesGroupById(new TimeSpan(fromTime, toTime), ids);

                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find CIFS for vServers with id "
                        + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        }
        catch (Exception e) {
            throw new NetappInterfaceException("Failed to get CIFS vServer summaries", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<CifsVserverMeasurements>>> getSummaries(Long containerId, Set<Long> cifsVserverIds, Date fromTime,
                                                                   Date toTime) throws NetappInterfaceException {
        logger.info("Query " + cifsVserverIds.size() + "cifsVserver summaries from " + fromTime + " to " + toTime
                + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<CifsVserverMeasurements>>> idToSummaries = new HashMap<>();

            List<Summary<CifsVserverMeasurements>> summaries = ComponentsFactoryRepository
                    .getCifsVserverPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                    .getSummariesGroupById(new TimeSpan(fromTime, toTime), cifsVserverIds);

            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);

            return idToSummaries;
        }
        catch (Exception e) {
            throw new NetappInterfaceException("Failed to get cifsVserver summaries", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }


    @Override
      public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
          logger.info("Query if " + objId + "cifsVserver has counters from " + fromTime + " to " + toTime);

          Connection connection = null;
          try {
              connection = ConnectionHelper.newConnection(dataSource);
              return ComponentsFactoryRepository.getCifsVserverPerformanceComponentsFactory()
                      .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
          } catch (Exception e) {
              throw new NetappInterfaceException("Failed to assert if cifsVserver has counters ", e);
          } finally {
              ConnectionHelper.close(connection);
          }
      }
}
