/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.cluster.interfaces.impl;

import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.cluster.interfaces.data.ClusterMeasurements;
import com.netapp.oci.netapp.performance.cluster.interfaces.local.ClusterPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.cluster.interfaces.remote.ClusterPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class ClusterPerformanceQueryBean implements ClusterPerformanceQueryRemote, ClusterPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<ClusterMeasurements>>> getSamples(Set<Long> clusterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
    	
    	long startTime = System.currentTimeMillis();
        logger.info("Query " + clusterIds.size() + "cluster samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
        	connection = ConnectionHelper.newConnection(dataSource);

        	Map<Long, List<SampleInfo<ClusterMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<ClusterMeasurements>>>();
        	containerIdToSamples.putAll(getSamples(connection, clusterIds.iterator().next(), clusterIds, fromTime, toTime));
        	return containerIdToSamples;
        } catch (Exception e) {
        	throw new NetappInterfaceException("Failed to get cluster samples", e);
        } finally {
        	ConnectionHelper.close(connection);
            logger.info("Total Execution time to query cluster samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<ClusterMeasurements>>> getSamples(Long containerId, Set<Long> clusterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + clusterIds.size() + "cluster samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, clusterIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get cluster samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query cluster samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<ClusterMeasurements>> getSamples(Long containerId, Long clusterId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(clusterId), fromTime, toTime).get(clusterId);
    }

    private Map<Long, List<SampleInfo<ClusterMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> clusterIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getClusterPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(clusterIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<ClusterMeasurements>> getLatestSample(Long clusterId, Set<Long> clusterIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + clusterIds.size() + "cluster samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getClusterPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(clusterIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get cluster samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<ClusterMeasurements>>> getSummaries(Set<Long> clusterIds, Date fromTime, Date toTime)
            throws NetappInterfaceException {
        logger.info("Query " + clusterIds.size() + "cluster summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<ClusterMeasurements>>> idToSummaries = new HashMap<>();
            

            List<Summary<ClusterMeasurements>> summaries = ComponentsFactoryRepository
                        .getClusterPerformanceComponentsFactory().createPersistenceFacade(connection, clusterIds.iterator().next())
                        .getSummariesGroupById(new TimeSpan(fromTime, toTime), clusterIds);

                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);

            return idToSummaries;
        }
        catch (Exception e) {
            throw new NetappInterfaceException("Failed to get cluster summaries", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<ClusterMeasurements>>> getSummaries(Long containerId, Set<Long> clusterIds, Date fromTime,
            Date toTime) throws NetappInterfaceException {
        logger.info("Query " + clusterIds.size() + "cluster summaries from " + fromTime + " to " + toTime
                + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<ClusterMeasurements>>> idToSummaries = new HashMap<>();

            List<Summary<ClusterMeasurements>> summaries = ComponentsFactoryRepository
                    .getClusterPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                    .getSummariesGroupById(new TimeSpan(fromTime, toTime), clusterIds);

            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);

            return idToSummaries;
        }
        catch (Exception e) {
            throw new NetappInterfaceException("Failed to get cluster summaries", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }
    
    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "cluster has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getClusterPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if cluster has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
