/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.disk;

import com.netapp.oci.netapp.performance.disk.interfaces.data.DiskMeasurements;
import com.netapp.oci.netapp.performance.disk.interfaces.data.DiskRawMeasurements;
import com.netapp.oci.netapp.performance.shared.PerformanceTableUtils;
import com.netapp.oci.platform.common.table.ObjectParser;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.tables.Column;
import com.netapp.oci.platform.framework.performance.tables.RatioMethod;
import com.netapp.oci.platform.framework.performance.tables.SqlMethod;

import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * @author Anton Sergeev
 */
public class DiskMetaData {
	public static final Column[] RAW_COLUMNS = new Column[]{
		new Column("base_for_disk_busy", "bigint unsigned"),
		new Column("cp_read_blocks", "bigint unsigned"),
		new Column("cp_read_latency", "bigint unsigned"),
		new Column("cp_reads", "bigint unsigned"),
		new Column("disk_busy", "bigint unsigned"),
		new Column("guarenteed_read_blocks", "bigint unsigned"),
		new Column("guarenteed_read_latency", "bigint unsigned"),
		new Column("guarenteed_reads", "bigint unsigned"),
		new Column("guarenteed_write_blocks", "bigint unsigned"),
		new Column("guarenteed_write_latency", "bigint unsigned"),
		new Column("guarenteed_writes", "bigint unsigned"),
		new Column("total_transfers", "bigint unsigned"),
		new Column("user_read_blocks", "bigint unsigned"),
		new Column("user_read_latency", "bigint unsigned"),
		new Column("user_reads", "bigint unsigned"),
		new Column("user_write_blocks", "bigint unsigned"),
		new Column("user_write_latency", "bigint unsigned"),
		new Column("user_writes", "bigint unsigned")
	};
    
    public static final Column[] SAMPLE_COLUMNS = new Column[]{
            new Column("cpReadBlocks", "float"),
            new Column("cpReadLatency", "float"),
            new Column("cpReads", "float"),
            new Column("diskBusy", "float"),
            new Column("baseForDiskBusy", "bigint unsigned"),
            new Column("guaranteedReadBlocks", "float"),
            new Column("guaranteedReadLatency", "float"),
            new Column("guaranteedReads", "float"),
            new Column("guaranteedWriteBlocks", "float"),
            new Column("guaranteedWriteLatency", "float"),
            new Column("guaranteedWrites", "float"),
            new Column("totalTransfers", "float"),
            new Column("userReadBlocks", "float"),
            new Column("userReadLatency", "float"),
            new Column("userReads", "float"),
            new Column("userWriteBlocks", "float"),
            new Column("userWriteLatency", "float"),
            new Column("userWrites", "float"),
            new Column("opmDiskThroughput", "float"),
            new Column("opmDiskLatency", "float"),
            new Column("totalOps", "float"),
    };
    
    public static final Column[] SUMMARY_COLUMNS = new Column[]{
            new Column("cpReadBlocks", "float"),
            new Column("cpReadLatency", "float"),
            new Column("cpReads", "float"),
            new Column("diskBusy", "float"),
            new Column("baseForDiskBusy", "bigint unsigned"),
            new Column("guaranteedReadBlocks", "float"),
            new Column("guaranteedReadLatency", "float"),
            new Column("guaranteedReads", "float"),
            new Column("guaranteedWriteBlocks", "float"),
            new Column("guaranteedWriteLatency", "float"),
            new Column("guaranteedWrites", "float"),
            new Column("totalTransfers", "float"),
            new Column("userReadBlocks", "float"),
            new Column("userReadLatency", "float"),
            new Column("userReads", "float"),
            new Column("userWriteBlocks", "float"),
            new Column("userWriteLatency", "float"),
            new Column("userWrites", "float"),
            new Column("opmDiskThroughput", "float"),
            new Column("opmDiskLatency", "float"),
            new Column("totalOps", "float"),
    };
    
    public static final SqlMethod[][] SAMPLE_TO_SUMMARY_METHODS = new SqlMethod[][]{
            {SqlMethod.AVG},
            {new RatioMethod("cpReads")},
            {SqlMethod.AVG},
            {new RatioMethod("baseForDiskBusy")},
            {SqlMethod.SUM},
            {SqlMethod.AVG},
            {new RatioMethod("guaranteedReads")},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {new RatioMethod("guaranteedWrites")},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {new RatioMethod("userReads")},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {new RatioMethod("userWrites")},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {new RatioMethod("totalOps")},
            {SqlMethod.AVG},
    };

    public static final SqlMethod[] SUMMARY_COMMON_METHODS = new SqlMethod[]{
            SqlMethod.AVG,
            new RatioMethod("cpReads"),
            SqlMethod.AVG,
            new RatioMethod("baseForDiskBusy"),
            SqlMethod.SUM,
            SqlMethod.AVG,
            new RatioMethod("guaranteedReads"),
            SqlMethod.AVG,
            SqlMethod.AVG,
            new RatioMethod("guaranteedWrites"),
            SqlMethod.AVG,
            SqlMethod.AVG,
            SqlMethod.AVG,
            new RatioMethod("userReads"),
            SqlMethod.AVG,
            SqlMethod.AVG,
            new RatioMethod("userWrites"),
            SqlMethod.AVG,
            SqlMethod.AVG,
            new RatioMethod("totalOps"),
            SqlMethod.AVG,
    };

    private static ObjectParser<DiskRawMeasurements> rawParser;
    private static ObjectParser<DiskMeasurements> sampleParser;

    static {
    	rawParser = new ObjectParser<DiskRawMeasurements>() {
            public void prepareObjectDataForInsert(PreparedStatementWrapper preparedStatement, DiskRawMeasurements objectData) throws TableException {
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getBaseForDiskBusy());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getCpReadBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getCpReadLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getCpReads());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getDiskBusy());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuarenteedReadBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuarenteedReadLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuarenteedReads());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuarenteedWriteBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuarenteedWriteLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuarenteedWrites());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getTotalTransfers());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserReadBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserReadLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserReads());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserWriteBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserWriteLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserWrites());
            }

            public DiskRawMeasurements parseObjectData(ResultSet resultSet, int startIndex) throws SQLException {
                int i = startIndex - 1;
                return new DiskRawMeasurements(
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i),
                        PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i));
            }
        };
        
        sampleParser = new ObjectParser<DiskMeasurements>() {
            public void prepareObjectDataForInsert(PreparedStatementWrapper preparedStatement, DiskMeasurements objectData) throws TableException {
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getCpReadBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getCpReadLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getCpReads());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getDiskBusy());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getBaseForDiskBusy());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuaranteedReadBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuaranteedReadLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuaranteedReads());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuaranteedWriteBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuaranteedWriteLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getGuaranteedWrites());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getTotalTransfers());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserReadBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserReadLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserReads());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserWriteBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserWriteLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getUserWrites());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getOpmDiskThroughput());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getOpmDiskLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getTotalOps());
            }

            public DiskMeasurements parseObjectData(ResultSet resultSet, int startIndex) throws SQLException {
                int i = startIndex - 1;
                return new DiskMeasurements(
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getLongNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i));
            }
        };
    }

    public static ObjectParser<DiskMeasurements> getSummaryParser() {
        return sampleParser;
    }

    public static ObjectParser<DiskMeasurements> getSampleParser() {
        return sampleParser;
    }    

    public static ObjectParser<DiskRawMeasurements> getRawParser() {
        return rawParser;
    }
}
