/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.disk.interfaces.data;

import com.netapp.oci.platform.util.schema.OCIEPerformanceTable;
import com.netapp.oci.platform.util.schema.PerformanceValue;

import java.io.Serializable;

/**
 * @author Anton Sergeev
 */

@OCIEPerformanceTable(
        tableName = "disk",
        cmObjectName = "disk",
        description = "shard by cluster"
)
public class DiskMeasurements implements Serializable {
    private static final long serialVersionUID = 1L;

    /**
     * Number of blocks transferred for consistency pofloat read operations per second
     */
    @PerformanceValue(
            counterName = "cp_read_blocks",
            desc = "Number of blocks transferred for consistency pofloat read operations per second"
    )
    private float cpReadBlocks;

    /**
     * Average latency per block in microseconds for consistency point read operations
     */
    @PerformanceValue(
            counterName = "cp_read_latency",
            desc = "Average latency per block in microseconds for consistency point read operations",
            counterType = "average",
            baseCounter = "cp_read_blocks"
    )
    private float cpReadLatency;

    /**
     * Number of disk read operations initiated each second for consistency point processing
     */
    @PerformanceValue(
            counterName = "cp_reads",
            desc = "Number of disk read operations initiated each second for consistency point processing"
    )
    private float cpReads;

    /**
     * Percentage of time there was at least one outstanding request to the disk
     */
    @PerformanceValue(
            counterName = "disk_busy",
            desc = "Percentage of time there was at least one outstanding request to the disk",
            counterType = "percent",
            baseCounter = "base_for_disk_busy"
    )
    private float diskBusy;
    
    /**
     * Disk operations used for calculate disk busy
     */
    private long baseForDiskBusy;

    /**
     * Number of blocks transferred for guaranteed read operations per second
     */
    @PerformanceValue(
            counterName = "guaranteed_read_blocks",
            desc = "Number of blocks transferred for guaranteed read operations per second"
    )
    private float guaranteedReadBlocks;

    /**
     * Average latency per block in microseconds for guaranteed read operations
     */
    @PerformanceValue(
            counterName = "guaranteed_read_latency",
            desc = "Average latency per block in microseconds for guaranteed read operations",
            counterType = "average",
            baseCounter = "guaranteed_read_blocks"

    )
    private float guaranteedReadLatency;

    /**
     * Number of disk read operations initiated each second for raid reconstruct or scrubbing activities
     */
    @PerformanceValue(
            counterName = "guaranteed_reads",
            desc = "Number of disk read operations initiated each second for raid reconstruct or scrubbing activities"
    )
    private float guaranteedReads;

    /**
     * Number of blocks transferred for guaranteed write operations per second
     */
    @PerformanceValue(
            counterName = "guaranteed_write_blocks",
            desc = "Number of blocks transferred for guaranteed write operations per second"
    )
    private float guaranteedWriteBlocks;

    /**
     * Average latency per block in microseconds for guaranteed write operations
     */
    @PerformanceValue(
            counterName = "guaranteed_write_latency",
            desc = "Average latency per block in microseconds for guaranteed write operations",
            counterType = "average",
            baseCounter = "guaranteed_write_blocks"
    )
    private float guaranteedWriteLatency;

    /**
     * Number of write read operations initiated each second for raid reconstruct or scrubbing activities
     */
    @PerformanceValue(
            counterName = "guaranteed_writes",
            desc = "Number of write read operations initiated each second for raid reconstruct or scrubbing activities"
    )
    private float guaranteedWrites;

    /**
     * Total number of disk operations involving data transfer initiated per second
     */
    @PerformanceValue(
            counterName = "total_transfers",
            desc = "Total number of disk operations involving data transfer initiated per second"
    )
    private float totalTransfers;

    /**
     * Number of blocks transferred for user read operations per second
     */
    @PerformanceValue(
            counterName = "user_read_blocks",
            desc = "Number of blocks transferred for user read operations per second"
    )
    private float userReadBlocks;

    /**
     * Average latency per block in microseconds for user read operations
     */
    @PerformanceValue(
            counterName = "user_read_latency",
            desc = "Average latency per block in microseconds for user read operations",
            counterType = "average",
            baseCounter = "user_read_blocks"
    )
    private float userReadLatency;

    /**
     * Number of disk read operations initiated each second for retrieving data or metadata associated with user requests
     */
    @PerformanceValue(
            counterName = "user_reads",
            desc = "Number of disk read operations initiated each second for retrieving data or metadata associated with user requests"
    )
    private float userReads;

    /**
     * Number of blocks transferred for user write operations per second
     */
    @PerformanceValue(
            counterName = "user_write_blocks",
            desc = "Number of blocks transferred for user write operations per second"
    )
    private float userWriteBlocks;

    /**
     * Average latency per block in microseconds for user write operations
     */
    @PerformanceValue(
            counterName = "user_write_latency",
            desc = "Average latency per block in microseconds for user write operations",
            counterType = "average",
            baseCounter = "user_write_blocks"
    )
    private float userWriteLatency;

    /**
     * Number of disk write operations initiated each second for storing data or metadata associated with user requests
     */
    @PerformanceValue(
            counterName = "user_writes",
            desc = "Number of disk write operations initiated each second for storing data or metadata associated with user requests"
    )
    private float userWrites;

    @PerformanceValue(
            
    		counterName = "opm_disk_throughput",
            desc = "Synthetic counter for OPM that is a sum of user_read_blocks + user_write_blocks + cp_read_blocks",
            counterType = "sum"
    )
    private float opmDiskThroughput;

    @PerformanceValue(
            
    		counterName = "opm_disk_latency",
            desc = "Synthetic counter for OPM that is a sum of latencies",
            counterType = "average Weighted Latency"
    )
    private float opmDiskLatency;
    
    @PerformanceValue(
            
    		counterName = "total_ops",
            desc = "Synthetic counter for OPM that is a sum of (user_reads, user_writes, cp_reads)",
            counterType = "sum"
    )
    private float totalOps;
    
    
    public DiskMeasurements() {
    }

    public DiskMeasurements(float cpReadBlocks, float cpReadLatency, float cpReads, float diskBusy, long baseForDiskBusy, float guaranteedReadBlocks,
                            float guaranteedReadLatency, float guaranteedReads, float guaranteedWriteBlocks, float guaranteedWriteLatency,
                            float guaranteedWrites, float totalTransfers, float userReadBlocks, float userReadLatency, float userReads,
                            float userWriteBlocks, float userWriteLatency, float userWrites, float opmDiskThroughput, float opmDiskLatency, float totalOps) {
        this.cpReadBlocks = cpReadBlocks;
        this.cpReadLatency = cpReadLatency;
        this.cpReads = cpReads;
        this.diskBusy = diskBusy;
        this.baseForDiskBusy = baseForDiskBusy;
        this.guaranteedReadBlocks = guaranteedReadBlocks;
        this.guaranteedReadLatency = guaranteedReadLatency;
        this.guaranteedReads = guaranteedReads;
        this.guaranteedWriteBlocks = guaranteedWriteBlocks;
        this.guaranteedWriteLatency = guaranteedWriteLatency;
        this.guaranteedWrites = guaranteedWrites;
        this.totalTransfers = totalTransfers;
        this.userReadBlocks = userReadBlocks;
        this.userReadLatency = userReadLatency;
        this.userReads = userReads;
        this.userWriteBlocks = userWriteBlocks;
        this.userWriteLatency = userWriteLatency;
        this.userWrites = userWrites;
        this.opmDiskThroughput = opmDiskThroughput;
        this.opmDiskLatency = opmDiskLatency;
        this.totalOps = totalOps;
    }
    
    public long getBaseForDiskBusy() {
		return baseForDiskBusy;
	}

	public void setBaseForDiskBusy(long baseForDiskBusy) {
		this.baseForDiskBusy = baseForDiskBusy;
	}

	public float getCpReadBlocks() {
        return cpReadBlocks;
    }

    public void setCpReadBlocks(float cpReadBlocks) {
        this.cpReadBlocks = cpReadBlocks;
    }

    public float getCpReadLatency() {
        return cpReadLatency;
    }

    public void setCpReadLatency(float cpReadLatency) {
        this.cpReadLatency = cpReadLatency;
    }

    public float getCpReads() {
        return cpReads;
    }

    public void setCpReads(float cpReads) {
        this.cpReads = cpReads;
    }

    public float getDiskBusy() {
        return diskBusy;
    }

    public void setDiskBusy(float diskBusy) {
        this.diskBusy = diskBusy;
    }

    public float getGuaranteedReadBlocks() {
        return guaranteedReadBlocks;
    }

    public void setGuaranteedReadBlocks(float guaranteedReadBlocks) {
        this.guaranteedReadBlocks = guaranteedReadBlocks;
    }

    public float getGuaranteedReadLatency() {
        return guaranteedReadLatency;
    }

    public void setGuaranteedReadLatency(float guaranteedReadLatency) {
        this.guaranteedReadLatency = guaranteedReadLatency;
    }

    public float getGuaranteedReads() {
        return guaranteedReads;
    }

    public void setGuaranteedReads(float guaranteedReads) {
        this.guaranteedReads = guaranteedReads;
    }

    public float getGuaranteedWriteBlocks() {
        return guaranteedWriteBlocks;
    }

    public void setGuaranteedWriteBlocks(float guaranteedWriteBlocks) {
        this.guaranteedWriteBlocks = guaranteedWriteBlocks;
    }

    public float getGuaranteedWriteLatency() {
        return guaranteedWriteLatency;
    }

    public void setGuaranteedWriteLatency(float guaranteedWriteLatency) {
        this.guaranteedWriteLatency = guaranteedWriteLatency;
    }

    public float getGuaranteedWrites() {
        return guaranteedWrites;
    }

    public void setGuaranteedWrites(float guaranteedWrites) {
        this.guaranteedWrites = guaranteedWrites;
    }

    public float getTotalTransfers() {
        return totalTransfers;
    }

    public void setTotalTransfers(float totalTransfers) {
        this.totalTransfers = totalTransfers;
    }

    public float getUserReadBlocks() {
        return userReadBlocks;
    }

    public void setUserReadBlocks(float userReadBlocks) {
        this.userReadBlocks = userReadBlocks;
    }

    public float getUserReadLatency() {
        return userReadLatency;
    }

    public void setUserReadLatency(float userReadLatency) {
        this.userReadLatency = userReadLatency;
    }

    public float getUserReads() {
        return userReads;
    }

    public void setUserReads(float userReads) {
        this.userReads = userReads;
    }

    public float getUserWriteBlocks() {
        return userWriteBlocks;
    }

    public void setUserWriteBlocks(float userWriteBlocks) {
        this.userWriteBlocks = userWriteBlocks;
    }

    public float getUserWriteLatency() {
        return userWriteLatency;
    }

    public void setUserWriteLatency(float userWriteLatency) {
        this.userWriteLatency = userWriteLatency;
    }

    public float getUserWrites() {
        return userWrites;
    }

    public void setUserWrites(float userWrites) {
        this.userWrites = userWrites;
    }

    public float getOpmDiskThroughput() {
		return opmDiskThroughput;
	}

	public void setOpmDiskThroughput(float opmDiskThroughput) {
		this.opmDiskThroughput = opmDiskThroughput;
	}

	public float getOpmDiskLatency() {
		return opmDiskLatency;
	}

	public void setOpmDiskLatency(float opmDiskLatency) {
		this.opmDiskLatency = opmDiskLatency;
	}

	public float getTotalOps() {
		return totalOps;
	}

	public void setTotalOps(float totalOps) {
		this.totalOps = totalOps;
	}

	@Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof DiskMeasurements)) return false;

        DiskMeasurements that = (DiskMeasurements) o;

        return Float.compare(that.cpReadBlocks, cpReadBlocks) == 0
                && Float.compare(that.cpReadLatency, cpReadLatency) == 0
                && Float.compare(that.cpReads, cpReads) == 0
                && Float.compare(that.diskBusy, diskBusy) == 0
                && that.baseForDiskBusy == baseForDiskBusy
                && Float.compare(that.guaranteedReadBlocks, guaranteedReadBlocks) == 0
                && Float.compare(that.guaranteedReadLatency, guaranteedReadLatency) == 0
                && Float.compare(that.guaranteedReads, guaranteedReads) == 0
                && Float.compare(that.guaranteedWriteBlocks, guaranteedWriteBlocks) == 0
                && Float.compare(that.guaranteedWriteLatency, guaranteedWriteLatency) == 0
                && Float.compare(that.guaranteedWrites, guaranteedWrites) == 0
                && Float.compare(that.totalTransfers, totalTransfers) == 0
                && Float.compare(that.userReadBlocks, userReadBlocks) == 0
                && Float.compare(that.userReadLatency, userReadLatency) == 0
                && Float.compare(that.userReads, userReads) == 0
                && Float.compare(that.userWriteBlocks, userWriteBlocks) == 0
                && Float.compare(that.userWriteLatency, userWriteLatency) == 0
                && Float.compare(that.userWrites, userWrites) == 0
        		&& Float.compare(that.opmDiskThroughput, opmDiskThroughput) == 0
        		&& Float.compare(that.opmDiskLatency, opmDiskLatency) == 0
        		&& Float.compare(that.totalOps, totalOps) == 0;
    }

    @Override
    public int hashCode() {
        int result = (cpReadBlocks != +0.0f ? Float.floatToIntBits(cpReadBlocks) : 0);
        result = 31 * result + (cpReadLatency != +0.0f ? Float.floatToIntBits(cpReadLatency) : 0);
        result = 31 * result + (cpReads != +0.0f ? Float.floatToIntBits(cpReads) : 0);
        result = 31 * result + (diskBusy != +0.0f ? Float.floatToIntBits(diskBusy) : 0);
        result = 31 * result + (int) (baseForDiskBusy ^ baseForDiskBusy >>> 32);
        result = 31 * result + (guaranteedReadBlocks != +0.0f ? Float.floatToIntBits(guaranteedReadBlocks) : 0);
        result = 31 * result + (guaranteedReadLatency != +0.0f ? Float.floatToIntBits(guaranteedReadLatency) : 0);
        result = 31 * result + (guaranteedReads != +0.0f ? Float.floatToIntBits(guaranteedReads) : 0);
        result = 31 * result + (guaranteedWriteBlocks != +0.0f ? Float.floatToIntBits(guaranteedWriteBlocks) : 0);
        result = 31 * result + (guaranteedWriteLatency != +0.0f ? Float.floatToIntBits(guaranteedWriteLatency) : 0);
        result = 31 * result + (guaranteedWrites != +0.0f ? Float.floatToIntBits(guaranteedWrites) : 0);
        result = 31 * result + (totalTransfers != +0.0f ? Float.floatToIntBits(totalTransfers) : 0);
        result = 31 * result + (userReadBlocks != +0.0f ? Float.floatToIntBits(userReadBlocks) : 0);
        result = 31 * result + (userReadLatency != +0.0f ? Float.floatToIntBits(userReadLatency) : 0);
        result = 31 * result + (userReads != +0.0f ? Float.floatToIntBits(userReads) : 0);
        result = 31 * result + (userWriteBlocks != +0.0f ? Float.floatToIntBits(userWriteBlocks) : 0);
        result = 31 * result + (userWriteLatency != +0.0f ? Float.floatToIntBits(userWriteLatency) : 0);
        result = 31 * result + (userWrites != +0.0f ? Float.floatToIntBits(userWrites) : 0);
        result = 31 * result + (opmDiskThroughput != +0.0f ? Float.floatToIntBits(opmDiskThroughput) : 0);
        result = 31 * result + (opmDiskLatency != +0.0f ? Float.floatToIntBits(opmDiskLatency) : 0);
        result = 31 * result + (totalOps != +0.0f ? Float.floatToIntBits(totalOps) : 0);
        return result;
    }

    @Override
    public String toString() {
        return "DiskMeasurements{" +
                "cpReadBlocks=" + cpReadBlocks +
                ", cpReadLatency=" + cpReadLatency +
                ", cpReads=" + cpReads +
                ", diskBusy=" + diskBusy +
                ", baseForDiskBusy=" + baseForDiskBusy +
                ", guaranteedReadBlocks=" + guaranteedReadBlocks +
                ", guaranteedReadLatency=" + guaranteedReadLatency +
                ", guaranteedReads=" + guaranteedReads +
                ", guaranteedWriteBlocks=" + guaranteedWriteBlocks +
                ", guaranteedWriteLatency=" + guaranteedWriteLatency +
                ", guaranteedWrites=" + guaranteedWrites +
                ", totalTransfers=" + totalTransfers +
                ", userReadBlocks=" + userReadBlocks +
                ", userReadLatency=" + userReadLatency +
                ", userReads=" + userReads +
                ", userWriteBlocks=" + userWriteBlocks +
                ", userWriteLatency=" + userWriteLatency +
                ", userWrites=" + userWrites +
                ", opmDiskThroughput=" + opmDiskThroughput +
                ", opmDiskLatency=" + opmDiskLatency +
                ", totalOps=" + totalOps +
                '}';
    }
}
