/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.disk.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.disk.interfaces.data.DiskMeasurements;
import com.netapp.oci.netapp.performance.disk.interfaces.local.DiskPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.disk.interfaces.remote.DiskPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;
import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class DiskPerformanceQueryBean implements DiskPerformanceQueryRemote, DiskPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<DiskMeasurements>>> getSamples(Set<Long> diskIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + diskIds.size() + "disk samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<DiskMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<DiskMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(diskIds);

            Map<Long, Set<Long>> clusterIdToDiskIds = NetAppModel.INSTANCE.getClusterIdByDiskIds(connection, diskIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToDiskIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, diskIds, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find disk with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get disk samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query disk samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<DiskMeasurements>>> getSamples(Long containerId, Set<Long> diskIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + diskIds.size() + "disk samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, diskIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get disk samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query disk samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<DiskMeasurements>> getSamples(Long containerId, Long diskId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(diskId), fromTime, toTime).get(diskId);
    }

    private Map<Long, List<SampleInfo<DiskMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> diskIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getDiskPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(diskIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<DiskMeasurements>> getLatestSample(Long clusterId, Set<Long> diskIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + diskIds.size() + "disk samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getDiskPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(diskIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get disk samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<DiskMeasurements>>> getSummaries(Set<Long> diskIds, Date fromTime, Date toTime)
            throws NetappInterfaceException {
        logger.info("Query " + diskIds.size() + "disk summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<DiskMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(diskIds);

            Map<Long, Set<Long>> clusterIdToDiskIds = NetAppModel.INSTANCE.getClusterIdByDiskIds(connection, diskIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToDiskIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<DiskMeasurements>> summaries = ComponentsFactoryRepository
                        .getDiskPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSummariesGroupById(new TimeSpan(fromTime, toTime), ids);

                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find disk with id "
                        + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        }
        catch (Exception e) {
            throw new NetappInterfaceException("Failed to get disk summaries", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<DiskMeasurements>>> getSummaries(Long containerId, Set<Long> diskIds, Date fromTime,
            Date toTime) throws NetappInterfaceException {
        logger.info("Query " + diskIds.size() + "disk summaries from " + fromTime + " to " + toTime
                + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<DiskMeasurements>>> idToSummaries = new HashMap<>();

            List<Summary<DiskMeasurements>> summaries = ComponentsFactoryRepository
                    .getDiskPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                    .getSummariesGroupById(new TimeSpan(fromTime, toTime), diskIds);

            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);

            return idToSummaries;
        }
        catch (Exception e) {
            throw new NetappInterfaceException("Failed to get disk summaries", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }
    
    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "disk has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getDiskPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if disk has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
