package com.netapp.oci.netapp.performance.fcplif.interfaces.impl;

import java.sql.Connection;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.db.NetAppPerformance;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.originator.handler.ClusterHandler;
import com.netapp.oci.netapp.originator.handler.VserverHandler;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.fcplif.interfaces.data.FcpLifVserverMeasurements;
import com.netapp.oci.netapp.performance.fcplif.interfaces.data.FcpLifVserverRawMeasurements;
import com.netapp.oci.netapp.performance.fcplif.interfaces.remote.FcpLifVserverPerformanceRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.shared.SimpleRawSampleConverter;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceDatasourceRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry.SampleType;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.SimpleUuidKey;
import com.onaro.sanscreen.util.Constant;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class FcpLifVserverPerformanceBean implements FcpLifVserverPerformanceRemote {
    public static final String PACKAGE_NAME = "fcpLifVserver";
    public static final String PACKAGE_NAME_RAW = PACKAGE_NAME + PerformanceConstants.RAW_PACKAGE_SUFFIX;
    private static final String FCPLIFVSERVER_PACKAGE = PerformanceConstants.PACKAGE_PREFIX + PACKAGE_NAME;
    private static final Logger logger = LoggerFactory.getLogger(FCPLIFVSERVER_PACKAGE);

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * use this bean to register reporting datasource
     */
    @EJB
    private NetappPerformanceDatasourceRegistry datasourceRegistry;

    /**
     * use this bean to get the registered listeners
     */
    @EJB
    private NetappPerformanceSampleListenerRegistry sampleListenerRegistry;
    
    public void reportFcpLifVserverSamples(Long datasourceId, String clusterKey, long time, Collection<RawSample<FcpLifVserverMeasurements>> rawSamples, long leaseTime, boolean isHistorical) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();
        Long containerId = null;
        Collection<Sample<FcpLifVserverMeasurements>> samples = null;

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(FCPLIFVSERVER_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(FCPLIFVSERVER_PACKAGE, clusterKey, datasourceId, leaseTime)) {
                logger.info("cluster key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(FCPLIFVSERVER_PACKAGE, clusterKey, leaseTime) + " (clusterKey:" + clusterKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            containerId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (containerId == null) {
                logger.warn("Failed to find cluster id from clusterKey:" + containerId + " datasourceId:" + datasourceId);
            } else {
                samples = new SimpleRawSampleConverter<FcpLifVserverMeasurements>(containerId, keyCache, VserverHandler.TAG, originId).toSamples(rawSamples, time, firstReport);
                ComponentsFactoryRepository.getFcpLifVserverPerformanceComponentsFactory().createPersistenceFacade(connection, clusterKey, containerId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME, datasourceId, startTime, System.currentTimeMillis() - startTime,
                		clusterKey, rawSamples.size(), samples.size(), leaseTime);                
            }            
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert fcplifVserver performance, clusterKey:" + clusterKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }
        
        // invoke all the registered listeners for this sample
        try {
        	if (samples != null) {	
        		Map<String, PerformanceSampleListener> listeners = sampleListenerRegistry.getSampleListeners(SampleType.FCP_LIF_VSERVER);                
        		if (listeners != null) {
        			logger.info("There are " + listeners.size() + " listeners for fcpLifVserver stats.");
        			Map<Long, Collection<Sample<FcpLifVserverMeasurements>>> rslt = new HashMap<>();
        			rslt.put(containerId, samples);

        			long begin = System.currentTimeMillis();;                	
        			for (String listenerName : listeners.keySet()) {
        				long iBegin = System.currentTimeMillis();
        				listeners.get(listenerName).processFcpLifVserver(rslt, isHistorical);
        				long iDelta = System.currentTimeMillis() - iBegin;
        				logger.debug(listenerName + " took " + iDelta + " milliseconds to process fcpLifVserver samples.");
        			}                	
        			long delta = System.currentTimeMillis() - begin;

        			if (delta > PerformanceConstants.SAMPLE_LISTENER_MAX_ELAPSED_TIME) {
        				logger.warn("It took " + delta + " milliseconds for the fcpLifVserver sample listeners to complete their process.");
        			}
        		}               
        	}
        } catch(Exception e) {
        	logger.error("Unable to invoke registered listeners for FcpLifVserver; ", e);
        }
    }

    /**
     * @inherit
     */
    public void reportFcpLifVserverRawSamples(Long datasourceId, 
                                       String clusterKey, 
                                       long time, 
                                       Collection<RawSample<FcpLifVserverRawMeasurements>> rawSamples, 
                                       long leaseTime) throws NetappInterfaceException {

        Connection connection = null;
        long startTime = System.currentTimeMillis();

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(FCPLIFVSERVER_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(FCPLIFVSERVER_PACKAGE, clusterKey, datasourceId, leaseTime)) {
                logger.info("Cluster key is already registered by:" + 
                            datasourceRegistry.getRegisterDatasourceId(FCPLIFVSERVER_PACKAGE, clusterKey, leaseTime) + 
                            " (Cluster:" + clusterKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            Long clusterId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (clusterId == null) {
                logger.warn("Failed to find clusterId id from clusterKey:" + clusterId + " datasourceId:" + datasourceId);
            } else {
                Collection<Sample<FcpLifVserverRawMeasurements>> samples
                        = new SimpleRawSampleConverter<FcpLifVserverRawMeasurements>(clusterId, keyCache, VserverHandler.TAG, originId).toSamples(rawSamples, time, firstReport);
                ComponentsFactoryRepository.getFcpLifVserverPerformanceRawComponentsFactory().createPersistenceFacade(connection, clusterKey, clusterId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME_RAW, datasourceId, startTime, System.currentTimeMillis() - startTime,
                		clusterKey, rawSamples.size(), samples.size(), leaseTime);
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert fcplifVserver raw performance, clusterKey:" + clusterKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

}
