package com.netapp.oci.netapp.performance.fcplif.interfaces.impl;

import java.sql.Connection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.fcplif.interfaces.data.FcpLifVserverMeasurements;
import com.netapp.oci.netapp.performance.fcplif.interfaces.local.FcpLifVserverPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.fcplif.interfaces.remote.FcpLifVserverPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import com.onaro.sanscreen.util.Constant;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class FcpLifVserverPerformanceQueryBean implements FcpLifVserverPerformanceQueryRemote, FcpLifVserverPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<FcpLifVserverMeasurements>>> getSamples(Set<Long> fcpLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + fcpLifIds.size() + "fcpLifVserver samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<FcpLifVserverMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<FcpLifVserverMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(fcpLifIds);

            Map<Long, Set<Long>> clusterIdToFcpLifIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, fcpLifIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToFcpLifIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, fromTime, toTime, ids));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find fcpLifVserver with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpLifVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query fcpLifVserver samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<FcpLifVserverMeasurements>>> getSamples(Long containerId, Set<Long> fcpLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + fcpLifIds.size() + "fcpLifVserver samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, fromTime, toTime, fcpLifIds);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpLifVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query fcpLifVserver samples: " + (System.currentTimeMillis() - startTime));

        }
    }


    @Override
    public List<SampleInfo<FcpLifVserverMeasurements>> getSamples(Long containerId, Long fcpLifId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(fcpLifId), fromTime, toTime).get(fcpLifId);

    }

    private Map<Long, List<SampleInfo<FcpLifVserverMeasurements>>> getSamples(Connection connection, Long containerId, Date fromTime, Date toTime, Set<Long> ids) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getFcpLifVserverPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(ids, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<FcpLifVserverMeasurements>> getLatestSample(Long vServerId, Set<Long> fcpLifIds) throws NetappInterfaceException {
        logger.info("Query " + fcpLifIds.size() + "fcpLifVserver samples for vServer:" + vServerId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getFcpLifVserverPerformanceComponentsFactory().createPersistenceFacade(connection, vServerId)
                            .getLatestSample(fcpLifIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpLif samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long,  List<Summary<FcpLifVserverMeasurements>>> getSummaries(Set<Long> vServerIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vServerIds.size() + " fcpLif vServer summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long,  List<Summary<FcpLifVserverMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(vServerIds);

            Map<Long, Set<Long>> clusterIdTovServerIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, vServerIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdTovServerIds.entrySet()) {
                Long clusterId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<FcpLifVserverMeasurements>> summaries = ComponentsFactoryRepository.getFcpLifVserverPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find fcpLifVserver with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpLifVserver summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long,  List<Summary<FcpLifVserverMeasurements>>> getSummaries(Long clusterId, Set<Long> vServerIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vServerIds.size() + " fcpLif vServer summaries from " + fromTime + " to " + toTime
                + " for containerId:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long,  List<Summary<FcpLifVserverMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<FcpLifVserverMeasurements>> summaries = ComponentsFactoryRepository.getFcpLifVserverPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), vServerIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpLifVserver summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + " fcpLifVserver has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getFcpLifVserverPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if fcpLifVserver has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
