/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.fcpport.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.fcpport.interfaces.data.FcpPortMeasurements;
import com.netapp.oci.netapp.performance.fcpport.interfaces.local.FcpPortPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.fcpport.interfaces.remote.FcpPortPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Nancy Walia
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class FcpPortPerformanceQueryBean implements FcpPortPerformanceQueryRemote, FcpPortPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<FcpPortMeasurements>>> getSamples(Set<Long> fcpPortIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + fcpPortIds.size() + "fcp port samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<FcpPortMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<FcpPortMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(fcpPortIds);

            Map<Long, Set<Long>> clusterIdToFcPortIds = NetAppModel.INSTANCE.getClusterIdByFcPortIds(connection, fcpPortIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToFcPortIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, fcpPortIds, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find fcPort with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpPort samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query fcp port samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<FcpPortMeasurements>>> getSamples(Long containerId, Set<Long> fcpPortIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + fcpPortIds.size() + "fcp port samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, fcpPortIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpPort samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query fcp port samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public List<SampleInfo<FcpPortMeasurements>> getSamples(Long containerId, Long fcpPortId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(fcpPortId), fromTime, toTime).get(fcpPortId);

    }

    private Map<Long, List<SampleInfo<FcpPortMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> fcpPortIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getFcpPortPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(fcpPortIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<FcpPortMeasurements>> getLatestSample(Long clusterId, Set<Long> fcpPortIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + fcpPortIds.size() + "fcpPort samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getFcpPortPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(fcpPortIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpPort samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    public Map<Long, List<Summary<FcpPortMeasurements>>> getSummaries(Set<Long> fcpPortIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + fcpPortIds.size() + "fcp port summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<FcpPortMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<Long>(fcpPortIds);

            Map<Long, Set<Long>> clusterIdTofcPortIds = NetAppModel.INSTANCE.getClusterIdByFcPortIds(connection, fcpPortIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdTofcPortIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<FcpPortMeasurements>> summaries = ComponentsFactoryRepository.getFcpPortPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find fcpPort with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpPort summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<FcpPortMeasurements>>> getSummaries(Long clusterId, Set<Long> fcpPortIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + fcpPortIds.size() + "fcp port summaries from " + fromTime + " to " + toTime
                + " for cluster:" + clusterId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<FcpPortMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<FcpPortMeasurements>> summaries = ComponentsFactoryRepository.getFcpPortPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, clusterId)
                    .getSummaries(new TimeSpan(fromTime, toTime), fcpPortIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcpPort summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "fcp port has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getFcpPortPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if fcp port has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}

