/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.iscsilif;

import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifMeasurements;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifRawMeasurements;
import com.netapp.oci.netapp.performance.shared.PerformanceTableUtils;
import com.netapp.oci.platform.common.table.ObjectParser;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.tables.Column;
import com.netapp.oci.platform.framework.performance.tables.RatioMethod;
import com.netapp.oci.platform.framework.performance.tables.SqlMethod;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * @author Al Velella
 *         Date: Aug 1, 2012
 *         Time: 9:49:01 AM
 */
public class IscsiLifMetaData {
    
    public static final Column[] RAW_COLUMNS = new Column[]{
        new Column("avg_latency", "bigint unsigned"),
        new Column("avg_other_latency", "bigint unsigned"),
        new Column("avg_read_latency", "bigint unsigned"),
        new Column("avg_write_latency", "bigint unsigned"),
        new Column("iscsi_other_ops", "bigint unsigned"),
        new Column("iscsi_read_ops", "bigint unsigned"),
        new Column("iscsi_write_ops", "bigint unsigned"),
        new Column("data_in_blocks", "bigint unsigned"),
        new Column("read_data", "bigint unsigned"),
        new Column("data_out_blocks", "bigint unsigned"),
        new Column("write_data", "bigint unsigned"),
        new Column("cmd_transfered", "bigint unsigned")
    };
    
    public static final Column[] SAMPLE_COLUMNS = new Column[]{
            new Column("dataInBlocks", "bigint unsigned"),
            new Column("dataOutBlocks", "bigint unsigned"),
            new Column("iscsiReadOps", "float"),
            new Column("iscsiWriteOps", "float"),
            new Column("protocolErrors", "float"),
            new Column("readData", "float"),
            new Column("writeData", "float"),
            new Column("cmdTransfered", "bigint unsigned"),
            new Column("avgLatency", "float"),
            new Column("avgOtherLatency", "float"),
            new Column("avgReadLatency", "float"),
            new Column("avgWriteLatency", "float"),
            new Column("iscsiOtherOps", "float"),
            new Column("opmIscsiLifOps", "float"),
            new Column("opmIscsiLifthroughput", "float")
    };

    public static final Column[] SUMMARY_COLUMNS = new Column[]{
            new Column("dataInBlocks", "bigint unsigned"),
            new Column("dataOutBlocks", "bigint unsigned"),
            new Column("iscsiReadOps", "float"),
            new Column("iscsiWriteOps", "float"),
            new Column("protocolErrors", "float"),
            new Column("readData", "float"),
            new Column("writeData", "float"),
            new Column("cmdTransfered", "bigint unsigned"),
            new Column("avgLatency", "float"),
            new Column("avgOtherLatency", "float"),
            new Column("avgReadLatency", "float"),
            new Column("avgWriteLatency", "float"),
            new Column("iscsiOtherOps", "float"),
            new Column("opmIscsiLifOps", "float"),
            new Column("opmIscsiLifthroughput", "float")
    };

    public static final SqlMethod[][] SAMPLE_TO_SUMMARY_METHODS = new SqlMethod[][]{
            {SqlMethod.SUM},
            {SqlMethod.SUM},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {SqlMethod.SUM},
            {new RatioMethod("cmdTransfered")},
            {new RatioMethod("iscsiOtherOps")},
            {new RatioMethod("iscsiReadOps")},
            {new RatioMethod("iscsiWriteOps")},
            {SqlMethod.AVG},
            {SqlMethod.AVG},
            {SqlMethod.AVG}
    };

    public static final SqlMethod[] SUMMARY_COMMON_METHODS = new SqlMethod[]{
            SqlMethod.SUM,
            SqlMethod.SUM,
            SqlMethod.AVG,
            SqlMethod.AVG,
            SqlMethod.AVG,
            SqlMethod.AVG,
            SqlMethod.AVG,
            SqlMethod.SUM,
            new RatioMethod("cmdTransfered"),
            new RatioMethod("iscsiOtherOps"),
            new RatioMethod("iscsiReadOps"),
            new RatioMethod("iscsiWriteOps"),
            SqlMethod.AVG,
            SqlMethod.AVG,
            SqlMethod.AVG
    };

    private static ObjectParser<IscsiLifRawMeasurements> rawParser;
    private static ObjectParser<IscsiLifMeasurements> sampleParser;

    static {
        rawParser = new ObjectParser<IscsiLifRawMeasurements>() {
            public void prepareObjectDataForInsert(PreparedStatementWrapper preparedStatement, IscsiLifRawMeasurements objectData) throws TableException {
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getAvgLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getAvgOtherLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getAvgReadLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getAvgWriteLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getIscsiOtherOps());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getIscsiReadOps());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getIscsiWriteOps());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getDataInBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getReadData());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getDataOutBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getWriteData());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getCmdTransfered());
            }

            public IscsiLifRawMeasurements parseObjectData(ResultSet resultSet, int startIndex) throws SQLException {
                int i = startIndex - 1;
                return new IscsiLifRawMeasurements()
                    .setAvgLatency(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setAvgOtherLatency(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setAvgReadLatency(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setAvgWriteLatency(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setIscsiOtherOps(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setIscsiReadOps(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setIscsiWriteOps(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setDataInBlocks(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setReadData(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setDataOutBlocks(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setWriteData(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i))
                    .setCmdTransfered(PerformanceTableUtils.getBigIntegerNullable(resultSet, ++i));
            }
        };
        
        sampleParser = new ObjectParser<IscsiLifMeasurements>() {
            public void prepareObjectDataForInsert(PreparedStatementWrapper preparedStatement, IscsiLifMeasurements objectData) throws TableException {
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getDataInBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getDataOutBlocks());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getIscsiReadOps());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getIscsiWriteOps());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getProtocolErrors());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getReadData());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getWriteData());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getCmdTransfered());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getAvgLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getAvgOtherLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getAvgReadLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getAvgWriteLatency());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getIscsiOtherOps());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getOps());
                PerformanceTableUtils.setNextNullable(preparedStatement, objectData.getThroughput());
            }

            public IscsiLifMeasurements parseObjectData(ResultSet resultSet, int startIndex) throws SQLException {
                int i = startIndex - 1;
                return new IscsiLifMeasurements(
                        PerformanceTableUtils.getLongNullable(resultSet, ++i),
                        PerformanceTableUtils.getLongNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getLongNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i),
                        PerformanceTableUtils.getFloatNullable(resultSet, ++i));
            }
        };
    }

    public static ObjectParser<IscsiLifMeasurements> getSummaryParser() {
        return sampleParser;
    }

    public static ObjectParser<IscsiLifMeasurements> getSampleParser() {
        return sampleParser;
    }

    public static ObjectParser<IscsiLifRawMeasurements> getRawParser() {
        return rawParser;
    }
}
