/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.iscsilif.interfaces.impl;

import java.sql.Connection;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.db.NetAppPerformance;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.originator.handler.LifHandler;
import com.netapp.oci.netapp.originator.handler.VserverHandler;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifMeasurements;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifRawMeasurements;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.remote.IscsiLifPerformanceRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.shared.SimpleRawSampleConverter;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceDatasourceRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry.SampleType;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.SimpleUuidKey;
import com.onaro.sanscreen.util.Constant;

/**
 * @author Al Velella
 *         Date: Jul 31, 2012
 *         Time: 5:29:17 PM
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class IscsiLifPerformanceBean implements IscsiLifPerformanceRemote {
    public static final String PACKAGE_NAME = "iscsilif";
    public static final String PACKAGE_NAME_RAW = PACKAGE_NAME + PerformanceConstants.RAW_PACKAGE_SUFFIX;
    private static final String ISCSILIF_PACKAGE = PerformanceConstants.PACKAGE_PREFIX + PACKAGE_NAME;
    private static final Logger logger = LoggerFactory.getLogger(ISCSILIF_PACKAGE);

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * use this bean to register reporting datasource
     */
    @EJB
    private NetappPerformanceDatasourceRegistry datasourceRegistry;

    /**
     * use this bean to get the registered listeners
     */
    @EJB
    private NetappPerformanceSampleListenerRegistry sampleListenerRegistry;
    
    public void reportIscsiLifSamples(Long datasourceId, String clusterKey, String vserverKey, long time, Collection<RawSample<IscsiLifMeasurements>> rawSamples, long leaseTime, boolean isHistorical) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();
        Long containerId = null;
        Collection<Sample<IscsiLifMeasurements>> samples = null;

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(ISCSILIF_PACKAGE, vserverKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(ISCSILIF_PACKAGE, vserverKey, datasourceId, leaseTime)) {
                logger.info("Vserver key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(ISCSILIF_PACKAGE, vserverKey, leaseTime) + " (vserverKey:" + vserverKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            containerId = keyCache.findId(originId, VserverHandler.TAG, vserverKey);
            if (containerId == null) {
                logger.warn("Failed to find containerId from vserverKey:" + vserverKey + " datasourceId:" + datasourceId);
            } else {
                samples = new SimpleRawSampleConverter<IscsiLifMeasurements>(containerId, keyCache, LifHandler.SubTag.NETWORK_LIF.getTagName(), originId).toSamples(rawSamples, time, firstReport);
                if(samples.isEmpty()){
                    logger.warn("Failed to convert IscsiLif raw-samples retrieved from the acquisiton into server-samples. clusterKey: {} datasourceId: {}", clusterKey, datasourceId);
                    return;
                }
                ComponentsFactoryRepository.getIscsiLifPerformanceComponentsFactory().createPersistenceFacade(connection, vserverKey, containerId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME, datasourceId, startTime, System.currentTimeMillis() - startTime,
                        vserverKey, rawSamples.size(), samples.size(), leaseTime);
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert iscsilif performance, vserverKey:" + vserverKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }
        
        // invoke all the registered listeners for this sample
        try {
        	if (samples != null) {	        
        		Map<String, PerformanceSampleListener> listeners = sampleListenerRegistry.getSampleListeners(SampleType.ISCSI_LIF);                
        		if (listeners != null) {
        			logger.info("There are " + listeners.size() + " listeners for iscsiLif stats.");
        			Map<Long, Collection<Sample<IscsiLifMeasurements>>> rslt = new HashMap<>();
        			rslt.put(containerId, samples);

        			long begin = System.currentTimeMillis();                	
        			for (String listenerName : listeners.keySet()) {
        				long iBegin = System.currentTimeMillis();
        				listeners.get(listenerName).processIscsiLif(rslt, isHistorical);
        				long iDelta = System.currentTimeMillis() - iBegin;
        				logger.debug(listenerName + " took " + iDelta + " milliseconds to process iscsiLif samples.");
        			}                	
        			long delta = System.currentTimeMillis() - begin;

        			if (delta > PerformanceConstants.SAMPLE_LISTENER_MAX_ELAPSED_TIME) {
        				logger.warn("It took " + delta + " milliseconds for the iscsiLif sample listeners to complete their process.");
        			}
        		}                
        	}
        } catch(Exception e) {
        	logger.error("Unable to invoke registered listeners for iscsiLif; " + e);
        }

    }

    public void reportIscsiLifRawSamples(Long datasourceId, String clusterKey, String vserverKey, long time, Collection<RawSample<IscsiLifRawMeasurements>> rawSamples, long leaseTime) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(ISCSILIF_PACKAGE, vserverKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(ISCSILIF_PACKAGE, vserverKey, datasourceId, leaseTime)) {
                logger.info("Vserver key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(ISCSILIF_PACKAGE, vserverKey, leaseTime) + " (vserverKey:" + vserverKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            Long vserverId = keyCache.findId(originId, VserverHandler.TAG, vserverKey);
            if (vserverId == null) {
                logger.warn("Failed to find vserver id from vserverKey:" + vserverKey + " datasourceId:" + datasourceId);
            } else {
                Collection<Sample<IscsiLifRawMeasurements>> samples
                        = new SimpleRawSampleConverter<IscsiLifRawMeasurements>(vserverId, keyCache, LifHandler.SubTag.NETWORK_LIF.getTagName(), originId).toSamples(rawSamples, time, firstReport);
                if(samples.isEmpty()){
                    logger.warn("Failed to convert IscsiLif raw-samples retrieved from the acquisiton into server-samples. clusterKey: {} datasourceId: {}", clusterKey, datasourceId);
                    return;
                }
                ComponentsFactoryRepository.getIscsiLifRawPerformanceComponentsFactory().createPersistenceFacade(connection, vserverKey, vserverId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME_RAW, datasourceId, startTime, System.currentTimeMillis() - startTime,
                        vserverKey, rawSamples.size(), samples.size(), leaseTime);
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert iscsilif raw performance, vserverKey:" + vserverKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
