/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.iscsilif.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifMeasurements;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.local.IscsiLifPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.remote.IscsiLifPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Al Velella
 *         Date: Jul 31, 2012
 *         Time: 5:29:40 PM
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class IscsiLifPerformanceQueryBean implements IscsiLifPerformanceQueryRemote, IscsiLifPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<IscsiLifMeasurements>>> getSamples(Set<Long> iscsiLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + iscsiLifIds.size() + "iscsi lif samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<IscsiLifMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<IscsiLifMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(iscsiLifIds);

            Map<Long, Set<Long>> vserverIdToIscsiLifIds = NetAppModel.INSTANCE.getVserverIdByNetworkLifIds(connection, iscsiLifIds);
            for (Map.Entry<Long, Set<Long>> entry : vserverIdToIscsiLifIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, fromTime, toTime, ids));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find iscsi lif with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsi lif samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query lif samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<IscsiLifMeasurements>>> getSamples(Long containerId, Set<Long> iscsiLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + iscsiLifIds.size() + "iscsi lif samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, fromTime, toTime, iscsiLifIds);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsi lif samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query lif samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<IscsiLifMeasurements>> getSamples(Long containerId, Long iscsiLifId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(iscsiLifId), fromTime, toTime).get(iscsiLifId);

    }

    private Map<Long, List<SampleInfo<IscsiLifMeasurements>>> getSamples(Connection connection, Long containerId, Date fromTime, Date toTime, Set<Long> ids) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getIscsiLifPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(ids, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<IscsiLifMeasurements>> getLatestSample(Long vServerId, Set<Long> iscsiLifIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + iscsiLifIds.size() + "iscsiLif samples for vServer:" + vServerId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getIscsiLifPerformanceComponentsFactory().createPersistenceFacade(connection, vServerId)
                            .getLatestSample(iscsiLifIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsiLif samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<IscsiLifMeasurements>>> getSummaries(Set<Long> iscsiLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + iscsiLifIds.size() + "iscsi lif summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<IscsiLifMeasurements>>>  idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(iscsiLifIds);

            Map<Long, Set<Long>> vserverIdToIscsiLifIds = NetAppModel.INSTANCE.getVserverIdByNetworkLifIds(connection, iscsiLifIds);
            for (Map.Entry<Long, Set<Long>> entry : vserverIdToIscsiLifIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<IscsiLifMeasurements>> summaries = ComponentsFactoryRepository.getIscsiLifPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find iscsi lif with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsi lif summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<IscsiLifMeasurements>>> getSummaries(Long vServerId, Set<Long> iscsiLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + iscsiLifIds.size() + "iscsi lif summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<IscsiLifMeasurements>>>  idToSummaries = new HashMap<>();
            List<Summary<IscsiLifMeasurements>> summaries = ComponentsFactoryRepository.getIscsiLifPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, vServerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), iscsiLifIds);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsi lif summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "iscsi lif has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getIscsiLifPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if iscsi lif has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
