package com.netapp.oci.netapp.performance.iscsilif.interfaces.impl;

import java.sql.Connection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifVserverMeasurements;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.local.IscsiLifVserverPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.remote.IscsiLifVserverPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)

public class IscsiLifVserverPerformanceQueryBean implements IscsiLifVserverPerformanceQueryRemote, IscsiLifVserverPerformanceQueryLocal{
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<IscsiLifVserverMeasurements>>> getSamples(Set<Long> iscsiLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + iscsiLifIds.size() + "iscsiLifVserver samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<IscsiLifVserverMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<IscsiLifVserverMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(iscsiLifIds);

            Map<Long, Set<Long>> clusterIdToFcpLifIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, iscsiLifIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToFcpLifIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, fromTime, toTime, ids));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find iscsiLifVserver with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsiLifVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query lif samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<IscsiLifVserverMeasurements>>> getSamples(Long containerId, Set<Long> iscsiLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + iscsiLifIds.size() + "iscsiLifVserver samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, fromTime, toTime, iscsiLifIds);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsiLifVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query iscsiLifVserver samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<IscsiLifVserverMeasurements>> getSamples(Long containerId, Long iscsiLifId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(iscsiLifId), fromTime, toTime).get(iscsiLifId);

    }

    private Map<Long, List<SampleInfo<IscsiLifVserverMeasurements>>> getSamples(Connection connection, Long containerId, Date fromTime, Date toTime, Set<Long> ids) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getIscsiLifVserverPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(ids, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<IscsiLifVserverMeasurements>> getLatestSample(Long clusterId, Set<Long> iscsiLifIds) throws NetappInterfaceException {
        logger.info("Query " + iscsiLifIds.size() + "iscsiLifVserver samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getIscsiLifVserverPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(iscsiLifIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsiLif samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    public Map<Long, List<Summary<IscsiLifVserverMeasurements>>> getSummaries(Set<Long> vServerIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vServerIds.size() + "iscsiLifVserver summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<IscsiLifVserverMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(vServerIds);

            Map<Long, Set<Long>> clusterIdToFcpLifIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, vServerIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToFcpLifIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<IscsiLifVserverMeasurements>> summaries = ComponentsFactoryRepository.getIscsiLifVserverPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find iscsiLifVserver with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsiLifVserver summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<IscsiLifVserverMeasurements>>> getSummaries(Long clusterId, Set<Long> vServerIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vServerIds.size() + "iscsiLifVserver summaries from " + fromTime + " to " + toTime + " for containerId:" + clusterId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<IscsiLifVserverMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<IscsiLifVserverMeasurements>> summaries = ComponentsFactoryRepository.getIscsiLifVserverPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, clusterId)
                    .getSummaries(new TimeSpan(fromTime, toTime), vServerIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get iscsiLifVserver summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "iscsiLifVserver has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getIscsiLifVserverPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if iscsiLifVserver has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
