/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.lun.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.lun.interfaces.data.LunMeasurements;
import com.netapp.oci.netapp.performance.lun.interfaces.local.LunPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.lun.interfaces.remote.LunPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class LunPerformanceQueryBean implements LunPerformanceQueryRemote, LunPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<LunMeasurements>>> getSamples(Set<Long> lunIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + lunIds.size() + "lun samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<LunMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<LunMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(lunIds);

            Map<Long, Set<Long>> clusterIdToLunIds = NetAppModel.INSTANCE.getClusterIdByLunIds(connection, lunIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToLunIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, fromTime, toTime, ids));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find lun with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lun samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query lun samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<LunMeasurements>>> getSamples(Long containerId, Set<Long> lunIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + lunIds.size() + "lun samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, fromTime, toTime, lunIds);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lun samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query lun samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public List<SampleInfo<LunMeasurements>> getSamples(Long containerId, Long lunId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(lunId), fromTime, toTime).get(lunId);

    }

    private Map<Long, List<SampleInfo<LunMeasurements>>> getSamples(Connection connection, Long containerId, Date
            fromTime, Date toTime, Set<Long> ids) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getLunPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(ids, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<LunMeasurements>> getLatestSample(Long clusterId, Set<Long> lunIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + lunIds.size() + "lun samples for vServer:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getLunPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(lunIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lun samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<LunMeasurements>>> getSummaries(Set<Long> lunIds, Date fromTime, Date toTime) throws
            NetappInterfaceException {
        logger.info("Query " + lunIds.size() + "lun samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<LunMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(lunIds);

            Map<Long, Set<Long>> clusterIdToLunIds = NetAppModel.INSTANCE.getClusterIdByLunIds(connection, lunIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToLunIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<LunMeasurements>> summaries = ComponentsFactoryRepository.getLunPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find lun with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lun summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<LunMeasurements>>> getSummaries(Long clusterId, Set<Long> lunIds, Date fromTime, Date toTime) throws
            NetappInterfaceException {
        logger.info("Query " + lunIds.size() + "lun samples from " + fromTime + " to " + toTime + ", for vServer id "
                + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<LunMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<LunMeasurements>> summaries = ComponentsFactoryRepository.getLunPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, clusterId)
                    .getSummaries(new TimeSpan(fromTime, toTime), lunIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lun summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "lun has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getLunPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if lun has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
