/*
 * $Id$
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.oci.netapp.performance.namespace.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.namespace.interfaces.data.NamespaceMeasurements;
import com.netapp.oci.netapp.performance.namespace.interfaces.local.NamespacePerformanceQueryLocal;
import com.netapp.oci.netapp.performance.namespace.interfaces.remote.NamespacePerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class NamespacePerformanceQueryBean implements NamespacePerformanceQueryRemote, NamespacePerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<NamespaceMeasurements>>> getSamples(Set<Long> namespaceIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + namespaceIds.size() + "Namespace samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<NamespaceMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<NamespaceMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(namespaceIds);

            Map<Long, Set<Long>> clusterIdToNamespaceIds = NetAppModel.INSTANCE.getClusterIdByNamespaceIds(connection, namespaceIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToNamespaceIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, fromTime, toTime, ids));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find Namespace with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Namespace samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query Namespace samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<NamespaceMeasurements>>> getSamples(Long containerId, Set<Long> namespaceIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + namespaceIds.size() + "Namespace samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, fromTime, toTime, namespaceIds);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Namespace samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query Namespace samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public List<SampleInfo<NamespaceMeasurements>> getSamples(Long containerId, Long namespaceId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(namespaceId), fromTime, toTime).get(namespaceId);

    }

    private Map<Long, List<SampleInfo<NamespaceMeasurements>>> getSamples(Connection connection, Long containerId, Date
            fromTime, Date toTime, Set<Long> ids) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getNamespacePerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(ids, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<NamespaceMeasurements>> getLatestSample(Long clusterId, Set<Long> namespaceIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + namespaceIds.size() + "Namespace samples for vServer:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getNamespacePerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(namespaceIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Namespace samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NamespaceMeasurements>>> getSummaries(Set<Long> namespaceIds, Date fromTime, Date toTime) throws
            NetappInterfaceException {
        logger.info("Query " + namespaceIds.size() + "Namespace samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<NamespaceMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(namespaceIds);

            Map<Long, Set<Long>> clusterIdToNamespaceIds = NetAppModel.INSTANCE.getClusterIdByNamespaceIds(connection, namespaceIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToNamespaceIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<NamespaceMeasurements>> summaries = ComponentsFactoryRepository.getNamespacePerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find Namespace with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Namespace summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NamespaceMeasurements>>> getSummaries(Long clusterId, Set<Long> namespaceIds, Date fromTime, Date toTime) throws
            NetappInterfaceException {
        logger.info("Query " + namespaceIds.size() + "Namespace samples from " + fromTime + " to " + toTime + ", for vServer id "
                + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<NamespaceMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<NamespaceMeasurements>> summaries = ComponentsFactoryRepository.getNamespacePerformanceComponentsFactory()
                    .createPersistenceFacade(connection, clusterId)
                    .getSummaries(new TimeSpan(fromTime, toTime), namespaceIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Namespace summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "Namespace has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getNamespacePerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if Namespace has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
