/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.networklif.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.networklif.interfaces.data.NetworkLifMeasurements;
import com.netapp.oci.netapp.performance.networklif.interfaces.local.NetworkLifPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.networklif.interfaces.remote.NetworkLifPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Nancy Walia
 *         Date: Feb 13, 2012
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class NetworkLifPerformanceQueryBean implements NetworkLifPerformanceQueryRemote, NetworkLifPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<NetworkLifMeasurements>>> getSamples(Set<Long> networkLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + networkLifIds.size() + "network lif samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<NetworkLifMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<NetworkLifMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(networkLifIds);

            Map<Long, Set<Long>> vserverIdToNetworkLifIds = NetAppModel.INSTANCE.getVserverIdByNetworkLifIds(connection, networkLifIds);
            for (Map.Entry<Long, Set<Long>> entry : vserverIdToNetworkLifIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, fromTime, toTime, ids));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find network lif with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get networkLif samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query network port samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<NetworkLifMeasurements>>> getSamples(Long containerId, Set<Long> networkLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + networkLifIds.size() + "network lif samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, fromTime, toTime, networkLifIds);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get networkLif samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query network port samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public List<SampleInfo<NetworkLifMeasurements>> getSamples(Long containerId, Long networkLifId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(networkLifId), fromTime, toTime).get(networkLifId);

    }

    private Map<Long, List<SampleInfo<NetworkLifMeasurements>>> getSamples(Connection connection, Long containerId, Date fromTime, Date toTime, Set<Long> ids) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getNetworkLifPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(ids, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<NetworkLifMeasurements>> getLatestSample(Long vServerId, Set<Long> networkLifIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + networkLifIds.size() + "networkLif samples for vServer:" + vServerId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getNetworkLifPerformanceComponentsFactory().createPersistenceFacade(connection, vServerId)
                            .getLatestSample(networkLifIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get networkLif samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NetworkLifMeasurements>>> getSummaries(Set<Long> networkLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + networkLifIds.size() + "network lif summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<NetworkLifMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(networkLifIds);

            Map<Long, Set<Long>> vserverIdToNetworkLifIds = NetAppModel.INSTANCE.getVserverIdByNetworkLifIds(connection, networkLifIds);
            for (Map.Entry<Long, Set<Long>> entry : vserverIdToNetworkLifIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<NetworkLifMeasurements>> summaries = ComponentsFactoryRepository.getNetworkLifPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find network lif with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get networkLif summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NetworkLifMeasurements>>> getSummaries(Long vServerId, Set<Long> networkLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + networkLifIds.size() + "network lif summaries from " + fromTime + " to " + toTime +
        "for vServer Id" + vServerId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<NetworkLifMeasurements>>> idToSummaries = new HashMap<>();
                List<Summary<NetworkLifMeasurements>> summaries = ComponentsFactoryRepository.getNetworkLifPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, vServerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), networkLifIds);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get networkLif summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "network lif has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getNetworkLifPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if network lif has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}

