package com.netapp.oci.netapp.performance.networklif.interfaces.impl;

import java.sql.Connection;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.db.NetAppPerformance;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.originator.handler.ClusterHandler;
import com.netapp.oci.netapp.originator.handler.VserverHandler;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.networklif.interfaces.data.NetworkLifVserverMeasurements;
import com.netapp.oci.netapp.performance.networklif.interfaces.data.NetworkLifVserverRawMeasurements;
import com.netapp.oci.netapp.performance.networklif.interfaces.remote.NetworkLifVserverPerformanceRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.shared.SimpleRawSampleConverter;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceDatasourceRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry.SampleType;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.SimpleUuidKey;
import com.onaro.sanscreen.util.Constant;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class NetworkLifVserverPerformanceBean implements NetworkLifVserverPerformanceRemote {
    public static final String PACKAGE_NAME = "networklifVserver";
    public static final String PACKAGE_NAME_RAW = PACKAGE_NAME + PerformanceConstants.RAW_PACKAGE_SUFFIX;
    private static final String NETWORKLIF_PACKAGE = PerformanceConstants.PACKAGE_PREFIX + PACKAGE_NAME;
    private static final Logger logger = LoggerFactory.getLogger(NETWORKLIF_PACKAGE);

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * use this bean to register reporting datasource
     */
    @EJB
    private NetappPerformanceDatasourceRegistry datasourceRegistry;

    /**
     * use this bean to get the registered listeners
     */
    @EJB
    private NetappPerformanceSampleListenerRegistry sampleListenerRegistry;
    
	@Override
	public void reportLifVserverRawData(Long datasourceId, String clusterKey, long time, Collection<RawSample<NetworkLifVserverRawMeasurements>> rawSamples, long leaseTime) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(NETWORKLIF_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(NETWORKLIF_PACKAGE, clusterKey, datasourceId, leaseTime)) {
                logger.info("Vserver key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(NETWORKLIF_PACKAGE, clusterKey, leaseTime) + " (clusterKey:" + clusterKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            Long clusterId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (clusterId == null) {
            	logger.warn("Failed to find cluster id from clusterKey:" + clusterKey + " datasourceId:" + datasourceId);
            } else {
            	// Convert sample uuids to object ids using the key cache
            	// VserverHandler.TAG is used to identify the object to look up in the key cache 
                Collection<Sample<NetworkLifVserverRawMeasurements>> samples
                        = new SimpleRawSampleConverter<NetworkLifVserverRawMeasurements>(clusterId, keyCache, VserverHandler.TAG, originId).toSamples(rawSamples, time, firstReport);
            	ComponentsFactoryRepository.getLifVserverPerformanceRawComponentsFactory().createPersistenceFacade(connection, clusterKey, clusterId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME_RAW, datasourceId, startTime, System.currentTimeMillis() - startTime,
                        clusterKey, rawSamples.size(), samples.size(), leaseTime);                
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert raw cifs:vserver performance, clusterKey:" + clusterKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }				
	}

	@Override
	public void reportLifVserverData(Long datasourceId, String clusterKey, long time, Collection<RawSample<NetworkLifVserverMeasurements>> rawSamples, long leaseTime, boolean isHistorical) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();
        Long containerId = null;
        Collection<Sample<NetworkLifVserverMeasurements>> samples = null;

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(NETWORKLIF_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(NETWORKLIF_PACKAGE, clusterKey, datasourceId, leaseTime)) {
                logger.info("Vserver key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(NETWORKLIF_PACKAGE, clusterKey, leaseTime) + " (clusterKey:" + clusterKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            containerId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (containerId == null) {
                logger.warn("Failed to find containerId from clusterKey:" + clusterKey + " datasourceId:" + datasourceId);
            } else {
                samples = new SimpleRawSampleConverter<NetworkLifVserverMeasurements>(containerId, keyCache, VserverHandler.TAG, originId).toSamples(rawSamples, time, firstReport);
                ComponentsFactoryRepository.getLifVserverPerformanceComponentsFactory().createPersistenceFacade(connection, clusterKey, containerId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME, datasourceId, startTime, System.currentTimeMillis() - startTime,
                		clusterKey, rawSamples.size(), samples.size(), leaseTime);
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert networklif performance, clusterKey:" + clusterKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }
                
        // invoke all the registered listeners for this sample
        try {
        	if (samples != null) {
        		Map<String, PerformanceSampleListener> listeners = sampleListenerRegistry.getSampleListeners(SampleType.NETWORK_LIF_VSERVER);                
        		if (listeners != null) {
        			logger.info("There are " + listeners.size() + " listeners for networkLifVserver stats.");
        			Map<Long, Collection<Sample<NetworkLifVserverMeasurements>>> rslt = new HashMap<>();
        			rslt.put(containerId, samples);

        			long begin = System.currentTimeMillis();                	
        			for (String listenerName : listeners.keySet()) {
        				long iBegin = System.currentTimeMillis();
        				listeners.get(listenerName).processNetworkLifVserver(rslt, isHistorical);
        				long iDelta = System.currentTimeMillis() - iBegin;
        				logger.debug(listenerName + " took " + iDelta + " milliseconds to process networkLifVserver samples.");
        			}                	
        			long delta = System.currentTimeMillis() - begin;

        			if (delta > PerformanceConstants.SAMPLE_LISTENER_MAX_ELAPSED_TIME) {
        				logger.warn("It took " + delta + " milliseconds for the networkLifVserver sample listeners to complete their process.");
        			}
        		}                                                                                                
        	}
        } catch(Exception e) {
        	logger.error("Unable to invoke registered listeners for networkLifVserver; ", e);
        }
	}
}

