package com.netapp.oci.netapp.performance.networklif.interfaces.impl;

import java.sql.Connection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.networklif.interfaces.data.NetworkLifVserverMeasurements;
import com.netapp.oci.netapp.performance.networklif.interfaces.local.NetworkLifVserverPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.networklif.interfaces.remote.NetworkLifVserverPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import com.onaro.sanscreen.util.Constant;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class NetworkLifVserverPerformanceQueryBean implements NetworkLifVserverPerformanceQueryRemote, NetworkLifVserverPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<NetworkLifVserverMeasurements>>> getSamples(Set<Long> lifVserverIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + lifVserverIds.size() + "lifVserver samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            
            Map<Long, List<SampleInfo<NetworkLifVserverMeasurements>>> containerIdToSamples = new HashMap<>();
              
            Set<Long> remainingIds = new HashSet<Long>(lifVserverIds);
            Map<Long, Set<Long>> clusterIdToLifVserverIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, lifVserverIds);
            
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToLifVserverIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find lifVserver with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            
            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lifVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query lifVserver samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public Map<Long, List<SampleInfo<NetworkLifVserverMeasurements>>> getSamples(Long containerId, Set<Long> lifVserverIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + lifVserverIds.size() + "lifVserver samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, lifVserverIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lifVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query lifVserver samples: " + (System.currentTimeMillis() - startTime));
        }
    }
    

    @Override
    public List<SampleInfo<NetworkLifVserverMeasurements>> getSamples(Long containerId, Long lifVserverIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(lifVserverIds), fromTime, toTime).get(lifVserverIds);

    }
    
    private Map<Long, List<SampleInfo<NetworkLifVserverMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> lifVserverIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getLifVserverPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(lifVserverIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<NetworkLifVserverMeasurements>> getLatestSample(Long vserverId, Set<Long> lifVserverIds) throws NetappInterfaceException {
        logger.info("Query " + lifVserverIds.size() + "lifVserver samples for vserver:" + vserverId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getLifVserverPerformanceComponentsFactory().createPersistenceFacade(connection, vserverId)
                            .getLatestSample(lifVserverIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lifVserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NetworkLifVserverMeasurements>>> getSummaries(Set<Long> vServerIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vServerIds.size() + " lifVserver summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<NetworkLifVserverMeasurements>>> idToSummaries = new HashMap<>();

            Set<Long> remainingIds = new HashSet<>(vServerIds);
            Map<Long, Set<Long>> clusterIdToLifVserverIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, vServerIds);

            for (Map.Entry<Long, Set<Long>> entry : clusterIdToLifVserverIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();
                remainingIds.removeAll(ids);
                List<Summary<NetworkLifVserverMeasurements>> summaries = ComponentsFactoryRepository.getLifVserverPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find lifVserver with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lifVserver summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NetworkLifVserverMeasurements>>> getSummaries(Long clusterId, Set<Long> vServerIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vServerIds.size() + " lifVserver summaries from " + fromTime + " to " + toTime +
        " for cluster id " + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<NetworkLifVserverMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<NetworkLifVserverMeasurements>> summaries = ComponentsFactoryRepository.getLifVserverPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), vServerIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get lifVserver summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
    	logger.info("Query if " + objId + "lifVserver has counters from " + fromTime + " to " + toTime);

    	Connection connection = null;
    	try {
    		connection = ConnectionHelper.newConnection(dataSource);
    		return ComponentsFactoryRepository.getLifVserverPerformanceComponentsFactory()
    				.createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
    	} catch (Exception e) {
    		throw new NetappInterfaceException("Failed to assert if lifVserver has counters ", e);
    	} finally {
    		ConnectionHelper.close(connection);
    	}
    }
	
}
