/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.networkport.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.networkport.interfaces.data.NetworkPortMeasurements;
import com.netapp.oci.netapp.performance.networkport.interfaces.local.NetworkPortPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.networkport.interfaces.remote.NetworkPortPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Nancy Walia
 */

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class NetworkPortPerformanceQueryBean implements NetworkPortPerformanceQueryRemote, NetworkPortPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    public Map<Long, List<SampleInfo<NetworkPortMeasurements>>> getSamples(Set<Long> networkPortIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + networkPortIds.size() + "network port samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<NetworkPortMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<NetworkPortMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(networkPortIds);

            Map<Long, Set<Long>> clusterIdToNetworkPortIds = NetAppModel.INSTANCE.getClusterIdByNetworkPortIds(connection, networkPortIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToNetworkPortIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find networkPort with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get network port samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query network port samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<NetworkPortMeasurements>>> getSamples(Long containerId, Set<Long> networkPortIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + networkPortIds.size() + "network port samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, networkPortIds, fromTime, toTime);

        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get network port samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query node samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public List<SampleInfo<NetworkPortMeasurements>> getSamples(Long containerId, Long networkPortId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(networkPortId), fromTime, toTime).get(networkPortId);

    }

    private Map<Long, List<SampleInfo<NetworkPortMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> networkPortIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getNetworkPortPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(networkPortIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<NetworkPortMeasurements>> getLatestSample(Long clusterId, Set<Long> networkPortIds) throws NetappInterfaceException {
        logger.info("Query " + networkPortIds.size() + "networkPort samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getNetworkPortPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(networkPortIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get networkPort samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NetworkPortMeasurements>>> getSummaries(Set<Long> networkPortIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + networkPortIds.size() + "network port summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<NetworkPortMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(networkPortIds);

            Map<Long, Set<Long>> clusterIdToNetworkPortIds = NetAppModel.INSTANCE.getClusterIdByNetworkPortIds(connection, networkPortIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToNetworkPortIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<NetworkPortMeasurements>> summaries = ComponentsFactoryRepository.getNetworkPortPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find networkPort with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get network port summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NetworkPortMeasurements>>> getSummaries(Long clusterId, Set<Long> networkPortIds,
                                                                          Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + networkPortIds.size() + "network port summaries from " + fromTime + " to " + toTime +
        " for cluster id " + clusterId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<NetworkPortMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<NetworkPortMeasurements>> summaries = ComponentsFactoryRepository.getNetworkPortPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), networkPortIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get network port summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "network port has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getNetworkPortPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if network port has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
