package com.netapp.oci.netapp.performance.nfs.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppPerformance;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.originator.handler.ClusterHandler;
import com.netapp.oci.netapp.originator.handler.VserverHandler;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.nfs.interfaces.data.NfsMeasurements;
import com.netapp.oci.netapp.performance.nfs.interfaces.remote.NfsPerformanceRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.shared.SimpleRawSampleConverter;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceDatasourceRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry.SampleType;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.SimpleUuidKey;
import com.onaro.sanscreen.util.Constant;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class NfsPerformanceBean implements NfsPerformanceRemote {
    public static final String PACKAGE_NAME = "nfs";
    public static final String PACKAGE_NAME_RAW = PACKAGE_NAME + PerformanceConstants.RAW_PACKAGE_SUFFIX;

    private static final String NFS_PACKAGE = PerformanceConstants.PACKAGE_PREFIX + PACKAGE_NAME;
    private static final Logger LOG = LoggerFactory.getLogger(NFS_PACKAGE);

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = DataSource.class,
            authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * use this bean to register reporting data source
     */
    @EJB
    private NetappPerformanceDatasourceRegistry dataSourceRegistry;

    /**
     * use this bean to get the registered listeners
     */
    @EJB
    private NetappPerformanceSampleListenerRegistry sampleListenerRegistry;

    @Override
    public void reportNfsData(Long dataSourceId, String clusterKey, long time,
                              Collection<RawSample<NfsMeasurements>> rawSamples, long leaseTime, boolean isHistorical)
            throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();
        Long clusterId = null;
        Collection<Sample<NfsMeasurements>> samples = null;

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !dataSourceRegistry.isRegistered(NFS_PACKAGE, clusterKey, dataSourceId, leaseTime);
            if (!dataSourceRegistry.register(NFS_PACKAGE, clusterKey, dataSourceId, leaseTime)) {
                LOG.info("Cluster key is already registered by:" + dataSourceRegistry.getRegisterDatasourceId(NFS_PACKAGE,
                        clusterKey, leaseTime) + " (clusterKey:" + clusterKey + " dataSourceId:" + dataSourceId + ")");
                return;
            }
            connection = ConnectionHelper.newConnection(dataSource);
            KeyCache keyCache = KeyCache.get();
            long originId = keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            clusterId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (clusterId == null) {
                LOG.warn("Failed to find clusterId from clusterKey:" + clusterKey + " dataSourceId:" + dataSourceId);
            } else {
                // Convert sample uuids to object ids using the key cache
                // AggregateHandler.TAG is used to identify the object to look up in the key cache
                samples = new SimpleRawSampleConverter<NfsMeasurements>(clusterId, keyCache, VserverHandler.TAG, originId)
                        .toSamples(rawSamples, time, firstReport);
                if (samples.isEmpty()) {
                    LOG.warn("Failed to convert Nfs raw-samples retrieved from the acquisition into server-samples. " +
                            "clusterKey: {} dataSourceId: {}", clusterKey, dataSourceId);
                    return;
                }
                ComponentsFactoryRepository.getNfsPerformanceComponentsFactory().createPersistenceFacade(connection,
                        clusterKey, clusterId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME, dataSourceId, startTime,
                        System.currentTimeMillis() - startTime, clusterKey, rawSamples.size(), samples.size(), leaseTime);
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert nfs performance, clusterKey:" +
                    clusterKey + " dataSourceId:" + dataSourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }

        // invoke all the registered listeners for this sample
        try {
            if (samples != null) {
                Map<String, PerformanceSampleListener> listeners = sampleListenerRegistry.getSampleListeners(SampleType.NFS);
                if (listeners != null) {
                    LOG.info("There are " + listeners.size() + " listeners for nfs stats.");
                    Map<Long, Collection<Sample<NfsMeasurements>>> rslt = new HashMap<>();
                    rslt.put(clusterId, samples);

                    long begin = System.currentTimeMillis();
                    for (String listenerName : listeners.keySet()) {
                        long iBegin = System.currentTimeMillis();
                        listeners.get(listenerName).processNfs(rslt, isHistorical);
                        long iDelta = System.currentTimeMillis() - iBegin;
                        LOG.debug(listenerName + " took " + iDelta + " milliseconds to process nfs samples.");
                    }
                    long delta = System.currentTimeMillis() - begin;

                    if (delta > PerformanceConstants.SAMPLE_LISTENER_MAX_ELAPSED_TIME) {
                        LOG.warn("It took " + delta + " milliseconds for the nfs sample listeners to complete their process.");
                    }
                }
            }
        } catch (Exception e) {
            LOG.error("Unable to invoke registered listeners for nfs; ", e);
        }
    }
}
