/* Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.nfs.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.nfs.interfaces.data.Nfsv3Measurements;
import com.netapp.oci.netapp.performance.nfs.interfaces.local.Nfsv3PerformanceQueryLocal;
import com.netapp.oci.netapp.performance.nfs.interfaces.remote.Nfsv3PerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class Nfsv3PerformanceQueryBean implements Nfsv3PerformanceQueryRemote, Nfsv3PerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<Nfsv3Measurements>>> getSamples(Set<Long> nfsIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + nfsIds.size() + "nfs samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<Nfsv3Measurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<Nfsv3Measurements>>>();

            Set<Long> remainingIds = new HashSet<Long>(nfsIds);
            Map<Long, Set<Long>> clusterIdToCifsVserverIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, nfsIds);

            for (Map.Entry<Long, Set<Long>> entry : clusterIdToCifsVserverIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find nfsv3 with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nfs samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query nfs samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public Map<Long, List<SampleInfo<Nfsv3Measurements>>> getSamples(Long containerId, Set<Long> nfsIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + nfsIds.size() + "nfs samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, nfsIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nfs samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query nfs samples: " + (System.currentTimeMillis() - startTime));
        }
    }
    

    @Override
    public List<SampleInfo<Nfsv3Measurements>> getSamples(Long containerId, Long nfsId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(nfsId), fromTime, toTime).get(nfsId);

    }

    private Map<Long, List<SampleInfo<Nfsv3Measurements>>> getSamples(Connection connection, Long containerId, Set<Long> nfsIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getNfsv3PerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(nfsIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<Nfsv3Measurements>> getLatestSample(Long clusterId, Set<Long> nfsIds) throws NetappInterfaceException {
        logger.info("Query " + nfsIds.size() + "nfs samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getNfsv3PerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(nfsIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nfs samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<Nfsv3Measurements>>> getSummaries(Set<Long> vServerIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vServerIds.size() + "nfs summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<Nfsv3Measurements>>>  idToSummaries = new HashMap<>();

            Set<Long> remainingIds = new HashSet<>(vServerIds);
            Map<Long, Set<Long>> clusterIdToCifsVserverIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, vServerIds);

            for (Map.Entry<Long, Set<Long>> entry : clusterIdToCifsVserverIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);

                List<Summary<Nfsv3Measurements>> summaries = ComponentsFactoryRepository.getNfsv3PerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find nfsv3 with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nfs summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<Nfsv3Measurements>>> getSummaries(Long clusterId, Set<Long> vServerIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vServerIds.size() + "nfs summaries from " + fromTime + " to " + toTime
                + " for cluster" + clusterId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<Nfsv3Measurements>>>  idToSummaries = new HashMap<>();
            List<Summary<Nfsv3Measurements>> summaries = ComponentsFactoryRepository.getNfsv3PerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), vServerIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nfs summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
      public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
          logger.info("Query if " + objId + "nfs has counters from " + fromTime + " to " + toTime);

          Connection connection = null;
          try {
              connection = ConnectionHelper.newConnection(dataSource);
              return ComponentsFactoryRepository.getNfsv3PerformanceComponentsFactory()
                      .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
          } catch (Exception e) {
              throw new NetappInterfaceException("Failed to assert if nfs has counters ", e);
          } finally {
              ConnectionHelper.close(connection);
          }
      }
}
