/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.nic.interfaces.impl;


import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.nic.interfaces.data.NicMeasurements;
import com.netapp.oci.netapp.performance.nic.interfaces.local.NicPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.nic.interfaces.remote.NicPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class NicPerformanceQueryBean implements NicPerformanceQueryRemote, NicPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<NicMeasurements>>> getSamples(Set<Long> nicIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + nicIds.size() + "nic samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<NicMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<NicMeasurements>>>();

            Set<Long> remainingIds = new HashSet<Long>(nicIds);
            Map<Long, Set<Long>> clusterIdToNicIds = NetAppModel.INSTANCE.getClusterIdByNetworkPortIds(connection, nicIds);

            for (Map.Entry<Long, Set<Long>> entry : clusterIdToNicIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find nic with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nic samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query nic samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public Map<Long, List<SampleInfo<NicMeasurements>>> getSamples(Long containerId, Set<Long> nicIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + nicIds.size() + "nic samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, nicIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nic samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query nic samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<NicMeasurements>> getSamples(Long containerId, Long nicId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(nicId), fromTime, toTime).get(nicId);

    }

    private Map<Long, List<SampleInfo<NicMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> nicIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getNicPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(nicIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<NicMeasurements>> getLatestSample(Long id, Set<Long> nicIds) throws NetappInterfaceException {
        logger.info("Query " + nicIds.size() + "nic samples for :" + id);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getNicPerformanceComponentsFactory().createPersistenceFacade(connection, id)
                            .getLatestSample(nicIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nic samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NicMeasurements>>> getSummaries(Set<Long> nicIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + nicIds.size() + "nic summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<NicMeasurements>>>  idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(nicIds);
            Map<Long, Set<Long>> clusterIdToNicIds = NetAppModel.INSTANCE.getClusterIdByNetworkPortIds(connection, nicIds);

            for (Map.Entry<Long, Set<Long>> entry : clusterIdToNicIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<NicMeasurements>> summaries = ComponentsFactoryRepository.getNicPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find nic with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nic summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NicMeasurements>>> getSummaries(Long clusterId, Set<Long> nicIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + nicIds.size() + "nic summaries from " + fromTime + " to " + toTime
                + " for cluster id " + clusterId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<NicMeasurements>>>  idToSummaries = new HashMap<>();
                List<Summary<NicMeasurements>> summaries = ComponentsFactoryRepository.getNicPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), nicIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nic summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
      public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
          logger.info("Query if " + objId + "nic has counters from " + fromTime + " to " + toTime);

          Connection connection = null;
          try {
              connection = ConnectionHelper.newConnection(dataSource);
              return ComponentsFactoryRepository.getNicPerformanceComponentsFactory()
                      .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
          } catch (Exception e) {
              throw new NetappInterfaceException("Failed to assert if nic has counters ", e);
          } finally {
              ConnectionHelper.close(connection);
          }
      }
}
