/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.nvmffclif.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.nvmffclif.interfaces.data.NvmfFcLifMeasurements;
import com.netapp.oci.netapp.performance.nvmffclif.interfaces.local.NvmfFcLifPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.nvmffclif.interfaces.remote.NvmfFcLifPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class NvmfFcLifPerformanceQueryBean implements NvmfFcLifPerformanceQueryRemote, NvmfFcLifPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<NvmfFcLifMeasurements>>> getSamples(Set<Long> nvmfFcLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + nvmfFcLifIds.size() + "fcp lif samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<NvmfFcLifMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<NvmfFcLifMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(nvmfFcLifIds);

            Map<Long, Set<Long>> vserverIdToNvmfFcLifIds = NetAppModel.INSTANCE.getVserverIdByNvmfFcLifIds(connection, nvmfFcLifIds);
            for (Map.Entry<Long, Set<Long>> entry : vserverIdToNvmfFcLifIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, fromTime, toTime, ids));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find fcp lif with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcp lif samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query fcp lif samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<NvmfFcLifMeasurements>>> getSamples(Long containerId, Set<Long> nvmfFcLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + nvmfFcLifIds.size() + "fcp lif samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, fromTime, toTime, nvmfFcLifIds);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcp lif samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query fcp lif samples: " + (System.currentTimeMillis() - startTime));

        }
    }


    @Override
    public List<SampleInfo<NvmfFcLifMeasurements>> getSamples(Long containerId, Long nvmfFcLifId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(nvmfFcLifId), fromTime, toTime).get(nvmfFcLifId);

    }

    private Map<Long, List<SampleInfo<NvmfFcLifMeasurements>>> getSamples(Connection connection, Long containerId, Date fromTime, Date toTime, Set<Long> ids) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getNvmfFcLifPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(ids, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<NvmfFcLifMeasurements>> getLatestSample(Long vServerId, Set<Long> nvmfFcLifIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + nvmfFcLifIds.size() + "nvmfFcLif samples for vServer:" + vServerId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getNvmfFcLifPerformanceComponentsFactory().createPersistenceFacade(connection, vServerId)
                            .getLatestSample(nvmfFcLifIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get nvmfFcLif samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long,  List<Summary<NvmfFcLifMeasurements>>> getSummaries(Set<Long> nvmfFcLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + nvmfFcLifIds.size() + "fcp lif summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long,  List<Summary<NvmfFcLifMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(nvmfFcLifIds);

            Map<Long, Set<Long>> vserverIdToNvmfFcLifIds = NetAppModel.INSTANCE.getVserverIdByNvmfFcLifIds(connection, nvmfFcLifIds);
            for (Map.Entry<Long, Set<Long>> entry : vserverIdToNvmfFcLifIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<NvmfFcLifMeasurements>> summaries = ComponentsFactoryRepository.getNvmfFcLifPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find fcp lif with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcp lif summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<NvmfFcLifMeasurements>>> getSummaries(Long vServerId, Set<Long> nvmfFcLifIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + nvmfFcLifIds.size() + "fcp lif summaries from " + fromTime + " to " + toTime);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<NvmfFcLifMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<NvmfFcLifMeasurements>> summaries = ComponentsFactoryRepository.getNvmfFcLifPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, vServerId)
                    .getSummaries(new TimeSpan(fromTime, toTime), nvmfFcLifIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get fcp lif summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + " fcp lif has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getNvmfFcLifPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if fcp lif has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
