/*
 * Copyright (c) 2014 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.opmheadroom.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppPerformance;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.originator.handler.ClusterHandler;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.opmheadroom.interfaces.data.OpmHeadRoomCPUMeasurements;
import com.netapp.oci.netapp.performance.opmheadroom.interfaces.remote.OpmHeadRoomCPUPerformanceRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceDatasourceRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry.SampleType;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.SimpleUuidKey;
import com.onaro.sanscreen.util.Constant;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;

import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

/**
 * EJB to report on Opm Headroom CPU performance counters.
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class OpmHeadRoomCPUPerformanceBean implements OpmHeadRoomCPUPerformanceRemote {

    public static final String PACKAGE_NAME = "opm_headroom_cpu";
    private static final String OPM_HEADROOM_CPU_PACKAGE = PerformanceConstants.PACKAGE_PREFIX + PACKAGE_NAME;
    private static final Logger logger = LoggerFactory.getLogger(OPM_HEADROOM_CPU_PACKAGE);


    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * use this bean to register reporting datasource
     */
    @EJB
    private NetappPerformanceDatasourceRegistry datasourceRegistry;

    /**
     * use this bean to get the registered listeners
     */
    @EJB
    private NetappPerformanceSampleListenerRegistry sampleListenerRegistry;
    
    @Override
    public void reportOpmHeadRoomCPUSamples(Long datasourceId, String clusterKey, long time, long leaseTime, boolean isHistorical) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();
        Long containerId = null;
        Collection<Sample<OpmHeadRoomCPUMeasurements>> samples = new LinkedList<Sample<OpmHeadRoomCPUMeasurements>>();

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(OPM_HEADROOM_CPU_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(OPM_HEADROOM_CPU_PACKAGE, clusterKey, datasourceId, leaseTime)) {
                logger.info("Cluster key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(OPM_HEADROOM_CPU_PACKAGE, clusterKey, leaseTime) + " (clusterKey:" + clusterKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            containerId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (containerId == null) {
                logger.warn("Failed to find cluster id from clusterKey:" + clusterKey + " datasourceId:" + datasourceId);
            } else {
            	if(firstReport){
            		samples.add(new Sample<OpmHeadRoomCPUMeasurements>(containerId, time-1));
            	} else {
                		samples.add(new Sample<OpmHeadRoomCPUMeasurements>(containerId, time));
            	}
            	
            	if(samples.isEmpty()){
                    logger.warn("Failed to convert resoure headroom CPU raw-samples retrieved from the acquisiton into server-samples. clusterKey: {} datasourceId: {}", clusterKey, datasourceId);
                    return;
                }
                ComponentsFactoryRepository.getOpmHeadRoomCPUComponentsFactory().createPersistenceFacade(connection, clusterKey, containerId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME, datasourceId, startTime, System.currentTimeMillis() - startTime,
                        clusterKey, 1, samples.size(), leaseTime);
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert resoure headroom CPU performance data, clusterKey:"
                    + clusterKey + " datasourceId:" + datasourceId + " , time " + time + ", lease time " + leaseTime, e);
        } finally {
            ConnectionHelper.close(connection);
        }
        
		// invoke all the registered listeners for this sample        
        try {
        	if (samples != null) {
        		Map<String, PerformanceSampleListener> listeners = sampleListenerRegistry.getSampleListeners(SampleType.OPM_HEADROOM_CPU);                
        		if (listeners != null) {
        			logger.info("There are " + listeners.size() + " listeners for resoure headroom CPU stats.");
        			Map<Long, Collection<Sample<OpmHeadRoomCPUMeasurements>>> rslt = new HashMap<>();
        			rslt.put(containerId, samples);

        			long begin = System.currentTimeMillis();                	
        			for (String listenerName : listeners.keySet()) {
        				long iBegin = System.currentTimeMillis();
        				listeners.get(listenerName).processOpmHeadRoomCPU(rslt, isHistorical);
        				long iDelta = System.currentTimeMillis() - iBegin;
        				logger.debug(listenerName + " took " + iDelta + " milliseconds to process resoure headroom CPU samples.");
        			}                	
        			long delta = System.currentTimeMillis() - begin;

        			if (delta > PerformanceConstants.SAMPLE_LISTENER_MAX_ELAPSED_TIME) {
        				logger.warn("It took " + delta + " milliseconds for the resoure headroom CPU sample listeners to complete their process.");
        			}
        		}                                                                                                               
        	}
        } catch(Exception e) {
        	logger.error("Unable to invoke registered listeners for resoure headroom CPU; ", e);
        }        
    }
}
