/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.processor.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.processor.interfaces.data.ProcessorMeasurements;
import com.netapp.oci.netapp.performance.processor.interfaces.local.ProcessorPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.processor.interfaces.remote.ProcessorPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Nancy Walia
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class ProcessorPerformanceQueryBean implements ProcessorPerformanceQueryRemote, ProcessorPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<ProcessorMeasurements>>> getSamples(Set<Long> processorIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + processorIds.size() + "processor samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<ProcessorMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<ProcessorMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(processorIds);

            Map<Long, Set<Long>> clusterIdByProcessorIds = NetAppModel.INSTANCE.getClusterIdByProcessorIds(connection, processorIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByProcessorIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find processor with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get processor samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query processor samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public Map<Long, List<SampleInfo<ProcessorMeasurements>>> getSamples(Long containerId, Set<Long> processorIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + processorIds.size() + "processor samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, processorIds, fromTime, toTime);

        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get processor samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query processor samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<ProcessorMeasurements>> getSamples(Long containerId, Long processorId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(processorId), fromTime, toTime).get(processorId);

    }

    private Map<Long, List<SampleInfo<ProcessorMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> processorIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getProcessorPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(processorIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<ProcessorMeasurements>> getLatestSample(Long clusterId, Set<Long> processorIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + processorIds.size() + "processor samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getProcessorPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(processorIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get processor samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<ProcessorMeasurements>>> getSummaries(Set<Long> processorIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + processorIds.size() + "processor summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<ProcessorMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(processorIds);

            Map<Long, Set<Long>> clusterIdByNodeIds = NetAppModel.INSTANCE.getClusterIdByProcessorIds(connection, processorIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByNodeIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<ProcessorMeasurements>> summaries = ComponentsFactoryRepository.getProcessorPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find processor with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get processor summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<ProcessorMeasurements>>> getSummaries(Long clusterId, Set<Long> processorIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + processorIds.size() + "processor summaries from " + fromTime + " to " + toTime
                + " for cluster " + clusterId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<ProcessorMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<ProcessorMeasurements>> summaries = ComponentsFactoryRepository.getProcessorPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), processorIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get processor summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
      public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
          logger.info("Query if " + objId + "processor has counters from " + fromTime + " to " + toTime);

          Connection connection = null;
          try {
              connection = ConnectionHelper.newConnection(dataSource);
              return ComponentsFactoryRepository.getProcessorPerformanceComponentsFactory()
                      .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
          } catch (Exception e) {
              throw new NetappInterfaceException("Failed to assert if processor has counters ", e);
          } finally {
              ConnectionHelper.close(connection);
          }
      }
}

