/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.qosservicecenter.interfaces.impl;


import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.qosservicecenter.interfaces.local.QosServiceCenterPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.qosservicecenter.interfaces.remote.QosServiceCenterPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.qosservicecenter.interfaces.data.QosServiceCenterMeasurements;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class QosServiceCenterPerformanceQueryBean implements QosServiceCenterPerformanceQueryRemote, QosServiceCenterPerformanceQueryLocal {

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<QosServiceCenterMeasurements>>> getSamples(Set<Long> qosServiceCenterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosServiceCenterIds.size() + " qos service center samples from " + fromTime + " to " + toTime);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<QosServiceCenterMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<QosServiceCenterMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(qosServiceCenterIds);

            Map<Long, Set<Long>> clusterIdByServiceCenterIds = NetAppModel.INSTANCE.getClusterIdByQosServiceCenterIds(connection, qosServiceCenterIds);

            for (Map.Entry<Long, Set<Long>> entry : clusterIdByServiceCenterIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find QoS service center with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS service center samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos service center samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public Map<Long, List<SampleInfo<QosServiceCenterMeasurements>>> getSamples(Long containerId, Set<Long> qosServiceCenterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosServiceCenterIds.size() + " qos service center samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, qosServiceCenterIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS service center samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos service center samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<QosServiceCenterMeasurements>> getSamples(Long containerId, Long qosServiceCenterId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(qosServiceCenterId), fromTime, toTime).get(qosServiceCenterId);

    }

    private Map<Long, List<SampleInfo<QosServiceCenterMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> qosServiceCenterIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getQosServiceCenterPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(qosServiceCenterIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<QosServiceCenterMeasurements>> getLatestSample(Long clusterId, Set<Long> qosServiceCenterIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosServiceCenterIds.size() + "qosServiceCenter samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getQosServiceCenterPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(qosServiceCenterIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get qosServiceCenter samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<QosServiceCenterMeasurements>>> getSummaries(Set<Long> qosServiceCenterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + qosServiceCenterIds.size() + "qos service center summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<QosServiceCenterMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(qosServiceCenterIds);

            Map<Long, Set<Long>> clusterIdByServiceCenterIds = NetAppModel.INSTANCE.getClusterIdByQosServiceCenterIds(connection, qosServiceCenterIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByServiceCenterIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<QosServiceCenterMeasurements>> summaries = ComponentsFactoryRepository.getQosServiceCenterPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find QoS service center with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS service center summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<QosServiceCenterMeasurements>>> getSummaries(Long clusterId, Set<Long> qosServiceCenterIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + qosServiceCenterIds.size() + "qos service center summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<QosServiceCenterMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<QosServiceCenterMeasurements>> summaries = ComponentsFactoryRepository.getQosServiceCenterPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), qosServiceCenterIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS service center summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "qos service center has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getQosServiceCenterPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if QoS service center has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
