/*
 * $Id$
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.oci.netapp.performance.qosworkload.interfaces.impl;

import java.sql.Connection;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.db.NetAppPerformance;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.originator.handler.ClusterHandler;
import com.netapp.oci.netapp.originator.handler.QosWorkloadHandler;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.data.QosFileWorkloadMeasurements;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.data.QosLunFileWorkloadMeasurements;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.remote.QosFileWorkloadPerformanceRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.shared.SimpleRawSampleConverter;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceDatasourceRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.PersistenceFacade;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.SimpleUuidKey;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class QosFileWorkloadPerformanceBean implements QosFileWorkloadPerformanceRemote {
    public static final String PACKAGE_NAME = "qos_file_workload";
    private static final String QOSWORKLOAD_PACKAGE = PerformanceConstants.PACKAGE_PREFIX + PACKAGE_NAME;
    private static final Logger logger = LoggerFactory.getLogger(QOSWORKLOAD_PACKAGE);

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * use this bean to register reporting datasource
     */
    @EJB
    private NetappPerformanceDatasourceRegistry datasourceRegistry;

    /**
     * use this bean to get the registered listeners
     */
    @EJB
    private NetappPerformanceSampleListenerRegistry sampleListenerRegistry;

    @Override
    public void reportQosFileWorkloadRemoteSamples(Long datasourceId, String clusterKey, long time, Collection<RawSample<QosFileWorkloadMeasurements>> rawSamples, long leaseTime, boolean isHistorical) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();
        Long containerId = null;
        Collection<Sample<QosFileWorkloadMeasurements>> samples = null;

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(QOSWORKLOAD_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(QOSWORKLOAD_PACKAGE, clusterKey, datasourceId, leaseTime)) {
                logger.info("Cluster key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(QOSWORKLOAD_PACKAGE, clusterKey, leaseTime) + " (clusterKey:" + clusterKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            containerId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (containerId == null) {
                logger.warn("Failed to find containerId from clusterKey:" + clusterKey + " datasourceId:" + datasourceId);
            } else {
                samples = new SimpleRawSampleConverter<QosFileWorkloadMeasurements>(containerId, keyCache, QosWorkloadHandler.TAG, originId).toSamples(rawSamples, time, firstReport);
                if(samples.isEmpty()){
                    logger.warn("Failed to convert QosWorkload(File) raw-samples retrieved from the acquisiton into server-samples. clusterKey: {} datasourceId: {}", clusterKey, datasourceId);
                    return;
                }
                final PersistenceFacade<QosFileWorkloadMeasurements,QosFileWorkloadMeasurements> persistenceFacade = ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterKey, containerId);
                persistenceFacade.insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME, datasourceId, startTime, System.currentTimeMillis() - startTime,
                        clusterKey, rawSamples.size(), samples.size(), leaseTime);

            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert qos file workload performance, clusterKey:" + clusterKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }

        // invoke all the registered listeners for this sample
        try {
            if (samples != null) {
                Map<String, PerformanceSampleListener> listeners = sampleListenerRegistry.getSampleListeners(NetappPerformanceSampleListenerRegistry.SampleType.QOS_FILE_WORKLOAD);
                if (listeners != null) {
                    logger.info("There are " + listeners.size() + " listeners for QosWorkload stats.");
                    Map<Long, Collection<Sample<QosFileWorkloadMeasurements>>> rslt = new HashMap<>();
                    rslt.put(containerId, samples);

                    long begin = System.currentTimeMillis();
                    for (String listenerName : listeners.keySet()) {
                        long iBegin = System.currentTimeMillis();
                        listeners.get(listenerName).processQosFileWorkload(rslt, isHistorical);
                        long iDelta = System.currentTimeMillis() - iBegin;
                        logger.debug(listenerName + " took " + iDelta + " milliseconds to process qosFileWorkload samples.");
                    }
                    long delta = System.currentTimeMillis() - begin;

                    if (delta > PerformanceConstants.SAMPLE_LISTENER_MAX_ELAPSED_TIME) {
                        logger.warn("It took " + delta + " milliseconds for the QosWorkload sample listeners to complete their process.");
                    }
                }
            }
        } catch(Exception e) {
            logger.error("Unable to invoke registered listeners for QosFileWorkload; ", e);
        }
    }
}
