/*
 * $Id$
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.oci.netapp.performance.qosworkload.interfaces.impl;

import java.sql.Connection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import com.onaro.commons.util.CollectionPrinter;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.data.QosFileWorkloadMeasurements;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.data.QosLunFileWorkloadMeasurements;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.local.QosFileWorkloadPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.remote.QosFileWorkloadPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class QosFileWorkloadPerformanceQueryBean implements QosFileWorkloadPerformanceQueryRemote,QosFileWorkloadPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<QosFileWorkloadMeasurements>>> getSamples(Set<Long> QosWorkloadIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + QosWorkloadIds.size() + "qos file workload samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<QosFileWorkloadMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<QosFileWorkloadMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(QosWorkloadIds);

            Map<Long, Set<Long>> clusterIdByWorkloadIds = NetAppModel.INSTANCE.getClusterIdByQosVolumeWorkloadIds(connection, QosWorkloadIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByWorkloadIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find QoS file workload with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS file workload samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos file workload detail samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<QosFileWorkloadMeasurements>>> getSamples(Long containerId, Set<Long> QosWorkloadIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + QosWorkloadIds.size() + " qos file workload samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, QosWorkloadIds, fromTime, toTime);

        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get file QoS workload samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos file workload detail samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public List<SampleInfo<QosFileWorkloadMeasurements>> getSamples(Long containerId, Long QosWorkloadId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(QosWorkloadId), fromTime, toTime).get(QosWorkloadId);

    }

    private Map<Long, List<SampleInfo<QosFileWorkloadMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> QosWorkloadIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(QosWorkloadIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<QosFileWorkloadMeasurements>> getLatestSample(Long clusterId, Set<Long> QosWorkloadIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + QosWorkloadIds.size() + "Qos File Workload samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(QosWorkloadIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Qos File Workload samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<QosFileWorkloadMeasurements>>> getSummaries(Set<Long> QosWorkloadIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + QosWorkloadIds.size() + "qos file workload summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<QosFileWorkloadMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(QosWorkloadIds);

            Map<Long, Set<Long>> clusterIdByWorkloadIds = NetAppModel.INSTANCE.getClusterIdByQosVolumeWorkloadIds(connection, QosWorkloadIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByWorkloadIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<QosFileWorkloadMeasurements>> summaries = ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find QoS file workload with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS file workload summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<QosFileWorkloadMeasurements>>> getSummaries(Long custerId, Set<Long> QosWorkloadIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + QosWorkloadIds.size() + "qos file workload summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<QosFileWorkloadMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<QosFileWorkloadMeasurements>> summaries = ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, custerId)
                    .getSummaries(new TimeSpan(fromTime, toTime), QosWorkloadIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS file workload summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "qos file workload has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if QoS file workload has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public SampleInfo<QosFileWorkloadMeasurements> getFirstSampleAfterTime(Long clusterId, Long QosWorkloadId, long time) throws NetappInterfaceException {
        logger.info("Get " + QosWorkloadId + " with first Qos File Workload sample for cluster: " + clusterId  + " after time: " + time);
        Connection connection = null;
        SampleInfo<QosFileWorkloadMeasurements> result = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Sample<QosFileWorkloadMeasurements> sample = ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                    .getFirstSampleAfterTime(connection, QosWorkloadId, time);
            if(sample != null) {
                result = new SampleInfo<QosFileWorkloadMeasurements>(new Date(sample.getTime()), sample.getSampleData());
            }
            return result;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get first Qos File Workload sample", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public SampleInfo<QosFileWorkloadMeasurements> getLastSampleBeforeTime(Long clusterId, Long QosWorkloadId, long time) throws NetappInterfaceException {
        logger.info("Get " + QosWorkloadId + " with last Qos File Workload sample for cluster:" + clusterId + " before time: " + time);
        Connection connection = null;
        SampleInfo<QosFileWorkloadMeasurements> result = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Sample<QosFileWorkloadMeasurements> sample = ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                    .getLastSampleBeforeTime(connection, QosWorkloadId, time);
            if(sample != null ){
                result = new SampleInfo<QosFileWorkloadMeasurements>(new Date(sample.getTime()), sample.getSampleData());
            }
            return result;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get last Qos File Workload sample ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     *
     * @param clusterId
     * @param QosWorkloadId
     * @param time
     * @return summary of QosLunFileWorkloadMeasurements
     * @throws NetappInterfaceException
     */
    public Summary<QosFileWorkloadMeasurements> getFirstSummaryAfterTime(Long clusterId, Long QosWorkloadId, long time) throws NetappInterfaceException {
        logger.info("Get " + QosWorkloadId + " with Qos File Workload summary for cluster:" + clusterId + " after time: " + time);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Summary<QosFileWorkloadMeasurements> summary = ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                    .getFirstSummaryAfterTime(connection, QosWorkloadId, time);
            return summary;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Qos File Workload summary after time ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     *
     * @param clusterId
     * @param QosWorkloadId
     * @param time
     * @return summary of QosLunFileWorkloadMeasurements
     * @throws NetappInterfaceException
     */
    public Summary<QosFileWorkloadMeasurements>  getLastSummaryBeforeTime(Long clusterId, Long QosWorkloadId, long time) throws NetappInterfaceException {
        logger.info("Get " + QosWorkloadId + " with Qos File Workload summary for cluster:" + clusterId + "before time: "+ time);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Summary<QosFileWorkloadMeasurements> summary = ComponentsFactoryRepository.getQosFileWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                    .getLastSummaryBeforeTime(connection, QosWorkloadId, time);
            return summary;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Qos File Workload summary before time ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
