/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.qosworkload.interfaces.impl;

import java.sql.Connection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;

import com.onaro.commons.util.CollectionPrinter;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.data.QosVolumeWorkloadMeasurements;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.local.QosVolumeWorkloadPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.remote.QosVolumeWorkloadPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class QosVolumeWorkloadPerformanceQueryBean implements QosVolumeWorkloadPerformanceQueryRemote, QosVolumeWorkloadPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<QosVolumeWorkloadMeasurements>>> getSamples(Set<Long> qosVolumeWorkloadIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosVolumeWorkloadIds.size() + "qos volume workload samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<QosVolumeWorkloadMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<QosVolumeWorkloadMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(qosVolumeWorkloadIds);

            Map<Long, Set<Long>> clusterIdByWorkloadIds = NetAppModel.INSTANCE.getClusterIdByQosVolumeWorkloadIds(connection, qosVolumeWorkloadIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByWorkloadIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find QoS volume workload with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS volume workload samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos volume workload detail samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<QosVolumeWorkloadMeasurements>>> getSamples(Long containerId, Set<Long> qosVolumeWorkloadIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosVolumeWorkloadIds.size() + " qos volume workload samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, qosVolumeWorkloadIds, fromTime, toTime);

        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS volume workload samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos volume workload detail samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public List<SampleInfo<QosVolumeWorkloadMeasurements>> getSamples(Long containerId, Long qosVolumeWorkloadId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(qosVolumeWorkloadId), fromTime, toTime).get(qosVolumeWorkloadId);

    }

    private Map<Long, List<SampleInfo<QosVolumeWorkloadMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> qosVolumeWorkloadIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getQosVolumeWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(qosVolumeWorkloadIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<QosVolumeWorkloadMeasurements>> getLatestSample(Long clusterId, Set<Long> qosVolumeWorkloadIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosVolumeWorkloadIds.size() + "qosVolumeWorkload samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getQosVolumeWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(qosVolumeWorkloadIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get qosVolumeWorkload samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<QosVolumeWorkloadMeasurements>>> getSummaries(Set<Long> qosVolumeWorkloadIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + qosVolumeWorkloadIds.size() + "qos volume workload summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<QosVolumeWorkloadMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(qosVolumeWorkloadIds);

            Map<Long, Set<Long>> clusterIdByWorkloadIds = NetAppModel.INSTANCE.getClusterIdByQosVolumeWorkloadIds(connection, qosVolumeWorkloadIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByWorkloadIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<QosVolumeWorkloadMeasurements>> summaries = ComponentsFactoryRepository.getQosVolumeWorkloadPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find QoS volume workload with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS volume workload summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<QosVolumeWorkloadMeasurements>>> getSummaries(Long custerId, Set<Long> qosVolumeWorkloadIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + qosVolumeWorkloadIds.size() + "qos volume workload summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<QosVolumeWorkloadMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<QosVolumeWorkloadMeasurements>> summaries = ComponentsFactoryRepository.getQosVolumeWorkloadPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, custerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), qosVolumeWorkloadIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS volume workload summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "qos volume workload has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getQosVolumeWorkloadPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if QoS volume workload has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public SampleInfo<QosVolumeWorkloadMeasurements> getFirstSampleAfterTime(Long clusterId, Long qosVolumeWorkloadId, long time) throws NetappInterfaceException {
        logger.info("Get " + qosVolumeWorkloadId + " with first qosVolumeWorkload sample for cluster: " + clusterId  + " after time: " + time);
        Connection connection = null;
        SampleInfo<QosVolumeWorkloadMeasurements> result = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Sample<QosVolumeWorkloadMeasurements> sample = ComponentsFactoryRepository.getQosVolumeWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                    .getFirstSampleAfterTime(connection, qosVolumeWorkloadId, time);
            if(sample != null) {
                result = new SampleInfo<QosVolumeWorkloadMeasurements>(new Date(sample.getTime()), sample.getSampleData());
            }
            return result;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get first qosVolumeWorkload sample", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public SampleInfo<QosVolumeWorkloadMeasurements> getLastSampleBeforeTime(Long clusterId, Long qosVolumeWorkloadId, long time) throws NetappInterfaceException {
        logger.info("Get " + qosVolumeWorkloadId + " with last qosVolumeWorkload sample for cluster:" + clusterId + " before time: " + time);
        Connection connection = null;
        SampleInfo<QosVolumeWorkloadMeasurements> result = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Sample<QosVolumeWorkloadMeasurements> sample = ComponentsFactoryRepository.getQosVolumeWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                    .getLastSampleBeforeTime(connection, qosVolumeWorkloadId, time);
            if(sample != null ){
                result = new SampleInfo<QosVolumeWorkloadMeasurements>(new Date(sample.getTime()), sample.getSampleData());
            }
            return result;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get last qosVolumeWorkload sample ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     *
     * @param clusterId
     * @param qosVolumeWorkloadId
     * @param time
     * @return summary of QosVolumeWorkloadMeasurements
     * @throws NetappInterfaceException
     */
    public Summary<QosVolumeWorkloadMeasurements> getFirstSummaryAfterTime(Long clusterId, Long qosVolumeWorkloadId, long time) throws NetappInterfaceException {
        logger.info("Get " + qosVolumeWorkloadId + " with qosVolumeWorkload summary for cluster:" + clusterId + " after time: " + time);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Summary<QosVolumeWorkloadMeasurements> summary = ComponentsFactoryRepository.getQosVolumeWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                    .getFirstSummaryAfterTime(connection, qosVolumeWorkloadId, time);
            return summary;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get qosVolumeWorkload summary after time ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     *
     * @param clusterId
     * @param qosVolumeWorkloadId
     * @param time
     * @return summary of QosVolumeWorkloadMeasurements
     * @throws NetappInterfaceException
     */
    public Summary<QosVolumeWorkloadMeasurements>  getLastSummaryBeforeTime(Long clusterId, Long qosVolumeWorkloadId, long time) throws NetappInterfaceException {
        logger.info("Get " + qosVolumeWorkloadId + " with qosVolumeWorkload summary for cluster:" + clusterId + "before time: "+ time);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Summary<QosVolumeWorkloadMeasurements> summary = ComponentsFactoryRepository.getQosVolumeWorkloadPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                    .getLastSummaryBeforeTime(connection, qosVolumeWorkloadId, time);
            return summary;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get qosVolumeWorkload summary before time ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

}
