/*
 * Copyright (c) 2014 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.qosworkloaddblade.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppPerformance;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.originator.handler.ClusterHandler;
import com.netapp.oci.netapp.originator.handler.QosWorkloadNodeRelationshipHandler;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.qosworkloaddblade.interfaces.data.QosWorkloadQueueDBladeMeasurements;
import com.netapp.oci.netapp.performance.qosworkloaddblade.interfaces.remote.QosWorkloadQueueDBladePerformanceRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.shared.SimpleRawSampleConverter;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceDatasourceRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry.SampleType;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.SimpleUuidKey;
import com.onaro.sanscreen.util.Constant;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * EJB to report on Workload Queue DBlade performance counters.
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class QosWorkloadQueueDBladePerformanceBean implements QosWorkloadQueueDBladePerformanceRemote {

    public static final String PACKAGE_NAME = "qos_workload_queue_dblade";
    private static final String QOSWORKLOAD_PACKAGE = PerformanceConstants.PACKAGE_PREFIX + PACKAGE_NAME;
    private static final Logger logger = LoggerFactory.getLogger(QOSWORKLOAD_PACKAGE);

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * use this bean to register reporting datasource
     */
    @EJB
    private NetappPerformanceDatasourceRegistry datasourceRegistry;

    /**
     * use this bean to get the registered listeners
     */
    @EJB
    private NetappPerformanceSampleListenerRegistry sampleListenerRegistry;
    
    /**
     * Report Workload Queue DBlade performance counters.
     *
     * @param datasourceId  id of the data source reporting performance data
     * @param clusterKey  the Cluster Key Identifier
     * @param time the sample time
     * @param rawSamples Workload Queue DBlade performance data
     * @param leaseTime  the time for datasource to be registered in the in the registry (in milliseconds)
     * @throws NetappInterfaceException
     */
    @Override
    public void reportQosWorkloadQueueDBladeRemoteSamples(Long datasourceId, String clusterKey, long time, Collection<RawSample<QosWorkloadQueueDBladeMeasurements>> rawSamples, long leaseTime, boolean isHistorical) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();
        Long containerId = null;
        Collection<Sample<QosWorkloadQueueDBladeMeasurements>> samples = null;

        try {

            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(QOSWORKLOAD_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(QOSWORKLOAD_PACKAGE, clusterKey, datasourceId, leaseTime)) {
                if (logger.isInfoEnabled()) {
                    logger.info(
                            new StringBuilder("Cluster key is already registered by:")
                                    .append(datasourceRegistry.getRegisterDatasourceId(QOSWORKLOAD_PACKAGE, clusterKey, leaseTime))
                                    .append(" (clusterKey:").append(clusterKey).append(" datasourceId:")
                                    .append(datasourceId).append(")").toString());
                }
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);

            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            containerId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (containerId == null) {
                logger.warn("Failed to find containerId from clusterKey:" + clusterKey + " datasourceId:" + datasourceId);
            } else {
                samples = new SimpleRawSampleConverter<QosWorkloadQueueDBladeMeasurements>(containerId, keyCache, QosWorkloadNodeRelationshipHandler.TAG, originId).toSamples(rawSamples, time, firstReport);
                if(samples.isEmpty()){
                    logger.warn("Failed to convert QosWorkloadQueueDBlade raw-samples retrieved from the acquisiton into server-samples. clusterKey: {} datasourceId: {}", clusterKey, datasourceId);
                    return;
                }
                ComponentsFactoryRepository.getQosWorkloadQueueDBladePerformanceComponentsFactory().createPersistenceFacade(connection, clusterKey, containerId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME, datasourceId, startTime, System.currentTimeMillis() - startTime,
                        clusterKey, rawSamples.size(), samples.size(), leaseTime);
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert qos workload queue dBlade performance data, clusterKey:"
                    + clusterKey + " datasourceId:" + datasourceId + " time:" + time + " leaseTime:" + leaseTime, e);
        } finally {
            ConnectionHelper.close(connection);
        }
        
		// invoke all the registered listeners for this sample        
        try {
        	if (samples != null) {
        		Map<String, PerformanceSampleListener> listeners = sampleListenerRegistry.getSampleListeners(SampleType.QOS_WORKLOAD_QUEUE_DBLADE);                
        		if (listeners != null) {
        			logger.info("There are " + listeners.size() + " listeners for QosWorkloadDetail stats.");
        			Map<Long, Collection<Sample<QosWorkloadQueueDBladeMeasurements>>> rslt = new HashMap<>();
        			rslt.put(containerId, samples);

        			long begin = System.currentTimeMillis();                	
        			for (String listenerName : listeners.keySet()) {
        				long iBegin = System.currentTimeMillis();
        				listeners.get(listenerName).processQosWorkloadQueueDBlade(rslt, isHistorical);
        				long iDelta = System.currentTimeMillis() - iBegin;
        				logger.debug(listenerName + " took " + iDelta + " milliseconds to process QosWorkloadQueueDBlade samples.");
        			}                	
        			long delta = System.currentTimeMillis() - begin;

        			if (delta > PerformanceConstants.SAMPLE_LISTENER_MAX_ELAPSED_TIME) {
        				logger.warn("It took " + delta + " milliseconds for the QosWorkloadQueueDBlade sample listeners to complete their process.");
        			}
        		}                                                                                                               
        	}
        } catch(Exception e) {
        	logger.error("Unable to invoke registered listeners for QosWorkloadQueueDBlade; ", e);
        }
    }
        
}
