/*
 * Copyright (c) 2014 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.qosworkloaddblade.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.qosworkloaddblade.interfaces.data.QosWorkloadQueueDBladeMeasurements;
import com.netapp.oci.netapp.performance.qosworkloaddblade.interfaces.local.QosWorkloadQueueDBladePerformanceQueryLocal;
import com.netapp.oci.netapp.performance.qosworkloaddblade.interfaces.remote.QosWorkloadQueueDBladePerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * Query interface for querying workload queue dblade counters.
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class QosWorkloadQueueDBladePerformanceQueryBean implements QosWorkloadQueueDBladePerformanceQueryRemote, QosWorkloadQueueDBladePerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    /**
     * Gets QoS workload queue dBlade samples for period of time
     *
     * @param qosWorkloadNodeRelationshipIds set of QoS workload node relationship ids
     * @param fromTime                       beginning of the time period
     * @param toTime                         end of the time period
     * @return map of QoS workload node relationship id to list ordered by time of QoS workload queue dBlade samples data
     * @throws com.netapp.oci.netapp.interfaces.data.NetappInterfaceException
     *          if anything went wrong
     */
    @Override
    public Map<Long, List<SampleInfo<QosWorkloadQueueDBladeMeasurements>>> getSamples(Set<Long> qosWorkloadNodeRelationshipIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosWorkloadNodeRelationshipIds.size() + "qos workload queue dBlade samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<QosWorkloadQueueDBladeMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<QosWorkloadQueueDBladeMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(qosWorkloadNodeRelationshipIds);

            Map<Long, Set<Long>> clusterIdByWorkloadNodeRelationshipIds = NetAppModel.INSTANCE.getClusterIdByQosWorkloadNodeRelationshipIds(connection, qosWorkloadNodeRelationshipIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByWorkloadNodeRelationshipIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException(
                        new StringBuilder("Failed to find QoS workload queue dBlade with id ")
                                .append(CollectionPrinter.LONG.toString(remainingIds))
                                .append(", from Time : ").append(fromTime)
                                .append(", to Time ").append(toTime).toString());
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException(new StringBuilder("Failed to find QoS workload queue dBlade ")
                    .append(", from Time : ").append(fromTime)
                    .append(", to Time ").append(toTime).toString(), e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos workload queue dBlade detail samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    /**
     * Gets QoS workload queue dBlade samples for period of time
     *
     * @param clusterId                      cluster id
     * @param qosWorkloadNodeRelationshipIds set of QoS workload node relationship ids
     * @param fromTime                       beginning of the time period
     * @param toTime                         end of the time period
     * @return map of QoS workload node relationship id to list ordered by time of QoS workload queue dBlade samples data
     * @throws com.netapp.oci.netapp.interfaces.data.NetappInterfaceException
     *          if anything went wrong
     */
    @Override
    public Map<Long, List<SampleInfo<QosWorkloadQueueDBladeMeasurements>>> getSamples(Long clusterId, Set<Long> qosWorkloadNodeRelationshipIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosWorkloadNodeRelationshipIds.size() + "qos workload queue dBlade samples from " + fromTime + " to " + toTime + " for containerId:" + clusterId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, clusterId, qosWorkloadNodeRelationshipIds, fromTime, toTime);

        } catch (Exception e) {
            throw new NetappInterfaceException(new StringBuilder("Failed to find QoS workload queue dBlade with id ")
                    .append("clusterId=").append(clusterId)
                    .append(", from Time : ").append(fromTime)
                    .append(", to Time ").append(toTime).toString(), e);

        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos workload queue dBlade detail samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    /**
     * Gets QoS workload queue dBlade samples for period of time
     *
     * @param containerId                     cluster id
     * @param qosWorkloadNodeRelationshipId Qos Workload Node Relationship  id
     * @param fromTime                      beginning of the time period
     * @param toTime                        end of the time period
     * @return list ordered by time of QoS workload queue dBlade samples data
     * @throws com.netapp.oci.netapp.interfaces.data.NetappInterfaceException
     *          if anything went wrong
     */
    @Override
    public List<SampleInfo<QosWorkloadQueueDBladeMeasurements>> getSamples(Long containerId, Long qosWorkloadNodeRelationshipId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(qosWorkloadNodeRelationshipId), fromTime, toTime).get(qosWorkloadNodeRelationshipId);

    }

    /**
     * Retrieve the latest sample for the given qosVolume id's (qosVolumeIds). The map returned will contain only qosWorkloadQueueDBlade ids that has a sample.
     * qosWorkloadQueueDBlade ids with no samples will not be returned in the map. An empty map will be returned if no samples are found for all the qosWorkloadQueueDBlade ids passed.
     *
     * @param containerId                      cluster id
     * @param qosWorkloadNodeRelationshipIds set of QoS workload node relationship ids
     * @param fromTime                      beginning of the time period
     * @param toTime                        end of the time period
     * @return map of qosWorkloadQueueDBlade id to latest sample SampleInfo<QosWorkloadQueueDBladeMeasurements>. qosWorkloadQueueDBlade ids with no samples will not be included in this map.
     * @throws com.netapp.oci.netapp.interfaces.data.NetappInterfaceException
     *          if anything went wrong
     */
    private Map<Long, List<SampleInfo<QosWorkloadQueueDBladeMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> qosWorkloadNodeRelationshipIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getQosWorkloadQueueDBladePerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(qosWorkloadNodeRelationshipIds, new TimeSpan(fromTime, toTime)));
    }

    /**
     * Retrieve the latest sample for the given qosVolume id's (qosVolumeIds). The map returned will contain only qosWorkloadQueueDBlade ids that has a sample.
     * qosWorkloadQueueDBlade ids with no samples will not be returned in the map. An empty map will be returned if no samples are found for all the qosWorkloadQueueDBlade ids passed.
     *
     * @param clusterId                      cluster id
     * @param qosWorkloadNodeRelationshipIds set of QoS workload node relationship ids
     * @return map of qosWorkloadQueueDBlade id to latest sample SampleInfo<QosWorkloadQueueDBladeMeasurements>. qosWorkloadQueueDBlade ids with no samples will not be included in this map.
     * @throws com.netapp.oci.netapp.interfaces.data.NetappInterfaceException
     *          if anything went wrong
     */
    @Override
    public Map<Long, SampleInfo<QosWorkloadQueueDBladeMeasurements>> getLatestSample(Long clusterId, Set<Long> qosWorkloadNodeRelationshipIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosWorkloadNodeRelationshipIds.size() + "qosWorkloadQueueDBlade samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getQosWorkloadQueueDBladePerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(qosWorkloadNodeRelationshipIds));
        } catch (Exception e) {
            throw new NetappInterfaceException(new StringBuilder("Failed to find QoS workload queue dBlade ")
                    .append(", Cluster Id : ").append(clusterId).toString(), e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos workload queue nBlade detail samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    /**
     * Gets QoS workload queue dBlade summaries for QoS workload node relationship ids
     *
     *
     * @param qosWorkloadNodeRelationshipIds set of QoS workload node relationship ids
     * @param fromTime                       beginning of the time period
     * @param toTime                         end of the time period
     * @return map of QoS workload node relationship id to summary data
     * @throws com.netapp.oci.netapp.interfaces.data.NetappInterfaceException if anything went wrong
     */
    @Override
    public Map<Long, List<Summary<QosWorkloadQueueDBladeMeasurements>>> getSummaries(Set<Long> qosWorkloadNodeRelationshipIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + qosWorkloadNodeRelationshipIds.size() + "qos workload queue dBlade summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<QosWorkloadQueueDBladeMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(qosWorkloadNodeRelationshipIds);

            Map<Long, Set<Long>> clusterIdByWorkloadIds = NetAppModel.INSTANCE.getClusterIdByQosWorkloadNodeRelationshipIds(connection, qosWorkloadNodeRelationshipIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByWorkloadIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<QosWorkloadQueueDBladeMeasurements>> summaries = ComponentsFactoryRepository.getQosWorkloadQueueDBladePerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException(
                        new StringBuilder("Failed to find QoS workload queue dBlade with id ")
                                .append(CollectionPrinter.LONG.toString(remainingIds))
                                .append(", from Time : ").append(fromTime)
                                .append(", to Time ").append(toTime).toString());

            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS workload queue dBlade summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     *
     * @param clusterId cluster id
     * @param qosWorkloadNodeRelationshipIds set of QoS workload node relationship ids
     * @param fromTime                       beginning of the time period
     * @param toTime                         end of the time period
     * @return
     * @throws NetappInterfaceException
     */
    @Override
    public Map<Long, List<Summary<QosWorkloadQueueDBladeMeasurements>>> getSummaries(Long clusterId, Set<Long> qosWorkloadNodeRelationshipIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + qosWorkloadNodeRelationshipIds.size() + "qos workload queue dBlade summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<QosWorkloadQueueDBladeMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<QosWorkloadQueueDBladeMeasurements>> summaries = ComponentsFactoryRepository.getQosWorkloadQueueDBladePerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), qosWorkloadNodeRelationshipIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS workload queue dBlade summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     * Returns true if samples exist over the time period
     *
     * @param clusterId
     * @param qosWorkloadNodeRelationshipId
     * @param fromTime
     * @param toTime
     * @return true if samples exist
     * @throws com.netapp.oci.netapp.interfaces.data.NetappInterfaceException if anything went wrong
     */
    @Override
    public Boolean areStatisticsAvailable(Long clusterId, Long qosWorkloadNodeRelationshipId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + qosWorkloadNodeRelationshipId + "qos workload queue dBlade has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getQosWorkloadQueueDBladePerformanceComponentsFactory()
                    .createPersistenceFacade(connection, clusterId).doSamplesExist(new TimeSpan(fromTime, toTime), qosWorkloadNodeRelationshipId);
        } catch (Exception e) {
            throw new NetappInterfaceException(new StringBuilder("Failed to find QoS workload queue dBlade ")
                    .append("clusterId=").append(clusterId)
                    .append(", from Time : ").append(fromTime)
                    .append(", to Time ").append(toTime).toString(), e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
