/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppPerformance;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.originator.handler.ClusterHandler;
import com.netapp.oci.netapp.originator.handler.QosWorkloadDetailHandler;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.data.QosWorkloadDetailMeasurements;
import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.remote.QosWorkloadDetailPerformanceRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.shared.SimpleRawSampleConverter;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceDatasourceRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry.SampleType;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.SimpleUuidKey;
import com.onaro.sanscreen.util.Constant;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class QosWorkloadDetailPerformanceBean implements QosWorkloadDetailPerformanceRemote {

    public static final String PACKAGE_NAME = "qos_workload_detail";
    private static final String QOSWORKLOAD_DETAIL_PACKAGE = PerformanceConstants.PACKAGE_PREFIX + PACKAGE_NAME;
    private static final Logger logger = LoggerFactory.getLogger(QOSWORKLOAD_DETAIL_PACKAGE);

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * use this bean to register reporting datasource
     */
    @EJB
    private NetappPerformanceDatasourceRegistry datasourceRegistry;

    /**
     * use this bean to get the registered listeners
     */
    @EJB
    private NetappPerformanceSampleListenerRegistry sampleListenerRegistry;
    
    @Override
    public void reportQosWorkloadDetailSamples(Long datasourceId, String clusterKey, long time, Collection<RawSample<QosWorkloadDetailMeasurements>> rawSamples, long leaseTime, boolean isHistorical) throws NetappInterfaceException {
        Connection connection = null;
        long startTime = System.currentTimeMillis();
        Long containerId = null;
        Collection<Sample<QosWorkloadDetailMeasurements>> samples = null;

        try {
            // DO NOT change the order of the line
            // otherwise firstReport will be always true
            boolean firstReport = !datasourceRegistry.isRegistered(QOSWORKLOAD_DETAIL_PACKAGE, clusterKey, datasourceId, leaseTime);
            if (!datasourceRegistry.register(QOSWORKLOAD_DETAIL_PACKAGE, clusterKey, datasourceId, leaseTime)) {
                logger.info("Cluster key is already registered by:" + datasourceRegistry.getRegisterDatasourceId(QOSWORKLOAD_DETAIL_PACKAGE, clusterKey, leaseTime) + " (clusterKey:" + clusterKey + " datasourceId:" + datasourceId + ")");
                return;
            }

            connection = ConnectionHelper.newConnection(dataSource);
            
            KeyCache keyCache = KeyCache.get();
            long originId =  keyCache.findOriginatorByClusterKey(new SimpleUuidKey(clusterKey));
            containerId = keyCache.findId(originId, ClusterHandler.TAG, clusterKey);
            if (containerId == null) {
                logger.warn("Failed to find containerId from clusterKey:" + clusterKey + " datasourceId:" + datasourceId);
            } else {
                samples = new SimpleRawSampleConverter<QosWorkloadDetailMeasurements>(containerId, keyCache, QosWorkloadDetailHandler.TAG,originId).toSamples(rawSamples, time, firstReport);
                if(samples.isEmpty()){
                    logger.warn("Failed to convert QosWorkloadDetail raw-samples retrieved from the acquisiton into server-samples. clusterKey: {} datasourceId: {}", clusterKey, datasourceId);
                    return;
                }
                ComponentsFactoryRepository.getQosWorkloadDetailPerformanceComponentsFactory().createPersistenceFacade(connection, clusterKey, containerId).insert(samples);
                NetAppPerformance.INSTANCE.updateHistory(connection, time, PACKAGE_NAME, datasourceId, startTime, System.currentTimeMillis() - startTime,
                        clusterKey, rawSamples.size(), samples.size(), leaseTime);
            }
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to insert qos volume workload detail performance, clusterKey:" + clusterKey + " datasourceId:" + datasourceId, e);
        } finally {
            ConnectionHelper.close(connection);
        }
        
		// invoke all the registered listeners for this sample        
        try {
        	if (samples != null) {
        		Map<String, PerformanceSampleListener> listeners = sampleListenerRegistry.getSampleListeners(SampleType.QOS_WORKLOAD_DETAIL);                
        		if (listeners != null) {
        			logger.info("There are " + listeners.size() + " listeners for QosWorkloadDetail stats.");
        			Map<Long, Collection<Sample<QosWorkloadDetailMeasurements>>> rslt = new HashMap<>();
        			rslt.put(containerId, samples);

        			long begin = System.currentTimeMillis();                	
        			for (String listenerName : listeners.keySet()) {
        				long iBegin = System.currentTimeMillis();
        				listeners.get(listenerName).processQosWorkloadDetail(rslt, isHistorical);
        				long iDelta = System.currentTimeMillis() - iBegin;
        				logger.debug(listenerName + " took " + iDelta + " milliseconds to process QosWorkloadDetail samples.");
        			}                	
        			long delta = System.currentTimeMillis() - begin;

        			if (delta > PerformanceConstants.SAMPLE_LISTENER_MAX_ELAPSED_TIME) {
        				logger.warn("It took " + delta + " milliseconds for the QosWorkloadDetail sample listeners to complete their process.");
        			}
        		}                                                                                                               
        	}
        } catch(Exception e) {
        	logger.error("Unable to invoke registered listeners for QosWorkloadDetail; ", e);
        }
    }
        
}

