/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.data.QosWorkloadDetailMeasurements;
import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.local.QosWorkloadDetailPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.remote.QosWorkloadDetailPerformanceQueryRemote;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class QosWorkloadDetailPerformanceQueryBean implements QosWorkloadDetailPerformanceQueryRemote, QosWorkloadDetailPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<QosWorkloadDetailMeasurements>>> getSamples(Set<Long> qosWorkloadDetailIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosWorkloadDetailIds.size() + " qos workload detail samples from " + fromTime + " to " + toTime);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<QosWorkloadDetailMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<QosWorkloadDetailMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(qosWorkloadDetailIds);

            Map<Long, Set<Long>> clusterIdByWorkloadIds = NetAppModel.INSTANCE.getClusterIdByQosWorkloadDetailIds(connection, qosWorkloadDetailIds);

            for (Map.Entry<Long, Set<Long>> entry : clusterIdByWorkloadIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, ids, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find QoS workload detail with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS volume workload detail samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos workload detail samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public Map<Long, List<SampleInfo<QosWorkloadDetailMeasurements>>> getSamples(Long containerId, Set<Long> qosWorkloadDetailIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosWorkloadDetailIds.size() + " qos workload detail samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return getSamples(connection, containerId, qosWorkloadDetailIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS volume workload detail samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query qos workload detail samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<QosWorkloadDetailMeasurements>> getSamples(Long containerId, Long qosWorkloadDetailId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(qosWorkloadDetailId), fromTime, toTime).get(qosWorkloadDetailId);

    }

    private Map<Long, List<SampleInfo<QosWorkloadDetailMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> qosWorkloadDetailIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getQosWorkloadDetailPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(qosWorkloadDetailIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<QosWorkloadDetailMeasurements>> getLatestSample(Long clusterId, Set<Long> qosWorkloadDetailIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + qosWorkloadDetailIds.size() + "qosWorkloadDetail samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getQosWorkloadDetailPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(qosWorkloadDetailIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get qosWorkloadDetail samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<QosWorkloadDetailMeasurements>>> getSummaries(Set<Long> qosWorkloadDetailIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + qosWorkloadDetailIds.size() + "qos workload detail summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<QosWorkloadDetailMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(qosWorkloadDetailIds);

            Map<Long, Set<Long>> clusterIdByWorkloadIds = NetAppModel.INSTANCE.getClusterIdByQosWorkloadDetailIds(connection, qosWorkloadDetailIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByWorkloadIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<QosWorkloadDetailMeasurements>> summaries = ComponentsFactoryRepository.getQosWorkloadDetailPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find QoS workload detail with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS volume workload detail summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<QosWorkloadDetailMeasurements>>> getSummaries(Long clusterId, Set<Long> qosWorkloadDetailIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + qosWorkloadDetailIds.size() + "qos workload detail summaries from " + fromTime + " to " + toTime);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<QosWorkloadDetailMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<QosWorkloadDetailMeasurements>> summaries = ComponentsFactoryRepository.getQosWorkloadDetailPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), qosWorkloadDetailIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get QoS volume workload detail summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "qos workload has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getQosWorkloadDetailPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if QoS workload has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}

