package com.netapp.oci.netapp.performance.shared;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.shared.interfaces.local.NetappPerformanceSampleListenerRegistry.SampleType;
import com.netapp.oci.netapp.performance.shared.interfaces.local.PerformanceSampleListener;

/**
 * Performance sample listener registration. The registration is keyed by the sample type and
 * the listener is keyed by the listener name. For example, a graphite listener is registered
 * to watch Aggregate samples, the registration is saved as:
 * 		[SampleType, "graphite"->PerformanceSampleListener]
 * 
 * @author Cheng Sun
 *
 */
public class SampleListenerRegistry {
	private static final Logger logger = LoggerFactory.getLogger(SampleListenerRegistry.class.getName());
	
    private final ConcurrentHashMap<SampleType, Map<String, PerformanceSampleListener>> sampleToListeners = 
    		new ConcurrentHashMap<>(30, 0.75f, 1);

    /**
     * Register a listener identified by thes listener name to the sample type.
     * 
     * @param sampleType
     * @param listenerName
     * @param listener
     */
    public synchronized void register(SampleType sampleType, String listenerName, PerformanceSampleListener listener) throws NetappInterfaceException{
    	if (sampleType == null || listenerName == null || listener == null) {
    		throw new NetappInterfaceException("sampleType/listenerName/listener are required to register a sample listener!");
    	}
    	
    	Map<String, PerformanceSampleListener> listeners = sampleToListeners.get(sampleType);
    	if (listeners == null) {
    		Map<String, PerformanceSampleListener> newListener = new HashMap<>();
    		newListener.put(listenerName, listener);
    		sampleToListeners.put(sampleType, newListener);
    	} else {
    		PerformanceSampleListener existlistener = listeners.get(listenerName);
    		if (existlistener == null) {
    			listeners.put(listenerName, listener);
    		} else {
    			String err = "SampleListener[" + listenerName + "] has been registered for sample[" + sampleType + 
    					"]. The registration is ignored.";
    			logger.warn(err);
    			throw new NetappInterfaceException(err);    			
    		}
    	}
    }

    /**
     * Unregister the listener identified by the listenerName to the input sample type.
     * 
     * @param sampleType
     * @param listenerName
     * @throws NetappInterfaceException
     */
    public synchronized void unregister(SampleType sampleType, String listenerName) throws NetappInterfaceException{
    	String errMsg = null;
    	if (sampleType == null || listenerName == null) {
    		errMsg = "sampleType/listenerName are required to unregister a sample listener!";
    		logger.error(errMsg);
    		throw new NetappInterfaceException(errMsg);
    	}

    	Map<String, PerformanceSampleListener> listeners = sampleToListeners.remove(sampleType);
    	if (listeners == null) {
    		errMsg = "There is no listener registered for [" + sampleType + "], nothing to unregister.";
			logger.warn(errMsg);
			throw new NetappInterfaceException(errMsg);
    	} else {
    		if (listeners.containsKey(listenerName)) {
    			listeners.remove(listenerName);
    		} else {
    			errMsg = "SampleListener[" + listenerName + "] isn't registered for sample[" + sampleType +
    					"], nothing to unregister.";
    			logger.warn(errMsg);
    			throw new NetappInterfaceException(errMsg);
    		}
    		
    		// clear up the listeners
    		if (listeners.isEmpty()) {
    			sampleToListeners.remove(sampleType);
    		}
    	}
    }

    /**
     * Get all the listeners that subscribe the input sample type.
     *  
     * @param sampleType
     * @return
     */
    public Map<String, PerformanceSampleListener> getListeners(SampleType sampleType) {
    	return sampleToListeners.get(sampleType);
    }
}
