/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.shared.interfaces.impl;

import com.netapp.oci.netapp.originator.JmsSender;
import com.netapp.oci.netapp.originator.SummaryInfo;
import com.netapp.oci.netapp.performance.SummaryCalculator;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.onaro.commons.types.Period;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Resource;
import javax.annotation.security.RunAs;
import javax.ejb.*;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.TextMessage;
import javax.sql.DataSource;
import javax.transaction.UserTransaction;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@MessageDriven(activationConfig = {
        @ActivationConfigProperty(propertyName = "destination", propertyValue = "java:/queue/SummaryQueue"),
        @ActivationConfigProperty(propertyName = "acknowledgeMode", propertyValue = "Auto-acknowledge"),
        @ActivationConfigProperty(propertyName = "maxSession", propertyValue = "1"),
        @ActivationConfigProperty(propertyName = "destinationType", propertyValue = "javax.jms.Queue")})
@TransactionManagement(value= TransactionManagementType.BEAN)
@RunAs("admin")
public class NetappPerformanceSummaryJob implements MessageListener
{
  private static final Logger logger = LoggerFactory.getLogger( NetappPerformanceSummaryJob.class );
  private static final Map<Long,Long> performanceTimeMap;

  static
  {
    HashMap<Long,Long> map = new HashMap<>();
    performanceTimeMap = Collections.synchronizedMap( map );
  }

  @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
  private DataSource sqlDataSource;

    @Resource
    MessageDrivenContext ctx;


    public void onMessage(Message inMessage)
  {
    if( logger.isDebugEnabled() ) logger.debug( "Starting performance process for " );
    String status = "failure";
    long dataSourceId = -1;
    String dateStr = " no time ";

    try
    {
      //This should never happen
      if( !(inMessage instanceof TextMessage) )
        logger.error( "Message of wrong type: " + inMessage.getClass().getName() );
      else
      {
        //Parse the data in the text message
        TextMessage msg = (TextMessage) inMessage;

        if( logger.isTraceEnabled() ) logger.debug( "MESSAGE BEAN: Message received: datasourceid = ["+dataSourceId+"]    "+
                  "message = ["+msg.getText()+"]   message payload = ["+msg.getPropertyNames()+"]");
        dataSourceId = msg.getLongProperty( "dataSourceId" );
        long timeToSummarize = msg.getLongProperty( "timeToSummarize" );

        SimpleDateFormat format = new SimpleDateFormat( "MM/dd/yy hh:mm:ss a" );
        dateStr = format.format( new Date( timeToSummarize ) );
        logger.info( "Executing summary for datasource " + dataSourceId + " at " + dateStr );

        boolean isFirstSample= msg.getBooleanProperty("firstSample");
        executeSummary( sqlDataSource, dataSourceId, timeToSummarize, isFirstSample );

        if( logger.isTraceEnabled() ) logger.debug( "MESSAGE BEAN: Summarization completed: datasourceid = ["+dataSourceId+"]    "+
                  "message = ["+msg.getText()+"]   message payload = ["+msg.getPropertyNames()+"]");
        status = "success";
      }
    }
    catch( Throwable t )
    {
      logger.error( "Error process message", t );
    }
    finally
    {
      if( logger.isDebugEnabled() ) logger.debug( "Completed performance summary for " + dataSourceId + " at " + dateStr + " was a " + status + "." );
    }
  }

  private SummaryInfo fillSummaryInfo(final SummaryInfo summaryInfo, Connection connection) throws Exception
  {
    //Read data source name from database
    String sql = "SELECT d.id, d.name FROM acquisition.ds d where d.id = " + summaryInfo.getDataSourceId();
    PreparedStatementWrapper preparedStatement = PreparedStatementWrapper.create( connection, sql, new TableResultParserDefault<Object>()
    {

      //fill summary data
      @Override
      public List<Object> parseObject(ResultSet resultSet) throws SQLException
      {
        summaryInfo.setDataSourceId(resultSet.getLong(1));
        summaryInfo.setDataSourceName(resultSet.getString(2));
        return null;
      }
    } );
    preparedStatement.executeQueryEx();
    return summaryInfo;
  }

  /**
   * Execute the summary process
   *
   * @param sqlDataSource
   * @param dataSourceId
   * @param time
   * @throws Exception
   */
  public void executeSummary(DataSource sqlDataSource, long dataSourceId, long time, boolean isFirstSample) throws Exception
  {
    Connection connection = null;
    try
    {
      //Execute the summary process
      connection = ConnectionHelper.newConnection( sqlDataSource );
      executeSummary( connection, dataSourceId, time, isFirstSample );
    }
    finally
    {
      if( connection != null )
        ConnectionHelper.close( connection );
    }
  }

  /**
   * Execute the summary process
   *
   * @param connection
   * @param dataSourceId
   * @param summaryTime
   * @throws Exception
   */
  public void executeSummary(Connection connection, long dataSourceId, long summaryTime, boolean isFirstSample) throws Exception
  {
    long execStartTime = System.currentTimeMillis();
    long summaryElapseTime = 0;
    long currentElapseTime = 0;
    long totalElapse = 0;

    SummaryInfo summaryInfo = new SummaryInfo();
    summaryInfo.setDataSourceId(dataSourceId);

    try
    {
      //The summary info
      summaryInfo = fillSummaryInfo( summaryInfo, connection );

      // Set time as null if its a firstSample
      if(isFirstSample) {
          SummaryCalculator.calculateSummary(ctx.getUserTransaction(), connection, dataSourceId, null);
      } else {
          SummaryCalculator.calculateSummary(ctx.getUserTransaction(), connection, dataSourceId, summaryTime);
      }
      totalElapse = System.currentTimeMillis() - execStartTime;

      summaryInfo.setSuccess( true );
      summaryInfo.setDuration( totalElapse );
      summaryInfo.setSummaryTime(summaryTime);
      summaryInfo.setStartTime(execStartTime);
    }
    catch( Throwable t )
    {
      summaryInfo.setSuccess( false );
      summaryInfo.setErrorMessage( t.getMessage() );
      logger.error( "Error trying to do summary for dataSource " + dataSourceId + " at time " + summaryTime, t );
    }
    finally
    {
      //Send summary completed message
      JmsSender sender = JmsSender.INSTANCE;
      sender.sendSummaryNotification( summaryInfo );
    }

    if( totalElapse > 20000 )
    {
      logger.warn( "method 'execute(...)' in class " + this.getClass() + " used (ms) " + totalElapse );
      logger.warn( "summary used " + summaryElapseTime );
      logger.warn( "current sample used " + currentElapseTime );
    }
  }
}
