package com.netapp.oci.netapp.performance.shared.interfaces.impl;

import com.onaro.commons.types.Period;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.jms.*;
import javax.naming.InitialContext;
import java.util.*;

public class PerformanceSummaryJobStart
{
  private static final Logger logger = LoggerFactory.getLogger( PerformanceSummaryJobStart.class );

  private boolean force;
  private long dataSourceId;
  private long sampleTime;
  private static final Map<Long,Long> performanceTimeMap;

  // Map used to determine summarization time
  static {
      HashMap<Long,Long> map = new HashMap<>();
      performanceTimeMap = Collections.synchronizedMap(map);
  }

  public PerformanceSummaryJobStart(long dataSourceId, long sampleTime)
  {
    this.dataSourceId = dataSourceId;
    this.sampleTime = sampleTime;
  }

  public PerformanceSummaryJobStart(long dataSourceId, long sampleTime, boolean force)
  {
    this.dataSourceId = dataSourceId;
    this.sampleTime = sampleTime;
    this.force = force;
  }

  public boolean isForce()
  {
    return force;
  }

  public void setForce(boolean force)
  {
    this.force = force;
  }

  public long getDataSourceId()
  {
    return dataSourceId;
  }

  public void setDataSourceId(long dataSourceId)
  {
    this.dataSourceId = dataSourceId;
  }

  public long getSampleTime()
  {
    return sampleTime;
  }

  public void setSampleTime(long sampleTime)
  {
    this.sampleTime = sampleTime;
  }

  public void execute()
  {
    Connection conn = null;
    Session session = null;
    MessageProducer producer = null;
    InitialContext ic = null;

    try
    {
      if( logger.isDebugEnabled() && force )
        logger.debug( "Force start of summary process for " + dataSourceId + " for time " + sampleTime);
      else
        logger.debug( "Check if summary process needed for " + dataSourceId + " for time " + sampleTime);

      ic = new InitialContext();
      // Creating connection factory
      ConnectionFactory cnf = (ConnectionFactory) ic.lookup( "java:/ConnectionFactory" );
      // Creating Connection
      conn = cnf.createConnection();
      conn.start();

      // Creating Session
      session = conn.createSession( false, Session.AUTO_ACKNOWLEDGE );
      Destination destination = (Destination) ic.lookup( "java:/queue/SummaryQueue" );

      // Creating producer
      producer = session.createProducer( destination );

      // Creating Messages
      TextMessage msg = session.createTextMessage( "Message" );
      msg.setText( "PerformanceSummaryJobStartMessage" );
      msg.setLongProperty( "dataSourceId", dataSourceId);
      Calendar calendar = Calendar.getInstance();
      calendar.setTime(new Date(sampleTime));
      calendar.add( Calendar.HOUR, -1 );
      calendar.set(Calendar.MILLISECOND, 0);
      calendar.set(Calendar.SECOND, 0);
      calendar.set(Calendar.MINUTE, 0);
      long timeToSummarize = calendar.getTimeInMillis();
      msg.setLongProperty( "timeToSummarize", timeToSummarize);

      msg.setBooleanProperty( "force", force );

      if(performanceTimeMap.get(dataSourceId) == null) {
          msg.setBooleanProperty("firstSample", true); //no previous samples
      } else {
          msg.setBooleanProperty("firstSample",false);
      }

      if( logger.isTraceEnabled() ) logger.trace( "Determining is its an hourly sample : datasourceid = ["+dataSourceId+"]    " +
                "message = ["+ msg.getText()+"]");

      // Run only if its a hourly sample or force flag is set. Message Queue only contains messages every hour for a datasource.
      if( force || shouldExecuteHourlySummary( dataSourceId, sampleTime ) ) {
          if( logger.isTraceEnabled() ) logger.trace( "Before message sent to Queue : datasourceid = ["+dataSourceId+"]    " +
                  "message = ["+ msg.getText()+"]   message payload= ["+msg+"]");

          producer.send( msg );
          logger.info( "Summary message sent to Queue : Data source id = ["+dataSourceId+"]    " +
                  "message = ["+ msg.getText()+"] message payload= ["+msg+"]");

          //only update if the current time is greater the last time
          if( performanceTimeMap.get( dataSourceId ) == null || performanceTimeMap.get( dataSourceId ) < sampleTime )
              performanceTimeMap.put( dataSourceId, sampleTime );
      }
    }
    catch( Throwable t )
    {
      logger.error( "Error processing performance summary.", t );
    }
    finally
    {
      try {
        if( producer != null )
          producer.close();
      } catch( Throwable t ) {
        logger.error( "Error cleaning up performance summary process.", t );
      }
      try {
        if( session != null )
          session.close();
      } catch( Throwable t ) {
        logger.error( "Error cleaning up performance summary process.", t );
      }
      try {
        if( conn != null )
          conn.close();
      } catch( Throwable t ) {
        logger.error( "Error cleaning up performance summary process.", t );
      }
       try {
        if( ic != null )
          ic.close();
      } catch( Throwable t ) {
        logger.error( "Error cleaning up performance summary process.", t );
      }

      if( logger.isDebugEnabled() && force )
        logger.debug( "Finished forced summary process for " + dataSourceId + " for time " + sampleTime);
      else
        logger.debug( "Finished check and execute summary process for " + dataSourceId + " for time " + sampleTime);
    }
  }

    /**
     * Tests of see if the summary process needs to be executed for this hour
     * @param dataSourceId
     * @param sampleTime
     * @return
     * @throws Exception
     */
    public boolean shouldExecuteHourlySummary( long dataSourceId, long sampleTime ) throws Exception
    {
        boolean executeSummary = false;

        Long lastSampleTime = performanceTimeMap.get( dataSourceId );
        if( lastSampleTime == null )
            executeSummary = true;
        else
        {
            //Convert the sampleTime to the hour of day this is sures that we are executing for an hour that needs it.
            Calendar calendar = Calendar.getInstance();
            calendar.setTimeInMillis( lastSampleTime );
            int lastReportHourOfDay = calendar.get( Calendar.HOUR_OF_DAY );
            calendar.setTimeInMillis( sampleTime );
            int reportHourOfDay = calendar.get( Calendar.HOUR_OF_DAY );

            // if must use "!=" and not "<" 23:00 is bigger then 00:00
            if( lastReportHourOfDay != reportHourOfDay || sampleTime - lastSampleTime > Period.HOUR.milliseconds() )
                executeSummary = true;
        }

        return executeSummary;
  }
}
