/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.wafl.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.wafl.interfaces.data.WaflMeasurements;
import com.netapp.oci.netapp.performance.wafl.interfaces.local.WaflPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.wafl.interfaces.remote.WaflPerformanceQueryRemote;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import org.slf4j.Logger;
import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class WaflPerformanceQueryBean implements WaflPerformanceQueryRemote, WaflPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger =org.slf4j.LoggerFactory.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<WaflMeasurements>>> getSamples(Set<Long> waflIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + waflIds.size() + "wafl samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<WaflMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<WaflMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(waflIds);

            Map<Long, Set<Long>> clusterIdToWaflIds = NetAppModel.INSTANCE.getClusterIdByNodeIds(connection, waflIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToWaflIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, waflIds, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find wafl with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get wafl samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query wafl samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<WaflMeasurements>>> getSamples(Long containerId, Set<Long> waflIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + waflIds.size() + "wafl samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, waflIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get wafl samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query wafl samples: " + (System.currentTimeMillis() - startTime));
        }
    }

    @Override
    public List<SampleInfo<WaflMeasurements>> getSamples(Long containerId, Long waflId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(waflId), fromTime, toTime).get(waflId);
    }

    private Map<Long, List<SampleInfo<WaflMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> waflIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getWaflPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(waflIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<WaflMeasurements>> getLatestSample(Long clusterId, Set<Long> waflIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + waflIds.size() + "wafl samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getWaflPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(waflIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get wafl samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<WaflMeasurements>>> getSummaries(Set<Long> waflIds, Date fromTime, Date toTime)
            throws NetappInterfaceException {
        logger.info("Query " + waflIds.size() + "wafl summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<WaflMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(waflIds);

            Map<Long, Set<Long>> clusterIdToWaflIds = NetAppModel.INSTANCE.getClusterIdByNodeIds(connection, waflIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdToWaflIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<WaflMeasurements>> summaries = ComponentsFactoryRepository
                        .getWaflPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSummariesGroupById(new TimeSpan(fromTime, toTime), ids);

                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find wafl with id "
                        + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        }
        catch (Exception e) {
            throw new NetappInterfaceException("Failed to get wafl summaries", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<WaflMeasurements>>> getSummaries(Long containerId, Set<Long> waflIds, Date fromTime,
            Date toTime) throws NetappInterfaceException {
        logger.info("Query " + waflIds.size() + "wafl summaries from " + fromTime + " to " + toTime
                + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<WaflMeasurements>>> idToSummaries = new HashMap<>();

            List<Summary<WaflMeasurements>> summaries = ComponentsFactoryRepository
                    .getWaflPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                    .getSummariesGroupById(new TimeSpan(fromTime, toTime), waflIds);

            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);

            return idToSummaries;
        }
        catch (Exception e) {
            throw new NetappInterfaceException("Failed to get wafl summaries", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }
    
    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "wafl has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getWaflPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if wafl has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
