/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.startup;

import com.netapp.oci.netapp.Param;
import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.originator.*;
import com.netapp.oci.netapp.originator.handler.*;
import com.netapp.oci.platform.acq.interfaces.data.PackagePollStatus;
import com.netapp.oci.platform.common.table.DBMetaDataUtil;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.originator.OriginatorChanges;
import com.netapp.oci.platform.originator.PlatformRegistry;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.runner.DbRunner;
import com.netapp.oci.platform.plugin.data.PlatformPlugin;
import com.netapp.oci.server.UpdateTaskSynchronizer;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import com.onaro.sanscreen.util.Schema;

import java.sql.Connection;
import java.util.*;

/**
 * @author Al Velella
 *         Date: Nov 10, 2011
 *         Time: 5:28:03 PM
 */
public class NetAppPlugin implements PlatformPlugin {

    private static final long serialVersionUID = 1L;

    public static final String CONFIG_DB = "netapp_model";
    public static final String PERFORMANCE_DB = "netapp_performance";
    public static final String CHANGE_RECORD_TABLE = CONFIG_DB + "." + "change_record";
    public static final String CHANGED_ATTRIBUTE_TABLE = CONFIG_DB + "." + "changed_attribute";
    public static final String CHANGE_DELETE_DETAIL_TABLE = CONFIG_DB + "." + "change_delete_detail";

    private static Map<String, Set<String>> performanceDatabasesToTables = null;
    private OriginatorChangesHandler originatorChangesHandler;
    private String buildNumber;

    public NetAppPlugin() {
        this.originatorChangesHandler = new OriginatorChangesHandler();
        this.buildNumber = ResourceBundle.getBundle("com.netapp.oci.netapp.about").getString("build");
    }

    @Override
    public Schema getSchema() {
        return Schema.NETAPP;
    }

    @Override
    public String getBuildNumber() {
        return buildNumber;
    }

    @Override
    public void deleteTables(Connection connection) throws ServerInterfaceException {
        try {
            Registry.INSTANCE.deleteTables(connection);
            PreparedStatementWrapper.createForUpdate(connection, "DELETE FROM " + CHANGE_RECORD_TABLE).executeUpdateEx();
            PreparedStatementWrapper.createForUpdate(connection, "DELETE FROM " + CHANGED_ATTRIBUTE_TABLE).executeUpdateEx();
            PreparedStatementWrapper.createForUpdate(connection, "DELETE FROM " + CHANGE_DELETE_DETAIL_TABLE).executeUpdateEx();
        } catch (TableException e) {
            throw new ServerInterfaceException("Failed to delete tables", e);
        }
    }

    @Override
    public Map<String, XmlHandler> getTagToXmlHandlers() {
        return Registry.INSTANCE.getTagToXmlHandlers();
    }

    @Override
    public Map<String, DbRunner> getTagToDbRunners() {
        return Registry.INSTANCE.getTagToDbRunners();
    }

    @Override
    public Map<String, String> getTagToResourceType() {
        return Registry.INSTANCE.getTagToResourceType();
    }

    @Override
    public void postOriginatorUpdate(Connection connection, long originId, long datasourceId, long time) throws ServerInterfaceException {
        new PostOriginatorUpdate().postProcess(connection, originId, datasourceId, time);
    }

    @Override
    public Map<String, List<String>> getTagToChangeDetailAttributes() {
       return Registry.INSTANCE.getTagToChangeDetailAttributes();
    }

    @Override
    public Set<String> getRootTagnames() {
        return Collections.singleton(ClusterHandler.TAG);
    }

    @Override
    public long getRootId(Connection connection, long originId) throws TableException {
        return NetAppModel.INSTANCE.getClusterIdByOriginatorId(connection, originId);
    }

    @Override
    public String getUpdateHistoryTableName() {
        return "netapp_model.update_history";
    }

    @Override
    public void logUpdateHistory(Connection connection, long originatorId, long startTime, long updateTime, long duration, int physicalChangeCount, int runnerCount) throws TableException {

        PreparedStatementWrapper.createForUpdate(connection,
                "INSERT INTO " +
                        getUpdateHistoryTableName() +
                        "  (createtime, starttime, duration, originid, physicalchange, runnerchange)" +
                        " VALUES" +
                        "  (?, ?, ?, ?, ?, ?)")
                .setNext(updateTime)
                .setNext(startTime)
                .setNext(duration)
                .setNext(originatorId)
                .setNext(physicalChangeCount)
                .setNext(runnerCount)
                .executeUpdateEx();
    }


    @Override
    public String getDeleteDetailTableName() {
        return CHANGE_DELETE_DETAIL_TABLE;
    }

    @Override
    public Map<String, String> getDeleteDetailColumnToTagname() {
        Map<String, String> columnToTagname = new HashMap<String, String>();
        columnToTagname.put("clusterid", ClusterHandler.TAG);
        columnToTagname.put("nodeid", NodeHandler.TAG);
        columnToTagname.put("vserverid", VserverHandler.TAG);
        columnToTagname.put("volumeid", VolumeHandler.TAG);
        columnToTagname.put("aggregateid", AggregateHandler.TAG);
        columnToTagname.put("name", null);
        return columnToTagname;
    }

    @Override
    public List<String> getTableOrder() {
        return Collections.unmodifiableList(Registry.INSTANCE.getTableOrder());
    }

    @Override
    public Set<String> getConfigurationDatabases() {
        return new HashSet<String>(Arrays.asList(CONFIG_DB, "netapp_model_view"));
    }

    public Map<String, Set<String>> getPerformanceDatabasesAndTables(Connection connection) throws ServerInterfaceException {
        try {
            if (NetAppPlugin.performanceDatabasesToTables == null) {
                NetAppPlugin.initializePerformanceTables(connection);
            }
            return performanceDatabasesToTables;
        } catch (TableException e) {
            throw new ServerInterfaceException("Failed to load list of databases and tables", e);
        }
    }

    @Override
    public void handleOriginatorChanges(Connection connection, long datasourceId, OriginatorChanges changes) throws ServerInterfaceException {
        originatorChangesHandler.logOriginatorChangesInfo(connection, changes);
    }

    @Override
    public void logOriginatorRemove(Connection connection, long originId, long clusterId, long updateTime) throws ServerInterfaceException {
        originatorChangesHandler.logOriginatorRemove(connection, originId, clusterId, updateTime);
    }

    @Override
    public void logObjectsPurgeJob(Connection connection, long clusterId, Map<String,Set<Long>> tagToIdsBeingDeleted, long updateTime) throws ServerInterfaceException, TableException {
        new OriginatorChangesTable().insertChangedRecordBatch(connection, Collections.singletonMap(ChangeType.PURGED, tagToIdsBeingDeleted), clusterId, updateTime);
    }

    @Override
    public void changesCommitted(Connection connection, long datasourceId, String datasourceName, OriginatorChanges changes) throws ServerInterfaceException {
        originatorChangesHandler.sendChangeNotification(connection, datasourceId, datasourceName, changes);
    }

    @Override
    public void notifyOriginatorRemove(Connection connection, long datasourceId, String datasourceName, long clusterId, long updateTime) throws ServerInterfaceException {
        JmsSender.INSTANCE.sendChangeNotification(datasourceId, datasourceName, clusterId, updateTime, ChangeType.DELETED);
    }

    @Override
    public void prune(Connection connection, long startTime) throws ServerInterfaceException {
        try {
            long historyRecordLifeTime = Param.get(Param.Key.HISTORY_LIFETIME, long.class);
            long changesRecordLifeTime = Param.get(Param.Key.CHANGES_LIFETIME, long.class);

            PruneTask task = new PruneTask(historyRecordLifeTime, changesRecordLifeTime);
            UpdateTaskSynchronizer.getInstance().execute(task);
            int changesCount = task.getAllChangesCount();
            if (changesCount > 0) {
                logUpdateHistory(connection, PruneTask.PRUNE_TASK_ORIGIN_ID, startTime, task.getUpdateTime(), System.currentTimeMillis() - startTime, changesCount, 0);
            }
        } catch (Exception e) {
            throw new ServerInterfaceException("Error while pruning NetApp Schema", e);
        }

    }

    @Override
    public void handlePackageMessage(Connection connection, PackagePollStatus pollStatus) throws ServerInterfaceException {
        JmsSender.INSTANCE.sendPackageNotification(connection, pollStatus);
    }

    private static synchronized void initializePerformanceTables(Connection connection) throws TableException {
        if (performanceDatabasesToTables == null) {
            final Map<String, Set<String>> tempDatabaseToTables = new HashMap<String, Set<String>>();
            final Set<String> tempTables = DBMetaDataUtil.getTableWithNames(connection, PERFORMANCE_DB, "sample_");
            tempTables.addAll(DBMetaDataUtil.getTableWithNames(connection, PERFORMANCE_DB, "summary_"));
            tempTables.addAll(DBMetaDataUtil.getTableWithNames(connection, PERFORMANCE_DB, "update_history"));
            tempDatabaseToTables.put(PERFORMANCE_DB, Collections.unmodifiableSet(tempTables));
            performanceDatabasesToTables = Collections.unmodifiableMap(tempDatabaseToTables);
        }
    }
    public <T> T getParam(String key, Class<T> clazz, T defaultValue){
        return Param.get(key, clazz, defaultValue);

    }

    @Override
    public PlatformRegistry getRegistry() {
        return Registry.INSTANCE;
    }
}
