/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.upgrade.version_1;

import com.netapp.oci.platform.common.table.StringCache;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.netapp.oci.platform.upgrade.UpgradeRunner;
import com.netapp.oci.platform.upgrade.Upgrader;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

/**
 * @author Al Velella
 *         Date: 12/5/12
 *         Time: 6:00 PM
 */
public class Upgrade_1_0_0_69 implements Upgrader {
    private static class ProblemRecord {
        private ProblemRecord(long qosId, long vserverId, long clusterId, String clusterUuid) {
            this.qosId = qosId;
            this.vserverId = vserverId;
            this.clusterId = clusterId;
            this.clusterUuid = clusterUuid;
        }

        public long qosId;
        public long vserverId;
        public long clusterId;
        public String clusterUuid;
    }

    private TableResultParserDefault<ProblemRecord> fixRecordParser = new TableResultParserDefault<ProblemRecord>() {
        @Override
        public ProblemRecord parseObject(ResultSet resultSet) throws SQLException {
            return new ProblemRecord(resultSet.getLong(1),
                    resultSet.getLong(2),
                    resultSet.getLong(3),
                    StringCache.get(resultSet.getString(4)));
        }
    };

    @Override
    public void run(UpgradeRunner upgradeRunner) throws TableException {
        updateTable(upgradeRunner, "sample_qos_volume_workload", true);
        updateTable(upgradeRunner, "summary_qos_volume_workload", false);
        updateSampleCurrentTable(upgradeRunner);
    }

    private void updateSampleCurrentTable(UpgradeRunner upgradeRunner) throws TableException {
        // find candidate records
        List<ProblemRecord> problemRecords = upgradeRunner.sqlQuery(
                " SELECT" +
                        "   sc.objid AS objid, vs.objid AS vserverId, c.objid AS clusterId, c.uuid AS clusterUuid" +
                        " FROM" +
                        "   netapp_performance.sample_current_qos_volume_workload sc" +
                        " JOIN" +
                        "   netapp_model.vserver vs" +
                        " ON" +
                        "   sc.containerid=vs.objid" +
                        " JOIN" +
                        "   netapp_model.cluster c" +
                        " ON" +
                        "   vs.clusterid=c.objid", fixRecordParser).executeQueryEx();

        // find current partitions
        String performanceDb = upgradeRunner.getPerformanceDb();
        Set<String> partitionNames = new HashSet<String>(upgradeRunner.sqlQuery(
                " SELECT" +
                        "   partition_name" +
                        " FROM" +
                        "   information_schema.PARTITIONS" +
                        " WHERE" +
                        "   table_schema='" + performanceDb + "' AND table_name='sample_current_qos_volume_workload'", TableResultParserDefault.STRING).executeQueryEx());


        // add cluster partitions
        for (ProblemRecord problemRecord : problemRecords) {
            String partitionName = "p" + problemRecord.clusterId;
            if (!partitionNames.contains(partitionName)) {
                upgradeRunner.sqlUpdate("ALTER TABLE netapp_performance.sample_current_qos_volume_workload ADD PARTITION (PARTITION " + partitionName + " VALUES IN (" + problemRecord.clusterId + "))").executeUpdateEx();
                partitionNames.add(partitionName);
            }
        }

        // move data from vserver to cluster
        for (ProblemRecord problemRecord : problemRecords) {
            if (upgradeRunner.sqlFunctionQuery("SELECT COUNT(*) from netapp_performance.sample_current_qos_volume_workload WHERE objid=" + problemRecord.qosId) != 1) {
                upgradeRunner.sqlUpdate(
                        " UPDATE" +
                                "   netapp_performance.sample_current_qos_volume_workload" +
                                " SET" +
                                "   containerkey=" + problemRecord.clusterUuid +
                                "   containerid=" + problemRecord.clusterId +
                                " WHERE" +
                                "   objid=" + problemRecord.qosId
                ).executeUpdateEx();

                String vserverPartition = "p" + problemRecord.vserverId;
                if (partitionNames.contains(vserverPartition)) {
                    upgradeRunner.sqlUpdate("ALTER TABLE netapp_performance.sample_current_qos_volume_workload DROP PARTITION " + vserverPartition).executeUpdateEx();
                    partitionNames.remove(vserverPartition);
                }
            }
        }
    }

    private void updateTable(UpgradeRunner upgradeRunner, String tablePrefix, boolean partition) throws TableException {
        // look at the created tables and find their container id
        String performanceDb = upgradeRunner.getPerformanceDb();
        Set<String> tables = new HashSet<String>(upgradeRunner.sqlQuery(
                " SELECT" +
                        "   TABLE_NAME" +
                        " FROM" +
                        "   information_schema.TABLES" +
                        " WHERE" +
                        "   table_schema='" + performanceDb + "' AND table_name LIKE '" + tablePrefix + "_%'", TableResultParserDefault.STRING).executeQueryEx());
        Set<Long> candidateIds = new HashSet<Long>();
        for (String table : tables) {
            candidateIds.add(Long.valueOf(table.substring(tablePrefix.length() + 1)));
        }

        // see how many of these container ids are vservers
        upgradeRunner.createTemporaryTableForJoin(candidateIds, "u69temp");
        List<ProblemRecord> problemRecords = upgradeRunner.sqlQuery(
                " SELECT" +
                        "   0 AS objid, vs.objid AS vserverId, c.objid AS clusterId, c.uuid AS clusterUuid" +
                        " FROM" +
                        "   netapp_model.vserver vs" +
                        " JOIN" +
                        "   u69temp t" +
                        " ON" +
                        "   t.objid=vs.objid" +
                        " JOIN" +
                        "   netapp_model.cluster c" +
                        " ON" +
                        "   c.objid=vs.clusterid", fixRecordParser).executeQueryEx();

        for (ProblemRecord problemRecord : problemRecords) {
            String vserverBasedName = tablePrefix + "_" + problemRecord.vserverId;
            String clusterBasedName = tablePrefix + "_" + problemRecord.clusterId;
            if (!tables.contains(clusterBasedName)) {
                // easy case, just rename the table
                upgradeRunner.sqlUpdate("RENAME TABLE " + performanceDb + "." + vserverBasedName + " TO " + performanceDb + "." + clusterBasedName).executeUpdateEx();
                tables.add(clusterBasedName);
            } else if (tables.contains(clusterBasedName) && tables.contains(vserverBasedName)){
                // merge table
                upgradeRunner.sqlUpdate("INSERT IGNORE INTO " + performanceDb + "." + clusterBasedName + " (SELECT * FROM " + performanceDb + "." + vserverBasedName + ")").executeUpdateEx();
                upgradeRunner.sqlUpdate("DROP TABLE " + performanceDb + "." + vserverBasedName).executeUpdateEx();
            }
            tables.remove(vserverBasedName);
        }

        if (partition) {
            for (String table : tables){
                // drop all the partitions so we can start the numbering from 0
                upgradeRunner.sqlUpdate("ALTER TABLE netapp_performance." + table + " REMOVE PARTITIONING").executeUpdateEx();

                long minDate = upgradeRunner.sqlFunctionQuery("SELECT MIN(time) FROM netapp_performance." + table);
                long maxDate = upgradeRunner.sqlFunctionQuery("SELECT MAX(time) FROM netapp_performance." + table);
                if (maxDate == 0) {
                    maxDate = System.currentTimeMillis();
                    minDate = maxDate;
                }

                // get midnight tonight
                List<Date> times = new ArrayList<Date>(30);
                Calendar currentCal = Calendar.getInstance();
                currentCal.setTime(new Date(maxDate));
                Calendar midnightCal = Calendar.getInstance();
                midnightCal.clear();
                midnightCal.set(currentCal.get(Calendar.YEAR), currentCal.get(Calendar.MONTH), currentCal.get(Calendar.DAY_OF_MONTH));
                midnightCal.add(Calendar.DAY_OF_MONTH, 1);

                // build partitions
                while (midnightCal.getTime().getTime() > minDate) {
                    times.add(midnightCal.getTime());
                    midnightCal.add(Calendar.DAY_OF_MONTH, -1);
                }

                // max 30 days
                while (times.size() > 30) {
                    times.remove(times.size() - 1);
                }
                Collections.reverse(times);

                // add partitions back
                StringBuilder sql = new StringBuilder("ALTER TABLE netapp_performance.").append(table).append(" PARTITION BY RANGE(time) (");
                int i = 0;
                for (Date time : times) {
                    if (i > 0) {
                        sql.append(',');
                    }
                    sql.append(" PARTITION p").append(i).append(" VALUES LESS THAN (").append(time.getTime()).append(")");
                    i++;
                }
                sql.append(')');
                upgradeRunner.sqlUpdate(sql.toString()).executeUpdateEx();
            }
        }
    }
}
