/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.upgrade.version_1_2;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import org.slf4j.Logger;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.netapp.oci.platform.upgrade.UpgradeRunner;
import com.netapp.oci.platform.upgrade.Upgrader;

/**
 * @author Abhishek Hiregoudar, Bharath Ramaprasad
 */
public class Upgrade_1_2_0_20 implements Upgrader {

	private static final String QOS_WORKLOAD_DETAIL = "qos_workload_detail";
	private static final String QOS_WORKLOAD_CONSTITUENT = "qos_workload_constituent";
	private static final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger(Upgrade_1_2_0_20.class);
	
	private static final List<String> bakConfigTableList = new ArrayList<String>(){{
		add("bak_qos_workload_detail");
		add("bak_qos_workload_constituent");
	}};

	
	@Override
    public void run(UpgradeRunner upgradeRunner) throws TableException {
		
        //Rename specific older config and perf tables for backup
		renameTablesForBackup(upgradeRunner);
        
		//Drop FK constraints on Backed up Config tables 
        dropForeignkeyConstraints(upgradeRunner);
       
        //Create Specific Old & New QosPerformance Sample & Summary Tables
        createQosPerformanceTables(upgradeRunner);
        
        //Populate qwnr config table based on the cluster version
        populateQosWorkloadNodeRelationshipConfigTable(upgradeRunner);            	 
    }
	
    /**
     * Renames specific older config and perf tables for backup
     * @param upgradeRunner - The main thread which runs the upgrader 
     * @throws TableException - In case of creation failure
     */
    private void renameTablesForBackup(UpgradeRunner upgradeRunner) throws TableException {
    	/**
    	 * Map< key=table_name, value=schema_name >
    	   SqlQuery should not return a list of maps, as a map suffices. This calls for a different parseResult implementation as part of TableResultParserDefault
    	 */    	
    		final List<Map<String, String>> tables = upgradeRunner.sqlQuery(
    			"SELECT table_name, table_schema" +
    			" FROM information_schema.tables" +
    			" WHERE table_schema IN ('" + upgradeRunner.getPerformanceDb() + "', '" + upgradeRunner.getConfigurationDb() + "')" +
    			" AND" +
    			" (table_name like 'sample_qos_workload_detail_%'" +
    			" OR table_name like 'summary_qos_workload_detail_%'" + 
    			" OR table_name like 'sample_qos_service_center_%'" +
    			" OR table_name like 'summary_qos_service_center_%'" +
    			" OR table_name like 'sample_qos_workload_constituent_%'" +
    			" OR table_name like 'summary_qos_workload_constituent_%'" +
    			" OR table_name like 'summary_qos_volume_workload_%'" +
    			" OR table_name like 'sample_qos_volume_workload_%'" +
    			" OR table_name like '"+ QOS_WORKLOAD_CONSTITUENT +"'"+
    			" OR table_name like '"+ QOS_WORKLOAD_DETAIL +"')", TableResultParserDefault.newStringStringMap()).executeQueryEx();

    	
    	// Renaming perf tables and config tables to bak_<table_name> For example : Renaming from 'netapp_performance.sample_qos_service_center_%' 
    	// To 'netapp_performance.bak_sample_qos_service_center_%'.
    	    	
    	for (Map<String, String> table : tables){
    		for (Map.Entry<String, String> entry :  table.entrySet()){
    			if(!entry.getKey().equals(QOS_WORKLOAD_DETAIL))
    				upgradeRunner.sqlUpdate("ALTER TABLE " + entry.getValue() + "." + entry.getKey() + " RENAME TO " + entry.getValue() + ".bak_" + entry.getKey()).executeUpdateEx();
    			else
    			{
    				//copy the QOS_WORKLOAD_DETAIL config table to a temporary table which will be used for Data Migration 
   
    				String copyWkldDetailStructure = "CREATE TABLE "+  upgradeRunner.getConfigurationDb() 
    												  + ".bak_qos_workload_detail LIKE " + upgradeRunner.getConfigurationDb() + "." + QOS_WORKLOAD_DETAIL;
    				String copyWkldDetailConfig = "INSERT "+  upgradeRunner.getConfigurationDb() 
    												  + ".bak_qos_workload_detail select * from " + upgradeRunner.getConfigurationDb() + "." + QOS_WORKLOAD_DETAIL; 
    				    				
    				upgradeRunner.sqlUpdate(copyWkldDetailStructure).executeUpdateEx();
    				upgradeRunner.sqlUpdate(copyWkldDetailConfig).executeUpdateEx();
    								
    			}                
    		}
    	}    	
    	logger.info("older config and perf tables are backed up");
    }
    
    /**
     * Drops FK constraints on Backed up Config tables
     * @param upgradeRunner - The main thread which runs the upgrader 
     * @throws TableException - In case of creation failure
     */
	private void dropForeignkeyConstraints(UpgradeRunner upgradeRunner) throws TableException{
		// drop all the FK constraints on bak config tables
		for(String bakConfigTable : bakConfigTableList)
		{
			String dropConstraintsQuery = "select constraint_name from information_schema.TABLE_CONSTRAINTS where table_name = '" + bakConfigTable +  "' AND constraint_type = 'FOREIGN KEY'" +
					" AND table_schema = '" + upgradeRunner.getConfigurationDb() + "'";
	        List<String> fkConstraintNames = upgradeRunner.sqlQuery(dropConstraintsQuery, TableResultParserDefault.STRING).executeQueryEx(); 			
	           
	        for(String constraintName : fkConstraintNames){
				String dropQuery = "ALTER TABLE "+  upgradeRunner.getConfigurationDb() + "." + bakConfigTable  +  
								  " DROP FOREIGN KEY " + constraintName ;		
				upgradeRunner.sqlUpdate(dropQuery).executeUpdateEx();			
			}

		}
	}
	
    /**
     * Populates QosWorkloadNodeRelationship config table based on the cluster version - populates for SN clusters only. 
     * @param upgradeRunner - The main thread which runs the upgrader 
     * @throws TableException - In case of creation failure
     */
    private void populateQosWorkloadNodeRelationshipConfigTable(UpgradeRunner upgradeRunner) throws TableException {
        
    	//Fetch the current objid sequence from management.id table to initialize the range for auto-incrementing objId 
         List<Long> objIdList = upgradeRunner.sqlQuery(
                "SELECT \n" +
                        "    value\n" +
                        "FROM\n" +
                        "    management.id\n" +
                        "WHERE\n" +
                        "    id = 1", TableResultParserDefault.LONG).executeQueryEx();
       
         if (objIdList.size() > 0) {
            Long objId = objIdList.get(0);
            
            //Create and populate a temp table with objids and config data
            String tempTableName = upgradeRunner.getConfigurationDb() + ".tmp_qos_workload_node_relationship";
            
            //Drop Temp table
            String dropTempTable = "DROP TEMPORARY TABLE IF EXISTS " + tempTableName;
            upgradeRunner.sqlUpdate(dropTempTable).executeUpdateEx();
            
            //Create & populate Temp table
            String createTempTable_DDL_DML = "CREATE TEMPORARY TABLE " + tempTableName + " (\n" +
                    "   objid BIGINT(20) NOT NULL AUTO_INCREMENT,\n" +
                    "   originid BIGINT(20) NOT NULL,\n" +
                    "   uuid VARCHAR(255) NOT NULL ,\n" +
                    "   resourceKey VARCHAR(512)  NOT NULL,\n" +
                    "   clusterId BIGINT(20) NOT NULL,\n" +
                    "   nodeId BIGINT(20) NOT NULL,\t\n" +
                    "   workloadId BIGINT(20) NOT NULL,\t\n" +
                    "   checksum INT(11) NOT NULL,\n" +
                    "   changeDetailChecksum INT(11) NOT NULL,\n" +
                    "   createtime BIGINT(20) NOT NULL,\n" +
                    "   updatetime BIGINT(20) NOT NULL,\n" +
                    "   objState ENUM('LIVE','DEAD'),\n" +
                    "   PRIMARY KEY (objid)\n" +
                    ") AUTO_INCREMENT = " + objId + " select \n" +
                    "   n.originid originid,\n" +
                    "   CONCAT(n.uuid, '.', SUBSTRING_INDEX(qvw.uuid, '.', -1)) UUID,\n" + 
                    "   CONCAT(n.uuid, ':type=qos_workload_node_relationship,workload_uuid=', SUBSTRING_INDEX(qvw.uuid, '.', -1)) resourceKey,\n" +
                    "    n.clusterId clusterId,\n" +
                    "    n.objid nodeId,\n" +
                    "    qvw.objid workloadId,\n" +
                    "    0 checksum,\n" +
                    "    0 changeDetailChecksum,\n" +
                    "    qvw.createtime createtime,\n" + 
                    "    qvw.updatetime updatetime,\n" + 
                    "    qvw.objState objState\n" + 
                    " from\n" +
                     	 upgradeRunner.getConfigurationDb() + ".qos_volume_workload qvw\n"+
                    " 	  INNER JOIN\n"+             
                    upgradeRunner.getConfigurationDb() + ".node n ON qvw.clusterId = n.clusterId\n"+
                    "	  INNER JOIN\n" +  
                    upgradeRunner.getConfigurationDb() + ".cluster c ON c.objId =  n.clusterId WHERE c.versionGeneration = 8 AND c.versionMajor = 2\n";
                    
            upgradeRunner.sqlUpdate(createTempTable_DDL_DML).executeUpdateEx();
           
            //Insert into qos_workload_node_relationship table.
            String qwnrInsertDML = "INSERT INTO netapp_model.qos_workload_node_relationship \n" +
                    "SELECT \n" +
                    "    objid,\n" +
                    "    originid,\n" +
                    "    uuid,\n" +
                    "    resourceKey,\n" +
                    "    clusterId,\n" +
                    "    nodeId,\n" +
                    "    workloadId,\n" +
                    "    checksum,\n" +
                    "    changeDetailChecksum,\n" +
                    "    createtime,\n" +
                    "    updatetime,\n" +
                    "    objState\n" +
                    "FROM\n" +
                    "    " + tempTableName;
            upgradeRunner.sqlUpdate(qwnrInsertDML).executeUpdateEx();
            
            //Update the object id sequence with the used ids plus one. Round it to the next thousand as objIds are incremented in thousands.
            String updateObjIdSeq = "UPDATE management.id \n" +
                    "SET \n" +
                    "    value = value + IFNULL((1000 * ((((SELECT \n" +
                    "            max(objid)\n" +
                    "        FROM\n" +
                    "         " + upgradeRunner.getConfigurationDb() + ".qos_workload_node_relationship) - value) DIV 1000) + 1)), 0)\n" +
                    "WHERE\n" +
                    "    id = 1";
            upgradeRunner.sqlUpdate(updateObjIdSeq).executeUpdateEx();
        } else {
            throw new TableException("Unable to query the management.id table");
        }
         logger.info("Populated Qos workload node relationship config table based on the cluster version"); 
    }
    
    /**
     * Create QosPerformance Sample & Summary Tables
     * @param upgradeRunner - The main thread which runs the upgrader 
     * @throws TableException - In case of creation failure
     */
    private void createQosPerformanceTables(UpgradeRunner upgradeRunner) throws TableException{
    	
    	/* Retrieve the already registered datasources - cluster ids*/    	
    	List<Long> clusterIdList = upgradeRunner.sqlQuery(
                		"SELECT \n" +
                        " 	objid\n" +
                        "FROM\n" +
                        "    " + upgradeRunner.getConfigurationDb() + ".cluster\n", TableResultParserDefault.LONG).executeQueryEx();
    	
    	 for(Long clusterId : clusterIdList){
    		
        	String queueDbladeSampleDDL = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".sample_qos_workload_queue_dblade_"+ clusterId +" (\n"+
  			      "objid BIGINT(20) NOT NULL DEFAULT '0',\n" + 
  			      " time BIGINT(20) NOT NULL DEFAULT '0',\n" +
  			      " empty TINYINT(1) DEFAULT NULL,\n" + 
  			      " cpuDBladeResidenceTime BIGINT(20) DEFAULT NULL,\n" +
  			      " cpuDBladeUtilServiceTime BIGINT(20) DEFAULT NULL,\n" +
  			      " cpuDBladeBackgroundServiceTime BIGINT(20) DEFAULT NULL,\n" +
  			      " diskHDDBackgroundServiceTime BIGINT(20) DEFAULT NULL,\n" +
  			      " diskSSDBackgroundServiceTime BIGINT(20) DEFAULT NULL,\n" +
  			      " delayWAFLSuspCPWaitTime BIGINT(20) DEFAULT NULL,\n" +
  			      " delayDiskIOWaitTime BIGINT(20) DEFAULT NULL,\n" +
  			      " collectionPeriod BIGINT(20) DEFAULT NULL,\n"+
  			      " PRIMARY KEY (`objid`,`time`),\n"+
  			      " KEY `time_index` (`time`)\n" +  
  			      ") ENGINE=INNODB DEFAULT CHARSET=utf8\n";    		
    		upgradeRunner.sqlUpdate(queueDbladeSampleDDL).executeUpdateEx();
    		
        	String queueDbladeSummaryDDL = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".summary_qos_workload_queue_dblade_"+ clusterId +" (\n"+
     			   "objid bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " fromtime bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " totime bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " cpuDBladeResidenceTime BIGINT(20) DEFAULT NULL,\n" + 
     			   " cpuDBladeUtilServiceTime BIGINT(20) DEFAULT NULL,\n" + 
     			   " cpuDBladeBackgroundServiceTime BIGINT(20) DEFAULT NULL,\n" + 
     			   " diskHDDBackgroundServiceTime BIGINT(20) DEFAULT NULL,\n" + 
     			   " diskSSDBackgroundServiceTime BIGINT(20) DEFAULT NULL,\n" + 
     			   " delayWAFLSuspCPWaitTime BIGINT(20) DEFAULT NULL,\n" + 
     			   " delayDiskIOWaitTime BIGINT(20) DEFAULT NULL,\n" + 
     			   " collectionPeriod BIGINT(20) DEFAULT NULL,\n"+ 
     			   " PRIMARY KEY (objid, fromtime, totime),\n" + 
     			   " KEY totime_index (totime),\n" + 
     			   " KEY fromtime_totime_index (fromtime, totime)\n"+
     			   ") ENGINE=InnoDB DEFAULT CHARSET=utf8\n";
        	upgradeRunner.sqlUpdate(queueDbladeSummaryDDL).executeUpdateEx();
     	
        	String queueNbladeSampleDDL = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".sample_qos_workload_queue_nblade_"+ clusterId +" (\n"+
     			   "objid bigint(20) NOT NULL DEFAULT '0',\n"+
     			   " time bigint(20) NOT NULL DEFAULT '0',\n"+
     			   " empty tinyint(1) DEFAULT NULL,\n"+
     			   " readOps float DEFAULT NULL,\n"+
     			   " writeOps float DEFAULT NULL,\n"+
     			   " otherOps float DEFAULT NULL,\n"+ 
     			   " readLatency float DEFAULT NULL,\n"+
     			   " writeLatency float DEFAULT NULL,\n"+ 
     			   " otherLatency float DEFAULT NULL,\n" + 
     			   " readData float DEFAULT NULL,\n" + 
     			   " writeData float DEFAULT NULL,\n" + 
     			   " cpuNBladeResidenceTime bigint(20) DEFAULT NULL,\n" + 
     			   " cpuNBladeUtilServiceTime bigint(20) DEFAULT NULL,\n" + 
     			   " delayClusterInterconnectWaitTime bigint(20) DEFAULT NULL,\n" + 
     			   " delayQosLimitWaitTime bigint(20) DEFAULT NULL,\n" + 
     			   " delayNetworkWaitTime bigint(20) DEFAULT NULL,\n" + 
     			   " collectionPeriod bigint(20) DEFAULT NULL,\n" + 
     			   " PRIMARY KEY (objid,time),\n" + 
     			   " KEY time_index (time)\n"+ 
     			   ") ENGINE=InnoDB DEFAULT CHARSET=utf8\n"; 
     		upgradeRunner.sqlUpdate(queueNbladeSampleDDL).executeUpdateEx();
     	
     		String queueNbladeSummaryDDL = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".summary_qos_workload_queue_nblade_"+ clusterId +" (\n" + 
     			   "objid bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " fromtime bigint(20) NOT NULL DEFAULT '0',\n"+ 
     			   " totime bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " readOps float DEFAULT NULL,\n" + 
     			   " writeOps float DEFAULT NULL,\n" + 
     			   " otherOps float DEFAULT NULL,\n " + 
     			   " readLatency float DEFAULT NULL,\n " + 
     			   " writeLatency float DEFAULT NULL,\n " + 
     			   " otherLatency float DEFAULT NULL,\n " + 
     			   " readData float DEFAULT NULL,\n" + 
     			   " writeData float DEFAULT NULL,\n " +
     			   " cpuNBladeResidenceTime bigint(20) DEFAULT NULL,\n" + 
     			   " cpuNBladeUtilServiceTime bigint(20) DEFAULT NULL,\n" + 
     			   " delayClusterInterconnectWaitTime bigint(20) DEFAULT NULL,\n" + 
     			   " delayQosLimitWaitTime bigint(20) DEFAULT NULL,\n" + 
     			   " delayNetworkWaitTime bigint(20) DEFAULT NULL,\n" + 
     			   " collectionPeriod bigint(20) DEFAULT NULL,\n" + 
     			   " PRIMARY KEY (objid, fromtime, totime),\n" + 
     			   " KEY totime_index (totime),\n" + 
     			   " KEY fromtime_totime_index (fromtime, totime)\n"+ 
     			   ") ENGINE=InnoDB DEFAULT CHARSET=utf8\n";
     		upgradeRunner.sqlUpdate(queueNbladeSummaryDDL).executeUpdateEx();
     		
     	String qosWkldDetailSampleDDL = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".sample_qos_workload_detail_"+ clusterId +" (\n" +
     			   "objid bigint(20) NOT NULL DEFAULT '0',\n"+
     			   " time bigint(20) NOT NULL DEFAULT '0',\n"+ 
     			   " empty tinyint(1) DEFAULT NULL,\n" + 
     			   " serviceTime bigint(20) DEFAULT NULL,\n"+
     			   " waitTime bigint(20) DEFAULT NULL,\n" + 
     			   " collectionPeriod bigint(20) DEFAULT NULL,\n" +
     			   " PRIMARY KEY (objid, time),\n" + 
     			   " KEY time_index (time)\n" + 
     			   ") ENGINE=InnoDB DEFAULT CHARSET=utf8\n";
     	upgradeRunner.sqlUpdate(qosWkldDetailSampleDDL).executeUpdateEx();
     	
     	String qosWkldDetailSummaryDDL = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".summary_qos_workload_detail_"+ clusterId +" (\n" + 
     			   "objid bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " fromtime bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " totime bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " serviceTime bigint(20) DEFAULT NULL,\n" +
     			   " waitTime bigint(20) DEFAULT NULL,\n" + 
     			   " collectionPeriod bigint(20) DEFAULT NULL,\n" + 
     			   " PRIMARY KEY (objid, fromtime, totime),\n" + 
     			   " KEY totime_index (totime),\n" + 
     			   " KEY fromtime_totime_index (fromtime, totime)\n" + 
     			   ") ENGINE=InnoDB DEFAULT CHARSET=utf8\n"; 
     	upgradeRunner.sqlUpdate(qosWkldDetailSummaryDDL).executeUpdateEx();
     	
     	String qosSvcCenterSampleDDL  = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".sample_qos_service_center_"+ clusterId +" (\n"+
     			   "objid bigint(20) NOT NULL DEFAULT '0',\n"+
     			   " time bigint(20) NOT NULL DEFAULT '0',\n"+ 
     			   " empty tinyint(1) DEFAULT NULL,\n" + 
     			   " waitTime bigint(20) DEFAULT NULL,\n" + 
     			   " utilization float DEFAULT NULL,\n" + 
     			   " collectionPeriod bigint(20) DEFAULT NULL,\n" + 
     			   " PRIMARY KEY (objid, time),\n" + 
     			   " KEY time_index (time)\n" + 
     			   ") ENGINE=InnoDB DEFAULT CHARSET=utf8\n";
     	upgradeRunner.sqlUpdate(qosSvcCenterSampleDDL).executeUpdateEx();

     	String qosSvcCenterSummaryDDL = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".summary_qos_service_center_"+ clusterId +" (\n"+
     			   "objid bigint(20) NOT NULL DEFAULT '0',\n"+ 
     			   " fromtime bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " totime bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " waitTime bigint(20) DEFAULT NULL,\n" + 
     			   " utilization float DEFAULT NULL,\n" + 
     			   " collectionPeriod bigint(20) DEFAULT NULL,\n" + 
     			   " PRIMARY KEY (objid, fromtime, totime),\n" + 
     			   " KEY totime_index (totime),\n" + 
     			   " KEY fromtime_totime_index (fromtime, totime)\n" +
     			   ") ENGINE=InnoDB DEFAULT CHARSET=utf8\n"; 
     	upgradeRunner.sqlUpdate(qosSvcCenterSummaryDDL).executeUpdateEx();

     	String qosVolumeWkldSampleDDL = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".sample_qos_volume_workload_"+ clusterId +" (\n"+
     			   "objid bigint(20) NOT NULL DEFAULT '0',\n"+
     			   " time bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " empty tinyint(1) DEFAULT NULL,\n" + 
     			   " latency float DEFAULT NULL,\n" + 
     			   " ops float DEFAULT NULL,\n" + 
     			   " totalData float DEFAULT NULL,\n" + 
     			   " readData float DEFAULT NULL,\n" + 
     			   " readLatency float DEFAULT NULL,\n" + 
     			   " readOps float DEFAULT NULL,\n" + 
     			   " writeData float DEFAULT NULL,\n" + 
     			   " writeLatency float DEFAULT NULL,\n" + 
     			   " writeOps float DEFAULT NULL,\n" + 
     			   " readIOPercentCache float DEFAULT NULL,\n" + 
     			   " readIOPercentExtCache float DEFAULT NULL,\n" + 
     			   " readIOPercentDisk float DEFAULT NULL,\n" + 
     			   " readIOPercentBambooSsd float DEFAULT NULL,\n" + 
     			   " readIOPercentHyaHdd float DEFAULT NULL,\n" + 
     			   " readIOPercentHyaCache float DEFAULT NULL,\n" + 
     			   " readIOPercentHyaNonCache float DEFAULT NULL,\n" + 
     			   " PRIMARY KEY (objid, time),\n" +
     			   " KEY time_index (time)\n" + 
     			   ") ENGINE=InnoDB DEFAULT CHARSET=utf8\n";
     	upgradeRunner.sqlUpdate(qosVolumeWkldSampleDDL).executeUpdateEx();

     	String qosVolumeWkldSummaryDDL = "CREATE TABLE " + upgradeRunner.getPerformanceDb() + ".summary_qos_volume_workload_"+ clusterId +" (\n" + 
     			   "objid bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " fromtime bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " totime bigint(20) NOT NULL DEFAULT '0',\n" + 
     			   " latency float DEFAULT NULL,\n" + 
     			   " ops float DEFAULT NULL,\n" + 
     			   " totalData float DEFAULT NULL,\n" + 
     			   " readData float DEFAULT NULL,\n" + 
     			   " readLatency float DEFAULT NULL,\n" + 
     			   " readOps float DEFAULT NULL,\n" + 
     			   " writeData float DEFAULT NULL,\n" + 
     			   " writeLatency float DEFAULT NULL,\n" + 
     			   " writeOps float DEFAULT NULL,\n" + 
     			   " readIOPercentCache float DEFAULT NULL,\n" + 
     			   " readIOPercentExtCache float DEFAULT NULL,\n" + 
     			   " readIOPercentDisk float DEFAULT NULL,\n" + 
     			   " readIOPercentBambooSsd float DEFAULT NULL,\n" + 
     			   " readIOPercentHyaHdd float DEFAULT NULL,\n" + 
     			   " readIOPercentHyaCache float DEFAULT NULL,\n" + 
     			   " readIOPercentHyaNonCache float DEFAULT NULL,\n" + 
     			   " PRIMARY KEY (objid, fromtime, totime),\n" + 
     			   " KEY totime_index (totime),\n" + 
     			   " KEY fromtime_totime_index (fromtime, totime)\n" + 
     			   ") ENGINE=InnoDB DEFAULT CHARSET=utf8\n";
     	upgradeRunner.sqlUpdate(qosVolumeWkldSummaryDDL).executeUpdateEx();
    	
    	 }
    	
    	if(clusterIdList.size() > 0){
    		logger.info("Created QosPerformance Sample & Summary Tables");
    	}
    }
}
