package com.netapp.oci.platform.common;

import com.netapp.oci.platform.common.interfaces.data.NestedStackTraceException;

import javax.interceptor.AroundInvoke;
import javax.interceptor.InvocationContext;

import org.slf4j.LoggerFactory;

/**
 * @author Ran Gilboa
 *         Date: Nov 3, 2010
 *         Time: 11:02:38 AM
 */
public class ServerInterfaceInterceptor {
    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(ServerInterfaceInterceptor.class);
    private int slowExecutionThresholdMs = 10000;

    public ServerInterfaceInterceptor() {
        String slowExecutionSeconds = System.getProperty("server.slow_execution_threshold_seconds");
        if (slowExecutionSeconds != null) {
            try {
                slowExecutionThresholdMs = Integer.parseInt(slowExecutionSeconds) * 1000;
            } catch (NumberFormatException e) {
                logger.warn("server.slow_execution_threshold_seconds is not an integer. value is : " + slowExecutionSeconds);
            }
        }
    }

    @AroundInvoke
    public Object logException(InvocationContext context) throws Exception {
        long start = System.currentTimeMillis();
        try {
            return context.proceed();
        } catch (NestedStackTraceException e) {
            logger.warn("Exception happened in method: " + context.getMethod() + " Parameters " + getParameterSampleText(context) + "\n" +
                    e.toString() + "\n" + e.getNestedStackTrace());
            throw e;
        } catch (Exception e) {
            logger.error("Exception (not nested stack trace exception) happened in method: " + context.getMethod() + " Parameters " + getParameterSampleText(context), e);
            throw e;
        } finally {
            long time = System.currentTimeMillis() - start;
            if (time > slowExecutionThresholdMs) {
                logger.warn(" Method: " + context.getMethod() + " with parameters " + getParameterSampleText(context) + " used " + time + " milliseconds");
            }
        }
    }

    private String getParameterSampleText(InvocationContext context) {
        StringBuilder sb = new StringBuilder();
        Object[] parameters = context.getParameters();
        sb.append("{");
        for (int i = 0; i < parameters.length; i++) {
            if (i != 0) {
                sb.append(", ");
            }
            sb.append(toString(parameters[i]));
        }
        sb.append("}");

        return sb.toString();
    }

    private String toString(Object obj) {
        if (obj instanceof Iterable) {
            StringBuilder sb = new StringBuilder();
            Iterable iterable = (Iterable) obj;
            sb.append("[");
            int i = 0;
            for (Object element : iterable) {
                if (i == 4) {
                    sb.append("...");
                    break;
                }
                if (i != 0) {
                    sb.append(", ");
                }
                sb.append(toString(element));
                i++;
            }
            sb.append("]");
            return sb.toString();
        } else {
            return String.valueOf(obj);
        }
    }
}
