/*
 * Copyright (c) 2010 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.common;

import com.onaro.commons.types.Period;
import java.io.Serializable;
import java.util.*;

/**
 * Helper class to understand how big is time span.
 */
public class TimeSpan implements Serializable, Comparable<TimeSpan> {
    private static final long serialVersionUID = -7838279009417985009L;

    /**
     * Start of the time span
     */
    private long fromTime;

    /**
     * End of the time span
     */
    private long toTime;


    /**
     * Copy constructor
     *
     * @param copy - time span to copy
     */
    public TimeSpan(TimeSpan copy) {
        this(copy.fromTime, copy.toTime);
    }

    /**
     * Copy constructor
     *
     * @param fromTime - from time
     * @param toTime   - to time
     */
    public TimeSpan(Date fromTime, Date toTime) {
        this(fromTime.getTime(), toTime.getTime());
    }

    public TimeSpan(long fromTime, long toTime) {
        if (toTime < fromTime) {
            throw new IllegalArgumentException("FromTime (" + fromTime + ") is less then toTime (" + toTime + ") - can't make time span out of it");
        }

        this.fromTime = fromTime;
        this.toTime = toTime;
    }

    public long getFromTime() {
        return fromTime;
    }

    public Date getFromTimeAsDate() {
        return new Date(fromTime);
    }

    public void setFromTime(long fromTime) {
        this.fromTime = fromTime;
    }

    public long getToTime() {
        return toTime;
    }

    public Date getToTimeAsDate() {
        return new Date(toTime);
    }

    public void setToTime(long toTime) {
        this.toTime = toTime;
    }

    /**
     * Checks if time spans overlap
     *
     * @param span the time span to test if overlap this time span
     * @return <code>true</code> if time spans overlap
     *         <code>false</code> otherwise
     */
    public boolean overlaps(TimeSpan span) {
        return overlaps(span.fromTime, span.toTime);
    }

    /**
     * Checks if time spans overlap
     *
     * @param fromTime the begining of the time span to test if overlap this time span
     * @param toTime   the end of the time span to test if overlap this time span
     * @return <code>true</code> if time spans overlap
     *         <code>false</code> otherwise
     */
    public boolean overlaps(long fromTime, long toTime) {
        return fromTime <= this.toTime && toTime >= this.fromTime;
    }

    /**
     * Checks if this time span consumes another timespan
     *
     * @param span the time span to test if it is consumed by this time span
     * @return <code>true</code> if this time span consumes
     *         <code>false</code> otherwise
     */
    public boolean consumes(TimeSpan span) {
        return span.fromTime >= this.fromTime && span.toTime <= this.toTime;
    }

    /**
     * Checks if time belongs to the time span
     *
     * @param time - time to check
     * @return <code>true</code> if time belongs to the span
     *         <code>false</code> otherwise
     */
    public boolean belongs(long time) {
        return fromTime <= time && toTime >= time;
    }

    /**
     * Returns size of the time span in milliseconds
     *
     * @return time span size in milliseconds
     */
    public long size() {
        return toTime - fromTime;
    }

    /**
     * Returns time span - result of overlap of two time spans
     *
     * @param timeSpan - another time span
     * @return time span of overlapping time between time spans,
     *         <code>null</code> if there was no overlapping
     */
    public TimeSpan getOverlap(TimeSpan timeSpan) {
        if (overlaps(timeSpan)) {
            return new TimeSpan(Math.max(this.fromTime, timeSpan.fromTime), Math.min(this.toTime, timeSpan.toTime));
        }

        return null;
    }

    /**
     * Returns join of two adjacent or overlapping time spans
     *
     * @param timeSpan - another time span
     * @return time span that is a join of two time spans,
     *         <code>null</code> if time spans are not adjacent
     */
    public TimeSpan getJoin(TimeSpan timeSpan) {
        TimeSpan result = new TimeSpan(Math.min(fromTime, timeSpan.getFromTime()), Math.max(toTime, timeSpan.getToTime()));

        if (result.size() > size() + timeSpan.size() + 1) {
            return null;
        } else {
            return result;
        }
    }

    /**
     * Substracts time span out of current time span
     *
     * @param timeSpan - time span to substract
     * @return new time span result of the substraction,
     *         <code>null</code> in case the result of substraction
     *         is empty time span
     */
    public TimeSpan substract(TimeSpan timeSpan) {
        if (timeSpan == null) {
            return new TimeSpan(this);
        }

        if (overlaps(timeSpan)) {
            if (belongs(timeSpan.fromTime)) {
                if (belongs(timeSpan.toTime)) {
                    // return earliest time span
                    if (timeSpan.toTime + 1 < this.toTime) {
                        return new TimeSpan(timeSpan.toTime + 1, this.toTime);
                    }
                } else {
                    if (this.fromTime < timeSpan.fromTime - 1) {
                        return new TimeSpan(this.fromTime, timeSpan.fromTime - 1);
                    }
                }
            } else if (belongs(timeSpan.toTime)) {
                if (timeSpan.toTime + 1 < this.toTime) {
                    return new TimeSpan(timeSpan.toTime + 1, this.toTime);
                }
            }
        }

        return null;
    }

    /**
     * Calculates full hour after from and to time of the current timespan are shifted left
     * For example 9:20 - 10:30 becomes 9:00 - 10:00
     *             9:00 - 9:45  becomes 9:00 - 10:00
     *
     * @return new time span object with whole hours only
     */
    public TimeSpan getLeftShiftWholeHoursTimeSpan() {
        GregorianCalendar calendar = new GregorianCalendar();
        if (this.toTime - this.fromTime < Period.HOUR.milliseconds()) {
            calendar.setTimeInMillis(this.toTime + Period.HOUR.milliseconds());
        } else {
            calendar.setTimeInMillis(this.toTime);
        }

        if (calendar.get(Calendar.MILLISECOND) > 0 ||
                calendar.get(Calendar.SECOND) > 0 ||
                calendar.get(Calendar.MINUTE) > 0) {

            calendar.set(Calendar.MILLISECOND, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MINUTE, 0);
        }
        long toTime = calendar.getTimeInMillis();

        calendar.setTimeInMillis(this.fromTime);
        if (calendar.get(Calendar.MILLISECOND) > 0 ||
                calendar.get(Calendar.SECOND) > 0 ||
                calendar.get(Calendar.MINUTE) > 0) {

            calendar.set(Calendar.MILLISECOND, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MINUTE, 0);
        }
        long fromTime = calendar.getTimeInMillis();

        return new TimeSpan(fromTime, toTime);
    }

    /**
     * Calculates the smallest time-span containing this time-span having only whole hours.
     *
     * @return a time-span containign the given one made of whole hours only
     */
    public TimeSpan getWholeHoursTimeSpan() {
        GregorianCalendar calendar = new GregorianCalendar();

        calendar.setTimeInMillis(toTime);
        if (calendar.get(Calendar.MILLISECOND) > 0 ||
                calendar.get(Calendar.SECOND) > 0 ||
                calendar.get(Calendar.MINUTE) > 0) {

            calendar.set(Calendar.MILLISECOND, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MINUTE, 0);

            calendar.add(Calendar.HOUR_OF_DAY, 1);
        }
        long toTime = calendar.getTimeInMillis();

        calendar.setTimeInMillis(fromTime);
        if (calendar.get(Calendar.MILLISECOND) > 0 ||
                calendar.get(Calendar.SECOND) > 0 ||
                calendar.get(Calendar.MINUTE) > 0) {

            calendar.set(Calendar.MILLISECOND, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MINUTE, 0);
        }
        long fromTime = calendar.getTimeInMillis();

        return new TimeSpan(fromTime, toTime);
    }


    /**
     * Splits time span into several time spans that correspond to full calendar hours within
     * original time span.
     *
     * @return list of time spans that correspond to full calendar hours within original time span
     *         sorted by time with latest span being the first
     */
    public List<TimeSpan> splitIntoHours() {
        return splitIntoPeriods(Period.HOUR);
    }

    /**
     * Splits time span into several time spans that correspond to full calendar days within
     * original time span.
     *
     * @return list of time spans that correspond to full calendar days within original time span
     *         sorted by time with latest span being the first
     */
    public List<TimeSpan> splitIntoDays() {
        return splitIntoPeriods(Period.DAY);
    }

    /**
     * Splits time span into several time spans that correspond to full calendar period within
     * original time span.
     *
     * @param period - time period for splitting, for example - hour, day, week, year
     * @return list of time spans that correspond to full calendar period within original time span
     *         sorted by time with latest span being the first
     */
    public List<TimeSpan> splitIntoPeriods(Period period) {
        List<TimeSpan> result = new LinkedList<TimeSpan>();

        // the idea is to start with to time and keep rolling down periods one after another till time reaches From time
        GregorianCalendar calendar = new GregorianCalendar();

        // all time spans are usually ending with 9999s cause that is end time of any calendar period like hour or day or week
        calendar.setTimeInMillis(getToTime() + 1);

        // reset weeks, days, hours, minutes, seconds, milliseconds for from depending on the period
        if (Period.WEEK.equals(period)) {
            calendar.set(Calendar.DAY_OF_WEEK, calendar.getMinimum(Calendar.DAY_OF_WEEK));
            calendar.set(Calendar.HOUR_OF_DAY, 0);
            calendar.set(Calendar.MINUTE, 0);
        }

        if (Period.DAY.equals(period)) {
            calendar.set(Calendar.HOUR_OF_DAY, 0);
            calendar.set(Calendar.MINUTE, 0);
        }

        if (Period.HOUR.equals(period)) {
            calendar.set(Calendar.MINUTE, 0);
        }
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);

        long timeTemp = calendar.getTimeInMillis();
        final int dstSavings = TimeZone.getDefault().getDSTSavings();
        if (isTransitionIntoDST()) {
            timeTemp -= dstSavings;
        } else if (isTransitionOutDST()) {
            timeTemp += dstSavings;
        }



        while (timeTemp - period.milliseconds() > getFromTime()) {
            result.add(new TimeSpan(timeTemp - period.milliseconds(), timeTemp - 1));
            timeTemp -= period.milliseconds();
        }

        if (timeTemp > getFromTime()) {
            result.add(new TimeSpan(timeTemp - period.milliseconds(), timeTemp - 1));
        }

        return result;
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("fromTime=").append(fromTime).append(" (").append(new Date(fromTime)).append(")");
        sb.append("<---->");
        sb.append("toTime=").append(toTime).append(" (").append(new Date(toTime)).append(")");
        return sb.toString();
    }

    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        final TimeSpan timeSpan = (TimeSpan) o;

        return fromTime == timeSpan.fromTime && toTime == timeSpan.toTime;
    }

    public int hashCode() {
        int result;
        result = (int) (fromTime ^ (fromTime >>> 32));
        result = 29 * result + (int) (toTime ^ (toTime >>> 32));
        return result;
    }

    public TimeSpan offset(long offsetMillis) {
        if (offsetMillis == 0) return this;
        else return new TimeSpan(fromTime - offsetMillis, toTime - offsetMillis);
    }

    public int compareTo(TimeSpan o) {
        if (getFromTime() != o.getFromTime())
            return Long.valueOf(getFromTime()).compareTo(Long.valueOf(o.getFromTime()));
        if (getToTime() != o.getToTime())
            return Long.valueOf(getToTime()).compareTo(Long.valueOf(o.getToTime()));
        return 0;//To change body of implemented methods use File | Settings | File Templates.
    }

    public boolean isTransitionIntoDST() {
        TimeZone timeZone = TimeZone.getDefault();
        return !timeZone.inDaylightTime(getFromTimeAsDate()) && timeZone.inDaylightTime(new Date(getToTime() + 1));
    }

    public boolean isTransitionOutDST() {
        TimeZone timeZone = TimeZone.getDefault();
        return timeZone.inDaylightTime(getFromTimeAsDate()) && !timeZone.inDaylightTime(new Date(getToTime() + 1));
    }

    /**
     * when the Client Select the "All" option fromTime is 0
     */
    public boolean isAll(){
        return fromTime == 0;
    }

    public void subtractFromByHourForHourlyTimeSpan() {
        if (size() < 2 * Period.HOUR.milliseconds()) {
            fromTime -= Period.HOUR.milliseconds();
        }
    }

    /**
     * Generates time span object for calendar hour
     *
     * @param time - any point of time that belongs to a calendar hour that will be generated
     * @return time span equal to an hour
     */
    public static TimeSpan generateHourTimeSpan(long time) {
        Calendar calendar = new GregorianCalendar();
        calendar.setTime(new Date(time));
        calendar.set(Calendar.MINUTE, 0);
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);

        long from = calendar.getTime().getTime();

        calendar.set(Calendar.MINUTE, 59);
        calendar.set(Calendar.SECOND, 59);
        calendar.set(Calendar.MILLISECOND, 999);
        long to = calendar.getTime().getTime();

        return new TimeSpan(from, to);
    }
}
