/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.common.param;

import com.netapp.oci.platform.common.table.*;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

/**
 * @author Al Velella
 *         Date: Oct 9, 2012
 *         Time: 5:06:12 PM
 */
public class PluginParam {

    private final String tableName;

    private static final String[] NO_STRINGS = new String[0];

    private Map<String, String> keyToValue = new HashMap<String, String>();

    public static class KeyToValue {
        private String key;
        private String value;

        public KeyToValue(String key, String value) {
            this.key = key;
            this.value = value;
        }

        public String getKey() {
            return key;
        }

        public String getValue() {
            return value;
        }

        @Override
        public String toString() {
            return key + " = " + value;
        }
    }

    public PluginParam(String configDB) {
        this.tableName = configDB + ".param";
    }

    public synchronized void set(String param, int value) throws ParamException {
        set(param, String.valueOf(value));
    }

    public synchronized void set(Connection connection, String param, int value) throws ParamException {
        set(connection, param, String.valueOf(value));
    }

    public synchronized void set(Connection connection, String param, long value) throws ParamException {
        set(connection, param, String.valueOf(value));
    }

    public synchronized void set(String param, long value) throws ParamException {
        set(param, String.valueOf(value));
    }

    public synchronized void set(String param, boolean value) throws ParamException {
        if (value) {
            set(param, 1);
        } else {
            set(param, 0);
        }
    }

    public synchronized void set(Connection connection, String param, boolean value) throws ParamException {
        if (value) {
            set(connection, param, 1);
        } else {
            set(connection, param, 0);
        }
    }

    public synchronized String get(String param, String defaultValue) {
        try {
            return getString(param);
        } catch (ParamException e) {
            return defaultValue;
        }
    }

    public synchronized String getString(String param) throws ParamException {
        return getValue(param);
    }

    public synchronized boolean getBoolean(String param) throws ParamException {
        return getInt(param) == 1;
    }

    public synchronized boolean get(String param, boolean defaultValue) {
        try {
            return getInt(param) == 1;
        } catch (ParamException e) {
            return defaultValue;
        }
    }

    public synchronized int getInt(String param) throws ParamException {
        String value = getString(param);
        try {
            return Integer.parseInt(value);
        } catch (NumberFormatException e) {
            throw new ParamException("Illegal value for parameter " + param + " , value " + value + " in param table (is not int)", e);
        }
    }

    public synchronized float getFloat(String param) throws ParamException {
        String value = getString(param);
        try {
            return Float.parseFloat(value);
        } catch (NumberFormatException e) {
            throw new ParamException("Illegal value for parameter " + param + " , value " + value + " in param table (is not float)", e);
        }
    }

    public synchronized int get(String param, int defaultValue) {
        try {
            return getInt(param);
        } catch (ParamException e) {
            return defaultValue;
        }
    }

    public synchronized float get(String param, float defaultValue) {
        try {
            return getFloat(param);
        } catch (ParamException e) {
            return defaultValue;
        }
    }

    public synchronized long getLong(String param) throws ParamException {
        String value = getString(param);
        try {
            return Long.parseLong(value);
        } catch (NumberFormatException e) {
            throw new ParamException("Illegal value for parameter " + param + " , value " + value + " in param table (is not long)", e);
        }
    }

    public synchronized long get(String param, long defaultValue) {
        try {
            return getLong(param);
        } catch (ParamException e) {
            return defaultValue;
        }
    }

    public synchronized String[] get(String param) {
        Map<String, String> keyToValue = getKeyToValue();
        String sCount = keyToValue.get(param + "_COUNT");
        if (sCount != null) {
            int count = Integer.parseInt(sCount);
            String[] values = new String[count];
            for (int i = 0; i < count; i++) {
                values[i] = keyToValue.get(param + "_" + i);
            }
            return values;
        }
        return NO_STRINGS;
    }

    public synchronized void set(String param, String[] values) throws ParamException {
        Connection connection = null;
        try {
            Map<String, String> keyToValue = getKeyToValue();

            connection = ConnectionHelper.newConnectionNoEjb();

            // delete from db
            String prefix = param + "_";
            deletePrefix(connection, prefix);

            // delete from cache
            for (Iterator<Map.Entry<String, String>> iterator = keyToValue.entrySet().iterator(); iterator.hasNext();) {
                Map.Entry<String, String> entry = iterator.next();
                if (entry.getKey().startsWith(prefix)) {
                    iterator.remove();
                }
            }

            // insert into db and cache
            if (values.length > 0) {
                String sCount = String.valueOf(values.length);
                updateParam(connection, param + "_COUNT", sCount);
                keyToValue.put(param + "_COUNT", sCount);
                for (int i = 0; i < values.length; i++) {
                    updateParam(connection, param + "_" + i, values[i]);
                    keyToValue.put(param + "_" + i, values[i]);
                }
            }
        } catch (Exception e) {
            clearCache();
            throw new ParamException("Failed to set param:" + param + " values:" + (values == null ? "null" : Arrays.asList(values)), e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    public synchronized void set(String param, String value) throws ParamException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            set(connection, param, value);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    public synchronized void set(Connection connection, String param, String value) throws ParamException {
        try {
            // values can not be null
            if (value == null) {
                value = "";
            }

            Map<String, String> keyToValue = getKeyToValue();
            keyToValue.put(param, value);
            updateParam(connection, param, value);
        } catch (Exception e) {
            clearCache();
            throw new ParamException("Failed to set parameter " + param + " with value " + value + " on param table", e);
        }
    }

    private String getValue(String param) throws ParamException {
        String value = getKeyToValue().get(param);
        if (value == null) {
            clearCache();
            throw new ParamException("Failed to get value for param:" + param);
        }
        return value;
    }

    /**
     * Reset the cache of param. We need that, for example, when we swap a database using DbConfig.
     */
    public synchronized void clearCache() {
        keyToValue.clear();
    }

    public synchronized Map<String, String> getKeyToValue() {
        if (keyToValue.isEmpty()) {
            Connection connection = null;
            try {
                keyToValue = new HashMap<String, String>();
                connection = ConnectionHelper.newConnectionNoEjb();
                for (KeyToValue value : findAllParams(connection)) {
                    keyToValue.put(value.getKey(), value.getValue());
                }
                return keyToValue;
            } catch (Exception e) {
                keyToValue.clear();
                throw new IllegalStateException("Could not retrieve key to value map from param table", e);
            } finally {
                ConnectionHelper.close(connection);
            }
        } else {
            return keyToValue;
        }
    }

    private void deletePrefix(Connection connection, String prefix) throws ParamException {
        String sql = "DELETE FROM " + tableName + "WHERE param LIKE \"" + prefix + "%\";";
        try {
            PreparedStatementWrapper.createForUpdate(connection, sql).executeUpdateEx();
        } catch (TableException e) {
            throw new ParamException("Failed to run " + sql, e);
        }
    }

    private void updateParam(Connection connection, String key, String value) throws ParamException {
        try {
            int updatedRows = PreparedStatementWrapper.createForUpdate(connection, "INSERT INTO " + tableName + " VALUES(?, ?) ON DUPLICATE KEY UPDATE value=?")
                .setNext(key)
                .setNext(value)
                .setNext(value)
                .executeUpdateEx();
            if (updatedRows == 0) {
                throw new ParamException("Could not set key, " + key + " does not exist in param table");
            }
        } catch (TableException e) {
            throw new ParamException("Failed to update parameter " + key + " with value " + value + " on param table", e);
        }
    }

    private Collection<KeyToValue> findAllParams(Connection connection) throws TableException {
        return PreparedStatementWrapper.create(connection, "SELECT * FROM " + tableName, new TableResultParserDefault<KeyToValue>(){
            @Override
            public KeyToValue parseObject(ResultSet resultSet) throws SQLException {
                String name = resultSet.getString(1);
                String value = StringCache.get(resultSet.getString(2));
                return new PluginParam.KeyToValue(name, value);
            }
        }).executeQueryEx();
    }
}
