	/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.common.table;

import com.netapp.oci.platform.common.LoginConstant;
import com.netapp.oci.platform.common.util.Assert;
import com.onaro.commons.table.DbConfig;
import org.slf4j.Logger;

import javax.naming.InitialContext;
import javax.sql.DataSource;
import java.sql.*;
import java.util.Properties;
import java.util.Stack;

/**
 * Light-weight context object. It provides access to {@link Connection} object on demand.
 * <p/>
 * To use,
 * <code>
 * Connection connection = null;
 * try {
 * connection = ConnectionHelper.newConnection(dataSource);
 * ...
 * } finally {
 * ConnectionHelper.close(connection);
 * }
 * </code>
 * <p/>
 * This class can detect Connection leaks if the above recommended coding pattern is missed. If the log4j DEBUG level is
 * turned on for its package, it can report the exact location where this type of leak occurs.
 * <p/>
 * This class is NOT multithread safe.
 *
 * @author Dongjun Sun
 */

public abstract class ConnectionHelper {
    private static final Logger logger =org.slf4j.LoggerFactory.getLogger(ConnectionHelper.class);
    private static final int THRESHOLD = 4;
    private static final boolean isDebugEnabled = Boolean.parseBoolean(System.getProperty("com.onaro.sanscreen.util.table.connection.helper", "false"));

    private Connection connection;
    private boolean closed = false;
    private static ThreadLocal<Object> referenceCount = new ThreadLocal<Object>();

    /**
     * Cosntructs a new {@link Connection} object that is wrapper around connection helper.
     * This factory method should be used in places where container
     * environment is not peresent. It checkes the system proeprty "no.jndi".
     * If set to true, a {@link Connection} object will be constructed when requested from the JDBC setting. Otherwise
     * it is supposed to be running within a J2EE container and uses JNDI to lookup the datasource interface.
     *
     * @return the newly constructed {@link ConnectionHelper} object.
     */
    public static Connection newConnectionNoEjb() {
        checkReferenceCount();
        boolean useJNDI = "false".equalsIgnoreCase(System.getProperty("no.jndi", "false"));
        ConnectionHelper connectionHelper;
        if (useJNDI) {
            connectionHelper = newJndiContext(null);
        } else {
            connectionHelper = newJdbcContext();
        }

        return new PrefixConnection(connectionHelper);
    }

    /**
     * Creates lazy instantiated connection from datasource
     *
     * @param dataSource the {@link DataSource} object
     * @return the newly constructed {@link Connection} object
     */
    public static Connection newConnection(DataSource dataSource) {
        checkReferenceCount();
        Assert.isNotNull(dataSource, "DataSource cannot be null");

        return new PrefixConnection(newJndiContext(dataSource));
    }

    @SuppressWarnings("unchecked")
    private static void checkReferenceCount() {
        if (isDebugEnabled) {
            boolean leaked = false;
            String errMessage = null;
            if (referenceCount.get() != null) {
                Stack<StackTraceElement[]> stack = (Stack<StackTraceElement[]>) referenceCount.get();
                if (stack.size() >= THRESHOLD) {
                    errMessage = formatErrorMessage(stack);
                    leaked = true;
                }
            }

            if (leaked) {
                logger.error(errMessage);
                referenceCount.remove();
                throw new IllegalStateException(errMessage);
            }
        }
    }

    @SuppressWarnings("unchecked")
    private static void incReferenceCount() {
        if (isDebugEnabled) {
            Stack<StackTraceElement[]> stack = (Stack<StackTraceElement[]>) referenceCount.get();
            if (stack == null) {
                stack = new Stack<StackTraceElement[]>();
            }
            stack.push(getStackTrace());
            referenceCount.set(stack);
        }
    }

    @SuppressWarnings("unchecked")
    private static void decReferenceCount() {
        if (isDebugEnabled) {
            Stack<StackTraceElement[]> stack = (Stack<StackTraceElement[]>) referenceCount.get();
            if (stack != null) {
                stack.pop();
                if (stack.isEmpty()) {
                    referenceCount.remove();
                }
            }
        }
    }

    private static ConnectionHelper newJndiContext(DataSource ds) {
        if (ds != null) {
            return new JndiConnectionHelper(ds);
        }

        try {
            ds = (DataSource) new InitialContext().lookup(JndiConnectionHelper.DATASOUCE_JNDI_NAME);
            return new JndiConnectionHelper(ds);
        } catch (Exception e) {
            throw new IllegalStateException("Failed to create ConnectionHelper", e);
        }
    }

    private static ConnectionHelper newJdbcContext() {
        try {
            final String JDBC_DRIVER = System.getProperty("JDBC_DRIVER", "org.mariadb.jdbc.Driver");
            Class.forName(JDBC_DRIVER);
        } catch (ClassNotFoundException e) {
            throw new IllegalStateException(e);
        }

        final String jdbcHost = System.getProperty("JDBC_HOST", DbConfig.getMySqlHost());
        final String jdbcDbName = System.getProperty("JDBC_DB_NAME", DbConfig.getSanscreenDbName());
        final String jdbcUrl = System.getProperty("JDBC_URL",
                "jdbc:mysql://" + jdbcHost + "/" + jdbcDbName + "?user=root&amp;password=");
        final String jdbcUser = System.getProperty("JDBC_USER", LoginConstant.DB_USER);
        final String jdbcPassword = System.getProperty("JDBC_PASSWORD", LoginConstant.DB_PASSWORD);

        Properties jdbcProps = new Properties();
        jdbcProps.put("user", jdbcUser);
        jdbcProps.put("password", jdbcPassword);
        jdbcProps.put("cachePrepStmts", "true");

        return new JdbcConnectionHelper(jdbcUrl, jdbcProps);
    }

    protected abstract Connection doGetConnection() throws SQLException;

    /**
     * This method allocates database connection, it is being called from invocation handler
     *
     * @return connection from connection helper
     * @throws SQLException if failed to allocate connection
     */
    Connection getConnection() throws SQLException {
        if (closed) {
            throw new IllegalStateException("ConnectionHelper has already been closed!");
        }

        if (connection == null) {
            connection = doGetConnection();
            if (isDebugEnabled) {
                logger.debug("open Connection(" + connection.hashCode() + "): " + getStackTraceInfo());
            }
        }

        return connection;
    }

    private static String formatErrorMessage(Stack<StackTraceElement[]> stack) {
        StringBuilder builder = new StringBuilder();

        builder.append("Call stack exceeds threashold, most recent calling stacking follows\n");
        for (StackTraceElement[] stackTraceElements : stack) {
            builder.append("########## Start of the calling stack ##########\n");
            for (StackTraceElement stackTraceElement : stackTraceElements) {
                builder.append("\t");
                builder.append(stackTraceElement.toString());
                builder.append("\n");
            }
            builder.append("########## End of the calling stack ##########\n");
            builder.append("\n");
        }

        return builder.toString();
    }

    private static StackTraceElement[] getStackTrace() {
        return Thread.currentThread().getStackTrace();
    }

    private static String getStackTraceInfo() {
        StackTraceElement[] stackTraceElements = Thread.currentThread().getStackTrace();
        for (StackTraceElement se : stackTraceElements) {
            String className = se.getClassName();
            if (!className.startsWith("java.") &&
                    !className.contains("table.ConnectionHelper") &&
                    !className.contains("server.TransactionContext") &&
                    !className.contains("$Proxy")) {
                return className + "(" + se.getLineNumber() + ")";
            }
        }

        return ConnectionHelper.class.getSimpleName();
    }

    /**
     * Calling the method <code>close</code> on a <code>ConnectionHelper</code> object that is already closed is a no-op.
     */
    void close() {
        this.connection = null;
        this.closed = true;
        decReferenceCount();
    }

    /**
     * Helper method to close database connection properly wethere it is null or not
     *
     * @param connection - database connection
     */
    public static void close(Connection connection) {
        if (connection != null) {
            try {
                int hash = connection.hashCode();
                connection.close();
                if (isDebugEnabled) {
                    logger.debug("close Connection(" + hash + "): " + getStackTraceInfo());
                }
            } catch (SQLException e) {
                logger.error("Failed to close connection", e);
            }
        }
    }

    /**
     * Helper method to close database result set properly wethere it is null or not
     *
     * @param resultSet - database resultSet
     */
    public static void close(ResultSet resultSet) {
        if (resultSet != null) {
            try {
                resultSet.close();
            } catch (SQLException e) {
                logger.error("Failed to close result set", e);
            }
        }
    }

    /**
     * Helper method to close database statement properly wethere it is null or not
     *
     * @param statement - database statement
     */
    public static void close(Statement statement) {
        if (statement != null) {
            try {
                statement.close();
            } catch (SQLException e) {
                logger.error("Failed to close statement", e);
            }
        }
    }

    static class JndiConnectionHelper extends ConnectionHelper {
        static final String DATASOUCE_JNDI_NAME = "java:/onaro_pool";

        private DataSource ds;

        JndiConnectionHelper(DataSource ds) {
            this.ds = ds;
            incReferenceCount();
        }

        protected Connection doGetConnection() throws SQLException {
            return ds.getConnection();
        }
    }

    static class JdbcConnectionHelper extends ConnectionHelper {
        private String jdbcUrl;
        private Properties jdbcProps;

        JdbcConnectionHelper(String jdbcUrl, Properties jdbcProps) {
            this.jdbcUrl = jdbcUrl;
            this.jdbcProps = jdbcProps;
            incReferenceCount();
        }

        protected Connection doGetConnection() throws SQLException {
            return DriverManager.getConnection(jdbcUrl, jdbcProps);
        }
    }
}
