/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.common.table;

import org.slf4j.Logger;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.net.URL;
import java.sql.*;
import java.sql.Date;
import java.util.*;

/**
 * This wrapper around PreparedStatement extends the interface with convenient functions:
 * index-less setNext for every type of parameter and function to execute a query.
 * <p/>
 * The way to use the class:
 * preparedStatementWrap = table.getPrepareStatement(sql)
 * preparedStatementWrap.setNext("s").setNext(1).setNext((long)11111111);
 * preparedStatementWrap.executeQueryEx();
 * <p/>
 * Or shorter (assume you are in a table subclass)
 * getPrepareStatement(sql).setNext("s").setNext(1).setNext((long)11111111).executeQueryEx();
 * <p/>
 * T - generic type which corresponds to the Table class parameter.
 * <p/>
 * Date: Apr 10, 2007
 * Time: 10:14:18 AM
 *
 * @author yarom.gabay
 */
public class PreparedStatementWrapper<T> implements PreparedStatement {

    static transient final Logger logger =org.slf4j.LoggerFactory.getLogger(PreparedStatementWrapper.class);
    private PreparedStatement preparedStatement;
    private TableResultParser<T> tableResultParser;
    private String sql;  // this is needed for logging purposes only
    private TableAutoGeneratedKeyParser tableAutoGeneratedKeyParser;
    private int insertIndex = 1;
    private Queue<Object> sqlParameters = new LinkedList<Object>(); // we keep this so we can print the sql nicely. 

    //////////////////////////////////////////////////////////////////////
    // Constructor and factories
    //////////////////////////////////////////////////////////////////////

    public PreparedStatementWrapper(PreparedStatement preparedStatement, TableResultParser<T> tableResultParser, String sql, TableAutoGeneratedKeyParser tableAutoGeneratedKeyParser) {
        this.preparedStatement = preparedStatement;
        this.tableResultParser = tableResultParser;
        this.sql = sql;
        this.tableAutoGeneratedKeyParser = tableAutoGeneratedKeyParser;

        if (sql.toLowerCase().contains("myisam")) {
            logger.warn("SQL contains myisam keyword: " + sql);
        }
    }

    public static <T> PreparedStatementWrapper<T> createForUpdate(Connection connection, String sql) throws TableException {
        PreparedStatement preparedStatement = null;
        try {
            preparedStatement = connection.prepareStatement(sql, Statement.RETURN_GENERATED_KEYS);
            preparedStatement.setFetchSize(Integer.MIN_VALUE);
            return new PreparedStatementWrapper<T>(preparedStatement, null, sql, null);
        } catch (SQLException e) {
            close(preparedStatement, null);
            throw new TableException("Failed to prepareStatement for " + PrefixConnection.translateSql(sql), e);
        }
    }

    public static <T> PreparedStatementWrapper<T> create(Connection connection, String sql, TableResultParser<T> resultParser) throws TableException {
        PreparedStatement preparedStatement = null;
        try {
            preparedStatement = connection.prepareStatement(sql, Statement.RETURN_GENERATED_KEYS);
            preparedStatement.setFetchSize(Integer.MIN_VALUE);
            return new PreparedStatementWrapper<T>(preparedStatement, resultParser, sql, null);
        } catch (SQLException e) {
            close(preparedStatement, null);
            throw new TableException("Failed to prepareStatement for " + PrefixConnection.translateSql(sql), e);
        }
    }

    public static <T> PreparedStatementWrapper<T> create(Connection connection, String sql, TableResultParser<T> resultParser, TableAutoGeneratedKeyParser tableAutoGeneratedKeyParser) throws TableException {
        PreparedStatement preparedStatement = null;
        try {
            preparedStatement = connection.prepareStatement(sql, Statement.RETURN_GENERATED_KEYS);
            preparedStatement.setFetchSize(Integer.MIN_VALUE);
            return new PreparedStatementWrapper<T>(preparedStatement, resultParser, sql, tableAutoGeneratedKeyParser);
        } catch (SQLException e) {
            close(preparedStatement, null);
            throw new TableException("Failed to prepareStatement for " + PrefixConnection.translateSql(sql), e);
        }
    }

    public static <T> long executeFunctionQuery(Connection connection, String sql, long defaultValue) throws TableException {
        PreparedStatement preparedStatement = null;
        try {
            preparedStatement = connection.prepareStatement(sql, Statement.RETURN_GENERATED_KEYS);
            preparedStatement.setFetchSize(Integer.MIN_VALUE);
            return new PreparedStatementWrapper<T>(preparedStatement, null, sql, null).executeFunctionQueryEx(defaultValue);
        } catch (SQLException e) {
            close(preparedStatement, null);
            throw new TableException("Failed to prepareStatement for " + PrefixConnection.translateSql(sql), e);
        }
    }

    //////////////////////////////////////////////////////////////////////
    // Non interface Methods (add other types as needed)
    // Note: These methods throw TableException so it's more convenient to use in the tables.
    //    They also take care of closing the preparedStatement in the case of an exception.
    //////////////////////////////////////////////////////////////////////

    public PreparedStatementWrapper<T> setNext(String value) throws TableException {
        try {
            preparedStatement.setString(insertIndex++, value);
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }
    public PreparedStatementWrapper<T> setNext(BigDecimal value) throws TableException {
        try {
            preparedStatement.setBigDecimal(insertIndex++, value);
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(byte value[]) throws TableException {
        try {
            preparedStatement.setBytes(insertIndex++, value);
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + Arrays.toString(value) + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(Date value) throws TableException {
        try {
            preparedStatement.setDate(insertIndex++, value);
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(java.util.Date value) throws TableException {
        try {
            preparedStatement.setDate(insertIndex++, new Date(value.getTime()));
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(Timestamp value) throws TableException {
        try {
            if (value == null)
                return setNextNull(Types.TIMESTAMP);

            preparedStatement.setTimestamp(insertIndex++, value);
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }


    public PreparedStatementWrapper<T> setNext(float value) throws TableException {
        try {
            preparedStatement.setFloat(insertIndex++, value);
            sqlParameters.add(Float.valueOf(value));
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(Float value) throws TableException {
        try {
            if (value == null)
                return setNextNull(Types.FLOAT);

            preparedStatement.setFloat(insertIndex++, value.floatValue());
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(double value) throws TableException {
        try {
            preparedStatement.setDouble(insertIndex++, value);
            sqlParameters.add(Double.valueOf(value));
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(Double value) throws TableException {
        try {
            if (value == null)
                return setNextNull(Types.DOUBLE);

            preparedStatement.setDouble(insertIndex++, value.doubleValue());
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNextNull(int sqlType) throws TableException {
        try {
            preparedStatement.setNull(insertIndex++, sqlType);
            sqlParameters.add(null);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set null for type " + sqlType + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(Integer value) throws TableException {
        try {
            if (value == null)
                return setNextNull(Types.INTEGER);

            preparedStatement.setLong(insertIndex++, value.intValue());
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(int value) throws TableException {
        try {
            preparedStatement.setInt(insertIndex++, value);
            sqlParameters.add(Integer.valueOf(value));
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(long value) throws TableException {
        try {
            preparedStatement.setLong(insertIndex++, value);
            sqlParameters.add(Long.valueOf(value));
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(Long value) throws TableException {
        try {
            if (value == null)
                return setNextNull(Types.BIGINT);

            preparedStatement.setLong(insertIndex++, value.longValue());
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(boolean value) throws TableException {
        try {
            preparedStatement.setBoolean(insertIndex++, value);
            sqlParameters.add(Boolean.valueOf(value));
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    public PreparedStatementWrapper<T> setNext(Boolean value) throws TableException {
        try {
            if (value == null)
                return setNextNull(Types.BOOLEAN);

            preparedStatement.setBoolean(insertIndex++, value.booleanValue());
            sqlParameters.add(value);
            return this;
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("failed to set param=" + value + " - " + this, e);
        }
    }

    /*
     The functions: executeUpdateEx, executeQueryEx and executeFunctionQueryEx
     takes care of exception handling and closing preparedStatement and resultSet for the user.
    */

    /**
     * Executes an update query
     *
     * @throws TableException
     */
    
    public int executeUpdateEx() throws TableException {
        try {
            int result = preparedStatement.executeUpdate();
            resetParameters();
            parseAutoGeneratedKey();
            return result;
        } catch (Exception e) {
            // burt1136354 intercept and rethrow the exception to avoid the next catch block below
            // We can't use instanceof because we should be connector and framework agnostic
            if (e.getCause().toString().contains("java.sql.SQLIntegrityConstraintViolationException: Cannot delete or update a parent row: a foreign key constraint fails")) {
                throw new TableException("Foreign key constraint: " + this, e);
            }
            if (e.getCause() instanceof SQLException) {
                SQLException sqlEx = (SQLException) e.getCause();
                //1507: Error in list of partitions (this error means that the partition was already dropped.)
                //1517: Duplicate partition names (this error means that the partition was already created.)
                //When these two errors seen, no need to throw exception and stop the insert action.
                if (sqlEx.getErrorCode() == 1507 && sqlEx.getSQLState().equalsIgnoreCase("HY000") ||
                        sqlEx.getErrorCode() == 1517 && sqlEx.getSQLState().equalsIgnoreCase("HY000")) {
                    logger.info(" Ignore the exception: " + e);
                    return 0;
                }
            }

            throw new TableException("Failed to execUpdate: " + this, e);

        }
        finally {
            closePreparedStatement(preparedStatement);
        }
    }

    /**
     * Executes a query.
     *
     * @return return list of row objects defined by the particular table (in parseResult).
     * @throws TableException
     */
    public List<T> executeQueryEx() throws TableException {
        ResultSet resultSet = null;
        try {
            resultSet = preparedStatement.executeQuery();
            resetParameters();
            return tableResultParser.parseResult(resultSet);
        } catch (Exception e) {
            throw new TableException("Failed to execQuery: " + this, e);
        } finally {
            closeResultSet(resultSet);
            closePreparedStatement(preparedStatement);
        }
    }

    /**
     * Executes a query.
     *
     * @return return a single result.
     * @throws TableException if not a single result found in the resultSet
     */
    public T executeQuerySingleResult() throws TableException {
        List<T> result = executeQueryEx();
        if (result.size() != 1) {
            throw new TableException("Found " + result.size() + " rows instead of a single one");
        }

        return result.get(0);
    }

    /**
     * Executes a query whose result is a single long value (SUM, COUNT, etc.).
     *
     * @return result of the query
     * @throws TableException
     */
    public long executeFunctionQueryEx() throws TableException {
        ResultSet resultSet = null;
        try {
            resultSet = preparedStatement.executeQuery();
            resetParameters();
            return parseFunctionResult(resultSet, 0);
        } catch (Exception e) {
            throw new TableException("Failed to execFunctionQuery: " + this, e);
        } finally {
            closeResultSet(resultSet);
            closePreparedStatement(preparedStatement);
        }
    }

    /**
     * Executes a query whose result is a sigle long value (SUM, COUNT, etc.).
     *
     * @return result of the query
     * @throws TableException
     */
    public long executeFunctionQueryEx(long defaultValue) throws TableException {
        ResultSet resultSet = null;
        try {
            resultSet = preparedStatement.executeQuery();
            resetParameters();
            return parseFunctionResult(resultSet, defaultValue);
        } catch (Exception e) {
            throw new TableException("Failed to execFunctionQuery: " + this, e);
        } finally {
            closeResultSet(resultSet);
            closePreparedStatement(preparedStatement);
        }
    }

    private long parseFunctionResult(ResultSet resultSet, long defaultValue) throws SQLException {
        if (resultSet.next()) {
            long result = resultSet.getLong(1);
            if (!resultSet.wasNull()) {
                return result;
            }
        }

        return defaultValue;
    }


    public int[] executeBatchEx() throws TableException {
        try {
            int[] result = preparedStatement.executeBatch();
            resetParameters();
            return result;
        } catch (Exception e) {
            throw new TableException("Failed to executeBatchEx:" + this, e);
        } finally {
            closePreparedStatement(preparedStatement);
        }
    }

    public void addBatchEx() throws TableException {
        try {
            preparedStatement.addBatch();
            resetParameters();
        } catch (Exception e) {
            closePreparedStatement(preparedStatement);
            throw new TableException("Failed to addBatch:" + this, e);
        }
    }

    /**
     * We need this function for in order to log a readable message.
     *
     * @return the raw sql query (with question marks) of the preparedStatement.
     */
    public String getSql() {
        return sql;
    }
    
    /**
     * Close preparedStatement;
     */
    public void closePreparedStatement() {
        try {
            if (preparedStatement != null)
                preparedStatement.close();
        } catch (Exception e) {
            logger.error("Failed to close preparedStatement for " + this, e);
        }
    }

    /**
     * Takes care of closing preparedStatement;
     */
    private void closePreparedStatement(PreparedStatement preparedStatement) {
        try {
            if (preparedStatement != null)
                preparedStatement.close();
        } catch (Exception e) {
            logger.error("Failed to close preparedStatement for " + this, e);
        }
    }

    /**
     * Takes care of closing resultSet;
     */
    private void closeResultSet(ResultSet resultSet) {
        try {
            if (resultSet != null)
                resultSet.close();
        } catch (Exception e) {
            logger.error("Failed to close resultSet", e);
        }
    }

    private void resetParameters() {
        insertIndex = 1;
        sqlParameters.clear();
    }

    private void parseAutoGeneratedKey() throws SQLException {
        if (tableAutoGeneratedKeyParser != null) {
            ResultSet generatedKeys = preparedStatement.getGeneratedKeys();
            tableAutoGeneratedKeyParser.parseKey(generatedKeys);
        }
    }

    //////////////////////////////////////////////////////////////////////
    // Forwarding Object Methods
    //////////////////////////////////////////////////////////////////////

    public int hashCode() {
        return preparedStatement.hashCode();
    }

    public boolean equals(Object obj) {
        return this == obj || !(obj == null || getClass() != obj.getClass()) && preparedStatement.equals(obj);
    }

    /**
     * Uses {@code sqlParameters} to replace the ? with real values.
     *
     * @return string representing the sql
     */
    public String toString() {
        StringBuilder buf = new StringBuilder();
        Queue<Object> parameters = new LinkedList<Object>(this.sqlParameters);
        String[] tokens = getSql().split("\\?");

        for (int i = 0; i < tokens.length; i++) {
            buf.append(tokens[i]);
            if ((i < tokens.length - 1) || (i == tokens.length - 1 && getSql().endsWith("?"))) {
                if (!parameters.isEmpty()) {
                    Object obj = parameters.remove();
                    if (obj instanceof String) {
                        String s = (String) obj;
                        if (s.contains("\"")) {
                            buf.append('\'').append(s).append('\'');
                        } else {
                            buf.append('"').append(s).append('"');
                        }
                    } else {
                        buf.append(obj);
                    }
                } else {
                    buf.append("?");
                }
            }
        }
        return buf.toString();
    }

    //////////////////////////////////////////////////////////////////////
    // Forwarding Interface Methods (from PreparedStatement and Statement)
    //////////////////////////////////////////////////////////////////////

    public ResultSet executeQuery() throws SQLException {
        ResultSet resultSet = preparedStatement.executeQuery();
        resetParameters();
        return resultSet;
    }

    public int executeUpdate() throws SQLException {
        int result = preparedStatement.executeUpdate();
        resetParameters();
        return result;
    }

    public boolean execute() throws SQLException {
        boolean result = preparedStatement.execute();
        resetParameters();
        return result;
    }

    public ResultSet executeQuery(String sql) throws SQLException {
        ResultSet resultSet = preparedStatement.executeQuery(sql);
        resetParameters();
        return resultSet;
    }

    public int executeUpdate(String sql) throws SQLException {
        int result = preparedStatement.executeUpdate(sql);
        resetParameters();
        return result;
    }

    public boolean execute(String sql) throws SQLException {
        boolean result = preparedStatement.execute(sql);
        resetParameters();
        return result;
    }

    public int executeUpdate(String sql, int autoGeneratedKeys) throws SQLException {
        int result = preparedStatement.executeUpdate(sql, autoGeneratedKeys);
        resetParameters();
        return result;
    }

    public int executeUpdate(String sql, int columnIndexes[]) throws SQLException {
        int result = preparedStatement.executeUpdate(sql, columnIndexes);
        resetParameters();
        return result;
    }

    public int executeUpdate(String sql, String columnNames[]) throws SQLException {
        int result = preparedStatement.executeUpdate(sql, columnNames);
        resetParameters();
        return result;
    }

    public boolean execute(String sql, int autoGeneratedKeys) throws SQLException {
        boolean result = preparedStatement.execute(sql, autoGeneratedKeys);
        resetParameters();
        return result;
    }

    public boolean execute(String sql, int columnIndexes[]) throws SQLException {
        boolean result = preparedStatement.execute(sql, columnIndexes);
        resetParameters();
        return result;
    }

    public boolean execute(String sql, String columnNames[]) throws SQLException {
        boolean result = preparedStatement.execute(sql, columnNames);
        resetParameters();
        return result;
    }

    public int[] executeBatch() throws SQLException {
        int[] result = preparedStatement.executeBatch();
        resetParameters();
        return result;
    }

    public void addBatch() throws SQLException {
        preparedStatement.addBatch();
        resetParameters();
    }

    public void clearBatch() throws SQLException {
        preparedStatement.clearBatch();
        resetParameters();
    }

    public void clearParameters() throws SQLException {
        preparedStatement.clearParameters();
        resetParameters();
    }

    public void setNull(int parameterIndex, int sqlType) throws SQLException {
        preparedStatement.setNull(parameterIndex, sqlType);
    }

    public void setBoolean(int parameterIndex, boolean x) throws SQLException {
        preparedStatement.setBoolean(parameterIndex, x);
    }

    public void setByte(int parameterIndex, byte x) throws SQLException {
        preparedStatement.setByte(parameterIndex, x);
    }

    public void setShort(int parameterIndex, short x) throws SQLException {
        preparedStatement.setShort(parameterIndex, x);
    }

    public void setInt(int parameterIndex, int x) throws SQLException {
        preparedStatement.setInt(parameterIndex, x);
    }

    public void setLong(int parameterIndex, long x) throws SQLException {
        preparedStatement.setLong(parameterIndex, x);
    }

    public void setFloat(int parameterIndex, float x) throws SQLException {
        preparedStatement.setFloat(parameterIndex, x);
    }

    public void setDouble(int parameterIndex, double x) throws SQLException {
        preparedStatement.setDouble(parameterIndex, x);
    }

    public void setBigDecimal(int parameterIndex, BigDecimal x) throws SQLException {
        preparedStatement.setBigDecimal(parameterIndex, x);
    }

    public void setString(int parameterIndex, String x) throws SQLException {
        preparedStatement.setString(parameterIndex, x);
    }

    public void setBytes(int parameterIndex, byte x[]) throws SQLException {
        preparedStatement.setBytes(parameterIndex, x);
    }

    public void setDate(int parameterIndex, Date x) throws SQLException {
        preparedStatement.setDate(parameterIndex, x);
    }

    public void setTime(int parameterIndex, Time x) throws SQLException {
        preparedStatement.setTime(parameterIndex, x);
    }

    public void setTimestamp(int parameterIndex, Timestamp x) throws SQLException {
        preparedStatement.setTimestamp(parameterIndex, x);
    }

    public void setAsciiStream(int parameterIndex, InputStream x, int length) throws SQLException {
        preparedStatement.setAsciiStream(parameterIndex, x, length);
    }

    @SuppressWarnings({"deprecation"})
    @Deprecated
    public void setUnicodeStream(int parameterIndex, InputStream x, int length) throws SQLException {
        preparedStatement.setUnicodeStream(parameterIndex, x, length);
    }

    public void setBinaryStream(int parameterIndex, InputStream x, int length) throws SQLException {
        preparedStatement.setBinaryStream(parameterIndex, x, length);
    }

    public void setObject(int parameterIndex, Object x, int targetSqlType, int scale) throws SQLException {
        preparedStatement.setObject(parameterIndex, x, targetSqlType, scale);
    }

    public void setObject(int parameterIndex, Object x, int targetSqlType) throws SQLException {
        preparedStatement.setObject(parameterIndex, x, targetSqlType);
    }

    public void setObject(int parameterIndex, Object x) throws SQLException {
        preparedStatement.setObject(parameterIndex, x);
    }


    public void setCharacterStream(int parameterIndex, Reader reader, int length) throws SQLException {
        preparedStatement.setCharacterStream(parameterIndex, reader, length);
    }

    public void setRef(int i, Ref x) throws SQLException {
        preparedStatement.setRef(i, x);
    }

    public void setBlob(int i, Blob x) throws SQLException {
        preparedStatement.setBlob(i, x);
    }

    public void setClob(int i, Clob x) throws SQLException {
        preparedStatement.setClob(i, x);
    }

    public void setArray(int i, Array x) throws SQLException {
        preparedStatement.setArray(i, x);
    }

    public ResultSetMetaData getMetaData() throws SQLException {
        return preparedStatement.getMetaData();
    }

    public void setDate(int parameterIndex, Date x, Calendar cal) throws SQLException {
        preparedStatement.setDate(parameterIndex, x, cal);
    }

    public void setTime(int parameterIndex, Time x, Calendar cal) throws SQLException {
        preparedStatement.setTime(parameterIndex, x, cal);
    }

    public void setTimestamp(int parameterIndex, Timestamp x, Calendar cal) throws SQLException {
        preparedStatement.setTimestamp(parameterIndex, x, cal);
    }

    public void setNull(int paramIndex, int sqlType, String typeName) throws SQLException {
        preparedStatement.setNull(paramIndex, sqlType, typeName);
    }

    public void setURL(int parameterIndex, URL x) throws SQLException {
        preparedStatement.setURL(parameterIndex, x);
    }

    public ParameterMetaData getParameterMetaData() throws SQLException {
        return preparedStatement.getParameterMetaData();
    }


    public void close() throws SQLException {
        preparedStatement.close();
    }

    public int getMaxFieldSize() throws SQLException {
        return preparedStatement.getMaxFieldSize();
    }

    public void setMaxFieldSize(int max) throws SQLException {
        preparedStatement.setMaxFieldSize(max);
    }

    public int getMaxRows() throws SQLException {
        return preparedStatement.getMaxRows();
    }

    public void setMaxRows(int max) throws SQLException {
        preparedStatement.setMaxRows(max);
    }

    public void setEscapeProcessing(boolean enable) throws SQLException {
        preparedStatement.setEscapeProcessing(enable);
    }

    public int getQueryTimeout() throws SQLException {
        return preparedStatement.getQueryTimeout();
    }

    public void setQueryTimeout(int seconds) throws SQLException {
        preparedStatement.setQueryTimeout(seconds);
    }

    public void cancel() throws SQLException {
        preparedStatement.cancel();
    }

    public SQLWarning getWarnings() throws SQLException {
        return preparedStatement.getWarnings();
    }

    public void clearWarnings() throws SQLException {
        preparedStatement.clearWarnings();
    }

    public void setCursorName(String name) throws SQLException {
        preparedStatement.setCursorName(name);
    }


    public ResultSet getResultSet() throws SQLException {
        return preparedStatement.getResultSet();
    }

    public int getUpdateCount() throws SQLException {
        return preparedStatement.getUpdateCount();
    }

    public boolean getMoreResults() throws SQLException {
        return preparedStatement.getMoreResults();
    }

    public void setFetchDirection(int direction) throws SQLException {
        preparedStatement.setFetchDirection(direction);
    }

    public int getFetchDirection() throws SQLException {
        return preparedStatement.getFetchDirection();
    }

    public void setFetchSize(int rows) throws SQLException {
        preparedStatement.setFetchSize(rows);
    }

    public int getFetchSize() throws SQLException {
        return preparedStatement.getFetchSize();
    }

    public int getResultSetConcurrency() throws SQLException {
        return preparedStatement.getResultSetConcurrency();
    }

    public int getResultSetType() throws SQLException {
        return preparedStatement.getResultSetType();
    }

    public void addBatch(String sql) throws SQLException {
        preparedStatement.addBatch(sql);
    }

    public Connection getConnection() throws SQLException {
        return preparedStatement.getConnection();
    }

    public boolean getMoreResults(int current) throws SQLException {
        return preparedStatement.getMoreResults(current);
    }

    public ResultSet getGeneratedKeys() throws SQLException {
        return preparedStatement.getGeneratedKeys();
    }

    public int getResultSetHoldability() throws SQLException {
        return preparedStatement.getResultSetHoldability();
    }


    ///////////////////////////

    public void setRowId(int parameterIndex, RowId x) throws SQLException {
        preparedStatement.setRowId(parameterIndex, x);
    }

    public void setNString(int parameterIndex, String value) throws SQLException {
        preparedStatement.setNString(parameterIndex, value);
    }

    public void setNCharacterStream(int parameterIndex, Reader value, long length) throws SQLException {
        preparedStatement.setNCharacterStream(parameterIndex, value, length);
    }

    public void setNClob(int parameterIndex, NClob value) throws SQLException {
        preparedStatement.setNClob(parameterIndex, value);
    }

    public void setClob(int parameterIndex, Reader reader, long length) throws SQLException {
        preparedStatement.setClob(parameterIndex, reader, length);
    }

    public void setBlob(int parameterIndex, InputStream inputStream, long length) throws SQLException {
        preparedStatement.setBlob(parameterIndex, inputStream, length);
    }

    public void setNClob(int parameterIndex, Reader reader, long length) throws SQLException {
        preparedStatement.setNClob(parameterIndex, reader, length);
    }

    public void setSQLXML(int parameterIndex, SQLXML xmlObject) throws SQLException {
        preparedStatement.setSQLXML(parameterIndex, xmlObject);
    }

    public void setAsciiStream(int parameterIndex, InputStream x, long length) throws SQLException {
        preparedStatement.setAsciiStream(parameterIndex, x, length);
    }

    public void setBinaryStream(int parameterIndex, InputStream x, long length) throws SQLException {
        preparedStatement.setBinaryStream(parameterIndex, x, length);
    }

    public void setCharacterStream(int parameterIndex, Reader reader, long length) throws SQLException {
        preparedStatement.setCharacterStream(parameterIndex, reader, length);
    }

    public void setAsciiStream(int parameterIndex, InputStream x) throws SQLException {
        preparedStatement.setAsciiStream(parameterIndex, x);
    }

    public void setBinaryStream(int parameterIndex, InputStream x) throws SQLException {
        preparedStatement.setBinaryStream(parameterIndex, x);
    }

    public void setCharacterStream(int parameterIndex, Reader reader) throws SQLException {
        preparedStatement.setCharacterStream(parameterIndex, reader);
    }

    public void setNCharacterStream(int parameterIndex, Reader value) throws SQLException {
        preparedStatement.setNCharacterStream(parameterIndex, value);
    }

    public void setClob(int parameterIndex, Reader reader) throws SQLException {
        preparedStatement.setClob(parameterIndex, reader);
    }

    public void setBlob(int parameterIndex, InputStream inputStream) throws SQLException {
        preparedStatement.setBlob(parameterIndex, inputStream);
    }

    public void setNClob(int parameterIndex, Reader reader) throws SQLException {
        preparedStatement.setNClob(parameterIndex, reader);
    }

    public boolean isClosed() throws SQLException {
        return preparedStatement.isClosed();
    }

    public void setPoolable(boolean poolable) throws SQLException {
        preparedStatement.setPoolable(poolable);
    }

    public boolean isPoolable() throws SQLException {
        return preparedStatement.isPoolable();
    }

    public <T> T unwrap(Class<T> iface) throws SQLException {
        return preparedStatement.unwrap(iface);
    }

    public boolean isWrapperFor(Class<?> iface) throws SQLException {
        return preparedStatement.isWrapperFor(iface);
    }

    private static void close(PreparedStatement preparedStatement, ResultSet resultSet) {
        try {
            if (resultSet != null)
                resultSet.close();
        } catch (SQLException e) {
            logger.error("Failed to close resultSet", e);
        }
        try {
            if (preparedStatement != null)
                preparedStatement.close();
        } catch (SQLException e) {
            logger.error("Failed to close preparedStatement", e);
        }
    }

    @Override
    public boolean isCloseOnCompletion() throws SQLException {
        return preparedStatement.isCloseOnCompletion();
    }

    @Override
    public void closeOnCompletion() throws SQLException{
        preparedStatement.closeOnCompletion();
    }
}
