/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.common.table;

import java.util.Map;
import java.util.WeakHashMap;
import java.util.Collections;
import java.lang.ref.WeakReference;
import java.sql.ResultSet;

/**
 * Based on the fact that {@link String} is immutable,  this cache is designed to eliminate duplicate copies of the same
 * value resulting from repeated retrieval of string-values from database, files and other external sources.
 * <p>In particular, this should be used everywhere a string-value is rerieved from a JDBC {@link ResultSet} because
 * {@link ResultSet#getString(int)} always constructs a string based on a <b>new</b> buffer, thus if the same value is read
 * multiple times from the DB, it'll be stored using multiple byte-arrays in the memory, instead of just one.
 * <p>Should be used as in the following example:<pre>
 * StringCache.get(resultSet.getString(3))
 * </pre>
 */
public class StringCache {

    /**
     * Because a {@link WeakHashMap} is used to hold the cached strings, both the key and values in the map are wrapped
     * in a {@link WeakReference} making it possible for the GC to reclaim "weakly reachable" strings (that is, there's
     * no strong reference to the string).
     */
    static private Map<String, WeakReference<String>> map = Collections.synchronizedMap(new WeakHashMap<String, WeakReference<String>>());

    public static String get(String val) {
        if (val == null)
            return null;
        WeakReference<String> res = map.get(val);
        if (res == null) {
            res = new WeakReference<String>(val);
            map.put(val, res);
            return val;
        }

        String resultRef = res.get();
        if (resultRef == null) {
            res = new WeakReference<String>(val);
            map.put(val, res);
            return val;
        }
        return resultRef;
    }
}
