/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.common.table;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

/**
 * General:
 * 1) Get methods throw an exception
 * 2) Find methods return null
 */
abstract public class Table<T> extends TableResultParserDefault<T> {

    static transient final Logger logger = LoggerFactory.getLogger("com.onaro.sanscreen.util.table");

    /**
     * Name of the table
     */
    private String name;

    /**
     * Full name
     */
    private String fullName;

    /**
     * Database the table is part of
     */
    private String database;

    /**
     * Something to be refactored. Connection to the database
     */
    private final Connection connection;

    /**
     * result parser
     */
    protected TableResultParserDefault<T> tableParser;

    /**
     * used only by those OriginTableWrappers
     */
    protected Table() {
        connection = null;
    }

    /**
     * Creates instance of table class
     *
     * @param connection - database connection
     * @param database   - name of the database table is part of
     * @param name       - name of the table within that database
     */
    protected Table(Connection connection, String database, String name) {
        this(connection, database, name, null);
    }

    protected Table(Connection connection, String database, String name, TableResultParserDefault<T> tableParser) {
        this.name = name;
        this.connection = connection;
        this.database = database;
        this.fullName = database + "." + name;
        if (tableParser == null) {
            this.tableParser = this;
        } else {
            this.tableParser = tableParser;
        }
    }


    /**
     * Gets full name of the table which is database + name
     *
     * @return full name of the table
     */
    public String getFullName() {
        return fullName;
    }

    /**
     * Gets table name without database prefix
     *
     * @return the table name
     */
    public String getName() {
        return name;
    }

    /**
     * Sets table name
     *
     * @param name - table name
     */
    public void setName(String name) {
        this.name = name;
        this.fullName = database + "." + name;
    }

    public String getDatabase() {
        return database;
    }

    public void setDatabase(String database) {
        this.database = database;
        this.fullName = database + "." + name;
    }

    public boolean isTableExist() throws TableException {
        return isTableExist(connection, database, name);
    }

    public static boolean isTableExist(Connection connection, String database, String name) throws TableException {
        return PreparedStatementWrapper.create(connection, "SHOW TABLES FROM " + database + " LIKE '" + name + "'", new TableResultParser<Boolean>() {
            public List<Boolean> parseResult(ResultSet resultSet) throws SQLException {
                return Collections.singletonList(resultSet.next() ? Boolean.TRUE : Boolean.FALSE);
            }
        }).executeQueryEx().iterator().next().booleanValue();
    }

    public Connection getConnection() {
        return connection;
    }

    /**
     * This method added for child classes to override so that they can be tested
     * using TestTable class
     *
     * @param tableObject - table object
     * @throws TableException if failed to insert
     */
    public void insert(T tableObject) throws TableException {
    }

    /**
     * This method added for child classes to override so that they can be tested
     * using TestTable class
     *
     * @param tableObjects - collections table object
     * @throws TableException if failed to insert
     */
    public void insert(Collection<T> tableObjects) throws TableException {
    }

    /**
     * Derived table classes: "implement me"
     *
     * @param resultSet
     * @return List containing the table result set
     * @throws SQLException
     */
    abstract public T parseObject(ResultSet resultSet) throws SQLException;

    /**
     * @return number of rows for this table
     * @throws TableException
     */
    public int getCount() throws TableException {
        return (int) getPrepareStatementEx("SELECT COUNT(*) FROM " + fullName).executeFunctionQueryEx();
    }

    /**
     * @param objId
     * @return number of rows for objId
     * @throws TableException
     */
    public int getCountById(long objId) throws TableException {
        return (int) getPrepareStatementEx("SELECT COUNT(*) FROM " + fullName + " WHERE objid=?").setNext(objId).executeFunctionQueryEx();
    }


    /**
     * delete all rows from table
     *
     * @throws TableException
     */
    public void truncate() throws TableException {
        getPrepareStatementEx("TRUNCATE " + fullName).executeUpdateEx();
    }

    /**
     * delete all rows from table
     *
     * @throws TableException
     */
    public void deleteAll() throws TableException {
        deleteAll(connection, fullName);
    }

    /**
     * delete all rows from table
     *
     * @param connection    - database connection
     * @param tableFullName - full name o fthe table to delete data for
     * @throws TableException
     */
    public static void deleteAll(Connection connection, String tableFullName) throws TableException {
        PreparedStatementWrapper.createForUpdate(connection, "DELETE FROM " + tableFullName).executeUpdateEx();
    }

    /**
     * delete rows (by objId) from table
     *
     * @throws TableException
     */
    public void deleteById(long objId) throws TableException {
        getPrepareStatementEx("DELETE FROM " + fullName + " WHERE objid=?").setNext(objId).executeUpdateEx();
    }

    /**
     * @deprecated instead of getPrepareStatement use one of the following methods
     *             <li>{@link #getPrepareStatementEx(String)}
     *             <li>{@link #getPrepareStatementEx(String, TableResultParser)} .
     */
    protected PreparedStatementWrapper<T> getPrepareStatement(String sql) throws SQLException {
        PreparedStatement preparedStatement = connection.prepareStatement(sql);
        preparedStatement.setFetchSize(Integer.MIN_VALUE);
        return new PreparedStatementWrapper<T>(preparedStatement, this, sql, null);
    }

    /**
     * The function takes care of exceptions and closing the preparedStatement in such case.
     *
     * @param sql
     * @return PreparedStatementWrapper use the the set Next inseted of setT(int i, T)
     * @throws TableException if failed to create the PreparedStatementWrapper
     */
    public PreparedStatementWrapper<T> getPrepareStatementEx(String sql) throws TableException {
        return PreparedStatementWrapper.create(connection, sql, tableParser);
    }

    /**
     * As the previous function, but allows to use a TableResultParser other than this.
     *
     * @param sql               the sql to exec
     * @param tableResultParser alternative TableResultParser thnen the Table default
     * @return PreparedStatementWrapper use the the set Next inseted of setT(int i, T)
     * @throws TableException if failed to create the PreparedStatementWrapper
     */
    public <K> PreparedStatementWrapper<K> getPrepareStatementEx(String sql, TableResultParser<K> tableResultParser) throws TableException {
        return PreparedStatementWrapper.create(connection, sql, tableResultParser);
    }

    /**
     * @param objId
     * @return the object
     * @throws TableException if the object is null or database error
     */
    public T get(long objId) throws TableException {
        T object = find(objId);
        if (object == null) {
            throw new TableException("Failed to find objId:" + objId);
        }
        return object;
    }

    /**
     * @param objId
     * @return null if object not found
     * @throws TableException if there is database error
     */
    public T find(long objId) throws TableException {
        List<T> rows = getPrepareStatementEx("SELECT * FROM " + fullName + " WHERE objid=?" + " LIMIT 1").setNext(objId).executeQueryEx();
        if (rows.size() == 0) {
            return null;
        }
        return rows.iterator().next();
    }

    /**
     * find all rows from table
     *
     * @throws TableException
     */
    public List<T> findAll() throws TableException {
        return getPrepareStatementEx("SELECT * FROM " + fullName).executeQueryEx();
    }

    /**
     * @return the Max objId from the table
     * @throws TableException
     */
    public long getMaxObjId() throws TableException {
        return getPrepareStatementEx("SELECT MAX(objId) FROM " + fullName).executeFunctionQueryEx();
    }

    /**
     * Method check if there is data in the table,
     * optimized for maximum performance
     *
     * @return <code>true</code> if table has data <code>false</code> otherwise
     */
    public boolean hasData() throws TableException {
        return !getPrepareStatementEx("SELECT * FROM " + fullName + " LIMIT 1").executeQueryEx().isEmpty();
    }

    /**
     * @return runs 'show create table <table-name>' and return an sql corresponding to the create table statement.
     * @throws TableException
     */
    public String getCreateTable() throws TableException {
        TableResultParser<String> tableResultParser = new TableResultParserDefault<String>() {
            public String parseObject(ResultSet resultSet) throws SQLException {
                return resultSet.getString(2);
            }
        };

        return getPrepareStatementEx("SHOW CREATE TABLE " + fullName, tableResultParser).executeQuerySingleResult();
    }

    /**
     * rename the table name
     *
     * @throws TableException if failed to rename (probably table exist)
     */
    public void renameTable(String newTableName) throws TableException {
        getPrepareStatementEx("RENAME TABLE " + fullName + "  TO " + database + "." + newTableName).executeUpdateEx();
        setName(newTableName);
    }

    /**
     * Drop the table
     *
     * @throws TableException - if failed to drop the table
     */
    public void dropTable() throws TableException {
        getPrepareStatementEx("DROP TABLE IF EXISTS " + fullName).executeUpdateEx();
    }

    /**
     * Drop the table
     *
     * @throws TableException - if failed to drop the table
     */
    public void dropDatabase() throws TableException {
        getPrepareStatementEx("DROP database IF EXISTS " + database).executeUpdateEx();
    }

    protected void closeConnection(PreparedStatement preparedStatement, ResultSet resultSet) {
        try {
            if (resultSet != null)
                resultSet.close();
        } catch (SQLException e) {
            logger.error("Failed to close resultSet", e);
        }
        try {
            if (preparedStatement != null)
                preparedStatement.close();
        } catch (SQLException e) {
            logger.error("Failed to close preparedStatement", e);
        }
    }

    /**
     * Creates a temporary table containing the specified IDs, which is used for join queries.
     *
     * @param idTableName the name of the IDs table
     * @param ids         the ids to insert into the table
     * @throws TableException if failed to create the table or to insert IDs into it
     */
    protected void createIdsTemporaryTableForJoin(String idTableName, Collection<Long> ids) throws TableException {
        createIdsTemporaryTableForJoin(connection, idTableName, new HashSet<Long>(ids));
    }

    /**
     * Creates a temporary table containing the specified string IDs, which is used for join queries.
     *
     * @param idTableName the name of the IDs table
     * @param ids         the String ids to insert into the table
     * @throws TableException if failed to create the table or to insert IDs into it
     */
    protected void createStringIdsTemporaryTableForJoin(String idTableName, Collection<String> ids) throws TableException {
        createStringIdsTemporaryTableForJoin(connection, idTableName, new HashSet<String>(ids));
    }

    /**
     * Drop a temporary table containing the specified ids, which is used for join queries.
     *
     * @param connection
     * @param idTableName
     * @throws TableException
     */
    public static void dropTemporaryTable(Connection connection, String idTableName, Logger logger) throws TableException {
        String sql = "DROP TEMPORARY TABLE IF EXISTS " + idTableName;
        boolean debug = logger != null && logger.isDebugEnabled();
        if (debug) {
            logger.info(sql + ";");
        }
        PreparedStatementWrapper.createForUpdate(connection, sql).executeUpdateEx();
    }

    public static void createIdsTemporaryTableForJoin(Connection connection, String idTableName, Set<Long> ids) throws TableException {
        createIdsTemporaryTableForJoin(connection, idTableName, ids, null);
    }

    /**
     * Creates a temporary table containing the specified ids, which is used for join queries.
     *
     * @param connection
     * @param idTableName
     * @param ids
     * @throws TableException
     */
    public static void createIdsTemporaryTableForJoin(Connection connection, String idTableName, Set<Long> ids, Logger logger) throws TableException {
        boolean debug = logger != null && logger.isDebugEnabled();

        String sql = "DROP TEMPORARY TABLE IF EXISTS " + idTableName;
        if (debug) {
            logger.info(sql + ";");
        }
        PreparedStatementWrapper.createForUpdate(connection, sql).executeUpdateEx();

        sql = "CREATE TEMPORARY TABLE " + idTableName + " (objid bigint(20), primary key(objid)) engine=InnoDB";
        if (debug) {
            logger.info(sql + ";");
        }
        PreparedStatementWrapper.createForUpdate(connection, sql).executeUpdateEx();
        if (ids.size() == 0) {
            return;
        }
        populateTempTable(connection, idTableName, ids, logger);


    }

    public static void populateTempTable(Connection connection, String idTableName, Set<Long> ids, org.slf4j.Logger logger) throws TableException {
        boolean debug = logger != null && logger.isDebugEnabled();

        String sql;//breaks ids set into 20000 records unit. It is because if set size is too big, the generated
        //string can cause OutOfMemoryException
        final int MAX_LENGTH = 20000;

        Iterator iterator = ids.iterator();
        while (iterator.hasNext()) {
            int i = 0;
            StringBuilder insertStatement = new StringBuilder("INSERT INTO ").append(idTableName).append(" VALUES ");
            boolean first = true;
            while (i++ < MAX_LENGTH && iterator.hasNext()) {
                if (!first) {
                    insertStatement.append(',');
                }
                insertStatement.append('(').append(iterator.next()).append(')');
                first = false;
            }
            sql = insertStatement.toString();
            if (debug) {
                logger.info(sql + ";");
            }
            PreparedStatementWrapper.createForUpdate(connection, sql).executeUpdateEx();
        }
    }

    /**
     * Deletes all the rows in the temporary table.  But, the table remains.
     *
     * @param connection        SQL connection
     * @param tableName         temp table name
     * @param logger            sql trace logger
     * @throws TableException   if errors were encountered while creating temporary table.
     */
    public static void clearTempTable(Connection connection, String tableName, Logger logger) throws TableException {
        boolean debug = logger != null && logger.isDebugEnabled();
        final String deleteTableRowsSql = "DELETE FROM " + tableName;
        if (debug) {
            logger.info(deleteTableRowsSql + ";");
        }
        PreparedStatementWrapper.createForUpdate(connection, deleteTableRowsSql).executeUpdateEx();
    }

    /**
     * Creates a temporary table containing the specified ids, which is used for join queries.
     *
     * @param connection
     * @param idTableName
     * @param ids
     * @throws TableException
     */
    public static void createStringIdsTemporaryTableForJoin(Connection connection, String idTableName, Set<String> ids) throws TableException {
        PreparedStatementWrapper.createForUpdate(connection, "DROP TEMPORARY TABLE IF EXISTS " + idTableName).executeUpdateEx();
        PreparedStatementWrapper.createForUpdate(connection, "CREATE TEMPORARY TABLE " + idTableName + " (objid varchar(255), primary key(objid)) engine=InnoDB").executeUpdateEx();
        if (ids.size() == 0) {
            return;
        }

        StringBuilder insertStatement = new StringBuilder("INSERT INTO ").append(idTableName).append(" VALUES ");
        boolean first = true;
        for (String id : ids) {
            if (!first) {
                insertStatement.append(',');
            }
            insertStatement.append("('").append(id).append("')");
            first = false;
        }
        PreparedStatementWrapper.createForUpdate(connection, insertStatement.toString()).executeUpdateEx();
    }

    public static void createKeyToIntValueTemporaryTableForJoin(Connection connection, String idTableName, Map<Long, Integer> keyToValue) throws TableException {
        PreparedStatementWrapper.createForUpdate(connection, "DROP TEMPORARY TABLE IF EXISTS " + idTableName).executeUpdateEx();
        PreparedStatementWrapper.createForUpdate(connection, "CREATE TEMPORARY TABLE " + idTableName + " (objid bigint(20), value int(11), primary key(objid, value)) engine=InnoDB").executeUpdateEx();
        if (keyToValue.size() == 0) {
            return;
        }

        //breaks ids set into 20000 records unit. It is because if set size is too big, the generated
        //string can cause OutOfMemoryException
        final int MAX_LENGTH = 20000;

        Iterator<Long> iterator = keyToValue.keySet().iterator();
        while (iterator.hasNext()) {
            int i = 0;
            StringBuilder insertStatement = new StringBuilder("INSERT INTO ").append(idTableName).append(" VALUES ");
            boolean first = true;
            while (i++ < MAX_LENGTH && iterator.hasNext()) {
                if (!first) {
                    insertStatement.append(',');
                }
                Long key = iterator.next();
                Integer value = keyToValue.get(key);
                insertStatement.append('(').append(key).append(',').append(value).append(')');
                first = false;
            }
            PreparedStatementWrapper.createForUpdate(connection, insertStatement.toString()).executeUpdateEx();
        }
    }

    public static void main(String[] args) {
        Connection conn = ConnectionHelper.newConnectionNoEjb();
        Set<Long> ls = new HashSet<Long>(30000);
        for (int i = 1; i < 30000; i++)
            ls.add(new Long(i));

        try {
            Table.createIdsTemporaryTableForJoin(conn, "test", ls);
        } catch (TableException e) {
            e.printStackTrace();
        }
    }
}
