/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.common.table;


import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

/**
 * @author Yarom Gabay
 */
public abstract class TableResultParserDefault<T> implements TableResultParser<T> {

    public static final TableResultParserDefault<String> STRING = new TableResultParserDefault<String>() {
        public String parseObject(ResultSet resultSet) throws SQLException {
            return StringCache.get(resultSet.getString(1));
        }
    };

    public static final TableResultParserDefault<Long> LONG = new TableResultParserDefault<Long>() {
        public Long parseObject(ResultSet resultSet) throws SQLException {
            return Long.valueOf(resultSet.getLong(1));
        }
    };


    public static final TableResultParserDefault<Boolean> BOOLEAN = new TableResultParserDefault<Boolean>() {
        public Boolean parseObject(ResultSet resultSet) throws SQLException {
            return Boolean.valueOf(resultSet.getBoolean(1));
        }
    };

    public static final TableResultParserDefault<Map<Long, Long>> LONG_TO_LONG_MAP = new TableResultParserDefault<Map<Long, Long>>() {
        private Map<Long, Long> longToLong = new HashMap<Long, Long>();

        @Override
        public List<Map<Long, Long>> parseResult(ResultSet resultSet) throws SQLException {
            while (resultSet.next()) {
                parseObject(resultSet);
            }
            return Collections.singletonList(longToLong);
        }

        @Override
        public Map<Long, Long> parseObject(ResultSet resultSet) throws SQLException {
            longToLong.put(resultSet.getLong(1), resultSet.getLong(2));
            return null;
        }
    };

    public static TableResultParserDefault<Map<Long, String>> newLongStringMap() {
        final Map<Long, String> longToString = new HashMap<Long, String>();
        return new TableResultParserDefault<Map<Long, String>>() {
            @Override
            public List<Map<Long, String>> parseResult(ResultSet resultSet) throws SQLException {
                while (resultSet.next()) {
                    parseObject(resultSet);
                }
                return Collections.singletonList(longToString);
            }

            @Override
            public Map<Long, String> parseObject(ResultSet resultSet) throws SQLException {
                longToString.put(resultSet.getLong(1), StringCache.get(resultSet.getString(2)));
                return null;
            }
        };
    }

    /**
     * Factory method to create a table parser, that can parse result sets with (String, String) key value pairs.
     */
    public static TableResultParserDefault<Map<String, String>> newStringStringMap() {
        final Map<String, String> stringToString = new HashMap<String, String>();
        return new TableResultParserDefault<Map<String, String>>() {
            @Override
            public List<Map<String, String>> parseResult(ResultSet resultSet) throws SQLException {
                while (resultSet.next()) {
                    parseObject(resultSet);
                }
                return Collections.singletonList(stringToString);
            }

            @Override
            public Map<String, String> parseObject(ResultSet resultSet) throws SQLException {
                stringToString.put(StringCache.get(resultSet.getString(1)), StringCache.get(resultSet.getString(2)));
                return null;
            }
        };
    }

    public List<T> parseResult(ResultSet resultSet) throws SQLException {
        LinkedList<T> res = new LinkedList<T>();
        while (resultSet.next()) {
            res.add(parseObject(resultSet));
        }
        return res;
    }

    public abstract T parseObject(ResultSet resultSet) throws SQLException;
}