package com.netapp.oci.platform.framework.performance;

import java.sql.Connection;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.tables.SqlMethod;

/**
 * Persistence facade class.
 *
 * @author Anton Sergeev
 */
public class PersistenceFacade<C, S> {

    protected Connection connection;

    protected PerformancePackageFactory<C, S> packageFactory;

    public PersistenceFacade(Connection connection, PerformancePackageFactory<C, S> packageFactory) throws PerformanceException {
        this.packageFactory = packageFactory;
        this.connection = connection;
        createTables();
    }

    /**
     * Saves new samples to db
     *
     * @param samples containing at most a single sample per monitored object (i.e. switch port, or virtual-machine)
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to save the samples in the DB or if duplicates were found
     */
    public void insert(Collection<Sample<C>> samples) throws PerformanceException {
        packageFactory.getPersistenceStrategy().execute(connection, samples);
    }
    
    /**
     * Saves new summaries to db
     *
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to save the summaries in the DB
     * */
    public void insertHourlySummaries(Connection connection, Long hourToSummarize) throws PerformanceException {
        packageFactory.getPersistenceStrategy().insertHourlySummaries(connection, hourToSummarize);
    }

    /**
     * Update the daily summary table with the samples
     *
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to save the summaries in the DB
     * */
    public void executeDailySummaries(Collection<Sample<C>> samples) throws PerformanceException {
        packageFactory.getPersistenceStrategy().executeDailySummary(connection, samples);
    }

    /**
     * Retrieves Summary<S> from the summary tables.hadar
     *
     * @param ids      identifies the object for which to retrieve samples (ojId)
     * @param timeSpan start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to retrieve samples or compressed-samples
     */
    public List<Summary<S>> getSummaries(TimeSpan timeSpan, Collection<Long> ids) throws PerformanceException {
        return packageFactory.getSummaryFacade().getSummaries(connection, timeSpan, ids);
    }

    /**
     * Retrieves samples from the compressed and not-compressed tables.
     *
     * @param id       identifies the object for which to retrieve samples (ojId)
     * @param timeSpan start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to retrieve samples or compressed-samples
     */
    public List<Sample<C>> getSamples(Long id, TimeSpan timeSpan) throws PerformanceException {
        return packageFactory.getSampleFacade().getSamples(connection, id, timeSpan);
    }

    public Map<Long, List<Sample<C>>> getSamples(Set<Long> ids, TimeSpan timeSpan) throws PerformanceException {
        return packageFactory.getSampleFacade().getSamples(connection, ids, timeSpan);
    }

    /**
     * Retrieves the latest sample for the given object ids.
     *
     * @param ids       identifies the object for which to retrieve samples (ojId)
     * @return UnOrdered Samples
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to retrieve samples 
     */
    public Map<Long, Sample<C>> getLatestSample(Set<Long> ids) throws PerformanceException {
        return packageFactory.getSampleFacade().getLatestSample(connection, ids);
    }

    /**
     * Retrieves samples from the compressed and not-compressed tables group by time.
     *
     * @param ids       identifies the object for which to retrieve samples (ojId)
     * @param timeSpan start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to retrieve samples or compressed-samples
     */
    public List<Sample<C>> getSamplesGroupByTime(Set<Long> ids, TimeSpan timeSpan) throws PerformanceException {
        return packageFactory.getSampleFacade().getSamplesGroupByTime(connection, ids, timeSpan);
    }

    /**
     * Retrieves samples from the compressed and not-compressed tables group by time.
     *
     * @param timeSpan start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to retrieve samples or compressed-samples
     */
    public List<Sample<C>> getSamplesGroupByTime(TimeSpan timeSpan) throws PerformanceException {
        return packageFactory.getSampleFacade().getSamplesGroupByTime(connection, timeSpan);
    }

    /**
     * returns true if samples exist over the time period
     *
     * @param timeSpan start and end times filtering the samples
     * @return true if samples exist
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to retrieve samples
     */
    public Boolean doSamplesExist(TimeSpan timeSpan, Long id) throws PerformanceException {
        return packageFactory.getSampleFacade().doSamplesExist(connection, timeSpan, id);
    }


    /**
     * For all objects (i.e. port), gets a single summary of samples of the entire requested period (that is, a single summary
     * per object).
     *
     * @param timeSpan the period for which to get the summary
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for each object during the requested period - indexed by the object's ID
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if any of the queries failed
     */
    public Map<Long, Summary<S>> getSummaryById(TimeSpan timeSpan, SqlMethod[] methods) throws PerformanceException {
        return getSummaryById(timeSpan, null, methods);
    }

    /**
     * For each object (i.e. port), gets a merge summary of samples of the entire requested period (that is, a single summary
     * per object). The summary is calculated from samples not yet processed and from summaries of older samples.
     *
     * @param timeSpan the period for which to get the summary
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if any of the queries failed
     */
    public List<Summary<S>> getMergedSummariesGroupById(TimeSpan timeSpan, SqlMethod[] methods) throws PerformanceException {
        return getMergedSummariesGroupById(timeSpan, null, methods);
    }


    public List<Summary<S>> getMergedSummariesGroupById(TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws PerformanceException {
        return packageFactory.getSummaryFacade().getMergedSummariesGroupById(connection, timeSpan, ids, methods);
    }
    
    /**
     * Get a list of sample summaries in the entire requested period
     *
     * @param timeSpan the period for which to get the summaries
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if any of the queries failed
     */
    public List<Summary<S>> getSummariesGroupById(TimeSpan timeSpan) throws PerformanceException {
        return getSummariesGroupById(timeSpan, Collections.<Long>emptyList());
    }

    /**
     * Get a list of sample summaries in the entire requested period
     *
     * @param timeSpan the period for which to get the summaries
     * @param ids  a list of sample object ids
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if any of the queries failed
     */
    public List<Summary<S>> getSummariesGroupById(TimeSpan timeSpan, Collection<Long> ids) throws PerformanceException {
        return packageFactory.getSummaryFacade().getSummaries(connection, timeSpan, ids);
    }

    /**
     * Gets a merge summary of the entire requested period for all objects.
     *
     * @param timeSpan the period for which to get the summary
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for all objects during the requested period.
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException
     *          if any of the queries failed
     */
    public S getMergedSummaries(TimeSpan timeSpan, SqlMethod[] methods) throws PerformanceException {
        return packageFactory.getSummaryFacade().getMergedSummaries(connection, timeSpan, methods);
    }

    /**
     * get a hourly patern for a given time period
     *
     * @param timeSpan the period for which to get the summary
     * @param ids      the seslected summary objects
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if any of the queries failed
     */
    public List<Summary<S>> getHourlyPattern(TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws PerformanceException {
        return packageFactory.getSummaryFacade().getHourlyPattern(connection, timeSpan, ids, methods);
    }


    /**
     * For each of the objects (i.e. port) who's ID was specified, gets a single summary of samples of the entire requested
     * period (that is, a single summary per object). The summary is calculated from samples not yet processed and from
     * summaries of older samples.
     *
     * @param timeSpan the period for which to get the summary
     * @param ids      identifies the objects for which summaries are requested, can be <code>null</code> - in this
     *                 case data will be returned for all objects
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if any of the queries failed
     */
    public Map<Long, Summary<S>> getSummaryById(TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws PerformanceException {
        return packageFactory.getSummaryFacade().getMergedSummaryById(connection, timeSpan, ids, methods);
    }

    /**
     * Create all the tables if they don't already exist.
     *
     * @throws PerformanceException if failed to create any of the tables
     */
    public void createTables() throws PerformanceException {
        packageFactory.getTableManagement().createTables(connection);
    }

    /**
     * Drops all performance tables
     *
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to drop them all
     */
    public void dropTables() throws PerformanceException {
        packageFactory.getTableManagement().dropTables(connection);
    }

    /**
     * @return the PerformanceComponentsFactory passed by the constructor
     */
    public PerformancePackageFactory<C, S> getPerformancePackageFactory() {
        return packageFactory;
    }

    /**
     * return first sample after afterTime
     *
     * @param connection
     * @param id
     * @param afterTime
     * @return
     * @throws PerformanceException
     */
    public Sample<C> getFirstSampleAfterTime(Connection connection, Long id, long afterTime) throws PerformanceException {
        return packageFactory.getSampleFacade().getFirstSampleAfterTime(connection, id, afterTime);
    }

    /**
     * return last sample before beforeTime
     *
     * @param connection
     * @param id
     * @param beforeTime
     * @return
     * @throws PerformanceException
     */
    public Sample<C> getLastSampleBeforeTime(Connection connection, Long id, long beforeTime) throws PerformanceException {
        return packageFactory.getSampleFacade().getLastSampleBeforeTime(connection, id, beforeTime);
    }

    /**
     * return first summary after afterTime
     *
     * @param connection
     * @param id
     * @param afterTime
     * @return
     * @throws PerformanceException
     */
    public Summary<S> getFirstSummaryAfterTime(Connection connection, Long id, Long afterTime) throws PerformanceException {
        return packageFactory.getSummaryFacade().getFirstSummaryAfterTime(connection, id, afterTime);
    }

    /**
     * return last summary before beforeTime
     *
     * @param connection
     * @param id
     * @param beforeTime
     * @return
     * @throws PerformanceException
     */
    public Summary<S> getLastSummaryBeforeTime(Connection connection, Long id, Long beforeTime) throws PerformanceException {
        return packageFactory.getSummaryFacade().getLastSummaryBeforeTime(connection, id, beforeTime);
    }
}