package com.netapp.oci.platform.framework.performance;

import java.sql.Connection;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.tables.SqlMethod;

/**
 * Facade for simple summary queries
 *
 * @author Anton Sergeev
 */
public interface SummaryFacade<S> {
    /**
     * For all objects (i.e. port), gets a single summary of samples of the entire requested period (that is, a single summary
     * per object). The summary is calculated from samples not yet processed and from summaries of older samples.
     *
     * @param connection - database connection
     * @param timeSpan the period for which to get the summary
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for each object during the requested period - indexed by the object's ID
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if any of the queries failed
     */
    Map<Long, Summary<S>> getMergedSummaryById(Connection connection, TimeSpan timeSpan, SqlMethod[] methods) throws PerformanceException;

    /**
     * For each of the objects (i.e. port) who's ID was specified, gets a single summary of samples of the entire requested
     * period (that is, a single summary per object). The summary is calculated from samples not yet processed and from
     * summaries of older samples.
     *
     * @param connection - database connection
     * @param timeSpan the period for which to get the summary
     * @param ids      identifies the objects for which summaries are requested, can be <code>null</code> - in this
     *                 case data will be returned for all objects
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws PerformanceException if any of the queries failed
     */
    Map<Long, Summary<S>> getMergedSummaryById(Connection connection, TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws PerformanceException;

    /**
     * For each object (i.e. port), gets a merge summary of samples of the entire requested period (that is, a single summary
     * per object). The summary is calculated from samples not yet processed and from summaries of older samples.
     *
     * @param connection - database connection
     * @param timeSpan the period for which to get the summary
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws PerformanceException if any of the queries failed
     */
    List<Summary<S>> getMergedSummariesGroupById(Connection connection, TimeSpan timeSpan, SqlMethod[] methods) throws PerformanceException;

    List<Summary<S>> getMergedSummariesGroupById(Connection connection, TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws PerformanceException;

    /**
     * Gets a merge summary of the entire requested period for all objects.
     *
     * @param connection - database connection
     * @param timeSpan the period for which to get the summary
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for all objects during the requested period.
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException
     *          if any of the queries failed
     */
    S getMergedSummaries(Connection connection, TimeSpan timeSpan, SqlMethod[] methods) throws PerformanceException;

    /**
     * Retrieves Summary<S> from the summary tables
     *
     * @param connection - database connection
     * @param ids      identifies the object for which to retrieve samples (ojId)
     * @param timeSpan start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException if failed to retrieve samples or compressed-samples
     */
    public List<Summary<S>> getSummaries(Connection connection, TimeSpan timeSpan, Collection<Long> ids) throws PerformanceException;

    /**
     * Retrieves All Summary<S> from the summary tables
     *
     * @param connection - database connection
     * @param timeSpan start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws PerformanceException if failed to retrieve samples or compressed-samples
     */
    public List<Summary<S>> getSummaries(Connection connection, TimeSpan timeSpan) throws PerformanceException;

    /**
     * get a hourly patern for a given time period
     *
     * @param connection - database connection
     * @param timeSpan the period for which to get the summary
     * @param ids      the seslected summary objects
     * @param methods - sql methods to apply towards summary columns
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws PerformanceException if any of the queries failed
     */
    List<Summary<S>> getHourlyPattern(Connection connection, TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws PerformanceException;

    /**
     * get first summary after afterTime
     *
     * @param id
     * @param afterTime
     * @return
     * @throws PerformanceException
     */
    public Summary<S> getFirstSummaryAfterTime(Connection connection, Long id, Long afterTime) throws PerformanceException;

    /**
     * get last summary before beforeTime
     *
     * @param id
     * @param beforeTime
     * @return
     * @throws PerformanceException
     */
    public Summary<S> getLastSummaryBeforeTime(Connection connection, Long id, Long beforeTime) throws PerformanceException;

}
