package com.netapp.oci.platform.framework.performance.data;

import java.io.Serializable;
import com.netapp.oci.platform.common.TimeSpan;

/**
 * Repersents the summary of multiple {@link Sample}s measured for a single object (e.g. switch-port) over a period of
 * time. For each measurement a child-class provides the formula and the storage (class member) for
 * calculating and persisting the summary of the measured values.
 */
public class Summary<S> implements Serializable, Comparable<Summary> {
    /** */
    private static final long serialVersionUID = 8255557641847664237L;

    /**
     * Identifies the object for which this instance summarizes samples.
     */
    private Long id;

    /**
     * The period of time summarized into this instance.
     */
    private TimeSpan timeSpan;

    /**
     * Summary data
     */
    private S summaryData;


    public Summary(S summaryData) {
        this.summaryData = summaryData;
    }

    public Summary(Long id, TimeSpan timeSpan, S summaryData) {
        this.id = id;
        this.timeSpan = timeSpan;
        this.summaryData = summaryData;
    }


    public S getSummaryData() {
        return summaryData;
    }

    public void setSummaryData(S summaryData) {
        this.summaryData = summaryData;
    }

    public Long getId() {
        return id;
    }

    public void setId(Long id) {
        this.id = id;
    }

    public TimeSpan getTimeSpan() {
        return timeSpan;
    }

    public void setTimeSpan(TimeSpan timeSpan) {
        this.timeSpan = timeSpan;
    }

    public int compareTo(Summary summary) {
        if (timeSpan.getFromTime() == summary.getTimeSpan().getFromTime()) {
            return 0;
        }

        return getTimeSpan().getFromTime() < summary.getTimeSpan().getFromTime() ? -1 : 1;
    }

    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Summary summary = (Summary) o;

        if (id != summary.id) return false;
        if (timeSpan != null ? !timeSpan.equals(summary.timeSpan) : summary.timeSpan != null) return false;
        if (summaryData != null ? !summaryData.equals(summary.summaryData) : summary.summaryData != null) return false;

        return true;
    }

    public int hashCode() {
        int result;
        result = (int) (id ^ (id >>> 32));
        result = 31 * result + (timeSpan != null ? timeSpan.hashCode() : 0);
        result = 31 * result + (summaryData != null ? summaryData.hashCode() : 0);
        return result;
    }


    public String toString() {
        return "Summary{" +
                "id=" + id +
                ", timeSpan=" + timeSpan +
                ", summaryData=" + summaryData +
                '}';
    }
}
