package com.netapp.oci.platform.framework.performance.tables;

import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.Table;

import java.sql.Connection;
import java.util.Arrays;

/**
 * Table that uses collection of columns and indexes for create and insert statements
 *
 * @author Anton Sergeev
 */
public abstract class AbstractConfigurableTable<T> extends Table<T> {

    public static String SEPARATOR = System.getProperty("line.separator");

    /**
     * Table columns
     */
    private Column[] columns;

    /**
     * Table indexes
     */
    private Index[] indexes;

    protected AbstractConfigurableTable(Connection connection, String database, String name, Column[] columns,
                                        Index[] indexes) {
        super(connection, database, name);
        this.columns = columns;
        this.indexes = indexes;
    }

    /**
     * Convinience method that allows to append extra columns to table definition
     * when child inherits from this class
     *
     * @param left  - array that should be on the left in the result array
     * @param right - array that should be on the right in the result array
     * @return merged array
     */
    protected static Column[] mergeColumns(Column[] left, Column[] right) {
        Column[] result = Arrays.copyOf(left, right.length + left.length);
        System.arraycopy(right, 0, result, left.length, right.length);
        return result;
    }

    public Column[] getColumns() {
        return columns;
    }

    public Index[] getIndexes() {
        return indexes;
    }

    /**
     * Creates the database table needed for keeping sample summaries.
     *
     * @throws com.netapp.oci.platform.common.table.TableException
     *          if creation failed
     */
    public void createTable() throws TableException {
        getPrepareStatementEx(getCreateSql()).executeUpdateEx();
    }

    public String getCreateSql() {
        return getCreateTableSql(false);
    }

    protected String getCreateTempTableSql() {
        return getCreateTableSql(true);
    }

    private String getCreateTableSql(boolean tempTable) {
        String tempKeyWord = tempTable ? "TEMPORARY" : "";

        StringBuilder sql = new StringBuilder();

        sql.append("CREATE ").append(tempKeyWord).append(" TABLE IF NOT EXISTS ").append(getFullName()).append(" (\r\n");
        int i = 0;
        for (Column column : getColumns()) {
            if (i != 0) {
                sql.append(", \r\n");
            }
            i++;
            sql.append(column.toSql());
        }
        for (Index index : getIndexes()) {
            sql.append(", \r\n").append(index.value());
        }
        sql.append(") ENGINE = InnoDB");

        return sql.toString();
    }

    protected String getDropTempTableSql() {
        return "DROP TEMPORARY TABLE IF EXISTS " + getFullName();
    }

    public String getInsertSql() {
        StringBuilder sql = new StringBuilder();
        sql.append("INSERT INTO ").append(getFullName()).append(" (");
        int i = 0;
        for (Column column : columns) {
            if (i != 0) {
                sql.append(", ");
            }
            i++;
            sql.append(column.getName());
        }
        sql.append(") VALUES (");
        i = 0;
        for (Column column : columns) {

            if (i != 0) {
                sql.append(", ");
            }
            i++;
            if (column.isAutoIncrement())
                sql.append("NULL");
            else
                sql.append("?");
        }
        sql.append(")");
        return sql.toString();
    }
}
