package com.netapp.oci.platform.framework.performance.tables;

import com.netapp.oci.platform.common.util.Assert;

import java.util.Map;
import java.util.HashMap;
import java.sql.Types;

/**
 * @author Anton Sergeev
 */
public class Column {
	/* maximum varchar length */
	private static int MAX_VARCHAR_LEN = 1024;
	
    /* the column name */
    private String name;

    /* the column type */
    private String type;

    /* Column java SQL type */
    private int sqlType;

    /* Column can be null */
    private boolean canBeNull;

    /* Column is auto Increment*/
    private boolean autoIncrement;

    public Column(String name, String type) {
        this(name, type, true); // default is less restrictive, nullable
    }

    public Column(String name, String type, boolean canBeNull) {
        this(name, type, canBeNull, false);
    }

    public Column(String name, String type, boolean canBeNull, boolean autoIncrement) {
        this.name = name;
        this.type = type;
        this.sqlType = getSqlTypeFromType(type);
        this.canBeNull = canBeNull;
        this.autoIncrement = autoIncrement;
    }


    public boolean canBeNull() {
        return canBeNull;
    }

    public boolean isAutoIncrement() {
        return autoIncrement;
    }

    public String getName() {
        return name;
    }

    public String getType() {
        return type;
    }

    public int getSqlType() {
        return sqlType;
    }

    public String toSql() {
        StringBuilder builder = new StringBuilder();
        builder.append(getName()).append(" ").append(getType());
        
        // add max length to the varchar DDL
        if (getType().equalsIgnoreCase("varchar")) {
        	builder.append("(" + MAX_VARCHAR_LEN + ")");
        }
        
        if (!canBeNull()) {
            builder.append(" NOT NULL ");
        }

        if (isAutoIncrement())
            builder.append(" AUTO_INCREMENT ");
        return builder.toString();
    }

    public String toString() {
        return toSql();
    }

    private static Map<String, Integer> typeToSqlType = new HashMap<String, Integer>();

    static {
        typeToSqlType.put("int", Types.INTEGER);
        typeToSqlType.put("smallint", Types.SMALLINT);
        typeToSqlType.put("tinyint", Types.TINYINT);
        typeToSqlType.put("mediumint", Types.INTEGER);
        typeToSqlType.put("bigint", Types.INTEGER);
        typeToSqlType.put("bool", Types.BOOLEAN);
        typeToSqlType.put("float", Types.FLOAT);
        typeToSqlType.put("double", Types.DOUBLE);
        typeToSqlType.put("varchar", Types.VARCHAR);
        typeToSqlType.put("bigint unsigned", Types.BIGINT);
    }

    private static int getSqlTypeFromType(String type) {
        Assert.isNotNull(type);
        String key;
        int position = type.indexOf("(");
        if (position == -1) {
            key = type;
        } else {
            key = type.substring(0, position);
        }

        Assert.isTrue(typeToSqlType.containsKey(key), "Can't figure out sql type for type " + type);
        return typeToSqlType.get(key).intValue();
    }
}
