package com.netapp.oci.platform.framework.performance.tables;

import com.netapp.oci.platform.common.table.ObjectParser;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.data.Sample;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;

/**
 * User: Vinod M
 * Date: 10/21/15
 * Time: 1:27 PM
 */
public class DailySummaryTable<C> extends AbstractConfigurableTable<Sample<C>> {
    private final static Logger LOG = LoggerFactory.getLogger(DailySummaryTable.class);

//    public static final int BATCH_SIZE = SampleTable.BATCH_SIZE;
    public static String TABLE_NAME_PREFIX = "summary_daily_";
    private static Column[] SUMMARY_TABLE_COLUMNS = {
            new Column("objid", "bigint(20)"),
            new Column("date", "bigint(20)"),
            new Column("noOfSamples", "int(11)"),
    };
    private static Index[] SUMMARY_TABLE_INDEXES = {
            new Index(true, false, SUMMARY_TABLE_COLUMNS[0], SUMMARY_TABLE_COLUMNS[1]), // primary key
            new Index(SUMMARY_TABLE_COLUMNS[1]),
    };
    private final SqlMethod[] summarizationMethods;
    private final ObjectParser<C> sampleDataParser;


    public DailySummaryTable(Connection connection, String suffixName, Column[] columns,
                                SqlMethod[] summarizationMethods, ObjectParser<C> sampleDataParser) {
        super(connection, "netapp_performance", TABLE_NAME_PREFIX + suffixName,
                mergeColumns(SUMMARY_TABLE_COLUMNS, columns), SUMMARY_TABLE_INDEXES);
        if (columns.length != summarizationMethods.length) {
            throw new IllegalArgumentException("Number of columns should be equal to the number of summarization methods");
        }

        this.summarizationMethods = summarizationMethods;
        this.sampleDataParser = sampleDataParser;
    }

    @Override
    public Sample<C> parseObject(ResultSet resultSet) throws SQLException {
        throw new UnsupportedOperationException("Daily summary read not supported.");
    }

    @Override
    public void insert(Collection<Sample<C>> samples) throws TableException {
        PreparedStatementWrapper<Sample<C>> preparedStatement = getPrepareStatementEx(getInsertSql());

        try {
            int numOfSamples = 0;
            for (Sample<C> sample : samples) {
                if (!sample.isNull()) { // Ignore null samples
                    ++numOfSamples;
                    prepareSystemDataForInsert(preparedStatement, sample);
                    sampleDataParser.prepareObjectDataForInsert(preparedStatement, sample.getSampleData());
                    preparedStatement.executeUpdate();
/* mariadb-java-client-1.2.3 batch insert does not work for insets with ON DUPLICATE KEY UPDATE
                    preparedStatement.addBatch();

                    if (numOfSamples % BATCH_SIZE == 0) {
                        preparedStatement.executeBatch();
                    }
*/
                }
            }

/*
            // Insert remaining samples
            if (numOfSamples % BATCH_SIZE != 0) {
                preparedStatement.executeBatch();
            }
*/
        } catch (SQLException e) {
            throw new TableException("Failed to executeUpdate: " + this, e);
        } finally {
            preparedStatement.closePreparedStatement();
        }
    }

    public String getInsertSql() {
        StringBuilder sql = new StringBuilder();
        sql.append("INSERT INTO ").append(getFullName()).append(" (");
        int colIdx = 0;
        final Column[] columns = super.getColumns();
        for (Column column : columns) {
            if (colIdx != 0) {
                sql.append(", ");
            }
            colIdx++;
            sql.append(column.getName());
        }
        sql.append(") VALUES (");
        colIdx = 0;
        for (Column column : columns) {
            if (colIdx != 0) {
                sql.append(", ");
            }
            colIdx++;
            if (column.isAutoIncrement())
                sql.append("NULL");
            else
                sql.append("?");
        }
        sql.append(")\n");
        sql.append("ON DUPLICATE KEY UPDATE \n");
        for (colIdx = 3; colIdx < columns.length; colIdx++) { //Start with the 4th column. The first three are objId, date & noOfSamples
            SqlMethod summarizationMethod = summarizationMethods[colIdx - 3];
            sql.append(columns[colIdx].getName()).append(" = ");
            sql.append(summarizationMethod.wrapColumn(columns[colIdx].getName()));
            sql.append(",\n    ");
        }
        sql.append("noOfSamples = (noOfSamples + 1)\n") ;
        LOG.debug("Daily SQL to execute " + sql.toString());
        return sql.toString();
    }

    /**
     * Prepares system columns for insert if there are any and returns the number of columns prepared
     *
     * @param preparedStatement prepared statement wrapper
     * @param sample
     * @return number of columns prepared
     * @throws TableException if failed to prepare any of the columns
     */
    protected int prepareSystemDataForInsert(PreparedStatementWrapper preparedStatement, Sample<C> sample) throws TableException {
        Calendar sampleDay = Calendar.getInstance();
        sampleDay.setTime(new Date(sample.getTime()));
        sampleDay.set(Calendar.HOUR_OF_DAY, 0);
        sampleDay.set(Calendar.MINUTE, 0);
        sampleDay.set(Calendar.SECOND, 0);
        sampleDay.set(Calendar.MILLISECOND, 0);
        preparedStatement.setNext(sample.getId());
        preparedStatement.setNext(sampleDay.getTimeInMillis());
        //No of samples will be set to 1 to start, and then incremented.
        preparedStatement.setNext(1);
        return 3;
    }
}
