package com.netapp.oci.platform.framework.performance.tables;

import com.onaro.commons.types.Period;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.table.ObjectParser;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.data.Summary;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.GregorianCalendar;

/**
 * Summary table that knows how to insert summaries based on values from related sample table
 * Usually when performance package uses this kind of summary table it means that
 * a) summarization process is simple and can be discribed using SQL functions like MAX, SUM, AVG etc
 * b) SummaryCalculator is not implemented for such package since there is no need for it
 *
 * @author Anton Sergeev
 */
public class InsertSelectSummaryTable<S> extends SummaryTable<S> {
	private final static Logger LOG = LoggerFactory.getLogger(InsertSelectSummaryTable.class);
	
    private ObjectParser<S> objectParser;

    public InsertSelectSummaryTable(Connection connection, String suffixName, Column[] columns, ObjectParser<S> objectParser) {
        super(connection, suffixName, columns);
        this.objectParser = objectParser;
    }

    @Override
    protected void prepareSummaryDataForInsert(PreparedStatementWrapper preparedStatement, S summaryData) throws TableException {
        objectParser.prepareObjectDataForInsert(preparedStatement, summaryData);
    }

    public void insert(SampleTable sampleTable, Long hourToSummarize, SqlMethod [][] summaryMethods) throws TableException {

        //If the hour to summarize is null, then summarize the hour of the last sample in the summary table.
        TimeSpan tableTimeSpan = (hourToSummarize == null) ? sampleTable.getTableTimeSpan() : null;
        if (hourToSummarize == null && tableTimeSpan == null) {
            // no need to do anything table is empty
            return;
        }

        Calendar calendar = new GregorianCalendar();
        calendar.setTimeInMillis(hourToSummarize == null ? tableTimeSpan.getToTime() : hourToSummarize);

        // this will point to the beginning of the hour
        calendar.set(Calendar.MILLISECOND, 0);
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MINUTE, 0);

        long fromTime = calendar.getTimeInMillis();
        // this will point to the end of the hour
        long toTime = fromTime + Period.HOUR.milliseconds() - 1;

        StringBuilder sql = new StringBuilder();

        sql.append("INSERT INTO ").append(getFullName()).append("\n");

        sql.append("SELECT objid,?");
        sql.append(SqlMethod.getWrappedColumns("", sampleTable.getSampleColumns(), summaryMethods));
        sql.append("\n");

        sql.append("FROM ").append(sampleTable.getFullName()).append("\n");
        sql.append("WHERE empty = 0 and time >= ? and time <= ? GROUP BY objid \n");
        sql.append(getOnDuplicateKeyUpdate("", sampleTable.getSampleColumns(), summaryMethods));

        final PreparedStatementWrapper<Summary<S>> prepareStatementEx = getPrepareStatementEx(sql.toString());
        prepareStatementEx.setNext(fromTime);
        prepareStatementEx.setNext(fromTime);
        prepareStatementEx.setNext(toTime);
        prepareStatementEx.executeUpdateEx();
    }

    protected S parseSummaryData(ResultSet resultSet, int startIndex) throws SQLException {
        return objectParser.parseObjectData(resultSet, startIndex);
    }
    
    private String getOnDuplicateKeyUpdate(String prefix, Column[] columns, SqlMethod[][] methods) {
        StringBuilder sb = new StringBuilder();
        sb.append("ON DUPLICATE KEY UPDATE \n");
        for (int i = 0; i < columns.length; i++) {
            boolean columnSummarized = false;
            Column column = columns[i];
            SqlMethod[] columnMethods = methods[i];
            for (int j = 0; j < columnMethods.length; j++) {
            	
                if (columnMethods[j] == null){
                    continue;
                }
                
                if (j != 0) {
                    sb.append(",\r\n    ");
                }
                sb.append(prefix + column.getName()).append(" = ");
                sb.append("VALUES(").append(column.getName()).append(")");

                //At this point we know column is being summarized                
                columnSummarized = true;
            }

            // Add comma if column is being summarized
            if (columnSummarized) {
                sb.append(",\r\n    ");
            }
        }
        
        // Remove comma if it's last character of string builder excluding  last 6 chars ("\r\n    ")
        int lastIndex = sb.lastIndexOf(",");
        int length = sb.length();
        if (lastIndex != -1 && lastIndex == (length - 7)){
            sb.delete(lastIndex, length);        
        }
        return sb.toString();
    }
    
}
