/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.framework.performance.tables;

import com.netapp.oci.platform.common.table.ObjectParser;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.data.Sample;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;

/**
 * Implementation of the current sample table that uses partitions mechanism
 *
 * @author Anton Sergeev
 */
public class PartitionedCurrentSampleTable<C> extends CurrentSampleTable<C> {
    private Long containerId;

    private static Column[] CURRENT_PARTITIONED_SAMPLE_TABLE_COLUMNS = {
            new Column("containerid", "bigint(20)"),
    };

    public PartitionedCurrentSampleTable(Connection connection, ObjectParser<C> sampleDataParser, String containerKey, Long containerId, String suffixName, Column[] columns) {
        super(connection, sampleDataParser, containerKey, suffixName, mergeColumns(CURRENT_PARTITIONED_SAMPLE_TABLE_COLUMNS, columns));
        this.containerId = containerId;
    }

    @Override
    protected int prepareSystemDataForInsert(PreparedStatementWrapper wrapper) throws TableException {
        int numberOfColumns = super.prepareSystemDataForInsert(wrapper);
        wrapper.setNext(containerId);
        return numberOfColumns + 1;
    }

    /**
     * Inserts a sample collection. It also performs management of underlying partitions. Most of the insert operations
     * are expected to succeed with the current partition. When MySQL gives error 1526 and SQLSTATE HY000 new partition
     * is added.See <a href="http://dev.mysql.com/doc/refman/5.1/en/error-messages-server.html>
     * MySQL Server Error Codes and Messages</a>
     *
     * @param samples the samples to be written to the database
     * @throws com.netapp.oci.platform.common.table.TableException
     *
     */
    public void insert(Collection<Sample<C>> samples) throws TableException {
        try {
            deleteFromPartition();
            super.insert(samples);
        } catch (TableException e) {
            if (e.getCause() instanceof SQLException) {
                SQLException sqlEx = (SQLException) e.getCause();
                if (sqlEx.getErrorCode() == 1526 && sqlEx.getSQLState().equalsIgnoreCase("HY000")) {
                    addPartition();
                    super.insert(samples);
                    return;
                }
                // fall through for other SQLException, like primary key violation
            }

            throw e; // rethrow the exception for situation other than partition rollover
        }
    }

    private void addPartition() throws TableException {
        final String partitionAddSql = "ALTER TABLE " + getFullName() + " ADD PARTITION (" + SEPARATOR +
                "PARTITION p" + containerId + " VALUES in (" + containerId + "))";
        getPrepareStatementEx(partitionAddSql).executeUpdateEx();
    }

    private void deleteFromPartition() throws TableException {
        final String partitionDropSql = "DELETE FROM " + getFullName() + " WHERE containerId = " + containerId;
        getPrepareStatementEx(partitionDropSql).executeUpdateEx();
    }

    @Override
    protected C parseSampleData(ResultSet resultSet, int startIndex) throws SQLException {
        return super.parseSampleData(resultSet, startIndex + 1);
    }
}
