/*
 * Copyright (c) 2010 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.framework.performance.tables;

import com.netapp.oci.platform.common.table.PrefixConnection;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.ObjectParser;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

public class PartitionedSampleTable<C> extends SampleTable<C> {

    /**
     * number of partitions to keep
     */
    private int maxPartitions;

    private final LinkedList<Long> allPartitionOrdinals = new LinkedList<Long>();

    public PartitionedSampleTable(Connection connection, ObjectParser<C> sampleDataParser, String suffixName, Column[] columns, int maxPartitions) {
        super(connection, sampleDataParser, suffixName, columns);
        this.maxPartitions = maxPartitions;
        initAllPartitionOrdinals();
    }

    private void initAllPartitionOrdinals() {
        for (long i = 0; i <= maxPartitions; i++) {
            allPartitionOrdinals.add(i);
        }
    }

    /**
     * Inserts a sample collection. It also performs management of underlying partitions. Most of the insert operations
     * are expected to succeed with the current partition. When MySQL gives error 1526 and SQLSTATE HY000 new partition
     * is added and old partitions could be dropped.See <a href="http://dev.mysql.com/doc/refman/5.1/en/error-messages-server.html>
     * MySQL Server Error Codes and Messages</a>
     *
     * @param samples the samples to be written to the database
     * @throws TableException
     */
    public void insert(Collection<Sample<C>> samples) throws TableException {
        try {
            super.insert(samples);
        } catch (TableException e) {
            if (e.getCause() instanceof SQLException) {
                SQLException sqlEx = (SQLException) e.getCause();
                if (sqlEx.getErrorCode() == 1526 && sqlEx.getSQLState().equalsIgnoreCase("HY000")) {
                    rollOverPartitions(samples.iterator().next().getTime());
                    super.insert(samples);
                    return;
                }

                // fall through for other SQLException, like primary key violation
            }

            throw e; // rethrow the exception for situation other than partition rollover
        }
    }

    private void rollOverPartitions(long sampleTime) throws TableException {
        LinkedList<String> partitionNames = findPartitionsByOrdinalOrder();
        if (partitionNames.size() < maxPartitions) {
            addNewPartition(sampleTime, getNewPartitionName(partitionNames));
        } else {
            while (partitionNames.size() > maxPartitions) {
                String droppedPartitionName = dropPartition(partitionNames.getFirst());
                partitionNames.remove(droppedPartitionName);
            }
            
            addNewPartition(sampleTime, getNewPartitionName(partitionNames));
        }

    }

    private String dropPartition(String droppedPartitionName) throws TableException {
        final String partitionDropSql = "ALTER TABLE " + getFullName() + " DROP PARTITION " + droppedPartitionName;
        getPrepareStatementEx(partitionDropSql).executeUpdateEx();
        return droppedPartitionName;
    }

    private void addNewPartition(long sampleTime, String newPartitionName) throws TableException {
        final String partitionAddSql = "ALTER TABLE " + getFullName() + " ADD PARTITION (" + SEPARATOR +
                "PARTITION " + newPartitionName + " VALUES LESS THAN (" + PerformanceUtils.getMidnightTime(sampleTime) + "))";
        getPrepareStatementEx(partitionAddSql).executeUpdateEx();
    }

    public LinkedList<String> findPartitionsByOrdinalOrder() throws TableException {
        final String partitionSelectSql =
                "SELECT partition_name" + SEPARATOR +
                        "FROM information_schema.partitions" + SEPARATOR +
                        "WHERE table_name=?" + SEPARATOR +
                        " AND table_schema=?" + SEPARATOR +
                        "ORDER BY partition_ordinal_position";
        String schemaName = PrefixConnection.translateDbPrefix(getDatabase());

        return (LinkedList<String>) getPrepareStatementEx(partitionSelectSql, STRING).setNext(getName()).setNext(schemaName).executeQueryEx();
    }

    private String getNewPartitionName(final List<String> partitionNames) {
        for (Long ordinal : allPartitionOrdinals) {
            String newPartitionName = "p" + ordinal;
            if (!partitionNames.contains(newPartitionName)) {
                return newPartitionName;
            }
        }

        throw new IllegalArgumentException("Incorrect partitions: " + partitionNames + ", ordinals: " + allPartitionOrdinals);
    }

    public String getCreateSql() {
        return super.getCreateSql() + SEPARATOR +
                " PARTITION BY RANGE(time) (" + SEPARATOR +
                "   PARTITION p0 VALUES LESS THAN (" + PerformanceUtils.getMidnightTime() + ")" + SEPARATOR +
                ")";
    }

}