/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */
package com.netapp.oci.platform.framework.performance.tables;

import com.netapp.oci.platform.common.param.PluginParam;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

/**
 * @author Anton Sergeev
 */
public class PerformanceTableUtils {
    /** Default number of days to store sample data */
    public static final int DEFAULT_SAMPLE_TABLES_LIFESPAN_DAYS = 30;
    
    /** Default number of days to store sample raw data */
    public static final int DEFAULT_SAMPLE_RAW_TABLES_LIFESPAN_DAYS = 30;
    
    /** Default number of days to store summary data */
    public static final int DEFAULT_SUMMARY_TABLES_LIFESPAN_DAYS = 390;
    
    /** Default number of days to store update history data */
    public static final int DEFAULT_UPDATE_HISTORY_LIFESPAN_DAYS = 7;

    private static final String PERFORMANCE_UPDATE_CURRENT_SAMPLE = "PERFORMANCE_UPDATE_CURRENT_SAMPLE";

    private static final boolean DEFAULT_PERFORMANCE_UPDATE_CURRENT_SAMPLE = true;

    /**
     * Double null value
     */
    public static final double DOUBLE_NULL = -1;

    /**
     * Float null value
     */
    public static final float FLOAT_NULL = -1;

    /**
     * Int null value
     */
    public static final int INT_NULL = -1;

    /**
     * Short null value, short is used for storing percentage values so -1 after toShort() becomes -100
     */
    public static final int SHORT_NULL = -100;

    public static double getDoubleNullable(ResultSet resultSet, int fieldIndex) throws SQLException {
        double doubleValue = resultSet.getDouble(fieldIndex);
        if (resultSet.wasNull()) {
            doubleValue = DOUBLE_NULL;
        }

        return doubleValue;        
    }

    public static void setNextNullable(PreparedStatementWrapper preparedStatement, double doubleValue) throws TableException {
        if (doubleValue == DOUBLE_NULL) {
            preparedStatement.setNextNull(Types.DOUBLE);
        } else {
            preparedStatement.setNext(doubleValue);
        }
    }

    public static int getIntNullable(ResultSet resultSet, int fieldIndex) throws SQLException {
        int intValue = resultSet.getInt(fieldIndex);
        if (resultSet.wasNull()) {
            intValue = INT_NULL;
        }

        return intValue;        
    }

    public static float getFloatNullable(ResultSet resultSet, int fieldIndex) throws SQLException {
        float floatValue = resultSet.getInt(fieldIndex);
        if (resultSet.wasNull()) {
            floatValue = FLOAT_NULL;
        }

        return floatValue;
    }

    public static short getShortNullable(ResultSet resultSet, int fieldIndex) throws SQLException {
        short shortValue = resultSet.getShort(fieldIndex);
        if (resultSet.wasNull()) {
            shortValue = SHORT_NULL;
        }

        return shortValue;
    }

    public static void setNextNullable(PreparedStatementWrapper preparedStatement, short shortValue) throws TableException {
        if (shortValue == SHORT_NULL) {
            preparedStatement.setNextNull(Types.INTEGER);
        } else {
            preparedStatement.setNext(shortValue);
        }
    }

    public static void setNextNullable(PreparedStatementWrapper preparedStatement, int intValue) throws TableException {
        if (intValue == INT_NULL) {
            preparedStatement.setNextNull(Types.INTEGER);
        } else {
            preparedStatement.setNext(intValue);
        }
    }

    public static void setNextNullable(PreparedStatementWrapper preparedStatement, float floatValue) throws TableException {
        if (floatValue == FLOAT_NULL) {
            preparedStatement.setNextNull(Types.FLOAT);
        } else {
            preparedStatement.setNext(floatValue);
        }
    }

    public static boolean getUpdateCurrentSample(PluginParam param) {
        return param.get(PERFORMANCE_UPDATE_CURRENT_SAMPLE, DEFAULT_PERFORMANCE_UPDATE_CURRENT_SAMPLE);
    }

    /**
     * Converts percentage value to DB safe percentage value.
     * Pretty much if percentage is 100% we convert it to 99.999% to fit database
     * data type of double(5,3)
     * Also we ceil values like 0.0001234 to 0.001 so that they don't end up being zeros in database
     *
     * @param percentage - original percentage value
     * @return safe percentage value
     */
    public static double toSafePercentage(double percentage) {
        if (percentage > 99.999) {
            return 99.999;
        }

        return Math.ceil(percentage * 1000) / 1000;
    }
}
