/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */
package com.netapp.oci.platform.framework.performance.tables;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.table.ObjectParser;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.SimpleTypeQuery;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParser;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.netapp.oci.platform.common.util.Assert;
import com.netapp.oci.platform.framework.performance.data.Sample;

public class SampleTable<C> extends AbstractConfigurableTable<Sample<C>> {
    private final static Logger LOG = LoggerFactory.getLogger(SampleTable.class);
    
    public static final int BATCH_SIZE = 8000;
    
    public static String TABLE_NAME_PREFIX = "sample_";

    protected static Column[] SAMPLE_TABLE_COLUMNS = {
            new Column("objid", "bigint(20)"),
            new Column("time", "bigint(20)"),
            new Column("empty", "bool"),
    };

    protected static Index[] SAMPLE_TABLE_INDEXES = {
            new Index(true, false, SAMPLE_TABLE_COLUMNS[0], SAMPLE_TABLE_COLUMNS[1]), // primary key
            new Index(SAMPLE_TABLE_COLUMNS[1])
    };

    private ObjectParser<C> sampleDataParser;

    public SampleTable(Connection connection, ObjectParser<C> sampleDataParser, String suffixName, Column[] columns) {
        super(connection, "netap_performance", TABLE_NAME_PREFIX + suffixName,
                mergeColumns(SAMPLE_TABLE_COLUMNS, columns), SAMPLE_TABLE_INDEXES);
        this.sampleDataParser = sampleDataParser;
    }

    /**
     * Persists samples into the database.
     *
     * @param samples the samples to be written to the database
     * @throws TableException if failed to add the samples
     */
    @Override
    public void insert(Collection<Sample<C>> samples) throws TableException {
        PreparedStatementWrapper<Sample<C>> preparedStatement = getPrepareStatementEx(getInsertSql());
   
        try {
            int numOfSamples = 0;
            for (Sample<C> sample : samples) {
                ++numOfSamples;
                preparedStatement.setNext(sample.getId());
                preparedStatement.setNext(sample.getTime());

                if (!sample.isNull()) {
                    preparedStatement.setNext(false);
                }
                else {
                    preparedStatement.setNext(true);
                }
                int numberOfSystemColumns = prepareSystemDataForInsert(preparedStatement);

                if (!sample.isNull()) {
                    prepareSampleDataForInsert(preparedStatement, sample.getSampleData());
                }
                else {
                    Column[] columns = getColumns();
                    for (int i = numberOfSystemColumns + 3; i < columns.length; i++) {
                        preparedStatement.setNextNull(columns[i].getSqlType());
                    }
                }
                preparedStatement.addBatch();
                
                if (numOfSamples % BATCH_SIZE == 0) {
                    preparedStatement.executeBatch();
                }
            }
            
            // Insert remaining samples
            if (numOfSamples % BATCH_SIZE != 0) {
                preparedStatement.executeBatch();
            }
        }
        catch (SQLException e) {
            throw new TableException("Failed to executeUpdate: " + this, e);
        }
        finally {
            preparedStatement.closePreparedStatement();
        }
    }

    /**
     * Prepares system columns for insert if there are any and returns the number of columns prepared
     *
     * @param wrapper prepared statement wrapper
     * @return number of columns prepared
     * @throws TableException if failed to prepare any of the columns
     */
    protected int prepareSystemDataForInsert(PreparedStatementWrapper wrapper) throws TableException {
        return 0;
    }

    protected void prepareSampleDataForInsert(PreparedStatementWrapper wrapper, C sampleData) throws TableException {
        sampleDataParser.prepareObjectDataForInsert(wrapper, sampleData);
    }

    @Override
    public Sample<C> parseObject(ResultSet resultSet) throws SQLException {
        long id = resultSet.getLong(1);
        long time = resultSet.getLong(2);
        boolean empty = resultSet.getBoolean(3);

        return new Sample<C>(Long.valueOf(id), time, empty ? null : parseSampleData(resultSet, 4));
    }

    protected C parseSampleData(ResultSet resultSet, int startIndex) throws SQLException {
        return sampleDataParser.parseObjectData(resultSet, startIndex);
    }

    /**
     * Creates the database table needed for keeping samples. The columns of this table depends on the
     * associated with the sample and are determined by child-classes.
     *
     * @throws TableException if creation failed
     */
    @Override
    public void createTable() throws TableException {
        getPrepareStatementEx(getCreateSql()).executeUpdateEx();
    }

    public void createCopyTable() throws TableException {
        getPrepareStatementEx(getCreateTempTableSql()).executeUpdateEx();
    }

    public void dropCopyTable() throws TableException {
        getPrepareStatementEx(getDropTempTableSql()).executeUpdateEx();
    }

    public Column[] getSampleColumns() {
        Column[] columns = getColumns();
        return Arrays.copyOfRange(columns, 3, columns.length);
    }

    /**
     * Gets the samples recorded for a single object (i.e. switch port) over a period of time.
     *
     * @param timeSpan the period in which samples sould be retrieved
     * @param id       identifies the object for which to retrieve samples
     * @return samples ordered by time ascending
     * @throws TableException if failed to retrieve the samples
     */
    public List<Sample<C>> getSamples(TimeSpan timeSpan, Long id) throws TableException {
        PreparedStatementWrapper<Sample<C>> preparedStatement;
        if (timeSpan.isAll()) {
            preparedStatement = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE objId = ? ORDER BY time").
                    setNext(id);
        } else {
            preparedStatement = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE objId = ? and time>=? AND time<=? ORDER BY time").
                    setNext(id).
                    setNext(timeSpan.getFromTime()).
                    setNext(timeSpan.getToTime());
        }

        return preparedStatement.executeQueryEx();
    }


    /**
     * Copies data from current table to new table based on time
     *
     * @param timeSpan      - reflects how much data should be copied
     * @param fullTableName - full table name of the table to copy to
     * @throws TableException if failed to copy
     */
    public void copySamples(TimeSpan timeSpan, String fullTableName) throws TableException {
        if(LOG.isDebugEnabled()) {
            LOG.debug("(copySamples) timeSpan: " + timeSpan);
        }
        
        PreparedStatementWrapper<Sample<C>> preparedStatement = getPrepareStatementEx("INSERT INTO " + fullTableName + " SELECT * FROM " + getFullName() + " FORCE INDEX (time_index) WHERE time >= ? AND time <= ?");
        preparedStatement.setNext(timeSpan.getFromTime());
        preparedStatement.setNext(timeSpan.getToTime());
        preparedStatement.executeUpdateEx();
    }

    /**
     * Gets the samples recorded for multiple objects (i.e. switch ports) over a period of time.
     *
     * @param timeSpan the period in which samples sould be retrieved
     * @param ids      identifies the objects for which to retrieve samples
     * @return samples ordered by ID then by time ascending
     * @throws com.netapp.oci.platform.common.table.TableException
     *          if failed to retrieve the samples
     */
    public List<Sample<C>> getSamples(TimeSpan timeSpan, Collection<Long> ids) throws TableException {
        String idTableName = getFullName() + "_ids";
        createIdsTemporaryTableForJoin(idTableName, ids);

        PreparedStatementWrapper<Sample<C>> preparedStatement;
        if (timeSpan.isAll()) {
            preparedStatement = getPrepareStatementEx("SELECT s.* FROM " + getFullName() + " s, " + idTableName + " i WHERE s.objId = i.objid ORDER BY s.objid, s.time");
        } else {
            preparedStatement = getPrepareStatementEx("SELECT s.* FROM " + getFullName() + " s, " + idTableName + " i WHERE s.objId = i.objid and time>=? AND time<=? ORDER BY s.objid, s.time").
                    setNext(timeSpan.getFromTime()).
                    setNext(timeSpan.getToTime());
        }

        return preparedStatement.executeQueryEx();
    }

    /**
     * Retrieve the latest sample for the given object ids.
     *
     * @param ids identifies the objects for which to retrieve samples
     * @return Unordered samples
     * @throws com.netapp.oci.platform.common.table.TableException
     *          if failed to retrieve the samples
     */
    public List<Sample<C>> getLatestSample(Collection<Long> ids) throws TableException {
        PreparedStatementWrapper<Sample<C>> preparedStatement;
        if (ids.size() == 1) {
            // Query for one object sample, not need of temp table
            preparedStatement = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE objId = ? ORDER BY objid desc, time desc LIMIT 1").setNext(ids.iterator().next());
        } else {
            String idTableName = getFullName() + "_ids";
            createIdsTemporaryTableForJoin(idTableName, ids);
            String sql = "SELECT \n" +
                    "    s1 . *\n" +
                    "FROM\n" +
                    "    " + getFullName() + " s1\n" +
                    "        INNER JOIN\n" +
                    "    (SELECT \n" +
                    "        objid, max(time) time\n" +
                    "    FROM\n" +
                    "        " + getFullName() + "\n" +
                    "    INNER JOIN " + idTableName + " USING (objid)\n" +
                    "    GROUP BY objid) s2 USING (objid , time)";
            preparedStatement = getPrepareStatementEx(sql);
        }
        return preparedStatement.executeQueryEx();
    }

    /**
     * Gets the period for which there are samples in the table. This is the minimum and maximum value of the time column
     * of all the rows in the table.
     *
     * @return the period for which there are samples in the table
     * @throws com.netapp.oci.platform.common.table.TableException
     *          if failed to query the DB
     */
    public TimeSpan getTableTimeSpan() throws TableException {
        List<TimeSpan> timeSpans = getPrepareStatementEx("SELECT MAX(time), MIN(time) FROM " + getFullName(), new TableResultParser<TimeSpan>() {
            public List<TimeSpan> parseResult(ResultSet resultSet) throws SQLException {
                if (resultSet.next()) {
                    long max = resultSet.getLong(1);
                    if (!resultSet.wasNull()) {
                        return Collections.singletonList(new TimeSpan(resultSet.getLong(2), max));
                    }
                }
                return Collections.emptyList();
            }
        }).executeQueryEx();

        if (timeSpans.size() > 0) {
            return timeSpans.get(0);
        }

        return null;
    }

    /**
     * Get samples for a specific period ordered by time. The samples are retrieved in "chunks" to limit the memory used
     * during the processing. Samples are retreived only when the "next" method is invoked.
     *
     * @param timeSpan  limits samples to a specific period
     * @param chunkSize - size of the chunks
     * @return an iterator of lists of samples
     * @throws com.netapp.oci.platform.common.table.TableException
     *          if query failed
     */
    public Iterator<List<Sample<C>>> getSampleChunks(TimeSpan timeSpan, int chunkSize) throws TableException {
        return new SampleChunkIterator(timeSpan, getIds(timeSpan), chunkSize);
    }

    /**
     * Get samples for a specific period ordered by time. The samples are retrieved in "chunks" to limit the memory used
     * during the processing and are limited to the requested IDs. Samples are retreived only when the "next" method is invoked.
     *
     * @param timeSpan  limits samples to a specific period
     * @param ids       the IDs of the objects for which samples are requested, never <code>null</code>
     * @param chunkSize - size of the chunks
     * @return an iterator of lists of samples
     * @throws TableException if query failed
     */
    public Iterator<List<Sample<C>>> getSampleChunks(TimeSpan timeSpan, Collection<Long> ids, int chunkSize) throws TableException {
        Assert.isNotNull(ids);
        return new SampleChunkIterator(timeSpan, ids, chunkSize);
    }

    /**
     * Delete all the samples which were recorded prior to the specified time.
     *
     * @param beforeTime earliest sample time to keep in the DB
     * @throws TableException if deletion failed
     */
    public void deleteOldestSamples(long beforeTime) throws TableException {
        getPrepareStatementEx("DELETE FROM " + getFullName() + " WHERE time <= ?").setNext(beforeTime).executeUpdateEx();
    }

    /**
     * Loads all the IDs of objects for which there are samples in the specified time-span
     *
     * @param timeSpan the time in which to look for IDs
     * @return list of IDs of objects having samples in the specified period
     * @throws com.netapp.oci.platform.common.table.TableException
     *          if failed to query for the IDs
     */
    public Collection<Long> getIds(TimeSpan timeSpan) throws TableException {
        SimpleTypeQuery<Long> query = new SimpleTypeQuery<Long>(getConnection());
        PreparedStatementWrapper<Long> preparedStatement;
        if (timeSpan.isAll()) {
            preparedStatement = query.getPrepareStatementEx("SELECT DISTINCT objid FROM " + getFullName());
        } else {
            preparedStatement = query.getPrepareStatementEx("SELECT DISTINCT objid FROM " + getFullName() + " WHERE time >= ? AND time <= ?");
            preparedStatement.setNext(timeSpan.getFromTime());
            preparedStatement.setNext(timeSpan.getToTime());
        }
        return preparedStatement.executeQueryEx();
    }

    /**
     * For selected objIds, gets a merge sample of the entire requested period, grouped by time (NOT by objId)
     *
     * @param timeSpan the period for which to get the summary
     * @param ids      the selected summary objects
     * @param methods  sql methods to apply to column when selecting summaries
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws TableException if any of the queries failed
     */
    public List<Sample<C>> getMergedSamplesGroupByTime(TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws TableException {
        Assert.isNotNull(ids);
        final PreparedStatementWrapper<Sample<C>> prepareStatement = getPrepareStatementEx(getMergedSamplesGroupByTimeSql(ids, methods, timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }
        return prepareStatement.executeQueryEx();
    }

    /**
     * gets a merge sample of the entire requested period, grouped by time (NOT by objId)
     *
     * @param timeSpan the period for which to get the summary
     * @param methods  sql methods to apply to column when selecting summaries
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws TableException if any of the queries failed
     */
    public List<Sample<C>> getMergedSamplesGroupByTime(TimeSpan timeSpan, SqlMethod[] methods) throws TableException {
        final PreparedStatementWrapper<Sample<C>> prepareStatement = getPrepareStatementEx(getMergedSamplesGroupByTimeSql(methods, timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }
        return prepareStatement.executeQueryEx();
    }

    private String getMergedSamplesGroupByTimeSql(SqlMethod[] methods, TimeSpan timeSpan) {
        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n" +
                "   -1,\r\n" +
                "   time,\r\n" +
                "s.empty");
        sql.append(SqlMethod.getWrappedColumns("s.", getSampleColumns(), methods));
        sql.append("\r\n");
        sql.append(" FROM ").append(getFullName()).append(" s");
        sql.append("\r\n");
        if (!timeSpan.isAll()) {
            sql.append(" WHERE s.time>=? AND s.time<? \r\n");
        }
        sql.append(" GROUP BY s.time");
        return sql.toString();
    }

    private String getMergedSamplesGroupByTimeSql(Collection<Long> ids, SqlMethod[] methods, TimeSpan timeSpan) throws TableException {
        String idTableName = getFullName() + "_ids";
        createIdsTemporaryTableForJoin(idTableName, ids);

        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n" +
                "   -1,\r\n" +
                "   s.time,\r\n" +
                "s.empty");
        sql.append(SqlMethod.getWrappedColumns("s.", getSampleColumns(), methods));
        sql.append("\r\n");
        sql.append("FROM ").append(getFullName()).append(" s, ").append(idTableName).append(" i ");
        sql.append("\r\n");
        sql.append(" WHERE s.objId = i.objid ");
        if (!timeSpan.isAll()) {
            sql.append(" AND s.time>=? AND s.time<? \r\n");
        }
        sql.append(" GROUP BY s.time");
        return sql.toString();
    }

    /**
     * returns true if the samples exist over a period of time.
     *
     * @param timeSpan the period in which samples sould be retrieved
     * @param id
     * @return samples ordered by ID then by time ascending
     * @throws com.netapp.oci.platform.common.table.TableException
     *          if failed to retrieve the samples
     */
    public Boolean doSamplesExist(TimeSpan timeSpan, Long id) throws TableException {

        PreparedStatementWrapper<Boolean> preparedStatement;
        preparedStatement = getPrepareStatementEx("SELECT EXISTS(SELECT 1 FROM " + getFullName() + " s  WHERE s.objId = ? and time>=? AND time<=?)", TableResultParserDefault.BOOLEAN).
                setNext(id).
                setNext(timeSpan.getFromTime()).
                setNext(timeSpan.getToTime());

        return preparedStatement.executeQuerySingleResult();

    }

    /**
     * returns first sample after timestamp
     *
     * @param id
     * @param afterTime
     * @return one sample
     * @throws TableException
     */
    public Sample<C> getFirstSampleAfterTime(Long id, long afterTime) throws TableException {
        PreparedStatementWrapper<Sample<C>> preparedStatement;
        preparedStatement = getPrepareStatementEx("SELECT * FROM " + getFullName() + " s WHERE s.objId = ? and time > ? and empty=0 order by time limit 1")
                .setNext(id)
                .setNext(afterTime);
        List<Sample<C>> result = preparedStatement.executeQueryEx();
        if(result!=null && !result.isEmpty()) {
            return result.get(0);
        }
        else {
            return null;
        }
    }

    /**
     * return last sample before timestamp
     *
     * @param id
     * @param beforeTime
     * @return oneSample
     * @throws TableException
     */
    public Sample<C> getLastSampleBeforeTime(Long id, long beforeTime) throws TableException {
        PreparedStatementWrapper<Sample<C>> preparedStatement;
        preparedStatement = getPrepareStatementEx("SELECT * FROM " + getFullName() + " s WHERE s.objId = ? and time < ? and empty=0 order by time desc limit 1")
                .setNext(id)
                .setNext(beforeTime);
        List<Sample<C>> result = preparedStatement.executeQueryEx();
        if(result!=null && !result.isEmpty()) {
            return result.get(0);
        }
        else {
            return null;
        }
    }

    private class SampleChunkIterator implements Iterator<List<Sample<C>>> {
        private TimeSpan timeSpan;

        private Iterator<Collection<Long>> idChunkIterator;

        public SampleChunkIterator(TimeSpan timeSpan, Collection<Long> ids, int chunkSize) throws TableException {
            this.timeSpan = timeSpan;

            if (timeSpan == null) {
                return;
            }

            if (ids != null) {
                idChunkIterator = splitCollection(ids, chunkSize).iterator();
            }
        }

        /**
         * Split collection of values into collection of chunks of values
         *
         * @param values - collection of values
         * @param size   - siz eof each chunk
         * @return collection of collections of values
         */
        public <T> Collection<Collection<T>> splitCollection(Collection<T> values, int size) {
            Collection<T> chunk = null;
            Collection<Collection<T>> chunks = new LinkedList<Collection<T>>();


            for (T value : values) {
                if (chunk == null) {
                    chunk = new LinkedList<T>();
                }

                chunk.add(value);

                if (chunk.size() == size) {
                    chunks.add(chunk);
                    chunk = null;
                }
            }

            if (chunk != null && chunk.size() > 0) {
                chunks.add(chunk);
            }

            return chunks;
        }

        public boolean hasNext() {
            return idChunkIterator != null && idChunkIterator.hasNext();
        }

        public List<Sample<C>> next() {
            try {
                return getSamples(timeSpan, idChunkIterator.next());
            } catch (TableException e) {
                NoSuchElementException noSuchElementException = new NoSuchElementException("Failed to get next chunk of samples");
                noSuchElementException.initCause(e);
                throw noSuchElementException;
            }
        }

        public void remove() {
            idChunkIterator.remove();
        }
    }
}
