package com.netapp.oci.platform.framework.performance.tables;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import com.onaro.commons.types.Period;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.util.Assert;
import com.netapp.oci.platform.framework.performance.data.Summary;

public abstract class SummaryTable<S> extends AbstractConfigurableTable<Summary<S>> {
    public static String TABLE_NAME_PREFIX = "summary_";

    private static Column[] SUMMARY_TABLE_COLUMNS = {
            new Column("objid", "bigint(20)"),
            new Column("fromtime", "bigint(20)"),
    };

    private static Index[] SUMMARY_TABLE_INDEXES = {
            new Index(true, false, SUMMARY_TABLE_COLUMNS[0], SUMMARY_TABLE_COLUMNS[1]), // primary key
            new Index(SUMMARY_TABLE_COLUMNS[1])
    };

    public SummaryTable(Connection connection, String suffixName, Column[] columns) {
        super(connection, "netapp_performance", TABLE_NAME_PREFIX + suffixName,
                mergeColumns(SUMMARY_TABLE_COLUMNS, columns), SUMMARY_TABLE_INDEXES);
    }


    /**
     * Saves to the DB the collection of summary objects
     *
     * @param summaries - collection of summary objects
     * @throws TableException if failed to summarize the samples or insert the rows to the DB
     */
    public void insert(Collection<Summary<S>> summaries) throws TableException {
        PreparedStatementWrapper<Summary<S>> preparedStatement = getPrepareStatementEx(getInsertSql());

        for (Summary<S> summary : summaries) {
            preparedStatement.setNext(summary.getId());
            preparedStatement.setNext(summary.getTimeSpan().getFromTime());

            prepareSummaryDataForInsert(preparedStatement, summary.getSummaryData());
            preparedStatement.addBatchEx();
        }

        preparedStatement.executeBatchEx();
    }

    protected abstract void prepareSummaryDataForInsert(PreparedStatementWrapper preparedStatement, S summaryData) throws TableException;

    public Summary<S> parseObject(ResultSet resultSet) throws SQLException {
        return new Summary<S>(
                resultSet.getLong(1),
                new TimeSpan(resultSet.getLong(2), resultSet.getLong(2)+Period.HOUR.milliseconds()-1),
                parseSummaryData(resultSet, 3));
    }

    protected abstract S parseSummaryData(ResultSet resultSet, int startIndex) throws SQLException;

    public TimeSpan getTableTimeSpan() throws TableException {
        long min = getPrepareStatementEx("SELECT MIN(fromtime) FROM " + getFullName()).executeFunctionQueryEx(-1);
        long max = getPrepareStatementEx("SELECT MAX(fromtime) FROM " + getFullName()).executeFunctionQueryEx(-1);
        if (min <= 0 && max <= 0)
            return null;

        max = max + Period.HOUR.milliseconds() - 1;
        return new TimeSpan(min, max);
    }

    /**
     * For each requested object (id) for which samples were recorded during the specified time, get the summary of all
     * the samples.
     *
     * @param timeSpan defines the start and end time for which to get the sumary
     * @param ids      IDs of objects for which summaries are requested
     * @return list containing, for each object, a single summary ({@link S}) which represents the summary of all the
     *         samples of this object in the specified period
     * @throws TableException if failed to fetch the summaries from the DB
     */
    public List<Summary<S>> getSummaries(TimeSpan timeSpan, Collection<Long> ids) throws TableException {
        Assert.isNotNull(ids);
        final PreparedStatementWrapper<Summary<S>> prepareStatement = getPrepareStatementEx(getSelectSql(ids, timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }

        return prepareStatement.executeQueryEx();
    }

    /**
     * get the summary of all obj ids
     *
     * @param timeSpan defines the start and end time for which to get the sumary
     * @return list containing, for each object, a single summary ({@link S}) which represents the summary of all the
     *         samples of this object in the specified period
     * @throws TableException if failed to fetch the summaries from the DB
     */
    public List<Summary<S>> getSummaries(TimeSpan timeSpan) throws TableException {
        final PreparedStatementWrapper<Summary<S>> prepareStatement = getPrepareStatementEx(getSelectSql(timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }

        return prepareStatement.executeQueryEx();
    }


    /**
     * Gets a merge summary of the entire requested period for all objects.
     * The summary is calculated from samples not yet processed and from summaries of older samples.
     *
     * @param timeSpan the period for which to get the summary
     * @param methods  sql methods to apply to column when selecting summaries
     * @return a summary of the samples recorded for ALL objects during the requested period.
     * @throws TableException if any of the queries failed
     */
    public S getMergedSummaries(TimeSpan timeSpan, SqlMethod[] methods) throws TableException {
        final PreparedStatementWrapper<Summary<S>> prepareStatement = getPrepareStatementEx(getMergedSummariesSql(methods, timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }
        List<Summary<S>> summaries = prepareStatement.executeQueryEx();
        if (summaries.size() == 0)
            return null;
        return summaries.get(0).getSummaryData();
    }

    /**
     * gets a merge summary of the entire requested period (that is, a single summary
     * per object). The summary is calculated from samples not yet processed and from summaries of older samples.
     *
     * @param timeSpan the period for which to get the summary
     * @param ids      the seslected summary objects
     * @param methods  sql methods to apply to column when selecting summaries
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws TableException if any of the queries failed
     */
    public List<Summary<S>> getMergedSummariesGroupById(TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws TableException {
        Assert.isNotNull(ids);
        final PreparedStatementWrapper<Summary<S>> prepareStatement = getPrepareStatementEx(getMergedSummariesGroupByIdSql(ids, methods, timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }

        return prepareStatement.executeQueryEx();
    }

    /**
     * For each object (i.e. port), gets a merge summary of samples of the entire requested period (that is, a single summary
     * per object). The summary is calculated from samples not yet processed and from summaries of older samples.
     *
     * @param timeSpan the period for which to get the summary
     * @param methods  sql methods to apply to column when selecting summaries
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws TableException if any of the queries failed
     */
    public List<Summary<S>> getMergedSummariesGroupById(TimeSpan timeSpan, SqlMethod[] methods) throws TableException {
        final PreparedStatementWrapper<Summary<S>> prepareStatement = getPrepareStatementEx(getMergedSummariesGroupByIdSql(methods, timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }

        return prepareStatement.executeQueryEx();
    }
    
    /**
     * Get a list of sample summaries for the entire requested period.
     *
     * @param timeSpan the period for which to get the summary
     * @return a list of sample summaries for the requested period.
     * @throws TableException if any of the queries failed
     */
    public List<Summary<S>> getSummariesGroupById(TimeSpan timeSpan) throws TableException {
        final PreparedStatementWrapper<Summary<S>> prepareStatement = getPrepareStatementEx(getSummariesGroupByIdSql(timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }

        return prepareStatement.executeQueryEx();
    }

    /**
     * Get a list of sample summaries for the given ids and entire requested period.
     *
     * @param timeSpan the period for which to get the summary
     * @param ids  a list of sample object ids
     * @return a list of the sample summaries for the given ids in the requested period.
     * @throws TableException if any of the queries failed
     */
    public List<Summary<S>> getSummariesGroupById(TimeSpan timeSpan, Collection<Long> ids) throws TableException {
        Assert.isNotNull(ids);
        final PreparedStatementWrapper<Summary<S>> prepareStatement = getPrepareStatementEx(getSummariesGroupByIdSql(ids, timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }

        return prepareStatement.executeQueryEx();
    }

    /**
     * get a hourly patern for a given time period
     *
     * @param timeSpan the period for which to get the summary
     * @param methods  sql methods to apply to column when selecting summaries
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws TableException if any of the queries failed
     */
    public List<Summary<S>> getHourlyPattern(TimeSpan timeSpan, SqlMethod[] methods) throws TableException {
        final PreparedStatementWrapper<Summary<S>> prepareStatement = getPrepareStatementEx(getHourlyPatternSql(methods, timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }
        return prepareStatement.executeQueryEx();
    }

    /**
     * get a hourly patern for a given time period for aselected ids
     *
     * @param timeSpan the period for which to get the summary
     * @param ids      the seslected summary objects
     * @param methods  sql methods to apply to column when selecting summaries
     * @return a summary of the samples recorded for each object during the requested period.
     * @throws com.netapp.oci.platform.common.table.TableException if any of the queries failed
     */
    public List<Summary<S>> getHourlyPattern(TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws TableException {
        Assert.isNotNull(ids);
        final PreparedStatementWrapper<Summary<S>> prepareStatement = getPrepareStatementEx(getHourlyPatternSql(ids, methods, timeSpan));
        if (!timeSpan.isAll()) {
            prepareStatement.setNext(timeSpan.getFromTime()).setNext(timeSpan.getToTime());
        }
        return prepareStatement.executeQueryEx();
    }

    public Column[] getSummaryColumns() {
        Column[] columns = getColumns();
        return Arrays.copyOfRange(columns, 3, columns.length);
    }

    /**
     * returns first sample after timestamp
     *
     * @param id
     * @param afterTime
     * @return one sample
     * @throws TableException
     */
    public Summary<S> getFirstSummaryAfterTime(Long id, long afterTime) throws TableException {
        PreparedStatementWrapper<Summary<S>> preparedStatement;
        preparedStatement = getPrepareStatementEx("SELECT * FROM " + getFullName() + " s WHERE s.objId = ? and time > ? and empty=0 order by time limit 1")
                .setNext(id)
                .setNext(afterTime);
        List<Summary<S>> result = preparedStatement.executeQueryEx();
        if(result!=null && !result.isEmpty()) {
            return result.get(0);
        }
        else {
            return null;
        }
    }

    /**
     * return last sample before timestamp
     *
     * @param id
     * @param beforeTime
     * @return oneSample
     * @throws TableException
     */
    public Summary<S> getLastSummaryBeforeTime(Long id, long beforeTime) throws TableException {
        PreparedStatementWrapper<Summary<S>> preparedStatement;
        preparedStatement = getPrepareStatementEx("SELECT * FROM " + getFullName() + "  s WHERE s.objId = ? and time < ? and empty=0 order by time desc limit 1")
                .setNext(id)
                .setNext(beforeTime);
        List<Summary<S>> result = preparedStatement.executeQueryEx();
        if(result!=null&& !result.isEmpty()) {
            return result.get(0);
        }
        else {
            return null;
        }
    }

    /**
     * @return the select all sql
     */
    private String getSelectSql(TimeSpan timeSpan) {
        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n");
        int i = 0;
        for (Column column : getColumns()) {
            if (i != 0) {
                sql.append(",\r\n");
            }
            i++;
            sql.append("   s.").append(column.getName());
        }
        sql.append("\r\n");
        sql.append(" FROM ").append(getFullName()).append(" s");
        sql.append("\r\n");
        if (!timeSpan.isAll())
            sql.append(" WHERE s.fromtime>=? AND s.fromtime<=?");
        return sql.toString();
    }

    /**
     * @param ids the ids is insert into temporary table (used for join)
     * @return return the select by ids sql for a given ids
     * @throws com.netapp.oci.platform.common.table.TableException - failed to create / fill the temporary ids table
     */
    private String getSelectSql(Collection<Long> ids, TimeSpan timeSpan) throws TableException {
        String idTableName = getFullName() + "_ids";
        createIdsTemporaryTableForJoin(idTableName, ids);
        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n");
        int i = 0;
        for (Column column : getColumns()) {
            if (i != 0) {
                sql.append(",\r\n");
            }
            i++;
            sql.append("   s.").append(column.getName());
        }
        sql.append("\r\n");
        sql.append("FROM ").append(getFullName()).append(" s, ").append(idTableName).append(" i ");
        sql.append("\r\n");
        sql.append("WHERE s.objId = i.objid \r\n");
        if (!timeSpan.isAll())
            sql.append(" AND fromtime>=? AND fromtime<=?");
        return sql.toString();
    }


    private String getMergedSummariesSql(SqlMethod[] methods, TimeSpan timeSpan) {
        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n" +
                "   -1 ,\r\n" +
                "   min(s.fromtime),\r\n" +
                "   max(s.fromtime)");
        sql.append(SqlMethod.getWrappedColumns("s.", getSummaryColumns(), methods));
        sql.append("\r\n");
        sql.append(" FROM ").append(getFullName()).append(" s");
        sql.append("\r\n");
        if (!timeSpan.isAll())
            sql.append(" WHERE s.fromtime>=? AND s.fromtime<?");
        return sql.toString();
    }

    private String getMergedSummariesGroupByIdSql(SqlMethod[] methods, TimeSpan timeSpan) {
        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n" +
                "   s.objid,\r\n" +
                "   min(s.fromtime),\r\n" +
                "   max(s.fromtime)");
        sql.append(SqlMethod.getWrappedColumns("s.", getSummaryColumns(), methods));
        sql.append("\r\n");
        sql.append(" FROM ").append(getFullName()).append(" s");
        sql.append("\r\n");
        if (!timeSpan.isAll())
            sql.append(" WHERE s.fromtime>=? AND s.fromtime<?");
        sql.append(" GROUP BY s.objid");
        return sql.toString();
    }

    private String getMergedSummariesGroupByIdSql(Collection<Long> ids, SqlMethod[] methods, TimeSpan timeSpan) throws TableException {
        String idTableName = getFullName() + "_ids";
        createIdsTemporaryTableForJoin(idTableName, ids);

        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n" +
                "   s.objid,\r\n" +
                "   min(s.fromtime),\r\n" +
                "   max(s.fromtime)");
        sql.append(SqlMethod.getWrappedColumns("s.", getSummaryColumns(), methods));
        sql.append("\r\n");
        sql.append("FROM ").append(getFullName()).append(" s, ").append(idTableName).append(" i ");
        sql.append("\r\n");
        sql.append(" WHERE s.objId = i.objid \r\n");
        if (!timeSpan.isAll())
            sql.append(" AND s.fromtime>=? AND s.fromtime<? \r\n");
        sql.append(" GROUP BY s.objid");
        return sql.toString();
    }
    
    private String getSummariesGroupByIdSql(TimeSpan timeSpan) {
        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n" +
                "   s.objid,\r\n" +
                "   s.fromtime,\r\n" +
                "   s.fromtime");
        sql.append(getWrappedColumns("s.", getSummaryColumns()));
        sql.append("\r\n");
        sql.append(" FROM ").append(getFullName()).append(" s");
        sql.append("\r\n");
        if (!timeSpan.isAll())
            sql.append(" WHERE s.fromtime>=? AND s.fromtime<?");
        return sql.toString();
    }

    private String getSummariesGroupByIdSql(Collection<Long> ids, TimeSpan timeSpan) throws TableException {
        String idTableName = getFullName() + "_ids";
        createIdsTemporaryTableForJoin(idTableName, ids);

        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n" +
                "   s.objid,\r\n" +
                "   s.fromtime,\r\n" +
                "   s.fromtime");
        sql.append(getWrappedColumns("s.", getSummaryColumns()));
        sql.append("\r\n");
        sql.append("FROM ").append(getFullName()).append(" s, ").append(idTableName).append(" i ");
        sql.append("\r\n");
        sql.append(" WHERE s.objId = i.objid \r\n");
        if (!timeSpan.isAll())
            sql.append(" AND s.fromtime>=? AND s.fromtime<? \r\n");
        return sql.toString();
    }
    
    private String getHourlyPatternSql(SqlMethod[] methods, TimeSpan timeSpan) {
        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n" +
                "   s.objid,\r\n" +
                "   (HOUR(from_unixtime(s.fromtime/1000))) fromHour,\r\n" +
                "   (HOUR(from_unixtime(s.fromtime/1000)) + 1) toHour");
        sql.append(SqlMethod.getWrappedColumns("s.", getSummaryColumns(), methods));
        sql.append("\r\n");
        sql.append(" FROM ").append(getFullName()).append(" s");
        sql.append("\r\n");
        if (!timeSpan.isAll())
            sql.append(" WHERE s.fromtime>=? AND s.fromtime<? \r\n");
        sql.append("GROUP BY s.objid, fromHour");
        return sql.toString();
    }

    private String getHourlyPatternSql(Collection<Long> ids, SqlMethod[] methods, TimeSpan timeSpan) throws TableException {
        String idTableName = getFullName() + "_ids";
        createIdsTemporaryTableForJoin(idTableName, ids);

        StringBuilder sql = new StringBuilder();
        sql.append("SELECT \r\n" +
                "   s.objid,\r\n" +
                "   (HOUR(from_unixtime(s.fromtime/1000))) fromHour,\r\n" +
                "   (HOUR(from_unixtime(s.fromtime/1000)) + 1) toHour");
        sql.append(SqlMethod.getWrappedColumns("s.", getSummaryColumns(), methods));
        sql.append("\r\n");
        sql.append("FROM ").append(getFullName()).append(" s, ").append(idTableName).append(" i ");
        sql.append("\r\n");
        sql.append(" WHERE s.objId = i.objid  \r\n");

        if (!timeSpan.isAll())
            sql.append(" AND s.fromtime>=? AND s.fromtime<? \r\n");
        sql.append("GROUP BY s.objid, fromHour");
        return sql.toString();
    }
    
    private String getWrappedColumns(String prefix, Column[] columns) {
        StringBuilder sql = new StringBuilder();
        for (int i = 0; i < columns.length; i++) {
            Column column = columns[i];
            sql.append(",\r\n    ");
            sql.append(prefix).append(column.getName());
            sql.append(" AS ").append(column.getName());
        }
        return sql.toString();
    }


}
