/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.framework.performance.tables;

import com.onaro.commons.types.Period;

import com.netapp.oci.platform.common.table.Table;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.TimeSpan;

import java.sql.Connection;

/**
 * Abstract table class that has methods common for tables with fromtime column
 * Also utility class
 *
 * @author Anton Sergeev
 */
public abstract class TimeSpanTable<T> extends Table<T> {

    protected TimeSpanTable(Connection connection, String database, String name) {
        super(connection, database, name);
    }

    /**
     * Ensures that time span table has certain amount of data for given period of time
     * @param timeSpanTable - time span table to process
     * @param lifeSpanMillis - time period to keep data for
     * @throws TableException - in case failed to run db queries
     */
    public static void ensureLifeSpan(Table timeSpanTable, long lifeSpanMillis) throws TableException {
        TimeSpan timeSpan = getTableTimeSpan(timeSpanTable);
        if (timeSpan != null) {
            long deleteBefore = timeSpan.getToTime() - lifeSpanMillis;
            deleteBeforeTime(timeSpanTable, deleteBefore);
        }
    }

    public int deleteBeforeTime(long time) throws TableException {
        return deleteBeforeTime(this, time);
    }

    public static int deleteBeforeTime(Table timeSpanTable, long time) throws TableException {
        return timeSpanTable.getPrepareStatementEx("DELETE FROM " + timeSpanTable.getFullName() + " WHERE fromtime < ?").setNext(time).executeUpdateEx();
    }

    public TimeSpan getTableTimeSpan() throws TableException {
        return getTableTimeSpan(this);
    }

    public static TimeSpan getTableTimeSpan(Table timeSpanTable) throws TableException {
        long min = timeSpanTable.getPrepareStatementEx("SELECT MIN(fromtime) FROM " + timeSpanTable.getFullName()).executeFunctionQueryEx(-1);
        long max = timeSpanTable.getPrepareStatementEx("SELECT MAX(fromtime) FROM " + timeSpanTable.getFullName()).executeFunctionQueryEx(-1);
        if (min <= 0 && max <= 0)
            return null;

        max = max + Period.HOUR.milliseconds() - 1;
        return new TimeSpan(min, max);
    }

    public void deleteLastRows(int rows) throws TableException {
        deleteLastRows(this, rows);
    }

    public static void deleteLastRows(Table timeSpanTable, int rows) throws TableException {
        timeSpanTable.getPrepareStatementEx("DELETE FROM " + timeSpanTable.getFullName() + " ORDER BY fromtime LIMIT ?").setNext(rows).executeUpdateEx();
    }
}
