/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.framework.performance.test;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.test.AbstractTestCase4;
import com.netapp.oci.platform.framework.performance.PerformancePackageFactory;
import com.netapp.oci.platform.framework.performance.PersistenceFacade;
import com.netapp.oci.platform.framework.performance.SampleFacade;
import com.netapp.oci.platform.framework.performance.SummaryFacade;
import com.netapp.oci.platform.framework.performance.data.CurrentSampleStatus;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.tables.*;
import com.onaro.commons.types.Period;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import java.lang.reflect.Method;
import java.sql.Connection;
import java.util.*;

/**
 * Test case for any performance package. Every performance package written should extend this testcase
 * and provide factories needed
 *
 * @author Anton Sergeev
 */
public abstract class AbstractPerformancePackageTestCase<C, S> extends AbstractTestCase4 {
    protected static final long DATASOURCE_ID = 1L;

    @Override
    @Before
    public void setUp() throws Exception {
        super.setUp();

        Connection connection = getDataSource().getConnection();
        getPerformancePackageFactory(getFirstContainerKey()).getTableManagement().createTables(connection);
        connection.close();
    }


    @Override
    @After
    public void tearDown() throws Exception {
        Connection connection = getDataSource().getConnection();
        Set<String> containerKeys = new HashSet<String>(getContainerKeyToContainerId().keySet());
        containerKeys.add("");  // some junits start testing without a valid container setup, make sure we clean those tables up
        for (String containerKey : containerKeys) {
            getPerformancePackageFactory(containerKey).getTableManagement().dropTables(connection);
            CurrentSampleTable<C> currentSampleTable = getCurrentSampleTable(connection, containerKey);
            currentSampleTable.deleteAll(); // since current sample tables do not get dropped
            new CurrentSampleStatusTable(connection, currentSampleTable.getDatabase()).deleteAll();
        }
        connection.close();

        super.tearDown();
    }

    /**
     * Each performance package, for example storage performance, vm performance etc should
     * have test case with implemented method for package factory
     *
     * @return package factory for package
     */
    protected abstract PerformancePackageFactory<C, S> getPerformancePackageFactory(String containerKey);

    protected abstract SampleTable<C> getSampleTable(Connection connection, Long containerId);

    protected abstract CurrentSampleTable<C> getCurrentSampleTable(Connection connection, String containerKey);

    protected abstract SummaryTable<S> getSummaryTable(Connection connection, Long containerId);

    protected abstract SqlMethod[] getCommonSqlMethods();

    protected abstract void reportRawSamples(Long datasourceId, String containerKey, long time, Collection<RawSample<C>> rawSamples) throws Exception;

    /**
     * Provides environment for the test to generate data for.
     * Map of device keys to corresponding collection of performance object keys
     */
    protected abstract Map<String, List<String>> getContainerKeyToObjectKeysMap();

    /**
     * Resolves given performance object key into id that can be used with
     * different facades
     */
    protected abstract Map<String, Map<String, Long>> getContainerKeyToObjectKeyToId() throws Exception;

    protected abstract Map<String, Long> getContainerKeyToContainerId() throws Exception;

    /**
     * Imports ODATA that supports method #getDeviceKeyToObjectKeyMap
     * The reason we need this is because every performance package uses some sort of
     * lookup to resolve object keys against existing inventory
     */
    protected abstract void importFoundationXmlData() throws Exception;

    /**
     * Generates performance sample data for testing
     *
     * @param previousSampleData previous sample data, may be null if this is the first sample
     * @return performance sample data
     */
    protected abstract C generateNextSampleData(C previousSampleData);

    /**
     * Generates performance summary data for testing
     *
     * @param previousSummaryData previous summary data, may be null if this is the first summary
     * @return performance summary data
     */
    protected abstract S generateNextSummaryData(S previousSummaryData);

    /**
     * Generates one hour worth of samples
     *
     * @param id              - id to generate samples for
     * @param timeSpan        - time span to generate samples for
     * @param pollingInterval - interval between samples
     * @return list of samples for one calendar hour
     */
    protected List<Sample<C>> generateSamples(long id, TimeSpan timeSpan, long pollingInterval) {
        return generateSamples(Collections.singleton(id), timeSpan, pollingInterval).values().iterator().next();
    }

    /**
     * Generates time span worth of samples
     *
     * @param ids             - ids to generate samples for
     * @param timeSpan        - time span to generate samples for
     * @param pollingInterval - interval between samples
     * @return list of samples for one calendar hour
     */
    protected Map<Long, List<Sample<C>>> generateSamples(Set<Long> ids, TimeSpan timeSpan, long pollingInterval) {
        Map<Long, List<Sample<C>>> samples = new HashMap<Long, List<Sample<C>>>();

        for (Long id : ids) {
            C sampleData = null;
            for (long time = timeSpan.getFromTime(); time < timeSpan.getToTime(); time += pollingInterval) {
                sampleData = generateNextSampleData(sampleData);

                List<Sample<C>> samplesList = samples.get(id);
                if (samplesList == null) {
                    samplesList = new ArrayList<Sample<C>>();
                    samples.put(id, samplesList);
                }
                samplesList.add(new Sample<C>(id, time, sampleData));
            }
        }

        return samples;
    }

    /**
     * Generates time span object for calendar hour
     *
     * @param time - any point of time that belongs to a calendar hour that will be generated
     * @return time span equal to an hour
     */
    protected TimeSpan generateHourTimeSpan(long time) {
        Calendar calendar = new GregorianCalendar();
        calendar.setTime(new Date(time));
        calendar.set(Calendar.MINUTE, 0);
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);

        long from = calendar.getTime().getTime();

        calendar.set(Calendar.MINUTE, 59);
        calendar.set(Calendar.SECOND, 59);
        calendar.set(Calendar.MILLISECOND, 999);
        long to = calendar.getTime().getTime();

        return new TimeSpan(from, to);
    }

    protected Map<Long, List<Sample<C>>> reportPerformanceData(TimeSpan timeSpan, long pollingInterval) throws Exception {
        return reportPerformanceData(getContainerKeyToContainerId().keySet(), timeSpan, pollingInterval);
    }

    protected Map<Long, List<Sample<C>>> reportPerformanceData(Set<String> containerKeys, TimeSpan timeSpan, long pollingInterval) throws Exception {
        Map<Long, List<Sample<C>>> result = new HashMap<Long, List<Sample<C>>>();
        C sampleData = null;
        Map<String, List<String>> deviceKeyToObjectKeys = getContainerKeyToObjectKeysMap();
        Map<String, Map<String, Long>> deviceKeyToObjectKeyToId = getContainerKeyToObjectKeyToId();
        for (long time = timeSpan.getFromTime(); time <= timeSpan.getToTime(); time += pollingInterval) {
            for (String deviceKey : containerKeys) {
                Collection<RawSample<C>> rawSamples = new LinkedList<RawSample<C>>();
                for (String objectKey : deviceKeyToObjectKeys.get(deviceKey)) {
                    sampleData = generateNextSampleData(sampleData);
                    RawSample<C> rawSample = new RawSample<C>(objectKey, sampleData);
                    rawSamples.add(rawSample);
                    // fill out result
                    Long objId = deviceKeyToObjectKeyToId.get(deviceKey).get(objectKey);
                    List<Sample<C>> objectSamples = result.get(objId);
                    if (objectSamples == null) {
                        objectSamples = new LinkedList<Sample<C>>();
                        result.put(objId, objectSamples);
                    }
                    objectSamples.add(new Sample<C>(objId, time, sampleData));
                }
                reportRawSamples(DATASOURCE_ID, deviceKey, time, rawSamples);
            }
        }

        return result;
    }

    @Test
    public void testHashCodeEqualsImplemented() throws Exception {
        C c = generateNextSampleData(null);

        Method hashCodeMethod = c.getClass().getMethod("hashCode");
        Assert.assertFalse(c.getClass().getSimpleName() + " did not implemented hashCode()", hashCodeMethod.getDeclaringClass().equals(Object.class));

        Method equalsMethod = c.getClass().getMethod("equals", Object.class);
        Assert.assertFalse(c.getClass().getSimpleName() + " did not implemented equals(Object o)", equalsMethod.getDeclaringClass().equals(Object.class));

        S s = generateNextSummaryData(null);

        hashCodeMethod = s.getClass().getMethod("hashCode");
        Assert.assertFalse(s.getClass().getSimpleName() + " did not implemented hashCode()", hashCodeMethod.getDeclaringClass().equals(Object.class));

        equalsMethod = s.getClass().getMethod("equals", Object.class);
        Assert.assertFalse(s.getClass().getSimpleName() + " did not implemented equals(Object o)", equalsMethod.getDeclaringClass().equals(Object.class));
    }

    @Test
    public void testSampleTable() throws Exception {
        Connection connection = null;
        Long id = Long.valueOf(1);
        Set<Long> ids = Collections.singleton(id);

        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            SampleTable<C> sampleTable = getSampleTable(connection, getFirstContainerId());

            sampleTable.dropTable();
            Assert.assertFalse(sampleTable.getFullName() + " should not exist", sampleTable.isTableExist());
            sampleTable.createTable();
            // second create should not throw any duplicate exceptions - it should be flexible enough
            // to recognize if table already exists
            sampleTable.createTable();
            Assert.assertTrue(sampleTable.getFullName() + " should exist", sampleTable.isTableExist());

            TimeSpan hour = generateHourTimeSpan(System.currentTimeMillis());
            List<Sample<C>> samples = generateSamples(id, hour, 10 * Period.MINUTE.milliseconds());
            // test write/read from table
            sampleTable.insert(samples);

            List<Sample<C>> dbSamples = sampleTable.getSamples(hour, id);

            Collections.sort(dbSamples);
            Assert.assertEquals(sampleTable.getFullName() + " insert and select are not equals size", samples.size(), dbSamples.size());
            AssertUtil.assertEquals(samples, dbSamples);

            // test getting samples for list of ids
            dbSamples.clear();
            dbSamples = sampleTable.getSamples(hour, ids);

            Collections.sort(dbSamples);
            Assert.assertEquals(sampleTable.getFullName() + " insert and select by ids are not equals size", samples.size(), dbSamples.size());
            AssertUtil.assertEquals(samples, dbSamples);

            // test getting samples by chunks for all samples
            dbSamples.clear();
            Iterator<List<Sample<C>>> sampleChunks = sampleTable.getSampleChunks(hour, 10);
            while (sampleChunks.hasNext()) {
                dbSamples.addAll(sampleChunks.next());
            }

            Collections.sort(dbSamples);
            AssertUtil.assertEquals(samples, dbSamples);

            // test getting samples by chunks for list of ids
            dbSamples.clear();
            sampleChunks = sampleTable.getSampleChunks(hour, ids, 10);
            while (sampleChunks.hasNext()) {
                dbSamples.addAll(sampleChunks.next());
            }

            Collections.sort(dbSamples);
            AssertUtil.assertEquals(samples, dbSamples);

            // test table time span
            TimeSpan tableTimeSpan = sampleTable.getTableTimeSpan();
            TimeSpan joinSpan = hour.getJoin(tableTimeSpan);
            Assert.assertEquals("Should be an hour time span", hour, joinSpan);

            // test deleting old samples
            int numberOfSamplesExpected = samples.size();
            Assert.assertEquals("Number of samples", numberOfSamplesExpected, sampleTable.getCount());

            long deleteBefore = tableTimeSpan.getFromTime() + 2 * 10 * Period.MINUTE.milliseconds() - 1;
            sampleTable.deleteOldestSamples(deleteBefore);

            Assert.assertEquals("Should be 2 samples less ", numberOfSamplesExpected - 2, sampleTable.getCount());

        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Test
    public void testEngineOfSampleTables() throws Exception {
        Connection connection = null;
        try {
            String engine = "INNODB";
            connection = getDataSource().getConnection();
            SampleTable<C> sampleTable = getSampleTable(connection, getFirstContainerId());

            Assert.assertTrue("The engine of the sample table is not InnoDB", sampleTable.getCreateSql().toUpperCase().contains(engine));
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Test
    public void testSampleTableDuplicateKey() throws Exception {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            SampleTable<C> sampleTable = getSampleTable(connection, getFirstContainerId());

            sampleTable.dropTable();
            sampleTable.createTable();

            Sample<C> sample = new Sample<C>(Long.valueOf(1), System.currentTimeMillis(), generateNextSampleData(null));
            List<Sample<C>> samples = Collections.singletonList(sample);
            // test write/read from table
            sampleTable.insert(samples);
            try {
                sampleTable.insert(samples);
                Assert.assertTrue("Second insert should create Duplicate key exception ", false);
            } catch (TableException e) {
                // expecting duplicate
            }

            sample.setId(Long.valueOf(2));
            sampleTable.insert(samples);

            sample.setId(Long.valueOf(1));
            sample.setTime(System.currentTimeMillis() + 1);
            sampleTable.insert(samples);

            Assert.assertEquals("Number of samples", 3, sampleTable.getCount());
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Test
    public void testCurrentSampleTableNoDuplicateKey() throws Exception {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            for (String containerKey : getContainerKeyToContainerId().keySet()) {
                CurrentSampleTable<C> currentSampleTable = getCurrentSampleTable(connection, containerKey);

                Sample<C> sample = new Sample<C>(Long.valueOf(1), System.currentTimeMillis(), generateNextSampleData(null));
                List<Sample<C>> samples = Collections.singletonList(sample);
                // test write/read from table
                currentSampleTable.insert(samples);
                try {
                    currentSampleTable.insert(samples);
                } catch (TableException e) {
                    Assert.assertTrue("Second insert should NOT create Duplicate key exception since it is update last sample time", false);
                }

                sample.setId(Long.valueOf(2));
                currentSampleTable.insert(samples);

                sample.setId(Long.valueOf(1));
                sample.setTime(System.currentTimeMillis() + 1);
                currentSampleTable.insert(samples);

                Assert.assertEquals("Number of current samples is equal to number of ids", 1, currentSampleTable.getCount());
            }
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Test
    public void testSummaryTable() throws Exception {
        Connection connection = null;
        Long id = Long.valueOf(1);
        Set<Long> ids = Collections.singleton(id);

        try {
            connection = getDataSource().getConnection();
            SummaryTable<S> summaryTable = getSummaryTable(connection, getFirstContainerId());

            // test create table
            summaryTable.dropTable();
            Assert.assertFalse(summaryTable.getFullName() + " should not exist", summaryTable.isTableExist());
            summaryTable.createTable();
            // second create should not throw any duplicate exceptions - it should be flexible enough
            // to recognize if table already exists
            summaryTable.createTable();
            Assert.assertTrue(summaryTable.getFullName() + " should exist", summaryTable.isTableExist());

            // generates summary
            long now = System.currentTimeMillis();
            TimeSpan hour1 = generateHourTimeSpan(now);
            Summary<S> summary1 = new Summary<S>(id, hour1, generateNextSummaryData(null));
            AssertUtil.validateHourTimeSpan(summary1.getTimeSpan());

            TimeSpan hour2 = generateHourTimeSpan(now + Period.HOUR.milliseconds());
            Summary<S> summary2 = new Summary<S>(id, hour2, generateNextSummaryData(summary1.getSummaryData()));
            AssertUtil.validateHourTimeSpan(summary2.getTimeSpan());

            // test write/read
            List<Summary<S>> summaries = new ArrayList<Summary<S>>();
            summaries.add(summary1);
            summaries.add(summary2);
            summaryTable.insert(summaries);
            List<Summary<S>> dbSummaries = summaryTable.getSummaries(hour1, ids);
            Assert.assertEquals(1, dbSummaries.size());
            Assert.assertTrue(equalsSummary(summary1.getSummaryData(), dbSummaries.get(0).getSummaryData()));
            //assertEquals(summary1, dbSummaries.get(0));

            dbSummaries = summaryTable.getSummaries(hour2, ids);
            Assert.assertEquals(1, dbSummaries.size());
            Assert.assertTrue(equalsSummary(summary2.getSummaryData(), dbSummaries.get(0).getSummaryData()));
            //assertEquals(summary2, dbSummaries.get(0));

            dbSummaries = summaryTable.getSummaries(hour1.getJoin(hour2), ids);
            Assert.assertEquals(2, dbSummaries.size());
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Test
    public void testSummaryMergeSqlQueries() throws Exception {
        Connection connection = null;
        long id1 = 1;
        long id2 = 2;
        try {
            connection = getDataSource().getConnection();
            SummaryTable<S> summaryTable = getSummaryTable(connection, getFirstContainerId());

            summaryTable.dropTable();
            summaryTable.createTable();
            long now = Period.DAY.subtract(System.currentTimeMillis());

            TimeSpan hour1 = generateHourTimeSpan(now);
            Summary<S> summary1 = new Summary<S>(id1, hour1, generateNextSummaryData(null));

            TimeSpan hour2 = hour1.offset(-Period.HOUR.milliseconds());
            Summary<S> summary2 = new Summary<S>(id2, hour2, generateNextSummaryData(null));

            List<Summary<S>> summaries = new LinkedList<Summary<S>>();
            summaries.add(summary1);
            summaries.add(summary2);

            // test write/read
            summaryTable.insert(summaries);

            S dbSummaryData1 = summaryTable.getMergedSummaries(hour1, getCommonSqlMethods());
            S dbSummaryData2 = summaryTable.getMergedSummaries(hour2, getCommonSqlMethods());
            Assert.assertNotNull(dbSummaryData1);
            Assert.assertNotNull(dbSummaryData2);
            Summary<S> dbSummary1 = new Summary<S>(id1, hour1, dbSummaryData1);
            Summary<S> dbSummary2 = new Summary<S>(id2, hour2, dbSummaryData2);

            List<Summary<S>> dbSummaries1 = summaryTable.getMergedSummariesGroupById(hour1, getCommonSqlMethods());
            Assert.assertEquals("Number of rows", 1, dbSummaries1.size());
            Summary<S> groupSummaryId1 = dbSummaries1.get(0);
            Assert.assertEquals(dbSummary1, groupSummaryId1);

            List<Summary<S>> dbSummaries2 = summaryTable.getMergedSummariesGroupById(hour2, getCommonSqlMethods());
            Assert.assertEquals("Number of rows", 1, dbSummaries2.size());
            Summary<S> groupSummaryId2 = dbSummaries2.get(0);
            Assert.assertEquals(dbSummary2, groupSummaryId2);

            List<Summary<S>> dbSummaries = summaryTable.getMergedSummariesGroupById(hour1, Collections.singletonList(Long.valueOf(1L)), getCommonSqlMethods());
            Assert.assertEquals("Number of rows", 1, dbSummaries.size());
            Assert.assertEquals(dbSummary1, dbSummaries.get(0));

            dbSummaries = summaryTable.getMergedSummariesGroupById(hour2, Collections.singletonList(Long.valueOf(1L)), getCommonSqlMethods());
            Assert.assertEquals("Number of rows", 0, dbSummaries.size());

            dbSummaries = summaryTable.getMergedSummariesGroupById(hour1, Collections.singletonList(Long.valueOf(2L)), getCommonSqlMethods());
            Assert.assertEquals("Number of rows", 0, dbSummaries.size());

            dbSummaries = summaryTable.getMergedSummariesGroupById(hour2, Collections.singletonList(Long.valueOf(2L)), getCommonSqlMethods());
            Assert.assertEquals("Number of rows", 1, dbSummaries.size());
            Assert.assertEquals(dbSummary2, dbSummaries.get(0));

            dbSummaries = summaryTable.getHourlyPattern(hour1.getJoin(hour2), Collections.singletonList(Long.valueOf(2L)), getCommonSqlMethods());
            Assert.assertEquals("Number of rows", 1, dbSummaries.size());

            dbSummaries = summaryTable.getHourlyPattern(hour1.getJoin(hour2), getCommonSqlMethods());
            Assert.assertEquals("Number of rows", 2, dbSummaries.size());

        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Test
    public void testReportRawDataAndGetSamplesSummaries() throws Exception {
        importFoundationXmlData();
        long now = System.currentTimeMillis();
        // report 1 day of performance data
        long fromTime = now - 5 * Period.HOUR.milliseconds();
        TimeSpan wholeTimeSpan = new TimeSpan(fromTime, now);
        Map<Long, List<Sample<C>>> objIdToSamples = reportPerformanceData(wholeTimeSpan, 10 * Period.MINUTE.milliseconds());
        Set<Sample<C>> latestSamples = new HashSet<Sample<C>>(); // we need this collection to test current sample table
        for (List<Sample<C>> allSamples : objIdToSamples.values()) {
            latestSamples.add(allSamples.get(allSamples.size() - 1)); // last sample is latest according to "reportPerformanceData" function
        }

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(getDataSource());

            // test current sample table and status table
            String containerKey = getFirstContainerKey();
            Long containerId = getFirstContainerId();
            CurrentSampleTable<C> currentSampleTable = getCurrentSampleTable(connection, containerKey);
            List<Sample<C>> allSamples = currentSampleTable.findAll();
            Assert.assertEquals("There should be equal number of current samples and latest samples", latestSamples.size(), allSamples.size());
            for (Sample<C> sample : allSamples) {
                Assert.assertTrue(latestSamples.contains(sample));
            }

            CurrentSampleStatusTable currentSampleStatusTable = new CurrentSampleStatusTable(connection, currentSampleTable.getDatabase());
            SampleFacade<C> sampleFacade = getPerformancePackageFactory(containerKey).getSampleFacade();
            SummaryFacade<S> summaryFacade = getPerformancePackageFactory(containerKey).getSummaryFacade();

            List<CurrentSampleStatus> currentSampleStatuses = currentSampleStatusTable.findAll();
            Assert.assertEquals(getContainerKeyToObjectKeyToId().size(), currentSampleStatuses.size());

            Map<String, CurrentSampleStatus> containerKeyToStatus = new HashMap<String, CurrentSampleStatus>();
            for (CurrentSampleStatus status : currentSampleStatuses) {
                containerKeyToStatus.put(status.getSampleContainerKey(), status);
            }

            // test current samples logic
            List<Sample<C>> latestContainerSamples = currentSampleTable.findAllByContainerKey(containerKey);
            Assert.assertEquals("There should be equal number of current samples and objids involved for given containerkey " + containerKey,
                    getContainerKeyToObjectKeyToId().get(containerKey).values().size(), latestContainerSamples.size());
            Assert.assertTrue(containerKeyToStatus.containsKey(containerKey));
            CurrentSampleStatus currentSampleStatus = containerKeyToStatus.get(containerKey);
            Assert.assertEquals(getPerformancePackageFactory(containerKey).getPersistenceStrategy().getPerformancePackageId(), currentSampleStatus.getPerformancePackageId());
            Assert.assertEquals(containerKey, currentSampleStatus.getSampleContainerKey());
            Assert.assertEquals(containerId, currentSampleStatus.getSampleContainerId());
            Assert.assertEquals(latestContainerSamples.iterator().next().getTime(), currentSampleStatus.getLastSampleTime());

            Collection<Long> ids = getContainerKeyToObjectKeyToId().get(containerKey).values();
            sampleFacade.getSamplesGroupByTime(connection, new HashSet<Long>(ids), wholeTimeSpan);
            for (Long id : ids) {
                List<Sample<C>> samples = sampleFacade.getSamples(connection, id, wholeTimeSpan);
                Collections.sort(samples);
                Assert.assertEquals(samples.get(0).getTime(), fromTime);
                Assert.assertEquals(samples.get(samples.size() - 1).getTime(), now);

                // get summaries returns all overlapping summaries for given timespan
                List<Summary<S>> summaries = summaryFacade.getSummaries(connection, wholeTimeSpan, Collections.singleton(id));
                Collections.sort(summaries);
                Assert.assertEquals(summaries.get(0).getTimeSpan().getFromTime(), generateHourTimeSpan(fromTime).getFromTime());
                Assert.assertEquals(summaries.get(summaries.size() - 1).getTimeSpan().getToTime() + 1, generateHourTimeSpan(now).getFromTime());

                // get merged returns only rightmost of the two overlapping timespans for given timestamp
                Assert.assertNotNull(summaryFacade.getMergedSummaries(connection, wholeTimeSpan, getCommonSqlMethods()));
                summaries = summaryFacade.getMergedSummariesGroupById(connection, wholeTimeSpan, getCommonSqlMethods());
                Collections.sort(summaries);
                Assert.assertEquals(summaries.get(0).getTimeSpan().getFromTime(), generateHourTimeSpan(fromTime + Period.HOUR.milliseconds()).getFromTime());
                Assert.assertEquals(summaries.get(summaries.size() - 1).getTimeSpan().getToTime() + 1, generateHourTimeSpan(now).getFromTime());

                // get merged returns only rightmost of the two overlapping timespans for given timestamp
                summaries = summaryFacade.getHourlyPattern(connection, wholeTimeSpan, Collections.singleton(id), getCommonSqlMethods());
                Assert.assertEquals(5, summaries.size());
            }
        } finally {
            ConnectionHelper.close(connection);
        }

//        doTestReportRawDataAndGetSamplesSummariesMore(wholeTimeSpan);
    }

    @Test
    public void testPersistanceFacade() throws Exception {
        importFoundationXmlData();
        Connection connection = null;

        try {
            connection = getDataSource().getConnection();

            // just recreate tables to run subsequent queries on new tables - tests logic of drop and create
            // so that drop method drops all tables. if drop method doesn't drop all then create method might
            // raise exception that one of tables already exists
            String containerKey = getFirstContainerKey();
            PersistenceFacade<C, S> persistenceFacade = new PersistenceFacade<C, S>(connection, getPerformancePackageFactory(containerKey)) {
                };
            persistenceFacade.dropTables();
            persistenceFacade.createTables();

            long now = System.currentTimeMillis();
            long fromTime = now - 5 * Period.HOUR.milliseconds();
            TimeSpan wholeTimeSpan = new TimeSpan(fromTime, now);

            Map<Long, List<Sample<C>>> idToSamples = reportPerformanceData(Collections.singleton(containerKey), wholeTimeSpan, 10 * Period.MINUTE.milliseconds());

            for (Long id : idToSamples.keySet()) {
                // test read samples
                Set<Long> ids = Collections.singleton(id);
                List<Sample<C>> samples = idToSamples.get(id);
                List<Sample<C>> dbSamples = persistenceFacade.getSamples(id, wholeTimeSpan);

                AssertUtil.assertEquals(samples, dbSamples);

                // test getting summary for collections of ids
                Map<Long, Summary<S>> dbSummaries = persistenceFacade.getSummaryById(wholeTimeSpan, ids, getCommonSqlMethods());
                Assert.assertEquals(dbSummaries.size(), 1);
                Summary<S> summary = dbSummaries.values().iterator().next();
                Assert.assertEquals(summary.getId(), id);

                List<Summary<S>> mergedSummaries = persistenceFacade.getMergedSummariesGroupById(wholeTimeSpan, ids, getCommonSqlMethods());
                Assert.assertEquals(mergedSummaries.size(), 1);
                summary = mergedSummaries.iterator().next();
                AssertUtil.validateHourTimeSpan(summary.getTimeSpan());
            }

            // test getting ALL summaries
            Map<Long, Summary<S>> dbSummaries = persistenceFacade.getSummaryById(wholeTimeSpan, getCommonSqlMethods());
            Assert.assertEquals(dbSummaries.size(), 3);
            Summary<S> summary = dbSummaries.values().iterator().next();
            Assert.assertTrue(idToSamples.keySet().contains(summary.getId()));

            // same as above but testing with NULL ids collection
            dbSummaries = persistenceFacade.getSummaryById(wholeTimeSpan, null, getCommonSqlMethods());
            Assert.assertEquals(dbSummaries.size(), 3);
            for (Summary<S> dbSummary : dbSummaries.values()) {
                Assert.assertTrue(idToSamples.keySet().contains(dbSummary.getId()));
            }

            // test getting merged summary
            S mergedSummaryData = persistenceFacade.getMergedSummaries(wholeTimeSpan, getCommonSqlMethods());
            Assert.assertNotNull(mergedSummaryData);

            List<Summary<S>> mergedSummaries = persistenceFacade.getMergedSummariesGroupById(wholeTimeSpan, getCommonSqlMethods());
            Assert.assertEquals(mergedSummaries.size(), 3);
            for (Summary<S> mergedSummary : mergedSummaries) {
                Assert.assertTrue(idToSamples.keySet().contains(mergedSummary.getId()));
                AssertUtil.validateHourTimeSpan(mergedSummary.getTimeSpan());
            }
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    public boolean equalsSummary(S s1, S s2) {
        return s1.equals(s2);
    }

    private String getFirstContainerKey() throws Exception {
        Set<String> containerKeys = getContainerKeyToContainerId().keySet();
        return containerKeys.isEmpty() ? null : containerKeys.iterator().next();
    }

    private Long getFirstContainerId() throws Exception {
        String key = getFirstContainerKey();
        return getContainerKeyToContainerId().get(key);
    }
}
