package com.netapp.oci.platform.framework.performance.util;

import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.platform.framework.performance.data.Sample;

import java.util.*;

import org.slf4j.LoggerFactory;

/**
 * Converts raw samples to regular samples
 */
public abstract class AbstractRawSampleConverter<C> implements RawSampleConverter<C> {
    protected AbstractRawSampleConverter() {
    }

    protected org.slf4j.Logger getLogger() {
        return LoggerFactory.getLogger(getClass());
    }

    /**
     * Converts raw-samples retrieved from the acquisiton into server-samples, setting the sample time and taking care
     * for creating zeroed samples for the first report.
     *
     * @param rawSamples  sent by the acquisition, containing at most a single sample per monitored object (e.g. switch port,
     *                    or virtual-machine
     * @param time        the time that samples were retrieved by the server
     * @param firstReport true if this is the first report from the acqusition data-source, meaning that for each sample
*                    the server will "invent" another zeroed sample immediately before the reported sample    @return the samples
     * @throws PerformanceException if found multiple samples for the same ID
     */
    public Collection<Sample<C>> toSamples(Collection<RawSample<C>> rawSamples, long time, boolean firstReport) throws PerformanceException {
        /**
         * IDs are used for checking if raw-samples contains the same id twice
         */
        Set<Long> ids = new HashSet<Long>();

        List<Sample<C>> samples = new LinkedList<Sample<C>>();
        for (RawSample<C> rawSample : rawSamples) {
            Long id = getIdFromKey(rawSample.getKey());
            if (id == null){
                // subclasses do the logging of warning message
                continue;
            }
            Sample<C> sample = new Sample<C>(id, time, rawSample.getCounters());
            if (ids.contains(sample.getId())) {
                getLogger().error("Detected multiples samples for ID={}, Raw Sample Key={}, sample={}", sample.getId(), rawSample.getKey(), sample.toString());
                throw new PerformanceException("Detected multiples samples for ID=" + sample.getId() + "  Raw Sample Key" + rawSample.getKey());
            }

            ids.add(sample.getId());

            if (firstReport) {
                /**
                 * insert null sample as first sample - done to mark interruptions in acquisition reporting of performance data
                 */
                samples.add(new Sample<C>(sample.getId(), time - 1));
            }

            samples.add(sample);
        }

        return samples;
    }

    /**
     * Translates a string-based object identifiers, such as WWN, to a long-number (objId). String-based keys are used
     * by the acquisition to identify the objects on which {@link RawSample}s are reported. Numbers are used internally
     * by the server to identify objects.
     *
     * @param key string based object identifier (OID from OData)  @return long number based object identifier
     * @throws PerformanceException if failed to find object ID by the specified key
     */
    protected abstract Long getIdFromKey(String key) throws PerformanceException;
}
