package com.netapp.oci.platform.framework.performance.util;

import java.sql.Connection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.SampleFacade;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.framework.performance.tables.SampleTable;
import com.netapp.oci.platform.framework.performance.tables.SqlMethod;

/**
 * Abstract implementation of the sample facade. It is working under assumption that all
 * samples are stored in one sample table
 *
 * @author Anton Sergeev
 */
public abstract class AbstractSampleFacade<C> implements SampleFacade<C> {

    /**
     * Logger instance
     */
    private org.slf4j.Logger logger;


    /**
     * Abstract method to get instance of sample table
     *
     * @param connection - database connection
     * @return sample table
     */
    protected abstract SampleTable<C> getSampleTable(Connection connection);


    /**
     * Abstract method to get the SQL methods for Grouping by Time
     *
     * @return Sql methods
     */
    protected abstract SqlMethod[] getGroupByTimeSqlMethod();

    protected AbstractSampleFacade(Logger logger) {
        this.logger = logger;
    }

    protected AbstractSampleFacade() {
    }

    public Logger getLogger() {
        return logger;
    }

    public List<Sample<C>> getSamples(Connection connection, Long id, TimeSpan timeSpan) throws PerformanceException {
        List<Sample<C>> samples = new LinkedList<Sample<C>>();
        SampleTable<C> sampleTable = getSampleTable(connection);
        try {
            TimeSpan sampleTableTimeSpan = sampleTable.getTableTimeSpan();

            if (sampleTableTimeSpan != null && sampleTableTimeSpan.overlaps(timeSpan)) {
                // select portion of data from sample table
                samples.addAll(sampleTable.getSamples(timeSpan, id));
            }

            return samples;
        } catch (TableException ex) {
            throw new PerformanceException("Failed getting samples from database", ex);
        }
    }

    public Map<Long, List<Sample<C>>> getSamples(Connection connection, Set<Long> ids, TimeSpan timeSpan) throws PerformanceException {
        Map<Long, List<Sample<C>>> result = new HashMap<Long, List<Sample<C>>>(ids.size());

        // Get the samples recorded for a single object.
        // Skip temporary table joins optimization when the query request contains a single object.
        //
        if(ids.size() == 1){
            final Long id = ids.iterator().next();
            result.put(id, getSamples(connection, id, timeSpan));
            return result;
        }
        
        // Get the samples recorded for multiple objects.
        // This involves creation of a temporary table containing the specified IDs, which is used for Join queries.
        //
        SampleTable<C> sampleTable = getSampleTable(connection);
        try {
            TimeSpan sampleTableTimeSpan = sampleTable.getTableTimeSpan();

            if (sampleTableTimeSpan != null && sampleTableTimeSpan.overlaps(timeSpan)) {
                // samples are ordered by objid and time
                Long currentId = Long.MIN_VALUE;
                List<Sample<C>> currentList = Collections.emptyList();
                for (Sample<C> sample : sampleTable.getSamples(timeSpan, ids)) {
                    if (!sample.getId().equals(currentId)) {
                        currentId = sample.getId();
                        currentList = new LinkedList<Sample<C>>();
                        result.put(currentId, currentList);
                    }
                    currentList.add(sample);
                }
            }


            return result;
        } catch (TableException ex) {
            throw new PerformanceException("Failed getting samples from database", ex);
        }
    }

    public Map<Long, Sample<C>> getLatestSample(Connection connection, Set<Long> ids) throws PerformanceException {
        Map<Long, Sample<C>> result = new HashMap<Long, Sample<C>>(ids.size());
        SampleTable<C> sampleTable = getSampleTable(connection);
        try {
            for (Sample<C> sample : sampleTable.getLatestSample(ids)) {
                result.put(sample.getId(), sample);
            }
        } catch (TableException ex) {
            throw new PerformanceException("Failed getting samples from database", ex);
        }
        return result;
    }

    public List<Sample<C>> getSamplesGroupByTime(Connection connection, Set<Long> ids, TimeSpan timeSpan) throws PerformanceException {
        List<Sample<C>> samples = new LinkedList<Sample<C>>();
        SampleTable<C> sampleTable = getSampleTable(connection);
        try {
            TimeSpan sampleTableTimeSpan = sampleTable.getTableTimeSpan();

            if (sampleTableTimeSpan != null && sampleTableTimeSpan.overlaps(timeSpan)) {
                // select portion of data from sample table
                samples.addAll(sampleTable.getMergedSamplesGroupByTime(timeSpan, ids, getGroupByTimeSqlMethod()));
            }

            Collections.sort(samples);

            return samples;
        } catch (TableException ex) {
            throw new PerformanceException("Failed getting samples from database", ex);
        }
    }

    public List<Sample<C>> getSamplesGroupByTime(Connection connection, TimeSpan timeSpan) throws PerformanceException {
        List<Sample<C>> samples = new LinkedList<Sample<C>>();
        SampleTable<C> sampleTable = getSampleTable(connection);
        try {
            TimeSpan sampleTableTimeSpan = sampleTable.getTableTimeSpan();

            if (sampleTableTimeSpan != null && sampleTableTimeSpan.overlaps(timeSpan)) {
                // select portion of data from sample table
                samples.addAll(sampleTable.getMergedSamplesGroupByTime(timeSpan, getGroupByTimeSqlMethod()));
            }

            Collections.sort(samples);

            return samples;
        } catch (TableException ex) {
            throw new PerformanceException("Failed getting samples from database", ex);
        }
    }

    /**
     * returns true if the samples exist over a period of time.
     *
     * @param timeSpan the period in which samples sould be retrieved
     * @param id
     * @return true if samples exist
     * @throws com.netapp.oci.platform.common.table.TableException
     *          if failed to retrieve the samples
     */
    public Boolean doSamplesExist(Connection connection, TimeSpan timeSpan, Long id) throws PerformanceException {
        try {
            SampleTable<C> sampleTable = getSampleTable(connection);

            return sampleTable.doSamplesExist(timeSpan, id);
        } catch (TableException ex){
            throw new PerformanceException("Failed to query sample table ", ex);
        }
    }

    /**
     * returns first sample after timestamp
     *
     * @param id
     * @param afterTime
     * @return
     * @throws PerformanceException
     */
    public Sample<C> getFirstSampleAfterTime(Connection connection, Long id, long afterTime) throws PerformanceException {
        try {
            SampleTable<C> sampleTable = getSampleTable(connection);
            return sampleTable.getFirstSampleAfterTime(id, afterTime);
        }catch (TableException ex) {
            throw new PerformanceException("Failed to get first sample from sample table ", ex);
        }
    }

    /**
     * returns last sample before timestamp
     *
     * @param id
     * @param beforeTime
     * @return
     * @throws PerformanceException
     */
    public Sample<C> getLastSampleBeforeTime(Connection connection, Long id, long beforeTime) throws PerformanceException {
        try {
            SampleTable<C> sampleTable = getSampleTable(connection);
            return sampleTable.getLastSampleBeforeTime(id, beforeTime);
        }catch (TableException ex) {
            throw new PerformanceException("Failed to get last sample from sample table ", ex);
        }
    }
}
