package com.netapp.oci.platform.framework.performance.util;

import java.sql.Connection;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.onaro.commons.types.Period;
import com.onaro.commons.util.MapOfLists;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.SummaryFacade;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.tables.SampleTable;
import com.netapp.oci.platform.framework.performance.tables.SqlMethod;
import com.netapp.oci.platform.framework.performance.tables.SummaryTable;

/**
 * Abstract implementation of summary facade. It is acting under assumption
 * there is only one summary table.
 *
 * @author Anton Sergeev
 */
public abstract class AbstractSummaryFacade<C, S> implements SummaryFacade<S> {
    /**
     * Chunk size is how many object ids will be processed at a time to determine
     * summaries for samples of those objects.
     */
    private static final int CHUNK_SIZE = 500;

    /**
     * Abstract method to get instance of summary table
     *
     * @param connection - database connection
     * @return summary table
     */
    protected abstract SummaryTable<S> getSummaryTable(Connection connection);

    /**
     * Abstract method to get instance of sample table
     *
     * @param connection - database connection
     * @return sample table
     */
    protected abstract SampleTable<C> getSampleTable(Connection connection);

    protected SummaryCalculator<C, S> getSummaryCalculator() {
        return null;
    }

    public List<Summary<S>> getMergedSummariesGroupById(Connection connection, TimeSpan timeSpan, SqlMethod[] methods) throws PerformanceException {
        return getMergedSummariesGroupById(connection, timeSpan, null, methods);
    }

    public List<Summary<S>> getMergedSummariesGroupById(Connection connection, TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws PerformanceException {
        try {
            // merged summary can work only on the summaryTable, if it less then an hour
            // add an hour
            if (timeSpan.size() < Period.HOUR.milliseconds()) {
                timeSpan.setFromTime(timeSpan.getToTime() - Period.HOUR.milliseconds());
            }
            SummaryTable<S> summaryTable = getSummaryTable(connection);
            if (ids == null || ids.size() == 0) {
                return summaryTable.getMergedSummariesGroupById(timeSpan, methods);
            } else {
                return summaryTable.getMergedSummariesGroupById(timeSpan, ids, methods);
            }
        } catch (TableException e) {
            throw new PerformanceException("Failed to get merged summaries", e);
        }
    }

    public S getMergedSummaries(Connection connection, TimeSpan timeSpan, SqlMethod[] methods) throws PerformanceException {
        try {
            // merged summary can work only on the summaryTable, if it less then an hour
            // add an hour
            if (timeSpan.size() < Period.HOUR.milliseconds()) {
                timeSpan.setFromTime(timeSpan.getToTime() - Period.HOUR.milliseconds());
            }
            SummaryTable<S> summaryTable = getSummaryTable(connection);
            return summaryTable.getMergedSummaries(timeSpan, methods);
        } catch (TableException e) {
            throw new PerformanceException("Failed to get merged summaries", e);
        }
    }

    public List<Summary<S>> getSummaries(Connection connection, TimeSpan timeSpan, Collection<Long> ids) throws PerformanceException {
        if (timeSpan.size() < Period.HOUR.milliseconds()) {
            timeSpan.setFromTime(timeSpan.getToTime() - Period.HOUR.milliseconds());
        }
        try {
            SummaryTable<S> summaryTable = getSummaryTable(connection);
            if (ids == null || ids.size() == 0) {
                return summaryTable.getSummaries(timeSpan);
            } else {
                return summaryTable.getSummaries(timeSpan, ids);
            }
        } catch (TableException e) {
            throw new PerformanceException("Failed to get Summaries", e);
        }
    }

    public List<Summary<S>> getSummaries(Connection connection, TimeSpan timeSpan) throws PerformanceException {
        if (timeSpan.size() < Period.HOUR.milliseconds()) {
            timeSpan.setFromTime(timeSpan.getToTime() - Period.HOUR.milliseconds());
        }
        try {
            SummaryTable<S> summaryTable = getSummaryTable(connection);
            return summaryTable.getSummaries(timeSpan);
        } catch (TableException e) {
            throw new PerformanceException("Failed to get Summaries", e);
        }
    }

    public Map<Long, Summary<S>> getMergedSummaryById(Connection connection, TimeSpan timeSpan, SqlMethod[] methods) throws PerformanceException {
        return getMergedSummaryById(connection, timeSpan, null, methods);
    }

    public Map<Long, Summary<S>> getMergedSummaryById(Connection connection, TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws PerformanceException {
        SummaryTable<S> summaryTable = getSummaryTable(connection);
        SampleTable<C> sampleTable = getSampleTable(connection);

        try {
            Map<Long, Summary<S>> summaryById;
            if (ids == null || ids.size() == 0) {
                summaryById = PerformanceUtils.toMap(summaryTable.getMergedSummariesGroupById(timeSpan, methods));
            } else {
                summaryById = PerformanceUtils.toMap(summaryTable.getMergedSummariesGroupById(timeSpan, ids, methods));
            }

            TimeSpan sampleTimeSpan = sampleTable.getTableTimeSpan();
            TimeSpan summaryTimeSpan = summaryTable.getTableTimeSpan();
            TimeSpan sampleSummaryTimeSpan = summaryTimeSpan != null ? timeSpan.substract(summaryTimeSpan) : timeSpan;

            if (sampleTimeSpan != null && sampleSummaryTimeSpan != null) {
                sampleSummaryTimeSpan = sampleTimeSpan.getWholeHoursTimeSpan().getOverlap(sampleSummaryTimeSpan);
                SummaryCalculator<C, S> summaryCalculator = getSummaryCalculator();

                if (sampleSummaryTimeSpan != null && summaryCalculator != null) {
                    Iterator<List<Sample<C>>> sampleChunkIterator;
                    if (ids == null) {
                        sampleChunkIterator = sampleTable.getSampleChunks(sampleSummaryTimeSpan, CHUNK_SIZE);
                    } else {
                        sampleChunkIterator = sampleTable.getSampleChunks(sampleSummaryTimeSpan, ids, CHUNK_SIZE);
                    }

                    // calculate the summary of the sample and merge it with the summary from the summary table
                    while (sampleChunkIterator.hasNext()) {
                        MapOfLists<Long, Sample<C>> samplesById = new MapOfLists<Long, Sample<C>>();
                        for (Sample<C> sample : sampleChunkIterator.next()) {
                            samplesById.add(Long.valueOf(sample.getId()), sample);
                        }

                        for (Map.Entry<Long, List<Sample<C>>> entry : samplesById.entrySet()) {
                            List<Sample<C>> samples = entry.getValue();
                            if (samples.size() > 0) {
                                Summary<S> summary = summaryCalculator.calculateSummary(samples, sampleSummaryTimeSpan);

                                Summary<S> summaryFromSummaries = summaryById.get(entry.getKey());
                                if (summaryFromSummaries != null) {
                                    summary = summaryCalculator.mergeSummaries(summary, summaryFromSummaries);
                                }

                                summaryById.put(entry.getKey(), summary);
                            }
                        }
                    }
                }
            }

            return summaryById;

        } catch (TableException ex) {
            throw new PerformanceException("Failed to get summaries from db", ex);
        }
    }

    public List<Summary<S>> getHourlyPattern(Connection connection, TimeSpan timeSpan, Collection<Long> ids, SqlMethod[] methods) throws PerformanceException {
        TimeSpan newTimeSpan = new TimeSpan(timeSpan);

        // hourly Pattern can work only on days
        if (newTimeSpan.size() < Period.DAY.milliseconds()) {
            newTimeSpan.setFromTime(newTimeSpan.getFromTime() - Period.DAY.milliseconds());
        }

        try {
            SummaryTable<S> summaryTable = getSummaryTable(connection);

            if (ids == null || ids.size() == 0) {
                return summaryTable.getHourlyPattern(newTimeSpan, methods);
            } else {
                return summaryTable.getHourlyPattern(newTimeSpan, ids, methods);
            }
        } catch (TableException e) {
            throw new PerformanceException("Failed to get merged summaries", e);
        }
    }

    @Override
    public Summary<S> getFirstSummaryAfterTime(Connection connection, Long id, Long afterTime) throws PerformanceException {
        try {
            SummaryTable<S> summaryTable = getSummaryTable(connection);
            return summaryTable.getFirstSummaryAfterTime(id, afterTime);
        }catch (TableException ex) {
            throw new PerformanceException("Failed to get first sample from summary table ", ex);
        }
    }

    @Override
    public Summary<S> getLastSummaryBeforeTime(Connection connection, Long id, Long beforeTime) throws PerformanceException {
        try {
            SummaryTable<S> summaryTable = getSummaryTable(connection);
            return summaryTable.getLastSummaryBeforeTime(id, beforeTime);
        }catch (TableException ex) {
            throw new PerformanceException("Failed to get last sample from summary table ", ex);
        }
    }

}
