/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.framework.performance.util;

import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.table.Table;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.TableManagement;
import com.netapp.oci.platform.framework.performance.tables.SampleTable;
import org.slf4j.Logger;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashSet;
import java.util.Set;

/**
 * Abstract implementation of the table management interface.
 * This implementation is in most cases what you need for your performance package.
 * It manages sample.
 *
 * @author Anton Sergeev
 */
public abstract class AbstractTableManagement<C> implements TableManagement {
    /**
     * Caching all the created tables, avoid executing SQL 'SHOW TABLES like ..'
     */
    private static final Set<String> createdTableNames = new HashSet<String>();

    protected org.slf4j.Logger logger;

    protected AbstractTableManagement(Logger logger) {
        this.logger = logger;
    }

    /**
     *  This is the method to fill the cache shortly after MEGA startup so that first acquisition is as fast as subsequent ones.
     *
     * @param connection   --database connection
     */
    public static synchronized void init(Connection connection) {
        org.slf4j.Logger localLogger = org.slf4j.LoggerFactory.getLogger(AbstractTableManagement.class);
        Statement statement = null;
        try {
            statement = connection.createStatement();
            ResultSet resultSet = statement.executeQuery("select table_name from information_schema.tables where table_schema = 'netapp_performance' AND (" +
                    "table_name like 'sample_%' or table_name like 'summary_%')");
            while (resultSet.next())  {
                String tableName = resultSet.getString(1);
                createdTableNames.add("netapp_performance." + tableName);
            }

        } catch (Exception e)   {
            localLogger.error("",e);
        } finally {
            try {
                if(statement!=null) {
                    statement.close();
                }
            } catch (SQLException ex) {
                localLogger.error("",ex);
            }

        }
    }
    /**
     * Abstract method to get instance of sample table
     *
     * @param connection - database connection
     * @return sample table
     */
    protected abstract SampleTable<C> getSampleTable(Connection connection);

    /**
     * Adapter method to create more tables if performance package relies on more tables
     * @param connection - database connection
     * @throws TableException if failed to create more tables
     */
    protected void createMoreTables(Connection connection) throws TableException {
    }

    /**
     * Adapter method to drop more tables if performance package relies on more tables
     * @param connection - database connection
     * @throws com.netapp.oci.platform.common.table.TableException if failed to drop more tables
     */
    protected void dropMoreTables(Connection connection) throws TableException {
    }

    /**
     * Adapter method to find out if more tables required by package exist
     * @param connection - database connection
     * @return indicator if tables exist
     * @throws com.netapp.oci.platform.common.table.TableException if failed to find out if more tables
     */
    protected boolean moreTablesExist(Connection connection) throws TableException {
        return true;
    }

    public void createTables(Connection connection) throws PerformanceException {
        SampleTable<C> sampleTable = getSampleTable(connection);

        if (isMarkCreated(sampleTable))
            return;
        try {
            logger.info("Table Management: Database tables don't exist - creating them.");
            sampleTable.createTable();//this createTable using sql command as "CREATE TABLE IF NOT EXIST", so will never see "table already exists" error
            logger.info("Table Management: Created table " + sampleTable.getFullName());
            createMoreTables(connection);
            markCreated(sampleTable);
        } catch (TableException ex) {
            throw new PerformanceException("Failed to create performance tables", ex);
        }
    }

    public void dropTables(Connection connection) throws PerformanceException {
        try {
            SampleTable<C> sampleTable = getSampleTable(connection);
            sampleTable.dropTable();
            logger.info("Table Management: Dropped table " + sampleTable.getFullName());
            dropMoreTables(connection);
        } catch (TableException ex) {
            throw new PerformanceException("Failed to drop performance tables", ex);
        } finally {
            clearCaches();
            init(connection);
        }
    }

    public static synchronized boolean isMarkCreated(Table table) {
        return createdTableNames.contains(table.getFullName());
    }
    
    public static synchronized boolean isMarkCreated(String fullTableName) {
        return createdTableNames.contains(fullTableName);
    }

    public static synchronized void markCreated(Table table) {
        createdTableNames.add(table.getFullName());
    }

    public static synchronized void unmarkCreated(SampleTable sampleTable) {
        createdTableNames.remove(sampleTable.getFullName());
    }
    
    public static synchronized void clearCaches() {
        createdTableNames.clear();
    }

}
