/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */
package com.netapp.oci.platform.framework.performance.util;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.param.PluginParam;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.tables.SampleTable;
import com.netapp.oci.platform.framework.performance.tables.SummaryTable;
import com.onaro.commons.types.Period;
import com.onaro.commons.util.MapOfLists;

import java.sql.Connection;
import java.util.*;

/**
 * Summarization strategy based on usage of summary calculator
 *
 * @author Anton Sergeev
 */
public abstract class CalculatorSummarizePersistenceStrategy<C, S> extends SummarizePersistenceStrategy<C, S> {
    protected CalculatorSummarizePersistenceStrategy(org.slf4j.Logger logger, PluginParam param, String sampleContainerKey, Long sampleContainerId) {
        super(logger, param, sampleContainerKey, sampleContainerId);
    }

    /**
     * Creates instance of summary calculator
     *
     * @return summary calculator instance
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException
     *          if failed to create SummaryCalculator
     */
    protected abstract SummaryCalculator<C, S> getSummaryCalculator() throws PerformanceException;

    protected void summarize(Connection connection, SampleTable<C> sampleTable, Long time) throws PerformanceException {
        try {
            List<Sample<C>> samples = sampleTable.findAll();
            if (samples.size() == 0)
                return;

            // Samples must be sorted by time in increasing order - it was decided not to use "ORDER BY" in the
            // SQL due to estimates java will be faster. Also note that this sorting is using sample's "compareTo()"
            // method which is assumed to look at time
            Collections.sort(samples);

            TimeSpan timeSpan = getHourTimeSpan(samples);

            MapOfLists<Long, Sample<C>> samplesById = new MapOfLists<Long, Sample<C>>();
            for (Sample<C> sample : samples) {
                samplesById.add(sample.getId(), sample);
            }

            getLogger().info("SummarizePersistenceStrategy: Insert into tables");

            persistSamples(connection, samplesById, timeSpan);
        } catch (TableException ex) {
            throw new PerformanceException("Failed to select all samples from table " + sampleTable.getFullName(), ex);
        }
    }

    protected void persistSamples(Connection connection, MapOfLists<Long, Sample<C>> idToSamples, TimeSpan timeSpan) throws PerformanceException {
        SummaryTable<S> summaryTable = getSummaryTable(connection);
        SummaryCalculator<C, S> summaryCalculator = getSummaryCalculator();
        for (Map.Entry<Long, List<Sample<C>>> entry : idToSamples.entrySet()) {
            List<Sample<C>> hourSamples = entry.getValue();
            Collections.sort(hourSamples);
            persistHourOfSamples(summaryTable, summaryCalculator, hourSamples, timeSpan);
        }
    }

    protected void persistHourOfSamples(SummaryTable<S> summaryTable, SummaryCalculator<C, S> summaryCalculator, List<Sample<C>> hourSamples, TimeSpan timeSpan) throws PerformanceException {
        // summary can be null if there are two samples and one the datasource stop between the samples
        try {
            if (hourSamples == null || hourSamples.size() < 2) {
                getLogger().warn("Failed to persist hour of samples, minimum number of samples should be 2, timeSpan:" + timeSpan);
            } else {
                Summary<S> summary = summaryCalculator.calculateSummary(hourSamples, timeSpan);
                summaryTable.insert(Collections.singleton(summary));
            }
        } catch (Exception ex) {
            getLogger().error("Failed to persist hour of samples, timeSpan:" + timeSpan, ex);
            throw new PerformanceException("Failed to persist hour of samples, timeSpan:" + timeSpan, ex);
        }
    }

    /**
     * get the from hour time to hour time
     *
     * @param samples list of the hour sample to process, assumed to be sorted by time
     * @return the TimeSpan from hour time to hour time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException
     *          there can be two resons
     *          <ol start=1>
     *          <li> the samples are not sorted by time and fromTime > toTime
     *          <li> the time interval between the first hour and last hour are more then one hour
     *          <li> The failed cause of Database exception
     *          </ol>
     */
    protected TimeSpan getHourTimeSpan(List<Sample<C>> samples) throws PerformanceException {
        Sample<C> firstSample = samples.get(0);
        long fromTime = firstSample.getTime();

        Sample<C> lastSample = samples.get(samples.size() - 1);
        long toTime = lastSample.getTime();

        if (toTime - fromTime > Period.HOUR.milliseconds()) {
            throw new PerformanceException("Samples contains more the one hour of data, fromTime:" + fromTime + " toTime:" + toTime);
        }

        GregorianCalendar calendar = new GregorianCalendar();

        // reset minutes, seconds, milliseconds for from depending on the period
        calendar.setTimeInMillis(fromTime);
        calendar.set(Calendar.MINUTE, 0);
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);

        return new TimeSpan(calendar.getTimeInMillis(), calendar.getTimeInMillis() + Period.HOUR.milliseconds() - 1);
    }
}
