package com.netapp.oci.platform.framework.performance.util;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Class is used mostly to remove generics from datasource registry EJB, since they cause problem for obfuscation.
 * Also it is nice way of encapsulation datasource management logic in simple java class
 *
 * @author Anton Sergeev
 */
public class DatasourceRegistry {
    /**
     * Hold for every package and key the datasourceId and the last report time
     */
    private final Map<RegistryKey, RegistryValue> keyToValue = new HashMap<RegistryKey, RegistryValue>();

    public boolean isRegistered(String pckg, String key, Long datasourceId, long leaseTime) {
        synchronized (keyToValue) {
            RegistryKey registryKey = new RegistryKey(pckg, key);
            RegistryValue registryValue = keyToValue.get(registryKey);
            if (registryValue == null)
                return false;

            if (!datasourceId.equals(registryValue.getDatasourceId()))
                return false;

            long timeDiff = System.currentTimeMillis() - registryValue.getLastReportTime();
            boolean res = timeDiff <= leaseTime;
            if (!res) {
                LoggerFactory.getLogger(pckg).info("Unregister " + registryKey + " for " + registryValue.getDatasourceId());
                keyToValue.remove(registryKey);
            }

            return res;
        }
    }

    public boolean register(String pckg, String key, Long datasourceId, long leaseTime) {
        synchronized (keyToValue) {
            RegistryKey registryKey = new RegistryKey(pckg, key);
            RegistryValue registryValue = keyToValue.get(registryKey);
            long curTime = System.currentTimeMillis();
            if (registryValue == null) {
                registryValue = new RegistryValue(datasourceId, curTime);
                keyToValue.put(registryKey, registryValue);
               org.slf4j.LoggerFactory.getLogger(pckg).info("Register " + registryKey + " with " + datasourceId);
                return true;
            }

            Long registeredId = registryValue.getDatasourceId();
            if (registeredId.equals(datasourceId)) {
                registryValue.setLastReportTime(curTime);
                return true;
            }

            long timeDiff = curTime - registryValue.getLastReportTime();
            if (timeDiff > leaseTime) {
                keyToValue.put(registryKey, new RegistryValue(datasourceId, curTime));
                LoggerFactory.getLogger(pckg).info("Register " + registryKey + " with " + datasourceId);
                return true;
            }

           org.slf4j.LoggerFactory.getLogger(pckg).info("Failed to register " + registryKey + " with " + datasourceId + ", already regitered with " + registeredId);

            return false;
        }
    }

    public Long getRegisterDatasourceId(String pckg, String key, long leaseTime) {
        synchronized (keyToValue) {
            RegistryKey registryKey = new RegistryKey(pckg, key);
            RegistryValue registryValue = keyToValue.get(registryKey);
            if (registryValue == null) {
                return null;
            }

            long curTime = System.currentTimeMillis();
            long timeDiff = curTime - registryValue.getLastReportTime();

            if (timeDiff > leaseTime) {
                LoggerFactory.getLogger(pckg).info("Unregister " + registryKey + " for " + registryValue.getDatasourceId());
                keyToValue.remove(registryKey);
                return null;
            }

            return registryValue.getDatasourceId();
        }
    }

    public long getLastReportTime(String pckg, String key, Long datasourceId) {
        synchronized (keyToValue) {
            RegistryKey registryKey = new RegistryKey(pckg, key);
            RegistryValue registryValue = keyToValue.get(registryKey);
            if (registryValue != null && registryValue.getDatasourceId().equals(datasourceId)) {
                registryValue.getLastReportTime();
            }

            return -1;
        }
    }

    public long getFirstReportTime(String pckg, String key, Long datasourceId) {
        synchronized (keyToValue) {
            RegistryKey registryKey = new RegistryKey(pckg, key);
            RegistryValue registryValue = keyToValue.get(registryKey);
            if (registryValue != null && registryValue.getDatasourceId().equals(datasourceId)) {
                registryValue.getFirstReportTime();
            }

            return -1;
        }
    }

    public void clearRegistery() {
        synchronized (keyToValue) {
            keyToValue.clear();
        }
    }

    public void unregisterPackage(String pckg) {
        synchronized (keyToValue) {
            Set<RegistryKey> keySet = new HashSet<RegistryKey>(keyToValue.keySet());
            for (RegistryKey registryKey : keySet) {
                if (registryKey.getPckg().equals(pckg)) {
                    keyToValue.remove(registryKey);
                }
            }
        }
    }

    class RegistryKey {

        private String pckg;

        private String key;

        public RegistryKey(String pckg, String key) {
            this.pckg = pckg;
            this.key = key;
        }

        public String getPckg() {
            return pckg;
        }

        public String getKey() {
            return key;
        }

        public boolean equals(Object o) {
            if (this == o)
                return true;
            if (o == null || getClass() != o.getClass())
                return false;

            RegistryKey that = (RegistryKey) o;

            if (!key.equals(that.key))
                return false;
            return pckg.equals(that.pckg);

        }

        public int hashCode() {
            int result;
            result = pckg.hashCode();
            result = 31 * result + key.hashCode();
            return result;
        }


        public String toString() {
            return "RegistryKey{" +
                    "pckg='" + pckg + '\'' +
                    ", key='" + key + '\'' +
                    '}';
        }
    }

    class RegistryValue {
        private Long datasourceId;
        private long lastReportTime;
        private long firstReportTime;


        public RegistryValue(Long datasourceId, long firstReportTime) {
            this.datasourceId = datasourceId;
            this.lastReportTime = firstReportTime;
            this.firstReportTime = firstReportTime;
        }


        public Long getDatasourceId() {
            return datasourceId;
        }

        public long getLastReportTime() {
            return lastReportTime;
        }

        public long getFirstReportTime() {
            return firstReportTime;
        }

        public void setLastReportTime(long lastReportTime) {
            this.lastReportTime = lastReportTime;
        }

        public boolean equals(Object o) {
            if (this == o)
                return true;
            if (o == null || getClass() != o.getClass())
                return false;

            RegistryValue that = (RegistryValue) o;

            if (lastReportTime != that.lastReportTime)
                return false;
            if (firstReportTime != that.firstReportTime)
                return false;
            if (!datasourceId.equals(that.datasourceId))
                return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = datasourceId.hashCode();
            result = 31 * result + (int) (lastReportTime ^ (lastReportTime >>> 32));
            return result;
        }
    }
}
