/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */
package com.netapp.oci.platform.framework.performance.util;

import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.param.PluginParam;
import com.netapp.oci.platform.common.table.Table;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.UpdateTable;
import com.netapp.oci.platform.framework.performance.tables.SampleTable;
import com.netapp.oci.platform.framework.performance.tables.SummaryTable;
import com.netapp.oci.platform.framework.performance.tables.TimeSpanTable;

import java.sql.Connection;

/**
 * Persistence strategy that provides summarization of the performance samples.
 *
 * @author Anton Sergeev
 */
public abstract class SummarizePersistenceStrategy<C, S> extends AbstractPersistenceStrategy<C> {

    /**
     * The sample table is renamed from performance.sample_xxx to performance.sample_xxx_to_summarize
     */
    public static final String TEMP_TABLE_SUFFIX = "_to_summarize";

    protected abstract SampleTablePreparator getSampleTablePreparator();

    protected SummarizePersistenceStrategy(org.slf4j.Logger logger, PluginParam param, String sampleContainerKey, Long sampleContainerId) {
        super(logger, param, sampleContainerKey, sampleContainerId);
    }

    /**
     * Creates instance of summary table
     *
     * @param connection - database connection
     * @return instance of summary table
     */
    protected abstract SummaryTable<S> getSummaryTable(Connection connection);


    /**
     * Returns the length of time period in milliseconds for which data should be stored in
     * performance tables
     *
     * @return time in milliseconds
     */
    protected abstract long getTablesLifeSpan();

    public void prepare(Connection connection, Long sampleTime) throws PerformanceException {
        /*
         * Rename the sampleTable from performance.sample_xxx to performance.sample_xxx_to_summarize
         *
         * There can be two resons for exception
         *                              <ol start=1>
         *                              <li> The processing of the for the previous hour did not ended (is still in process)
         *                              <li> The rename failed cause of Database exception
         *                              </ol>
         */

        getLogger().info("SummarizePersistenceStrategy: Prepare Start");

        String tableName;
        String database;
        try {
            SampleTable<C> sampleTable = getSampleTable(connection);
            tableName = sampleTable.getName();
            database = sampleTable.getDatabase();

            String tempTableName = tableName + TEMP_TABLE_SUFFIX;

            Table tempSampleTable = new UpdateTable(connection, database, tempTableName);

            // call preparator specific to performance package
            // just a note: preparator may decide not to create to_summarize table
            // that means it is not time yet to run C&S process
            getSampleTablePreparator().prepareSampleData(sampleTable, tempSampleTable.getName(), sampleTime);
        } finally {
            getLogger().info("SummarizePersistenceStrategy: Prepare Ended");
        }
    }

    public void complete(Connection connection, Long sampleTime) throws PerformanceException {
        SampleTable<C> sampleTable = getSampleTable(connection);
        try {
            getLogger().debug("SummarizePersistenceStrategy: Job Start");
            summarize(connection, sampleTable, sampleTime);
        } catch (Exception e) {
            getLogger().error("SummarizePersistenceStrategy: Failed to run summarize()", e);
        } finally {
            getLogger().debug("SummarizePersistenceStrategy: Job Ended ");
        }
    }

    /**
     * Performs summarization depending on the algorithm chosen
     *
     * @param sampleTable - sample table to summarize
     * @param connection  - database connection
     * @throws PerformanceException if failed summarization
     */
    protected abstract void summarize(Connection connection, SampleTable<C> sampleTable, Long sampleTime) throws PerformanceException;

    protected void shrinkTablesSize(Connection connection) throws PerformanceException {
        try {
            long lifeSpan = getTablesLifeSpan();
            TimeSpanTable.ensureLifeSpan(getSummaryTable(connection), lifeSpan);
        } catch (TableException ex) {
            getLogger().error("",ex);
            throw new PerformanceException("Failed to delete data from summary table");
        }
    }
}
