package com.netapp.oci.platform.framework.performance.util.preparators;

import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.framework.performance.tables.SampleTable;
import com.netapp.oci.platform.framework.performance.util.SampleTablePreparator;

import java.util.Calendar;
import java.util.GregorianCalendar;

/**
 * The idea of this implementation lies in moving portion of the data that is ready
 * for compression and summarization from sample table to a temporary table.
 * Instead of renaming whole table.
 * <p/>
 * The reason for this approach is the fact that sample data about last hour might come with
 * certain delay which disables renaming approach.
 * <p/>
 * Acts under assumption that it got triggered when sample table doesn't have samples from the
 * begining of new hour but only samples of the previous hour
 *
 * @author Anton Sergeev
 */
public class CopyPreparator implements SampleTablePreparator {
    /**
     * The number of hours to hold in sample table before copying
     */
    private int hoursToHold;

    private org.slf4j.Logger logger;

    public CopyPreparator(int hoursToHold, org.slf4j.Logger logger) {
        this.hoursToHold = hoursToHold;
        this.logger = logger;
    }

    public void prepareSampleData(SampleTable sampleTable, String tempTableName, Long sampleTime) throws PerformanceException {
        try {
            TimeSpan tableTimeSpan = (sampleTime == null) ? sampleTable.getTableTimeSpan() : null;
            if (sampleTime == null && tableTimeSpan == null) {
                // no need to do anything table is empty
                return;
            }

            Calendar calendar = new GregorianCalendar();
            calendar.setTimeInMillis(sampleTime == null ? tableTimeSpan.getToTime() : sampleTime);

            // that will point to the begining of the last hour with data
            calendar.set(Calendar.MILLISECOND, 0);
            calendar.set(Calendar.SECOND, 0);
            calendar.set(Calendar.MINUTE, 0);

            // moves pointer to the begining of the period to C&S
            calendar.add(Calendar.HOUR, -hoursToHold);
            long fromTime = calendar.getTimeInMillis();

            // adds one hour
            calendar.add(Calendar.HOUR, 1);
            long toTime = calendar.getTimeInMillis() - 1;

            // if we have data to C&S then proceed with it
            if (tableTimeSpan == null || toTime >= tableTimeSpan.getFromTime()) {
                copySamples(sampleTable, tempTableName, fromTime, toTime);
            }
        } catch (TableException ex) {
            logger.error("Preparation step failed in copy preparator", ex);
            throw new PerformanceException("Preparation step failed in copy preparator", ex);
        }
    }

    private void copySamples(SampleTable sampleTable, String tempTableName, long fromTime, long toTime) throws TableException {
        String oldName = sampleTable.getName();
        sampleTable.setName(tempTableName);
        String fullTempName = sampleTable.getFullName();
        sampleTable.createCopyTable();
        sampleTable.setName(oldName);
        sampleTable.copySamples(new TimeSpan(fromTime, toTime), fullTempName);
    }
}