package com.netapp.oci.platform.job;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import javax.annotation.security.RunAs;
import javax.ejb.EJB;
import javax.ejb.Singleton;
import javax.ejb.Startup;
import javax.sql.DataSource;

import com.onaro.management.session.DailyManagementJob;
import com.onaro.management.session.ManagementRemote;
import com.onaro.sanscreen.acquisition.sessions.DSAuditsPurgeOldEntries;
import com.onaro.sanscreen.server.sessions.LogSchedulerJob;
import com.onaro.sanscreen.server.sessions.LogSchedulerSession;
import org.quartz.CronScheduleBuilder;
import org.quartz.CronTrigger;
import org.quartz.JobBuilder;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.TriggerBuilder;
import org.quartz.impl.StdSchedulerFactory;
import org.slf4j.Logger;

import com.netapp.oci.netapp.ems.shared.interfaces.impl.NetappEmsEventPurgeJob;
import com.netapp.oci.netapp.performance.shared.interfaces.impl.NetappPerformanceDailyJob;
import com.netapp.oci.netapp.performance.shared.interfaces.impl.NetappPerformanceSummaryPurgeJob;
import com.netapp.oci.platform.plugin.remote.PlatformPluginRemote;

/**
 * Schedule quartz jobs and start scheduler
 *
 * @author Kosala Almeda
 *         Jan 16, 2015
 */
@Startup
@Singleton
@RunAs("admin") // @RunAs annotation does not work here. SecurityContext.setOutgoingRunAs method is used instead
public class JobSchedulerBean {

	public static final String QRTZ_JOB_GROUP = "AcqJob";

	private static final Logger logger =org.slf4j.LoggerFactory.getLogger(JobSchedulerBean.class);

	// ----- Data sources -----

	@Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSourceOnaroPool;

	@Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool_acq", type = DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
	private DataSource dataSourceAcqOnaroPool;

	@Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool_model", type = DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSourceModelOnaroPool;

	// ----- EJBs -----

	@EJB
    private ManagementRemote managementBean;

	@EJB
    private LogSchedulerSession logSchedulerBean;

	@EJB
	private PlatformPluginRemote platformPluginBean;


	@PostConstruct
	public void init() throws SchedulerException {

		try {

			Scheduler sched = new StdSchedulerFactory().getScheduler();

			// EJBs and DataSources
			JobDataMap jobDataMap = new JobDataMap();
			jobDataMap.put("dataSourceOnaroPool", dataSourceOnaroPool);
			jobDataMap.put("dataSourceAcqOnaroPool", dataSourceAcqOnaroPool);
			jobDataMap.put("dataSourceModelOnaroPool", dataSourceModelOnaroPool);
			jobDataMap.put("managementBean", managementBean);
			jobDataMap.put("logSchedulerBean", logSchedulerBean);
			jobDataMap.put("platformPluginBean", platformPluginBean);

			// ----- LogSchedulerJob -----

			JobDetail logSchedulerJob = JobBuilder.newJob(LogSchedulerJob.class)
					.withIdentity(LogSchedulerJob.JOB_NAME, QRTZ_JOB_GROUP)
					.usingJobData(jobDataMap)
					.build();

			CronTrigger logSchedulerTrigger = TriggerBuilder.newTrigger()
					.withIdentity(LogSchedulerJob.JOB_NAME, QRTZ_JOB_GROUP)
					.withSchedule(CronScheduleBuilder.cronSchedule(LogSchedulerJob.CRON_TRIGGER))
					.build();

			sched.scheduleJob(logSchedulerJob, logSchedulerTrigger);

			// ----- DailyManagementJob -----

			JobDetail dailyManagementJob = JobBuilder.newJob(DailyManagementJob.class)
					.withIdentity(DailyManagementJob.JOB_NAME, QRTZ_JOB_GROUP)
					.usingJobData(jobDataMap)
					.build();

			CronTrigger dailyManagementTrigger = TriggerBuilder.newTrigger()
					.withIdentity(DailyManagementJob.JOB_NAME, QRTZ_JOB_GROUP)
					.withSchedule(CronScheduleBuilder.cronSchedule(DailyManagementJob.CRON_TRIGGER))
					.build();

			sched.scheduleJob(dailyManagementJob, dailyManagementTrigger);

			// ----- DSAuditsPurgeOldEntries -----

			JobDetail dSAuditsPurgeOldEntriesJob = JobBuilder.newJob(DSAuditsPurgeOldEntries.class)
					.withIdentity(DSAuditsPurgeOldEntries.JOB_NAME, QRTZ_JOB_GROUP)
					.usingJobData(jobDataMap)
					.build();

			CronTrigger dSAuditsPurgeOldEntriesTrigger = TriggerBuilder.newTrigger()
					.withIdentity(DSAuditsPurgeOldEntries.JOB_NAME, QRTZ_JOB_GROUP)
					.withSchedule(CronScheduleBuilder.cronSchedule(DSAuditsPurgeOldEntries.CRON_TRIGGER))
					.build();

			sched.scheduleJob(dSAuditsPurgeOldEntriesJob, dSAuditsPurgeOldEntriesTrigger);

			// ----- ConnectionPoolPingJob -----

			JobDetail connectionPoolPingJob = JobBuilder.newJob(ConnectionPoolPingJob.class)
					.withIdentity(ConnectionPoolPingJob.JOB_NAME, QRTZ_JOB_GROUP)
					.usingJobData(jobDataMap)
					.build();

			CronTrigger connectionPoolPingTrigger = TriggerBuilder.newTrigger()
					.withIdentity(ConnectionPoolPingJob.JOB_NAME, QRTZ_JOB_GROUP)
					.withSchedule(CronScheduleBuilder.cronSchedule(ConnectionPoolPingJob.CRON_TRIGGER))
					.build();

			sched.scheduleJob(connectionPoolPingJob, connectionPoolPingTrigger);

			// ----- PruneJob -----

			JobDetail pruneJob = JobBuilder.newJob(PruneJob.class)
					.withIdentity(PruneJob.JOB_NAME, QRTZ_JOB_GROUP)
					.usingJobData(jobDataMap)
					.build();

			CronTrigger pruneTrigger = TriggerBuilder.newTrigger()
					.withIdentity(PruneJob.JOB_NAME, QRTZ_JOB_GROUP)
					.withSchedule(CronScheduleBuilder.cronSchedule(PruneJob.CRON_TRIGGER))
					.build();

			sched.scheduleJob(pruneJob, pruneTrigger);

			// ----- NetappPerformanceDailyJob -----

			JobDetail netappPerformanceDailyJob = JobBuilder.newJob(NetappPerformanceDailyJob.class)
					.withIdentity(NetappPerformanceDailyJob.JOB_NAME, QRTZ_JOB_GROUP)
					.usingJobData(jobDataMap)
					.build();

			CronTrigger netappPerformanceDailyTrigger = TriggerBuilder.newTrigger()
					.withIdentity(NetappPerformanceDailyJob.JOB_NAME, QRTZ_JOB_GROUP)
					.withSchedule(CronScheduleBuilder.cronSchedule(NetappPerformanceDailyJob.CRON_TRIGGER))
					.build();

			sched.scheduleJob(netappPerformanceDailyJob, netappPerformanceDailyTrigger);

			// ----- NetappPerformanceSummaryPurgeJob -----

			JobDetail netappPerformanceSummaryPurgeJob = JobBuilder.newJob(NetappPerformanceSummaryPurgeJob.class)
					.withIdentity(NetappPerformanceSummaryPurgeJob.JOB_NAME, QRTZ_JOB_GROUP)
					.usingJobData(jobDataMap)
					.build();

			CronTrigger netappPerformanceSummaryPurgeTrigger = TriggerBuilder.newTrigger()
					.withIdentity(NetappPerformanceSummaryPurgeJob.JOB_NAME, QRTZ_JOB_GROUP)
					.withSchedule(CronScheduleBuilder.cronSchedule(NetappPerformanceSummaryPurgeJob.CRON_TRIGGER))
					.build();

			sched.scheduleJob(netappPerformanceSummaryPurgeJob, netappPerformanceSummaryPurgeTrigger);


			// ----- NetappEMSEventPurgeJob -----
			JobDetail netappEmsEventPurgeJob = JobBuilder.newJob(NetappEmsEventPurgeJob.class)
                    .withIdentity(NetappEmsEventPurgeJob.JOB_NAME, QRTZ_JOB_GROUP)
                    .usingJobData(jobDataMap)
                    .build();

            CronTrigger netappEmsEventPurgeTrigger = TriggerBuilder.newTrigger()
                    .withIdentity(NetappEmsEventPurgeJob.JOB_NAME, QRTZ_JOB_GROUP)
                    .withSchedule(CronScheduleBuilder.cronSchedule(NetappEmsEventPurgeJob.CRON_TRIGGER))
                    .build();

            sched.scheduleJob(netappEmsEventPurgeJob, netappEmsEventPurgeTrigger);

			// ----- ... -----

			sched.start();

			logger.info("QRTZ jobs scheduled");

		} catch (SchedulerException e) {
			logger.error("Quartz Job Scheduling Failed", e);
			throw e;
		}
	}

}
