package com.netapp.oci.platform.originator;

import javax.naming.NamingException;

import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.Table;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.runner.DbRunner;
import com.netapp.oci.platform.plugin.data.PlatformPlugin;
import com.netapp.oci.server.TransactionContext;
import com.netapp.oci.server.TransactionContextException;
import com.netapp.oci.server.UpdateTask;
import com.netapp.oci.server.UpdateTaskException;
import com.onaro.commons.table.DbConfig;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import com.onaro.sanscreen.server.sessions.ImplicitCommitCheckRemote;
import com.onaro.sanscreen.server.sessions.SessionFactory;
import org.slf4j.Logger;

import java.sql.Connection;
import java.util.*;

/**
 * @author Iti Behari
 *         Date: 5/23/13
 */
public class OriginatorRemoveTask extends UpdateTask {
    private static final long serialVersionUID = 1L;
    private static Logger logger =org.slf4j.LoggerFactory.getLogger("com.netapp.oci.platform.originator");
    private final PlatformPlugin platformPlugin;
    int totalObjCount = 0;
    private long rootId;

    public OriginatorRemoveTask(final Connection connection, final PlatformPlugin platformPlugin, final Long originId, String originatorName) throws UpdateTaskException, OriginatorUpdateException {
        super(originatorName, originId);
        this.platformPlugin = platformPlugin;
        PlatformRegistry platformRegistry = platformPlugin.getRegistry();
        final OriginatorStateCache stateCache = OriginatorStateCacheFactory.getOriginatorCache(connection, originId, platformRegistry);
        this.totalObjCount = stateCache.getTotalObjCount();
        this.originatorName = originatorName;
    }

    public boolean run(final TransactionContext transactionContext) throws UpdateTaskException {
        logger.info("Begin removing Originator: " + getOriginatorId());
        boolean throwImplicitCommitCheckError = false;
        try {
            final Connection connection = transactionContext.getConnection();
            long time = transactionContext.getTime();
            initImplicitCommitCheck(connection, time);
            rootId = platformPlugin.getRootId(connection, getOriginatorId());
            deleteOriginatorDataFromAllTables(connection);
            logOriginatorChanges(connection, time);
            clearCache();
            throwImplicitCommitCheckError = true;
            return true; // What does it mean?
        } catch (Exception e) {
            throw new UpdateTaskException("Unable to remove Originator " + getOriginatorId(), e);
        } finally {
            assertImplicitCommitCheck(getOriginatorId(), originatorName, updateTime, throwImplicitCommitCheckError);
        }
    }

    private void clearCache() {
        OriginatorStateCacheFactory.removeOriginatorCache(getOriginatorId());
        KeyCache.removeOriginatorCache(getOriginatorId());
        KeyCache.commitOriginator(platformPlugin.getSchema(), getOriginatorId());
    }

    private void logOriginatorChanges(final Connection connection, long updateTime) throws ServerInterfaceException {
        platformPlugin.logOriginatorRemove(connection, getOriginatorId(), rootId, updateTime);

    }


    private void deleteOriginatorDataFromAllTables(final Connection connection) throws TableException {
        Map<String,DbRunner> tagToDbRunner = platformPlugin.getTagToDbRunners();
        final Map<String, String> subTableToBaseTable = platformPlugin.getRegistry().getSubTableToBaseTable();
        // Need to build a lookup of table name to tag. This is awkward. Must avoid Iti. I always made her work around this
        Map<String, String> tableToTag = buildTableToTagLookupMap(platformPlugin.getTagToXmlHandlers());

        // start delete
        List<String> deleteTableOrder = new ArrayList<String>(platformPlugin.getTableOrder());
        Collections.reverse(deleteTableOrder);
        for (String tableName : deleteTableOrder) {
            executeDeleteDbRunner(connection, tagToDbRunner, tableName, tableToTag.get(tableName));
            delete(connection, subTableToBaseTable, tableName);
        }
    }

    private void delete(Connection connection, Map<String, String> subTableToBaseTable, String tableName) throws TableException {
        if(subTableToBaseTable.containsKey(tableName)){
            // delete base table
            final String baseTable = subTableToBaseTable.get(tableName);
            PreparedStatementWrapper.createForUpdate(connection, String.format("DELETE FROM %s WHERE originId = ? ", baseTable)).setNext(getOriginatorId()).executeUpdateEx();
        }
        // and delete
        PreparedStatementWrapper.createForUpdate(connection, String.format("DELETE FROM %s WHERE originId = ? ", tableName)).setNext(getOriginatorId()).executeUpdateEx();
    }

    private void executeDeleteDbRunner(Connection connection, Map<String, DbRunner> tagToDbRunner, String tableName, String tag) throws TableException {
        // execute delete db runner but only for those objects that are still live
        DbRunner runner = tagToDbRunner.get(tag);
        if (runner != null) {
            Set<Long> ids = new HashSet<Long>(PreparedStatementWrapper.create(connection,
                    String.format("SELECT objid FROM %s WHERE originId=? AND objState='LIVE'", tableName),
                    TableResultParserDefault.LONG)
                    .setNext(getOriginatorId()).executeQueryEx());
            if (!ids.isEmpty()) {
                String tempTable = tableName + "_temp";
                String tempDatabase = tempTable.substring(0, tempTable.indexOf('.'));
                Table.createIdsTemporaryTableForJoin(connection, tempTable, ids, logger);
                runner.getDeleteExec(tempDatabase, ids).execute(connection, 0); // time parameter isn't used
            }
        }
    }

    private Map<String, String> buildTableToTagLookupMap(Map<String, XmlHandler> tagToXmlHandler) {
        Map<String, String> tableToTag = new HashMap<String, String>();
        for (Map.Entry<String, XmlHandler> entry : tagToXmlHandler.entrySet()) {
            tableToTag.put(entry.getValue().getTableName(), entry.getKey());
        }
        return tableToTag;
    }

    private void initImplicitCommitCheck(Connection connection, long updateTime) throws TableException, TransactionContextException {
        PreparedStatementWrapper.createForUpdate(connection, "UPDATE " + DbConfig.getSanscreenDbName() + ".implicit_commit SET starttime=" + updateTime).executeUpdateEx();
    }

    private void assertImplicitCommitCheck(long originId, String originatorName, long time, boolean throwException) throws UpdateTaskException {
        try {
            ImplicitCommitCheckRemote implicitCommitCheckRemote = SessionFactory.createSession(ImplicitCommitCheckRemote.JNDI_NAME);
            implicitCommitCheckRemote.assertImplicitCommitCheck(time, throwException);
        } catch (NamingException e) {
            logger.error("Error checking implicit commit, originId:" + originId + " originatorName:" + originatorName, e);
            if (throwException) {
                throw new UpdateTaskException("Error checking implicit commit, originId:" + originId + " originatorName:" + originatorName, e);
            }
        }
    }

    public boolean isDeleteRequired() {
        return totalObjCount > 0;
    }

    public long getUpdateTime() {
        return updateTime;
    }

    public int getTotalObjCount() {
        return totalObjCount;
    }

    public long getRootId() {
        return rootId;
    }
}
