package com.netapp.oci.platform.originator;

import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.KeyCacheException;
import com.netapp.oci.platform.originator.key.UuidKey;
import org.slf4j.Logger;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;


/**
 * A class to manage the slim cache of Object state for "cacheless" originator update handlers.  Tracks, by
 * tag, the "state" of an object, as well as it's most current checksum.     One will exist per Originator.   This
 * object is NOT thread safe.   The purpose here is twofold:   1) to cache the DB row checksum of each object to
 * allow quick comparison on new originator reports, and 2) to track the 'state' of the objects in the processing
 * of a originator report.
 * <p/>
 * <p>Assumes that only there is only one report session per originator at any given time!
 * <p/>
 * <p>State is initialized to DELETED for each existing object on a new originator report session, then changed as
 * parsing happens (to NO_CHANGE, or UPDATE for instance).  Additions are also processed by adding a new object
 * state with a CREATE state.   After all objects are parsed, the entire tree is walked and those that were not
 * touched (left in DELETE) are processed as a bulk delete and removed from the tree.
 *
 * @author Jason Sprague
 *         Date: Oct 20, 2011
 *         Time: 11:46:42 AM
 */

public final class OriginatorStateCache {

    private static final Logger logger = org.slf4j.LoggerFactory.getLogger("com.netapp.oci.platform.originatordump");

    public static class ChangeData {
        private Map<String, Set<Long>> addedObjs;
        private Map<String, Set<Long>> updatedObjs;
        private Map<String, Set<Long>> deletedObjs;
        private Map<String, Set<Long>> deletedHistoryObjs;

        public ChangeData(Map<String, Set<Long>> addedObjs, Map<String, Set<Long>> updatedObjs, Map<String, Set<Long>> deletedObjs, Map<String, Set<Long>> deletedHistoryObjs) {
            this.addedObjs = addedObjs;
            this.updatedObjs = updatedObjs;
            this.deletedObjs = deletedObjs;
            this.deletedHistoryObjs = deletedHistoryObjs;
        }

        public Map<String, Set<Long>> getAddedObjs() {
            return addedObjs;
        }

        public Map<String, Set<Long>> getUpdatedObjs() {
            return updatedObjs;
        }

        public Map<String, Set<Long>> getDeletedObjs() {
            return deletedObjs;
        }

        public Map<String, Set<Long>> getDeletedHistoryObjs() {
            return deletedHistoryObjs;
        }
    }

    //Tag Name -> UUID -> object state.
    private final Map<String, Map<UuidKey, ObjState>> cache;

    private final long originId;

    private PlatformRegistry platformRegistry;

    OriginatorStateCache(final Connection connection, final long originId, PlatformRegistry platformRegistry) throws OriginatorUpdateException {
        cache = new HashMap<String, Map<UuidKey, ObjState>>();
        this.originId = originId;

        this.platformRegistry = platformRegistry;

        initFromDB(connection, platformRegistry.getTagToXmlHandlers().values());
    }

    /**
     * Add a new object to the state cache.   Flags new state with CREATED state.
     *
     * @param tagName  tag name for the object (e.g. Vserver, Node, etc.)
     * @param uuid     the UUID of the object as reported
     * @param objId    the internal objId of the object
     * @param checksum the current checksum for data on the object.
     * @return the obj id of the object.
     */
    public long addObjState(final String tagName, final UuidKey uuid, final long objId, final long checksum,
                            final long changeDetailChecksum) {
        final Map<UuidKey, ObjState> objStates = getKeyToObjStateMap(tagName);
        final ObjState newState = new ObjState(objId, ObjState.State.CREATED, checksum, changeDetailChecksum);
        objStates.put(uuid, newState);
        return newState.getObjid();
    }

    /**
     * Flag the state of an existing object as 'UPDATED' - meaning some piece of data changed in the report.
     *
     * @param tagName  tag name for the object (e.g. Vserver, Node, etc.)
     * @param uuid     the UUID of the object as reported
     * @param checksum the current checksum for data on the object.
     * @return the obj id of the object.
     */
    public Long updateObjState(final String tagName, final UuidKey uuid, final long checksum, final long changeDetailChecksum) {
        final ObjState state = getObjectState(tagName, uuid);
        state.setState(ObjState.State.UPDATED);
        state.setChecksum(checksum);
        state.setChangeDetailChecksum(changeDetailChecksum);
        return state.getObjid();
    }

    /**
     * Flag the state of an existing object as 'UPDATED_TO_OFFLINE_OBJ' - meaning some piece of data changed in the report.
     *
     * @param tagName  tag name for the object (e.g. Vserver, Node, etc.)
     * @param uuid     the UUID of the object as reported
     * @param checksum the current checksum for data on the object.
     * @return the obj id of the object.
     */
    public Long updateOfflineObjState(final String tagName, final UuidKey uuid, final long checksum, final long changeDetailChecksum) {
        final ObjState state = getObjectState(tagName, uuid);
        state.setState(ObjState.State.UPDATED_TO_OFFLINE_OBJ);
        state.setChecksum(checksum);
        state.setChangeDetailChecksum(changeDetailChecksum);
        return state.getObjid();
    }


    /**
     * Flag the state of an existing object as 'NO_CHANGE' - meaning no date changed during the report.
     *
     * @param tagName tag name for the object (e.g. Vserver, Node, etc.)
     * @param uuid    the UUID of the object as reported
     * @return the obj id of the object.
     */
    public long retainObjState(final String tagName, final UuidKey uuid) {
        final ObjState state = getObjectState(tagName, uuid);
        state.setState(ObjState.State.NO_CHANGE);
        return state.getObjid();
    }

    /**
     * Flag the state of an existing object as 'NO_CHANGE_OFFLINE_OBJ' - meaning no date changed during the report.
     *
     * @param tagName tag name for the object (e.g. Vserver, Node, etc.)
     * @param uuid    the UUID of the object as reported
     * @return the obj id of the object.
     */
    public long retainOfflineObjState(final String tagName, final UuidKey uuid) {
        final ObjState state = getObjectState(tagName, uuid);
        state.setState(ObjState.State.NO_CHANGE_OFFLINE_OBJ);
        return state.getObjid();
    }


    /**
     * Replace the current key with the passed in key
     *
     * @param tagName tag name for the object (e.g. vVserver, Node, etc.)
     * @param uuidKey the shared key
     */
    public void replaceKey(String tagName, UuidKey uuidKey) {
        final Map<UuidKey, ObjState> objStates = getKeyToObjStateMap(tagName);
        objStates.put(uuidKey, objStates.remove(uuidKey));
    }

    /**
     * Modify the current keys with the passed in new keys
     *
     * @param tagName tag name for the object (e.g. vServer, Node, etc.)
     * @param uuidKeyMap the map of old uuidKeys to new UuidKeys
     */
    public synchronized void modifyUuidKeys(String tagName, Map<UuidKey, UuidKey> uuidKeyMap) {
       final Map<UuidKey, ObjState> objStates = getKeyToObjStateMap(tagName);
       for(Map.Entry<UuidKey, UuidKey> entry : uuidKeyMap.entrySet()) {
           ObjState objState = objStates.remove(entry.getKey());
           if (objState != null) {
               objStates.put(entry.getValue(), objState);
           }
       }

    }


    /**
     * Get the object state map for the given tag name.
     *
     * @param tagName tag name for the object (e.g. Vserver, Node, etc.)
     * @return requested map of object UUID -> object state.
     */
    public Map<UuidKey, ObjState> getKeyToObjStateMap(final String tagName) {
        if (!cache.containsKey(tagName)) {
            final Map<UuidKey, ObjState> objStates = new HashMap<UuidKey, ObjState>();
            cache.put(tagName, objStates);
        }
        return cache.get(tagName);
    }

    /**
     * Get the state of a single object.
     *
     * @param tagName tag name for the object (e.g. Vserver, Node, etc.)
     * @param uuid    the UUID of the object as reported
     * @return the object state.
     */
    public ObjState getObjectState(final String tagName, final UuidKey uuid) {
        final Map<UuidKey, ObjState> objStates = getKeyToObjStateMap(tagName);
        return objStates != null ? objStates.get(uuid) : null;
    }

    /**
     * Get all object states currently flagged for deletion.
     *
     * @return Map<tagName, Map<uuid, ObjState>> Map of all stateObjects currently flagged for deletion grouped by tagName
     */
    public Map<String, Map<UuidKey, ObjState>> getObjDeletes() {
        return getObjectsWithState(EnumSet.of(ObjState.State.DELETED, ObjState.State.DELETED_OFFLINE_OBJ));
    }


    /**
     * Get all <u> offline </u> object states currently offlined.
     *
     * @return Map<tagName, Map<uuid, ObjState>> Map of all stateObjects currently flagged for deletion grouped by tagName
     */
    public Map<String, Map<UuidKey, ObjState>> getOfflineObjDeletes() {

        final Map<String, Map<UuidKey, ObjState>> tagToOfflinedObjectsToState = new HashMap<String, Map<UuidKey, ObjState>>();
        final Set<String> tagsThatCanBeOfflined = platformRegistry.getTagToDependentChildrenForOfflineTags().keySet();

        for (String tag : tagsThatCanBeOfflined) {
            final Map<UuidKey, ObjState> objsForTag = cache.get(tag);
            if (objsForTag != null) {

                final Map<UuidKey, ObjState> offlinedObjStates = new HashMap<UuidKey, ObjState>();
                for (Map.Entry<UuidKey, ObjState> objForTagEntry : objsForTag.entrySet()) {
                    final ObjState objState = objForTagEntry.getValue();

                    if (objState.getState() == ObjState.State.DELETED_OFFLINE_OBJ) {
                        offlinedObjStates.put(objForTagEntry.getKey(), objState);
                    }
                }

                if (!offlinedObjStates.isEmpty()) {
                    tagToOfflinedObjectsToState.put(tag, offlinedObjStates);
                }
            }
        }
        return tagToOfflinedObjectsToState;
    }

    private Map<String, Map<UuidKey, ObjState>> getObjectsWithState(EnumSet<ObjState.State> statesToLookFor) {
        final Map<String, Map<UuidKey, ObjState>> tagToDeletes = new HashMap<String, Map<UuidKey, ObjState>>();
        for (final Map.Entry<String, Map<UuidKey, ObjState>> entry : cache.entrySet()) {

            final Map<UuidKey, ObjState> tagValues = entry.getValue();
            final Map<UuidKey, ObjState> deletedObjStates = new HashMap<UuidKey, ObjState>();

            for (final Map.Entry<UuidKey, ObjState> uuidToObjStateEntry : tagValues.entrySet()) {
                ObjState objState = uuidToObjStateEntry.getValue();
                if (statesToLookFor.contains(objState.getState())) {
                    deletedObjStates.put(uuidToObjStateEntry.getKey(), objState);
                }
            }

            if (!deletedObjStates.isEmpty()) {
                tagToDeletes.put(entry.getKey(), deletedObjStates);
            }

        }

        return tagToDeletes;
    }

    public Map<UuidKey, ObjState> getObjectsWithTagAndState(String tagName, EnumSet<ObjState.State> statesToLookFor) {
        final Map<UuidKey, ObjState> objStateMap = cache.get(tagName);
        final Map<UuidKey, ObjState> foundObjStates = new HashMap<UuidKey, ObjState>();

        if (objStateMap != null) {
            for (final Map.Entry<UuidKey, ObjState> uuidToObjStateEntry : objStateMap.entrySet()) {
                ObjState objState = uuidToObjStateEntry.getValue();
                if (statesToLookFor.contains(objState.getState())) {
                    foundObjStates.put(uuidToObjStateEntry.getKey(), objState);
                }
            }
        }

        return foundObjStates;
    }


    /**
     * Clear the cache.   Should be called once originator updates have been successfully persisted.
     * <p/>
     * <p>Walks the tree of state objects.   If the object is marked in deleted state, it is removed.
     */
    public Map<String, List<UuidKey>> clear() {
        final Map<String, List<UuidKey>> tagToUuid = new HashMap<String, List<UuidKey>>();
        for (final Map.Entry<String, Map<UuidKey, ObjState>> entry : cache.entrySet()) {

            final String tag = entry.getKey();
            final Map<UuidKey, ObjState> stateCache = entry.getValue();

            for (final Iterator<Map.Entry<UuidKey, ObjState>> it = stateCache.entrySet().iterator(); it.hasNext(); ) {
                Map.Entry<UuidKey, ObjState> stateCacheEntry = it.next();
                final ObjState state = stateCacheEntry.getValue();
                if (EnumSet.of(ObjState.State.DELETED, ObjState.State.DELETED_OFFLINE_OBJ).contains(state.getState())) {
                    UuidKey uuid = stateCacheEntry.getKey();
                    it.remove();

                    List<UuidKey> uuids = tagToUuid.get(tag);
                    if (uuids == null) {
                        uuids = new ArrayList<UuidKey>();
                        tagToUuid.put(tag, uuids);
                    }
                    uuids.add(uuid);
                }
            }
        }
        return tagToUuid;
    }

    /**
     * Provides information about the changes caused because of the latest originator update.   Should be called once
     * originator updates have been calculated .i.e. the update XML has been parsed.
     *
     * @return an array of <code>Map</code>s which contains the objIds values keyed by the XML tag name.  The list of
     *         elements in this array are in the following order: added objs, updated objs and deleted objs.
     * @throws OriginatorUpdateException 
     */
    public ChangeData getChanges(Connection connection) throws OriginatorUpdateException {
        Map<String, Set<Long>> addedObjs = new HashMap<String, Set<Long>>();
        Map<String, Set<Long>> updatedObjs = new HashMap<String, Set<Long>>();
        Map<String, Set<Long>> deletedObjs = new HashMap<String, Set<Long>>();
        Map<String, Set<Long>> deletedHistoryObjs = new HashMap<String, Set<Long>>();

        Map<String, XmlHandler> tagToXmlHandlers = platformRegistry.getTagToXmlHandlers();
        for (final Map.Entry<String, Map<UuidKey, ObjState>> entry : cache.entrySet()) {
        	
            final String tagName = entry.getKey();
            final XmlHandler xmlHandler = tagToXmlHandlers.get(tagName);
            final String parentTagName = platformRegistry.getParentRelationshipTagName(tagName);
            Map<UuidKey, ObjState> parentTagStateCache = null;
            if (parentTagName != null) {
                parentTagStateCache = cache.get(parentTagName);
            }
            final Map<UuidKey, ObjState> stateCache = entry.getValue();

            for (final UuidKey key : stateCache.keySet()) {
                final Long objId = stateCache.get(key).getObjid();
                final ObjState.State changeLogStatus = calculateChangeLogStatus(stateCache, parentTagStateCache, key);
                final String changeLogTagName = (parentTagName != null) ? parentTagName : tagName;
                if (changeLogStatus == ObjState.State.DELETED || changeLogStatus == ObjState.State.DELETED_OFFLINE_OBJ) {
                    if (xmlHandler.keepHistory()) {
                        add(deletedHistoryObjs, changeLogTagName, objId);
                    } else {
                        add(deletedObjs, changeLogTagName, objId);

                    }
                } else if (changeLogStatus == ObjState.State.CREATED) {
                    add(addedObjs, changeLogTagName, objId);
                } else if (changeLogStatus == ObjState.State.UPDATED || changeLogStatus == ObjState.State.UPDATED_TO_OFFLINE_OBJ) {
                    add(updatedObjs, changeLogTagName, objId);
                }
            }
        }
        
        for (final Map.Entry<String, Map<UuidKey, ObjState>> entry : cache.entrySet()) {
            final XmlHandler xmlHandler = tagToXmlHandlers.get(entry.getKey());            
            xmlHandler.adjustDeletedObjects(connection, deletedObjs, deletedHistoryObjs);
        }        
        
        return new ChangeData(addedObjs, updatedObjs, deletedObjs, deletedHistoryObjs);
    }

    /**
     * Is something in the tag being deleted
     *
     * @param tagname tag name to check
     * @return true if any object in that cache is being deleted for the tagname
     */
    public boolean isObjStateDelete(String tagname) {
        Collection<ObjState> cacheValues = cache.containsKey(tagname)
                ? cache.get(tagname).values()
                : Collections.<ObjState>emptyList();
        for (ObjState objState : cacheValues) {
            if (objState.getState() == ObjState.State.DELETED) {
                return true;
            }
        }
        return false;
    }

    /**
     * Calculates the change status for the currently processed tag.  It takes into account a tag's parent tag state,
     * if any, to determine its change status.
     * <p/>
     * For example, a parent tag might not have been changed.  But, its relationship tag might have got added during
     * this originator update.  In this case, the change status for its object type should be "updated" instead of
     * "added".
     *
     * @param stateCache       currently processed tag's state cache.
     * @param parentStateCache state cache of the parent of the currently processed tag.
     * @param key              key to the state cache. usually, UUID>
     * @return
     */
    public ObjState.State calculateChangeLogStatus(Map<UuidKey, ObjState> stateCache,
                                                   Map<UuidKey, ObjState> parentStateCache, UuidKey key) {
        ObjState.State currentTagState = stateCache.get(key).getState();

        // if there is no parent for the current tag, just return the current status.
        if (parentStateCache == null) {
            return currentTagState;
        }

        // if the control reaches here, it means that there is a parent tag associated with the current tag.

        // if nothing is changed on the current tag, just return "no change" as the change status of the parent tag
        // will determine the element's final change status.
        if (currentTagState == ObjState.State.NO_CHANGE || currentTagState == ObjState.State.NO_CHANGE_OFFLINE_OBJ) {
            return ObjState.State.NO_CHANGE;
        }

        ObjState.State parentTagState = parentStateCache.get(key).getState();
        if (parentTagState == currentTagState) {
            return ObjState.State.NO_CHANGE;  // do not add a new one.  to prevent duplicates
        }

//        assert parentTagState == ObjState.State.NO_CHANGE;

        return ObjState.State.UPDATED;
    }

    /**
     * Adds the provided object to the given map using the given <code>tagName</code> as the key.
     *
     * @param objsByTag the map to which the object needs to be added.
     * @param tagName   key for the 'to be added' object.
     * @param objId     the objectId to be added.
     */
    private void add(Map<String, Set<Long>> objsByTag, String tagName, Long objId) {
        Set<Long> objs = objsByTag.get(tagName);
        if (objs == null) {
            objs = new HashSet<Long>();
            objsByTag.put(tagName, objs);
        }
        objs.add(objId);
    }

    /**
     * Reset the state of every object to DELETED in preparation for processing a new originator update.
     */
    public void reset(boolean preserveOfflineObjectsFlag) {
        for (final Map.Entry<String, Map<UuidKey, ObjState>> entry : cache.entrySet()) {
            for (final ObjState objState : entry.getValue().values()) {
                final ObjState.State state = objState.getState();
                if (!preserveOfflineObjectsFlag) {
                    objState.setState(ObjState.State.DELETED);
                } else if (state == ObjState.State.UPDATED_TO_OFFLINE_OBJ || state == ObjState.State.NO_CHANGE_OFFLINE_OBJ) {
                    objState.setState(ObjState.State.DELETED_OFFLINE_OBJ);
                } else if (state != ObjState.State.PRESERVED) {
                    objState.setState(ObjState.State.DELETED);
                }
            }
        }
    }

    public void log() {
        if (logger.isDebugEnabled()) {
            for (final Map.Entry<String, Map<UuidKey, ObjState>> tagEntry : cache.entrySet()) {
                String tag = tagEntry.getKey();
                for (Map.Entry<UuidKey, ObjState> uuidEntry : tagEntry.getValue().entrySet()) {
                    logger.info(tag + ":" + uuidEntry.getKey() + ":" + uuidEntry.getValue().getState());
                }
            }
        }
    }


    /**
     * Initialize object state (and checksums) from DB.   Should only be called once at startup.
     *
     * @param connection DB connection
     * @throws OriginatorUpdateException for various connection and SQL failures.
     */
    private void initFromDB(final Connection connection, Collection<XmlHandler> xmlHandlers) throws OriginatorUpdateException {
        cache.clear();
        for (XmlHandler xmlHandler : xmlHandlers) {
            final String tagName = xmlHandler.getTagName();
            try {

                /**
                 * SELECT
                 *   o.objid, o.uuid, o.checksum, o.changeDetailChecksum
                 * FROM
                 *   <table> o
                 * WHERE originId = ?
                 *   AND inactive = false -- only if the table supports history
                 */
                String sql =
                        " SELECT" +
                                "   o.objid, o.uuid, o.checksum, o.changeDetailChecksum" +
                                " FROM" +
                                "   " + xmlHandler.getTableName() + " o" +
                                " WHERE" +
                                "   o.originId=? AND o.objState = 'LIVE'";
                if (xmlHandler.keepHistory()) {
                    sql = sql + " AND o.inactive=false";
                }

                // load originator data
                final KeyCache keyCache = KeyCache.get();
                final Map<Long, UuidKey> idToUuidKey = keyCache.getIdToKey(tagName, originId);
                List<OriginatorData> originatorDatas = PreparedStatementWrapper.create(connection, sql,
                        new TableResultParserDefault<OriginatorData>() {
                            @Override
                            public OriginatorData parseObject(ResultSet resultSet) throws SQLException {
                                return new OriginatorData(Long.valueOf(resultSet.getLong(1)), resultSet.getString(2), resultSet.getLong(3), resultSet.getLong(4));
                            }
                        }).setNext(originId).executeQueryEx();

                // build state cache
                for (OriginatorData originatorData : originatorDatas) {
                    final Map<UuidKey, ObjState> objStates;
                    Long objid = originatorData.getObjid();
                    long checksum = originatorData.getChecksum();
                    long changeDetailChecksum = originatorData.getChangeDetailChecksum();

                    if (cache.containsKey(tagName)) {
                        objStates = cache.get(tagName);
                    } else {
                        objStates = new HashMap<UuidKey, ObjState>();
                        cache.put(tagName, objStates);
                    }

                    UuidKey uuidKey = idToUuidKey.get(objid);
                    if (uuidKey == null) {
                        throw new KeyCacheException("Originator " + originId + ":" + tagName + ":" + objid + ": is not known to the keycache");
                    } else {
                        // This UUID is in the keycache. Conserve memory and re-use the stored objid
                        objid = keyCache.findId(originId, tagName, uuidKey);
                    }
                    objStates.put(uuidKey, new ObjState(objid, ObjState.State.DELETED, checksum, changeDetailChecksum));
                }
            } catch (final Exception te) {
                throw new OriginatorUpdateException("Unable to init handlers", te);
            }
        }
    }

    public int getTotalObjCount() {
        return cache.values().size();
    }

    private static class OriginatorData {
        private Long objid;
        private String uuid;
        private long checksum;
        private long changeDetailChecksum;

        private OriginatorData(Long objid, String uuid, long checksum, long changeDetailChecksum) {
            this.objid = objid;
            this.uuid = uuid;
            this.checksum = checksum;
            this.changeDetailChecksum = changeDetailChecksum;
        }

        public Long getObjid() {
            return objid;
        }

        public String getUuid() {
            return uuid;
        }

        public long getChecksum() {
            return checksum;
        }

        public long getChangeDetailChecksum() {
            return changeDetailChecksum;
        }
    }
}
