/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator;

import java.sql.Connection;
import java.util.HashMap;
import java.util.Map;

/**
 * Factory to get state cache for originators.  Note these 'cache' objects are not thread safe.   Usage patterns
 * assume each instance will be used by one 'update report' for a unique originator at any given time, though updates to
 * different state caches can happen concurrently.
 *
 * @author Jason Sprague
 *         Date: Oct 20, 2011
 *         Time: 8:52:00 PM
 */
public final class OriginatorStateCacheFactory {

    private final static Map<Long, OriginatorStateCache> origIdToCache= new HashMap<Long, OriginatorStateCache>();

    private  OriginatorStateCacheFactory() {}

    /**
     * Get a handle to the state cache for the given originator.  Note that the state cache is a 'singleton'
     * for each originator ID (persisted in memory).
     *
     *
     * @param connection  DB connection.
     * @param originatorId  the originator ID.
     * @param platformRegistry the registry for the schema
     * @return  the originator cache.
     * @throws OriginatorUpdateException   for problems initializing the cache.
     */
    public synchronized static OriginatorStateCache getOriginatorCache(final Connection connection, final Long originatorId, PlatformRegistry platformRegistry)
            throws OriginatorUpdateException {

        final OriginatorStateCache stateCache;
        if (origIdToCache.containsKey(originatorId)) {
            stateCache = origIdToCache.get(originatorId);
        }
        else {
            stateCache = new OriginatorStateCache(connection, originatorId, platformRegistry);
            origIdToCache.put(originatorId, stateCache);
        }
        return stateCache;
    }

    /**
     * Remove the given originator cache permanently.
     *
     * @param originatorId   the originator ID.
     */
    public synchronized static void removeOriginatorCache (final Long originatorId) {
        if (originatorId == null) {
            origIdToCache.clear();
        } else {
            origIdToCache.remove(originatorId);
        }
    }
}
