/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator;

import com.netapp.oci.platform.common.table.*;
import com.netapp.oci.platform.originator.db.*;
import com.netapp.oci.platform.originator.handler.*;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.KeyCacheException;
import com.netapp.oci.platform.originator.key.UuidKey;
import com.netapp.oci.platform.originator.runner.DbRunner;
import com.netapp.oci.platform.originator.runner.UnlinkObjectRunner;
import com.netapp.oci.platform.plugin.data.PlatformPlugin;
import com.onaro.commons.util.MapOfSets;
import com.onaro.commons.util.ObjectUtils;
import com.onaro.sanscreen.server.originatorupdate.xml.Element;
import org.slf4j.Logger;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

/**
 * @author Al Velella
 *         Date: Sep 22, 2011
 *         Time: 9:51:29 AM
 */
public class OriginatorUpdate {

    private static class ObjectRef {
        protected static final ObjectRef NULL = new ObjectRef(null, null);

        private Long id;
        private String uuid;

        public ObjectRef(Long id, String uuid) {
            this.id = id;
            this.uuid = uuid;
        }
    }
    
    private static class DoNotDelete {

        private String objectTagName;
        private String objectKey;
        private Map<String, String> alternateKey;

        public DoNotDelete() {
            alternateKey = new HashMap<String, String>();
        }

        public String getObjectTagName() {
            return objectTagName;
        }

        public void setObjectTagName(String objectTagName) {
            this.objectTagName = objectTagName;
        }

        public String getObjectKey() {
            return objectKey;
        }

        public void setObjectKey(String objectKey) {
            this.objectKey = objectKey;
        }

        public Map<String, String> getAlternateKey() {
            return alternateKey;
        }

        public void setAlternateKey(Map<String, String> alternateKey) {
            this.alternateKey = alternateKey;
        }
    }

    private static Logger logger = org.slf4j.LoggerFactory.getLogger("com.netapp.oci.platform.originator");

    /**
     * A cached map of all the columns to be copied into a delete detail record when that object is deleted
     */
    private final static Map<String, Set<String>> tagToDeleteDetailColumnNamesCache = new HashMap<String, Set<String>>();

    /**
     * the cache origin id must be integer
     */
    private Long originId;

    /**
     * Time data was received in server
     */
    private long startTime;

    /**
     * Xml handlers for currently parsing schema
     */
    private Map<String, XmlHandler> tagToXmlHandlers;

    /**
     * DB Runner handlers for currently parsing schema
     */
    private Map<String, DbRunner> tagToDbRunner;

    /**
     * Tag name to resource type
     */
    private Map<String, String> tagToResourceType;

    private Map<String, RunnerOperation> tableToDbRunnerOp;

    private Map<String, List<String>> tagToChangeDetailAttributes;

    private Map<String, Map<String, List<ObjectQueryRunner>>> tagNameToParentTagNameToOfflineRunnerOperation;
    private Map<String, Map<String, List<ObjectQueryRunner>>> tagNameToParentTagNameToOnlineRunnerOperation;


    /**
     * Update history table
     */
    private String updateHistoryTableName;

    /**
     * root tags
     */
    private Set<String> rootTags;

    /**
     * Delete detail table
     */
    private String deleteDetailTableName;

    /**
     * List of columns in delete_detail
     */
    private Map<String, String> deleteDetailColumnToTagname;

    /**
     * Table hierarchy. This is used during database operations to make sure foreign key relations can be added/removed
     */
    private List<String> tableOrder;

    /**
     * Stack of objects
     */
    private Stack<ObjectRef> objectStack;

    /**
     * Cache of parsed ColumnData
     */
    private Map<String, ColumnData> tableToColumnData;

    private Map<String, AbstractBatchOperation> tableToInsertOperation;

    private Map<String, AbstractBatchOperation> tableToUpdateOperation;

    private Map<String, Set<AbstractTemporaryTableOperation>> tableToDeleteOperation;

    private Map<String, InsertDeleteDetailOperation> tableToInsertDeleteDetailOperation;

    private Map<String, UpdateDeleteDetailOperation> tableToUpdateDeleteDetailOperation;

    /**
     * This contains the list of select queries by table that we need to make to populate
     * <code>attributeChangeDetails</code>.  This is populated during XML file parsing and the select queries are
     * executed in batch at the end of XML file processing.
     */
    private Map<String, ChangeDetailQueryData> tableToChangeDetailQueryData;

    private List<RunnerOperation> cleanupDbRunnerOps;

    /**
     * Contains the set of attribute change details as a result of this originator update.
     */
    private Set<AttributeChangeDetail> attributeChangeDetails;

    private OriginatorStateCache stateCache;

    private KeyCache keyCache;

    private boolean preserveOfflineObjectsFlag;

    private PlatformRegistry platformRegistry;
    private Map<String, Set<UuidKey>> tagToUuidKeysToPurge;

    private DoNotDelete doNotDelete;
    private List<DoNotDelete> doNotDeleteList = new ArrayList<DoNotDelete>();

    public OriginatorUpdate(Connection connection, Long originId, long startTime, PlatformPlugin platformPlugin, boolean preserveOfflineObjectsFlag) throws OriginatorUpdateException {
        this.originId = originId;
        this.startTime = startTime;
        this.platformRegistry = platformPlugin.getRegistry();
        this.tagToXmlHandlers = platformPlugin.getTagToXmlHandlers();
        this.tagToDbRunner = platformPlugin.getTagToDbRunners();
        this.tagToResourceType = platformPlugin.getTagToResourceType();
        this.tableToDbRunnerOp = new HashMap<String, RunnerOperation>();
        this.updateHistoryTableName = platformPlugin.getUpdateHistoryTableName();
        this.deleteDetailTableName = platformPlugin.getDeleteDetailTableName();
        this.rootTags = platformPlugin.getRootTagnames();
        this.deleteDetailColumnToTagname = platformPlugin.getDeleteDetailColumnToTagname();
        this.tagToChangeDetailAttributes = platformPlugin.getTagToChangeDetailAttributes();
        this.tableOrder = platformPlugin.getTableOrder();
        this.objectStack = new Stack<ObjectRef>();
        this.tableToColumnData = new HashMap<String, ColumnData>();
        this.tableToInsertOperation = new HashMap<String, AbstractBatchOperation>();
        this.tableToUpdateOperation = new HashMap<String, AbstractBatchOperation>();
        this.tableToDeleteOperation = new HashMap<String, Set<AbstractTemporaryTableOperation>>();
        this.tableToInsertDeleteDetailOperation = new HashMap<String, InsertDeleteDetailOperation>();
        this.tableToUpdateDeleteDetailOperation = new HashMap<String, UpdateDeleteDetailOperation>();
        this.cleanupDbRunnerOps = new ArrayList<RunnerOperation>();
        this.tableToChangeDetailQueryData = new HashMap<String, ChangeDetailQueryData>();
        this.attributeChangeDetails = new HashSet<AttributeChangeDetail>();
        this.stateCache = OriginatorStateCacheFactory.getOriginatorCache(connection, originId, platformRegistry);
        this.keyCache = KeyCache.get();
        this.stateCache.reset(preserveOfflineObjectsFlag);
        loadColumnsForDeleteDetail(connection, this.tagToXmlHandlers.keySet(), this.deleteDetailColumnToTagname.keySet());
        this.tagNameToParentTagNameToOfflineRunnerOperation = new HashMap<String, Map<String, List<ObjectQueryRunner>>>();
        this.tagNameToParentTagNameToOnlineRunnerOperation = new HashMap<String, Map<String, List<ObjectQueryRunner>>>();
        this.preserveOfflineObjectsFlag = preserveOfflineObjectsFlag;
        this.platformRegistry = platformPlugin.getRegistry();
        this.tagToUuidKeysToPurge = new HashMap<String, Set<UuidKey>>();
    }

    public Long getOriginId() {
        return originId;
    }

    public void update(Element element) throws OriginatorUpdateException {
        // Special case for empty stack, but otherwise test the top of objectStack to see if there's a valid originObject
        // there. A non-null object means we are actively parsing elements of the XML
        if (objectStack.isEmpty() || objectStack.peek() != null) {
            // Lookup handler for this element
            XmlHandler databaseHandler = tagToXmlHandlers.get(element.tagName);

            ObjectRef object;
            if (databaseHandler == null) {
                if (objectStack.isEmpty()) {
                    // There's no handler for this element but there could be one for a child element. We will push a
                    // dummy originObject to the top of the stack. This preserves the stack depth and signals to the
                    // child elements that we are still actively parsing elements of the XML
                    object = ObjectRef.NULL;
                } else {
                    // There's no handler for this element but there could be one for a child element. We will push a copy
                    // of the current originObject to the top of the stack. This preserves the stack depth and signals
                    // to the child elements that we are still actively parsing elements of the XML
                    object = peekObjectRef();
                }
                
                Map<String, String> properties = element.prop;
                if (element.tagName.equals("DoNotDelete")) {
                    doNotDelete = new DoNotDelete();
                    doNotDelete.setObjectTagName(properties.get("ObjectTagName"));
                    doNotDelete.setObjectKey(properties.get("ObjectKey"));
                    doNotDeleteList.add(doNotDelete);
                }
                else if (element.tagName.equals("AlternateKey")) {
                    doNotDelete.getAlternateKey().put(properties.get("Name"), properties.get("Value"));
                }               
            } else {
                // There's a handler for this element. Run it and save its return originObject. If the handler returns a null
                // originObject it will stop parsing of child elements
                Map<String, String> properties = element.prop;
                final Map<UuidKey, ObjState> keyToObjState = stateCache.getKeyToObjStateMap(element.tagName);
                Map<String, List<XmlChildTagDetails>> tagToDependentChildrenForOfflineTags = platformRegistry.getTagToDependentChildrenForOfflineTags();

                try {
                    Long id = databaseHandler.handleObject(this, properties, keyToObjState,
                            tagToDependentChildrenForOfflineTags.containsKey(element.tagName) ? tagToDependentChildrenForOfflineTags.get(element.tagName) : Collections.<XmlChildTagDetails>emptyList());
                    object = new ObjectRef(id, properties.get(databaseHandler.getKeyAttr()));
                } catch (KeyCacheException e) {
                    throw new OriginatorUpdateException("Key-cache exception", e);
                }
            }
            pushObject(object);
        } else {
            // A parent handler set the top of of the originObjectStack to null, meaning we must ignore
            // all children handlers. Continue to push null to preserve stack depth and to signal to next child
            // that it should ignore parsing
            pushObject(null);
        }
    }

    public void queueDelete() throws OriginatorUpdateException {
        // do not generate delete-detail records if any of the root-nodes have been deleted
        boolean isRootDelete = isRootDelete();

        Map<String, Map<UuidKey, ObjState>> tagToDeletes = stateCache.getObjDeletes();

        Map<String, Map<String, List<String>>> tagToColumnsToUnlinkOnDelete = platformRegistry.getTagToColumnsToUnlinkOnDelete();
        for (String currentTagName : tagToDeletes.keySet()) {
            // Queue up delete xml handler operations
            final XmlHandler xmlHandler = tagToXmlHandlers.get(currentTagName);
            String tablename = xmlHandler.getTableName();
            String baseTableName = xmlHandler.getBaseTableName();
            boolean isHierarchy = baseTableName != null;

            Map<UuidKey, ObjState> keysToDeleteForCurrentTag = tagToDeletes.get(currentTagName);
            Set<Long> idsToDelete = new HashSet<Long>();

            String relationshipParentTagName = getParentRelationshipTagName(currentTagName);

            if (relationshipParentTagName == null) {
                AbstractTemporaryTableOperation deleteOp;
                AbstractTemporaryTableOperation baseDeleteOp = null;
                if (xmlHandler.keepHistory()) {
                    deleteOp = new MarkForDeletionOperation(tablename);
                } else {
                    if (isHierarchy) {
                        baseDeleteOp = new SoftDeleteOperation(baseTableName);
                    }
                    deleteOp = new SoftDeleteOperation(tablename);
                }
                for (ObjState objState : keysToDeleteForCurrentTag.values()) {
                    Long objid = objState.getObjid();
                    deleteOp.addId(objid);
                    idsToDelete.add(objid);
                    if (isHierarchy) {
                        baseDeleteOp.addId(objid);
                    }
                }
                Set<AbstractTemporaryTableOperation> deleteOps = new HashSet<AbstractTemporaryTableOperation>(2);
                deleteOps.add(deleteOp);
                if (isHierarchy) {
                    deleteOps.add(baseDeleteOp);
                }
                tableToDeleteOperation.put(tablename, deleteOps);

            } else {
                // special case for handling deletion of the relationship objects.
                // Hard delete the relationship objects if the parent object is still live
                // Soft delete the relationship object if the parent object is also dead.
                AbstractTemporaryTableOperation deleteOp = new DeleteOperation(tablename);
                AbstractTemporaryTableOperation softDeleteOp = new SoftDeleteOperation(tablename);

                for (Map.Entry<UuidKey, ObjState> keyToDelete : keysToDeleteForCurrentTag.entrySet()) {
                    UuidKey uuidKey = keyToDelete.getKey();
                    Long objid = keyToDelete.getValue().getObjid();
                    idsToDelete.add(objid);
                    boolean isParentAlsoGettingDeleted = tagToDeletes.containsKey(relationshipParentTagName) && tagToDeletes.get(relationshipParentTagName).containsKey(uuidKey);
                    if (isParentAlsoGettingDeleted) {
                        softDeleteOp.addId(objid);
                    } else {
                        deleteOp.addId(objid);
                        Set<UuidKey> uuidKeysToPurge = tagToUuidKeysToPurge.get(currentTagName);
                        if (uuidKeysToPurge == null) {
                            uuidKeysToPurge = new HashSet<UuidKey>();
                            tagToUuidKeysToPurge.put(currentTagName, uuidKeysToPurge);
                        }
                        uuidKeysToPurge.add(uuidKey);
                    }
                }
                Set<AbstractTemporaryTableOperation> operations = new HashSet<AbstractTemporaryTableOperation>(2);
                if (deleteOp.getOperationsCount() > 0) {
                    operations.add(deleteOp);
                }

                if (softDeleteOp.getOperationsCount() > 0) {
                    operations.add(softDeleteOp);
                }

                tableToDeleteOperation.put(tablename, operations);


            }

            // Queue up any runner cleanup operations. These will execute before the standard insert/update/delete xml handler
            // operations run.
            DbRunner dbRunner = tagToDbRunner.get(currentTagName);
            if (dbRunner != null) {
                String tempDatabase = deleteDetailTableName.substring(0, deleteDetailTableName.indexOf('.'));
                RunnerOperation runnerOperation = dbRunner.getDeleteExec(tempDatabase, idsToDelete);
                cleanupDbRunnerOps.add(runnerOperation);
            }

            // Queue up unlink object operations
            Map<String, List<String>> objectsToUnlink = tagToColumnsToUnlinkOnDelete.containsKey(currentTagName) ? tagToColumnsToUnlinkOnDelete.get(currentTagName) : Collections.<String, List<String>>emptyMap();
            if (!objectsToUnlink.isEmpty()) {
                cleanupDbRunnerOps.add(new UnlinkObjectRunner().getRunnerOp("management", xmlHandler.getTagName(), idsToDelete, objectsToUnlink));
            }
            // Generate changedetail if it's relationship object and contains attributes that need to be recorded
            boolean isRelationshipObject = getParentRelationshipTagName(currentTagName) != null;
            boolean shouldChangeDetailBeRecorded = !getChangeDetailAttributes(currentTagName).isEmpty();
            if (isRelationshipObject && shouldChangeDetailBeRecorded) {
                generateChangeDetailForDeletedRelationshipObjects(currentTagName, xmlHandler, tablename, keysToDeleteForCurrentTag);
            }

            // build sql to create change_delete_detail records
            String resourceType = tagToResourceType.get(currentTagName);
            if (!isRootDelete && resourceType != null && !deleteDetailColumnToTagname.isEmpty()) {
                Set<String> columnNames = tagToDeleteDetailColumnNamesCache.get(currentTagName);
                if (!columnNames.isEmpty()) {
                    String tempTableName = deleteDetailTableName + "_temp";

                    // create a delete-detail record and copy targeted columns
                    InsertDeleteDetailOperation insertDeleteOp
                            = new InsertDeleteDetailOperation(
                            columnNames,
                            tempTableName,
                            deleteDetailTableName,
                            tablename,
                            resourceType,
                            deleteDetailColumnToTagname.keySet(),
                            new ArrayList<ObjState>(keysToDeleteForCurrentTag.values()));
                    tableToInsertDeleteDetailOperation.put(tablename, insertDeleteOp);

                    // update a previously created delete-detail record and set the ancestor-deleted flag
                    for (Map.Entry<String, String> entry : deleteDetailColumnToTagname.entrySet()) {
                        if (currentTagName.equals(entry.getValue())) {
                            UpdateDeleteDetailOperation updateDeleteDetailOperation
                                    = new UpdateDeleteDetailOperation(tempTableName, deleteDetailTableName, idsToDelete, entry.getKey());
                            tableToUpdateDeleteDetailOperation.put(tablename, updateDeleteDetailOperation);
                            break;
                        }
                    }
                }
            }
        }
    }

    /**
     * Iterate over all objects marked to be deleted for the tag
     * and generate a changeDetail only if the relationship object is getting deleted but it's parent object is still live.
     *
     * @param tagName
     * @param xmlHandler
     * @param tablename
     * @param deletedObjStateMap
     */
    private void generateChangeDetailForDeletedRelationshipObjects(String tagName, XmlHandler xmlHandler, String tablename, Map<UuidKey, ObjState> deletedObjStateMap) {
        for (Map.Entry<UuidKey, ObjState> objStateForTag : deletedObjStateMap.entrySet()) {

            // isChangeDetailRequired figures out if the parent is still live but relationship object is getting deleted.
            if (isChangeDetailRequired(null, xmlHandler, objStateForTag.getKey())) {
                // We are going to create dummy rowData with null values because the object is being deleted.

                List<String> columnNamesToRecord = getChangeDetailAttributes(tagName);
                int numberOfCols = columnNamesToRecord.size();
                List<Object> rowDataToRecord = new ArrayList<Object>(numberOfCols);
                for (String ignored : columnNamesToRecord) {
                    rowDataToRecord.add(null);
                }
                addToChangeDetailQueryData(columnNamesToRecord, objStateForTag.getValue().getObjid(), rowDataToRecord, tagName, tablename);
            }
        }
    }

    private void pushObject(ObjectRef object) {
        objectStack.push(object);
    }

    private ObjectRef peekObjectRef() {
        return objectStack.peek();
    }

    public Long peekObjectId() {
        return objectStack.peek().id;
    }

    public String peekObjectUuid() {
        return objectStack.peek().uuid;
    }

    public Long peekObjectId(int position) {
        return objectStack.get(objectStack.size() - 1 - position).id;
    }

    public void popObject() {
        objectStack.pop();
    }

//    public boolean isDelete(String tag, String key) {
//        return ObjState.State.DELETED == stateCache.getObjectState(tag, platformRegistry.buildUuidKey(tag, key)).calcState();
//    }

    public int getChangeCount() {
        return getOperationsCount(tableToInsertOperation) + getOperationsCount(tableToUpdateOperation) + getOperationsCountForSetDbOperation(tableToDeleteOperation);
    }

    public int getRunnerCount() {
        int count = 0;
        for (RunnerOperation runnerOperation : tableToDbRunnerOp.values()) {
            count += runnerOperation.getCount();
        }
        for (RunnerOperation runnerOperation : cleanupDbRunnerOps) {
            count += runnerOperation.getCount();
        }
        return count;
    }

    public boolean isChanged() {
        return getChangeCount() > 0;
    }

    public Long insertObject(HandlerResult handlerResult, XmlHandler xmlHandler) {
        final String tableName = xmlHandler.getTableName();
        final RowData row = handlerResult.getRowData();

        addInsertOperation(tableName, xmlHandler.keepHistory(), handlerResult.getColumnData(), row);

        if (handlerResult.isHierarchy()) {
            addBaseInsertOperation(xmlHandler.getBaseTableName(), handlerResult.getColumnData(), row);
        }

        final String tagName = xmlHandler.getTagName();
        addRunnerOperations(handlerResult, tagName, tableName);

        Long objId = stateCache.addObjState(tagName, handlerResult.getUuidKey(), row.getObjid(), row.getChecksum(), row.getChangeDetailChecksum());

        if (isChangeDetailRequired(handlerResult, xmlHandler, handlerResult.getUuidKey())) {
            addToChangeDetailQueryData(xmlHandler, handlerResult);
        }
        return objId;
    }

    private void addInsertOperation(String tableName, boolean historyTable, ColumnData columnData, RowData row) {
        AbstractBatchOperation insertOp = tableToInsertOperation.get(tableName);
        if (insertOp == null) {
            insertOp = new InsertOperation(tableName, historyTable, columnData);
            tableToInsertOperation.put(tableName, insertOp);
        }
        insertOp.addRow(row);
    }

    private void addBaseInsertOperation(String tableName, ColumnData columnData, RowData row) {
        AbstractBatchOperation insertOp = tableToInsertOperation.get(tableName);
        if (insertOp == null) {
            insertOp = new BaseInsertOperation(tableName, columnData);
            tableToInsertOperation.put(tableName, insertOp);
        }
        insertOp.addRow(row);
    }

    public Long updateObject(HandlerResult handlerResult, XmlHandler xmlHandler, boolean isObjOffline) {
        final String tableName = xmlHandler.getTableName();
        RowData row = handlerResult.getRowData();
        addUpdateOperation(tableName, handlerResult.getColumnData(), xmlHandler.keepHistory(), row);
        if (handlerResult.isHierarchy()) {
            addBaseUpdateOperation(xmlHandler.getBaseTableName(), handlerResult.getColumnData(), row);
        }

        final String tagName = xmlHandler.getTagName();
        addRunnerOperations(handlerResult, tagName, tableName);


        if (handlerResult.hasChangeDetailReqAttributesChanged()) {
            addToChangeDetailQueryData(xmlHandler, handlerResult);
        }

        if (isObjOffline) {
            return stateCache.updateOfflineObjState(tagName, handlerResult.getUuidKey(), row.getChecksum(), row.getChangeDetailChecksum());
        }
        return stateCache.updateObjState(tagName, handlerResult.getUuidKey(), row.getChecksum(), row.getChangeDetailChecksum());
    }

    private void addUpdateOperation(String tableName, ColumnData columnData, boolean historyTable, RowData row) {
        AbstractBatchOperation updateOp = tableToUpdateOperation.get(tableName);
        if (updateOp == null) {

            updateOp = new UpdateOperation(tableName, historyTable, columnData);
            tableToUpdateOperation.put(tableName, updateOp);
        }
        updateOp.addRow(row);
    }

    private void addBaseUpdateOperation(String tableName, ColumnData columnData, RowData row) {
        AbstractBatchOperation updateOp = tableToUpdateOperation.get(tableName);
        if (updateOp == null) {

            updateOp = new BaseUpdateOperation(tableName, columnData);
            tableToUpdateOperation.put(tableName, updateOp);
        }
        updateOp.addRow(row);
    }

    public void addOfflineRunners(Long objId, String tagName, List<XmlChildTagDetails> offlineChildren) {
        addObjectQueryRunner(tagNameToParentTagNameToOfflineRunnerOperation, tagName, offlineChildren, objId);
    }

    public void addOnlineRunners(Long objId, String tagName, List<XmlChildTagDetails> onlineChildren) {
        addObjectQueryRunner(tagNameToParentTagNameToOnlineRunnerOperation, tagName, onlineChildren, objId);
    }

    private void addObjectQueryRunner(Map<String, Map<String, List<ObjectQueryRunner>>> tagNameToParentNameToObjectQueryRunner, String parentTagName, List<XmlChildTagDetails> dependentChildren, Long objId) {
        for (XmlChildTagDetails offlineChild : dependentChildren) {

            final String childTagName = offlineChild.getChildTagName();
            Map<String, List<ObjectQueryRunner>> parentTagToRunner = tagNameToParentNameToObjectQueryRunner.get(childTagName);

            if (parentTagToRunner == null) {
                parentTagToRunner = new HashMap<String, List<ObjectQueryRunner>>();
                tagNameToParentNameToObjectQueryRunner.put(childTagName, parentTagToRunner);
            }

            List<ObjectQueryRunner> runners = parentTagToRunner.get(parentTagName);
            if (runners == null) {
                ObjectQueryRunner runner = new ObjectQueryRunner(offlineChild);
                runners = new LinkedList<ObjectQueryRunner>();
                runners.add(runner);

                parentTagToRunner.put(parentTagName, runners);
            } else if (isNewRunnerNeeded(runners, offlineChild)) {
                ObjectQueryRunner runner = new ObjectQueryRunner(offlineChild);
                runners.add(runner);
            }

            for (ObjectQueryRunner it: runners) {
                it.addId(objId);
            }
        }
    }

    private boolean isNewRunnerNeeded(List<ObjectQueryRunner> runners, XmlChildTagDetails offlineChild) {
        for (ObjectQueryRunner it: runners) {
            if(it.createdFromSameXmlChildTagDetails(offlineChild)) {
                 return false;
            }
        }

        return true;
    }

    public String getParentRelationshipTagName(String tagName) {
        return platformRegistry.getParentRelationshipTagName(tagName);
    }

    private boolean isChangeDetailRequired(HandlerResult handlerResult, XmlHandler xmlHandler, UuidKey key) {
        String tagName = xmlHandler.getTagName();
        String parentTagName = getParentRelationshipTagName(tagName);

        Map<UuidKey, ObjState> objStates = stateCache.getKeyToObjStateMap(tagName);
        Map<UuidKey, ObjState> parentObjStates = null;
        if (parentTagName != null) {
            parentObjStates = stateCache.getKeyToObjStateMap(parentTagName);
        }

        ObjState.State logStatus = stateCache.calculateChangeLogStatus(objStates, parentObjStates, key);

        if (logStatus == ObjState.State.UPDATED || logStatus == ObjState.State.UPDATED_TO_OFFLINE_OBJ) {
            if (handlerResult != null) {
                return handlerResult.hasChangeDetailReqAttributesChanged();
            } else {
                // if there is no handlerResult, that means relationship object is deleted but not its parent object.
                // So, check if changeDetail is required for the relationship object.
                List<String> changeDetailAttributes = getChangeDetailAttributes(tagName);
                return !changeDetailAttributes.isEmpty();
            }
        }

        return false;
    }

    private void addToChangeDetailQueryData(XmlHandler xmlHandler, HandlerResult handlerResult) {

        //iterate over all attributeNames that need to be recorded and get their row value.
        List<String> changeDetailAttributes = getChangeDetailAttributes(xmlHandler.getTagName());
        List<Object> rowDataListToRecord = new LinkedList<Object>();
        List<Object> baseRowDataListToRecord = new LinkedList<Object>();
        for (String attributeName : changeDetailAttributes) {
            String colName = attributeName.toLowerCase();
            if (handlerResult.isBaseColumn(colName)) {
                baseRowDataListToRecord.add(handlerResult.getBaseRowDataValue(colName));
            }

            if (handlerResult.isSubColumn(colName)) {
                rowDataListToRecord.add(handlerResult.getRowDataValue(colName));
            }

            if (!baseRowDataListToRecord.isEmpty()) {
                addToChangeDetailQueryData(changeDetailAttributes, handlerResult.getObjId(), baseRowDataListToRecord, xmlHandler.getTagName(), xmlHandler.getBaseTableName());

            }
            if (!rowDataListToRecord.isEmpty()) {
                addToChangeDetailQueryData(changeDetailAttributes, handlerResult.getObjId(), rowDataListToRecord, xmlHandler.getTagName(), xmlHandler.getTableName());

            }
        }
    }

    private void addToChangeDetailQueryData(List<String> columnNamesToRecord, Long objId, List<Object> rowData, String tagName, String tableName) {
        ChangeDetailQueryData changeDetailQueryData = tableToChangeDetailQueryData.get(tableName);
        if (changeDetailQueryData == null) {
            changeDetailQueryData = new ChangeDetailQueryData(columnNamesToRecord, tagName, tableName);
            tableToChangeDetailQueryData.put(tableName, changeDetailQueryData);
        }
        changeDetailQueryData.addChangedRow(objId, rowData);
    }


    /**
     * Gather the detailed attribute change information by using the change detail query data that was stored while
     * parsing the XML schema.  Change detail query data contains for each table, what are the attributes that need
     * detailed change information as well as the rows that has at least one of those attributes whose values got
     * changed during this originator update.
     */
    @SuppressWarnings("deprecation")
    public void gatherDetailedAttributeChangeInfo(Connection connection) throws TableException {
        boolean debug = logger.isDebugEnabled();

        // do the change detail query for each table.
        Set<String> tables = tableToChangeDetailQueryData.keySet();
        if (tables.size() > 0) {
            String schemaName = tables.iterator().next().split("\\.")[0];
            String tempTableName = schemaName + ".change_detail_temp_table";
            try {
                // It seems that MySQL performance degrades, when the select query has lot of values in the "IN"
                // clause. Therefore, we create a temporary table to hold the obj ids and use JOIN.
                Table.createIdsTemporaryTableForJoin(connection, tempTableName, Collections.<Long>emptySet(), logger);

                for (String tableName : tables) {
                    final ChangeDetailQueryData changeDetailQueryData = tableToChangeDetailQueryData.get(tableName);

                    Set<Long> objIds = changeDetailQueryData.getChangedAttributeValues().keySet();

                    // populate the temporary table with the changed row's obj ids.

                    Table.populateTempTable(connection, tempTableName, objIds, logger);


                    // select query is of the form: select column1, column2 .., columnn, objid from table t JOIN tempTable tt ON tt.id = t.objId;
                    // where 'n' is the number of columns that need detailed change information for the table.
                    String sql = changeDetailQueryData.getSql(tempTableName);

                    PreparedStatementWrapper preparedStatement = PreparedStatementWrapper.create(connection, sql,
                            new TableResultParserDefault<Object>() {

                                // for each of the rows, check the DB value of each attribute against the value in originator
                                // update XML and if there is a change, add it to the attribute change details set for
                                // persisting into DB later.
                                @Override
                                public Object parseObject(final ResultSet resultSet) throws SQLException {
                                    Map<Long, List<Object>> changedRows = changeDetailQueryData.getChangedAttributeValues();
                                    Long objId = resultSet.getLong("objId");
                                    List<Object> changedRow = changedRows.get(objId);
                                    String tagName = changeDetailQueryData.getTagName();
                                    int columnIndex = 0;
                                    for (String column : changeDetailQueryData.getColumns()) {
                                        Object valueInDb = resultSet.getObject(columnIndex + 1);   // result set column index start from 1.
                                        Object valueInXml = changedRow.get(columnIndex);
                                        if (!ObjectUtils.isEqual(valueInDb, valueInXml)) {
                                            attributeChangeDetails.add(new AttributeChangeDetail(objId, tagName, column.toLowerCase(), valueInDb, valueInXml));
                                        }
                                        columnIndex++;
                                    }
                                    return null;
                                }
                            });
                    if (debug) {
                        logger.info(preparedStatement + ";");
                    }
                    preparedStatement.executeQueryEx();

                    // deletes all the rows in the temp table.  But, the table remains.
                    Table.clearTempTable(connection, tempTableName, logger);
                }
            } finally {
                Table.dropTemporaryTable(connection, tempTableName, logger);
            }
        }

        // at this point, we need not store change detail query data in memory any more.
        tableToChangeDetailQueryData = null;
    }

    public Long retainObject(HandlerResult handlerResult, String tagname, boolean isObjOffline) {
        final UuidKey uuid = handlerResult.getUuidKey();

        if (isObjOffline) {
            return stateCache.retainOfflineObjState(tagname, uuid);
        }
        return stateCache.retainObjState(tagname, uuid);
    }

    public void submit(Connection connection, long time) throws TableException {
        List<String> reverseTableOrder = new ArrayList<String>(this.tableOrder);
        Collections.reverse(reverseTableOrder);

        // debug state cache hook
        if (logger.isDebugEnabled()) {
            stateCache.log();
        }

        for (RunnerOperation runnerOperation : cleanupDbRunnerOps) {
            runnerOperation.execute(connection, time);
        }

        final Map<String, String> subTableToBaseTable = platformRegistry.getSubTableToBaseTable();

        for (String table : tableOrder) {

            if (subTableToBaseTable.containsKey(table)) {
                // Need to insert/update the base table first
                final String baseTable = subTableToBaseTable.get(table);
                DbOperation insertOp = tableToInsertOperation.get(baseTable);

                if (insertOp != null) {
                    insertOp.execute(connection, time);
                }
                DbOperation updateOp = tableToUpdateOperation.get(baseTable);
                if (updateOp != null) {
                    updateOp.execute(connection, time);
                }
            }

            DbOperation insertOp = tableToInsertOperation.get(table);

            if (insertOp != null) {
                insertOp.execute(connection, time);
            }
            DbOperation updateOp = tableToUpdateOperation.get(table);
            if (updateOp != null) {
                updateOp.execute(connection, time);
            }

        }
        for (String table : reverseTableOrder) {
            // insert delete detail record

            InsertDeleteDetailOperation detailOp = tableToInsertDeleteDetailOperation.get(table);
            if (detailOp != null) {
                detailOp.execute(connection, time);
            }

            // update any children delete detail records
            UpdateDeleteDetailOperation updateDetailOp = tableToUpdateDeleteDetailOperation.get(table);
            if (updateDetailOp != null) {
                updateDetailOp.execute(connection, time);
            }
            // and do the delete
            Set<AbstractTemporaryTableOperation> deleteOps = tableToDeleteOperation.get(table);

            if (deleteOps != null && !deleteOps.isEmpty()) {
                for (DbOperation deleteOp : deleteOps) {
                    deleteOp.execute(connection, time);
                }
            }
        }

        for (RunnerOperation runnerOperation : tableToDbRunnerOp.values()) {
            runnerOperation.execute(connection, time);
        }

        handleCacheOnSubmit();

    }

    private void handleCacheOnSubmit() {
        Map<String, List<UuidKey>> tagToUuid = stateCache.clear();

        // tagToUuid is a map of all uuid's grouped by tagName that need to be cleared from map. It currently contains keys that need to be hard-deleted as well as soft-deleted.
        // Remove all uuidKeys from the map that we need to hard-delete, so tagToUuid only contains keys that need to be soft-deleted only.
        for (Map.Entry<String, Set<UuidKey>> entry : tagToUuidKeysToPurge.entrySet()) {
            String tagName = entry.getKey();
            for (UuidKey uuidKeyToPurge : entry.getValue()) {
                tagToUuid.get(tagName).remove(uuidKeyToPurge);
            }
        }
        // Now soft-delete tagToUuids
        keyCache.softDelete(originId, tagToUuid);
        // Hard delete tagToUuidKeysToPurge
        keyCache.purgeRegularObjects(originId, tagToUuidKeysToPurge);
    }

    public Long getObjectIdByTagname(final String tagname, final String key) {
        if (key == null) {
            return null;
        }
        return getObjectIdByTagname(tagname, platformRegistry.buildUuidKey(tagname, key));
    }

    public Long getObjectIdByTagname(final String tagname, final UuidKey uuidKey) {
        final ObjState objState = stateCache.getObjectState(tagname, uuidKey);
        return objState != null ? objState.getObjid() : null;
    }

    public Long getRequiredObjectIdByTagname(final String tagname, final String key, String parseTagname) {
        Long objId = getObjectIdByTagname(tagname, key);
        if (objId == null) {
            throw new IllegalArgumentException("Failed to get " + tagname + " UUID '" + key + "' from " + parseTagname + " properties");
        }
        return objId;
    }

    public ColumnData getColumnData(String tablename) {
        return tableToColumnData.get(tablename);
    }

    public void cacheColumnData(String tablename, ColumnData columnData) {
        tableToColumnData.put(tablename, columnData);
    }

    private void addRunnerOperations(HandlerResult handlerResult, String tagname, String tablename) {
        DbRunner dbRunner = tagToDbRunner.get(tagname);
        if (dbRunner != null) {
            RunnerOperation runnerOperation = tableToDbRunnerOp.get(tablename);
            if (runnerOperation == null) {
                String tempDatabase = deleteDetailTableName.substring(0, deleteDetailTableName.indexOf('.'));
                runnerOperation = dbRunner.getUpdateExec(tempDatabase, Collections.<Long>emptySet());
                tableToDbRunnerOp.put(tablename, runnerOperation);
            }
            runnerOperation.addId(handlerResult.getObjId());
        }
    }

    private int getOperationsCount(Map<String, AbstractBatchOperation> typeToOperation) {
        int result = 0;
        for (AbstractBatchOperation operation : typeToOperation.values()) {
            result += operation.getOperationsCount();
        }
        return result;
    }


    private int getOperationsCountForSetDbOperation(Map<String, Set<AbstractTemporaryTableOperation>> typeToOperations) {
        int result = 0;
        for (Set<AbstractTemporaryTableOperation> operationSet : typeToOperations.values()) {
            for (AbstractTemporaryTableOperation dbOperation : operationSet) {
                result += dbOperation.getOperationsCount();
            }
        }
        return result;
    }

    public OriginatorChanges getOriginatorChanges(Connection connection, long updateTime) throws OriginatorUpdateException {
        return new OriginatorChangesImpl(originId, stateCache.getChanges(connection), updateTime, attributeChangeDetails);
    }

    public void preserveOfflineObjects(Connection connection) throws OriginatorUpdateException {
        try {
            for (Map.Entry<String, Map<String, List<ObjectQueryRunner>>> entry : tagNameToParentTagNameToOfflineRunnerOperation.entrySet()) {
                String tagName = entry.getKey();
                Map<String, List<ObjectQueryRunner>> parentTableToObjectQueryRunner = entry.getValue();
                Set<Long> offlineUUIDs = new HashSet<Long>();

                for (List<ObjectQueryRunner> objectQueryRunners : parentTableToObjectQueryRunner.values()) {
                    for (ObjectQueryRunner objectQueryRunner: objectQueryRunners) {
                        Map<Long, String> idToParentUuids = objectQueryRunner.findObjectIds(connection, originId);
                        for (Map.Entry<Long, String> idToParentUuidEntry : idToParentUuids.entrySet()) {
                            if (!isChildParentDeleted(objectQueryRunner.getParentTag(), idToParentUuidEntry.getValue())) {
                                offlineUUIDs.add(idToParentUuidEntry.getKey());
                            }
                        }
                    }
                }

                Collection<ObjState> objStatesForTag = stateCache.getKeyToObjStateMap(tagName).values();
                if (offlineUUIDs.size() > 0) {
                    for (ObjState objState : objStatesForTag) {
                        if (offlineUUIDs.contains(objState.getObjid())) {
                            objState.setState(ObjState.State.PRESERVED);
                        }
                    }
                }

            }
        } catch (TableException e) {
            throw new OriginatorUpdateException("Failed to preserve offline objects ", e);
        }
    }

    private boolean isChildParentDeleted(String parentTag, String parentUuid) {
        if (parentTag == null) {
            return false;
        } else {
            UuidKey parentUuidKey = platformRegistry.buildUuidKey(parentTag, parentUuid);
            return stateCache.getObjectState(parentTag, parentUuidKey).getState() == ObjState.State.DELETED;
        }
    }

    public void unpreserveOnlineObjects(Connection connection) throws OriginatorUpdateException {
        processDeletedOfflineObjects();
        try {
            for (Map.Entry<String, Map<String, List<ObjectQueryRunner>>> entry : tagNameToParentTagNameToOnlineRunnerOperation.entrySet()) {
                String tagName = entry.getKey();
                Map<String, List<ObjectQueryRunner>> parentTableToObjectQueryRunner = entry.getValue();
                Set<Long> onlineUUIDs = new HashSet<Long>();

                for (List<ObjectQueryRunner> objectQueryRunners : parentTableToObjectQueryRunner.values()) {
                    for (ObjectQueryRunner objectQueryRunner : objectQueryRunners) {
                        onlineUUIDs.addAll(objectQueryRunner.findObjectIds(connection, originId).keySet());
                    }
                }

                Collection<ObjState> objStatesForTag = stateCache.getKeyToObjStateMap(tagName).values();
                if (onlineUUIDs.size() > 0) {
                    for (ObjState objState : objStatesForTag) {
                        if (onlineUUIDs.contains(objState.getObjid()) && objState.getState() == ObjState.State.PRESERVED) {
                            objState.setState(ObjState.State.DELETED);
                        }
                    }
                }
            }
        } catch (TableException e) {
            throw new OriginatorUpdateException("Failed to preserve offline objects ", e);
        }
    }

    private void processDeletedOfflineObjects() {
        Map<String, Map<UuidKey, ObjState>> offlineObjDeletes = stateCache.getOfflineObjDeletes();

        for (String tagName : offlineObjDeletes.keySet()) {
            // Queue up delete xml handler operations
            Map<UuidKey, ObjState> deletedOfflineObjStateMap = offlineObjDeletes.get(tagName);
            final Map<String, List<XmlChildTagDetails>> tagToDependentChildrenForOfflineTags = platformRegistry.getTagToDependentChildrenForOfflineTags();

            for (ObjState objState : deletedOfflineObjStateMap.values()) {
                addOnlineRunners(objState.getObjid(), tagName, tagToDependentChildrenForOfflineTags.get(tagName));
            }

        }
    }

    public void deletePreservedChildren(Connection connection) throws OriginatorUpdateException, TableException {
        MapOfSets<String, XmlChildTagDetails> parentTagToXmlChildTagDetails = new MapOfSets<String, XmlChildTagDetails>();

        final Map<String, List<XmlChildTagDetails>> tagToDependentChildrenForOfflineTags = platformRegistry.getTagToDependentChildrenForOfflineTags();
        for (List<XmlChildTagDetails> xmlChildTagDetails : tagToDependentChildrenForOfflineTags.values()) {
            for (XmlChildTagDetails xmlChildTagDetail : xmlChildTagDetails) {
                String parentTag = xmlChildTagDetail.getParentTag();
                if (parentTag != null) {
                    parentTagToXmlChildTagDetails.add(parentTag, xmlChildTagDetail);
                }
            }
        }


        ObjectQueryRunner queryRunner = new ObjectQueryRunner();
        for (Map.Entry<String, Set<XmlChildTagDetails>> parent : parentTagToXmlChildTagDetails.entrySet()) {
            Map<UuidKey, ObjState> keyToObjState = stateCache.getObjectsWithTagAndState(parent.getKey(), EnumSet.of(ObjState.State.DELETED));
            Set<Long> parentIdsToBeDeleted = new HashSet<Long>(keyToObjState.size());
            for (ObjState parentObjState : keyToObjState.values()) {
                parentIdsToBeDeleted.add(parentObjState.getObjid());
            }

            for (XmlChildTagDetails childTagDetails : parent.getValue()) {
                Set<Long> childrenObjectsToMarkDelete = queryRunner.findObjectIds(connection, originId, childTagDetails.getChildTable(),
                        childTagDetails.getParentFK(), parentIdsToBeDeleted);

                if (childrenObjectsToMarkDelete.size() > 0) {
                    Collection<ObjState> objStatesForChildTag = stateCache.getKeyToObjStateMap(childTagDetails.getChildTagName()).values();
                    for (ObjState objState : objStatesForChildTag) {
                        if (childrenObjectsToMarkDelete.contains(objState.getObjid()) && objState.getState() == ObjState.State.PRESERVED) {
                            objState.setState(ObjState.State.DELETED);
                        }
                    }
                }

            }

        }
    }

    public List<String> getChangeDetailAttributes(String tagName) {
        List<String> changeDetailAttributes = this.tagToChangeDetailAttributes.get(tagName);
        return (changeDetailAttributes != null) ? changeDetailAttributes : Collections.<String>emptyList();
    }

    public KeyCache getIdCache() {
        return keyCache;
    }

    /**
     * Test if any root node is being deleted by this originator
     *
     * @return true if any root node is being deleted
     */
    private boolean isRootDelete() {
        boolean isRootDelete = false;
        for (String tagname : rootTags) {
            isRootDelete = stateCache.isObjStateDelete(tagname);
            if (isRootDelete) {
                break;
            }
        }
        return isRootDelete;
    }

    private void loadColumnsForDeleteDetail(Connection connection, Set<String> tagnames, Set<String> deleteDetailColumns) throws OriginatorUpdateException {
        if (tagToDeleteDetailColumnNamesCache.isEmpty() && !deleteDetailColumns.isEmpty()) {
            synchronized (tagToDeleteDetailColumnNamesCache) {
                if (tagToDeleteDetailColumnNamesCache.isEmpty()) {

                    // load the cache
                    for (String tagname : tagnames) {
                        String tablename = tagToXmlHandlers.get(tagname).getTableName();
                        int dot = tablename.indexOf('.');

                        // Anytime we go against the information_schema table we must be careful about our database prefix. The normal
                        // prefix substitution methods will not translate the database name in a sql argument
                        String tableSchema = PrefixConnection.translateDbPrefix(tablename.substring(0, dot));

                        String sql
                                = " SELECT" +
                                "   COLUMN_NAME" +
                                " FROM" +
                                "   information_schema.COLUMNS" +
                                " WHERE" +
                                "   TABLE_NAME='" + tablename.substring(dot + 1) + '\'' +
                                "   AND TABLE_SCHEMA='" + tableSchema + "'";
                        try {
                            List<String> columns = PreparedStatementWrapper.create(connection, sql, TableResultParserDefault.STRING).executeQueryEx();

                            // We have all the columns for the current table. Now filter out the ones that don't actually appear in the delete detail table
                            Set<String> columnsToCopy = new HashSet<String>();
                            tagToDeleteDetailColumnNamesCache.put(tagname, columnsToCopy);
                            for (String column : columns) {
                                column = column.toLowerCase();
                                if (deleteDetailColumns.contains(column)) {
                                    columnsToCopy.add(column);
                                }
                            }
                        } catch (TableException e) {
                            throw new OriginatorUpdateException("Failed to load table description for " + tablename, e);
                        }
                    }
                }
            }
        }
    }

    public boolean getPreserveOfflineObjectsFlag() {
        return preserveOfflineObjectsFlag;
    }

    public UuidKey buildKey(String tag, String key) {
        return platformRegistry.buildUuidKey(tag, key);
    }

    /**
     * Make sure the key-cache and state-cache are using the same key and not copies of the same key
     *
     * @param tag     tag for key
     * @param uuidKey the key to share
     * @throws KeyCacheException
     */
    public void replaceKey(String tag, UuidKey uuidKey) throws KeyCacheException {
        keyCache.replaceKey(this.originId, tag, uuidKey);
        stateCache.replaceKey(tag, uuidKey);
    }
    
    public void processDoNotDeleteList(Connection connection) throws OriginatorUpdateException {
        for (DoNotDelete doNotDelete : doNotDeleteList) {
            // Read Object Tag Name and the key
            String tagName = doNotDelete.getObjectTagName();
            String primaryKey = doNotDelete.getObjectKey();

            if (primaryKey == null) {
                // Use alternate key to retrieve primary key
                Map<String, String> alternateKey = doNotDelete.getAlternateKey();
                XmlHandler databaseHandler = tagToXmlHandlers.get(tagName);
                // Retrieve primary key. Alternate keys can resolve to multiple primary keys. 
                primaryKey = databaseHandler.alternateKeyToPrimaryKey(connection, this, alternateKey);
            }

            if (primaryKey != null) {
                UuidKey uuidKey = buildKey(tagName, primaryKey);
                // Retrieve the Object Status
                Map<UuidKey, ObjState> keyToObjState = stateCache.getKeyToObjStateMap(tagName);
                ObjState state = keyToObjState.get(uuidKey);
                // Proceed only if the object is marked deleted or delete offline.
                if (state != null
                        && (state.getState() == ObjState.State.DELETED || state.getState() == ObjState.State.DELETED_OFFLINE_OBJ)) {
                    // Preserve the child objects.
                    if (state.getState() == ObjState.State.DELETED) {
                        Map<String, List<XmlChildTagDetails>> tagToDependentChildrenForOfflineTags = platformRegistry
                                .getTagToDependentChildrenForOfflineTags();
                        if (tagToDependentChildrenForOfflineTags.get(tagName) != null) {
                            addOfflineRunners(state.getObjid(), tagName, tagToDependentChildrenForOfflineTags.get(tagName));
                        }
                    }
                    
                    state.setState(ObjState.State.NO_CHANGE_OFFLINE_OBJ);
                }
            }
        }
    }
    
}
