/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator;

import java.io.Reader;
import java.io.StringReader;
import java.sql.Connection;
import java.util.Collections;
import java.util.Set;

import javax.naming.NamingException;

import com.onaro.commons.table.DbConfig;
import com.onaro.commons.types.Period;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import com.onaro.sanscreen.server.sessions.ImplicitCommitCheckRemote;
import com.onaro.sanscreen.server.sessions.SessionFactory;
import com.onaro.sanscreen.util.Schema;
import com.onaro.sanscreen.util.xml.SaxParserWrapper;
import com.onaro.util.sessions.SessionCacheException;
import org.xml.sax.InputSource;

import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.plugin.data.PlatformPlugin;
import com.netapp.oci.server.TransactionContext;
import com.netapp.oci.server.TransactionContextException;
import com.netapp.oci.server.UpdateTask;
import com.netapp.oci.server.UpdateTaskException;

/**
 * @author Al Velella
 *         Date: Sep 21, 2011
 *         Time: 4:06:12 PM
 */
public class OriginatorUpdateTask extends UpdateTask {
    private static final long serialVersionUID = 1L;
    public static final String PRESERVE_OFFLINE_OBJECTS_KEY = "PRESERVE_OFFLINE_OBJECTS";
    public static final String SOFT_DELETE_PURGE_LIFETIME_PARAM_KEY = "SOFT_DELETE_PURGE_LIFETIME_DAYS";

    private OriginatorUpdate originatorUpdate;
    private PlatformPlugin platformPlugin;
    private long starttime;
    private Long datasourceId;
    private int changeCount;
    private boolean checkImplicitCommit;
    private OriginatorChanges originatorChanges;
    private boolean hasObjectsToPurge;

    /**
     * true if it Originator was removed (The datasource was removed or empty XML was sent to the server)
     */
    protected boolean originatorRemoved;

    public OriginatorUpdateTask(String originatorName, PlatformPlugin platformPlugin, Long originId, Long datasourceId, long starttime, boolean checkImplicitCommit) throws UpdateTaskException {
        super(originatorName, originId);
        this.originatorUpdate = null;
        this.platformPlugin = platformPlugin;
        this.starttime = starttime;
        this.datasourceId = datasourceId;
        this.checkImplicitCommit = checkImplicitCommit;
    }

    @Override
    public boolean run(TransactionContext transactionContext) throws UpdateTaskException {
        try {
            final Connection connection = transactionContext.getConnection();
            long time = transactionContext.getTime();
            initImplicitCommitCheck(transactionContext);

            if (changeCount > 0) {

                if (originatorRemoved) {
                    logger.info(taskClassName + ": Remove originator " + loggerPrefix);
                }

                // gather the detailed change info here as this needs to be run when no one updates the database.
                originatorUpdate.gatherDetailedAttributeChangeInfo(connection);

                // calculate originator changes after gathering the detailed change info.
                originatorChanges = originatorUpdate.getOriginatorChanges(connection, time);

                originatorUpdate.submit(connection, time);

                platformPlugin.handleOriginatorChanges(connection, datasourceId, originatorChanges);
            }

            if(hasObjectsToPurge){
                purge(connection, time);
                if(changeCount == 0){
                    originatorChanges = new OriginatorChangesImpl(originatorId,
                            new OriginatorStateCache.ChangeData(Collections.<String, Set<Long>>emptyMap(),
                                    Collections.<String, Set<Long>>emptyMap(),
                                    Collections.<String, Set<Long>>emptyMap(),
                                    Collections.<String, Set<Long>>emptyMap()),
                                    time,
                                    Collections.<AttributeChangeDetail>emptySet());
                }
            }
            platformPlugin.postOriginatorUpdate(connection, originatorId, datasourceId, time);

            platformPlugin.logUpdateHistory(connection, originatorId, starttime, time, System.currentTimeMillis() - starttime,
                    originatorUpdate.getChangeCount(), originatorUpdate.getRunnerCount());

            assertImplicitCommitCheck(transactionContext.getTime(), true);

            //clean up cache of chagnes until all the above succeed. The cache of changes will be used to rollback cache when exception is thrown
            if(changeCount > 0) {
                KeyCache.commitOriginator(platformPlugin.getSchema(), originatorId);
            }

            return changeCount > 0 || hasObjectsToPurge;

        } catch (UpdateTaskException e) {
            throw e;
        } catch (Exception e) {
            throw new UpdateTaskException("Failed to run NewOriginatorUpdateTask, origin:" + getOriginatorId() + " originatorName:" + getOriginatorName(), e);
        } finally {
            assertImplicitCommitCheck(transactionContext.getTime(), false);
        }
    }

    private void purge(Connection connection, long time) throws TableException, ServerInterfaceException {
        long softDeletedObjectsLifetime = platformPlugin.getParam(SOFT_DELETE_PURGE_LIFETIME_PARAM_KEY, long.class, Long.valueOf(0)) *  Period.DAY.milliseconds();
        new SoftDeletedObjectsPurger().purge(connection, platformPlugin, softDeletedObjectsLifetime, originatorId, time);
    }

    public boolean update(Connection connection, String xml, boolean validate, Schema schema) throws UpdateTaskException {
        return update(connection, new StringReader(xml), validate, schema);
    }

    /**
     * Reads and processes XML data from a stream
     *
     * @param connection database connection
     * @param reader     reader for the data XML
     * @param validate   flag to validate XDML schema
     * @param schema     XML schema
     * @return returns true if changes have been detected
     * @throws UpdateTaskException a generic exception in case anything goes wrong
     */
    public boolean update(Connection connection, Reader reader, boolean validate, Schema schema) throws UpdateTaskException {
        changeCount = 0;
        try {
            boolean preserveOfflineObjectsFlag = platformPlugin.getParam(PRESERVE_OFFLINE_OBJECTS_KEY, boolean.class, false);

            originatorUpdate = new OriginatorUpdate(connection, originatorId, starttime, platformPlugin, preserveOfflineObjectsFlag);
            KeyCache.reportOriginator(platformPlugin.getSchema(), originatorId);

            OriginatorUpdateHandler databaseCacheHandler = new OriginatorUpdateHandler(originatorUpdate, connection);
            SaxParserWrapper saxParser = new SaxParserWrapper(databaseCacheHandler);
            if (validate) {
                saxParser.setSchemaFileUri(schema.getFilename());
            }
            saxParser.run(new InputSource(reader));
            changeCount = databaseCacheHandler.getChangeCount();
            originatorRemoved = false;
            return databaseCacheHandler.isChanged();
        } catch (Exception e) {
            logger.error("Error when process from originId:" + originatorId + " originName:" + originatorName, e);
            throw new UpdateTaskException("Error when process from originId:" + originatorId + " originName:" + originatorName, e);
        }
    }

    public boolean remove(Connection connection) throws UpdateTaskException {
        try {
            KeyCache.reportOriginator(platformPlugin.getSchema(), originatorId);
            boolean preserveOfflineObjectsFlag = false;
            originatorUpdate = new OriginatorUpdate(connection, originatorId, starttime, platformPlugin, preserveOfflineObjectsFlag);
            originatorRemoved = true;
            originatorUpdate.queueDelete();
            changeCount = originatorUpdate.getChangeCount();
            return originatorUpdate.isChanged();
        } catch (OriginatorUpdateException e) {
            logger.error("Error when process from originId:" + originatorId + " originName:" + originatorName, e);
            throw new UpdateTaskException("Error when process from originId:" + originatorId + " originName:" + originatorName, e);
        } finally {
            OriginatorStateCacheFactory.removeOriginatorCache(originatorId);
            KeyCache.removeOriginatorCache(originatorId);
        }
    }

    public int getChangeCount() {
        return changeCount;
    }

    private void initImplicitCommitCheck(TransactionContext transactionContext) throws TableException, TransactionContextException {
        PreparedStatementWrapper.createForUpdate(transactionContext.getConnection(), "UPDATE " + DbConfig.getSanscreenDbName() + ".implicit_commit SET starttime=" + transactionContext.getTime()).executeUpdateEx();
    }

    private void assertImplicitCommitCheck(long time, boolean throwException) throws UpdateTaskException {
        if (checkImplicitCommit) {
            try {
                ImplicitCommitCheckRemote implicitCommitCheckRemote = SessionFactory.createSession(ImplicitCommitCheckRemote.JNDI_NAME);
                implicitCommitCheckRemote.assertImplicitCommitCheck(time, throwException);
            } catch (NamingException e) {
                logger.error("Error checking implicit commit, originId:" + getOriginatorId() + " originatorName:" + getOriginatorName(), e);
                if (throwException) {
                    throw new UpdateTaskException("Error checking implicit commit, originId:" + getOriginatorId() + " originatorName:" + getOriginatorName(), e);
                }
            }
        }
    }

    public OriginatorChanges getOriginatorChanges() {
        return originatorChanges;
    }

    public void rollback() {
        OriginatorStateCacheFactory.removeOriginatorCache(originatorId);
        KeyCache.rollbackOriginator(platformPlugin.getSchema(), originatorId);
    }

    public boolean isPurgingRequired(Connection connection, Schema schema, Long originId) throws TableException {
        boolean hasDeadObj = KeyCache.containsDeadObjects(originId);
        if(hasDeadObj){
            long softDeletedObjectsLifetime = platformPlugin.getParam(SOFT_DELETE_PURGE_LIFETIME_PARAM_KEY, long.class, Long.valueOf(0)) * Period.DAY.milliseconds();
            long cutOffTime = System.currentTimeMillis() - softDeletedObjectsLifetime;

            for (String tableName: platformPlugin.getTableOrder()){
                final String sql = String.format("SELECT count(*) FROM %s WHERE originId = %d AND updateTime < %d AND objState = 'DEAD' ", tableName, originId, cutOffTime);
                long count = PreparedStatementWrapper.executeFunctionQuery(connection, sql, 0);
                if(count > 0){
                    hasObjectsToPurge = true;
                    return true;
                }
            }

        }
        return false;

    }

    public void setHasObjectsToPurge(boolean purgingFlag) {
        this.hasObjectsToPurge = purgingFlag;
    }
}
