/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator;

import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.originator.handler.XmlChildTagDetails;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.key.UuidKey;
import com.netapp.oci.platform.originator.key.UuidKeyFactory;
import com.netapp.oci.platform.originator.runner.DbRunner;

import java.sql.Connection;
import java.util.*;

/**
 * @author Nancy Walia
 *         Date: Sep 10, 2012
 */
public abstract class PlatformRegistry {

    private final Map<String, XmlHandler> tagToXmlHandler;

    private final Map<String, DbRunner> tagToDbRunner;

    private final Map<String, String> tagToResourceType;

    private final Map<String, UuidKeyFactory> tagToUuidKeyFactory;

    private final List<String> tableOrder;

    private final Map<String, String> tagToHistoryTables;

    private final Map<String, String> tagToParentTag;

    private Map<String, Map<String, List<String>>> tagToColumnsToUnlinkOnDelete;

    private Map<String, List<XmlChildTagDetails>> tagToDependentChildrenForOfflineTags;

    private Map<String, String> subTableToBaseTable;

    protected PlatformRegistry() {
        this.tagToXmlHandler = new HashMap<String, XmlHandler>();
        this.tableOrder = new ArrayList<String>();
        this.tagToHistoryTables = new HashMap<String, String>();
        this.tagToDbRunner = new HashMap<String, DbRunner>();
        this.tagToResourceType = new HashMap<String, String>();
        this.tagToUuidKeyFactory = new HashMap<String, UuidKeyFactory>();
        this.tagToParentTag = new HashMap<String, String>();
        this.subTableToBaseTable = new HashMap<String, String>();
    }

    protected void addXmlHandler(final XmlHandler handler, String resourceType, UuidKeyFactory uuidKeyFactory) {
        tagToResourceType.put(handler.getTagName(), resourceType);
        tagToXmlHandler.put(handler.getTagName(), handler);
        tableOrder.add(handler.getTableName());
        if (handler.keepHistory()) {
            tagToHistoryTables.put(handler.getTagName(), handler.getTableName());
        }
        tagToUuidKeyFactory.put(handler.getTagName(), uuidKeyFactory);
        tagToParentTag.put(handler.getTagName(), null);
    }

    protected void addHierarchyXmlHandler(final XmlHandler handler, String resourceType, UuidKeyFactory uuidKeyFactory) {
        addXmlHandler(handler, resourceType, uuidKeyFactory);
        this.subTableToBaseTable.put(handler.getTableName(), handler.getBaseTableName());
    }

    /**
     * Register handler as child of a parent handler. The child and parent must share the exact same key (UUID).
     * This relationship triggers the following behavior for the parent and child objects:
     * <p/>
     * 1) They will share object IDs (for easy joins in the database)
     * 2) They will share key types (for reliable lookups in the parent and child state caches)
     * 3) They will share the same UuidKey key instance (for memory savings in state caches)
     * 4) They will share change records (using the parent's resource information)
     *
     * @param handler   child handler
     * @param parentTag parent tag
     */
    protected void addSelfRelationshipXmlHandler(final XmlHandler handler, String parentTag) {
        String resourceType = null; // no resource-type for relationship children
        UuidKeyFactory keyFactory = tagToUuidKeyFactory.get(parentTag); // use parent's key factory
        addXmlHandler(handler, resourceType, keyFactory);
        tagToParentTag.put(handler.getTagName(), parentTag);
    }

    public Map<String, XmlHandler> getTagToXmlHandlers() {
        return tagToXmlHandler;
    }

    public void addDbRunner(String tag, String updateSql, String deleteSql) {
        addDbRunner(tag,
                updateSql == null ? Collections.<String>emptyList() : Arrays.asList(updateSql),
                deleteSql == null ? Collections.<String>emptyList() : Arrays.asList(deleteSql));
    }

    public void addDbRunner(String tag, List<String> updateSqls, List<String> deleteSqls) {
        tagToDbRunner.put(tag, new DbRunner(tag, getClass(), updateSqls, deleteSqls));
    }

    public Map<String, DbRunner> getTagToDbRunners() {
        return tagToDbRunner;
    }

    public void deleteTables(Connection connection) throws TableException {
        List<String> tables = new ArrayList<String>(tableOrder);
        Collections.reverse(tables);
        for (String table : tables) {
            if(subTableToBaseTable.containsKey(table)){
                final String baseTable = subTableToBaseTable.get(table);
                PreparedStatementWrapper.createForUpdate(connection, "DELETE FROM " + baseTable).executeUpdateEx();

            }
            PreparedStatementWrapper.createForUpdate(connection, "DELETE FROM " + table).executeUpdateEx();
        }
    }

    public List<String> getTableOrder() {
        return tableOrder;
    }

    public Map<String, String> getHistoryTables() {
        if (tagToHistoryTables.size() > 0) {
            return Collections.unmodifiableMap(tagToHistoryTables);
        } else {
            return Collections.emptyMap();
        }
    }

    public String getResourceType(String tag) {
        return tagToResourceType.get(tag);
    }

    public Map<String, String> getTagToResourceType() {
        return tagToResourceType;
    }

    public UuidKey buildUuidKey(String tag, String key) {
        return tagToUuidKeyFactory.get(tag).key(key);
    }

    public String getParentRelationshipTagName(String tagName) {
        return tagToParentTag.get(tagName);
    }

    /**
     * @return Map of tag names -> map of table names of tags and corresponding FK to unlink
     */
    public Map<String, Map<String, List<String>>> getTagToColumnsToUnlinkOnDelete() {
        return tagToColumnsToUnlinkOnDelete;
    }

    protected void setTagToColumnsToUnlinkOnDelete(Map<String, Map<String, List<String>>> tagToColumnsToUnlinkOnDelete) {
        this.tagToColumnsToUnlinkOnDelete = tagToColumnsToUnlinkOnDelete;
    }

    /**
     * @return Map of tag names that can be offlined -> list of dependent children
     */
    public Map<String, List<XmlChildTagDetails>> getTagToDependentChildrenForOfflineTags() {
        return tagToDependentChildrenForOfflineTags;
    }


    protected void setTagToDependentChildrenForOfflineTags(Map<String, List<XmlChildTagDetails>> tagToDependentChildrenForOfflineTags) {
        this.tagToDependentChildrenForOfflineTags = tagToDependentChildrenForOfflineTags;
    }

    public Map<String, String> getSubTableToBaseTable() {
        return subTableToBaseTable;
    }

    public String getSubTable(String table) {
        return subTableToBaseTable.get(table);
    }
}