/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator;

import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.netapp.oci.platform.originator.db.DbOperation;
import com.netapp.oci.platform.originator.db.DeleteOperation;
import com.netapp.oci.platform.originator.db.RunnerOperation;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.runner.DbRunner;
import com.netapp.oci.platform.originator.runner.UnlinkObjectRunner;
import com.netapp.oci.platform.plugin.data.PlatformPlugin;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import org.slf4j.Logger;

import java.sql.Connection;
import java.util.*;

/**
 * @author Iti Behari
 *         Date: 6/4/13
 */
public class SoftDeletedObjectsPurger {
private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(SoftDeletedObjectsPurger.class);

    // JDBC-connector-agnostic foreign key constraint exception message
    private static String FK_CONSTRAINT_MSG = "java.sql.SQLIntegrityConstraintViolationException: Cannot delete or update a parent row: a foreign key constraint fails";

    public Map<String, Set<Long>> purge(Connection connection, PlatformPlugin plugin, long softDeletedObjsLifetime, long originId, long updateTime) throws TableException, ServerInterfaceException {
        logger.info("Begin purging for originator " + originId);
        long cutoffTime = updateTime - softDeletedObjsLifetime;
        Long rootId = Long.valueOf(plugin.getRootId(connection, originId));
        List<String> tables = getTablesInDeletionOrder(plugin);
        Map<String, String> tableToTagLookupMap = buildTableToTagLookupMap(plugin);

        Map<String, Set<Long>> tagToIdsBeingDeleted = new HashMap<String, Set<Long>>();
        int totalObjsBeingDeleted = 0;
        for (String tableName : tables) {
            Set<Long> objIdsToBeDeleted = queryObjIdsToBeDeletedForTable(connection, originId, tableName, cutoffTime);
            if (!objIdsToBeDeleted.isEmpty()) {
                totalObjsBeingDeleted += objIdsToBeDeleted.size();
                String tagName = tableToTagLookupMap.get(tableName);
                // Note: tableName and subTableName have the format "schema.table"
                deleteDataFromTable(connection, plugin, originId, tagName, tableName, objIdsToBeDeleted, updateTime, tagToIdsBeingDeleted, tableToTagLookupMap);
                recordIdsToBeDeleted(tagToIdsBeingDeleted, tagName, objIdsToBeDeleted);
                String subTableName = plugin.getRegistry().getSubTable(tableName);
                if (subTableName != null) {
                    String subTagName = tableToTagLookupMap.get(subTableName);
                    // For subtables: pass null for tag map, and don't re-record ids to be deleted
                    deleteDataFromTable(connection, plugin, originId, subTagName, subTableName, objIdsToBeDeleted, updateTime, null, tableToTagLookupMap);
                }
            }
        }

        if (!tagToIdsBeingDeleted.isEmpty()) {
            insertChangeRecord(connection, plugin, rootId, tagToIdsBeingDeleted, updateTime);
            clearCache(plugin, originId, tagToIdsBeingDeleted);
        }
        logger.info("Finished purging for originator  " + originId + ", total objects purged:" + totalObjsBeingDeleted);
        logger.debug("Tags that were affected by Purging : " + tagToIdsBeingDeleted.keySet());
        return tagToIdsBeingDeleted;
    }

    private void insertChangeRecord(Connection connection, PlatformPlugin plugin, long rootId, Map<String, Set<Long>> tagToIdsBeingDeleted, long updateTime) throws ServerInterfaceException, TableException {
        plugin.logObjectsPurgeJob(connection, rootId, tagToIdsBeingDeleted, updateTime);
    }


    private List<String> getTablesInDeletionOrder(PlatformPlugin plugin) {
        List<String> reversedTableOrder = new ArrayList<String>(plugin.getTableOrder());
        Collections.reverse(reversedTableOrder);
        return reversedTableOrder;
    }


    private void clearCache(PlatformPlugin plugin, Long originId, Map<String, Set<Long>> tagsToIdsToBeDeleted) {
        KeyCache.get().purgeSoftDeletedObjects(originId, tagsToIdsToBeDeleted);
    }

    private Map<String, String> buildTableToTagLookupMap(PlatformPlugin plugin) {
        Map<String, String> tableToTagName = new HashMap<String, String>();
        Map<String, XmlHandler> tagToXmlHandler = plugin.getTagToXmlHandlers();
        for (Map.Entry<String, XmlHandler> entry : tagToXmlHandler.entrySet()) {
            tableToTagName.put(entry.getValue().getTableName(), entry.getKey());
        }
        return tableToTagName;
    }

    /**
      * Record all the objId's to be deleted from the appropriate database table as defined by the tag handler.
      *
      * @param  tagToIdsBeingDeleted    map used to record tag/objIds to be deleted
      * @param  tagName                 tag handler name used as map key
      * @param  objIdsToBeDeleted       set of objIds to delete to be added/inserted into map via tag key
      * @return                         this method has no return value
      */
    private void recordIdsToBeDeleted(Map<String, Set<Long>> tagToIdsBeingDeleted, String tagName, Set<Long> objIdsToBeDeleted) {
        if (tagToIdsBeingDeleted.get(tagName) == null) {
            tagToIdsBeingDeleted.put(tagName, objIdsToBeDeleted);
        } else {
            Set<Long> ids = tagToIdsBeingDeleted.get(tagName);
            ids.addAll(objIdsToBeDeleted);
            tagToIdsBeingDeleted.put(tagName, ids);
        }
    }

    /**
      * Queue up and run any cleanup and object unlink runners associated with the tag/objIds being deleted.
      *
      * @param  connection              database connection
      * @param  plugin              PlatformPlugin instance with access to runners and registry
      * @param  tagName             tag handler name associated with objIds to be deleted
      * @param  schemaDotTable      schema and table in which objIds are to be deleted, in "schema.table" format
      * @param  objIdsToBeDeleted   Set of objIds to be deleted
      * @return this method has no return value
      */
    private void queueUpRunnerCleanupAndUnlinkObjectOperations(Connection connection, PlatformPlugin plugin, String tagName, String schemaDotTable, Set<Long> objIdsToBeDeleted, long updateTime) throws TableException {
        String tokens[] = schemaDotTable.split("\\.");
        String schema = tokens[0];

        // Queue up any runner cleanup operations. These will execute before the standard insert/update/delete xml handler
        // operations run.
        Map<String, DbRunner> tagToDbRunner = plugin.getTagToDbRunners();
        DbRunner dbRunner = tagToDbRunner.get(tagName);
        if (dbRunner != null) {
            RunnerOperation runnerOperation = dbRunner.getDeleteExec(schema, objIdsToBeDeleted);
            runnerOperation.execute(connection, updateTime);
        }
        Map<String, Map<String, List<String>>> tagToColumnsToUnlinkOnDelete = plugin.getRegistry().getTagToColumnsToUnlinkOnDelete();
        // Queue up unlink object operations
        Map<String, List<String>> objectsToUnlink = tagToColumnsToUnlinkOnDelete.containsKey(tagName) ? tagToColumnsToUnlinkOnDelete.get(tagName) : Collections.<String, List<String>>emptyMap();
        if (!objectsToUnlink.isEmpty()) {
            new UnlinkObjectRunner().getRunnerOp("management", tagName, objIdsToBeDeleted, objectsToUnlink).execute(connection, updateTime);
        }
    }

    /**
      * Convert a set of Longs to a String of Long values broken up by delimiter.
      *
      * @param  objIds      Set<Long> of objIds to convert to a String
      * @param  delimiter   Delimiter to use when building up String from Set<Long>
      * @return             String of delimited Set<Long> values
      */
    private String objIdsToString(Set<Long> objIds, String delimiter) {
        HashSet<String> strobjIdsToBeDeleted = new HashSet<>(objIds.size());
        objIds.forEach(i -> strobjIdsToBeDeleted.add(Long.toString(i)));
        return String.join(delimiter, strobjIdsToBeDeleted);
    }

    /**
      * Invoke runners then delete objIds, and set up for handling of (rare) foreign key constraints errors.
      *
      * @param  connection              database connection
      * @param  plugin                  PlatformPlugin instance used for cleanup and unlink operations
      * @param  originId                ensures we only delete dead objects from this originator id
      * @param  currentTagName          current tag handler name associated with objIds to be deleted
      * @param  schemaDotTable          schema and table in which objIds are to be deleted, in "schema.table" format
      * @param  objIdsToBeDeleted       Set of objIds to be deleted
      * @param  updateTime              not used, but passed through to runners
      * @param  tagToIdsBeingDeleted    map used to record tag/objIds to be deleted
      * @param  tableToTagLookupMap     lookup map of table to tag names
      * @return                         this method has no return value
      */
    private void deleteDataFromTable(Connection connection, PlatformPlugin plugin, Long originId, String currentTagName, String schemaDotTable, Set<Long> objIdsToBeDeleted, long updateTime, Map<String, Set<Long>> tagToIdsBeingDeleted, Map<String, String> tableToTagLookupMap) throws TableException {

        queueUpRunnerCleanupAndUnlinkObjectOperations(connection, plugin, currentTagName, schemaDotTable, objIdsToBeDeleted, updateTime);

        try {
            if (objIdsToBeDeleted != null && tagToIdsBeingDeleted != null) {
                recordIdsToBeDeleted(tagToIdsBeingDeleted, currentTagName, objIdsToBeDeleted);
            }
            DbOperation operation = new DeleteOperation(schemaDotTable, objIdsToBeDeleted);
            operation.execute(connection, updateTime);
        } catch (TableException te) {
            String cause = te.getCause().toString();
            // burt1136354: runners have already run, we need to handle foreign key
            // constraint errors here to keep the database and caches coherent
            if (cause.contains(FK_CONSTRAINT_MSG)) {
                logger.info(String.format("Caught FK constraint exception while deleting from %s with originId:%d, updateTime:%d, objIds:%s",
                                        schemaDotTable, originId, updateTime, objIdsToString(objIdsToBeDeleted, ",")));
                String tokens[] = schemaDotTable.split("\\.");
                String schema = tokens[0];
                String table = tokens[1];
                // Note: we have no way of knowing which id(s) caused the issue so we must delete one entry at a time
                for (Long objIdToBeDeleted: objIdsToBeDeleted) {
                    deleteEntryFromTable(connection, plugin, originId, false, currentTagName, schema, table, "objId", Long.toString(objIdToBeDeleted), updateTime, tagToIdsBeingDeleted, tableToTagLookupMap);
                }
            } else {
                logger.error(String.format("Unexpected non-FK constraint exception while deleting from %s with originId:%d, updateTime:%d, objIds:%s",
                                        schemaDotTable, originId, updateTime, objIdsToString(objIdsToBeDeleted, ",")));
                throw te;
            }
        }
    }

    /**
      * Delete objIds associated with foreign keys, recursively handling foreign key constraint errors.
      *
      * @param  connection              database connection
      * @param  plugin                  PlatformPlugin instance used for cleanup and unlink operations
      * @param  originId                ensures we only delete dead objects from this originator id
      * @param  isRecursivelyCalled     ensures we only call foreign table runners when called recursively
      * @param  currentTagName          current tag handler name associated with objIds to be deleted
      * @param  schema                  foreign/reference database name
      * @param  table                   foreign table name
      * @param  column                  foreign column name
      * @param  value                   foreign column value
      * @param  updateTime              not used, but passed through to runners
      * @param  tagToIdsBeingDeleted    map of tag keys and sets of objIds being deleted
      * @param  tableToTagLookupMap     lookup map of table to tag names
      * @return                         this method has no return value
      */
    // delete entry, recursively chasing down FK constraints and delete foreign table entries, retrying as necessary
    private void deleteEntryFromTable(Connection connection, PlatformPlugin plugin, Long originId, boolean isRecursivelyCalled, String currentTagName, String schema, String table, String column, String value, long updateTime, Map<String, Set<Long>> tagToIdsBeingDeleted, Map<String, String> tableToTagLookupMap) throws TableException {
        Set<Long> objIdsToBeDeleted = queryObjIdsToBeDeletedForEntry(connection, originId, schema, table, column, value);
        if (objIdsToBeDeleted != null && tagToIdsBeingDeleted != null) {
            recordIdsToBeDeleted(tagToIdsBeingDeleted, currentTagName, objIdsToBeDeleted);
        }

        if (isRecursivelyCalled) {
            String schemaDotTable = String.join("\\.", schema, table);
            queueUpRunnerCleanupAndUnlinkObjectOperations(connection, plugin, currentTagName, schemaDotTable, objIdsToBeDeleted, updateTime);
        }

        String sql = String.format("DELETE FROM %s.%s WHERE originId = ? AND objState = 'DEAD' AND objId in (%s)", schema, table, objIdsToString(objIdsToBeDeleted, ","));
        boolean done = false;
        while (!done) {
            try {
                PreparedStatementWrapper.createForUpdate(connection, sql).setNext(originId).executeUpdateEx();
                logger.info(String.format("Handling FK constraint. SQL:[%s]", sql));
                done = true;
            } catch (Exception e) {
                // Due to nested exceptions we have to go searching for foreign key constraints
                String cause = e.getCause().toString();
                if (cause.contains(FK_CONSTRAINT_MSG)) {
                    String realCause = cause.substring(cause.indexOf(FK_CONSTRAINT_MSG) + FK_CONSTRAINT_MSG.length());
                    String tokens[] = realCause.split("`");
                    // Example: //java.sql.SQLIntegrityConstraintViolationException: Cannot delete or update a parent row: a foreign key constraint fails (`netapp_model`.`disk_aggregate_relationship`, CONSTRAINT `disk_aggregate_relationship_ibfk_2` FOREIGN KEY (`aggregateId`) REFERENCES `aggregate` (`objid`))
                    String fk_schema = tokens[1];
                    String fk_table = tokens[3];
                    String fk_column = tokens[7];
                    String ref_table = tokens[9];
                    String ref_column = tokens[11];
                    String tagName = tableToTagLookupMap.get(fk_table);
                    deleteEntryFromTable(connection, plugin, originId, true, tagName, fk_schema, fk_table, fk_column, String.format("SELECT %s FROM %s.%s WHERE originId = %d AND objState = 'DEAD' AND %s IN (%s)", ref_column, fk_schema, ref_table, originId, column, value), updateTime, tagToIdsBeingDeleted, tableToTagLookupMap);
                } else {
                    // Bail out if we encounter any other kind of error. We should never get here
                    // TODO: log, delete *all* dead objects and restart/reboot?
                    logger.error(String.format("Unexpected non-FK constraint exception while handling FK constraint exception. SQL:[%s]", sql));
                    throw new TableException("Error encountered while recursively deleting foreign key constraints", e);
                }
            }
        }
    }

    private Set<Long> queryObjIdsToBeDeletedForEntry(Connection connection, Long originId, String schema, String table, String column, String value) throws TableException {
        final String sql = String.format("SELECT objId FROM %s.%s WHERE originId = ? AND objState = 'DEAD' AND %s IN (%s)", schema, table, column, value);
        List<Long> objIdsToBeDeleted = PreparedStatementWrapper.create(connection, sql, TableResultParserDefault.LONG)
                .setNext(originId)
                .executeQueryEx();
        return new HashSet<Long>(objIdsToBeDeleted);
    }

    private Set<Long> queryObjIdsToBeDeletedForTable(Connection connection, Long originId, String tableName, long cutOffTime) throws TableException {
        final String sql = String.format("SELECT objId FROM %s WHERE originId = ? AND updateTime < ? AND objState = 'DEAD' ", tableName);
        List<Long> ids = PreparedStatementWrapper.create(connection, sql, TableResultParserDefault.LONG)
                .setNext(originId)
                .setNext(cutOffTime)
                .executeQueryEx();
        return new HashSet<Long>(ids);
    }
}
