/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator.db;

import com.netapp.oci.platform.common.table.*;
import com.netapp.oci.platform.originator.handler.XmlChildTagDetails;
import org.slf4j.Logger;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

/**
 * @author Iti Behari
 *         Date: 2/14/13
 *         <p/>
 *         <p/>
 *         Finds objects for a tag with given foreign key relationships.
 *         For eg:- It finds lunIds for a specific aggregate
 *         For eg:- given an id for aggregate, the runner can help find its associated volumes, luns or qtrees using foreignKeyRelationship map
 *
 *         SQL is something (but not exactly) like:
 *         SELECT
 *           objId[, vserverUuid]
 *         FROM
 *           lun, [vserver,] tmp
 *         WHERE
 *           lun.volumeId = tmp.objId
 *           [AND lun.vserverid = vserver.objid];
 *
 *         OR:
 *
 *         SELECT
 *           objId[, vserverUuid]
 *         FROM
 *           lun, volume, vserver, tmp
 *         WHERE
 *           lun.volumeId = volume.objId
 *           AND lun.aggregateId = tmp.objId
 *           [AND vserver.objid = volume.vserverid];
 */
public class ObjectQueryRunner {

    private org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger("com.netapp.oci.platform.originator.db");
    private Map<String, String> tableToForeignKeyColumnMap;
    private Set<Long> ids;

    // Attributes for relationshipQuery
    private String parentTable;
    private String parentFK;
    private String parentTag;



    private boolean simpleQuery = true;

    /**
     * @param offlineChild - details about foreign key relationship in order to find objectIds
     *
     */
    public ObjectQueryRunner(XmlChildTagDetails offlineChild) {
        this.tableToForeignKeyColumnMap = offlineChild.getDependentTableToForeignKeyColumn();
        this.ids = new HashSet<Long>();
        this.parentTable = offlineChild.getParentTable();
        this.parentFK = offlineChild.getParentFK();
        this.parentTag = offlineChild.getParentTag();
        this.simpleQuery = false;
   }

    public ObjectQueryRunner(){
        this.simpleQuery = true;
    }



    public Set<Long> findObjectIds(Connection connection, Long origId, String childTableName, String foreignKeyColName, Set<Long> ids ) throws TableException {
        String tmpTableName = "netapp_model.offline_tmp_" + origId;
        String sql = buildSimpleSql(tmpTableName, childTableName, foreignKeyColName);
        return findObjectIds(connection, tmpTableName, sql, ids).keySet();


    }

    public Map<Long, String> findObjectIds(Connection connection, Long origId) throws TableException {

        if(simpleQuery){
            throw new IllegalStateException("All arguments are not provided in order to run the query"); //TODO pull simpleQuerycode out.
        }
        String tmpTableName = "netapp_model.offline_tmp_" + origId;
        String sql = buildSql(tmpTableName);
        return findObjectIds(connection, tmpTableName, sql, ids);
    }

    private Map<Long, String> findObjectIds(Connection connection, String tmpTableName, String sql, Set<Long> ids) throws TableException {
        logger.debug(sql + ";");
        Table.createIdsTemporaryTableForJoin(connection, tmpTableName, ids, logger);

        final Map<Long, String> idToParentUuid = new HashMap<Long, String>();
        PreparedStatementWrapper.create(connection, sql, new TableResultParser<Object>() {
            @Override
            public List<Object> parseResult(ResultSet resultSet) throws SQLException {
                while (resultSet.next()) {
                    idToParentUuid.put(Long.valueOf(resultSet.getLong(1)), parentTable == null ? null : StringCache.get(resultSet.getString(2)));
                }
                return Collections.emptyList();
            }
        }).executeQueryEx();
        return idToParentUuid;
    }

    private String buildSql(String tmpTableName) {

        int counter = tableToForeignKeyColumnMap.size();
        StringBuilder builder = new StringBuilder();
        String lastAlias = "t" + counter--;
        builder.append("SELECT t0.objId ");
        if (parentTable != null) {
            builder.append(", parentTbl.uuid AS parentUuid");
        }
        builder.append(" FROM ").append(tmpTableName).append(" ").append(lastAlias);
        for (Map.Entry<String, String> entry : tableToForeignKeyColumnMap.entrySet()) {
            String newAlias = "t" + counter--;
            String curTable = entry.getKey();
            String foreignKeyColumn = entry.getValue();
            builder.append(String.format(" JOIN %s %s ON %s.objId = %s.%s and %s.objState = 'LIVE'", curTable, newAlias, lastAlias, newAlias, foreignKeyColumn, newAlias));
            lastAlias = newAlias;
        }
        if (parentTable != null) {
            builder.append(String.format(" JOIN %s parentTbl ON parentTbl.objid=t0.%s and parentTbl.objState = 'LIVE'", parentTable, parentFK));
        }
        return builder.toString();
    }

    private String buildSimpleSql(String tmpTableName, String tableName, String foreignKeyColName) {
        return String.format(" SELECT tbl.objId FROM %s tmp JOIN %s tbl ON tmp.objId = tbl.%s", tmpTableName, tableName, foreignKeyColName);
    }

    public void addId(Long id) {
        ids.add(id);
    }


    public String getParentTag() {
        return parentTag;
    }

    public boolean createdFromSameXmlChildTagDetails(final XmlChildTagDetails rhs)
    {
        return cmpObjects(this.tableToForeignKeyColumnMap, rhs.getDependentTableToForeignKeyColumn()) &&
               cmpObjects(this.parentTable, rhs.getParentTable()) &&
               cmpObjects(this.parentFK, rhs.getParentFK()) &&
               cmpObjects(this.parentTag, rhs.getParentTag());
    }

    static private boolean cmpObjects(Object lhs, Object rhs){
        return (null == lhs)? (null == rhs): lhs.equals(rhs);
    }
}
