/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator.handler;

import com.netapp.oci.platform.originator.ObjState;
import com.netapp.oci.platform.originator.OriginatorUpdate;
import com.netapp.oci.platform.originator.OriginatorUpdateException;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.key.KeyCacheException;
import com.netapp.oci.platform.originator.key.UuidKey;
import com.onaro.sanscreen.server.dataobject.Utils;

import java.sql.Timestamp;
import java.util.*;

/**
 * @author Al Velella
 *         Date: Sep 30, 2011
 *         Time: 5:03:39 PM
 */
public class HandlerResult {
    private ObjState objState;
    private Map<String, String> properties;
    private boolean cachedColumnData;
    private ColumnData columnData;
    private RowData rowData;
    private List<String> changeDetailAttributes;
    private ObjState.State previousState;

    private UuidKey uuidKey;


    /**
     * Standard database fields are 255 in length
     */
    private static final int VARCHAR_255 = 255;
    private static final int VARCHAR_512 = 512;
    private static final int VARCHAR_1024 = 1024;
    private boolean isHierarchy;

    public HandlerResult(OriginatorUpdate originatorUpdate, XmlHandler xmlHandler, Map<String, String> properties,
                         String keyAttribute, Map<UuidKey, ObjState> keyToObjState, boolean isHierarchy,
                         Set<ColumnAttr> keyAttrs)
            throws OriginatorUpdateException, KeyCacheException {
        this.isHierarchy = isHierarchy;
        this.properties = properties;
        EnumSet<ColumnAttr> keyAttrEnums = EnumSet.of(ColumnAttr.REQUIRED, ColumnAttr.KEY);
        keyAttrEnums.addAll(keyAttrs);
        String uuid = (String) fetchValueFromProperties(keyAttribute, String.class, keyAttrEnums);

        // if parentTagName is not null, then is a child handler with properties inherited from a parent
        String tagName = xmlHandler.getTagName();
        String parentTagName = originatorUpdate.getParentRelationshipTagName(tagName);

        // Build a UuidKey for this object. Use the parent key factory if this is a child handler.
        this.uuidKey = originatorUpdate.buildKey(parentTagName == null ? tagName : parentTagName, uuid);
        this.objState = keyToObjState.get(uuidKey);
        this.previousState = objState == null ? null : objState.getState();

        // If no objstate then this is a new object
        if (this.objState == null) {
            initializeObjState(originatorUpdate, keyToObjState, tagName, parentTagName);
        }


        this.changeDetailAttributes = originatorUpdate.getChangeDetailAttributes(tagName);

        this.columnData = originatorUpdate.getColumnData(xmlHandler.getTableName());
        if (columnData == null) {
            this.columnData = initializeColumnData(originatorUpdate, xmlHandler, keyAttribute, keyAttrs);
        } else {
            // cachedColumnData = true, so the handler will skip building the already generated ColumnData
            this.cachedColumnData = true;
        }

        setRowData(initializeRowData(originatorUpdate.getOriginId()));

    }

    public HandlerResult(OriginatorUpdate originatorUpdate, XmlHandler xmlHandler, Map<String, String> properties, String keyAttribute, Map<UuidKey, ObjState> keyToObjState, boolean isHierarchy) throws OriginatorUpdateException, KeyCacheException {
        this(originatorUpdate, xmlHandler, properties, keyAttribute, keyToObjState, isHierarchy, EnumSet.noneOf(ColumnAttr.class));
    }

    public HandlerResult(OriginatorUpdate originatorUpdate, XmlHandler xmlHandler, Map<String, String> properties, String keyAttribute, Map<UuidKey, ObjState> keyToObjState) throws OriginatorUpdateException, KeyCacheException {
        this(originatorUpdate, xmlHandler, properties, keyAttribute, keyToObjState, false, EnumSet.noneOf(ColumnAttr.class));
    }


    private RowData initializeRowData(Long originId) {
        List<Object> values = new ArrayList<Object>();
        values.add(uuidKey);

        if (isHierarchy) {
            List<Object> baseValues = new ArrayList<Object>();
            baseValues.add(uuidKey);
            return new RowData(objState.getObjid(), originId, baseValues, values, uuidKey.hashCode(), 0);

        } else {
            return new RowData(this.objState.getObjid(), originId, values, this.uuidKey.hashCode(), 0);

        }
    }

    private ColumnData initializeColumnData(OriginatorUpdate originatorUpdate, XmlHandler xmlHandler, String keyAttribute,
                                            Set<ColumnAttr> keyAttrs) {
        this.cachedColumnData = false;

        List<String> names = new ArrayList<String>();
        names.add(keyAttribute.toLowerCase());

        List<Class<?>> types = new ArrayList<Class<?>>();
        types.add(UuidKey.class);

        List<EnumSet<ColumnAttr>> attributes = new ArrayList<EnumSet<ColumnAttr>>();
        EnumSet<ColumnAttr> keyAttrEnums = EnumSet.of(ColumnAttr.REQUIRED, ColumnAttr.KEY);
        keyAttrEnums.addAll(keyAttrs);
        attributes.add(keyAttrEnums);

        // cachedColumnData = false, so the handler will build ColumnData as the rest of the attributes are parsed
        ColumnData columnData;
        if (isHierarchy) {
            List<String> baseNames = new ArrayList<String>();
            baseNames.add(keyAttribute.toLowerCase());

            List<Class<?>> baseTypes = new ArrayList<Class<?>>();
            baseTypes.add(UuidKey.class);

            List<EnumSet<ColumnAttr>> baseAttributes = new ArrayList<EnumSet<ColumnAttr>>();
            baseAttributes.add(keyAttrEnums);

            columnData = new ColumnData(names, types, attributes, baseNames, baseTypes, baseAttributes);
        } else {
            columnData = createColumnData(names, types, attributes);
        }
        originatorUpdate.cacheColumnData(xmlHandler.getTableName(), this.columnData);

        return columnData;
    }

    protected ColumnData createColumnData(List<String> names, List<Class<?>> types, List<EnumSet<ColumnAttr>> attributes) {
        return new ColumnData(names, types, attributes);
    }

    private void initializeObjState(OriginatorUpdate originatorUpdate, Map<UuidKey, ObjState> keyToObjState, String tagName, String parentTagName) throws OriginatorUpdateException, KeyCacheException {
        // Generate a new object ID. If this is a child handler we want to re-use the parent's object ID
        KeyCache keyCache = originatorUpdate.getIdCache();
        Long objId = keyCache.reserveId(originatorUpdate.getOriginId(), tagName, uuidKey, parentTagName);

        // new object state
        this.objState = new ObjState(objId, ObjState.State.CREATED, 0, 0);
        keyToObjState.put(uuidKey, this.objState);

        // to conserve memory replace parent key with the newly generated child key
        if (parentTagName != null) {
            originatorUpdate.replaceKey(parentTagName, uuidKey);
        }
    }

    public ColumnData getColumnData() {
        return columnData;
    }

    protected void setRowData(RowData rowData) {
        this.rowData = rowData;
    }

    public RowData getRowData() {
        return rowData;
    }

    public Object getRowDataValue(String colName) {
        return rowData.getValues().get(columnData.getNames().indexOf(colName));

    }

    public Object getBaseRowDataValue(String colName) {
        return rowData.getBaseValues().get(columnData.getBaseNames().indexOf(colName));
    }

    public boolean isSubColumn(String colName) {
        return columnData.getNames().contains(colName);
    }

    public boolean isBaseColumn(String colName) {
        return isHierarchy && columnData.getBaseNames().contains(colName);
    }

    public ObjState.State calcState() {
        if (objState.getState() != ObjState.State.CREATED && getRowData().getChecksum() != objState.getChecksum()) {
            objState.setState(ObjState.State.UPDATED);
        }
        return objState.getState();
    }

    public ObjState.State getPreviousState() {
        return previousState;
    }

    public boolean hasChangeDetailReqAttributesChanged() {
        return getRowData().getChangeDetailChecksum() != objState.getChangeDetailChecksum();
    }

    public HandlerResult column(String xmlAttributeName, Long objid) {
        return column(xmlAttributeName, objid, EnumSet.noneOf(ColumnAttr.class));
    }

    public HandlerResult column(String xmlAttributeName, Long objId, EnumSet<ColumnAttr> attribute) {
        final boolean changeDetailRequired = changeDetailAttributes.contains(xmlAttributeName);
        if (!cachedColumnData) {
            columnData.add(xmlAttributeName.toLowerCase(), long.class, attribute, changeDetailRequired);
        }

        if (changeDetailRequired) {
            getRowData().addChangeDetailReqValue(objId);
        } else {
            getRowData().add(objId);
        }

        return this;
    }

    public HandlerResult column(String xmlAttributeName, Class<?> type) {
        return column(xmlAttributeName, type, EnumSet.noneOf(ColumnAttr.class));
    }

    public HandlerResult column(String xmlAttributeName, Class<?> type, EnumSet<ColumnAttr> attribute) {
        final boolean changeDetailRequired = changeDetailAttributes.contains(xmlAttributeName);
        if (!cachedColumnData) {
            columnData.add(xmlAttributeName.toLowerCase(), type, attribute, changeDetailRequired);
        }

        Object value = fetchValueFromProperties(xmlAttributeName, type, attribute);

        if (changeDetailRequired) {
            getRowData().addChangeDetailReqValue(value);
        } else {
            getRowData().add(value);
        }

        return this;
    }

    public HandlerResult subColumn(String xmlAttributeName, Long objid) {
        return subColumn(xmlAttributeName, objid, EnumSet.noneOf(ColumnAttr.class));
    }

    public HandlerResult subColumn(String xmlAttributeName, Long objId, EnumSet<ColumnAttr> attribute) {
        column(xmlAttributeName, objId, attribute);
        return this;
    }

    public HandlerResult subColumn(String xmlAttributeName, Class<?> type) throws OriginatorUpdateException {

        return subColumn(xmlAttributeName, type, EnumSet.noneOf(ColumnAttr.class));
    }

    public HandlerResult subColumn(String xmlAttributeName, Class<?> type, EnumSet<ColumnAttr> attribute) throws OriginatorUpdateException {
        if (!isHierarchy) {
            throw new OriginatorUpdateException("Does not represent 'base-sub' hierarchy relationship. ");
        }
        column(xmlAttributeName, type, attribute);
        return this;
    }

    public HandlerResult baseColumn(String xmlAttributeName, Long objid) throws OriginatorUpdateException {
        return baseColumn(xmlAttributeName, objid, EnumSet.noneOf(ColumnAttr.class));
    }

    public HandlerResult baseColumn(String xmlAttributeName, Long objId, EnumSet<ColumnAttr> attribute) throws OriginatorUpdateException {
        if (!isHierarchy) {
            throw new OriginatorUpdateException("Does not represent 'base-sub' hierarchy relationship. ");

        }
        final boolean changeDetailRequired = changeDetailAttributes.contains(xmlAttributeName);
        if (!cachedColumnData) {
            getColumnData().baseAdd(xmlAttributeName.toLowerCase(), long.class, attribute, changeDetailRequired);
        }

        if (changeDetailRequired) {
            getRowData().addBaseChangeDetailReqValue(objId);
        } else {
            getRowData().addBaseValue(objId);
        }

        return this;
    }

    public HandlerResult baseColumn(String xmlAttributeName, Class<?> type) throws OriginatorUpdateException {
        return baseColumn(xmlAttributeName, type, EnumSet.noneOf(ColumnAttr.class));
    }

    public HandlerResult baseColumn(String xmlAttributeName, Class<?> type, EnumSet<ColumnAttr> attribute) throws OriginatorUpdateException {
        if (!isHierarchy) {
            throw new OriginatorUpdateException("Does not represent 'base-sub' hierarchy relationship. ");
        }
        final boolean changeDetailRequired = changeDetailAttributes.contains(xmlAttributeName);
        if (!cachedColumnData) {
            getColumnData().baseAdd(xmlAttributeName.toLowerCase(), type, attribute, changeDetailRequired);
        }

        Object value = fetchValueFromProperties(xmlAttributeName, type, attribute);

        if (changeDetailRequired) {
            getRowData().addBaseChangeDetailReqValue(value);
        } else {
            getRowData().addBaseValue(value);
        }

        return this;
    }


    protected Object fetchValueFromProperties(String xmlAttributeName, Class<?> type, EnumSet<ColumnAttr> attribute) {
        String valueStr = properties.get(xmlAttributeName);
        if (valueStr == null && attribute.contains(ColumnAttr.REQUIRED)) {
            throw new IllegalArgumentException("Failed to get " + xmlAttributeName + " value from properties:" + properties);
        }

        Object value;
        if (valueStr == null) {
            value = null;
        } else if (type == String.class) {
            int colLength;
            if (attribute.contains(ColumnAttr.STRING_1024)) {
                colLength = VARCHAR_1024;
            } else if (attribute.contains(ColumnAttr.STRING_512)) {
                colLength = VARCHAR_512;
            } else {
                colLength = VARCHAR_255;
            }

            if (attribute.contains(ColumnAttr.TEXT)) {
                value = valueStr;
            }
            else {
                value = truncateString(valueStr, colLength);
            }
        } else if (type == Enum.class) {
            value = valueStr;
        } else if (type == long.class) {
            value = Long.valueOf(valueStr);
        } else if (type == int.class) {
            value = Integer.valueOf(valueStr);
        } else if (type == float.class) {
            value = Float.valueOf(valueStr);
        } else if (type == boolean.class) {
            value = Boolean.valueOf(valueStr);
        } else if (type == Timestamp.class) {
            Long valueLong = Long.valueOf(valueStr); //expect this in seconds.
            value = new Timestamp(valueLong * 1000); //needs milliseconds
        } else {
            throw new IllegalArgumentException("Unexpected type " + type);
        }
        return value;
    }

    private String truncateString(String valueStr, int length) {
        if (valueStr != null && valueStr.length() > length) {
            return valueStr.substring(0, length);
        }
        return valueStr;
    }


    public Long getObjId() {
        return this.objState.getObjid();
    }

    public UuidKey getUuidKey() {
        return uuidKey;
    }

    public boolean isHierarchy() {
        return isHierarchy;
    }
}
