/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator.handler;

import com.netapp.oci.platform.originator.ObjState;
import com.netapp.oci.platform.originator.OriginatorUpdate;
import com.netapp.oci.platform.originator.OriginatorUpdateException;
import com.netapp.oci.platform.originator.key.KeyCacheException;
import com.netapp.oci.platform.originator.key.UuidKey;
import com.onaro.sanscreen.server.tables.idgenerator.IdGenerator;
import com.onaro.sanscreen.server.tables.idgenerator.IdGeneratorException;

import java.sql.Connection;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author Al Velella Date: Sep 22, 2011 Time: 12:00:11 PM
 */
public abstract class XmlHandler {
	private String tagName;
	private String tableName;
	private String baseTableName;

	public XmlHandler(String tagName, String tableName) {
		this.tagName = tagName;
		this.tableName = tableName;
	}

	public XmlHandler(String tagName, String tableName, String baseTableName) {
		this.tagName = tagName;
		this.tableName = tableName;
		this.baseTableName = baseTableName;
	}

	public String getTagName() {
		return tagName;
	}

	public String getTableName() {
		return tableName;
	}

	/**
	 * Returns the key attribute.
	 * <p/>
	 * It is expected that handlers whose key attribute is different than the
	 * typical "uuid" attribute will override this method.
	 * 
	 * @return the key attribute
	 */
	public String getKeyAttr() {
		return "UUID";
	}

	/**
	 * Indicates whether the database records created by this handler needs to
	 * be preserved for future reference after it gets deleted.
	 * 
	 * @return <code>true</code> if the history needs to be kept. Else,
	 *         <code>false</code>
	 */
	public boolean keepHistory() {
		return false;
	}

	protected Long createId() throws OriginatorUpdateException {
		try {
			return IdGenerator.getInstance().getNextId();
		} catch (IdGeneratorException e) {
			throw new OriginatorUpdateException("Failed to create id", e);
		}
	}

	public boolean isObjectOffline(OriginatorUpdate originatorUpdate,
			Map<String, String> objectProperties) {
		return false;
	}

	protected boolean isObjectOffline(String offlinePropertyName,
			String expectedValue, Map<String, String> objectProperties) {
		final String actualValue = objectProperties.get(offlinePropertyName);
		return expectedValue.equals(actualValue);
	}

	/**
	 * Build compound UUID. First value must be non-null. Optional added values.
	 * Format will be "val1[,val2,val3,...,valN]".
	 * 
	 * @param firstValue
	 *            required, basis value for UUID.
	 * @param addedValues
	 *            optional var arg list of additional key values.
	 * @return UUID string key in format val1[,val2,val3,...,valN]
	 */
	protected String buildCompoundUUID(final Object firstValue,
			final Object... addedValues) {
		final StringBuilder uuid = new StringBuilder(firstValue.toString());
		for (final Object val : addedValues) {
			uuid.append(Character.valueOf(',')).append(val.toString());
		}
		return uuid.toString();
	}

	public Long handleObject(OriginatorUpdate originatorUpdate,
			Map<String, String> properties,
			Map<UuidKey, ObjState> keyToHashcode,
			List<XmlChildTagDetails> dependentChildrenForOfflineTags)
			throws OriginatorUpdateException, KeyCacheException {
		HandlerResult handlerResult = parseProperties(originatorUpdate,
				properties, keyToHashcode);
		ObjState.State state = handlerResult.calcState();

		boolean isObjOffline = originatorUpdate.getPreserveOfflineObjectsFlag()
				&& isObjectOffline(originatorUpdate, properties);
		if (isObjOffline && state == ObjState.State.UPDATED) {
			originatorUpdate.addOfflineRunners(handlerResult.getObjId(),
					getTagName(), dependentChildrenForOfflineTags);
		} else if (!isObjOffline
				&& (handlerResult.getPreviousState() == ObjState.State.DELETED_OFFLINE_OBJ || handlerResult
						.getPreviousState() == ObjState.State.PRESERVED)) {

			// DE5056 Burt 771357: child objects PRESERVED forever after a node
			// failover
			//
			// Consider the case where a node fails, and an aggregate with
			// children (aggregate, volume, lun...)
			// fails over. The children are marked PRESERVED because of the node
			// offline, but they
			// may also be reported accurately in the same report, under a new
			// node, resulting in a
			// new parent-child relationship.

			// When the original node comes back online, we won't find these
			// children to un-PRESERVE,
			// because they are associated with a new node.

			// So when we see PRESERVED objects that are online, we must honor
			// the object's report
			// of children and clear their PRESERVED state, in order to avoid
			// PERSERVEing them
			// forever.

			originatorUpdate.addOnlineRunners(handlerResult.getObjId(),
					getTagName(), dependentChildrenForOfflineTags);
		}

		if (state == ObjState.State.CREATED) {
			return originatorUpdate.insertObject(handlerResult, this);
		} else if (state == ObjState.State.UPDATED) {
			return originatorUpdate.updateObject(handlerResult, this,
					isObjOffline);
		} else {
			return originatorUpdate.retainObject(handlerResult, getTagName(),
					isObjOffline);
		}
	}

	public abstract HandlerResult parseProperties(
			OriginatorUpdate originatorUpdate, Map<String, String> properties,
			Map<UuidKey, ObjState> keyToObjState)
			throws OriginatorUpdateException, KeyCacheException;

	public String getBaseTableName() {
		return baseTableName;
	}

	public String alternateKeyToPrimaryKey(Connection connection,
			OriginatorUpdate originatorUpdate, Map<String, String> map)
			throws OriginatorUpdateException {
		return null;
	}

	/**
	 * Adjust the change types of deleted objects. 
	 * <p>
	 * For an example, if VolumeMoveJob is deleted because the volume is deleted, 
	 * then change type need be set to DELETED; if VolumeMoveJob is deleted because 
	 * the volume is moved again, then change type need be set to UPADTED 
	 * (in this case, the VolumeMoveJob is not marked as DEAD, instead its 
	 * inactive attribute is iset to true)
	 * 
	 * @param connection
	 *            Database connection
	 * @param deletedObjs
	 *            A map of tag name to deleted objects whose change types are
	 *            DELETED
	 * @param deletedHistoryObjs
	 *            A map of tag name to deleted objects whose change types are
	 *            UPDATED
	 * @throws OriginatorUpdateException
	 *             If fail to adjust the change types of deleted objects
	 */
	public void adjustDeletedObjects(Connection connection,
			Map<String, Set<Long>> deletedObjs,
			Map<String, Set<Long>> deletedHistoryObjs)
			throws OriginatorUpdateException {

	}

}
