/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator.key;

import com.netapp.oci.netapp.originator.handler.ClusterHandler;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParser;
import com.netapp.oci.platform.originator.OriginatorUpdateException;
import com.netapp.oci.platform.originator.PlatformRegistry;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.plugin.data.PlatformPlugin;
import com.onaro.sanscreen.server.tables.idgenerator.IdGenerator;
import com.onaro.sanscreen.server.tables.idgenerator.IdGeneratorException;
import com.onaro.sanscreen.util.Schema;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;


/**
 * @author Al Velella
 *         Date: Nov 20, 2012
 *         Time: 4:14:02 PM
 */
public class KeyCache {

    private static final KeyCache INSTANCE = new KeyCache();

    /**
     * Map of tag(Cluster/Volume) -> originId -> key(uuid) -> objid
     *
     * The inner map (originId -> (uuid -> objid)) must be thread safe as it will be accessed by multiple originators simultaneously.
     * The outer maps are built during plugin registration and do not need to be synchronized.
     */

    private static Map<String, ConcurrentHashMap<Long, Map<UuidKey, Long>>> tagToOriginIdToKeyToId = new HashMap<String, ConcurrentHashMap<Long, Map<UuidKey, Long>>>();

    /**
     * List of new UUIDs added by originator. If the originator transaction fails these IDs must be removed from the key-cache
     */
    private static Map<Long, List<KeyCommit>> commitOrigToKeyCommits = new HashMap<Long, List<KeyCommit>>();

     /**
     * Map of originId to Dead UUID keys
     */
    private static ConcurrentHashMap<Long, Map<String, Set<UuidKey>>> originIdToDeadKeys = new ConcurrentHashMap<Long, Map<String, Set<UuidKey>>>();

    /**
     * Platform registry
     */
    private static PlatformRegistry platformRegistry;

    public static boolean containsDeadObjects(Long originId) {
        Map<String, Set<UuidKey>> tagToDeadKeys = getTagToDeadKeys(originId);

        for (Set<UuidKey> keys : tagToDeadKeys.values()) {
            if (!keys.isEmpty()) {
                return true;
            }
        }

        return false;
    }

    private static class KeyCommit {
        enum CommitType {
            ADD,
            SOFT_DELETE,
            RESURRECT,
            PURGE_SOFT_DELETED,
            PURGE_REGULAR_DELETE
        }

        private String tag;
        private UuidKey uuidKey;
        private CommitType type;
        private Long objid;
        private Long originId;

        private KeyCommit(String tag, UuidKey uuidKey, CommitType type, Long objid, Long originId) {
            this.tag = tag;
            this.uuidKey = uuidKey;
            this.type = type;
            this.objid = objid;
            this.originId = originId;
        }

        @Override
        public String toString() {
            return tag + ':' + uuidKey + '-' + type;
        }
    }

    private KeyCache() {
    }

    public static void clearAll() throws TableException {
        tagToOriginIdToKeyToId = new HashMap<String, ConcurrentHashMap<Long, Map<UuidKey, Long>>>();
        commitOrigToKeyCommits = new HashMap<Long, List<KeyCommit>>();
        originIdToDeadKeys = new ConcurrentHashMap<Long, Map<String, Set<UuidKey>>>();
    }
    
    /**
     * Init an empty cache for the plugin
     *
     * @param plugin plugin to init
     */
    public static void clear(PlatformPlugin plugin) throws TableException {
        Map<String, ConcurrentHashMap<Long, Map<UuidKey, Long>>> tagToOriginIdToKeyToId = new HashMap<String, ConcurrentHashMap<Long, Map<UuidKey, Long>>>();
        for (Map.Entry<String, XmlHandler> entry : plugin.getTagToXmlHandlers().entrySet()) {
            String tagName = entry.getKey();
            final ConcurrentHashMap<Long, Map<UuidKey, Long>> originIdToKeyToId = new ConcurrentHashMap<Long, Map<UuidKey, Long>>();
            tagToOriginIdToKeyToId.put(tagName, originIdToKeyToId);
        }

        commitOrigToKeyCommits =  new HashMap<Long, List<KeyCommit>>();
        originIdToDeadKeys = new ConcurrentHashMap<Long, Map<String, Set<UuidKey>>>();
    }

    /**
     * Init plugin-specific key cache from database
     *
     * @param connection sql connection
     * @param plugin     plugin to init
     * @throws TableException exception
     */
    public static void init(Connection connection, PlatformPlugin plugin) throws TableException {
        platformRegistry = plugin.getRegistry();
        for (Map.Entry<String, XmlHandler> entry : plugin.getTagToXmlHandlers().entrySet()) {
            final String tagName = entry.getKey();
            // final Map<UuidKey, Long> keyToId = new HashMap<UuidKey, Long>();
            final ConcurrentHashMap<Long, Map<UuidKey, Long>> originIdToKeyToId = new ConcurrentHashMap<Long, Map<UuidKey, Long>>();
            tagToOriginIdToKeyToId.put(tagName, originIdToKeyToId);

            final XmlHandler handler = entry.getValue();

            String keyAttr = handler.getKeyAttr();
            String table = handler.getTableName();


            String sql = String.format("SELECT %s, objid,  originId, objState = 'DEAD' FROM %s ", keyAttr, table);
            if (handler.keepHistory()) {
                sql += " WHERE inactive=0";
            }
            PreparedStatementWrapper.create(connection, sql, new TableResultParser<Object>() {
                @Override
                public List<Object> parseResult(ResultSet resultSet) throws SQLException {
                    while (resultSet.next()) {
                        final UuidKey uuidKey = platformRegistry.buildUuidKey(tagName, resultSet.getString(1));
                        Long objId = Long.valueOf(resultSet.getLong(2));
                        Long originId = Long.valueOf(resultSet.getLong(3));
                        final boolean isObjDead = resultSet.getBoolean(4);

                        Map<UuidKey, Long> keyToId = originIdToKeyToId.get(originId);
                        if(keyToId == null)  {
                            keyToId = new HashMap<UuidKey, Long>();
                            originIdToKeyToId.put(originId, keyToId);
                            keyToId.put(uuidKey, objId);
                        }
                        else {
                            keyToId.put(uuidKey, objId);
                        }
                        if (isObjDead) {
                            Map<String, Set<UuidKey>> tagToDeadKeys = getTagToDeadKeys(originId);
                            Set<UuidKey> deadKeys = tagToDeadKeys.get(tagName);
                            if (deadKeys==null) {
                                deadKeys = new HashSet<UuidKey>();
                                tagToDeadKeys.put(tagName, deadKeys);
                            }
                            deadKeys.add(uuidKey);
                        }
                    }
                    return null;
                }
            }).executeQueryEx();
        }

    }


    /**
     * Retrieve plugin cache
     *
     * @return KeyCache instance
     */
    public static KeyCache get() {
        return INSTANCE;
    }

    /**
     * Checks if the request object is in dead state, if it is in dead state, the object is resurrected and its
     * old objId is returned.
     * Otherwise, generates a new ID for the requested object.
     * The first originator describing an object owns it. ID generation will fail if another originator attempts ID
     * creation.
     *
     * @param originId originator requesting ID creation
     * @param tagName  object name being created as corresponding to XML handler
     * @param keyValue key of the object to-be-created object
     * @return unique ID for the object
     * @throws OriginatorUpdateException - IDGenerator exception
     * @throws KeyCacheException         - originator attempted to create an ID for object already owned by a different originator
     */
    public Long reserveId(long originId, String tagName, UuidKey keyValue, String relationshipParentTagName) throws OriginatorUpdateException, KeyCacheException {
        // does this object already exist but is in dead state?
        if (isObjectDead(originId, tagName, keyValue)) {
            // If yes, resurrect the dead object.
            return resurrectDeadObj(tagName, originId, keyValue);
        }

        if (relationshipParentTagName == null) {
            Long objId = createId(); // create a new id
            return makeReservation(objId, originId, tagName, keyValue);
        } else {
            Long parentsObjId = findId(originId, relationshipParentTagName, keyValue); // re-use parents' objId
            if (parentsObjId == null) {
                throw new OriginatorUpdateException("Could not find parent UUID " + keyValue);
            }
            return makeReservation(parentsObjId.longValue(), originId, tagName, keyValue);
        }
    }


    public boolean isObjectDead(long originId, String tagName, UuidKey keyValue) {
        Map<String, Set<UuidKey>> tagToDeadKeys = getTagToDeadKeys(originId);
        return tagToDeadKeys.containsKey(tagName) && tagToDeadKeys.get(tagName).contains(keyValue);
    }

    private Long resurrectDeadObj(String tagName, Long originId, UuidKey keyValue) {
        Map<String, Set<UuidKey>> tagToDeadKeys = getTagToDeadKeys(originId);
        if(tagToDeadKeys.containsKey(tagName)){
            tagToDeadKeys.get(tagName).remove(keyValue);
            queueKeyCommit(originId, tagName, keyValue, KeyCommit.CommitType.RESURRECT);
        }
        return findId(originId, tagName, keyValue);
    }

    /**
     * Insert a key-cache reservation
     *
     * @param originId originator requesting ID creation
     * @param tagName  object name being created as corresponding to XML handler
     * @param keyValue key of the object to-be-created object
     * @param objid object ID
     * @return unique ID for the object
     * @throws OriginatorUpdateException - IDGenerator exception
     * @throws KeyCacheException         - originator attempted to create an ID for object already owned by a different originator
     */
    private Long makeReservation(long objid, long originId, String tagName, UuidKey keyValue) throws OriginatorUpdateException, KeyCacheException {
        Map<UuidKey, Long> keyToId = tagToOriginIdToKeyToId.get(tagName).get(originId);
        if(keyToId!=null) {
            Long oldObjid = keyToId.get(keyValue);
            if (oldObjid!=null && oldObjid != objid) {
                throw new KeyCacheException(tagName + " " + keyValue + " is owned by objid " + oldObjid + " and can't be used by objid " + objid + " for originId " + originId);
            }
            else {
                keyToId.put(keyValue, objid);
            }
        }
        else {
            keyToId = new HashMap<UuidKey, Long>();
            keyToId.put(keyValue, objid);
            tagToOriginIdToKeyToId.get(tagName).put(originId, keyToId);
        }
        queueKeyCommit(originId, tagName, keyValue, KeyCommit.CommitType.ADD);
        return objid;
    }

    /**
     * soft delete uuidkeys
     * <p/>
     * The object will stay in cache but is going to be added to deadUuidKeys list.
     *
     * @param originId  originator
     * @param tagToUuid tag/UUIDs to delete
     *
     */
    public void softDelete(Long originId, Map<String, List<UuidKey>> tagToUuid) {
        Map<String, Set<UuidKey>> tagToDeadKeys = getTagToDeadKeys(originId);

        for (Map.Entry<String, List<UuidKey>> entry : tagToUuid.entrySet()) {
            for (UuidKey uuid : entry.getValue()) {
                Set<UuidKey> deadKeys = tagToDeadKeys.get(entry.getKey());
                if(deadKeys == null){
                    deadKeys = new HashSet<UuidKey>();
                    tagToDeadKeys.put(entry.getKey(), deadKeys);
                }
                deadKeys.add(uuid);
                queueKeyCommit(originId, entry.getKey(), uuid, KeyCommit.CommitType.SOFT_DELETE);
            }
        }
    }

    /**
     * purges the given ids from cache
     *
     * @param originId originator
     * @param tagNameToIds
     */
    public void purgeSoftDeletedObjects(Long originId, Map<String, Set<Long>> tagNameToIds) {
        for (Map.Entry<String, Set<Long>> entry : tagNameToIds.entrySet()) {
            purgeSoftDeletedObjects(originId, entry.getKey(), entry.getValue());
        }
    }

    /**
     * purges the given ids from cache.
     *
     * @param originId originator
     * @param tagName
     * @param ids
     */
    private void purgeSoftDeletedObjects(Long originId, String tagName, Set<Long> ids) {
        Map<UuidKey, Long> uuidToId = tagToOriginIdToKeyToId.get(tagName).get(originId);
        Map<String, Set<UuidKey>> tagToDeadKeys = getTagToDeadKeys(originId);

        if (tagToDeadKeys.containsKey(tagName)) {
            Set<UuidKey> deadUuidKeys = tagToDeadKeys.get(tagName);
            Map<Long, UuidKey> idToKey = getIdToKey(tagName, originId);
            for (Long id : ids) {
                UuidKey key = idToKey.get(id);
                if(uuidToId != null) {
                    uuidToId.remove(key);
                }
                deadUuidKeys.remove(key);
                queueKeyCommit(originId, tagName, key, KeyCommit.CommitType.PURGE_SOFT_DELETED, id);
            }
        }
    }

    public void purgeRegularObjects(Long originId, Map<String,Set<UuidKey>> tagToUuidKeysToPurge) {

        for (Map.Entry<String, Set<UuidKey>> entry : tagToUuidKeysToPurge.entrySet()) {
            String tagName = entry.getKey();
            for (UuidKey uuidKey : entry.getValue()) {
                Map<UuidKey, Long> keyToId = tagToOriginIdToKeyToId.get(tagName).get(originId);
                if(keyToId!=null) {
                    Long oldObjId =  keyToId.get(uuidKey);
                    keyToId.remove(uuidKey);
                    queueKeyCommit(originId, tagName, uuidKey, KeyCommit.CommitType.PURGE_REGULAR_DELETE, oldObjId);
                }
            }
        }
    }

    private static Map<String, Set<UuidKey>> getTagToDeadKeys(Long originId) {
        return originIdToDeadKeys.computeIfAbsent(originId, k -> new HashMap<String, Set<UuidKey>>());
    }


    public static void removeOriginatorCache(Long originatorId) {
        for(ConcurrentHashMap<Long, Map<UuidKey, Long>> originIdToKeyToId : tagToOriginIdToKeyToId.values()) {
            Map<UuidKey, Long> keyToId = originIdToKeyToId.remove(originatorId);
            if(keyToId!=null) {
                keyToId.clear();
            }
        }
        commitOrigToKeyCommits.remove(originatorId);
        originIdToDeadKeys.remove(originatorId);
    }

    public Map<Long, UuidKey> getIdToKey(String tagName, Long originId) {
        Map<UuidKey, Long> uuidKeyToId = tagToOriginIdToKeyToId.get(tagName).get(originId);

        HashMap<Long, UuidKey> idToUuidKey = new HashMap<Long, UuidKey>();
        if(uuidKeyToId!=null) {
            for (Map.Entry<UuidKey, Long> entry : uuidKeyToId.entrySet()) {
                idToUuidKey.put(entry.getValue(), entry.getKey());
            }
        }
        return idToUuidKey;
    }

    /**
     * Return id of key
     *
     * @param originId the originator ID
     * @param tag  tag name
     * @param key key string
     * @return id or NULL if key is not in key cache
     */
    public Long findId(Long originId, String tag, String key) {
        UuidKey uuidKey = platformRegistry.buildUuidKey(tag, key);
        return findId(originId, tag, uuidKey);
    }

    public Long findId(Long originId, String tag, UuidKey key) {
        Long id = null;
        ConcurrentHashMap<Long, Map<UuidKey, Long>> originIdToKeyToId = tagToOriginIdToKeyToId.get(tag);
        if(originIdToKeyToId!=null)  {
            Map<UuidKey, Long> keyToId = originIdToKeyToId.get(originId);
            if(keyToId!=null) {
                id = keyToId.get(key);
            }
        }
        return id;
    }

    /**
     *
     * @param clusterKey  UuidKey for clusterKey
     * @return
     */
    public long findOriginatorByClusterKey(UuidKey clusterKey) {
        long originIdRet = -1L;

        ConcurrentHashMap<Long, Map<UuidKey, Long>>  originIdToKeyToId = tagToOriginIdToKeyToId.get(ClusterHandler.TAG);
        for(Map.Entry<Long, Map<UuidKey, Long>> entry : originIdToKeyToId.entrySet()) {
            Long originId = entry.getKey();
            Map<UuidKey, Long> keyToId =  entry.getValue();
            if(keyToId.containsKey(clusterKey))  {
                originIdRet = originId;
                break;
            }
        }
        return originIdRet;
    }

    /**
     * Originator is about to report and may be making ID reservation
     *
     * @param schema   schema for key-cache
     * @param originId originator that is reporting
     */
    public static void reportOriginator(Schema schema, Long originId) {
        List<KeyCommit> keyCommits = commitOrigToKeyCommits.get(originId);
        if (keyCommits == null) {
            keyCommits = new ArrayList<KeyCommit>();
            commitOrigToKeyCommits.put(originId, keyCommits);
        }
        keyCommits.clear();

    }

    /**
     * Persist any ID reservation created by the originator
     *
     * @param schema   schema for key-cache
     * @param originId originator that is committing
     */
    public static void commitOriginator(Schema schema, Long originId) {
        List<KeyCommit> keyCommits = commitOrigToKeyCommits.get(originId);
        if (keyCommits != null) {
            keyCommits.clear();
        }
    }

    /**
     * Rolls back all changes for the provided originator
     * If the change type was
     * ADD: removes the reservation from the cache
     * RESURRECT: resets the reservation  to DEAD' (which is disassociating the object from its originator)
     * DELETE: restores the reservation to mark the object 'LIVE' (i.e. associates the object back to its original originator)
     *
     * @param schema   schema for key-cache
     * @param originId originator that is rolling back
     */
    public static void rollbackOriginator(Schema schema, Long originId) {
        List<KeyCommit> keyCommits = commitOrigToKeyCommits.get(originId);
        Map<String, Set<UuidKey>> tagToDeadKeys = getTagToDeadKeys(originId);

        if (keyCommits != null) {
            for (KeyCommit keyCommit : keyCommits) {
                Map<UuidKey, Long> keyToIds = tagToOriginIdToKeyToId.get(keyCommit.tag).get(originId);
                if(keyToIds == null) {
                    continue;
                }
                switch (keyCommit.type) {
                    case ADD:
                        // remove it from the cache.
                        keyToIds.remove(keyCommit.uuidKey);
                        break;

                    case RESURRECT:
                        // soft Delete it
                        tagToDeadKeys.get(keyCommit.tag).add(keyCommit.uuidKey);
                        break;

                    case SOFT_DELETE:
                        // Resurrect it
                        tagToDeadKeys.get(keyCommit.tag).remove(keyCommit.uuidKey);
                        break;

                    case PURGE_SOFT_DELETED:
                        // soft-delete it
                        // We are doing put instead of putIfAbsent because we expect the reservation to be absent from the map.
                        keyToIds.put(keyCommit.uuidKey, keyCommit.objid);
                        tagToDeadKeys.get(keyCommit.tag).add(keyCommit.uuidKey);
                        break;

                    case PURGE_REGULAR_DELETE:
                        // Resurrect it
                        // We are doing put instead of putIfAbsent because we expect the reservation to be absent from the map.
                        keyToIds.put(keyCommit.uuidKey, keyCommit.objid);
                        break;
                }
            }
            keyCommits.clear();
        }
    }

    /**
     * Replace key object used in key-cache map with passed-in key
     *
     * @param originId originator
     * @param tagName tag name
     * @param uuidKey replacement key
     */
    public void replaceKey(Long originId, String tagName, UuidKey uuidKey) throws KeyCacheException {
        Map<UuidKey, Long> keyToId = tagToOriginIdToKeyToId.get(tagName).get(originId);
        if(keyToId!=null) {
            keyToId.put(uuidKey, keyToId.remove(uuidKey));
        }
    }

    /**
     * Modified uuidKeys with passed-in new uuidKeys
     * @param originId   originator
     * @param tagName tag name
     * @param uuidKeyMap   Map of old uuidKeys to new uuidKeys
     * @throws KeyCacheException
     */
    public synchronized void modifyUuidKeys(Long originId, String tagName, Map<UuidKey, UuidKey> uuidKeyMap) throws KeyCacheException {
        Map<UuidKey, Long> keyToId = tagToOriginIdToKeyToId.get(tagName).get(originId);
        if(keyToId!=null)   {
            for (Map.Entry<UuidKey, UuidKey> entry : uuidKeyMap.entrySet()) {
                Long id = keyToId.remove(entry.getKey());
                if (id != null) {
                    keyToId.put(entry.getValue(), id);
                }
            }
        }
    }


    private Long createId() throws OriginatorUpdateException {
        try {
            return IdGenerator.getInstance().getNextId();
        } catch (IdGeneratorException e) {
            throw new OriginatorUpdateException("Failed to create id", e);
        }
    }

    private void queueKeyCommit(Long originId, String tagName, UuidKey uuidKey, KeyCommit.CommitType commitType) {
        queueKeyCommit(originId, tagName, uuidKey, commitType, null);
    }

    private void queueKeyCommit(Long originId, String tagName, UuidKey uuidKey, KeyCommit.CommitType commitType, Long objid) {
        List<KeyCommit> uuidKeysToCommit = commitOrigToKeyCommits.get(originId);
        if (uuidKeysToCommit == null) {
            uuidKeysToCommit = new ArrayList<KeyCommit>();
            commitOrigToKeyCommits.put(originId, uuidKeysToCommit);
        }

        uuidKeysToCommit.add(new KeyCommit(tagName, uuidKey, commitType, objid, originId));
    }

}
