/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.plugin.data;

import com.netapp.oci.platform.acq.interfaces.data.PackagePollStatus;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.originator.OriginatorChanges;
import com.netapp.oci.platform.originator.PlatformRegistry;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.runner.DbRunner;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import com.onaro.sanscreen.util.Schema;

import java.io.Serializable;
import java.sql.Connection;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author Al Velella
 *         Date: Nov 10, 2011
 *         Time: 5:05:23 PM
 */
public interface PlatformPlugin extends Serializable {

    /**
     * Get schema
     *
     * @return
     */
    Schema getSchema();

    /**
     * Get build number
     *
     * @return
     */
    String getBuildNumber();

    /**
     * Delete data from all configuration tables
     *
     * @param connection connection
     * @throws ServerInterfaceException
     */
    void deleteTables(Connection connection) throws ServerInterfaceException;

    /**
     * Return XML Handlers to parse schema
     *
     * @return map of element-name to handler
     */
    Map<String, XmlHandler> getTagToXmlHandlers();

    /**
     * Return DB Runners for post-processing after a configuration update
     *
     * @return map of element-name to runners
     */
    Map<String, DbRunner> getTagToDbRunners();

    /**
     * Return map of tag to resource keys
     *
     * @return
     */
    Map<String, String> getTagToResourceType();

    /**
     * Used by the framework to do additional post processing after originatorUpdate
     *
     * @param connection
     * @param originId
     * @param datasourceId
     * @param time
     * @throws ServerInterfaceException
     */
    void postOriginatorUpdate(Connection connection, long originId, long datasourceId, long time) throws ServerInterfaceException;


    /**
     * Provides a <code>Map</code> of all the attributes for which detailed change information is needed, keyed by
     * the attribute's element tag name.
     *
     * @return <code>Map</code> of all the attributes for which detailed change information is needed.
     */
    Map<String, List<String>> getTagToChangeDetailAttributes();

    /**
     * The root tag
     *
     * @return set of root tag
     */
    Set<String> getRootTagnames();

    /**
     * The root Id
     *
     * @return id
     */
    long getRootId(Connection connection, long originId) throws TableException;

    /**
     * Return reference to update_history table
     *
     * @return table name
     */
    String getUpdateHistoryTableName();

    /**
     * Return reference to delete_detail table
     *
     * @return table name
     */
    String getDeleteDetailTableName();

    /**
     * Return map of columns to copy into delete_detail table. Each column is associated to a tagname
     *
     * @return map of column name to associated tag
     */
    Map<String, String> getDeleteDetailColumnToTagname();

    /**
     * Return table hierarchy
     *
     * @return return list of table names from higher to lower hierarchy
     */
    List<String> getTableOrder();

    /**
     * Plugin-implemented configuration databases
     *
     * @return database names
     */
    Set<String> getConfigurationDatabases();

    /**
     * Plugin-implemented performance databases with their corresponding tables.
     *
     * @return Map of database name -> set of table names in each database.
     */
    Map<String, Set<String>> getPerformanceDatabasesAndTables(Connection connection) throws ServerInterfaceException;

    /**
     * Used by the framework to inform the plugin about the changes caused by data source reporting.
     *
     * @param connection   Database connection.
     * @param datasourceId Datasource ID
     * @param changes      Changes information.
     * @throws ServerInterfaceException if there were any issues while handling change information.
     */
    void handleOriginatorChanges(Connection connection, long datasourceId, OriginatorChanges changes) throws ServerInterfaceException;

    /**
     * Used by the framework to inform the plugin about the changes are committed tot he database.
     *
     * @param connection     Database connection.
     * @param datasourceId   Datasource ID
     * @param datasourceName Datasource Name
     * @param changes        Changes information.
     * @throws ServerInterfaceException if there were any issues while handling change information.
     */
    void changesCommitted(Connection connection, long datasourceId, String datasourceName, OriginatorChanges changes) throws ServerInterfaceException;

    /**
     * Used by the framework to inform the plugin about the originator being removed from the database.
     *
     * @param connection Database connection.
     * @param originId   Originator Id
     * @param updateTime Update time
     * @throws ServerInterfaceException
     */
    void logOriginatorRemove(Connection connection, long originId, long rootId, long updateTime) throws ServerInterfaceException;


    /**
     * Used by the framework to inform the plugin about the objects being purged.
     *
     * @param connection Database connection.
     * @param updateTime Update time
     * @throws ServerInterfaceException
     */
    void logObjectsPurgeJob(Connection connection, long rootId, Map<String, Set<Long>> tagToIdsBeingDeleted, long updateTime) throws ServerInterfaceException, TableException;

    /**
     * Prunes expired database records.
     *
     * @param connection Database connection.
     * @throws ServerInterfaceException If unexpected issue arise.
     */
    void prune(Connection connection, long startTime) throws ServerInterfaceException;

    /**
     * Acquisition package status message
     *
     * @param connection Database connection
     * @param pollStatus Status information from acquisition
     * @throws ServerInterfaceException if there were any issues while handling package message
     */
    void handlePackageMessage(Connection connection, PackagePollStatus pollStatus) throws ServerInterfaceException;

    void logUpdateHistory(Connection connection, long originatorId, long startTime, long updateTime, long duration, int physicalChangeCount, int runnerCount) throws TableException;

    <T> T getParam(String key, Class<T> clazz, T defaultValue);

    /**
     * Return PlatformRegistry
     *
     * @return plugin-specific registry
     */
    PlatformRegistry getRegistry();

    /**
     * notify to appropriate parties that the originator is being removed.
     * @param connection
     * @param datasourceId
     * @param datasourceName
     * @param rootId
     * @param updateTime
     * @throws ServerInterfaceException
     */
    void notifyOriginatorRemove(Connection connection, long datasourceId, String datasourceName, long rootId, long updateTime) throws ServerInterfaceException;
}
