/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.upgrade;

import com.netapp.oci.platform.common.InternalServerException;
import com.netapp.oci.platform.common.table.*;
import com.netapp.oci.platform.db.SqlFile;
import org.slf4j.Logger;

import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collections;
import java.util.List;
import java.util.Set;

/**
 * @author Nancy Walia
 *         Date: Sep 10, 2012
 */
public class UpgradeRunner {
    private static Logger logger =org.slf4j.LoggerFactory.getLogger(UpgradeRunner.class);

    private Connection connection;
    private long starttime;
    private Version initialVersion;
    private Version currentVersion;
    private UpgradeRegistry upgradeRegistry;
    private String databasePrefix = "";

    private static final Version VERSION_0 = new Version(0, 0, 0, 0);

    public UpgradeRunner(UpgradeRegistry upgradeRegistry, Connection connection) {
        this.connection = connection;
        this.upgradeRegistry = upgradeRegistry;
    }

    public UpgradeRunner(UpgradeRegistry upgradeRegistry, Connection connection, String databasePrefix) {
        this.connection = connection;
        this.upgradeRegistry = upgradeRegistry;
        this.databasePrefix = databasePrefix;
    }

    public void run() throws InternalServerException {
        try {
            starttime = System.currentTimeMillis();

            currentVersion = getVersion();
            initialVersion = new Version(currentVersion);

            UpgradeRegistry.Entry dropViewUpgrader = upgradeRegistry.getDropViewsUpgrader();
            //There are no views in Cloud
            if (dropViewUpgrader != null)
                executeUpgrader(null, dropViewUpgrader);

            String configuarationDB = getConfigurationDb();
            if (currentVersion == VERSION_0) {
                logger.info("Initialize OCI " + configuarationDB + " database");
                createDB();
                currentVersion = getVersion();
            }

            Version latestVersion = upgradeRegistry.getLatestVersion();
            if (latestVersion.equals(currentVersion)) {
                logger.info("OCI " + configuarationDB + " database okay at version " + latestVersion);
            } else {
                logger.info("Starting OCI " + configuarationDB + " upgrade");
                
                UpgradeRegistry.Entry finishUpgraderEntry = upgradeRegistry.getFinishUpgrader();
                List<Version> versions = upgradeRegistry.getVersions();
                for (Version version : versions) {
                    if (version.compareTo(currentVersion) > 0) {
                        UpgradeRegistry.Entry entry = upgradeRegistry.getUpgrader(version);
                        executeUpgrader(version, entry);
                        currentVersion = version;
                        executeUpgrader(null, finishUpgraderEntry);
                    }
                }
            }
            UpgradeRegistry.Entry createViewUpgrader = upgradeRegistry.getCreateViewsUpgrader();
            //There are no views in Cloud
            if (createViewUpgrader != null)
                executeUpgrader(null, createViewUpgrader);
        } catch (Exception e) {
            logger.error("Failed to load upgraders", e);
            throw new InternalServerException("Failed to load upgraders", e);
        }
    }

    public long getStartTime() {
        return starttime;
    }

    public Version getInitialVersion() {
        return initialVersion;
    }

    public Version getCurrentVersion() {
        return currentVersion;
    }

    private void createDB() throws InternalServerException, IOException {
        List<String> createDBs = upgradeRegistry.getCreateDBs();

        // drop any old database
        Collections.reverse(createDBs);
        for (String createDB : createDBs) {
            String db = createDB.substring(createDB.indexOf("create_db_") + "create_db_".length(), createDB.indexOf(".sql"));
            sqlUpdate("DROP DATABASE IF EXISTS " + db).executeUpdateEx();
        }

        // create databases
        Collections.reverse(createDBs);
        for (String createDB : createDBs) {
            SqlFile.executeFile(connection, createDB, upgradeRegistry.getClass(), upgradeRegistry.getSqlDirBase(), "");
        }
    }

    private Version getVersion() throws InternalServerException {
    	String configurationDb = getConfigurationDb();
    	try {
    		return sqlQuery("SELECT major, minor, maintenance, patch FROM " + configurationDb + ".version", new TableResultParserDefault<Version>() {
    			@Override
    			public Version parseObject(ResultSet resultSet) throws SQLException {
    				return new Version(resultSet.getInt(1),
    						resultSet.getInt(2),
    						resultSet.getInt(3),
    						resultSet.getInt(4));
    			}
    		}).executeQueryEx().iterator().next();

    	}
    	catch(Exception e){
    		logger.info("Netapp_model version table missing, assuming fresh install. Reason: " + e.toString());
    		logger.info("Wiping out the database");
    		logger.info("Create " + configurationDb);
    		return VERSION_0;
    	}
    }
  
    private void executeUpgrader(Version version, UpgradeRegistry.Entry entry) throws TableException, IOException {
        logger.info(version == null ? entry.getDescription() : "Upgrader " + version + ": " + entry.getDescription());

        String sqlFile = entry.getSqlFile();
        if (sqlFile != null) {
            SqlFile.executeFile(connection, sqlFile, upgradeRegistry.getClass(), upgradeRegistry.getSqlDirBase(), "");
        }
        Upgrader upgrader = entry.getUpgrader();
        if (upgrader != null) {
            upgrader.run(this);
        }
    }

    public <T> PreparedStatementWrapper<T> sqlQuery(String sql, TableResultParserDefault<T> tableParser) throws TableException {
        return PreparedStatementWrapper.create(connection, sql, tableParser);
    }

    public <T> PreparedStatementWrapper<T> sqlUpdate(String sql) throws TableException {
        return PreparedStatementWrapper.createForUpdate(connection, sql);
    }

    public long sqlFunctionQuery(String sql) throws TableException {
        return PreparedStatementWrapper.executeFunctionQuery(connection, sql, -1);
    }

    public String getConfigurationDb() {
        return PrefixConnection.translateDbPrefix(upgradeRegistry.getConfigurationDb());
    }

    public String getPerformanceDb() {
        return PrefixConnection.translateDbPrefix(upgradeRegistry.getPerformanceDb());
    }

    public void createTemporaryTableForJoin(Set<Long> ids, String temp) throws TableException {
        Table.createIdsTemporaryTableForJoin(connection, temp, ids);
    }
}