/*
 * Copyright (c) 2010 NetApp
 * All rights reserved
 */

package com.netapp.oci.server;

import com.netapp.oci.platform.common.table.ConnectionHelper;

import javax.sql.DataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.Connection;

/**
 * Provides an in-memory state of the last time an {@link UpdateTask}, or
 * similar management task, was executed.  This class helps protect against
 * chronological issues which might arise when a back-dated task -- which is
 * potentially stale -- is executed.
 */
public class TransactionContext {

    private static final Logger LOG = LoggerFactory.getLogger(TransactionContext.class);

    /**
     * the transaction time
     */
    private final long time;

    private final Connection connection;

    /**
     * Cached time of last transaction.  Modifications to this variable should
     * be synchronized on {@link TransactionContext.class}.
     */
    private static volatile long lastUpdateTime = 0;

    /**
     *
     * @param clazz the running class using this
     * @throws TransactionContextException
     */
    public TransactionContext(Class<?> clazz) throws TransactionContextException {
        this(clazz, null);
    }

    /**
     *
     * @param clazz     the running class using this
     * @param dataSource the container's {@link javax.sql.DataSource}
     * @throws TransactionContextException
     */
    public TransactionContext(Class<?> clazz, DataSource dataSource) throws TransactionContextException {
        this.time = generateTime(clazz);
        if (dataSource == null) {
            this.connection = ConnectionHelper.newConnectionNoEjb();
        } else {
            this.connection = ConnectionHelper.newConnection(dataSource);
        }
    }

    /**
     * protect the server from negative changes to the clock
     *
     * @param clazz
     * @return
     * @throws TransactionContextException
     */
    private static synchronized long generateTime(Class<?> clazz) throws TransactionContextException {
        long lastUpdateTime = getLastUpdateTime();
        long currentTime = System.currentTimeMillis();

        final long delta = lastUpdateTime - currentTime;
        if (delta > 0) {
            // a time drift has occurred (e.g., system clock changed); these
            // should naturally resolve themselves once the delta disappears
            StringBuilder sb = new StringBuilder(128)
                    .append("[class=").append(clazz.getSimpleName())
                    .append(", currentTime=").append(currentTime)
                    .append(", lastUpdateTime=").append(lastUpdateTime)
                    .append("] unexpected time drift of ").append(delta / 1000L)
                    .append(" seconds");
            throw new TransactionContextException(sb.toString());
        }

        while (currentTime == lastUpdateTime) {
            currentTime = System.currentTimeMillis();
        }

        LOG.debug("{} generated time {}", clazz.getSimpleName(), currentTime);

        TransactionContext.lastUpdateTime = currentTime;

        return currentTime;
    }

    public static synchronized void setLastUpdateTime(long updatetime) {
        lastUpdateTime = Math.max(getLastUpdateTime(), updatetime);
    }

    public static synchronized long getLastUpdateTime() {
        if (lastUpdateTime <= 0) {
                lastUpdateTime = System.currentTimeMillis();
        }
        return lastUpdateTime;
    }

    public static synchronized void resetLastUpdateTime() {
        lastUpdateTime = 0;
    }

    public long getTime() {
        return time;
    }

    public Connection getConnection() {
        return connection;
    }

    public void close() {
        ConnectionHelper.close(connection);
    }

    public static void close(TransactionContext transactionContext) {
        if (transactionContext != null) {
            transactionContext.close();
        }
    }

}
