/*
 * $Id: //depot/prod/acquisition/main/tools/config/eclipse/.settings/org.eclipse.jdt.ui.prefs#1 $
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.sanscreen.server.api;

import java.io.IOException;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.ThreadPoolExecutor;

import javax.annotation.Resource;
import javax.ejb.EJB;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.Unmarshaller;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.sax.SAXSource;

import com.onaro.sanscreen.acquisition.sessions.AcquisitionFacadeSessionLocal;
import netapp.manage.NaElement;
import netapp.manage.NaServer;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession;
import com.netapp.oci.netapp.client.interfaces.data.EmsZapiManagerException;
import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.ems.EmsZapiManager;
import com.netapp.oci.netapp.ems.shared.interfaces.data.NetappEmsEvent;
import com.netapp.oci.netapp.ems.shared.interfaces.impl.NetappEmsEventUtility;
import com.netapp.oci.netapp.ems.shared.interfaces.local.EmsMsgListener;
import com.netapp.oci.netapp.ems.shared.interfaces.local.NetAppEmsEventPersistence;
import com.netapp.oci.netapp.ems.shared.interfaces.local.NetappEmsMsgListenerRegistry;
import com.netapp.oci.netapp.interfaces.data.NetappEms;
import com.netapp.oci.netapp.interfaces.data.NetappEms.EmsMessageInfo.Parameters;
import com.netapp.oci.netapp.interfaces.data.NetappEmsEventDto;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;

/**
 * REST endpoint to handle EMS Post request.
 * URI: /acquisition/ONTAP/ems
 */
public class EmsPostServlet extends HttpServlet {
    private static final long serialVersionUID = -1L;
    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(EmsPostServlet.class);

    private static final String SAX_FEATURE_VALIDATING = "http://xml.org/sax/features/validation";
    private static final String SAX_FEATURE_DTD = "http://apache.org/xml/features/nonvalidating/load-external-dtd";
    private static final String EMS_CORRECTIVE_ACTION_KEY = "action";
    private static final String EMS_CORRECTIVE_ACTION_NONE = "None";

    @EJB
    private NetAppEmsEventPersistence emsEventPersistenceBean;

    /**
     * use this bean to get the registered EMS listeners
     */
    @EJB
    private NetappEmsMsgListenerRegistry emsListenerRegistry;

    @EJB
    private AcquisitionServerSession acquisitionServerSession;

    @EJB
    private AcquisitionFacadeSessionLocal acquisitionLocal;

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private javax.sql.DataSource dataSource;

    @Override
    public void doPost(HttpServletRequest request, HttpServletResponse response)
        throws ServletException, IOException {

        try {
            JAXBContext jc = JAXBContext.newInstance(NetappEms.class);

            // set DTD and XSD validation off
            SAXParserFactory spf = SAXParserFactory.newInstance();
            spf.setFeature(SAX_FEATURE_DTD, false);
            spf.setFeature(SAX_FEATURE_VALIDATING, false);
            spf.setValidating(false);

            XMLReader xmlReader = spf.newSAXParser().getXMLReader();
            SAXSource source = new SAXSource(xmlReader, new InputSource(request.getReader()));

            Unmarshaller unmarshaller = jc.createUnmarshaller();
            NetappEms ems = (NetappEms)unmarshaller.unmarshal(source);

            if (acquisitionLocal.isEmsSupported(ems.getEmsMessageInfo().getMessageName())) {
                logger.info("Received ems = " + ems.getEmsMessageInfo().getMessageName() + " to process.");

                // fetch remediation and inject into NetappEms as a parameter
                Parameters.Parameter actionParameter = new Parameters.Parameter();
                actionParameter.setName(EMS_CORRECTIVE_ACTION_KEY);
                final String action = getEmsRemediation(ems.getEmsMessageOrigin().getClusterUuid(), ems.getEmsMessageInfo().getMessageName());
                actionParameter.setValue(action == null ? EMS_CORRECTIVE_ACTION_NONE : action);
                ems.getEmsMessageInfo().getParameters().getParameter().add(actionParameter);

                // Convert the message to entity object before persisting in the MEGA netapp_model
                NetappEmsEvent emsEntity = NetappEmsEventUtility.convertNetAppEms(ems);
                emsEventPersistenceBean.insertEvent(emsEntity);

                // Convert the message to a dto object list before sending it to ems listeners
                NetappEmsEventDto emsDto = NetappEmsEventUtility.convertToDto(emsEntity);
                // asynchronously process the EMS and return to the caller
                Map<String, EmsMsgListener> emsListeners = emsListenerRegistry.getEmsMsgListeners();
                if (!emsListeners.isEmpty()) {
                    logger.info("Sending EMS [" + ems.getEmsMessageInfo().getMessageName() + "] to listeners [" + emsListeners.keySet() + "] for processing.");
                    ThreadPoolExecutor executor = (ThreadPoolExecutor)request.getServletContext().getAttribute(AppContextListener.EMS_EXECUTOR_ID);
                    executor.execute(new EmsPostProcessor(new ArrayList<NetappEmsEventDto>(Arrays.asList(emsDto)), emsListeners.values()));
                }
            } else {
                logger.info("Received ems = " + ems.getEmsMessageInfo().getMessageName() + " which is not supported.");
            }

            // reply with 202 for acknowledge
            response.setStatus(HttpServletResponse.SC_ACCEPTED);
        } catch (Throwable e) {
            logger.error("EMS message is malformed; ", e);
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
        }
    }

    private String getEmsRemediation(final String clusterUUID, final String msgName) throws EmsZapiManagerException {
        Connection dbConnection = null;
        NaServer zapiConnection = null;

        try {
            dbConnection = ConnectionHelper.newConnection(dataSource);
            long clusterId = NetAppModel.INSTANCE.getClusterIdByClusterUUID(dbConnection, clusterUUID);
            Properties props = NetAppModel.INSTANCE.getClusterDatasourceProperties(dbConnection, clusterId);
            zapiConnection = EmsZapiManager.INSTANCE.createZAPIConnection(props);
            NaElement catalog = EmsZapiManager.INSTANCE.getEmsEventCatalogByMsgName(Arrays.asList(msgName), zapiConnection);
            NaElement attributesList = catalog.getChildByName("attributes-list");

            String action = "None";
            if (attributesList != null) {
                if (attributesList.getChildByName("event-catalog-info").hasChildren()) {
                    action = attributesList.getChildByName("event-catalog-info").getChildContent("action");
                }
            }

            return action;
        } catch (TableException e) {
            logger.error("Unable to locate ds by clusterUUID '{}', cluster is under discovery.", clusterUUID);
            return null;
        } finally {
            if (dbConnection != null) {
                ConnectionHelper.close(dbConnection);
            }

            if (zapiConnection != null) {
                zapiConnection.close();
            }
        }

    }

    @Override
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException {
        logger.info("Not supported!");
    }

}
