/*
 * $Id: //depot/prod/acquisition/main/tools/config/eclipse/.settings/org.eclipse.jdt.ui.prefs#1 $
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.sanscreen.server.api;

import java.io.ByteArrayInputStream;
import java.nio.charset.StandardCharsets;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.stream.Collectors;

import javax.ejb.EJB;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.Unmarshaller;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.sax.SAXSource;

import netapp.manage.NaAPIFailedException;
import netapp.manage.NaElement;
import netapp.manage.NaServer;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.ems.shared.interfaces.data.NetAppEmsEventException;
import com.netapp.oci.netapp.ems.shared.interfaces.data.NetappEmsEvent;
import com.netapp.oci.netapp.ems.shared.interfaces.impl.NetappEmsEventUtility;
import com.netapp.oci.netapp.ems.shared.interfaces.local.EmsMsgListener;
import com.netapp.oci.netapp.ems.shared.interfaces.local.NetAppEmsEventPersistence;
import com.netapp.oci.netapp.ems.EmsZapiManager;
import com.netapp.oci.netapp.interfaces.data.NetAppEmsConfig;
import com.netapp.oci.netapp.interfaces.data.NetappEmsEventDto;
import com.netapp.oci.netapp.interfaces.data.NetappEmsZapi;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;

/**
 * A thread to invoke registered EMS listeners to process the EMS message.
 */
public class EmsPullProcessor implements Runnable {
    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(EmsPullProcessor.class);

    private final long id;
    private static final long MILLISEC_IN_SEC = 1000;
    private final Connection connection;
    private static final String SAX_FEATURE_VALIDATING = "http://xml.org/sax/features/validation";
    private static final String SAX_FEATURE_DTD = "http://apache.org/xml/features/nonvalidating/load-external-dtd";

    private NetAppEmsEventPersistence netAppEmsEventPersistence;
    List<EmsMsgListener> listeners;

    public EmsPullProcessor(long id, Connection connection, NetAppEmsEventPersistence netAppEmsEventPersistence, List<EmsMsgListener> listeners) {
        this.id = id;
        this.connection = connection;
        this.netAppEmsEventPersistence = netAppEmsEventPersistence;
        this.listeners = listeners;
    }

    /**
     * The run method fetches ems events, inserts them in the database and
     * send them to all EMS listeners for processing
     * for a datasource.
     */
    @Override
    public void run() {
        NaServer zapiConnection = null;
        List<NetappEmsEvent> emsEntities = null;

        try {
            final String clusterUuid = NetAppModel.INSTANCE.getClusterUUIDByDatasourceID(connection, id);
            if (clusterUuid == null) {
                // cluster hasn't been discovered yet
                return;
            }

            List<NetAppEmsConfig> emsConfigs = NetAppModel.INSTANCE.getEmsConfigs(connection);
            final String clusterVersion = NetAppModel.INSTANCE.getClusterVersionByDsId(connection, id);
            List<NetAppEmsConfig> filterRlst = emsConfigs.stream()
                    .filter(emsConfig -> {
                        return emsConfig.getOntapVersion().compareTo(clusterVersion) <= 0 &&
                               !emsConfig.getState().equals(NetAppEmsConfig.EmsMessageState.TO_BE_DELETED) &&
                               !emsConfig.getState().equals(NetAppEmsConfig.EmsMessageState.DELETED);
                     })
                    .collect(Collectors.toList());

            // add the filter rules and remove duplicates
            Set<String> filterRules = new HashSet<>();
            filterRlst.forEach(rslt -> filterRules.add(rslt.getMessageFilterRule()));
            if (filterRules.isEmpty()) {
                // nothing to poll
                return;
            }

            // Make a zapi call to fetch events beginning from last received event + 1
            long lastEventTimestamp = netAppEmsEventPersistence.getLastEventTimestamp(clusterUuid);
            long fromTimestamp;

            // If no previous events stored for a datasource, get lastUpdateTime from netapp_model.cluster table for the cluster
            if (lastEventTimestamp == 0) {
                // timestamp in netapp_model.cluster tables is in ms, convert it to seconds
                lastEventTimestamp = NetAppModel.INSTANCE.getLastUpdateTimeByClusterUUID(connection, clusterUuid)/MILLISEC_IN_SEC;
                // if lastUpdateTime in netapp_model.cluster is also null, do not fetch any ems events for that cluster
                if (lastEventTimestamp == 0) {
                    logger.info("Not pulling ems events for cluster with uuid {}.", clusterUuid);
                    return;
                }
                fromTimestamp = lastEventTimestamp;
            } else {// fetch events beginning from last received event + 1
                fromTimestamp = lastEventTimestamp + 1;
            }
            if (!filterRules.isEmpty()) {
                Properties dsProperties = NetAppModel.INSTANCE.getDatasourceProperties(connection, id);
                zapiConnection = EmsZapiManager.INSTANCE.createZAPIConnection(dsProperties);
                NaElement zapiResponse = EmsZapiManager.INSTANCE.getEmsEvents(zapiConnection, filterRules, fromTimestamp);
                // unmarshal the events
                if (zapiResponse != null) {
                    NetappEmsZapi zapiEmsEvents = unmarshalZapiEmsEvents(zapiResponse);
                    if (zapiEmsEvents != null) {
                        emsEntities = NetappEmsEventUtility.convertNetAppEmsZapi(zapiEmsEvents, clusterUuid, zapiConnection);
                        // Persist the EMS message in the MEGA netapp_model.ems_event and netapp_model.ems_event_parameter table
                        // Before persisting sort them in increasing order of timestamp
                        if (emsEntities != null && !emsEntities.isEmpty()) {
                            Collections.reverse(emsEntities);
                            netAppEmsEventPersistence.insertEvents(emsEntities);
                            logger.info("Received " + emsEntities.size() + " by ZAPI pull for processing.");
                            // Convert the EMS events to a dto object list before sending it to ems listeners
                            List<NetappEmsEventDto> emsDtoList = new ArrayList<>();
                            for (NetappEmsEvent emsEvent : emsEntities) {
                                emsDtoList.add(NetappEmsEventUtility.convertToDto(emsEvent));
                            }

                            logger.info("Sending " + emsDtoList.size() + " EMS events to " + listeners.size() + " listeners for processing.");
                            for (EmsMsgListener listener : listeners) {
                                listener.process(emsDtoList);
                            }
                            logger.info(emsDtoList.size() + " EMS events have been processed by " + listeners.size() + " listeners.");
                        }
                    }
                } else {
                    logger.warn("No EMS events retrieved through ZAPI call for clusterUUid - " + clusterUuid);
                }
            }
        } catch (NaAPIFailedException e) {
            logger.error("Error fetching EMS events from ZAPI", e);
        } catch (NetAppEmsEventException e) {
            logger.error("Error converting EMS events or inserting EMS events in the database", e);
        } catch (TableException e) {
            logger.error("Error accessing database", e);
        } finally {
            if (zapiConnection != null) {
                zapiConnection.close();
            }
            ConnectionHelper.close(connection);
        }
    }

    private NetappEmsZapi unmarshalZapiEmsEvents(NaElement zapiResponse) {
        NetappEmsZapi zapiEmsEvents = null;
        try {
            JAXBContext jc = JAXBContext.newInstance(NetappEmsZapi.class);

            // set DTD and XSD validation off
            SAXParserFactory spf = SAXParserFactory.newInstance();
            spf.setFeature(SAX_FEATURE_DTD, false);
            spf.setFeature(SAX_FEATURE_VALIDATING, false);
            spf.setValidating(false);

            XMLReader xmlReader = spf.newSAXParser().getXMLReader();
            SAXSource source = new SAXSource(xmlReader, new InputSource(new ByteArrayInputStream(zapiResponse.toString().getBytes(StandardCharsets.UTF_8))));

            Unmarshaller unmarshaller = jc.createUnmarshaller();
            zapiEmsEvents = (NetappEmsZapi)unmarshaller.unmarshal(source);
        } catch (Exception e) {
            logger.error("Error unmarshalling EMS events from ZAPI", e);
        }
        return zapiEmsEvents;
    }
}
