package com.netapp.sanscreen.server.api;

import com.onaro.sanscreen.util.Schema;

import java.net.MalformedURLException;
import java.net.URL;

/**
 * Miscellaneous client utilities for OriginatorUpdateServlet
 * @author Daniel Nuriyev
 */
public class OriginatorUpdateClientUtils {

    /**
     * The Acquisition context root is located at /acquisition so need
     * to make sure all calls are sent there.
     */
    private static final String APPLICATION_CONTEXT_ROOT = "acquisition";

    /**
     * Creates URL for the originator update such as https://DOMAIN:443/api/originatorUpdate/SCHEMA/ORIGINATOR_ID
     * @param baseURL base URL such as https://domain.name.com:443/
     * @param schema schema name as defined in Schema enum
     * @param origanatorID data source originator ID
     * @return URL for the originator update
     */
    public static String createURL(String baseURL, Schema schema, String origanatorID) throws MalformedURLException {

        if(baseURL == null) throw new MalformedURLException("Null baseURL");

        baseURL = baseURL.trim();
        if(!baseURL.startsWith("http://") && !baseURL.startsWith("https://")){
            throw new MalformedURLException("Unsupported protocol in base URL: " + baseURL);
        }

        /*
        this line checks that baseURL is formatted correctly
        otherwise it will throw its own MalformedURLException
         */
        new URL(baseURL);

        if(schema == null) throw new MalformedURLException("Null schema");

        StringBuilder url = new StringBuilder(100);
        url.append(baseURL);

        if(!baseURL.endsWith("/")) url.append('/');

        return url.append(APPLICATION_CONTEXT_ROOT + "/originatorUpdate/")
                .append(schema).append('/')
                .append(origanatorID)
                .toString();

    }

    /**
     * Throws an exception if the response HTTP status code is not 200
     * @param statusCode HTTP status code
     * @param statusMessage
     * @param body
     * @param error
     * @throws Exception
     */
    public static void assertResponse(int statusCode, String statusMessage, String body, String error) throws Exception{

        if(statusCode < 200 || statusCode > 299){
            if(error == null) throw new Exception("Server returned HTTP status " + statusCode);
            else throw new Exception(error.trim());
        }
    }
    
}
