/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.sanscreen.server.api;

import com.netapp.oci.profiler.ProfilerService;
import com.netapp.oci.profiler.ProfilerServiceFactory;
import com.netapp.oci.server.UpdateTaskException;
import com.onaro.sanscreen.server.sessions.OriginatorUpdateSessionLocal;
import com.onaro.sanscreen.util.JndiHelper;
import com.onaro.sanscreen.util.Schema;

import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.InputStream;
import java.io.InputStreamReader;

/**
 * Reads XML as stream and passes it on for SAX parsing
 * The URL is https://DOMAIN/api/originatorUpdate/SHEMA/originId
 * Example: https://localhost/api/originatorUpdate/SANSCREEN/1
 * The response of the servlet is:
 *  HTTP Status 200 on success
 *  HTTP Status 500 on error, and the Java exception message 
 * @author Daniel Nuriyev
 */
public class OriginatorUpdateServlet extends HttpServlet {

    private static final long serialVersionUID = -1L;
    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(OriginatorUpdateServlet.class);

    public void doPost(HttpServletRequest request, HttpServletResponse response) {
        InputStream in = null;
        String pathInfo = null;
        try{
            //pathInfo looks like: /originatorUpdate/SCHEMA/originId  todo: get data source name from the URL itself.
            pathInfo = request.getPathInfo();
            logger.debug(pathInfo);
            String[] path = pathInfo.split("/");

            //if any of the params is null or wrong, the code will throw an exception
            Schema schema = Schema.valueOf(path[path.length - 2]);
            Long originId = new Long(path[path.length - 1]);

            final ProfilerService profilerService = ProfilerServiceFactory.getInstance();
            profilerService.log("Originator Update Start", originId);

            OriginatorUpdateSessionLocal ejb = (OriginatorUpdateSessionLocal) JndiHelper.getEJB(OriginatorUpdateSessionLocal.JNDI_NAME);
            in = request.getInputStream();
            ejb.getUpdate(new InputStreamReader(in, "UTF-8"), schema, originId, null);
            in.close();
            in = null;

            profilerService.log("Originator Update End", originId);

            //this is the protocol between this servlet and a client in case of success
            response.setStatus(HttpServletResponse.SC_OK);
        } catch (UpdateTaskException ute) {
            logger.error(ute.getMessage(), ute);
            // We can't provide details of the failure in the http response, because:
            //  1) the parser has closed the request input stream before acq has finished writing to it.
            //  2) writing to the response output stream will block both acq and server:
            //          acq is waiting for the server to finish reading the input stream
            //          server is waiting of acq to read the output stream.
            //  3) If server returns an error status, then the library used by acq won't
            //          allow access to the response document, only the response error stream
            //
            //  So instead of providing the full feedback in a response document, we will include
            //  the head of the stack trace in a header that acq may access
            response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
            final String rawStackTrace = ute.getNestedStackTrace();
            final int maxEndIndex = 1024;
            final int atIndex = rawStackTrace.indexOf("\tat ");
            final int endIndex = (atIndex > 0) ? atIndex : maxEndIndex;
            final String shortStackTrace = rawStackTrace.substring(0, endIndex);
            response.setHeader("OCIE_ORIGINATOR_UPDATE_FAILURE_REASON", shortStackTrace);

        } catch (Throwable x) {
            logger.error(x.getMessage(), x);
            throw new ServletException("Failed to process originator update request for " + pathInfo + " because of " + x.getMessage(), x);
        } finally {
            try{
                if (in != null) {
                    in.close();
                }
            } catch (Exception e) {
                logger.error(e.getMessage(), e);
            }
        }
    }
}
