package com.netapp.sanscreen.server.profiler;

import com.netapp.oci.platform.common.ServerSystem;
import com.netapp.oci.profiler.*;

import java.io.File;
import java.io.IOException;
import java.util.Timer;

@SuppressWarnings("unused")
/**
 * Profiler service implementation for the server JVM.
 */
public class ServerProfilerService extends BaseProfilerService {

    /**
     * Prefix for the output file.
     */
    private static final String PROFILE_OUTPUT_PREFIX = "Server";

    /**
     * Default sampling interval (in seconds) for the CPU and memory profiler task.
     */
    private static final String DEFAULT_SAMPLE_INTERVAL = "60";

    /**
     * System property that specifies the sampling interval (in seconds)
     */
    private static final String SAMPLE_INTERVAL_SYS_PROPERTY = "profiler.interval";

    public ServerProfilerService() throws ProfilerServiceException {
        try {
            String[] columnHeaders = new String[] {                    
                    CpuMemoryProfiler.CPU_USAGE,
                    CpuMemoryProfiler.USED_HEAP_SIZE,
                    CpuMemoryProfiler.CURRENT_HEAP_SIZE,
                    CpuMemoryProfiler.MAX_HEAP_SIZE,
                    "MySQL Data Size",
                    "MySQL CPU Usage",
                    "MySQL Mem Usage",
                    "Originator Update Start",
                    "Originator Update End"
            };

            String serverLogDir = System.getProperty("server.logs");
            File outputFile = new File(serverLogDir + File.separator + PROFILE_OUTPUT_PREFIX + "_" + System.currentTimeMillis() + ".csv");
            setDataStore(new CsvFileDataStore(outputFile, columnHeaders));
        } catch (IOException e) {
            throw new ProfilerServiceException(e);
        }
    }

    @Override
    public void start() {
        Timer timer = new Timer("CPU Memory Profiler");
        String mysqlDataDir = ServerSystem.getMySqlDataDir(System.getProperty("onaro.home"));

        Profiler dataSizeProfiler;
        if (ServerSystem.isLinux()) {
            // on linux we dedicated a disk to mysql
            dataSizeProfiler = new DiskPartitionSizeProfiler("MySQL Data Size", mysqlDataDir);
        } else {
            dataSizeProfiler = new DirSizeProfiler("MySQL Data Size", new File(mysqlDataDir));
        }

        ProfilerTimerTask task = new ProfilerTimerTask(this, new Profiler[] {
                new CpuMemoryProfiler(),
                dataSizeProfiler,
                new ExtProcCpuMemoryUsageProfiler(new String[] {"MySQL CPU Usage", "MySQL Mem Usage"}, "mysqld")
            }
        );
        int timePeriodInSecs = Integer.parseInt(System.getProperty(SAMPLE_INTERVAL_SYS_PROPERTY, DEFAULT_SAMPLE_INTERVAL));
        timer.scheduleAtFixedRate(task, getProfilingStartTime(), timePeriodInSecs * 1000);
    }
}
