/*
 * $Id: //depot/prod/acquisition/main/tools/config/eclipse/.settings/org.eclipse.jdt.ui.prefs#1 $
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */
package com.onaro.management.serverresource;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Base64;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import sun.security.provider.X509Factory;

public class ServerCertificateUtils {
    private static final Logger logger = LoggerFactory.getLogger(ServerCertificateUtils.class);

    private static final String SERVER_KEYSTORE_FILE = "/jboss/server/onaro/cert/server.keystore";

    private static final String CLIENT_KEYSTORE_FILE = "/jboss/server/onaro/cert/client.keystore";

    private static final String KEY_ALIAS = "keyentry";
    private static final String CLIENT_KEY_ALAIS = "client_cert";

    private static final String DEFAULT_KEYSTORE_PASSWORD = "changeit";

    private static final String DEFAULT_KEYSTORE_TYPE = "JKS";

    private static final String NEW_LINE = System.getProperty("line.separator");

    /**
     * x509Certificate format strings
     */
    private static final String CERT_CN = "CN";

    public static void main(String[] args) {

//        Properties props = System.getProperties();
//        props.setProperty("onaro.home", "/opt/netapp/essentials");
        System.out.println(System.getProperty("onaro.home"));
    }

    public static  String getClientPrivateKey() {
        return getPrivateKey(CLIENT_KEY_ALAIS, getClientKeystorePath());
    }

    public static List<String> getClientCertificates() {
        return getCertificates(CLIENT_KEY_ALAIS, getClientKeystorePath());
    }

    /**
     * Utility method to retrieve all the server certificates from the certificate chain and convert the certificate to PEM format.
     *
     * @return List<String> - server certificates encoded in Base64.
     *         null   - error occurred during retrieving server certificate
     */
    public static List<String> getServerCertificates() {
        return getCertificates(KEY_ALIAS,getServerKeystorePath());
    }

    private static List<String> getCertificates(String alias, String storePath) {

        Certificate[] certs = getRawCertificate(alias, storePath);
        List<String> certList = new ArrayList<>();
        try {
            if (certs != null && certs.length > 0) {

                // CA signed certificate and might have intermediate levels, we need to get all CAs in the chain
                for (int i = 0; i < certs.length; i++) {
                    StringBuilder pem = new StringBuilder();
                    pem.append(X509Factory.BEGIN_CERT + NEW_LINE)
                        .append(Base64.getEncoder().encodeToString(certs[i].getEncoded()))
                        .append(NEW_LINE + X509Factory.END_CERT + NEW_LINE);
                    certList.add(pem.toString());
                }
            }
        } catch (CertificateEncodingException e) {
            logger.error("Unable to convert server certificate to PEM fromat; ", e);
        }

        return certList;
    }

    /**
     * Utility method to get CN from server certificate which is the first one in the chain.
     *
     * @return String - server certificate encoded in Base64.
     *         null   - error occurred during retrieving server certificate CN
     */
    public static String getCNFromServerCertificate() {
        return getCNFromCertificate(getRawCertificate(KEY_ALIAS,getServerKeystorePath()));
    }

    public static String getCNFromClientCertificate() {
        return getCNFromCertificate(getRawCertificate(CLIENT_KEY_ALAIS,getClientKeystorePath()));
    }

    private static String getCNFromCertificate(Certificate[] certs) {

        try {

            if (certs != null && certs.length > 0) {
                Certificate cert = certs[0];

                if (cert instanceof X509Certificate) {

                    X509Certificate x509Cert = (X509Certificate)cert;
                    String dn = x509Cert.getSubjectDN().getName();

                    int start = dn.indexOf(CERT_CN);
                    if (start == -1) {
                        return null;
                    }
                    String cn = dn.substring(start + 3);
                    return cn.contains(",") ? cn.substring(0, cn.indexOf(",")) : cn;
                }
            }
        } catch (Exception e) {
            logger.error("Unable to get CN from server certificate; ", e);
        }

        return null;
    }

    private static String getServerKeystorePath() {
        return System.getProperty("onaro.home", ".") + SERVER_KEYSTORE_FILE;
    }

    private static String getClientKeystorePath() {
        return System.getProperty("onaro.home", ".") + CLIENT_KEYSTORE_FILE;
    }

    /**
     * Get the certificate from keystore.
     *
     * @return the server certificate - ordered with the user's certificate first followed by zero or more certificate authorities
     *         null - error occurred when retrieving server certificate
     */
    private static Certificate[] getRawCertificate(String alias, String storePath) {
        try {
            KeyStore keystore = KeyStore.getInstance(DEFAULT_KEYSTORE_TYPE);

            keystore.load(new FileInputStream(storePath), DEFAULT_KEYSTORE_PASSWORD.toCharArray());

            // the chain - ordered with the user's certificate first followed by zero or more certificate authorities
            return keystore.getCertificateChain(alias);
        } catch (KeyStoreException | NoSuchAlgorithmException | CertificateException | IOException e) {
            logger.error("Unable to retrieve server certificate; ", e);
            return null;
        }
    }


    public static String getPrivateKey(String alias, String storePath) {
        try {
            KeyStore keystore = KeyStore.getInstance(DEFAULT_KEYSTORE_TYPE);

            File f = new File(storePath);
            if (f.exists() && f.canRead()) {
                keystore.load(new FileInputStream(storePath), DEFAULT_KEYSTORE_PASSWORD.toCharArray());
            } else {
                return null;
            }

            Key key = keystore.getKey(alias, DEFAULT_KEYSTORE_PASSWORD.toCharArray());

            if (key != null) {
                StringBuilder pem = new StringBuilder();
                pem.append( "-----BEGIN PRIVATE KEY-----" + NEW_LINE);
                pem.append(Base64.getEncoder().encodeToString(key.getEncoded()));
                pem.append(NEW_LINE);
                pem.append( "-----END PRIVATE KEY-----" + NEW_LINE);

                return pem.toString();
            } else {
                return null;
            }
        } catch (KeyStoreException | NoSuchAlgorithmException | CertificateException | IOException | UnrecoverableKeyException e) {
            logger.error("Unable to retrieve server certificate; ", e);
            return null;
        }
    }
}
