/*
 * Copyright (c) 2010 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.acquisition.sessions;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.plugin.remote.PlatformPluginRemote;
import com.onaro.sanscreen.server.dataobject.acquisition.*;
import com.onaro.sanscreen.server.dataobject.acquisition.Package;
import com.onaro.sanscreen.server.sessions.SummarySessionEJB;
import com.onaro.sanscreen.server.tables.acquisition.*;
import com.onaro.sanscreen.util.Schema;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.*;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Properties;
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRED)
@Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
public class AcquisitionFacadeForClientSessionEJB implements AcquisitionFacadeForClientSession {
    private static transient final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger("com.onaro.sanscreen.acquisition.sessions");

    @EJB
    private AcquisitionServerSession acqServerSession;

    @EJB
    private PlatformPluginRemote platformPluginRemote;

    public List<DataSourceTypeVendorModel> getAllDataSourceVendorModels() throws AcquisitionUnitException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            return new DataSourceTypeVendorModelTable(connection).findAll();
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to get all VendorModels", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    public DataSourceConfiguration getNewDataSourceByType(final DataSourceTypeVendorModel vendorModel) throws AcquisitionUnitException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            // get attributes from the database ORDERED BY NAME
            final List<DsTypeAttribute> attrs = new DataSourceTypeAttrsTable(connection).findByDsTypeIdOrderByName(vendorModel.getDsTypeId());

            // only ds_type and ds_type_attr tables used
            return new DataSourceConfigurationImpl(vendorModel.getDsTypeId(),
                    vendorModel.getVendorDescription(),
                    vendorModel.getModelDescription(),
                    attrs);
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to get ds defaults", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    public DataSourceConfiguration getDataSourceById(final long ds_id) throws AcquisitionUnitException {

        // Create a new DataSourceConfigurationImpl object and populate it with attributes from appropriate tables
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();

            // Get "ds" table entry for the datasource
            final DataSourceNew dataSourceNew = new DataSourceTableNew(connection).getById(ds_id);

            // get "ds_type_atts" table entries for the datasource
            final List<DsTypeAttribute> attrs = new DataSourceTypeAttrsTable(connection).findByDsTypeIdOrderByName(dataSourceNew.getDsTypeId());

            // create new DataSourceConfigurationImpl object
            final DataSourceConfigurationImpl dataSourceConfig = new DataSourceConfigurationImpl(dataSourceNew.getDsTypeId(),
                    dataSourceNew.getVendor(),
                    dataSourceNew.getModel(),
                    attrs);

            // set the top level fields on dataSourceConfig
            dataSourceConfig.setId(dataSourceNew.getId());
            dataSourceConfig.setName(dataSourceNew.getName());
            dataSourceConfig.setAuId(dataSourceNew.getAuId());

            // get "ds_attrs" table entries for the datasource into a properties map
            final Properties propertiesFromDB = new DataSourceAttrsTable(connection).findByDsId(ds_id);

            // domain.id is stored by old implementations as a ds_attr, but is not needed
            propertiesFromDB.remove("domain.id");

            // now populate the datasource attributes
            dataSourceConfig.createOrUpdateAttributesFromDBProperties(propertiesFromDB);

            return dataSourceConfig;

        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to get ds defaults", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }

    }

    public List<Package> getLicensedPackages(final long dataSourceTypeId) throws AcquisitionUnitException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();

            final List<Package> result = new ArrayList<Package>();
            final List<DsTypeAttribute> attrs = new DataSourceTypeAttrsTable(connection).findByDsTypeIdOrderByName(dataSourceTypeId);

            DataSourceConfiguration dsConfig = new DataSourceConfigurationImpl(dataSourceTypeId, null, null, attrs);
            List<Package> packages = dsConfig.getPackages();
            for (Package pckg : packages) {
                Package.PackageName packageName = pckg.getName();
                switch (packageName) {
                    case NETAPP_FOUNDATION:  //we will always (and only) enable this if we are running essentials
                                             //No license requirement!
                                             //TODO - Revisit before release?
                        if (platformPluginRemote.isPluginRegisteredForSchema(Schema.NETAPP)) {
                            result.add(pckg);
                        }
                        break;
                    case FOUNDATION:
                        break;

                    case PERFORMANCE:
                        break;


                    case NETAPP_STORAGE_PERFORMANCE:   //we will always (and only) enable this if we are running essentials
                                                       //No license requirement!
                                                       //TODO - Revisit before release?
                        if (platformPluginRemote.isPluginRegisteredForSchema(Schema.NETAPP)) {
                            result.add(pckg);
                        }
                        break;
                    default:
                        throw new AcquisitionUnitException("Cannot get features for Package " + packageName.toString() + ". Not a vaild package.");
                }
            }

            return result;
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to get ds defaults", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    public void addNewDataSource(final DataSourceConfiguration newDataSource) throws AcquisitionUnitException {
        insertOrUpdateDataSource(newDataSource, true);
    }

    public void updateDataSource(final DataSourceConfiguration updatedDataSource) throws AcquisitionUnitException {
        insertOrUpdateDataSource(updatedDataSource, false);
    }

    private void insertOrUpdateDataSource(final DataSourceConfiguration dataSourceConfig, final boolean insert) throws AcquisitionUnitException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();

            // make sure AU exists for this ds
            final AcquisitionUnitTableNew acquisitionUnitTable = new AcquisitionUnitTableNew(connection);
            final AcquisitionUnitNew au = acquisitionUnitTable.get(dataSourceConfig.getAuId());
            if (au == null) {
                logger.error("Failed to find acquisitionUnitName");
                throw new AcquisitionUnitException("Failed to find acquisitionUnitName");
            }

            // create a properties object to insert into DS_ATTRS.
            // These come from the DataSourceAttributes under Packages in DataSourceConfiguration
            final Properties propertiesForDB = createPropertiesForDB(dataSourceConfig);

            final DataSourceTableNew dataSourceTable = new DataSourceTableNew(connection);
            final DataSourceAttrsTable dataSourceAttrsTable = new DataSourceAttrsTable(connection);

            // populate DataSourceNew fields (EXCEPT for the Properties map) from the DataSourceConfig fields
            final DataSourceNew dataSource = convertDataSourceConfigToDataSourceNew(dataSourceConfig);


            if (insert) {
                // a new datasource
                auditDataSourceAdd(dataSourceConfig, connection, propertiesForDB);

                // TODO: this shoud be done as DB constraint !!!
                // check that the datasource name is not already taken
                checkDataSourceNameIsUnique(dataSourceConfig, dataSourceTable);

                // add the new data source to the database
                dataSourceTable.insert(dataSource);
                dataSourceAttrsTable.insert(dataSource.getId(), propertiesForDB);
            } else {
                // an update to a datasource and its attributes
                // audit an acq unit change
                auditDataSourceAcqUnitChange(dataSourceConfig, connection, au, dataSource);

                // audit any property changes
                auditDataSourcePropertyChanges(dataSourceConfig, connection, dataSource);

                // persist the updated datasource and its attributes in the database
                updateDataSourceInDb(propertiesForDB, dataSourceTable, dataSourceAttrsTable, dataSource);

                acqServerSession.notifyDsConfigurationChange(dataSource.getId());
            }

            // Update time tracker of the latest changes
            updateLastUpdateTime();
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to add a new Data Source to the database.", e);
        }
        catch (AcquisitionServerException e) {
            throw new AcquisitionUnitException("Failed to add a new Data Source to the database.", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    private void updateDataSourceInDb(final Properties propertiesForDB, final DataSourceTableNew dataSourceTable, final DataSourceAttrsTable dataSourceAttrsTable,
                                      final DataSourceNew dataSource) throws TableException {
        dataSourceTable.updateWithoutManualAndActive(dataSource);
        // We need to remove all the list properties because if some middle rows were removed by client
        // we need to remove them from DB as well. Easier to remove all list properties and re-insert them.
        dataSourceAttrsTable.removeListProperties(dataSource.getId());
        // update does insert if needed
        dataSourceAttrsTable.update(dataSource.getId(), propertiesForDB);
    }

    private void auditDataSourcePropertyChanges(final DataSourceConfiguration editedDataSourceConfig, final Connection connection,
                                                final DataSourceNew dataSource) throws TableException, AcquisitionUnitException {

        // used for audit logging
        final String userName = currentContext.getCallerPrincipal().getName();

        // create a properties list from edited attributes
        final Properties editedProperties = createPropertiesForDB(editedDataSourceConfig);

        // create a properties list from current DB attributes
        final Properties currentDbProperties = new DataSourceAttrsTable(connection).findByDsId(editedDataSourceConfig.getId());

        // run through each property in the edited set and audit any changes
        for (final Object editedPropertyNameObject : editedProperties.keySet()) {
            final String editedPropertyName = (String) editedPropertyNameObject;
            final String editedPropertyValue = (String) editedProperties.get(editedPropertyName);
            final String currentDbPropertyValue = (String) currentDbProperties.get(editedPropertyName);

            final String[] nameParts = editedPropertyName.split("\\.");
            if (nameParts.length > 2 && nameParts[1].equals("transport"))
                continue;

            if (currentDbPropertyValue != null) {
                auditProperties(connection, dataSource, userName, editedProperties, editedPropertyName, editedPropertyValue, currentDbPropertyValue, nameParts);
            } else {
                // this is a new property, so log this
                if (logger.isDebugEnabled()) {
                    logger.debug("AcquisitionFacadeSessionEJB.updateProp() - new prop was added : " +
                            "name=" + editedPropertyName + ";" +
                            "originalValue=" + currentDbPropertyValue + ";" +
                            "new value=" + editedPropertyValue);
                }
            }

        }
    }

    private void auditProperties(
            final Connection connection,
            final DataSourceNew dataSource,
            final String userName,
            final Properties editedProperties,
            final String editedPropertyName,
            final String editedPropertyValue,
            final String currentDbPropertyValue,
            final String[] nameParts) {
        if (!editedPropertyValue.equals(currentDbPropertyValue)) {
            // the property has changed, so we audit appropriately
            // HEADSUP There are certain properties we want to create special log messages for.
            // TODO check if the attribute is marked "encrypt" here
//            if (nameParts.length > 1 && nameParts[1].contains("password")) {
//                AuditWriter.addAudit(connection, AuditType.DATASOURCE_UPDATED_PASSWORD, userName, dataSource.getName());
//            } else if ("performance.enabled".equals(editedPropertyName)) {
//                if ("true".equals(editedProperties.get(editedPropertyName))) {
//                    AuditWriter.addAudit(connection, AuditType.DATASOURCE_UPDATED_PERFORMANCE_ENABLED, userName, dataSource.getName());
//                } else {
//                    AuditWriter.addAudit(connection, AuditType.DATASOURCE_UPDATED_PERFORMANCE_DISABLED, userName, dataSource.getName());
//                }
//            } else if ("performance.poll".equals(editedPropertyName)) {
//                AuditWriter.addAudit(connection, AuditType.DATASOURCE_UPDATED_PERFORMANCE_POLLING_CHANGED, userName, dataSource.getName(), currentDbPropertyValue, editedPropertyValue);
//            } else if ("foundation.poll".equals(editedPropertyName)) {
//                AuditWriter.addAudit(connection, AuditType.DATASOURCE_UPDATED_FOUNDATION_POLLING_CHANGED, userName, dataSource.getName(), currentDbPropertyValue, editedPropertyValue);
//            } else {
//                // Use the generic
//                AuditWriter.addAudit(connection, AuditType.DATASOURCE_UPDATED, userName, dataSource.getName(), editedPropertyName, currentDbPropertyValue, editedPropertyValue);
//            }
        } else {
            // no change, so we do not audit
        }
    }

    private void checkDataSourceNameIsUnique(final DataSourceConfiguration dataSourceConfig, final DataSourceTableNew dataSourceTable) throws TableException, AcquisitionUnitException {
        final String dsName = dataSourceConfig.getName();
        if (dataSourceTable.findByName(dsName).size() > 0) {
            throw new AcquisitionUnitException("Name " + dsName + " already used by another datasource");
        }
    }

    private void auditDataSourceAcqUnitChange(
            final DataSourceConfiguration dataSourceConfig,
            final Connection connection,
            final AcquisitionUnitNew au,
            final DataSourceNew dataSource) {
        final String userName = currentContext.getCallerPrincipal().getName();
        // Checks if acq. unit name has changed
        if (au.getId() != dataSourceConfig.getAuId()) {
            dataSource.setAuId(au.getId());
//            AuditWriter.addAudit(connection, AuditType.DATASOURCE_UPDATED_ACQ_UNIT, userName, dataSource.getName(), au.getName(), au.getName());
        }
    }

    private void auditDataSourceAdd(
            final DataSourceConfiguration dataSourceConfig,
            final Connection connection,
            final Properties propertiesForDB) {
        final String userName = currentContext.getCallerPrincipal().getName();

//        final AuditType auditType;
//        if ("true".equals(propertiesForDB.getProperty("performance.enabled", "false"))) {
//            auditType = AuditType.DATASOURCE_ADDED_PERFORMANCE_ENABLED;
//        } else {
//            auditType = AuditType.DATASOURCE_ADDED;
//        }
//
//        AuditWriter.addAudit(connection,
//                auditType,
//                userName,
//                dataSourceConfig.getName(),
//                dataSourceConfig.getVendor(),
//                dataSourceConfig.getModel(),
//                propertiesForDB.getProperty("foundation.ip"));
    }

    private DataSourceNew convertDataSourceConfigToDataSourceNew(final DataSourceConfiguration dsc) {
        return new DataSourceNew(dsc.getAuId(), dsc.getDsTypeId(), dsc.getId(), dsc.getName(), null, dsc.getVendor(), dsc.getModel(),
                // the last line of fields is not needed for a new ds insert
                "", "", 0, 1, false);
    }

    /**
     * Used when creating/updating a datasource to compile a list of flat properties from all packages
     *
     * @param dataSourceConfiguration
     * @return list of flat properties ready to be placed in ds_atts table
     */
    private Properties createPropertiesForDB(final DataSourceConfiguration dataSourceConfiguration) {
        final Properties properties = new Properties();
        for (final Package pkg : dataSourceConfiguration.getPackages()) {
            final String packageName = pkg.getName().toString();
            for (final DataSourceAttribute attribute : pkg.getAttributes()) {
                if (attribute.getType().equals(DataSourceAttribute.AttributeType.LIST)) {
                    createListProperties(properties, packageName, attribute);
                } else {
                    createStandardProperty(properties, packageName, attribute);
                }
            }
        }
        return properties;
    }

    /**
     * Converts DataSourceAttribute objects to properties to store in the database
     *
     * @param properties  The Properties map holding the properties to add to the database table DS_ATTRS
     * @param packageName
     * @param attribute   The DataSourceAttribute object stored in the Package attribute list
     */
    private void createStandardProperty(final Properties properties, final String packageName, final DataSourceAttribute attribute) {
        final DataSourceSimpleAttribute simpleAttr = ((DataSourceSimpleAttribute) attribute);
        if (simpleAttr.getValueAsString() != null) {
            final String attributeName = simpleAttr.getName();
            final String propertyName = packageName + "." + attributeName;

            String propertyValueString = simpleAttr.prepareValueForDatabase();

            // foundation.poll and nas.poll need to be converted
            // from minutes to seconds and stored as integers
            if (attributeName.equals("poll") && (packageName.equals("foundation") || packageName.equals("nas") || packageName.equals("netappfoundation"))) {
                final Float valueInSeconds = (Float) simpleAttr.getValue() * 60;
                final Integer integerValueInSeconds = valueInSeconds.intValue();
                propertyValueString = integerValueInSeconds.toString();
            }

            properties.put(propertyName, propertyValueString);
        }
    }

    private void createListProperties(final Properties properties, final String packageName,
                                      final DataSourceAttribute attribute) {
        final DataSourceListAttribute listAttr = (DataSourceListAttribute) attribute;
        // Dump the list to properties
        final String listName = listAttr.getName();
        final String listPropertyPrefix = packageName + ".LIST." + listName;
        // create the properties representing the columns
        for (final DataSourceSimpleAttribute column : listAttr.getColumns()) {
            // columns are represented as the 0th row
            final String propertyKey = listPropertyPrefix + "." + "0" + "." + column.getName();
            properties.put(propertyKey, column.getDefaultValueAsString());
        }

        int rowNumber = 0;
        for (final DataSourceListAttribute.Row row : listAttr.getRows()) {
            rowNumber++;
            for (final DataSourceSimpleAttribute rowAttribute : row.getRowValues()) {
                // make a property
                if (rowAttribute.getValue() != null) {
                    final String propertyKey = listPropertyPrefix + "." + rowNumber + "." + rowAttribute.getName();
                    properties.put(propertyKey, rowAttribute.prepareValueForDatabase());
                }
            }
        }
    }


    @RolesAllowed({"guest", "admin", "acquisition"})
    public Collection<AcquisitionUnitNew> getAllAcquisitionUnit() throws AcquisitionUnitException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            return new AcquisitionUnitTableNew(connection).findAll();
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to get Acquisition Units", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    private void updateLastUpdateTime() {
        SummarySessionEJB.updateTimeDatasources = System.currentTimeMillis();
    }


    //////////////////////////////////////////
    @Resource
    private SessionContext currentContext;

}
