/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.acquisition.sessions;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.ejb.Remote;

import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionAuditMessage;
import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionUnit;
import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionUnitNew;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSource;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceAudit;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceDefaults;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceNew;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceStatus;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceVendorModel;
import com.onaro.sanscreen.server.tables.acquisition.Originator;
// import com.netapp.acquisition.aquisitionserverside.interfaces.data.Originator;
import com.onaro.sanscreen.types.ByteArrayWrapper;
import com.onaro.sanscreen.util.Schema;

/**
 * @author carl.chan-aldebol
 */
@Remote
public interface AcquisitionFacadeSession {
    String JNDI_NAME = "ejb:mega-app/server-core//AcquisitionFacadeSessionEJB!com.onaro.sanscreen.acquisition.sessions.AcquisitionFacadeSession";

    /**
     * Calls into the database to retrieve an OriginatorID from the server
     * 1) If datasourceID is null, an exception will be thrown in the server
     * 2) If datasourceID is non-null, and devicekey is null, and a datasource row exists in the database, the
     * originatorID will be returned. If none exists, it will be created and returned
     * 4) If datasourceID is non-null and devicekey is non-null, an existing datasourceID will be returned if present,
     * otherwise a new originatorID will be returned, and entered into the database.
     *
     * @param ds_objid
     * @param deviceKey
     *
     * @return The ID allocated for this device key
     *
     * @throws AcquisitionUnitException
     */
    public long getOriginatorID(final long ds_objid, final String deviceKey, final String originatorName) throws AcquisitionUnitException;

    /**
     * Calls into the database to retrieve an OriginatorID from the server
     * 1) If datasourceID is null, an exception will be thrown in the server
     * 2) If datasourceID is non-null, and devicekey is null, and a datasource row exists in the database, the
     * originatorID will be returned. If none exists, it will be created and returned
     * 4) If datasourceID is non-null and devicekey is non-null, an existing datasourceID will be returned if present,
     * otherwise a new originatorID will be returned, and entered into the database.
     *
     * If performDuplicateDatasourceCheck flag is set to true, then a check is performed to detect duplicate datasources of same type and same device key.
     *
     * @param ds_objid
     * @param deviceKey
     * @param performDuplicateDatasourceCheck
     *
     *
     * @return The ID allocated for this device key
     *
     * @throws AcquisitionUnitException
     */
    public long getOriginatorID(final long ds_objid, final String deviceKey, final String originatorName, final boolean performDuplicateDatasourceCheck) throws AcquisitionUnitException;

    public void deleteOriginatorID(long datasourceID, final String deviceKey) throws AcquisitionUnitException, RemoteException;

    public void deleteOriginatorID(Set<Long> originatorIDs) throws AcquisitionUnitException, RemoteException;

    public void stopReportOriginatorID(Set<Long> originatorIDs) throws AcquisitionUnitException, RemoteException;

    /**
     * Add DS without providing asEmsRestDest which is default to false, it means the caller doesn't register itself as a EMS REST push destination.
     *
     * @param auName
     * @param name
     * @param vendor
     * @param model
     * @param datasourceTypeId
     * @param manual
     * @param isActive
     * @param attrs
     * @return
     * @throws AcquisitionUnitException
     */
    public Long addDS(final String auName,
                      final String name,
                      final String vendor,
                      final String model,
                      final Integer datasourceTypeId,
                      final Integer manual,
                      final Boolean isActive,
                      final Properties attrs) throws AcquisitionUnitException;

    /**
     * <p>Enter a synchronous call waiting for the given Datasource ID to "start".   Start here means it was
     * recognized by the active AU and is scheduled and ready to poll.   Returns true if started after the call.
     * Maximum wait time is 5 seconds.  </p>
     *
     * <p>A false return here could mean a number of things: the AU is not active, the DataSource or AU no longer
     * exists, or we simply timed out on the wait.   </p>
     *
     * @param datasourceId   the ID of the DataSource to check for start.
     * @return   true if the DataSource is detected as "started' (active).   False if not.
     * @throws AcquisitionUnitException infrastructure issues looking data up.
     */
    public boolean waitForDataSourceStart (final Long datasourceId) throws AcquisitionUnitException;

    /**
     * retrun the audit messages bound with the auditID
     *
     * @param auditId
     *
     * @return
     */
    public List<AcquisitionAuditMessage> getAuditMessages(final Long auditId) throws AcquisitionUnitException;

    public void updateManageURL(Long datasourceId, String manageURL) throws AcquisitionUnitException;

    public DataSourceNew getDsById(final long ds_objid) throws AcquisitionUnitException;

    public void updateDsIsActiveForAu(AcquisitionUnitNew au, boolean active) throws AcquisitionUnitException;

    /**
     * This method is used by the data source to report that it finished an acquisition cycle and it has nothing new to
     * report. The reason for such a method is to act as a ping receiver between the acquisition and the server, with
     * out this method jboss disconnect the connection between them.
     */
    public void noChangeSinceLastReport(AcquisitionUnit au) throws AcquisitionUnitException;

    public Boolean update(String xml, Schema schema) throws AcquisitionUnitException;

    public Boolean updateCompressed(byte[] compressedXML, Schema schema) throws AcquisitionUnitException;

    public void restartDS(final long ds_objid) throws AcquisitionUnitException, RemoteException;

    public void updateManual(final long ds_objid, final int manual) throws AcquisitionUnitException, RemoteException;

    /**
     * Update datasource using a Property object
     *
     * @param auName    Acquisition unit name
     * @param dsObjid   Datasource object ID
     * @param attrs     Updated properties. Should contain all properties including ones not changes.
     * @throws AcquisitionUnitException exception
     */
    public void updateProp(final String auName, final long dsObjid, final Properties attrs) throws AcquisitionUnitException;

    /**
     * Update datasource using an XML string
     *
     * @param auName    Acquisition unit name
     * @param ds_objid  Datasource object ID
     * @param prop      Updated properties in a more hostile format. Should contain all properties including ones not changes.
     * @throws AcquisitionUnitException exception
     */
    public void updateProp(final String auName, final long ds_objid, final String prop) throws AcquisitionUnitException;

    public void stopReport(final long[] ids) throws AcquisitionUnitException, RemoteException;

    public void removeDSs(final long[] ids) throws AcquisitionUnitException, RemoteException;

    public void removeAllDataSource() throws AcquisitionUnitException, RemoteException;

    /**
     * <p>Remove a DataSource from the system.  This handles all aspects of the delete, including stopping the
     * DataSource, stopping and removing the originator reports, remove the reference in the acquisition schema.
     * By default the caller is not an EMS REST push destination.</p>
     *
     * <p>WARNING.  This will cause all data associated with the objects discovered by this DataSource to be
     * permanently removed.</p>
     *
     * @param id  object ID for the DataSources to be deleted.
     *
     * @throws AcquisitionUnitException if delete fails for some reason.
     * @throws RemoteException          for communication issues.
     */
    public void removeDataSource(final long id) throws AcquisitionUnitException, RemoteException;

    public void removeDsAudit(long[] originIds) throws AcquisitionUnitException, RemoteException;

    /**
     * Delete all the data source audit records related to origin Id, save the last two... <p/> This method is called
     * when right clicking on a data source and choosing "clearing history"
     */
    void removeDsAudit(long originId) throws AcquisitionUnitException;

    /**
     * Retrieves all the audit records related to an acquisition unit. For each data source of this acq unit, delete all
     * the audit records, save the last two... <p/> This method is called when right clicking on an acquisition unit and
     * choosing "clearing data source history"
     */
    public void removeDsAuditForAu(final String acquisitionUnitName) throws AcquisitionUnitException;

    public void updateDsName(long id, String newName) throws AcquisitionUnitException, RemoteException;

    public DataSource getDataSource(String ip, int deviceTypeId, long time) throws AcquisitionUnitException, RemoteException;

    public long getLastUpdate() throws AcquisitionUnitException, RemoteException;

    public DataSource[] getDsForAu(String acqUnitName) throws AcquisitionUnitException;

    public List<DataSourceNew> getDsForAuNew(String acqUnitName) throws AcquisitionUnitException;

    public DataSource[] getAllDs() throws AcquisitionUnitException;

    public List<DataSourceNew> getAllDsNew() throws AcquisitionUnitException;

    public List<DataSourceNew> getAllDsNew(boolean includeAttributes) throws AcquisitionUnitException;

    public DataSource[] getAllDsByTime(long date) throws AcquisitionUnitException;

    public DataSource[] getDsByIds(long[] ids) throws AcquisitionUnitException;

    public String getOriginatorNameOriginatorId(Long originId) throws AcquisitionUnitException;

    public String getDataSourceNameByDatasourceId(Long dsID) throws AcquisitionUnitException, RemoteException;

    public DataSource getOpenById(long id) throws VendorModelNotFoundException, AcquisitionUnitException, RemoteException;

    public DataSource getById(long id, long date) throws AcquisitionUnitException, RemoteException;

    public DataSourceAudit[] getAllDsHistoryById(long dsID) throws AcquisitionUnitException, RemoteException;

    public AcquisitionUnitNew getAcquisitionUnit(final String acquisitionUnitName) throws AcquisitionUnitException, RemoteException;

    public Collection<AcquisitionUnit> getAllAcquisitionUnit() throws AcquisitionUnitException, RemoteException;

    public AcquisitionUnit[] getAcquisitionUnitsByNames(String[] names, int status, boolean remoteOnly) throws AcquisitionUnitException, RemoteException;

    public void removeAcqUnitsByNames(String[] names) throws AcquisitionUnitException, RemoteException;

    public Collection<DataSourceVendorModel> geAllDatasourceTypeVendorModel() throws AcquisitionUnitException, RemoteException;

    public Map<Long, DataSourceNew> getOriginIdToDatasource() throws AcquisitionUnitException, RemoteException;

    public Map<Long, AcquisitionUnitNew> getAcquisitionUnitsByID() throws AcquisitionUnitException, RemoteException;

    /**
     * Returns an array of {@link DataSourceStatus} instances representing the current data source's status.
     * <p/>
     * The implementation of this method is encapsulated in {@link DataSourceStatusQueryImpl}, which the needed logic
     * has been coded to retrive Data source's status, based on document named:
     * "Data-source alerts Software Requirement Specification"
     */
    public DataSourceStatus[] getDataSourceStatusQuery(int timeOutInHours, int timeOutForSinglePollInHours, int numOfAttemps) throws AcquisitionUnitException;

    /**
     * Returns an array of {@link DataSourceStatus} instances representing the current data source's status.
     * <p/>
     * This method will take into account default settings used in the logic to determine if a specific data source
     * is in a Success or Failed state.
     * <p/>
     * The implementation of this method is encapsulated in {@link DataSourceStatusQueryImpl}
     */
    public DataSourceStatus[] getDataSourceStatusQuery() throws AcquisitionUnitException;

    public DataSourceVendorModel[] findAllDataSourceVendorModels() throws AcquisitionUnitException;

    public DataSourceVendorModel[] findAllVmByDataSourceType(long dataSourceType) throws AcquisitionUnitException;

    public DataSourceVendorModel[] findAllVmByVendor(final String vendor) throws AcquisitionUnitException;

    public DataSourceVendorModel[] findVmByDataSourceTypeAndVendor(long dataSourceType, String vendor) throws AcquisitionUnitException;

    public DataSourceVendorModel findVmByVendorModel(final String vendor, final String model) throws AcquisitionUnitException;

    public String getDsTypeProperty(long dataSourceType, final String name) throws AcquisitionUnitException;

    public Collection<DataSourceDefaults> getDsDefaults() throws AcquisitionUnitException;

    public ByteArrayWrapper getDsPolicy(long dataSourceType) throws AcquisitionUnitException;

    public String getDsPolicyClassName(long dataSourceType) throws AcquisitionUnitException;

    public String getDsDefaultName(long dataSourceType) throws AcquisitionUnitException;

    public String getDefaultDescription(long dataSourceType) throws AcquisitionUnitException;

    public Map<Long, Originator> getOriginatorIdToOriginator() throws AcquisitionUnitException;

    public Map<Long, com.netapp.acquisition.aquisitionserverside.interfaces.data.Originator> getOriginatorIdToAcqOriginator() throws AcquisitionUnitException;

    public void retireExpiredOriginators() throws AcquisitionUnitException;

    public void retireUnreportedOriginators(final Set<Long> activeOriginatorIDs, final Long dataSourceId) throws AcquisitionUnitException, RemoteException;

    public void updateOriginatorLastAccessTime(final long originatorId) throws AcquisitionUnitException;

    void removeDsAuditOnly(long[] dsIDs) throws AcquisitionUnitException;

    void updateDS(long[] dsIDs) throws AcquisitionUnitException;

    public long getLatestPerformancePollTimestamp(final long dataSourceId) throws AcquisitionUnitException;

}
