/*
 * $Id: //depot/prod/acquisition/main/tools/config/eclipse/.settings/org.eclipse.jdt.ui.prefs#1 $
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */
package com.onaro.sanscreen.acquisition.sessions;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.stream.Collectors;

import com.onaro.management.serverresource.ServerCertificateUtils;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceNew;
import netapp.manage.NaElement;
import netapp.manage.NaServer;

import com.netapp.oci.netapp.client.interfaces.data.EmsZapiManagerException;
import com.netapp.oci.netapp.ems.EmsZapiManager;
import com.netapp.oci.netapp.ems.EmsZapiManager.ClientKeyInformation;
import com.netapp.oci.netapp.ems.NetappEmsUtil;
import com.netapp.oci.netapp.interfaces.data.NetAppEmsConfig;

public final class AcquisitionFacadeUtil {

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(AcquisitionFacadeUtil.class);
    public static final String LONG_BOARD = "9.0";
    private static final String FILTER_RULE_INCLUDE = "include";
    private static final String CERT_TYPE = "server-ca";

    private static ClientKeyInformation registerClientKey(NaServer zapiConnection) throws EmsZapiManagerException {
        try {
            // Register client key before creating event destination that way we can register the ca and ca serial with the destination
            String clientKey = ServerCertificateUtils.getClientPrivateKey();
            // if we are unable to get a client key we skip registering it to the cluster
            if(clientKey != null) {
                String clientCert = ServerCertificateUtils.getClientCertificates().get(0);
                NaElement result = EmsZapiManager.INSTANCE.installClientKey(clientKey, clientCert, zapiConnection);
                // fattire(9.3) does not support return of ca and serial when we install the key
                // if the result does not contain the ca we have to query the cluster for it
                if(result == null || result.getChildByName("ca") == null) {
                    return getClientKeyInfo(zapiConnection);
                } else {
                    return new ClientKeyInformation(result.getChildContent("serial"), result.getChildContent("ca"));
                }
            } else {
                return new ClientKeyInformation();
            }
        } catch (EmsZapiManagerException ex) {
            // If it is a dupulicate entry retrieve the key
            if (ex.getErrno() == EmsZapiManager.EDUPLICATEENTRY) {
                return getClientKeyInfo(zapiConnection);
            }
            else {
                throw ex;
            }
        } catch (Exception ex) {
            throw ex;
        }
    }

    private static ClientKeyInformation getClientKeyInfo(NaServer zapiConnection) throws EmsZapiManagerException {
        String cn = ServerCertificateUtils.getCNFromClientCertificate();
        NaElement certificateResult = EmsZapiManager.INSTANCE.showSecurityCertificates(cn,EmsZapiManager.CLIENT_TYPE_KEY,zapiConnection);
        // Parse xml element down to the individual certificate
        NaElement certificateInfo = certificateResult.getChildByName("attributes-list").getChildByName("certificate-info");

        return new ClientKeyInformation(certificateInfo.getChildContent("serial-number"), certificateInfo.getChildContent("certificate-authority"));
    }

    /**
     * For EMS registration---install security-certificate, createFilterRule, createDestination, createFilter
     */
    public static void registerEms(String hostname, Properties attrs, List<NetAppEmsConfig> emsConfigs) throws EmsZapiManagerException {

        NaServer zapiConnection = null;

        try {
            zapiConnection = EmsZapiManager.INSTANCE.createZAPIConnection(attrs);

            // Don't register if Ontap version is lower than 9.0
            String clusterVersion = getDSClusterVersion(zapiConnection);
            if (NetappEmsUtil.compareOntapVersion(clusterVersion, LONG_BOARD) < 0) {
                return;
            }

            String destinationName = EmsZapiManager.INSTANCE.getDestinationName(hostname);
            String destinationURL = EmsZapiManager.INSTANCE.getDestinationURL(hostname);

            // Please keep the order of making zapi calls
            ClientKeyInformation clientKey = registerClientKey(zapiConnection);

            try {
                // run this zapi first since if the destination is already created, it would fail which means either it is already registered or is not able to register
                EmsZapiManager.INSTANCE.createEventNotificationDestination(destinationName, destinationURL, clientKey.getClientCertificateKeyAuthority(), clientKey.getClientCertificateKeySerial(), zapiConnection);
            } catch (EmsZapiManagerException ex) {
                if (ex.getErrno() != EmsZapiManager.EDUPLICATEENTRY) {
                    throw ex;
                } else {
                    // If already created modify with updated destination url and certificate values
                    try {
                        EmsZapiManager.INSTANCE.modifyEventNotificationDestination(destinationName, destinationURL, clientKey.getClientCertificateKeyAuthority(), clientKey.getClientCertificateKeySerial(), zapiConnection);
                    } catch (EmsZapiManagerException modifyException) {
                        throw modifyException;
                    }
                }
            }
            //createEventFilter
            String filterName = EmsZapiManager.INSTANCE.getFilterName(hostname);
            try {
                EmsZapiManager.INSTANCE.createEventFilter(filterName, zapiConnection);
            } catch (EmsZapiManagerException ex) {
                if (ex.getErrno() != EmsZapiManager.EDUPLICATEENTRY) {
                    //delete previously created eventNotificationDestination
                    EmsZapiManager.INSTANCE.deleteEventNotificationDestination(destinationName, destinationURL, zapiConnection);
                    throw ex;
                }
            }

            // Get filter rules that can be applied to this cluster
            Set<String> filterRules = getFilterRulesForAddDs(emsConfigs, clusterVersion);

            //createEventFilterRules
            for (String filterRule : filterRules) {
                try {
                    EmsZapiManager.INSTANCE.addFilterRule(filterName, filterRule, FILTER_RULE_INCLUDE, zapiConnection);
                } catch (EmsZapiManagerException ex) {
                    if (ex.getErrno() != EmsZapiManager.EDUPLICATEENTRY && ex.getErrno() != EmsZapiManager.E_EMS_NO_EVENT_MATCHED_THE_RULE) {
                        //delete previously 2 calls
                        EmsZapiManager.INSTANCE.deleteEventNotificationDestination(destinationName, destinationURL, zapiConnection);
                        EmsZapiManager.INSTANCE.deleteEventFilter(filterName, zapiConnection);
                        throw ex;
                    }
                }
            }

            List<String> destinations = new ArrayList<>();
            destinations.add(destinationName);
            try {
                EmsZapiManager.INSTANCE.createEventNotification(filterName, destinations, zapiConnection);
            } catch (EmsZapiManagerException ex) {
                if (ex.getErrno() != EmsZapiManager.EDUPLICATEENTRY) {
                    //delete previously 2 calls(filterRules will be deleted when filter is deleted)
                    EmsZapiManager.INSTANCE.deleteEventNotificationDestination(destinationName, destinationURL, zapiConnection);
                    EmsZapiManager.INSTANCE.deleteEventFilter(filterName, zapiConnection);
                    throw ex;
                }
            }

            try {
                List<String> certificates = ServerCertificateUtils.getServerCertificates();
                EmsZapiManager.INSTANCE.installSecurityCertificate(destinationName, CERT_TYPE, certificates, zapiConnection);
            } catch (EmsZapiManagerException ex) {
                if (ex.getErrno() != EmsZapiManager.EDUPLICATEENTRY) {
                    //delete previously 3 calls
                    EmsZapiManager.INSTANCE.deleteEventNotification(filterName, zapiConnection);
                    EmsZapiManager.INSTANCE.deleteEventNotificationDestination(destinationName, destinationURL, zapiConnection);
                    EmsZapiManager.INSTANCE.deleteEventFilter(filterName, zapiConnection);
                    throw ex;
                }
            }

            if (!emsNotificationCheck(destinationName, zapiConnection)) {
                logger.error("EMS notification test failed; It could be the CN of Server Certificate setting, which has to be DNS resolvable hostname or IP Address. CN = {}", hostname);
            }

        } catch (Exception ex) {
            logger.error("failed to register EMS: ", ex);
            throw ex;
        } finally {
            if (zapiConnection != null) {
                zapiConnection.close();
            }
        }
    }

    /**
     * Check if the EMS notification can go through.
     *
     * @param destinationName
     * @param zapiConnection
     * @return
     * @throws EmsZapiManagerException
     */
    private static boolean emsNotificationCheck(String destinationName, NaServer zapiConnection) throws EmsZapiManagerException {
        String nodeName = EmsZapiManager.INSTANCE.getClusterNodeName(zapiConnection);
        return EmsZapiManager.INSTANCE.emsNotificationDestinationCheck(destinationName, nodeName, zapiConnection);
    }

    /**
     * Get DS cluster version. We only get version down to: generation.major. We don't support minor version
     *
     * @param attrs
     * @return
     */
    private static String getDSClusterVersion(NaServer zapiConnection) {
        String version = "0.0";
        try {
            version = EmsZapiManager.INSTANCE.getDSClusterVersion(zapiConnection);
        }  catch(EmsZapiManagerException ex) {
            logger.error("Failed to get DS cluster version: {}", ex);
        }

        return version;
    }

    public static void unRegisterEms(String hostname, Properties dsProperties) throws EmsZapiManagerException {
        NaServer zapiConnection = null;

        try {
            zapiConnection = EmsZapiManager.INSTANCE.createZAPIConnection(dsProperties);

            // Don't unregister if Ontap version is lower than 9.0
            if (NetappEmsUtil.compareOntapVersion(getDSClusterVersion(zapiConnection), LONG_BOARD) < 0) {
                return;
            }

            String destinationName = EmsZapiManager.INSTANCE.getDestinationName(hostname);
            String destinationURL = EmsZapiManager.INSTANCE.getDestinationURL(hostname);
            String filterName = EmsZapiManager.INSTANCE.getFilterName(hostname);

            EmsZapiManager.INSTANCE.deleteEventNotification(filterName, zapiConnection);

            // Delete client key before deleting event notification destination because we query the event notification destination for the associated key
            EmsZapiManager.INSTANCE.deleteClientKey(destinationName, zapiConnection);

            EmsZapiManager.INSTANCE.deleteEventNotificationDestination(destinationName, destinationURL, zapiConnection);

            EmsZapiManager.INSTANCE.deleteEventFilter(filterName, zapiConnection);

            EmsZapiManager.INSTANCE.uninstallSecurityCertificate(destinationName, CERT_TYPE, zapiConnection);

        } catch (Exception ex) {
            logger.error("failed to unregister EMS: ", ex);
            throw new EmsZapiManagerException(ex.getMessage() == null ? "Failed to unregister EMS." : ex.getMessage());
        } finally {
            if (zapiConnection != null) {
                zapiConnection.close();
            }
        }
    }

    /**
     * Reload server certificate to the cluster.
     *
     * @param emsDestName
     * @param props
     * @throws EmsZapiManagerException
     */
    public static void reloadSecurityCertificate(String emsDestName, Properties props) throws EmsZapiManagerException {
        NaServer zapiConnection = null;
        try {
            zapiConnection = EmsZapiManager.INSTANCE.createZAPIConnection(props);
            EmsZapiManager.INSTANCE.uninstallSecurityCertificate(emsDestName, CERT_TYPE, zapiConnection);
            EmsZapiManager.INSTANCE.installSecurityCertificate(emsDestName, CERT_TYPE, ServerCertificateUtils.getServerCertificates(), zapiConnection);

        } catch (Exception ex) {
            logger.error("failed to unregister EMS: ", ex);
            throw new EmsZapiManagerException(ex.getMessage() == null ? "Failed to reload server certificate to cluster." : ex.getMessage());
        } finally {
            if (zapiConnection != null) {
                zapiConnection.close();
            }
        }
    }

    /**
     * Filter out the EmsConfig which version is higher than the cluster version, and its state is non-toBeDeleted||deleted
     *
     * @param emsConfigs
     * @param clusterVersion
     * @return
     */
    private static Set<String> getFilterRulesForAddDs(List<NetAppEmsConfig> emsConfigs, String clusterVersion) {
        Set<String> filterRules = new HashSet<>();

        // filter out unqualified emsConfigs
        List<NetAppEmsConfig> filterRlst = emsConfigs.stream()
            .filter(emsConfig -> {
                return NetappEmsUtil.compareOntapVersion(emsConfig.getOntapVersion(), clusterVersion) <= 0 &&
                       !emsConfig.getState().equals(NetAppEmsConfig.EmsMessageState.TO_BE_DELETED) &&
                       !emsConfig.getState().equals(NetAppEmsConfig.EmsMessageState.DELETED);
             })
            .collect(Collectors.toList());

        // add the filter rules and remove duplicates
        filterRlst.forEach(rslt -> filterRules.add(rslt.getMessageFilterRule()));

        return filterRules;
    }

    /**
     * Register with the ds to receive EMS REST push, and mark the ds for EMS destination enabled.
     *
     * @param ds
     * @param dsAttrs
     * @param connection
     * @return DataSourceNew - the DS whose filter rule is updated
     */
    public static DataSourceNew updateEmsToDs(String hostname, DataSourceNew ds, Properties dsAttrs, List<NetAppEmsConfig> emsToBeAdded,
        List<NetAppEmsConfig> emsToBeDeleted, List<NetAppEmsConfig> emsConfigAll) {
        NaServer zapiConnection = null;

        try {
            zapiConnection = EmsZapiManager.INSTANCE.createZAPIConnection(dsAttrs);

            final String clusterVersion = getDSClusterVersion(zapiConnection);
            final String filterName = EmsZapiManager.INSTANCE.getFilterName(hostname);
            String destinationName = EmsZapiManager.INSTANCE.getDestinationName(hostname);

            // Update with client authentication url and key when cluster version is 9.0 and above
            if(NetappEmsUtil.compareOntapVersion(clusterVersion, LONG_BOARD) >= 0 &&
               !EmsZapiManager.INSTANCE.isClientAuthenticationEnabled(destinationName, zapiConnection)) {
                String destinationURL = EmsZapiManager.INSTANCE.getDestinationURL(hostname);
                try {
                    ClientKeyInformation clientKey = registerClientKey(zapiConnection);
                    EmsZapiManager.INSTANCE.modifyEventNotificationDestination(destinationName, destinationURL, clientKey.getClientCertificateKeyAuthority(), clientKey.getClientCertificateKeySerial(), zapiConnection);
                } catch (EmsZapiManagerException modifyException) {
                    throw modifyException;
                }
            }

            // take care of adding filter rules
            //// filter out unqualified emsConfigs
            if (!emsToBeAdded.isEmpty()) {
                List<NetAppEmsConfig> filterRlstAdd = emsToBeAdded.stream()
                        .filter(emsConfig ->
                        NetappEmsUtil.compareOntapVersion(emsConfig.getOntapVersion(), clusterVersion) <= 0)
                        .collect(Collectors.toList());

                //// add the filter rules and remove duplicates
                Set<String> filterRulesAdd = getEmsFilterRulesForUpdate(filterRlstAdd, emsConfigAll);

                //// we do have filter rules to add to this DS
                if (!filterRulesAdd.isEmpty()) {
                    for (String filterRule : filterRulesAdd) {
                        try {
                            EmsZapiManager.INSTANCE.addFilterRule(filterName, filterRule, FILTER_RULE_INCLUDE, zapiConnection);
                        } catch (EmsZapiManagerException ex) {
                            if (ex.getErrno() != EmsZapiManager.EDUPLICATEENTRY && ex.getErrno() != EmsZapiManager.E_EMS_NO_EVENT_MATCHED_THE_RULE) {
                                throw ex;
                            }
                        }
                    }
                }
            }

            // take care of deleting filter rules
            // filter out unqualified emsConfigs for delete
            if (!emsToBeDeleted.isEmpty()) {
                List<NetAppEmsConfig> filterRlstDelete = emsToBeDeleted.stream()
                        .filter(emsConfig ->
                        NetappEmsUtil.compareOntapVersion(emsConfig.getOntapVersion(), clusterVersion) <= 0)
                        .collect(Collectors.toList());

                // add the filter rules and remove duplicates
                Set<String> filterRulesDelete = getEmsFilterRulesForUpdate(filterRlstDelete, emsConfigAll);

                // we do have filter rules to add to this DS
                if (!filterRulesDelete.isEmpty()) {
                    //deleteEventFilterRules
                    for (String filterRule : filterRulesDelete) {
                        try {
                            EmsZapiManager.INSTANCE.deleteEventFilterRule(filterName, filterRule, FILTER_RULE_INCLUDE, zapiConnection);
                        } catch (EmsZapiManagerException ex) {
                            if (ex.getErrno() != EmsZapiManager.EDUPLICATEENTRY && ex.getErrno() != EmsZapiManager.E_EMS_NO_EVENT_MATCHED_THE_RULE) {
                                throw ex;
                            }
                        }
                    }
                }
            }

            return ds;
        } catch (EmsZapiManagerException e) {
            logger.error("Unable to add EMS filter rule to cluster " + ds.getName(), e);
            return ds;
        } finally {
            if (zapiConnection != null) {
                zapiConnection.close();
            }
        }
    }

    /**
     * Get hostname from server certificate, which is set to CN in the certificate.
     *
    * @return hostName - from the server certificate CN
    *         null - error occurred retrieving CN from the certificate
    * @throws EmsZapiManagerException
    */
    public static String getHostName() {
        String cn = ServerCertificateUtils.getCNFromServerCertificate();
        if (cn == null) {
            throw new IllegalArgumentException("Unable to get CN from the server certificate.");
        } else {
            logger.info("CN of Server Certificate = {}, which has to be DNS resolvable hostname or IP Address.", cn);
            return cn;
        }
    }

    /**
     * Since filter rule supports wildcard character we need to make sure the input filter rule is the first one to add, or the last one to delete.
     *
     * @param emsConfigUpdate
     * @param emsConfigAll
     * @return the set of filter rules needed to be added or deleted
     */
    private static Set<String> getEmsFilterRulesForUpdate(final List<NetAppEmsConfig> emsConfigUpdate, final List<NetAppEmsConfig> emsConfigAll) {

        // get the unique filter rules for update
        Set<String> rulesRslt = new HashSet<>();
        Set<String> rulesForUpdate = new HashSet<>();
        emsConfigUpdate.forEach(rslt -> rulesForUpdate.add(rslt.getMessageFilterRule()));

        rulesForUpdate.forEach(rule -> {
            List<NetAppEmsConfig> configRslt = emsConfigAll.stream().filter(emsConfig -> {
                return emsConfig.getMessageFilterRule().equals(rule) && emsConfig.getState().equals(NetAppEmsConfig.EmsMessageState.ADDED);
            }).collect(Collectors.toList());

            // this is the first one for adding or last one for deleting
            if (configRslt.isEmpty()) {
                rulesRslt.add(rule);
            }
        });

        return rulesRslt;
    }

}
