/*
 * $Id: //depot/prod/acquisition/main/tools/config/eclipse/.settings/org.eclipse.jdt.ui.prefs#1 $
 *
 * Copyright (c) 2017 NetApp, Inc.
 * All rights reserved.
 */
package com.onaro.sanscreen.acquisition.sessions;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import netapp.manage.NaElement;
import netapp.manage.NaServer;

import com.netapp.oci.netapp.client.interfaces.data.EmsZapiManagerException;
import com.netapp.oci.netapp.ems.EmsZapiManager;

/**
 * Utility class to clear the EMS registry from ONTAP cluster.
 */
public final class ClearOntapEmsRegistry {

    @SuppressWarnings("unchecked")
    public static void main(String[] args) {
        if (args.length != 3 && args.length != 4) {
            StringBuilder buf = new StringBuilder();
            buf.append("Clear ONTAP EMS registries:\n\n");
            buf.append("\tClearOntapEmsRegistry <IP of the Cluster> <Cluster Login Username> <Cluster Login Password> <The Hostname of the Registered OCUM instance>\n\n");
            buf.append("** <The Hostname of the Registered OCUM instance> is optional. Clear all the EMS registries if it's not provided\n");
            System.out.println(buf);
            System.exit(-1);
        }

        String clusterIp = args[0];
        String username = args[1];
        String pswd = args[2];
        String ocumHost = args.length == 4 ? args[3] : null;

        NaServer zapiConnection = null;
        try {
            zapiConnection = EmsZapiManager.INSTANCE.createZAPIConnection(clusterIp, username, pswd);

            if (ocumHost == null) {
                NaElement rsp = getEmsNotificationDests(zapiConnection);
                List<String> dests = new ArrayList<>();

                if (rsp != null) {
                    NaElement attributesList = rsp.getChildByName("attributes-list");
                    if (attributesList != null) {
                        List<NaElement>  destInfoList = attributesList.getChildren();
                        if (destInfoList != null) {
                            for (NaElement destInfo : destInfoList) {
                                String destUrl = destInfo.getChildByName("destination").getChildByName("string").getContent();
                                String dest = null;
                                try {
                                    dest = new URL(destUrl).getHost();
                                } catch (MalformedURLException e) {
                                    System.out.println("Malformed URL, ignore it - " + e);
                                }

                                if (dest != null) {
                                    dests.add(dest);
                                }
                            }
                        }
                    }

                    for (String dest : dests) {
                        clearEmsRegistry(dest, zapiConnection);
                    }
                }
            } else {
                clearEmsRegistry(ocumHost, zapiConnection);
            }
        } catch (EmsZapiManagerException e) {
            throw new RuntimeException(e);
        } finally {
            if (zapiConnection != null) {
                zapiConnection.close();
            }
        }
    }

    /**
     * Clear an EMS registry identified by the destination hostname/ip.
     *
     * @param dest
     * @param zapiConnection
     */
    private static void clearEmsRegistry(String dest, NaServer zapiConnection) {
        String destinationName = EmsZapiManager.INSTANCE.getDestinationName(dest);
        String destinationURL = EmsZapiManager.INSTANCE.getDestinationURL(dest);
        String filterName = EmsZapiManager.INSTANCE.getFilterName(dest);

        try {
            EmsZapiManager.INSTANCE.deleteEventNotification(filterName, zapiConnection);

            // Delete client key before deleting event notification destination because we query the event notification destination for the associated key
            EmsZapiManager.INSTANCE.deleteClientKey(destinationName, zapiConnection);

            EmsZapiManager.INSTANCE.deleteEventNotificationDestination(destinationName, destinationURL, zapiConnection);

            EmsZapiManager.INSTANCE.deleteEventFilter(filterName, zapiConnection);

            EmsZapiManager.INSTANCE.uninstallSecurityCertificate(destinationName, "server-ca", zapiConnection);
        } catch (EmsZapiManagerException e) {
            System.out.println(dest + "Unable to remove EMS Post Destination <" + dest + "> from the cluster - " + e);
        }

        System.out.println(dest + " has been removed from the cluster as an EMS REST destination.");
    }

    /**
     * Get all the EMS REST destinations from the cluster.
     *
     * @param zapiConnection
     * @return
     */
    private static NaElement getEmsNotificationDests(NaServer zapiConnection) {
        try {
            NaElement rsp = EmsZapiManager.INSTANCE.showNotificationDestinations(zapiConnection);
            return rsp;
        } catch (EmsZapiManagerException e) {
            System.out.println(e);
            return null;
        }
    }

}
