/*
 * Copyright (c) 2010 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.acquisition.sessions;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import com.onaro.commons.types.Period;
import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionAuditMessage;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceAudit;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceNew;
import com.onaro.sanscreen.server.tables.acquisition.DataSourceAuditDetailsTable;
import com.onaro.sanscreen.server.tables.acquisition.DataSourceAuditTable;
import com.onaro.sanscreen.server.tables.acquisition.DataSourceTableNew;
import org.slf4j.Logger;

import com.netapp.oci.platform.common.ServerConstant;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;

/**
 * Handles everything related to data source audit sessions. This service provides means for data sources to update audit
 * entries for different packages running in a data source. It also automatically derives the data source status from the
 * combined status of all packages.
 *
 * A data source status is defined to be:
 * <ul> 'Stopped' if data source is stopped</ul>
 * <ul> 'Success' if <u>ALL</u> packages are successfull.</ul>
 * <ul> List of packages that failed: 'Service Insight, Application Insight failed' when some or all packages failed.</ul>
 * <ul> List of packages that are stuck: 'Service Insight not running' when the package didn't acquire for a long time.</ul>
 */
public class DataSourceAuditManager {
    private static transient final Logger logger =org.slf4j.LoggerFactory.getLogger("com.onaro.sanscreen.acquisition.sessions");

    private static final int ONE_DAY = 24 * 60 * 60 * 1000;

    /**
     * To prevent from going to the database too many times, this method keeps the last audit for each package (per DS)
     */
    private static Map<Long, Map<String, PackageAuditHistory>> auditHistoryCache = Collections.synchronizedMap(new HashMap<Long, Map<String, PackageAuditHistory>>());
    private static Map<Long, DataSourceAudit> openAuditSessionsCache = Collections.synchronizedMap(new HashMap<Long, DataSourceAudit>());

    public DataSourceAuditManager() {
    }

    public synchronized void initialize() {
        auditHistoryCache.clear();

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();

            final DataSourceTableNew dataSourceTable = new DataSourceTableNew(connection);
            final DataSourceAuditTable dataSourceAuditTable = new DataSourceAuditTable(connection);

            final List<DataSourceNew> allDS = dataSourceTable.getAll();
            for (final DataSourceNew dataSource : allDS) {
                final List<DataSourceAudit> statuses = dataSourceAuditTable.getAuditHistory(dataSource.getId());
                for (final DataSourceAudit audit : statuses) {
                    final PackageAuditHistory packageAuditHistory = findPackageAuditHistory(audit);

                    // Add this to the audit cache. Don't update the auditSessionCache because this audit entry
                    // does not reflect an open session
                    packageAuditHistory.push(audit);
                }
            }
        }
        catch (TableException e) {
            logger.error("Failed to load audit history...", e);

            // Fall though...
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     * Create a new audit session
     *
     * @param packageName
     * @param dsID
     * @param status
     *
     * @return Session ID (used by {@link #closeDsAudit} to update and close this session)
     *
     * @throws AcquisitionUnitException
     */
    public Long createDsAudit(final Long dsID, final String packageName, final String status) throws AcquisitionUnitException {
        Connection connection = null;
        try {
            if (logger.isDebugEnabled())
                logger.debug("Creating audit for datasource's originId: [" + dsID + "], with status: [" + status + "]");

            connection = ConnectionHelper.newConnectionNoEjb();

            final DataSourceAuditTable dataSourceAuditTable = new DataSourceAuditTable(connection);
            DataSourceAudit audit = new DataSourceAudit(packageName,
                                                        status,
                                                        "",
                                                        dsID,
                                                        new Date().getTime(),
                                                        ServerConstant.END_DATE,
                                                        1);


            final PackageAuditHistory packageAudits = findPackageAuditHistory(audit);
            synchronized (packageAudits) {
                dataSourceAuditTable.insert(audit);

                if (logger.isDebugEnabled())
                    logger.debug("data source audit record was created successfully, returning auditId: [" + audit.getId() + "]");

                updateAuditCache(audit, packageAudits);
            }

            // Now update the aggregated status of the datasource
            final DataSourceNew dataSource = findDataSource(audit.getDsId(), connection);
            if (dataSource == null)
                throw new IllegalStateException("Could not find data source " + audit.getDsId());

            final String newDataSourceStatus = figureOutDataSourceStatus(dataSource);
            if (!newDataSourceStatus.equals(dataSource.getStatus())) {
                final DataSourceTableNew dataSourceTable = new DataSourceTableNew(connection);
                dataSource.setStatus(newDataSourceStatus);
                dataSourceTable.update(dataSource);
            }

            return audit.getId();
        }
        catch (Exception e) {
            throw new AcquisitionUnitException("Failed to create audit originId:" + dsID, e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    public void updateAudit(final Long auditSessionID, final String newStatus) throws AcquisitionUnitException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();

            final DataSourceAuditTable dataSourceAuditTable = new DataSourceAuditTable(connection);

            // Get the audit record from the cache...
            DataSourceAudit audit = openAuditSessionsCache.get(auditSessionID);
            if (audit == null) {
                if (logger.isDebugEnabled())
                    logger.debug("updateAudit() - Failed to find audit id " + auditSessionID);

                // Try from the database...
                audit = dataSourceAuditTable.find(auditSessionID.longValue());
                if (audit == null) {
                    logger.warn("updateAudit() - Failed to find audit id " + auditSessionID);
                    return;
                }

                openAuditSessionsCache.put(auditSessionID, audit);
            }

            synchronized (findPackageAuditHistory(audit)) {
                audit.setStatus(newStatus);

                if (logger.isDebugEnabled())
                    logger.debug("Audit record was found using audit's ID: [" + auditSessionID + "], belongs to data source's OringId: [" + audit.getDsId() + "], current status: [" + audit.getStatus() + "]");

                dataSourceAuditTable.update(audit);
            }

            // Now update the aggregated status of the datasource
            updateDS(audit.getDsId(), connection);

            if (logger.isDebugEnabled())
                logger.debug("Audit record was updated using audit's ID: [" + audit.getId() + "], belongs to data source's OringId: [" + audit.getDsId() + "], current status: [" + audit.getStatus() + "]");
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to close audit id:" + auditSessionID, e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     * This method closes the audit session.
     *
     * <p/>The current record is an opened one. We have to look at the record
     * before that and decide if the current record should be merged into the last one. If so, we delete the current
     * record and increment the count in the last record. Otherwise, we should close the current record with the new
     * status. This way when a data source is continuously successfull, only one audit record will be written with an
     * incrementing count.
     */
    public void closeDsAudit(final Long auditSessionID, final String status, final List<AcquisitionAuditMessage> messages) throws AcquisitionUnitException {
        //record the messages

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();

            if (messages != null && messages.size() > 0)
                updateSampleAuditMessages(auditSessionID, messages, connection);


            final DataSourceAuditTable dataSourceAuditTable = new DataSourceAuditTable(connection);

            // Get the audit record from the cache...
            DataSourceAudit audit = openAuditSessionsCache.get(auditSessionID);
            if (audit == null) {
                if (logger.isDebugEnabled())
                    logger.debug("closeDsAudit() - Failed to find audit id " + auditSessionID);

                // Try from the database...
                audit = dataSourceAuditTable.find(auditSessionID.longValue());
                if (audit == null) {
                    logger.warn("closeDsAudit() - Failed to find audit id " + auditSessionID);
                    return;
                }

                openAuditSessionsCache.put(auditSessionID, audit);
            }

            final PackageAuditHistory packageAuditHistory = findPackageAuditHistory(audit);
            // Get the history of this session from the history cache.
            if (packageAuditHistory == null)
                throw new IllegalStateException("Could not find audit in cache");

            synchronized (packageAuditHistory) {
                // Update and close the audit
                audit.setStatus(status);
                audit.setEndTime(new Date().getTime());

                if (logger.isDebugEnabled())
                    logger.debug("Audit record was found using audit's ID: [" + auditSessionID + "], belongs to data source's OriginId: [" + audit.getDsId() + "], current status: [" + audit.getStatus() + "]");

                packageAuditHistory.push(audit);

                // Update DB (no need to update caches since we were working on the object at the top of stack)
                dataSourceAuditTable.update(audit);
            }

            // Clear old records - more than 2 month old
            final long oldStartDate = audit.getEndTime() - 2* Period.MONTH.milliseconds();
            dataSourceAuditTable.deleteByOldStartDate(audit.getDsId(), oldStartDate);

            // Now update the aggregated status of the datasource
            updateDS(audit.getDsId(), connection);

            // This audit session is closed. Remove it from the session cache
            openAuditSessionsCache.remove(auditSessionID);
        }
        catch (TableException e) {
            logger.error("Fail close audit id:" + auditSessionID, e);
            throw new AcquisitionUnitException("Failed to close audit id:" + auditSessionID, e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
       * When this method is used it means there was no change in the audit
       *
       * @param auditSessionID
       * @param status
       *
       * @throws AcquisitionUnitException
       */
    public void closeDsAudit(final Long auditSessionID, final String status) throws AcquisitionUnitException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();

            final DataSourceAuditTable dataSourceAuditTable = new DataSourceAuditTable(connection);

            // Get the audit record from the cache...
            DataSourceAudit audit = openAuditSessionsCache.get(auditSessionID);
            if (audit == null) {
                if (logger.isDebugEnabled())
                    logger.debug("closeDsAudit() - Failed to find audit id " + auditSessionID);

                // Try from the database...
                audit = dataSourceAuditTable.find(auditSessionID.longValue());
                if (audit == null) {
                    logger.warn("closeDsAudit() - Failed to find audit id " + auditSessionID);
                    return;
                }

                openAuditSessionsCache.put(auditSessionID, audit);
            }

            if (logger.isDebugEnabled())
                logger.debug("Audit record was found using audit's ID: [" + auditSessionID + "], belongs to data source's OringId: [" + audit.getDsId() + "], current status: [" + audit.getStatus() + "]");

            final PackageAuditHistory packageAuditHistory = findPackageAuditHistory(audit);
            if (packageAuditHistory == null)
                throw new IllegalStateException("Could not find audit in cache");

            synchronized (packageAuditHistory) {
                audit.setStatus(status);
                // Get the history of this session from the history cache.

                if (packageAuditHistory.hasHistory()) {

                    // Get rid of the open line
                    packageAuditHistory.pop();

                    final DataSourceAudit previousAuditRecord = packageAuditHistory.pop();

                    // Delete the current record (we will modify the previous one with updated count and endtime)
                    dataSourceAuditTable.deleteById(audit.getId());

                    // This audit session is closed. Remove it from the session cache
                    openAuditSessionsCache.remove(auditSessionID);

                    // The latest record is now the previous one with an incremented 'number of times' (we just deleted the session)
                    previousAuditRecord.setNumberOfTimes(previousAuditRecord.getNumberOfTimes() + 1);
                    previousAuditRecord.setEndTime(new Date().getTime());
                    audit = previousAuditRecord;

                }
                //upadate history ( we are using pop and push in order to set the last succsess time ).
                packageAuditHistory.push(audit);
                // Update DB
                dataSourceAuditTable.update(audit);
            }

            final DataSourceNew dataSource = findDataSource(audit.getDsId(), connection);

            final String newDataSourceStatus = figureOutDataSourceStatus(dataSource);
            if (!newDataSourceStatus.equals(dataSource.getStatus())) {
                final DataSourceTableNew dataSourceTable = new DataSourceTableNew(connection);
                dataSource.setStatus(newDataSourceStatus);
                dataSourceTable.update(dataSource);
            }


        }
        catch (TableException e) {
            logger.error("Fail close audit id:" + auditSessionID, e);
            throw new AcquisitionUnitException("Failed to close audit id:" + auditSessionID, e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     * @param auditSessionID
     * @param messages
     * @param connection
     *
     * @throws AcquisitionUnitException
     */
    private void updateSampleAuditMessages(final Long auditSessionID, final List<AcquisitionAuditMessage> messages, final Connection connection) throws AcquisitionUnitException {
        // TODO: This is a potential resource leak. Use the connection from the calling method instead.
        try {
            final DataSourceAuditDetailsTable dsadt = new DataSourceAuditDetailsTable(connection);
            for (final AcquisitionAuditMessage message : messages) {
                message.setAuditId(auditSessionID);
                dsadt.insert(message);
            }
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to insert messages for audit id:" + auditSessionID, e);
        }
    }



    public void updateDS(long dsID, Connection connection) throws TableException {
        // Now update the aggregated status of the datasource
        final DataSourceNew dataSource = findDataSource(dsID, connection);

        final String newDataSourceStatus = figureOutDataSourceStatus(dataSource);
        if (!newDataSourceStatus.equals(dataSource.getStatus())) {
            final DataSourceTableNew dataSourceTable = new DataSourceTableNew(connection);
            dataSource.setStatus(newDataSourceStatus);
            dataSourceTable.update(dataSource);
        }
    }

    public void stopReportingAuditForPacakge(final Long dsID, final String packageName) throws AcquisitionUnitException {
        // First check if this package is not marked as disabled already...
        final PackageAuditHistory packageAuditHistory = findPackageAuditHistory(dsID, packageName);

        final DataSourceAudit lastAudit = packageAuditHistory.peek();

        // If there is no history for this package, don't add any audit. This might indicate that the history is cleaned
        // or the package never ran (not licensed).
        if (lastAudit == null)
            return;

        // Avoid repetitions of the "Disabled" report...
        if (lastAudit.getStatus() != null && lastAudit.getStatus().equals("Disabled"))
            return;

        // This package is not marked as disabled. Let's do it now!
        final Long tempSessionID = createDsAudit(dsID, packageName, "Disabled");
        closeDsAudit(tempSessionID, "Disabled", null);
    }

    public void reportStandbyForPackage(final Long dsID, final String packageName) throws AcquisitionUnitException {
        // First check if this package is not marked as disabled already...
        final PackageAuditHistory packageAuditHistory = findPackageAuditHistory(dsID, packageName);

        final DataSourceAudit lastAudit = packageAuditHistory.peek();
        if (lastAudit != null && lastAudit.getStatus() != null && lastAudit.getStatus().equals("Stand-by"))
            return;

        // This package is not marked as disabled. Let's do it now!
        final Long tempSessionID = createDsAudit(dsID, packageName, "Stand-by");
        closeDsAudit(tempSessionID, "Stand-by", null);
    }

    private DataSourceNew findDataSource(final Long dsID, final Connection connection) throws TableException {
        // Now update the data source status (if neccessary). First try to fetch it from the cache (to spare a
        // DB query...)
        final DataSourceTableNew dataSourceTable = new DataSourceTableNew(connection);
        return dataSourceTable.getById(dsID);
    }

    private String figureOutDataSourceStatus(final DataSourceNew dataSourceData) {
        if (dataSourceData.getManual() == 2)
            return "Stopped";

        // Build a map of package statuses. The map is a sorted map to make sure that the packages are always reported
        // in the same order.
        final SortedMap<String, String> packageStatuses = new TreeMap<String, String>();

        final long now = new Date().getTime();

        // Lookup the history of this data source...
        final Map<String, PackageAuditHistory> auditsForDsPerPackage = auditHistoryCache.get(dataSourceData.getId());
        if (auditsForDsPerPackage == null)
            return "";

        // For each package, get the last audit reported for. This is how we determine the status of the package
        for (final Map.Entry<String, PackageAuditHistory> entryForPackage : auditsForDsPerPackage.entrySet()) {
            // Ignore the 'System package' that is used to report scheduled tasks like deleting originators, etc...
            if (entryForPackage.getKey().equals("System"))
                continue;

            final PackageAuditHistory auditStackForPackage = entryForPackage.getValue();

            final DataSourceAudit lastAudit = auditStackForPackage.peek();

            // If no history is available, don't report anything about this package
            if (lastAudit == null)
                continue;

            // Disabled packages are not considered when reporting status
            if (isDisabled(lastAudit))
                continue;

            if (isStandby(lastAudit)) {
                packageStatuses.put(lastAudit.getPackageName(), "stand-by");
                continue;
            }

            if (isRunning(lastAudit)) {
                packageStatuses.put(lastAudit.getPackageName(), "acquiring");
                continue;
            }

            if (!isStatusOK(lastAudit)) {
                packageStatuses.put(lastAudit.getPackageName(), "failed");
                continue;
            }

            // ASSUMPTION: If we're here - we assume package is successful

            // Check for package timeout (is package STUCK) (did not succeed for 24hours)
            if (packageStuck(auditStackForPackage, now)) {
                packageStatuses.put(lastAudit.getPackageName(), "not running");
                continue;
            }

            packageStatuses.put(lastAudit.getPackageName(), "success");
        }

        final StringBuilder sb = new StringBuilder(100);

        for (final Map.Entry<String, String> packageStatus : packageStatuses.entrySet())
            sb.append(packageStatus.getKey()).append(" ").append(packageStatus.getValue()).append(", ");

        if (sb.length() > 0) {
            sb.deleteCharAt(sb.length() - 1);
            sb.deleteCharAt(sb.length() - 1);
        }

        return sb.toString();
    }

    /**
     * Figures out if this package is stuck.
     *
     * <p/>This method will only catch DS that have reported success - but are STUCK - and not scheduled / reporting anymore.
     * This will apply in two of the following cases:
     *   1. Last audit was Success - but happenned more than a day ago
     *   2. Last closed audit was success, it is now acquiring - but stuck (in this case 'lastAudit' will be the success = the same as above)
     *
     * <p/>Take the last audit close time if successful or assume it never happened. Remember - this method is only called
     * after we already evaluated errors, stand-by, etc... We already concluded that the package was successful - and
     * we want to eliminate the option of it being STUCK.
     *
     * HEADSUP: Since we don't have an easy way to retrieve the package's poll interval from the server side, we will default to 1 day.
     *
     * @param auditStackForPackage
     * @param now
     *
     * @return <code>true</code> if package is stuck, <code>false</code> otherwise
     */
    private boolean packageStuck(final PackageAuditHistory auditStackForPackage, final long now) {
        final DataSourceAudit lastAudit = auditStackForPackage.peek();

        return (now - lastAudit.getEndTime()) > ONE_DAY;
    }

    private boolean isRunning(final DataSourceAudit audit) {
        if (audit == null || audit.getStatus() == null)
            return false;

        return audit.getStatus().startsWith("Acquiring") || audit.getStatus().startsWith("Sent");
    }

    private boolean isStandby(final DataSourceAudit audit) {
        if (audit == null || audit.getStatus() == null)
            return false;

        return audit.getStatus().equals("Stand-by");
    }

    private boolean isStatusOK(final DataSourceAudit audit) {
        if (audit == null || audit.getStatus() == null)
            return false;

        return audit.getStatus().equals("Success") || audit.getStatus().startsWith("Stopped reporting device");
    }

    private boolean isDisabled(final DataSourceAudit audit) {
        if (audit == null || audit.getStatus() == null)
            return false;

        return audit.getStatus().equals("Disabled");
    }

    private void updateAuditCache(final DataSourceAudit audit, final PackageAuditHistory packageAudits) {

        packageAudits.push(audit);
        openAuditSessionsCache.put(audit.getId(), audit);
    }

    private PackageAuditHistory findPackageAuditHistory(final DataSourceAudit audit) {
        return findPackageAuditHistory(audit.getDsId(), audit.getPackageName());
    }

    /**
     * Given a specific audit, find the right audit stack in the cache for this audit.
     */
    private synchronized PackageAuditHistory findPackageAuditHistory(final Long dsID, final String packageName) {
        // Keep this audit in the 'lastAuditEntries' cache
        Map<String, PackageAuditHistory> auditHistoryForPackage = auditHistoryCache.get(dsID);
        if (auditHistoryForPackage == null) {
            auditHistoryForPackage = new HashMap<String, PackageAuditHistory>();
            auditHistoryCache.put(dsID, auditHistoryForPackage);
        }

        PackageAuditHistory packageAudits = auditHistoryForPackage.get(packageName);
        if (packageAudits == null) {
            packageAudits = new PackageAuditHistory();
            auditHistoryForPackage.put(packageName, packageAudits);
        }

        return packageAudits;
    }

    public DataSourceAudit[] getAuditHistory(final long dsID) throws AcquisitionUnitException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            DataSourceAuditTable dataSourceAuditTable = new DataSourceAuditTable(connection);
            Collection<DataSourceAudit> col = dataSourceAuditTable.findAllByOriginId(dsID);
            //noinspection unchecked
            return col.toArray(new DataSourceAudit[col.size()]);
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to getAllHistoryById:" + dsID, e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    public static List<DataSourceAudit> getAuditHistoryAscending(final long dsID, final Connection connection) throws AcquisitionUnitException {
        try {

            DataSourceAuditTable dataSourceAuditTable = new DataSourceAuditTable(connection);

            // getAuditHistory orders results ascending by start time
            return dataSourceAuditTable.getAuditHistory(dsID);
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to getAllHistoryById:" + dsID, e);
        }
    }

    public static boolean isDsStatusOK(final String dataSourceStatus, final boolean acqUnitRunning) {
        if (!acqUnitRunning)
            return false;

        if (dataSourceStatus == null)
            return true;

        if (dataSourceStatus.length() == 0)
            return true;

        return !dataSourceStatus.equals("Stopped") && dataSourceStatus.indexOf("not running") == -1 && dataSourceStatus.indexOf("failed") == -1;
    }

    // finds the starttime of the nth most recent closed audit. If there are not n closed audits,
    // then the oldest starttime of any audit is returned. If there are no audits, 0 is returned.
    public Long getStartTimeForRecentAudits(final List<DataSourceAudit> auditsForDatasource, final int n) {

        // Get the audits for the datasource in descending order.
        //List<DataSourceAudit> auditsForDatasource = auditHistoryCache.
                //.findAllByOriginId(dataSource.getId());

        long result = 0;
        int i = 1;
        for (DataSourceAudit audit : auditsForDatasource) {
            result = audit.getStartTime();
            if (audit.isAuditSessionClosed()) {
                if (n == i) {
                    break;
                }
                i++;
            }
        }
        return result;
    }


    /**
     *
     * @param daysBack
     * @param connection
     * @throws SQLException
     *
     * Deletes all audits older than now - daysBack for active datasources.
     */
    public static void clearOldAudits(int daysBack, Connection connection) throws SQLException {

        // we are cleaning old entries from ds_audit every time we close existing audit session.
        // Look at closeDsAudit method - we remove 2 month old records there  

//        final String sql = "select * from acquisition.ds_audit " +
//                           "where " +
//                           "  starttime < unix_timestamp( subdate(now(), ?) ) * 1000" +
//                           "and " +
//                           "  ds_id in (SELECT id from ds where active=0)";
//        PreparedStatement statement = connection.prepareStatement(sql);
//        statement.setInt(1, daysBack);
//        statement.executeUpdate();

    }

    /**
     * Class representing the audit history for a package.
     * This data will be used for:
     * 1. Consolidating audit recordings (increasing count for same status)
     * 2. Identifying packages that are stuck (by recording last success time)
     */
    final static class PackageAuditHistory {
        private DataSourceAudit lastAudit = null;
        private DataSourceAudit previousAudit = null;

        PackageAuditHistory() {
        }

        void push(final DataSourceAudit audit) {
            previousAudit = lastAudit;
            lastAudit = audit;
        }

        public boolean hasHistory() {
            return lastAudit != null && previousAudit != null && lastAudit.getStatus() != null && previousAudit.getStatus() != null;
        }

        public DataSourceAudit pop() {
            final DataSourceAudit audit = lastAudit;
            lastAudit = previousAudit;
            previousAudit = null;

            return audit;
        }

        public DataSourceAudit peek() {
            return lastAudit;
        }
    }
}
