/*
 * Copyright (c) 2010 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.acquisition.sessions;

import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;
import com.onaro.commons.util.MapOfSets;
import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionUnitNew;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceAudit;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceNew;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceStatus;
import com.onaro.sanscreen.server.tables.acquisition.AcquisitionUnitTableNew;
import com.onaro.sanscreen.server.tables.acquisition.DataSourceTableNew;
import org.slf4j.LoggerFactory;

import java.sql.Connection;
import java.util.*;

public class DataSourceStatusQueryImpl {
    private static final org.slf4j.Logger logger =
            LoggerFactory
            .getLogger("com.onaro.sanscreen.acquisition.sessions");

    /**
     * Wrapper method.  Invoke {#getDataSourceStatusQuery} using default parameters.
     */
    public DataSourceStatus[] getDataSourceStatusQuery() throws AcquisitionUnitException {
        return getDataSourceStatusQuery(0, 0, 0);
    }

    /**
     * Returns an array of {@link DataSourceStatus} instances representing the current data source's status.
     * The following summarize the logic of this method:
     * <ul>
     * <li> Retrieve a list of ALL data sources in the Database.
     * <li> Build a map with the relationship between AcquisitionUnitName-To-AcquisitionUnitStatus
     * <li> For each data source, retrieve all the rows from DATASOURCEAUDIT table.
     * <li> Process rows collected from DATASOURCEAUDIT table based on its status.
     * <li> Runs a list of Logics that define conditions for witch a Datasource can be considered as FAILED or SUCCESSfully terminated
     * <li> Returns an array of {@link DataSourceStatus} representing the status of the data sources.
     * </ul>
     */
    @SuppressWarnings({"ConstantConditions"})
    public DataSourceStatus[] getDataSourceStatusQuery(final int timeOutInHours, final int timeoutForSinglePollHours, int numOfAttemps) throws AcquisitionUnitException {

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();

            final Map<Long, DataSourceStatus> dataSourceStatuses = new LinkedHashMap<Long, DataSourceStatus>();
            final List<DataSourceNew> dataSources = findAllDataSources(connection);
            final Map<Long, AcquisitionUnitNew> acquisitionUnits = findAllAquisitionUnits(connection);
            final MapOfSets<Long, Long> datasourceIdsToDeviceIds = findConflictingDataSources(connection);

            // Analyzing all the data source one by one.
            for (final DataSourceNew dataSource : dataSources) {

                // TODO: check if dataSource is stopped
                final AcquisitionUnitNew acqUnit = acquisitionUnits.get(Long.valueOf(dataSource.getAuId()));

                final List<DataSourceAudit> dataSourceAudits = DataSourceAuditManager.getAuditHistoryAscending(dataSource.getId(), connection);

                // Default timeout is one day
                final long timeoutMillis = (timeOutInHours != 0 ? timeOutInHours : 24) * 60 * 60 * 1000;
                final long timeoutForSinglePollMillis = (timeoutForSinglePollHours != 0 ? timeoutForSinglePollHours : 2) * 60 * 60 * 1000;
                if (numOfAttemps == 0) {
                    numOfAttemps = 1;
                }

                // Create the data source status object based on the audit history
                final DataSourceStatus dataSourceStatus = createDataSourceStatus(acqUnit, dataSource, dataSourceAudits, timeoutMillis, timeoutForSinglePollMillis, numOfAttemps, datasourceIdsToDeviceIds);

                dataSourceStatuses.put(Long.valueOf(dataSource.getId()), dataSourceStatus);
            }

            return dataSourceStatuses.values().toArray(new DataSourceStatus[dataSourceStatuses.size()]);
        }
        catch (Exception e) {
            logger.error("Failed to get a list of data sources.", e);
            throw new AcquisitionUnitException("Failed to get a list of data sources.", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     * Creates an instance of {@link DataSourceStatus} formatting some of the data members in a more user friendly
     * description.  (based on document named: "Data-source alerts Software Requirement Specification"
     */
    private DataSourceStatus createDataSourceStatus(final AcquisitionUnitNew acquisitionUnit, final DataSourceNew dataSource, final List<DataSourceAudit> dsAuditRows, final long timeoutMillis, final long timeoutForSinglePollMillis, final int numOfAttemps, MapOfSets<Long, Long> datasourceIdsToDeviceIds) {
        final DataSourceStatus dataSourceStatus = new DataSourceStatus(acquisitionUnit, dataSource);

        // Initialize the status with all relevant audits
        dataSourceStatus.initialize(dsAuditRows);

        // Process the audit records to identify alert status for each data source package
        //if (dataSourceStatus.getStatus() != DataSourceStatus.DSStatus.Stopped)
        if (dataSourceStatus.getStatus() == DataSourceStatus.DSStatus.Running) {
            processAlerts(dataSourceStatus, timeoutMillis, timeoutForSinglePollMillis, numOfAttemps, datasourceIdsToDeviceIds);
        }

        return dataSourceStatus;
    }

    /**
     * Encapsulates the calls to the "logics" that understand Datasources status.
     */
    private void processAlerts(final DataSourceStatus dataSourceStatus, final long timeoutMillis, final long timeoutForSinglePollMillis, final int numOfAttemps, MapOfSets<Long, Long> datasourceIdsToDeviceIds) {
        final long now = new Date().getTime();

        // Go over each package and process the alerts individually
        for (final DataSourceStatus.DataSourcePackage dataSourcePackage : dataSourceStatus.getPackages()) {
            if (dataSourcePackage.isDisabled()) {
                dataSourcePackage.setAlert(false, "Disabled");
                continue;
            }

            final int numOfSequentialErrors = dataSourcePackage.getNumOfSequentialErrors();
            final String timeoutInHours = Long.toString(timeoutMillis / 1000 / 60 / 60);
            final String timeoutInForSinglePollInHours = Long.toString(timeoutForSinglePollMillis / 1000 / 60 / 60);

            if (dataSourceStatus.getAcquisitionUnit().getStatus().equals(AcquisitionUnitNew.Status.NOT_CONNECTED)) {
                dataSourcePackage.setAlert(true, "Acquisition unit " + dataSourceStatus.getAcquisitionUnit().getName() + " is not running or not communicating. Last Datasource status: " + dataSourcePackage.getLastStatus());
            } else {
                // if the package has success as the last status, make sure it is not too long ago
                if (dataSourcePackage.isSuccessful()) {
                    // Make sure the package is not unsuccessful for too long.
                    if (now - dataSourcePackage.getLastSuccessTime() > timeoutMillis) {
                        dataSourcePackage.setAlert(true, "Did not run for more than " + timeoutInHours + " hours");
                    }
                }
                // if the last (completed) runs all have failed statuses for enough polls, report alert.
                if (numOfSequentialErrors > numOfAttemps) {
                    // The package will only go on alert status if number of sequentials errors exceeds the limit...
                    dataSourcePackage.setAlert(true, "Failed for more than " + Integer.toString(numOfAttemps) + " times: " + dataSourcePackage.getLastStatus());
                }
                // if the package has started but not finished with in the allowed timeout, set alert.
                if (dataSourcePackage.getLastStartTime() > dataSourcePackage.getLastAcquisitionTime()
                        && (dataSourcePackage.getLastStartTime() + timeoutForSinglePollMillis < now)) {
                    // The package has started but not finished after the timeout
                    dataSourcePackage.setAlert(true, "Datasource Package has not finished after " + timeoutInForSinglePollInHours + " hours ");
                }
                // if foundation package reported the same device as package from other datasource
                //TODO - These package names should be consolidated somewhere.
                if (dataSourcePackage.getPackageName().equals("Inventory")
                        && datasourceIdsToDeviceIds.containsKey(dataSourceStatus.getDataSource().getId())){
                        dataSourcePackage.setAlert(true, "Inventory Package has reported conflicting device(s)");
                }
            }
        }
    }

    /**
     * Gets the map of conflicting datasources reporting about same switches, storages and hosts
     *
     * @param connection database connection
     * @return map of datasource id to the set of device ids
     */
    @SuppressWarnings({"unchecked"})
    private MapOfSets<Long, Long> findConflictingDataSources(final Connection connection) throws AcquisitionUnitException {
        final DataSourceTableNew dataSourceTable = new DataSourceTableNew(connection);
        return new MapOfSets<Long, Long>();
    }

    /**
     * Retrieves an array of {@link DataSourceNew} which represents data sources in the database where DATASOURCE.changeid = 0
     *
     * @param connection
     */
    @SuppressWarnings({"unchecked"})
    private List<DataSourceNew> findAllDataSources(final Connection connection) throws AcquisitionUnitException {
        try {
            final DataSourceTableNew dataSourceTable = new DataSourceTableNew(connection);
            return dataSourceTable.findAll();
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to get a list of data sources that were not stopped manually.", e);
        }
    }

    /**
     * Returns a Map with the relationship AcquisitionUnitName-To-AcquisitionStatus
     */
    private Map<Long, AcquisitionUnitNew> findAllAquisitionUnits(final Connection connection) throws AcquisitionUnitException {
        try {
            final Map<Long, AcquisitionUnitNew> map = new HashMap<Long, AcquisitionUnitNew>();
            final AcquisitionUnitTableNew acquisitionUnitTable = new AcquisitionUnitTableNew(connection);
            final List<AcquisitionUnitNew> acqUnits = acquisitionUnitTable.findAll();
            for (final AcquisitionUnitNew acqUnit : acqUnits) {
                map.put(Long.valueOf(acqUnit.getId()), acqUnit);
            }

            return map;
        }
        catch (TableException e) {
            throw new AcquisitionUnitException("Failed to read AcquisitionUnitTable", e);
        }
    }

}
