package com.onaro.sanscreen.acquisition.sessions;

import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceVersion;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.SessionContext;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

/**
 * Retrieves a list of data source versions from the META-INF in the data source JAR files. </p> This allows the
 * administrator to view the exact versions of the data sources installed.
 *
 * @author Ophir Horn
 * @version 1.0
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRED)
@RolesAllowed({"guest", "user", "admin"})
public class DataSourceVersionsSessionEJB implements DataSourceVersionsSession {

    public final static String ONARO_VERSION_CUSTOMER = "onaro-version-customer";
    public final static String ONARO_VERSION_TIME = "onaro-version-time";
    public final static String ONARO_VERSION_CHECKSUM = "onaro-version-checksum";

    static transient final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger("com.onaro.sanscreen.acquisition.sessions");

    private String datasourcesDeployDir;

    @Resource
    public void setSessionContext(SessionContext context) throws RemoteException {
        datasourcesDeployDir = System.getProperty("jboss.server.base.dir") + "/deployments";
        logger.info("setSessionContext called: datasourcesDeplyDir set to " + datasourcesDeployDir); // TODO: Remove after Debugging
    }

    /**
     * Retrieves an array with the versions of all the data source JARs in the
     * deploy directory
     */
    public DataSourceVersion[] getDataSourceVersions() throws RemoteException {
        List<DataSourceVersion> versions = new LinkedList<DataSourceVersion>();

        final String deployDirPath = datasourcesDeployDir + "/datasources.war";
        File deployDir = new File(deployDirPath);
        if (!deployDir.isDirectory()) {
            logger.error("Deploy dir path does not point to a valid dir: " + deployDirPath);
            return new DataSourceVersion[0]; // no data sources found - valid when running demo version 
        }

        // Iterate over all the JAR files in the directory
        final String[] jarFiles = deployDir.list(new JarFilter());
        for (final String jarName : jarFiles) {
            DataSourceVersion version = extractVersionFromJar(jarName);

            if (version != null) {
                versions.add(version);
            }
        }

        return versions.toArray(new DataSourceVersion[versions.size()]);
    }

    /**
     * Load a JAR file identified by name (from the server's data sources deploy dir) and extracts
     * the version information inside this JAR (from meta-inf and deployee.xml)
     *
     * @param jarName JAR file name to investigate
     *
     * @return DataSourceVersion object with version info or null if this JAR is not a DS jar
     */
    private DataSourceVersion extractVersionFromJar(final String jarName) throws RemoteException {
        try {
            final String jarFullPath = datasourcesDeployDir + "/datasources.war/" + jarName;

            DataSourceVersion version = new DataSourceVersion();
            version.setJarName(jarName);

            // Open the JAR file and extract the version from the manifest
            JarFile jarFile = new JarFile(jarFullPath);

            // TODO: Extract the data source name from the deployee file inside the JAR.

            // Get the version properties from the JAR manifest
            Manifest manifest = jarFile.getManifest();
            Map attributes = manifest.getMainAttributes();

            // If this JAR does not have the Onaro-specific manifest version properties
            // then it is probably a library JAR needed by the data source, included in
            // the ARCHIVE parameter of the MLet file.
            if (!attributes.containsKey(new Attributes.Name(ONARO_VERSION_CUSTOMER))) {
                return null;
            }

            version.setCustomer((String) attributes.get(new Attributes.Name(ONARO_VERSION_CUSTOMER)));
            version.setTimestamp((String) attributes.get(new Attributes.Name(ONARO_VERSION_TIME)));
            version.setChecksum((String) attributes.get(new Attributes.Name(ONARO_VERSION_CHECKSUM)));

            return version;
        }
        catch (IOException e) {
            logger.error("Failed to open JAR file: " + jarName, e);
            throw new RemoteException("Failed to open JAR file: " + jarName, e);
        }
    }

    /**
     * File name filter that accepts only JAR files
     */
    private static class JarFilter implements FilenameFilter {
        public boolean accept(File dir, String name) {
            return name.toLowerCase().endsWith(".jar");
        }
    }
}
