package com.onaro.sanscreen.acquisition.sessions;

import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceNew;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import org.slf4j.Logger;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.annotation.security.RunAs;
import javax.ejb.*;
import javax.jms.*;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import java.util.HashMap;
import java.util.Map;

@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRED)
@RolesAllowed({"guest", "acquisition", "admin"})
@RunAs("guest")
public class ForceAcqListenerEJB implements ForceAcqListenerSession {
    private static transient final org.slf4j.Logger logger =org.slf4j.LoggerFactory.getLogger("com.onaro.sanscreen.acquisition.sessions");
    private static final Map<Long, TopicSubscriber> subscribers = new HashMap<Long, TopicSubscriber>();

    @Resource
    protected SessionContext sessionContext;

    private String topicName = "";

    /**
     * Called by each of the acquisition units to listen for events.
     * <p/>
     * <p/>This is a mechanism that tries to overcome firewall issues. Each acquisition unit "polls" the server
     * for events. The server internally uses JMS to receive these events asynchronically.
     *
     * @param acqUnitID Identified the acquisition unit listening for this event
     * @return The event or "empty" if nothing happened.
     * @throws ServerInterfaceException
     */
    public String listen(final long acqUnitID) throws ServerInterfaceException {
        try {
            final TopicSubscriber subscriber = getSubscriber(acqUnitID);

            // Wait for a message with a timeout...
            final Message message = subscriber.receive(1000);
            if (message == null)
                return "empty";

            if (message instanceof ObjectMessage) {
                final ObjectMessage omessage = (ObjectMessage) message;
                final DataSourceNew dsd = (DataSourceNew) omessage.getObject();

                if (dsd.getAuId() == acqUnitID) {
                    // The event name is the data source name. The acq framework will 'force' all the packages for
                    // this data soruce.
                    final String text = dsd.getName()/* + "/" + "foundation"*/;
                    logger.debug("ForceAcqListenerEJB.listen() return from object message text : " + text);
                    return text;
                }
            }


            if (message instanceof TextMessage) {
                TextMessage tmessage = (TextMessage) message;
                final String text = tmessage.getText();
                logger.debug("ForceAcqListenerEJB.listen() returns from text message text : " + text);
                return text;
            }

            logger.error("ForceAcqListenerEJB.listen() failed to decide and returns an 'empty' message");
            return "empty";
        }
        catch (JMSException e) {
            logger.error("ForceAcqListenerEJB.listen() failed with exception " + e.getMessage(), e);
            throw new EJBException(e.getMessage(), e);
        }
    }

    public TopicSubscriber getSubscriber(final long acqUnitID) throws ServerInterfaceException {
        TopicSubscriber subscriber = subscribers.get(acqUnitID);
        if (subscriber != null)
            return subscriber;

        // get TopicConnectionFactory from JNDI using res-refs
        TopicConnectionFactory tcf;
        TopicConnection topicConnection;
        TopicSession topicSession;
        Topic topic;

        try {
            Context ctx = new InitialContext();

            Object ObjTopicConnectionFactory = ctx.lookup("java:/ConnectionFactory");
            tcf = (TopicConnectionFactory) PortableRemoteObject.narrow(ObjTopicConnectionFactory, TopicConnectionFactory.class);

            Object objTopic = ctx.lookup("java:/topic/ForceAcquisitionTopic");
            topic = (Topic) PortableRemoteObject.narrow(objTopic, Topic.class);
        }
        catch (NamingException ne) {
            logger.warn("Init problems, session may not work correctly", ne);
            throw new ServerInterfaceException(ne.getMessage(), ne);
        }

        try {
            topicConnection = tcf.createTopicConnection();
            topicConnection.start();

            topicSession = topicConnection.createTopicSession(false, Session.AUTO_ACKNOWLEDGE);

            // get topic name
            if (logger.isDebugEnabled())
                logger.debug("DiscoverySession listen on topic: " + topicName);

            try {
                subscriber = topicSession.createSubscriber(topic);
            }
            catch (JMSException e) {
                throw new ServerInterfaceException("Failed to create subsriber - " + e.getMessage(), e);
            }

            subscribers.put(acqUnitID, subscriber);

            return subscriber;
        }
        catch (JMSException jmse) {
            logger.warn("Init problems, force acquisition may not work correctly", jmse);
            throw new ServerInterfaceException(jmse.getMessage(), jmse);
        }
    }
}
