/*
 * Copyright (c) 2010 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.acquisition.sessions;

import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;
import com.onaro.sanscreen.acquisition.identify.AcqIdentifyException;
import com.onaro.sanscreen.acquisition.identify.AcqStorageIdentifyTable;
import com.onaro.sanscreen.acquisition.identify.AcqVolumeIdentifyTable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.Map;


@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRED)
@RolesAllowed({"guest", "acquisition", "admin", "scheduler"})
public class IdentifySessionEJB implements IdentifySession {

    @Resource(name = "jdbc/OnaroDataSource",
            mappedName = "java:/onaro_pool",
            type = DataSource.class,
            authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    /**
     * Static reference to the system logger.
     */
    static transient final Logger logger = LoggerFactory.getLogger("com.onaro.sanscreen.acquisition.sessions");

    /**
     * @param storageIdentifier - the storage identifier (could be the serial number)
     * @return The storage IP associated with the identifier. if the identifier does not exist return null.
     * @throws DataSourceException
     */
    public String getStorageIp(String storageIdentifier) throws AcqIdentifyException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            AcqStorageIdentifyTable acqStorageIdentifyTable = new AcqStorageIdentifyTable(connection);
            return acqStorageIdentifyTable.getStorageIp(storageIdentifier);
        } catch (TableException e) {
            logger.error("Failed to get storage ip for identifier " + storageIdentifier, e);
            throw new AcqIdentifyException("Failed to get storage ip for identifier " + storageIdentifier, e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }


    /**
     * Set the storage ip for a given storage identifier
     *
     * @param storageIdentifier - The storage identifier (could be the serial number or WWN )
     * @param storageIp         - The storage IP associated with the identifier
     * @throws AcqIdentifyException
     */
    public void setStorageIp(String storageIdentifier, String storageIp) throws AcqIdentifyException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            AcqStorageIdentifyTable acqStorageIdentifyTable = new AcqStorageIdentifyTable(connection);
            acqStorageIdentifyTable.setStorageIp(storageIdentifier, storageIp);
        } catch (TableException e) {
            logger.error("Failed to set storage ip " + storageIp + " for identifier " + storageIdentifier, e);
            throw new AcqIdentifyException("Failed to set storage ip " + storageIp + " for identifier " + storageIdentifier, e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }


    /**
     * @param storageIdentifier the storage isentifier
     * @param volumeIdentifier  the volume identifier
     * @return the volume name if does not exist return null
     */
    public String getVolumeName(String storageIdentifier, String volumeIdentifier) throws AcqIdentifyException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            AcqVolumeIdentifyTable acqVolumeIdentifyTable = new AcqVolumeIdentifyTable(connection);
            return acqVolumeIdentifyTable.findVolumeName(storageIdentifier, volumeIdentifier);
        } catch (TableException e) {
            logger.error("Failed to get voluem name storageIdentifier " + storageIdentifier + ", volumeIdentifier " + volumeIdentifier, e);
            throw new AcqIdentifyException("Failed to get voluem name storageIdentifier " + storageIdentifier + ", volumeIdentifier " + volumeIdentifier, e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }


    /**
     * Set the voluem name for a giving storage identifier and volume identifier
     *
     * @param storageIdentifier the storage isentifier
     * @param volumeIdentifier  the volume identifier
     * @param volumeName        the volume name
     */
    public void setVolumeName(String storageIdentifier, String volumeIdentifier, String volumeName) throws AcqIdentifyException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            AcqVolumeIdentifyTable acqVolumeIdentifyTable = new AcqVolumeIdentifyTable(connection);
            acqVolumeIdentifyTable.setVolumeName(storageIdentifier, volumeIdentifier, volumeName);
        } catch (TableException e) {
            logger.error("Failed to set voluem name " + volumeName + " for storageIdentifier " + storageIdentifier + ", volumeIdentifier " + volumeIdentifier, e);
            throw new AcqIdentifyException("Failed to set voluem name " + volumeName + " for storageIdentifier " + storageIdentifier + ", volumeIdentifier " + volumeIdentifier, e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }


    /**
     * @param storageIdentifier the storage isentifier
     * @return a Map between volume identifier and volume Name if the storage does not exist return empty HashMap;
     */
    public Map<String, String> getVolumeIdentifierToName(String storageIdentifier) throws AcqIdentifyException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            AcqVolumeIdentifyTable acqVolumeIdentifyTable = new AcqVolumeIdentifyTable(connection);
            return acqVolumeIdentifyTable.getVolumeName(storageIdentifier);
        } catch (TableException e) {
            logger.error("Failed to get voluem name storageIdentifier " + storageIdentifier, e);
            throw new AcqIdentifyException("Failed to get voluem name storageIdentifier " + storageIdentifier, e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     * Set the volume identifier to name for a giving storage
     *
     * @param storageIdentifier      the storage isentifier
     * @param volumeIdentifierToName a Map between volume identifier and volume Name
     */
    public void setVolumeIdentifierToName(String storageIdentifier, Map volumeIdentifierToName) throws AcqIdentifyException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            AcqVolumeIdentifyTable acqVolumeIdentifyTable = new AcqVolumeIdentifyTable(connection);
            acqVolumeIdentifyTable.setVolumeIdentifierToName(storageIdentifier, volumeIdentifierToName);
        } catch (TableException e) {
            logger.error("Failed to get voluem name storageIdentifier " + storageIdentifier, e);
            throw new AcqIdentifyException("Failed to get voluem name storageIdentifier " + storageIdentifier, e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     * Delete all the data from the tables
     */
    public void deleteAll() {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            new AcqVolumeIdentifyTable(connection).deleteAll();
            new AcqStorageIdentifyTable(connection).deleteAll();
        } catch (TableException e) {
            logger.error("Failed to delete all data", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
