/*
 * Copyright (c) 2010 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.acquisition.sessions;

import javax.naming.NamingException;

import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;
import com.onaro.sanscreen.acquisition.vobject.AcquisitionUnitStatus;
import com.onaro.sanscreen.acquisition.vobject.Status;
import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionUnit;
import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionUnitAudit;
import com.onaro.sanscreen.server.sessions.SessionFactory;
import com.onaro.sanscreen.server.tables.acquisition.AcquisitionUnitTable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import java.rmi.RemoteException;
import java.sql.Connection;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;


@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRED)
@RolesAllowed({"guest", "scheduler", "acquisition", "admin"})
public class LeaseManagerSessionEJB implements LeaseManagerSession {
    private static transient final Logger logger =
            LoggerFactory
            .getLogger("com.onaro.sanscreen.acquisition.sessions");

    /**
     * This method will return the status of the Acquisition unit
     * <p/>
     * Used for unit testing only. See {@link #checkAcquisitionUnitLease} to check if the lease of an acquisition unit
     * is still valid
     */
    public int checkAcquisitionUnitStatus(final String acquisitionUnitName) throws LeaseManagerException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            AcquisitionUnitTable acquisitionUnitTable = new AcquisitionUnitTable(connection);
            AcquisitionUnit acqUnit = acquisitionUnitTable.findByName(acquisitionUnitName);
            return acqUnit.getStatusWithLeaseCheck();
        }
        catch (TableException e) {
            logger.error("Failed to checkAcquisitionUnitStatus acquisitionUnitName=" + acquisitionUnitName, e);
            throw new LeaseManagerException("Could not find AcquisitionUnit " + acquisitionUnitName + " for status", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     * This method will return the status of the Acquisition unit
     *
     * @return true if the acquisition unit is still active (and lease is valid)
     */
    public boolean checkAcquisitionUnitLease(final String acquisitionUnitName) throws LeaseManagerException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            AcquisitionUnitTable acquisitionUnitTable = new AcquisitionUnitTable(connection);
            AcquisitionUnit acqUnit = acquisitionUnitTable.findByName(acquisitionUnitName);
            if (acqUnit == null) {
                logger.error("Could not find AcquisitionUnit " + acquisitionUnitName + " for status");
                throw new LeaseManagerException("Could not find AcquisitionUnit " + acquisitionUnitName + " for status");
            }
            return acqUnit.isActive();
        }
        catch (TableException e) {
            logger.error("Failed to checkAcquisitionUnitLease acquisitionUnitName=" + acquisitionUnitName, e);
            throw new LeaseManagerException(e.getMessage(), e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }


    /**
     * This method will return acqusition unit name list
     */
    public Collection getAcquisitionUnits() throws LeaseManagerException, RemoteException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnectionNoEjb();
            AcquisitionUnitTable acquisitionUnitTable = new AcquisitionUnitTable(connection);
            List<AcquisitionUnitStatus> result = new LinkedList<AcquisitionUnitStatus>();

            Collection aulCollection = acquisitionUnitTable.findAll();
            for (Iterator iterator = aulCollection.iterator(); iterator.hasNext();) {
                try {
                    AcquisitionUnit acqUnit = (AcquisitionUnit) iterator.next();
                    AcquisitionUnitStatus status = getAcquistionUnitStatus(acqUnit);
                    result.add(status);
                }
                catch (LeaseManagerException e) {
                    if (logger.isDebugEnabled())
                        logger.debug("", e);

                    // Continue...
                }
            }

            return result;
        }
        catch (TableException e) {
            logger.error("Failed to getAcquisitionUnits", e);
            throw new LeaseManagerException("Could not read acq unit data", e);
        }
        finally {
            ConnectionHelper.close(connection);
        }
    }

    private AcquisitionUnitStatus getAcquistionUnitStatus(AcquisitionUnit data) throws LeaseManagerException, RemoteException {
        AcquisitionUnitStatus status = new AcquisitionUnitStatus(data.getName(), data.getIp(), data.getStatusWithLeaseCheck());

        // if the unit status is warn fill its data
        //        if (data.getStatus() == AcquisitionUnit.WARN)
        fillInfoList(status.getAcquisitionUnitName(), status.getInfoList());

        return status;
    }

    private void fillInfoList(String name, List infoList) throws LeaseManagerException, RemoteException {
        try {
            for (Iterator iterator = getAcquisitionUnitAuditSession().getAcquisitionUnitStatus(name).iterator(); iterator.hasNext();)
            {
                AcquisitionUnitAudit data = (AcquisitionUnitAudit) iterator.next();
                infoList.add(new Status(data.getTime(), data.getStatus()));
            }
        } catch (AcquisitionUnitException e) {
            logger.error("Couldn't get acquisition unit " + name + " info", e);
            throw new LeaseManagerException("Couldn't get acquisition unit " + name + " info", e);
        }
    }


    private AcquisitionUnitAuditSession getAcquisitionUnitAuditSession() throws RemoteException {
        try {
            return (AcquisitionUnitAuditSession) SessionFactory.createSession(AcquisitionUnitAuditSession.JNDI_NAME);
        } catch (NamingException e) {
            logger.error("Failed to createSession " + AcquisitionUnitAuditSession.JNDI_NAME, e);
            throw new RemoteException("Failed to createSession " + AcquisitionUnitAuditSession.JNDI_NAME, e);
        }
    }
}

