/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.client.simulator;

import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.server.TransactionContext;
import com.netapp.oci.server.TransactionContextException;
import com.netapp.sanscreen.server.api.OriginatorUpdateClientUtils;
import com.onaro.sanscreen.server.originatorupdate.OriginatorCacheChanges;
import com.onaro.sanscreen.server.sessions.OriginatorUpdateSession;
import com.onaro.sanscreen.server.sessions.StartupSession;
import com.onaro.sanscreen.server.tables.idgenerator.IdGenerator;
import com.onaro.sanscreen.util.Schema;
import com.onaro.sanscreen.util.xml.XMLUtil;
import com.onaro.util.sessions.SessionCache;
import com.onaro.util.sessions.SessionCacheException;

import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.sql.Connection;

public class Simulator {

    private OriginatorUpdateSession updateSession;
    private StartupSession startUpSession;

    private Connection connection;
    private TransactionContext transactionContext;

    private static final boolean DOUBLE_SUBMISSION_CHECK = false;

    private static Simulator instance;

    public Simulator() throws Exception {
        try {
            updateSession = createSession(OriginatorUpdateSession.JNDI_NAME);
            startUpSession = createSession(StartupSession.JNDI_NAME);

        } catch (Exception e) {
            throw new Exception("Failed to init simulator", e);
        }
    }

    public static Simulator getInstance() throws Exception {
        if (instance == null) {
            instance = new Simulator();
        }

        return instance;
    }

    public void resetIndex() throws Exception {
        IdGenerator.getInstance().reset();
    }

    /**
     * POSTs a file over HTTP to OriginatorUpdateServlet
     * @param file file to POST
     * @param schema schema of the XML file
     * @throws Exception any exception that happens here
     * @author Daniel Nuriyev
     */
    public void sendHttp(File file, Schema schema) throws Exception {

        if(!file.exists()) {
            throw new FileNotFoundException(file.getAbsolutePath());
        }

        BufferedReader r = null;

        try{
            InputStream fis = new FileInputStream(file);
            Reader fr = new InputStreamReader(fis, "UTF-8");
            r = new BufferedReader(fr);
            StringBuilder s = new StringBuilder((int)file.length());
            String line = null;
            while ((line = r.readLine()) != null) {
                s.append(line);
            }
            sendHttp(s.toString(), schema);
        } finally {
            try {
                if(r != null) {
                    r.close();
                }
            } catch(IOException ignore) {}
        }
    }

    /**
     * POST XML over HTTP to OriginatorUpdateServlet
     * @param xml XML to POST
     * @param schema schema of the XML
     * @throws Exception any exception that happens here
     * @author Daniel Nuriyev
     */
    public void sendHttp(String xml, Schema schema) throws Exception {

        OutputStreamWriter w = null;
        BufferedReader r = null;

        try{

            String originID = XMLUtil.getElementVal(xml, "Originator");

            String url = OriginatorUpdateClientUtils.createURL("http://localhost/", schema,  originID);

            URL urlObj = new URL(url);

            HttpURLConnection c = (HttpURLConnection)urlObj.openConnection();
            c.setDoOutput(true);

            w = new OutputStreamWriter(c.getOutputStream(), "UTF-8");
            w.write(xml);
            w.close();

            int code = c.getResponseCode();
            InputStream in = null;
            if(code >= 200 && code < 300){
                in = c.getInputStream();
            } else {
                in = c.getErrorStream();
            }
            r = new BufferedReader(new InputStreamReader(in));
            String line = null;
            StringBuilder inBuffer = new StringBuilder();
            while ((line = r.readLine()) != null) {
                inBuffer.append(line);
            }
            String body = inBuffer.toString();

            OriginatorUpdateClientUtils.assertResponse(code, null, body, body);

        } finally {
            try{ w.close(); }catch(Exception e){}
            try{ r.close(); }catch(Exception e){}
        }

    }

    public OriginatorCacheChanges send(String xml, Schema schema) throws Exception {
        OriginatorCacheChanges update = updateSession.getUpdate(xml, schema);// sendNoEjb(xml)
        if (DOUBLE_SUBMISSION_CHECK) {
            OriginatorCacheChanges reUpdate = updateSession.getUpdate(xml, schema);// sendNoEjb(xml)
            if (reUpdate.isChanged()) {
                System.out.println();
                System.out.println(xml);
                System.out.println();
                System.out.println(reUpdate);
                throw new IllegalStateException("There should be no change!, \r\n" + reUpdate);
            }
        }
        return update;
    }

    public boolean removeUpdate(Long originId) throws Exception {
        return updateSession.removeOriginUpdate(originId).isChanged();
    }


    public void clearAll() throws Exception {
        try {
            resetIndex();
            startUpSession.clearAllData();
            updateSession.resetAllCache();

            closeConnection();
            clearTransactionContext();
            TransactionContext.resetLastUpdateTime();

        } catch (Exception e) {
            throw new Exception("Failed to clear all data", e);
        }
    }

    /**
     * Closes and nils the Simulator's TransactionContext to force the next call to {@link #newTransactionContext()} to
     * create a new TransactionContext.
     */
    public void clearTransactionContext() {
        if (transactionContext != null) {
            TransactionContext.close(transactionContext);
            transactionContext = null;
        }
    }

    public StartupSession getStartUpSession() {
        return startUpSession;
    }
    public OriginatorUpdateSession getUpdateSession() {
        return updateSession;
    }

    public Connection getConnection() throws TableException {
        if (connection == null) {
            connection = ConnectionHelper.newConnectionNoEjb();
        }
        return connection;
    }

    public TransactionContext newTransactionContext() throws TransactionContextException {
        clearTransactionContext();
        if (transactionContext == null) {
            transactionContext = new TransactionContext(getClass());
        }

        return transactionContext;
    }

    public TransactionContext getTransactionContext() throws TransactionContextException {
        if (transactionContext == null) {
            newTransactionContext();
        }
        return transactionContext;
    }

    /**
     * Closes the Simulator's connection. A new connection will be opened the next time one getConnection() is called,
     * so it is best to call this method as the last thing done with the Simulator.
     */
    public void closeConnection() {
        if (connection != null) {
            ConnectionHelper.close(connection);
            connection = null;
        }
    }

    public void resetCache() {
    }

    private static SessionCache sessionCache = null;

    static public synchronized <E> E createSession(final String sessionJndiName) throws SessionCacheException {
        if (sessionCache == null) {
            sessionCache = new SessionCache(Simulator.class.getClassLoader(), "admin", "admin123", 0);
        }

        return (E) sessionCache.getSession(sessionJndiName);
    }
}
