/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.client.simulator;

import com.netapp.oci.platform.common.util.Assert;
import com.onaro.sanscreen.client.simulator.annotations.*;
import javassist.*;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

/**
 * This class generates the simulatormapping in memory and it then can be fed into Castor. The input is
 * the root of the simulator object (i.e., NasReport). The Castor's xml binding file will be generated by using and recursively
 * traversing the entire hierarchy.
 * <p/>
 * The objects you write need to have:
 * (1) XmlClass annotation
 * (2) Every field that has XML name different in the xml, should declare an XmlField annotation.
 *
 * @author yarom.gabay
 */
public class SimulatorMappingGenerator {
    CtClass rootType;
    Queue<CtClass> classQueue = new LinkedList<CtClass>();
    Set<CtClass> visitedClasses = new HashSet<CtClass>();
    ClassPool classPool = ClassPool.getDefault();


    public SimulatorMappingGenerator(Class<?> type) throws NotFoundException {
        this.rootType = classPool.get(type.getName());
        this.classQueue.add(rootType);
    }

    public String getSimulatorMapping() throws Exception {

        StringBuilder sb = new StringBuilder();
        sb.append("<!DOCTYPE databases PUBLIC \"-//EXOLAB/Castor Mapping DTD Version 1.0//EN\" \"http://castor.exolab.org/mapping.dtd\">");
        sb.append("\n");
        sb.append("<mapping>");
        sb.append("\n");
        while (classQueue.size() > 0) {
            CtClass type = classQueue.peek();
            classQueue.remove();
            visitedClasses.add(type);
            generateObjectBlock(type, sb);
        }

        sb.append("</mapping>");

        return sb.toString();
    }

    private void generateObjectBlock(CtClass type, StringBuilder sb) throws Exception {
        sb.append(String.format("<class name=\"%s\">", type.getName()));
        sb.append("\n");
        sb.append(String.format("<map-to xml=\"%s\"/>", getXmlName(type)));
        sb.append("\n");

        // examine the fields of the class.  All the fields becomes an attribute or nested element unless 'xmlIgnore'
        // annotation is specified for the field.
        for (CtField field : classPool.get(type.getName()).getDeclaredFields()) {
            if (!shouldIncludeField(field)) {
                continue;
            }
            if (isCollection(field)) {
                CtClass paramType = getTypeParameterOfCollection(field);
                checkAndAddToQueue(paramType);                
                appendFieldElement(sb, field.getName(), getXmlTypeOfType(paramType), true, getXmlNameOfField(field),
                        isElement(field));
            } else {
                checkAndAddToQueue(field.getType());
                appendFieldElement(sb, field.getName(), getXmlTypeOfField(field), false, getXmlNameOfField(field),
                        isElement(field));
            }            
        }

        // Examine the public methods for the class.  Only the methods that has 'XmlMethod' annotation specified gets 
        // consideration to be added either as an attribute or an element.
        for (CtMethod method: classPool.get(type.getName()).getDeclaredMethods()) {
            if (!shouldIncludeMethod(method)) {
                continue;
            }

            CtClass returnType = method.getReturnType();
            checkAndAddToQueue(returnType);

            String propertyName = makeFirstCharacterLowerCase(method.getName().substring(3));
            // we do not support collections for methods yet.
            appendFieldElement(sb, propertyName, getXmlTypeOfType(returnType), false, getXmlNameOfMethod(method), isElement(method));
        }

        sb.append("</class>");
        sb.append("\n");
    }

    private void appendFieldElement(StringBuilder sb, String propertyName, String typeShortName, boolean isCollection,
                                    String xmlName, boolean isElement) {
        sb.append(String.format("<field name=\"%s\" type=\"%s\"" , propertyName, typeShortName));
        if (isCollection) {
            sb.append(" collection=\"vector\"");
        }
        sb.append(">");
        sb.append("\n");
        sb.append("\t");
        sb.append(String.format("<xml name=\"%s\" node=\"%s\"/>", xmlName, isElement ? "element" : "attribute"));
        sb.append("\n");
        sb.append("</field>");
        sb.append("\n");
    }

    private boolean shouldIncludeMethod(CtMethod method) throws ClassNotFoundException {
        return !Modifier.isStatic(method.getModifiers()) && getAnnotationOfMethod(method, XmlMethod.class) != null;
    }

    private boolean shouldIncludeField(CtField field) throws ClassNotFoundException {
        return !Modifier.isStatic(field.getModifiers()) && getAnnotationOfField(field, XmlIgnore.class) == null;
    }

    /**
     * Get the type parameter of the collection. E.g., Vector<String> ==> String.
     */
    private CtClass getTypeParameterOfCollection(CtField field) throws ClassNotFoundException, NotFoundException, CannotCompileException, NoSuchFieldException {
        Class<?> declaringClass = getClass().getClassLoader().loadClass(field.getDeclaringClass().getName());
        String genericType = declaringClass.getDeclaredField(field.getName()).getGenericType().toString();
        String[] strings = genericType.split("<|>");
        Assert.isTrue(strings != null && strings.length > 1, "Failed parsing for generic type '" + genericType + "' for the value '" + field.getName() + "'");
        //noinspection ConstantConditions
        return classPool.get(strings[1]);
    }

    /**
     * @return whether this field is a collection (for now just Vector is considered)
     * @throws NotFoundException
     */
    private boolean isCollection(CtField field) throws NotFoundException {
        return field.getType().getName().equals("java.util.Vector");
    }

    /**
     * If this type should also geenrate mapping information add it to the processing queue.
     *
     * @param type
     * @throws ClassNotFoundException
     */
    private void checkAndAddToQueue(CtClass type) throws ClassNotFoundException {
        if (getAnnotationOfType(type, XmlClass.class) != null && !visitedClasses.contains(type)) {
            classQueue.add(type);
        }
    }


    /**
     * Determines whether this field should generate a 'attribute' or 'element' mapping
     *
     * @param field
     * @return
     * @throws ClassNotFoundException
     * @throws NotFoundException
     */
    private boolean isElement(CtField field) throws ClassNotFoundException, NotFoundException {
        XmlFieldType xmlField = getAnnotationOfField(field, XmlFieldType.class);
        if (xmlField != null && !xmlField.value().equals(FieldType.UNSPECIFIED)) {
            return xmlField.value().equals(FieldType.ELEMENT);
        }

        String fieldType = field.getType().getName();
        return isElement(fieldType);
    }

    private boolean isElement(String type) {
        return !type.equals("java.lang.String")
                && !(type.equals("int") || type.equals("java.lang.Integer"))
                && !(type.equals("long") || type.equals("java.lang.Long"))
                && !(type.equals("float") || type.equals("java.lang.Float"))
                && !(type.equals("double") || type.equals("java.lang.Double"))
                && !(type.equals("boolean") || type.equals("java.lang.Boolean"))
                && !type.equals("java.util.Date");
    }

    private boolean isElement(CtMethod method) throws ClassNotFoundException, NotFoundException {
        String returnType = method.getReturnType().getName();
        return isElement(returnType);
    }

    private String getXmlTypeOfField(CtField field) throws NotFoundException {
        return getXmlTypeOfType(field.getType());
    }

    /**
     * Determines the XML type this field should be mapped to. E.g. java.lang.String ==> string.
     *
     * @param type
     * @return
     */
    private String getXmlTypeOfType(CtClass type) {
        String fieldTypeName = type.getName();
        if (fieldTypeName.equals("java.lang.String")) {
            return "string";
        } else if (fieldTypeName.equals("int") || fieldTypeName.equals("java.lang.Integer")) {
            return "integer";
        } else if (fieldTypeName.equals("long") || fieldTypeName.equals("java.lang.Long")) {
            return "long";
        } else if (fieldTypeName.equals("boolean") || fieldTypeName.equals("java.lang.Boolean")) {
            return "boolean";
        } else if (fieldTypeName.equals("java.util.Date")) {
            return "date";
        } else {
            return fieldTypeName;
        }
    }


    /**
     * returns the XML name of the type. Determined by the XmlClass annotation.
     *
     * @param type
     * @return XML name of the type.     *
     * @throws ClassNotFoundException
     */
    private String getXmlName(CtClass type) throws ClassNotFoundException {
        if (type.equals(rootType)) {
            return "root";
        }

        XmlClass annotation = getAnnotationOfType(type, XmlClass.class);
        if (annotation != null && annotation.xmlBinding().length() > 0) {
            return annotation.xmlBinding();
        } else {
            return type.getSimpleName();
        }
    }

    /**
     * XML name of the field. determined by XmlField. The default is a capitalized version of the field (only first letter). E.g., originatorID ==> OriginatorId.
     *
     * @param field
     * @return
     * @throws Exception
     */
    private String getXmlNameOfField(CtField field) throws Exception {
        XmlBinding annotation = getAnnotationOfField(field, XmlBinding.class);
        if (annotation != null && annotation.value().length() > 0) {
            return annotation.value();
        } else {
            if (isCollection(field)) {
                CtClass paramType = getTypeParameterOfCollection(field);
                return paramType.getSimpleName();
            } else {
                return capitalizeFirstCharacter(field.getName());
            }
        }
    }

    private String getXmlNameOfMethod(CtMethod method) throws ClassNotFoundException {
        XmlMethod annotation = getAnnotationOfMethod(method, XmlMethod.class);
        if (annotation != null && annotation.xmlBinding().length() > 0) {
            return annotation.xmlBinding();
        } else {
            return method.getName().substring(3);
        }
    }

    private String capitalizeFirstCharacter(String name) {
        return name.substring(0, 1).toUpperCase() + name.substring(1);
    }

    private String makeFirstCharacterLowerCase(String name) {
        return name.substring(0, 1).toLowerCase() + name.substring(1);
    }

    private <T> T getAnnotationOfField(CtField field, Class<T> annotationClass) throws ClassNotFoundException {
        Object[] annotations = field.getAnnotations();
        return getAnnotation(annotationClass, annotations);
    }

    private <T> T getAnnotationOfType(CtClass type, Class<T> annotationClass) throws ClassNotFoundException {
        Object[] annotations = type.getAnnotations();
        return getAnnotation(annotationClass, annotations);
    }

    private <T> T getAnnotationOfMethod(CtMethod method, Class<T> annotationClass) throws ClassNotFoundException {
        Object[] annotations = method.getAnnotations();
        return getAnnotation(annotationClass, annotations);
    }

    private <T> T getAnnotation(Class<T> annotationClass, Object[] annotations) {
        if (annotations.length > 0 && annotationClass.isInstance(annotations[0])) {
            return annotationClass.cast(annotations[0]);
        } else {
            return null;
        }
    }

    /**
     * This method is used to get all the unconverted mappings. Eventually, when we convert all objects this will go away.
     *
     * @return
     */
    private String getLegacyMapping() throws IOException {
        BufferedReader stream = null;
        try {
            stream = new BufferedReader(new InputStreamReader(Simulator.class.getResourceAsStream("/simulator/simulatormapping-legacy.xml")));

            StringBuilder contents = new StringBuilder();

            String line = stream.readLine();
            while (line != null) {
                contents.append(line).append('\n');
                line = stream.readLine();
            }

            return contents.toString();
        } finally {
            if (stream != null) {
                stream.close();
            }
        }
    }
}
