/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.server.dataobject;

import com.netapp.oci.platform.common.ServerConstant;
import com.onaro.commons.types.AbstractType;
import com.onaro.sanscreen.server.tables.idgenerator.OidIdGenerator;

import java.util.Map;

public class Utils {

    /**
     * Standard database fields are 255 in length
     */
    public static final int VARCHAR_255_LIMIT = 255;

    public static String getValue(String key, Map<String, String> properties, String defaultValue) {
        String val = truncateString(properties.get(key));
        if (val != null) {
            return val;
        }
        return defaultValue;
    }

    public static AbstractType getValueOfInt(String key, Map<String, String> properties, AbstractType defaultAbstractType, Class<? extends AbstractType> abstractTypeClass) {
        String val = properties.get(key);
        if (val == null) {
            return defaultAbstractType;
        }

        int iVal = Integer.parseInt(val);
        AbstractType typeById = AbstractType.getById(abstractTypeClass, iVal);
        if (typeById == null) {
            throw new IllegalArgumentException("Failed to " + abstractTypeClass.getName() + " from int value " + val);
        }
        return typeById;
    }

    public static AbstractType getValueOfString(String key, Map<String, String> properties, Class<? extends AbstractType> abstractTypeClass) {
        String val = properties.get(key);
        if (val == null) {
            throw new IllegalArgumentException("Failed to get " + key + " value from properties:" + properties);
        }

        AbstractType typeByName = AbstractType.getByName(abstractTypeClass, val);
        if (typeByName == null) {
            throw new IllegalArgumentException("Failed to " + abstractTypeClass.getName() + " from string value " + val);
        }
        return typeByName;
    }

    public static AbstractType getValueOfString(String key, Map<String, String> properties, AbstractType defaultAbstractType, Class<? extends AbstractType> abstractTypeClass) {
        String val = properties.get(key);
        AbstractType type = defaultAbstractType;
        if (val != null) {
            AbstractType typeByName = AbstractType.getByName(abstractTypeClass, val);
            if (typeByName != null) {
                type = typeByName;
            }
        }
        return type;
    }

    public static String getUpperCaseValue(String key, Map<String, String> properties, String defaultValue) {
        String val = truncateString(properties.get(key));
        if (val != null) {
            return val.toUpperCase();
        }

        if (defaultValue == null) {
            return null;
        }

        return defaultValue.toUpperCase();
    }

    public static String getUpperCaseBigStringValue(String key, Map<String, String> properties, String defaultValue) {
        String val = properties.get(key);
        if (val != null) {
            return val.toUpperCase();
        }

        if (defaultValue == null) {
            return null;
        }

        return defaultValue.toUpperCase();
    }

    public static String getUpperCaseValue(String key, Map<String, String> properties) {
        String val = getUpperCaseValue(key, properties, null);
        if (val == null) {
            throw new IllegalArgumentException("Failed to get " + key + " value from properties:" + properties);
        }
        return val;
    }


    public static int getValue(String key, Map<String, String> properties, int defaultValue) {
        String val = properties.get(key);
        if (val != null) {
            return Integer.parseInt(val);
        }
        return defaultValue;
    }

    public static int getValue(String key, Map<String, String> properties) {
        String val = properties.get(key);
        if (val == null) {
            throw new IllegalArgumentException("Failed to get " + key + " value from properties:" + properties);
        }
        return Integer.parseInt(val);
    }

    public static String getValueAsString(String key, Map<String, String> properties) {
        String val = truncateString(properties.get(key));
        if (val == null) {
            throw new IllegalArgumentException("Failed to get " + key + " value from properties:" + properties);
        }
        return val;
    }

    public static long getValueAsLong(String key, Map<String, String> properties) {
        String val = properties.get(key);
        if (val == null) {
            throw new IllegalArgumentException("Failed to get " + key + " value from properties:" + properties);
        }
        return Long.parseLong(val);
    }

    public static float getValueAsFloat(String key, Map<String, String> properties) {
        String val = properties.get(key);
        if (val == null) {
            throw new IllegalArgumentException("Failed to get " + key + " value from properties:" + properties);
        }
        return Float.parseFloat(val);
    }

    public static float getValueAsFloat(String key, Map<String, String> properties, float defaultValue) {
        String val = properties.get(key);
        if (val == null) {
            return defaultValue;
        }
        float f = Float.parseFloat(val);
        if (Float.isNaN(f)) {
            return defaultValue;
        } else {
            return f;
        }
    }

    public static long getValue(String key, Map<String, String> properties, long defaultValue) {
        String val = properties.get(key);
        if (val != null) {
            return Long.parseLong(val);
        }
        return defaultValue;
    }

    public static float getValue(String key, Map<String, String> properties, float defaultValue) {
        String val = properties.get(key);
        if (val != null) {
            return Float.parseFloat(val);
        }
        return defaultValue;
    }

    public static boolean getValue(String key, Map<String, String> properties, boolean defaultValue) {
        String str = properties.get(key);
        if (str == null) {
            return defaultValue;
        }

        if (str.equals("0")) {
            return false;
        }
        str = str.toLowerCase();
        if (str.equals("false")) {
            return false;
        }
        if (str.equals("no")) {
            return false;
        }
        // if it's non false, it's true by definition
        return true;
    }

    /**
     * @param oid -
     * @return - Constant.LONG_NULL_ID if oid is null or empty and the OID-id otherwise.
     */
    public static Long findOidValue(String oid) {
        if (oid == null || oid.length() == 0) {
            return ServerConstant.LONG_NULL_ID;
        } else {
            return OidIdGenerator.getObjId(oid);
        }
    }

    /**
     * @param oid - input oid
     * @return - get the OID-id out of oid.
     * @throws IllegalArgumentException - on failure to find id.
     */
    public static Long getOidValue(String oid) {
        Long result = findOidValue(oid);
        if (ServerConstant.LONG_NULL_ID.equals(result)) {
            throw new IllegalArgumentException("Failed to get OID-id from oid " + oid);
        }
        return result;
    }

    public static String getStringValue(String key, Map<String, String> properties) {
        String val = truncateString(properties.get(key));
        if (val == null) {
            throw new IllegalArgumentException("Failed to get value " + key + " from properties :" + properties);
        }
        return val;
    }

    public static String getBigStringValue(String key, Map<String, String> properties) {
        String val = properties.get(key);
        if (val == null) {
            throw new IllegalArgumentException("Failed to get value " + key + " from properties :" + properties);
        }
        return val;
    }

    public static String findStringValue(String key, Map<String, String> properties) {
        return truncateString(properties.get(key));
    }

    public static boolean getBooleanValue(String key, Map<String, String> properties) {
        String value = properties.get(key);
        if (value == null) {
            throw new IllegalArgumentException("Failed to get value " + key + " from properties :" + properties);
        }
        return Boolean.valueOf(value).booleanValue();
    }

    private static String truncateString(String valueStr) {
        if (valueStr != null && valueStr.length() > VARCHAR_255_LIMIT) {
            return valueStr.substring(0, VARCHAR_255_LIMIT);
        }
        return valueStr;
    }
}
