package com.onaro.sanscreen.server.dataobject.acquisition;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Set;

/**
 * Data object for AcquisitionUnit.
 */
public class AcquisitionUnitNew implements Serializable {
    private static final long serialVersionUID = 1L;

    private final String name;
    private final long leaseContract;

    private String ip = null;
    private Status status = null;
    private long id = -1;
    private long nextLeaseRenewal = 0L;

    private final Set<AcqUnitAttribute> acqUnitAttributes = new HashSet<AcqUnitAttribute>();

    /**
     * Constructor - obsolete! only for legacy uses
     *
     * @param auOld
     *
     * @deprecated
     */
    public AcquisitionUnitNew(AcquisitionUnit auOld) {
        name = auOld.getName();
        ip = auOld.getIp();
        status = Status.fromInt(auOld.getStatus());
        leaseContract = auOld.getLeasePeriod();
        nextLeaseRenewal = auOld.getNextLeaseRenewal();
    }

    /**
     * Constructor
     *
     * @param name Name of acquisition unit (unique)
     * @param ip IP address
     * @param leaseContract Contract (millis) between server and acq unit for liveness reports.
     * If the acq unit did not report for more than the contract time, the acq unit is assumed dead
     */
    public AcquisitionUnitNew(final String name, final String ip, final long leaseContract) {
        this.name = name;
        this.ip = ip;
        this.leaseContract = leaseContract;
    }

    public AcquisitionUnitNew(final long id,
                              final String name,
                              final String ip,
                              final Status status,
                              final long leasePeriod,
                              final long nextLeaseRenewal) {
        this.name = name;
        this.ip = ip;
        this.leaseContract = leasePeriod;

        setId(id);
        setStatus(status);
        setNextLeaseRenewal(nextLeaseRenewal);
    }

    public long getId() {
        return id;
    }

    public void setId(final long id) {
        this.id = id;
    }

    public String getName() {
        return this.name;
    }

    public String getIp() {
        return this.ip;
    }

    /**
     * The status takes into account the status of the acquisition unit (based on the data base row) and the lease
     * status
     */
    public Status getStatus() {
        return status;
    }

    public Set<AcqUnitAttribute> getAcqUnitAttributes() {
        return acqUnitAttributes;
    }

    public void setStatus(Status status) {
        this.status = status;
    }

    public long getLeasePeriod() {
        return leaseContract;
    }

    public long getNextLeaseRenewal() {
        return this.nextLeaseRenewal;
    }

    public void setNextLeaseRenewal(long nextLeaseRenewal) {
        this.nextLeaseRenewal = nextLeaseRenewal;
    }

    public AcquisitionUnit getAcquisitionUnit() {
        return new AcquisitionUnit(name, ip, status.toInt(), (int) leaseContract, nextLeaseRenewal);
    }

    public void setIp(final String acqUnitIp) {
        this.ip = acqUnitIp;
    }

    public String toString() {
        return "AcauisitionUnitNew: id=" + id + ", name=" + name + ", ip=" + ip;
    }

    /**
     * @return true if the acquisition unit is active (status is active and lease not expired)
     */
    public boolean isActive() {
        if (status == Status.NOT_CONNECTED)
            return false;

        if (nextLeaseRenewal - System.currentTimeMillis() < 0)
            return false;

        return true;
    }

    public String calculateDataSourceState(boolean active, int manual) {
        if (!active && manual == 2)
            return "Stopped";

        if (!isActive())
            return "Acquisition not running";

        if (active && manual == 1)
            return "Running";

        if (!active && manual == 1)
            return "Starting";

        if (active && manual == 2)
            return "Stopping";

        return "";
    }

    public static enum Status implements Serializable {
        NOT_CONNECTED("Not connected"),
        CONNECTED("ok"),
        WARN("WARNING");

        private final static long serialVersionUID = 1L;
        private String status;

        Status(final String status) {
            this.status = status;
        }

        public String getStatus() {
            return status;
        }

        public String toString() {
            return status;
        }

        public int toInt() {
            switch (this) {
                case NOT_CONNECTED:
                    return -1;
                case CONNECTED:
                    return 1;
                case WARN:
                    return 2;
                default:
                    throw new IllegalStateException("Unknown AcquisitionUnit status : " + status);
            }
        }

        public static Status parseString(final String str) {
            for (Status s : Status.values()) {
                if (s.status.equals(str))
                    return s;
            }

            throw new IllegalArgumentException("Invalid AcquisitionUnitStatus type '" + str + "'");
        }

        public static Status fromInt(final int status) {
            switch (status) {
                case -1:
                    return NOT_CONNECTED;
                case 1:
                    return CONNECTED;
                default:
                    return WARN;
            }
        }
    }

    public static class AcqUnitAttribute implements Serializable {
        private static final long serialVersionUID = 1L;

        private final String name;
        private final String value;
        private final String type;
        private final String subType;

        private long id;

        public AcqUnitAttribute(final String type, final String subType, final String name, final String value) {
            this.type = type;
            this.subType = subType;
            this.name = name;
            this.value = value;
        }

        public AcqUnitAttribute(final long id, final String type, final String subType, final String name, final String value) {
            this.type = type;
            this.subType = subType;
            this.name = name;
            this.value = value;
            this.id = id;
        }

        public String getName() {
            return name;
        }

        public String getValue() {
            return value;
        }

        public String getType() {
            return type;
        }

        public long getId() {
            return id;
        }

        public void setId(final long id) {
            this.id = id;
        }

        public String getSubType() {
            return subType;
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final AcqUnitAttribute that = (AcqUnitAttribute) o;

            if (name != null ? !name.equals(that.name) : that.name != null) return false;
            if (subType != null ? !subType.equals(that.subType) : that.subType != null) return false;
            if (type != null ? !type.equals(that.type) : that.type != null) return false;
            if (value != null ? !value.equals(that.value) : that.value != null) return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = (name != null ? name.hashCode() : 0);
            result = 31 * result + (value != null ? value.hashCode() : 0);
            result = 31 * result + (type != null ? type.hashCode() : 0);
            result = 31 * result + (subType != null ? subType.hashCode() : 0);
            return result;
        }

        public String toString() {
            return "AU attribute: name=" + name + ", value=" + value + ", type=" + type;
        }
    }
}
