package com.onaro.sanscreen.server.dataobject.acquisition;

import java.util.*;
import java.io.Serializable;

public class DataSourceListAttribute extends DataSourceAttribute {
    private static final long serialVersionUID = 1L;

    private List<DataSourceSimpleAttribute> columns;

    public DataSourceListAttribute(final String name, final String description, final String label, final String role, final boolean editable, final boolean encrypt, final int guiorder, final List<DataSourceSimpleAttribute> columns) {
        super(AttributeType.LIST, name, description, label, role, editable, encrypt, guiorder);
        this.columns = columns;
    }

    public String getDefaultValueAsString() {
        return null;
    }

    public String getValueAsString() {
        return null;
    }

    public void setValueFromString(final String value) {
        throw new UnsupportedOperationException("Cannot set a value for a list type.");
    }

    private Rows values = new Rows();

    public List<DataSourceSimpleAttribute> getColumns() {
        Collections.sort(columns);
        return columns;
    }

    public Map<String, DataSourceSimpleAttribute> getColumnsAsMapByName() {
        final Map<String, DataSourceSimpleAttribute>  result = new TreeMap<String, DataSourceSimpleAttribute>();
        for (final DataSourceSimpleAttribute column : columns) {
            result.put(column.getName(), column);
        }
        return result;
    }


    public List<Row> getRows() {
        return values;
    }

    public Row createRow() {
        // for each column, create a new DataSourceSimpleAttribute and add to a new list
        final List<DataSourceSimpleAttribute> newRowAttributes =  new ArrayList<DataSourceSimpleAttribute>();

        for (final DataSourceSimpleAttribute column : columns) {

            final DataSourceSimpleAttribute copy = copy(column);
            newRowAttributes.add(copy);
        }
        return new Row(newRowAttributes);
    }

    private DataSourceSimpleAttribute copy(final DataSourceSimpleAttribute column) {
        final AttributeType type = column.getType();
        if (type.equals(AttributeType.BOOLEAN)) {
            return new DataSourceBooleanAttribute(column);
        } else
        if (type.equals(AttributeType.STRING)) {
            return new DataSourceStringAttribute(column);
        } else
        if (type.equals(AttributeType.INTEGER)) {
            return new DataSourceIntegerAttribute((DataSourceIntegerAttribute) column);
        } else
        if (type.equals(AttributeType.FLOAT)) {
            return new DataSourceFloatAttribute((DataSourceFloatAttribute) column);
        } else
        if (type.equals(AttributeType.ENUM)) {
            return new DataSourceEnumAttribute((DataSourceEnumAttribute) column);
        }
        return null;
    }


    public class Row implements Serializable {
        private static final long serialVersionUID = 1L;

        List<DataSourceSimpleAttribute> rowValues = new ArrayList<DataSourceSimpleAttribute>();

        Row(final List<DataSourceSimpleAttribute> rowValues) {
            this.rowValues = rowValues;
            Collections.sort(this.rowValues);
        }

        public DataSourceSimpleAttribute getAttributeByIndex(final int index) {
            return rowValues.get(index);
        }

        public List<DataSourceSimpleAttribute> getRowValues() {
            return Collections.unmodifiableList(rowValues);
        }

        public Map<String, DataSourceSimpleAttribute> getRowFieldsAsMapByName() {
            final Map<String, DataSourceSimpleAttribute>  result = new TreeMap<String, DataSourceSimpleAttribute>();
            for (final DataSourceSimpleAttribute field : rowValues) {
                result.put(field.getName(), field);
            }
            return result;
        }

    }

    /**
     * 
     */
    private class Rows extends ArrayList<Row> {
        private static final long serialVersionUID = 1L;

        public boolean add(final Row row) {
            validateRowAgainstColumns(row);
            return super.add(row);
        }

        private void validateRowAgainstColumns(final Row row) throws IllegalArgumentException {
            // make sure that the fields in the added row matches the columns
            for (final DataSourceSimpleAttribute rowField : row.rowValues) {
                // make sure the field is in the column map
                final DataSourceSimpleAttribute column = getColumnsAsMapByName().get(rowField.getName());
                if (column == null || !column.getType().equals(rowField.getType())) {
                    // the rowField does not match any column
                    throw new IllegalArgumentException("Added row does not match the columns for this list." +
                        "Field " + rowField.getName() + " is not a column or is the wrong type in list " + getName());
                }
            }
        }

        private void validateRowsAgainstColumns(final Collection<? extends Row> c) {
            for (final Row row : c) {
                validateRowAgainstColumns(row);
            }
        }

        public void add(final int index, final Row o) {
            validateRowAgainstColumns(o);
            super.add(index, o);
        }
        public boolean addAll(final Collection<? extends Row> c) {
            validateRowsAgainstColumns(c);
            return super.addAll(c);
        }

        public boolean addAll(final int index, final Collection<? extends Row> c) {
            validateRowsAgainstColumns(c);
            return super.addAll(index, c);
        }        
    }



}
