package com.onaro.sanscreen.server.dataobject.acquisition;

import java.io.Serializable;
import java.util.*;

/**
 * Data source status information. Aggregates information about a specific data source and all the packages it runs
 */
public class DataSourceStatus implements Serializable {
    private static final long serialVersionUID = 1L;

    private final AcquisitionUnitNew acquisitionUnit;
    final DataSourceNew dataSource;
    final Map<String, DataSourcePackage> packages = new HashMap<String, DataSourcePackage>();

    public DataSourceStatus(final AcquisitionUnitNew acquisitionUnit, final DataSourceNew dataSource) {
        this.acquisitionUnit = acquisitionUnit;
        this.dataSource = dataSource;
    }

    public AcquisitionUnitNew getAcquisitionUnit() {
        return acquisitionUnit;
    }

    public DataSourceNew getDataSource() {
        return dataSource;
    }

    /**
     * @return The bottom line status of the data source
     */
    public DSStatus getStatus() {
        if (dataSource.getManual() == 0)
            return DSStatus.Stopped;

        return DSStatus.Running;
    }

    public boolean isAlert() {
        boolean alert = false;
        for (final DataSourceStatus.DataSourcePackage dataSourcePackage : getPackages())
            alert |= dataSourcePackage.isAlert();

        return alert;
    }

    public final DataSourcePackage getPackageStatus(final String packageName) {
        return packages.get(packageName);
    }

    public final Collection<DataSourcePackage> getPackages() {
        return packages.values();
    }

    /**
     * Given a set of audits for this data source, this method resets the data source status according to this audit history
     */
    public void initialize(final List<DataSourceAudit> dsAuditRows) {
        packages.clear();

        for (final DataSourceAudit audit : dsAuditRows) {
            DataSourcePackage dataSourcePackage = packages.get(audit.getPackageName());
            if (dataSourcePackage == null) {
                dataSourcePackage = new DataSourcePackage(audit.getPackageName());
                packages.put(audit.getPackageName(), dataSourcePackage);
            }

            dataSourcePackage.updateAudit(audit);
        }
    }

    public static enum DSStatus {
        Stopped,
        Running
    }

    public static final class DataSourcePackage implements Serializable {
        private static final long serialVersionUID = 1L;

        private final String packageName;
        private String lastStatus = null;
        private long lastSuccessTime = 0;
        private long lastAcquisitionTime = 0;
        private boolean alert = false;
        private String alertMessage = null;
        private int sequentialErrors = 0;
        private long lastStartTime = 0;

        DataSourcePackage(final String packageName) {
            this.packageName = packageName;
        }

        public String getPackageName() {
            return packageName;
        }

        public long getLastSuccessTime() {
            return lastSuccessTime;
        }

        public long getLastStartTime() {
            return lastStartTime;
        }

        public boolean isAlert() {
            return alert;
        }

        public void setAlert(final boolean alert, final String alertMessage) {
            this.alert = alert;
            this.alertMessage = alertMessage;
        }

        public String getAlertMessage() {
            if (alertMessage == null)
                return lastStatus;
            
            return alertMessage;
        }

        /**
         * @return true if package is currently successful
         */
        public boolean isSuccessful() {
            return lastStatus != null && lastStatus.equals("Success");
        }

        public boolean isDisabled() {
            return lastStatus != null && lastStatus.equals("Disabled");
        }

        public int getNumOfSequentialErrors() {
            return sequentialErrors;
        }

        public long getLastAcquisitionTime() {
            return lastAcquisitionTime;
        }

        public String getLastStatus() {
            return lastStatus;
        }

        /**
         *
         * @param audit
         *
         * NOTE: Audits are processed in order, starting with the oldest
         * to the most recent.
         *
         */
        void updateAudit(final DataSourceAudit audit) {
            if (!audit.isAuditSessionClosed()) {
                lastStartTime = audit.getStartTime();
                return;
            }
            final String status= audit.getStatus();
            if ("success".equalsIgnoreCase(status) || "stand-by".equalsIgnoreCase(status)) {
                lastSuccessTime = audit.getEndTime();
                sequentialErrors = 0;
            }
            else {
                sequentialErrors += audit.getNumberOfTimes();
            }

            lastStatus = status;
            lastAcquisitionTime = audit.getEndTime();
            lastStartTime = audit.getStartTime();
        }
    }
}
