package com.onaro.sanscreen.server.interfaces.util;

import com.onaro.sanscreen.server.interfaces.data.DataObject;
import com.onaro.sanscreen.server.interfaces.data.DataObjectReference;

import java.util.*;

/**
 * @author Anton Sergeev
 */
public class DataObjectUtils {
    public static <D extends DataObject, T extends Collection<D>>
    Set<Long> getIdsFromCollection(T dataObjects) {
        Set<Long> ids = new HashSet<Long>(dataObjects.size());

        for (D dataObject: dataObjects) {
            ids.add(dataObject.getId());
        }

        return ids;
    }

    public static <D extends DataObject, T extends Collection<D>>
    Set<Long> getIdsFromNestedCollection(Collection<T> dataObjectCollections) {
        Set<Long> ids = new HashSet<Long>(dataObjectCollections.size()); // Size guess.

        for (T dataObjectCollection: dataObjectCollections) {
            for (D dataObject: dataObjectCollection) {
                ids.add(dataObject.getId());
            }
        }
        
        return ids;
    }

    public static <D extends DataObject, T extends Collection<D>>
    Map<Long, D> getIdToDataObject(T dataObjects) {
        Map<Long, D> result = new HashMap<Long, D>(dataObjects.size());

        for (D dataObject: dataObjects) {
            result.put(dataObject.getId(), dataObject);
        }

        return result;
    }

    public static <D extends DataObject, T extends Collection<D>>
    Map<Long, Map<Long, D>> getIdToDataObjectMap(Map<Long, T> dataObjects) {
        Map<Long, Map<Long, D>> result = new HashMap<Long, Map<Long, D>>(dataObjects.size());
         for (Map.Entry<Long, T> entry : dataObjects.entrySet()) {
             Long id = entry.getKey();
             Map<Long, D> map = new HashMap<Long, D>();
             for (D d : entry.getValue()) {
                 map.put(d.getId(), d);
             }
             result.put(id, map);
         }

        return result;
    }

    /**
     * Compare DataObjects by their IDs. This can be used to test if two DataObjects represent the same data.
     * <p>
     * The server only guarantees that DataObjects with the same ObjectType have unique IDs. To reflect this,
     * this method requires its arguments to be the same subtype of DataObject. Note DataObject itself is an
     * abstract type, therefore the arguments will be pure subtypes.
     * <p>
     * It turns out that in v6.0 the server actually assigns unique IDs to all DataObjects, regardless of
     * ObjectType. Therefore this method would be useful even if the arguments were different subtypes of
     * DataObject. However, this is not guaranteed by the server team (according to Dong Sun).
     * 
     * @param left
     * @param right
     * @return {@code true} if the arguments are both null or have the same IDs
     */
    public static <D extends DataObject> boolean isIdEqual (D left, D right) {
        if (left == null) {
            return (right == null);
        } else {
            return left.getId().equals (right.getId());
        }
    }
    
    /**
     * Get the hash code of a DataObject's ID. This can be used to implement hashCode() if {@link #isIdEqual} is used to implement equals().
     * @param dataObject
     * @return 0 if the argument is null, otherwise a hash code of the argument's ID
     */
    public static int getIdHashCode (DataObject dataObject) {
        return dataObject == null? 0 : dataObject.getId().hashCode();
    }

    public static Set<Long> extractIdsOutOfReferences(Set<? extends DataObjectReference> refs) {
        Set<Long> ids = new HashSet<Long>();
        for (DataObjectReference ref: refs) {
            ids.add(ref.getId());
        }

        return ids;
    }
}
