/*
 * Copyright (c) 2009 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.server.sessions;

import com.netapp.oci.platform.common.ServerConstant;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.Table;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.originator.OriginatorStateCacheFactory;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.plugin.data.PlatformPlugin;
import com.netapp.oci.platform.plugin.remote.PlatformPluginRemote;
import com.netapp.oci.server.TransactionContext;
import com.onaro.commons.exec.Exec;
import com.onaro.commons.table.DbConfig;
import com.onaro.sanscreen.acquisition.identify.AcqStorageIdentifyTable;
import com.onaro.sanscreen.acquisition.identify.AcqVolumeIdentifyTable;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import com.onaro.sanscreen.server.tables.acquisition.AcquisitionUnitAuditTable;
import com.onaro.sanscreen.server.tables.acquisition.AcquisitionUnitTableNew;
import com.onaro.sanscreen.server.tables.acquisition.DataSourceAuditTable;
import com.onaro.sanscreen.server.tables.acquisition.DataSourceTableNew;
import com.onaro.sanscreen.server.tables.idgenerator.IdGenerator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.*;
import javax.sql.DataSource;

import java.rmi.RemoteException;
import java.sql.Connection;

/**
 * @author Asaf Yigal
 * @version 1.0
 */
@Stateless
@TransactionAttribute(TransactionAttributeType.REQUIRED)
@Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
@RolesAllowed({"admin"})
public class StartupSessionEJB implements StartupSession {

    private static final Logger LOG = LoggerFactory.getLogger(StartupSessionEJB.class);

    @Resource
    private SessionContext sessionContext;

    @Resource(name = "jdbc/OnaroDataSource",
            mappedName = "java:/onaro_pool",
            type = DataSource.class,
            authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    @EJB
    OriginatorUpdateSession originatorUpdateSession;

    @EJB
    PlatformPluginRemote platformPluginRemote;

    @Override
    public void clearSanData() throws TableException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Table<?>[] tables = new Table[]{
                    new AcqStorageIdentifyTable(connection),
                    new AcqVolumeIdentifyTable(connection),
            };

            for (Table<?> table : tables) {
                table.deleteAll();
            }

            for (PlatformPlugin plugin : platformPluginRemote.getPlugins()) {
                plugin.deleteTables(connection);
                PreparedStatementWrapper.createForUpdate(connection, "DELETE FROM " + plugin.getUpdateHistoryTableName()).executeUpdateEx();
            }

            clearAllCaches();

            TransactionContext.getLastUpdateTime();
        } catch (Exception e) {
            throw new TableException("Failed to clear SanData", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    /**
     * Clears and resets all the memory cache we keep in the server.
     *
     * @throws RemoteException
     */
    @Override
    public void clearAllCaches() throws TableException {
        try {
            IdGenerator.getInstance().reset();
            OriginatorStateCacheFactory.removeOriginatorCache(null);
            KeyCache.clearAll();
            TransactionContext.resetLastUpdateTime();
        } catch (Exception e) {
            throw new TableException("Failed to clear all caches", e);
        }
    }

    @Override
    public void clearAllData() throws TableException {
        clearSanData();
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Table<?>[] tables = new Table[]{new DataSourceTableNew(connection),
                    new AcquisitionUnitAuditTable(connection),
                    new AcquisitionUnitTableNew(connection),
                    new DataSourceAuditTable(connection)};

            for (Table<?> table : tables) {
                table.deleteAll();
            }
        } catch (Exception e) {
            sessionContext.setRollbackOnly();
            throw new TableException("Failed to clear all data", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public void clearChanges() throws TableException {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public void restartServer() {
        String onaroHome = System.getProperty("onaro.home", "c:/program files/sanscreen");
        String command = "\"" + onaroHome + "/bin/restart_server.cmd" + "\"";
        LOG.info("Restart Server, Running:" + command);
        Exec exec = new Exec(command);
        exec.run(120000);
    }

    @Override
    public void restartAcq() {
        String onaroHome = System.getProperty("onaro.home", "c:/program files/sanscreen");
        String command = "\"" + onaroHome + "/bin/restart_acq.cmd" + "\"";
        LOG.info("Restart Acq, Running:" + command);
        Exec exec = new Exec(command);
        exec.run(60000);
    }

    /**
     * Changes the prefix of the database (to a customer). {@code null} means no prefix.
     *
     * @param mysqlHost
     * @param prefix
     */
    @Override
    public void setPrefixDatabase(String mysqlHost, String prefix) throws ServerInterfaceException {
        try {
            DbConfig.setMySqlHost(mysqlHost);
            DbConfig.setDatabasePrefix(prefix);

            clearAllCaches();
        } catch (TableException e) {
            throw new ServerInterfaceException("Failed to set prefix database: mysqlHost=" + mysqlHost + ", prefix=" + prefix, e);
        }
    }
}
