package com.onaro.sanscreen.server.tables.acquisition;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import com.onaro.commons.table.DbConfig;
import com.onaro.sanscreen.server.dataobject.acquisition.AcqAttribute;

import com.netapp.oci.platform.common.table.TableException;

public class DataSourceAttrsTable extends AcquisitionTable<AcqAttribute> {
    private static final String TABLE_NAME = "ds_attrs";

    public DataSourceAttrsTable(Connection connection) {
        super(connection, DataSourceAttrsTable.TABLE_NAME);
    }

    /**
     * This static method returns the full table name without initializing an instance. It will be aware of changes in customer prefix {@link com.onaro.commons.table.DbConfig}
     * @return
     */
    public static String FULL_TABLE_NAME() {
        return DbConfig.getAcquisitionDbName() + "." + TABLE_NAME;
    }

    public void insert(final long ds_id, final String name, final String value) throws TableException {
        getPrepareStatementEx("INSERT INTO " + getFullName() + " (ds_id, name, value) VALUES(?, ?, ?)")
                .setNext(ds_id)
                .setNext(name)
                .setNext(value)
                .executeUpdateEx();
    }

    public void insert(final long ds_id, final Properties attrs) throws TableException {
        for (Enumeration<?> enumeration = attrs.keys(); enumeration.hasMoreElements();) {
            final String name = (String) enumeration.nextElement();
            final String value = attrs.getProperty(name);

            insert(ds_id, name, value);
        }
    }

    public AcqAttribute parseObject(final ResultSet resultSet) throws SQLException {
        return new AcqAttribute(
                resultSet.getLong("ds_id"),
                resultSet.getString("name"),
                resultSet.getString("value"));
    }

    public Properties findByDsId(final long ds_id) throws TableException {
        List<AcqAttribute> rows = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE ds_id=?").setNext(ds_id).executeQueryEx();
        Properties result = new Properties();
        for (final AcqAttribute acqAttribute : rows) {
            result.put(acqAttribute.getName(), acqAttribute.getValue());
        }
        return result;
    }

    public HashMap<Long, String> getDsIdValueMap(String attributeName) throws TableException {
        List<AcqAttribute> rows = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE name=?").setNext(attributeName).executeQueryEx();
        HashMap<Long, String> dsIdValueMap = new HashMap<Long, String>();

        for (final AcqAttribute acqAttribute : rows) {
            dsIdValueMap.put(acqAttribute.getObjid(), acqAttribute.getValue());
        }
        return dsIdValueMap;
    }

    public void update(final long ds_id, final Properties attrs) throws TableException {
        for (Enumeration<?> enumeration = attrs.keys(); enumeration.hasMoreElements();) {
            final String name = (String) enumeration.nextElement();
            final String value = attrs.getProperty(name);

            List<AcqAttribute> foundAttrs = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE ds_id=? and name=?")
                    .setNext(ds_id)
                    .setNext(name)
                    .executeQueryEx();

            if (foundAttrs.size() == 1 && !foundAttrs.get(0).getValue().equals(value)) {
                update(ds_id, name, value);
            }

            if (foundAttrs.size() == 0) {
                insert(ds_id, name, value);
            }
        }
    }

    public void update(final long ds_id, final String name, final String value) throws TableException {
        getPrepareStatementEx("UPDATE " + getFullName() + " SET value=? WHERE ds_id=? AND name=?")
                .setNext(value)
                .setNext(ds_id)
                .setNext(name)
                .executeUpdateEx();
    }

    public void create(final Long ds_id, final String name, final String value) throws TableException {
        try {
            // First we try insert. If the record already exists we get an exception and we try update
            getPrepareStatementEx("INSERT INTO " + getFullName() + " (ds_id, name, value) VALUES(?, ?, ?)")
                    .setNext(ds_id)
                    .setNext(name)
                    .setNext(value)
                    .executeUpdateEx();
        }
        catch (TableException e) {
            getPrepareStatementEx("UPDATE " + getFullName() + " SET value=? WHERE ds_id=? AND name=?")
                    .setNext(value)
                    .setNext(ds_id)
                    .setNext(name)
                    .executeUpdateEx();
        }
    }

    public void removeListProperties(final Long ds_id) throws TableException {
        String listKeyWordString = "%.LIST.%";
        getPrepareStatementEx("DELETE FROM " + getFullName() + " WHERE ds_id=? AND name like ?")
                .setNext(ds_id)
                .setNext(listKeyWordString)
                .executeUpdateEx();
    }



}
