package com.onaro.sanscreen.server.tables.acquisition;

import com.netapp.oci.platform.common.table.TableException;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceAudit;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

public class DataSourceAuditTable extends AcquisitionTable<DataSourceAudit> {

    public static final String TABLE_NAME = "ds_audit";

    public DataSourceAuditTable(Connection connection) {
        super(connection, TABLE_NAME);
    }

    public DataSourceAudit parseObject(final ResultSet resultSet) throws SQLException {
        return new DataSourceAudit(resultSet.getLong("id"),
                                   resultSet.getString("package"),
                                   resultSet.getString("status"),
                                   resultSet.getString("uri"),
                                   resultSet.getLong("ds_id"),
                                   resultSet.getLong("starttime"),
                                   resultSet.getLong("endtime"),
                                   resultSet.getInt("numberoftimes"));
    }

    public void insert(final DataSourceAudit dataSourceAudit) throws TableException {
        final String sql = new StringBuilder()
                .append("INSERT INTO ").append(getFullName())
                .append(" (package, status, uri, ds_id, starttime, endtime, numberoftimes) ")
                .append("VALUES(?, ?, ?, ?, ?, ?, ?)").toString();

        getPrepareStatementEx(sql)
                .setNext(dataSourceAudit.getPackageName())
                .setNext(dataSourceAudit.getStatus())
                .setNext(dataSourceAudit.getURI())
                .setNext(dataSourceAudit.getDsId())
                .setNext(dataSourceAudit.getStartTime())
                .setNext(dataSourceAudit.getEndTime())
                .setNext(dataSourceAudit.getNumberOfTimes())
                .executeUpdateEx();

        dataSourceAudit.setId(getAutoGeneratedId());
    }

    public void update(final DataSourceAudit dataSourceAudit) throws TableException {
        final String sql = new StringBuilder()
                .append("UPDATE ").append(getFullName()).append(" SET ")
                .append("package=?, ")
                .append("status=?, ")
                .append("uri=?, ")
                .append("ds_id=?, ")
                .append("starttime=?, ")
                .append("endtime=?, ")
                .append("numberoftimes=? ")
                .append("WHERE id = ?").toString();

        getPrepareStatementEx(sql)
                .setNext(dataSourceAudit.getPackageName())
                .setNext(dataSourceAudit.getStatus())
                .setNext(dataSourceAudit.getURI())
                .setNext(dataSourceAudit.getDsId())
                .setNext(dataSourceAudit.getStartTime())
                .setNext(dataSourceAudit.getEndTime())
                .setNext(dataSourceAudit.getNumberOfTimes())
                .setNext(dataSourceAudit.getId())
                .executeUpdateEx();
    }

    /**
     * @param dsID Data Source ID
     *
     * @return all the audit records related to the given data source ID, sorted by the starttime (desc).
     *
     * @throws TableException
     */
    public List<DataSourceAudit> findAllByOriginId(long dsID) throws TableException {
        return getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE ds_id = ? ORDER BY starttime DESC")
                .setNext(dsID)
                .executeQueryEx();
    }

    /**
     * Finds all the records that belongs to the same Data source (ds_id) where the audit-record's key is
     * less than or equal than <code>id</code> passed in.
     * This method is used to delete Data source's audit records that belongs to the same data source and share same
     * "status" information.  In this way we'll avoid record duplicity.
     *
     * @param dsID Data source's id
     * @param auditID Data source audit's record's id
     *
     * @return all the audit records related to the given origin ID and ObjID, sorted by the starttime (desc).
     *
     * @throws TableException
     */
    public List<DataSourceAudit> findAllByOriginIdAndObjId(long dsID, final String packageName, long auditID) throws TableException {
        final String sql = "SELECT DISTINCT * FROM " + getFullName() + " WHERE ds_id = ? and package = ? and id <= ? ORDER BY starttime DESC";
        return getPrepareStatementEx(sql)
                .setNext(dsID)
                .setNext(packageName)
                .setNext(auditID)
                .executeQueryEx();
    }

    /**
     * Get the audit history of a specific data source. This will return audits for all packages, sorted by starttime DESC.
     *
     * @param dsID
     *
     * @return the statuses record of the DS related to the given origin ID
     *
     * @throws TableException
     */
    public List<DataSourceAudit> findStatuses(long dsID) throws TableException {
        final String sql = "SELECT * FROM " + getFullName() + " WHERE ds_id=? ORDER BY starttime DESC";
        return getPrepareStatementEx(sql)
                .setNext(dsID)
                .executeQueryEx();
    }

    /**
     * Get the audit history of a specific data source. This will return audits for all packages, sorted by starttime ASC.
     *
     * @param dsID
     *
     * @return the statuses record of the DS related to the given origin ID
     *
     * @throws com.netapp.oci.platform.common.table.TableException
     */
    public List<DataSourceAudit> getAuditHistory(long dsID) throws TableException {
        // TODO: Limit time to no more than week odf history
        final String sql = "SELECT * FROM " + getFullName() + " WHERE ds_id=? ORDER BY starttime ASC";
        return getPrepareStatementEx(sql)
                .setNext(dsID)
                .executeQueryEx();
    }

    /**
     * @return all the audit records related to the given origin ID, sorted by the origin ID and then starttime (desc).
     *
     * @throws com.netapp.oci.platform.common.table.TableException
     */
    public List<DataSourceAudit> findAllByAcquisitionUnit(final long au_id) throws TableException {
        final String sql = "SELECT DISTINCT dsa.* FROM " + getFullName() + " dsa, " + DataSourceTableNew.FULL_TABLE_NAME() + " ds " +
                           "WHERE dsa.ds_id = ds.id AND ds.au_id = ? ORDER BY dsa.ds_id, dsa.starttime DESC";
        return getPrepareStatementEx(sql)
                .setNext(au_id)
                .executeQueryEx();
    }

    public DataSourceAudit find(long auditId) throws TableException {
        final String sql = "SELECT * FROM " + getFullName() + " WHERE id=?";
        final List<DataSourceAudit> dataSourceAudits = getPrepareStatementEx(sql)
                .setNext(auditId)
                .executeQueryEx();

        if (dataSourceAudits == null || dataSourceAudits.size() != 1)
            return null;

        return dataSourceAudits.get(0);
    }

    public void deleteForDataSource(long dsID) throws TableException {
        getPrepareStatementEx("DELETE FROM " + getFullName() + " WHERE ds_id=?")
                .setNext(dsID)
                .executeUpdateEx();
    }

    public void deleteById(long id) throws TableException {
        getPrepareStatementEx("DELETE FROM " + getFullName() + " WHERE id=?")
                .setNext(id)
                .executeUpdateEx();
    }

    public void deleteByOldStartDate(long dsId, long time) throws TableException {
        getPrepareStatementEx("DELETE FROM " + getFullName() + " " +
                              "WHERE ds_id = ? and " +
                              " starttime < ?")
                .setNext(dsId)
                .setNext(time)
                .executeUpdateEx();
    }

}
