/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.server.tables.acquisition;

import com.netapp.oci.platform.common.table.TableException;
import com.onaro.commons.table.DbConfig;
import com.onaro.sanscreen.server.dataobject.acquisition.DataSourceNew;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

public class DataSourceTableNew extends AcquisitionTable<DataSourceNew> {

    private static final String TABLE_NAME = "ds";

    public DataSourceTableNew(final Connection connection) {
        super(connection, TABLE_NAME);
    }

    /**
     * This static method returns the full table name without initializing an instance. It will be aware of changes in customer prefix {@link DbConfig}
     * @return
     */
    public static String FULL_TABLE_NAME() {
        return DbConfig.getAcquisitionDbName() + "." + TABLE_NAME;
    }

    public DataSourceNew parseObject(final ResultSet resultSet) throws SQLException {
        return new DataSourceNew(
                resultSet.getLong("au_id"),
                resultSet.getLong("ds_type_id"),
                resultSet.getLong("id"),
                resultSet.getString("name"),
                null,
                resultSet.getString("vendor"),
                resultSet.getString("model"),
                resultSet.getString("status"),
                resultSet.getString("statusext"),
                resultSet.getLong("time"),
                resultSet.getInt("manual"),
                resultSet.getBoolean("active"));
    }

    public void insert(final DataSourceNew ds) throws TableException {
        final String sql =
                "INSERT INTO " + getFullName() +
                        "(ds_type_id, au_id, name, vendor, model, status, statusext, time, manual, active) " + " VALUES " +
                        "(?,          ?,     ?,    ?,      ?,     ?,      ?,         ?,    ?,      ?)";

        getPrepareStatementEx(sql)
                .setNext(ds.getDsTypeId())
                .setNext(ds.getAuId())
                .setNext(ds.getName())
                .setNext(ds.getVendor())
                .setNext(ds.getModel())
                .setNext(ds.getStatus())
                .setNext(ds.getStatusExt())
                .setNext(ds.getTime())
                .setNext(ds.getManual())
                .setNext(ds.isActive())
                .executeUpdateEx();

        ds.setId(getAutoGeneratedId());
    }

    public void update(final DataSourceNew ds) throws TableException {
        final String sql =
                "UPDATE " + getFullName() +
                        " SET au_id=?, ds_type_id=?, name=?, vendor=?, model=?, status=?, statusext=?, time=?, manual=?, active=? WHERE id=?";

        getPrepareStatementEx(sql)
                .setNext(ds.getAuId())
                .setNext(ds.getDsTypeId())
                .setNext(ds.getName())
                .setNext(ds.getVendor())
                .setNext(ds.getModel())
                .setNext(ds.getStatus())
                .setNext(ds.getStatusExt())
                .setNext(ds.getTime())
                .setNext(ds.getManual())
                .setNext(ds.isActive())
                .setNext(ds.getId())
                .executeUpdateEx();
    }

    public void updateLastStatus(final long ds_id, final String status) throws TableException {
        getPrepareStatementEx("UPDATE " + getFullName() + " SET status=? WHERE id=?")
                .setNext(status)
                .setNext(ds_id)
                .executeUpdateEx();
    }

    public void updateWithoutManualAndActive(final DataSourceNew ds) throws TableException {
        final String sql =
                "UPDATE " + getFullName() +
                        " SET au_id=?, ds_type_id=?, name=?, vendor=?, model=?, status=?, statusext=?, time=? WHERE id=?";

        getPrepareStatementEx(sql)
                .setNext(ds.getAuId())
                .setNext(ds.getDsTypeId())
                .setNext(ds.getName())
                .setNext(ds.getVendor())
                .setNext(ds.getModel())
                .setNext(ds.getStatus())
                .setNext(ds.getStatusExt())
                .setNext(ds.getTime())
                .setNext(ds.getId())
                .executeUpdateEx();
    }

    public void deleteDataSources(final long[] ids) throws TableException {
        for (final long id : ids)
            getPrepareStatementEx("DELETE FROM " + getFullName() + " WHERE id=?").setNext(id).executeUpdateEx();
    }

    public List<DataSourceNew> findByName(final String name) throws TableException {
        DataSourceAttrsTable dsAttrsTable = new DataSourceAttrsTable(getConnection());
        List<DataSourceNew> result = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE name=?").setNext(name).executeQueryEx();
        for (DataSourceNew ds : result)
            ds.setAttrs(dsAttrsTable.findByDsId(ds.getId()));

        return result;
    }

    public DataSourceNew getDS(final String ip, final int ds_type_id, final long time) throws TableException {
        final String sql = "SELECT ds.* FROM " + getFullName() + " ds, " + DataSourceAttrsTable.FULL_TABLE_NAME() + " da " +
                "WHERE ds.ds_type_id=? AND ds.id=da.ds_id AND da.name='ip' AND da.value=? AND ds.time <= ?";

        List<DataSourceNew> result = getPrepareStatementEx(sql).setNext(ds_type_id).setNext(ip).setNext(time).executeQueryEx();
        if (result == null || result.size() != 1)
            return null;

        return result.get(0);
    }

    public DataSourceNew getByOriginatorId(final long originatorId) throws TableException {
        final String sql = "SELECT ds.* FROM " + getFullName() + " ds, " + OriginatorTable.FULL_TABLE_NAME() + " o " +
                "WHERE ds.id=o.ds_id AND o.id=?";

        List<DataSourceNew> result = getPrepareStatementEx(sql).setNext(originatorId).executeQueryEx();
        if (result == null || result.size() != 1)
            return null;

        return result.get(0);
    }

    public List<DataSourceNew> findByTypeIdAndDeviceKey(final long dataSourceTypeId, final String deviceKey) throws TableException {
        final String sql = "SELECT ds.* FROM " + getFullName() + " ds, " + OriginatorTable.FULL_TABLE_NAME() + " o " +
                " WHERE o.ds_id = ds.id AND ds.ds_type_id = ? AND o.devicekey=? ";
         return getPrepareStatementEx(sql).setNext(dataSourceTypeId).setNext(deviceKey).executeQueryEx();
    }

    public List<DataSourceNew> findByAuName(final String au_name) throws TableException {
        final String sql = "SELECT ds.* FROM " + getFullName() + " ds, " + AcquisitionUnitTableNew.FULL_TABLE_NAME() + " au " +
                "WHERE ds.au_id=au.id AND au.name=?";

        return getPrepareStatementEx(sql).setNext(au_name).executeQueryEx();
    }

    public List<DataSourceNew> findByTime(final long time) throws TableException {
        return getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE time <=?").setNext(time).executeQueryEx();
    }

    public DataSourceNew getById(final Long id) throws TableException {
        List<DataSourceNew> rows = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE id=?").setNext(id).executeQueryEx();
        if (rows.size() == 0)
            return null;
        return rows.get(0);
    }

    public List<DataSourceNew> findByVendorAndModel(final String vendor, final String model) throws TableException {
        return getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE vendor=? AND model=?").setNext(vendor).setNext(model).executeQueryEx();
    }

    public List<DataSourceNew> findByTypeId(final long dataSourceTypeId) throws TableException {
        return getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE ds_type_id=?").setNext(dataSourceTypeId).executeQueryEx();
    }

    public DataSourceNew getByIdAndTime(final long id, final long time) throws TableException {
        List<DataSourceNew> result = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE ds.id=? AND time<=?").setNext(id).setNext(time).executeQueryEx();
        if (result == null || result.size() != 1)
            return null;

        return result.get(0);
    }

    public void updateIsActive(final long ds_id, final boolean active) throws TableException {
        getPrepareStatementEx("UPDATE " + getFullName() + " SET active=? WHERE id=?")
                .setNext(active)
                .setNext(ds_id)
                .executeUpdateEx();
    }

    public void updateManual(final long ds_id, final int manual) throws TableException {
        getPrepareStatementEx("UPDATE " + getFullName() + " SET manual=? WHERE id=?")
                .setNext(manual)
                .setNext(ds_id)
                .executeUpdateEx();
    }
}
