package com.onaro.sanscreen.server.tables.acquisition;

import com.netapp.oci.platform.common.table.TableException;
import com.onaro.commons.table.DbConfig;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class OriginatorTable extends AcquisitionTable<Originator> {
    private static final String TABLE_NAME = "originator";

    public OriginatorTable(Connection connection) {
        super(connection, TABLE_NAME);
    }

    /**
     * This static method returns the full table name without initializing an instance. It will be aware of changes in customer prefix {@link DbConfig}
     *
     * @return
     */
    public static String FULL_TABLE_NAME() {
        return DbConfig.getAcquisitionDbName() + "." + TABLE_NAME;
    }

    public Originator parseObject(final ResultSet resultSet) throws SQLException {
        return new Originator(
                resultSet.getLong("id"),
                resultSet.getLong("ds_id"),
                resultSet.getString("devicekey"),
                resultSet.getString("name"),
                resultSet.getLong("lastaccesstime")
        );
    }

    public void insert(final Originator originator) throws TableException {
        getPrepareStatementEx("INSERT INTO " + getFullName() + " (ds_id, devicekey, name, lastaccesstime) VALUES(?, ?, ?, ?)")
                .setNext(originator.getDsId())
                .setNext(originator.getDeviceKey())
                .setNext(originator.getName())
                .setNext(originator.getLastAccessTime())
                .executeUpdateEx();

        originator.setId(getAutoGeneratedId());
    }



    public Long getByDsIdAndDeviceKey(final long ds_id, final String deviceKey) throws TableException {
        final List<Originator> originators;
        if (deviceKey != null) {
            originators = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE devicekey=? and ds_id=?")
                    .setNext(deviceKey)
                    .setNext(ds_id)
                    .executeQueryEx();
        } else {
            originators = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE devicekey is NULL and ds_id=?")
                    .setNext(ds_id)
                    .executeQueryEx();
        }

        if (originators.size() == 0)
            return null;

        if (originators.size() == 1)
            return originators.get(0).getId();

        final String errorString = "OriginatorTable.getByDsIdAndDeviceKey(ds_id= " + ds_id + ";key=" + deviceKey + ") - originators.size=" + originators.size() + " - ERROR: more than one originator found !!!";
        throw new TableException(errorString);
    }

    public Set<Long> findByDsId(final long ds_id) throws TableException {
        final List<Originator> list = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE ds_id=?").setNext(ds_id).executeQueryEx();
        final Set<Long> result = new HashSet<Long>();
        for (Originator originatorRecord : list)
            result.add(originatorRecord.getId());

        return result;
    }

    public long update(final Long id, final String deviceKey, final String name) throws TableException {
        return getPrepareStatementEx("UPDATE " + getFullName() + " SET devicekey=?, name=? WHERE id=?")
                .setNext(deviceKey)
                .setNext(name)
                .setNext(id)
                .executeUpdateEx();

    }

    public long updateOriginatorLastAccessTime(final Long id) throws TableException {
        return getPrepareStatementEx("UPDATE " + getFullName() + " SET lastaccesstime=? WHERE id=?")
                .setNext(System.currentTimeMillis())
                .setNext(id)
                .executeUpdateEx();
    }

    public void delete(final Long id) throws TableException {
        getPrepareStatementEx("DELETE FROM " + getFullName() + " WHERE id=?")
                .setNext(id)
                .executeUpdateEx();
    }

    public void delete(final Long datasourceID, final String deviceKey) throws TableException {
        if (deviceKey == null) {
            getPrepareStatementEx("DELETE FROM " + getFullName() + " WHERE ds_id=? AND devicekey is NULL")
                    .setNext(datasourceID)
                    .executeUpdateEx();
        } else {
            getPrepareStatementEx("DELETE FROM " + getFullName() + " WHERE ds_id=? AND devicekey=?")
                    .setNext(datasourceID)
                    .setNext(deviceKey)
                    .executeUpdateEx();
        }
    }

    public List<Originator> getExpiredOriginators(final long originatorCutoffTimeMillis) throws TableException {
        final List<Originator> originators;
        originators = getPrepareStatementEx("SELECT * FROM " + getFullName() + " WHERE lastaccesstime < ?")
                .setNext(originatorCutoffTimeMillis)
                .executeQueryEx();
        return originators;
    }

}
