/*
 * Capacity
 *
 * Copyright (c) 2008 NetApp
 * All rights reserved
 */
package com.onaro.util;

import java.io.Serializable;

public class Capacity implements Serializable, Comparable<Capacity> {

    private final static long serialVersionUID = 1L;

    private final boolean allowNegative;
    private long bytes;

    private Capacity(final long bytes, final Unit unit, final boolean allowNegative) {
        if (bytes < 0 && !allowNegative)
            throw new IllegalArgumentException("capacity:" + bytes + " is illegal, can not be less then 0.");

        this.bytes = unit.toByte(bytes);
        this.allowNegative = allowNegative;
    }

    /**
     * Create a capacity object with the given value and unit
     */
    public static Capacity create(final long value, final Unit unit) {
        return new Capacity(value, unit, false);
    }

    /**
     * Create a capacity object by copying another capacity
     */
    public static Capacity create(final Capacity capacity) {
        return new Capacity(capacity.bytes, Unit.BYTE, false);
    }

    /**
     * Create an empty (0) capacity object with the given value and unit
     */
    public static Capacity create() {
        return new Capacity(0L, Unit.BYTE, false);
    }

    /**
     * Allow create Capacity with Negative number
     *
     * @return a capacity object (with 0 capacity) allowing negative numbers
     */
    public static Capacity createAllowNegative() {
        return new Capacity(0L, Unit.BYTE, true);
    }

    /**
     * Create a capacity object by copying another capacity
     */
    public static Capacity createAllowNegative(final Capacity capacity) {
        return new Capacity(capacity.bytes, Unit.BYTE, true);
    }

    /**
     * Allow create Capacity with Negative number
     *
     * @param value
     * @param unit
     *
     * @return a capacity object allowing negative numbers
     */
    public static Capacity createAllowNegative(final long value, final Unit unit) {
        return new Capacity(value, unit, true);
    }

    /**
     * @return The value of this capacity in the given unit
     */
    public final long value(final Unit unit) {
        return unit.fromByte(bytes);
    }

    /**
     * Used by acquisiton when generating XML out of capacity objects (in this case nulls should just be returned as nulls)
     *
     * @return Capacity value (in required units) or <code>null</code> if cu is null.
     */
    public static Long valueForXML(final Capacity cu, final Unit unit) {
        if (cu == null)
            return null;

        //noinspection UnnecessaryBoxing
        return Long.valueOf(unit.fromByte(cu.bytes));
    }

    /**
     * Add another capacity into this object
     *
     * @param c Capacity to add (if <code>null</code> it is treated as zero)
     */
    public void add(final Capacity c) {
        if (c == null)
            return;

        final long resultant = bytes + c.bytes;
        if (resultant < 0 && !allowNegative)
            throw new IllegalArgumentException("Capacity " + resultant + " is negative and not allowed by this object. (" +
                                               bytes + " + " + c.bytes + " = " + resultant + ")");

        bytes = resultant;
    }

    /**
     * Substract another capacity from this object
     *
     * @param c Capacity to substract (if <code>null</code> it is treated as zero)
     */
    public void substract(final Capacity c) {
        if (c == null)
            return;

        final long resultant  = bytes - c.bytes;
        if (resultant < 0 && !allowNegative)
            throw new IllegalArgumentException("Capacity " + resultant + " is negative and not allowed by this object. (" +
                                               bytes + " - " + c.bytes + " = " + resultant + ")");

        bytes = resultant;
    }

    /**
     * Add another capacity into this object
     *
     * @param bytes
     */
    public void add(final long bytes) {
        final long resultant = this.bytes + bytes;
        if (resultant < 0 && !allowNegative)
            throw new IllegalArgumentException("Capacity " + resultant + " is negative and not allowed by this object. (" +
                                               this.bytes + " + " + bytes + " = " + resultant + ")");

        this.bytes = resultant;
    }

    /**
     * Substract another capacity from this object
     *
     * @param bytes
     */
    public void substract(final long bytes) {
        final long resultant = this.bytes - bytes;
        if (resultant < 0 && !allowNegative)
            throw new IllegalArgumentException("Capacity " + resultant + " is negative and not allowed by this object. (" +
                                               this.bytes + " - " + bytes + " = " + resultant + ")");

        this.bytes = resultant;
    }

    @Override
    public boolean equals(Object o) {
        if (o == this)
            return true;
        if (!(o instanceof Capacity))
            return false;
        Capacity capacity = (Capacity) o;
        return bytes == capacity.bytes;
    }

    @Override
    public int hashCode() {
        return (int) (bytes ^ (bytes >>> 32));
    }


    public int compareTo(Capacity capacity) {
        if (capacity == null)
            throw new IllegalArgumentException("Capacity is null");

        if (bytes > capacity.bytes)
            return 1;

        if (bytes < capacity.bytes)
            return -1;

        return 0;
    }

    /**
     * @return String representation of this capacity object (so we can look at it in the debugger) in KB (1024 based)
     */
    @Override 
    public String toString() {
        return value(Unit.KB_1024) + "KB";
    }

    /**
     *
     * @param unit The capacity unit to evaluate to.
     * @return A string representation of the capacity of this object in the unit specified. The unit abbreviation
     * will also be concatenated to the resulting string as well.
     */
    public String toString(final Unit unit) {
        return String.valueOf(value(unit)).concat(unit.getAbbreviation());
    }

    /**
     * @return String representation of capacity in bytes
     */
    public String toStringB() {
        return toString(Unit.BYTE);
    }

    /**
     * @return String representation of capacity in kibibytes
     */
    public String toStringKiB() {
        return toString(Unit.KB_1000);
    }

    /**
     * @return String representation of capacity in mebibytes
     */
    public String toStringMiB() {
        return toString(Unit.MB_1000);
    }

    /**
     * @return String representation of capacity in gibibytes
     */
    public String toStringGiB() {
        return toString(Unit.GB_1000);
    }

    /**
     * @return String representation of capacity in tebibytes
     */
    public String toStringTiB() {
        return toString(Unit.TB_1000);
    }

    /**
     * @return String representation of capacity in pebibytes
     */
    public String toStringPiB() {
        return toString(Unit.PB_1000);
    }

    /**
     * @return String representation of capacity in kilobytes
     */
    public String toStringKB() {
        return toString(Unit.KB_1024);
    }

    /**
     * @return String representation of capacity in megabytes
     */
    public String toStringMB() {
        return toString(Unit.MB_1024);
    }

    /**
     * @return String representation of capacity in gigabytes
     */
    public String toStringGB() {
        return toString(Unit.GB_1024);
    }

    /**
     * @return String representation of capacity in terabytes
     */
    public String toStringTB() {
        return toString(Unit.TB_1024);
    }

    /**
     * @return String representation of capacity in petabytes
     */
    public String toStringPB() {
        return toString(Unit.PB_1024);
    }

    public enum Unit {
        BYTE(1, "B"),
        KB_1000(1000, "KiB"),
        MB_1000(KB_1000.factor * 1000, "MiB"),
        GB_1000(MB_1000.factor * 1000, "GiB"),
        TB_1000(GB_1000.factor * 1000, "TiB"),
        PB_1000(TB_1000.factor * 1000, "PiB"),
        KB_1024(1024, "KB"),
        MB_1024(KB_1024.factor * 1024, "MB"),
        GB_1024(MB_1024.factor * 1024, "GB"),
        TB_1024(GB_1024.factor * 1024, "TB"),
        PB_1024(TB_1024.factor * 1024, "PB");

        private final long factor;
        private final String abbreviation;

        public String getAbbreviation() {
            return this.abbreviation;
        }

        private Unit(final long factor, final String abbreviation) {
            this.factor = factor;
            this.abbreviation = abbreviation;
        }

        long fromByte(long byteValue) {
            return byteValue / factor;
        }

        long toByte(long byteValue) {
            return byteValue * factor;
        }
    }
}
