package com.onaro.util;

import java.util.LinkedList;
import java.util.Iterator;
import java.util.zip.Deflater;
import java.util.zip.DataFormatException;
import java.util.zip.Inflater;
import java.io.*;

public class Compression {

    /**
     * compresses a byte array, using ZLIB algorithm with the default strategy
     * ({@link Deflater#DEFAULT_STRATEGY}) and the default level ({@link Deflater#DEFAULT_COMPRESSION}).
     * @param input the byte array to compress
     * @return compressed the compressed data prefixed with a 4 byte (long) length
     */
    public static byte[] compressData(byte input[]) {
        return compressData(input, Deflater.DEFAULT_COMPRESSION, Deflater.DEFAULT_STRATEGY);
    }

    /**
     * compresses a byte array, using ZLIB algorithm with the default strategy
     * ({@link Deflater#DEFAULT_STRATEGY}).
     * @param input the byte array to compress
     * @param level compression level, see {@link Deflater#setLevel}
     * @return compressed the compressed data prefixed with a 4 byte (long) length
     */
    public static byte[] compressData(byte input[], int level) {
        return compressData(input, level, Deflater.DEFAULT_STRATEGY);
    }

    /**
     * compresses a byte array, using ZLIB algorithm
     * @param input the byte array to compress
     * @param level compression level, see {@link Deflater#setLevel}
     * @param strategy compression strategy, see {@link Deflater#setStrategy}
     * @return compressed the compressed data prefixed with a 4 byte (long) length
     */
    public static byte[] compressData(byte input[], int level, int strategy) {

        byte res[];

        LinkedList results = new LinkedList();

        // Compress the bytes
        Deflater compresser = new Deflater(level);
        compresser.setStrategy(strategy);
        compresser.setInput(input);
        compresser.finish();

        byte[] output = new byte[5120];

        // read the compression in small chunks
        int total = 0;
        int compressedDataLength = -1;
        while( !compresser.finished() ){
            compressedDataLength = compresser.deflate(output);
            total += compressedDataLength;
            if( compressedDataLength != 0 ) {
                results.add(new BufferAndLength( output, compressedDataLength));
                output = new byte[5120];
            }
        }

        res = new byte[total + 4];

        // copy into the result array
        putInt(res, 0, input.length);
        int location = 4;
        for (Iterator resultsIterator = results.iterator(); resultsIterator.hasNext() ; ){
            BufferAndLength item = (BufferAndLength)resultsIterator.next();
            System.arraycopy(item.buf, 0, res, location, item.len);
            location += item.len;
        }

        return res;

    }

    /**
     * decompresses a byte array, using ZLIB algorithm
     * @param input
     * @return decompressed data
     */
    public static byte[] decompressData(byte input[]) throws DataFormatException {

        int srcLen = getInt(input, 0);

        // Compress the bytes
        Inflater decompresser = new Inflater();
        decompresser.setInput(input, 4, input.length - 4);
        byte[] output = new byte[srcLen];

        // read the compression in small chunks
        int decompressedDataLength = decompresser.inflate(output);
        if (!decompresser.finished() || decompressedDataLength != srcLen){
            throw new DataFormatException("Corrupted input");
        }
        decompresser.end();

        return output;

    }

    /**
     * Convert object to byte array
     * Note: method does not compress
     * @param object
     * @return bytes array that represent this object
     * @throws IOException
     */
    public static byte[] toBytes(Object object) throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ObjectOutputStream oos = new ObjectOutputStream(baos);
        oos.writeObject(object);
        return baos.toByteArray();
    }

    /**
     * Convert byte array into object
     * Note: method does not compress 
     * @param bytes
     * @return consructed object from byte array
     * @throws IOException
     * @throws ClassNotFoundException
     */
    public static Object toObject(byte[] bytes) throws IOException, ClassNotFoundException {
        ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
        ObjectInputStream ois = new ObjectInputStream(bais);
        return ois.readObject();
    }

    /**
     * Convert object to a compressed byte array
     * @param object
     * @return compressed byet array
     * @throws IOException
     */
    public static byte[] toCompressedBytes(Object object) throws IOException {
        byte[] bytes = toBytes(object);
        return Compression.compressData(bytes);
    }

    /**
     * Convert compressed byte array into an object
     * @param compressedBytes
     * @return decomporessed object
     * @throws IOException
     * @throws ClassNotFoundException
     * @throws DataFormatException
     */
    public static Object toUncompressedObject(byte[] compressedBytes) throws IOException, ClassNotFoundException, DataFormatException {
        byte[] bytes = Compression.decompressData(compressedBytes);
        return toObject(bytes);
    }
    

    static int getInt(byte[] b, int off) {
	return ((b[off + 3] & 0xFF) << 0) +
	       ((b[off + 2] & 0xFF) << 8) +
	       ((b[off + 1] & 0xFF) << 16) +
	       ((b[off + 0] & 0xFF) << 24);
    }

    static void putInt(byte[] b, int off, int val) {
	b[off + 3] = (byte) (val >>> 0);
	b[off + 2] = (byte) (val >>> 8);
	b[off + 1] = (byte) (val >>> 16);
	b[off + 0] = (byte) (val >>> 24);
    }

    static class BufferAndLength {

        byte buf[];
        int len;

        public BufferAndLength(byte[] buf, int len) {
            this.buf = buf;
            this.len = len;
        }
    }

    public static boolean compareFloatWithPrecision(float f1, float f2, int prec) {
        float delta = 1.0f;
        String fs1 = "" + f1;
        String fs2 = "" + f2;
        String shortString = (fs1.length() > fs2.length()) ? fs2 : fs1;
        int dotIndex = shortString.indexOf(".");
        if (dotIndex >= prec) {
            //float is bigger than prec
            for (int n = 0; n < dotIndex - prec; n++) {
                delta *= 10.0f;
            }
        } else {
            for (int n = 0; n < prec - dotIndex; n++) {
                delta /= 10.0f;
            }
        }

        if (Math.abs(f1 - f2) <= delta) {
            return true;
        } else {
            return false;
        }
    }
}
