package com.onaro.util;

import java.io.Serializable;
import java.util.*;

public class Counter<E> implements Iterable<E>, Serializable {

    private static final long serialVersionUID = 3L;

    private Map<E, Count> map;

    public Counter() {
        this((Comparator<E>) null);
    }

    /**
     * Create a new Counter
     *
     * @param comparator if not null, TreeMap will be used to holds the counters and the entries will be ordered
     */
    public Counter(Comparator<E> comparator) {
        map = comparator != null ?
                new TreeMap<E, Count>(comparator) :
                new HashMap<E, Count>();
    }


    public Counter(Set<E> set) {
        this();
        merge(set);
    }

    public Counter(Counter<E> counter) {
        this();
        for (E e : counter) {
            set(e, counter.get(e));
        }
    }


    /**
     * Get the current count of an object
     *
     * @param obj an object
     *
     * @return the count of the object
     */
    public long get(E obj) {
        Count count = map.get(obj);
        return count != null ? count.getCount() : 0;
    }

    public long inc(E obj) {
        return inc(obj, 1);
    }

    public void merge(Counter<E> otherCounter) {
        if (otherCounter != null) {
            for (E obj : otherCounter) {
                long thisCount = get(obj);
                long otherCount = otherCounter.get(obj);
                set(obj, thisCount + otherCount);
            }
        }
    }

    /**
     * Increment the count for each of the members of the set
     */
    public void merge(Set<E> set) {
        if (set != null) {
            for (E obj : set) {
                inc(obj);
            }
        }
    }

    public long inc(E obj, long number) {
        Count count = map.get(obj);
        if (count == null) {
            count = new Count();
            map.put(obj, count);
        }

        return count.inc(number);
    }

    public long set(E obj, long number) {
        Count count = map.get(obj);
        if (number > 0) {
            if (count == null) {
                count = new Count();
                map.put(obj, count);
            }

            return count.set(number);
        }

        // avoid from keeping 0 counts
        if (count != null) {
            map.remove(obj);
            return count.getCount();
        }

        return 0;

    }


    public long sum() {
        long sum = 0;
        for (Count count : map.values()) {
            sum += count.getCount();
        }

        return sum;
    }

    public long max() {
        long max = 0;

        for (Count count : map.values()) {
            max = Math.max(max, count.getCount());
        }

        return max;
    }

    public long min() {
        if (!isEmpty()) {
            long min = Long.MAX_VALUE;

            for (Count count : map.values()) {
                min = Math.min(min, count.getCount());
            }

            return min;
        } else {
            return 0;
        }
    }

    public boolean isEmpty() {
        return map.isEmpty();
    }

    public Iterator<E> iterator() {
        return elements().iterator();
    }

    public Set<E> elements() {
        return map.keySet();
    }

    static class Count implements Serializable {
        private static final long serialVersionUID = 1L;
        private long count;

        public long getCount() {
            return count;
        }

        public long inc(long number) {
            if (number < 0) throw new IllegalArgumentException("number < 0");

            long old = count;
            count += number;
            return old;
        }

        @Override
        public String toString() {
            return String.valueOf(count);
        }

        /**
         * change the count to the specified number
         *
         * @param count new count to set
         *
         * @return old count
         */
        public long set(long count) {
            long old = this.count;
            this.count = count;
            return old;
        }
    }


    public String toString() {
        StringBuilder str = new StringBuilder();
        for (Map.Entry<E, Count> entry : map.entrySet()) {
            str.append(entry.getKey()).append(":").append(entry.getValue()).append("\r\n");
        }
        return str.toString();
    }
}
