package com.onaro.util.enumeration;

import org.slf4j.Logger;

import javax.swing.*;
import java.net.URL;
import java.util.*;

/**
 * A dictionary defining the presentation attributes of enumeration values. The attributes (which are name, icon) are
 * read from a <code>ResourceBundle</code>. <p/> The dictionary caches icons to reduce memory utilization and
 * construction time.
 */
public class EnumPresentation {

	private static final String PACKAGE_NAME = "com.onaro.util.enumeration";
	
	/**
	 * Logger for this class.
	 */
	static final Logger logger = org.slf4j.LoggerFactory.getLogger(PACKAGE_NAME);

	
	public static final String RB_BASENAME = PACKAGE_NAME + ".enumeration";
	
	/**
	 * Provides the default enumeration translation, and thus improve performance and reduce the number of times the
	 * enum-configuration-file is being referred.
	 */
	private static final Map<ClassLoader, EnumPresentation> classLoaderToPresentationMap = new HashMap<ClassLoader, EnumPresentation>(2);

	private ClassLoader classLoader;
	
	/**
	 * Gets the Server's default enumeration translation. This should not be called from the Client.
	 * In the Client, use the singleton SANscreenClientPlugin.getDefault().getEnumPresentation() instead.
	 * 
	 * @return the default translation
	 */
	public static EnumPresentation getDefaults() {
		ClassLoader classLoader = EnumPresentation.class.getClassLoader();
		return getDefaults(classLoader);
	}
	
	/**
	 * This is a hack/fix for an incorrect use of resource loading in the client.
	 */
	public static EnumPresentation getDefaults(ClassLoader classLoader) {
		EnumPresentation result = classLoaderToPresentationMap.get(classLoader);
		if (result == null)
		{
			ResourceBundle rb = ResourceBundle.getBundle(RB_BASENAME, Locale.getDefault(), classLoader);
			
			result = new EnumPresentation(classLoader, rb);
			classLoaderToPresentationMap.put(classLoader, result);
		}
		
		return result;
	}

	/**
	 * The resource bundle the attributes are read from.
	 */
	private ResourceBundle resourceBundle;

	/**
	 * A map of maps of <code>Attributes</code>. The top level map indexed by the enumeration type name. The second
	 * level map is indexed by the value itself.
	 */
	private Map<String,AttributesMap> enumAttributes = new HashMap<String,AttributesMap>();

	/**
	 * Map of the set of values of the enumeration types indexed by the type name. Each entry in the map is a list of
	 * values.
	 */
	private Map<String,Attributes[]> enumAttributeValues = new HashMap<String,Attributes[]>();

	/**
	 * Constructs using the given resources.
	 * 
	 * @param resourceBundle
	 *            resource file defining all the enumerations attributes.
	 */
	public EnumPresentation(ClassLoader classLoader, ResourceBundle resourceBundle) {
		
		if (classLoader == null) throw new IllegalArgumentException();
		if (resourceBundle == null) throw new IllegalArgumentException();
		
		this.resourceBundle = resourceBundle;
		this.classLoader = classLoader;
	}
	
	public final ResourceBundle getResourceBundle()
	{
		return resourceBundle;
	}

    /**
     * Get the attributes for the specified enum where the type prefix is the enum's full class name and the value
     * string is the enum's toString() result.<p/>
     *
     * @param value enum whose attributes are keyed in the associated resource bundle by "<i>full class name</i>.<i>enum value</i>".
     * @return Attributes found for this enum
     */
    public Attributes getAttributes(Enum value) {
        if (value == null) {
            return null;
        }
        return getAttributes(value.getClass().getName(), value);
    }
	/**
	 * Gets the attributes of an enumeration value.
	 * 
	 * @param enumType
	 *            the enum type name
	 * @param value
	 *            the value
	 */
	public Attributes getAttributes(String enumType, Object value) {
		if (value == null) {
			return null;
		}

		/**
		 * If the attributes are already cached, just return them. It might exist with null value because its not the
		 * first time it was searched for.
		 */
		String key = value.toString();
		AttributesMap attributeMap = enumAttributes.get(enumType);
		if (attributeMap != null) {
			if (attributeMap.containsKey(key)) {
				return attributeMap.get(key);
			}
		}

		/**
		 * Otherwise, try to fetch them.
		 */
		return buildAttributes(enumType, key);
	}

	public Attributes getAttributes(AttributesMap attributeMap, String enumType, Object value) {
		if (value == null) {
			return null;
		}

		/**
		 * If the attributes are already cached, just return them. It might exist with null value because its not the
		 * first time it was searched for.
		 */
		String key = value.toString();
		if (attributeMap != null) {
			if (attributeMap.containsKey(key)) {
				return attributeMap.get(key);
			}
		}

		/**
		 * Otherwise, try to fetch them.
		 */
		return buildAttributes(enumType, key);
	}

	/**
	 * Gets the attributes of an enumeration value of a specific context. A context allows creating two levels of
	 * enumeration types.
	 * 
	 * @param context
	 *            higher level enumeration type
	 * @param enumType
	 *            the enum type name
	 * @param value
	 *            the value
	 */
	public Attributes getAttributes(String context, String enumType, String value) {
		if (value == null) {
			return null;
		}

		StringBuffer s = new StringBuffer(context).append(".").append(enumType);
		enumType = s.toString();

		/**
		 * If the attributes are already cached, just return them. It might exist with null value because its not the
		 * first time it was searched for.
		 */
		AttributesMap attributeMap = enumAttributes.get(enumType);
		if (attributeMap != null) {
			Attributes attributes = attributeMap.get(value);
			if (attributes != null) {
				return attributes;
			}
		}

		/**
		 * Otherwise, try to fetch them.
		 */
		return buildAttributes(enumType, value);
	}

	public AttributesMap getEnumAttributes(String enumType) {
		AttributesMap attributesMap = enumAttributes.get(enumType);
		if (attributesMap == null) buildAttributeValuess(enumType + ".");
		return enumAttributes.get(enumType);
	}

	/**
	 * Gets the attributes of the values of the given type.
	 * 
	 * @param enumType
	 *            the requested type name
	 * @return array of value attributes of the given type
	 */
	public Attributes[] getValues(String enumType) {
		String typePrefix = enumType + ".";

		/**
		 * If the values are already cached, just return them. It might exist with null value because its not the first
		 * time it was searched for.
		 */
		if (enumAttributeValues.containsKey(typePrefix)) {
			return enumAttributeValues.get(typePrefix);
		}
		
		/**
		 * Otherwise, try to fetch them.
		 */
		return buildAttributeValuess(typePrefix);
	}

	/**
	 * Build the value attributes of an enumeration type from the resources. The new attributes are cached, if not
	 * found, a <code>null</code> is cached. The attributes are considered as defined (and is cached) if they contain
	 * the name attribute.
	 * 
	 * @param typePrefix
	 *            prefix of resource keys (type name followed by the separator)
	 * @return array of value attributes of the given type
	 */
	private Attributes[] buildAttributeValuess(String typePrefix) {
		int prefixLength = typePrefix.length();
		Set<String> typeValues = new LinkedHashSet<String>();

		for (Enumeration<String> keys = getResourceBundle().getKeys(); keys.hasMoreElements();) {
			String key = keys.nextElement();

			if (key.startsWith(typePrefix)) {

				int afterValuePos = key.lastIndexOf('.');
				// JDM: Disabling assertion to allow for keys with no text ex: zone.member.device.type..name=None
//				assert afterValuePos > prefixLength : "Key '" + key + " doesn't contain a valid value for type '"
//						+ typePrefix + "'";

				String enumKey = key.substring(0, afterValuePos);

				typeValues.add(enumKey);
			}
		}

		Attributes[] attributeValues = null;

		if (!typeValues.isEmpty()) {

			attributeValues = new Attributes[typeValues.size()];
			int valueNum = 0;
            for (String enumKey : typeValues) {
                String value = enumKey.substring(prefixLength, enumKey.length());
                enumKey = enumKey.substring(0, enumKey.length() - value.length() - 1);

                attributeValues[valueNum] = getAttributes(enumKey, value);
                ++valueNum;
            }
			Arrays.sort(attributeValues);
			enumAttributeValues.put(typePrefix, attributeValues);
		}

		return attributeValues;
	}

	/**
	 * Build the attributes of an enumeration value from the resources. The new attributes are cached,if not found, a
	 * <code>null</code> is cached. The attributes are considered as defined (and is cached) if they contain the name
	 * attribute.
	 * 
	 * @param enumType
	 *            type name and value for indexing the attributes map
	 * @param value
	 *            value name
	 * @return the requested attribute or <code>null</code> if it doesn't exist
	 */
	private Attributes buildAttributes(String enumType, String value) {
        Attributes attributes = new Attributes(value);

        attributes.enumType = enumType;

        String key = new StringBuffer(enumType).append(".").append(value).toString();

        try {
            attributes.name = getResourceBundle().getString(key + ".name");
        }
        catch (MissingResourceException e) {
        }

        try {
           String ordinalStr = getResourceBundle().getString(key + ".ordinal");
           attributes.ordinal = Integer.valueOf(Integer.parseInt(ordinalStr));
        }
        catch (MissingResourceException e) {
        }
        catch (NumberFormatException e) {
        }

        try {
            attributes.smallIconUrl = getIconUrl(key + ".icon");
            if (attributes.smallIconUrl != null) attributes.smallIcon = new ImageIcon(attributes.smallIconUrl);
        }
        catch (MissingResourceException e) {
        }

        try {
            attributes.largeIconUrl = getIconUrl(key + ".icon.large");
            if (attributes.largeIconUrl != null) attributes.largeIcon = new ImageIcon(attributes.largeIconUrl);
        }
        catch (MissingResourceException e) {
        }

        try {
            attributes.annotationTypeName = getResourceBundle().getString(key + ".annotationTypeName");
        }
        catch (MissingResourceException e) {
        }

        if (attributes.name == null) {
            attributes = null;
        }

        AttributesMap attributeMap = enumAttributes.get(enumType);
        if (attributeMap == null) {
            attributeMap = new AttributesMap();
            enumAttributes.put(enumType, attributeMap);
        }

        attributeMap.put(value, attributes);
        return attributes;
    }

	/**
	 * Gets and icon with the given name as a resource (reading it from the class path).
	 * 
	 * @param name
	 *            the icon's file name, may be null
	 * @return the requested icon or null if <code>name</code> is null
	 */
	public URL getIconUrl(String name) {
		String iconFile = getResourceBundle().getString(name);

		if (iconFile == null || iconFile.length() == 0) return null;

		return classLoader.getResource(iconFile);
	}

	public class AttributesMap {
		private Map<String,Attributes> attributeMap;

		public AttributesMap() {
			attributeMap = new HashMap<String,Attributes>();
		}

		protected Attributes get(String enumType) {
			return attributeMap.get(enumType);
		}

		private void put(String value, Attributes attributes) {
			attributeMap.put(value, attributes);
		}

		protected boolean containsKey(String key) {
			return attributeMap.containsKey(key);
		}
	}

	public static class Attributes implements Comparable<Attributes> {
		public String value;
		
		public String enumType;

		public String name;

		public URL smallIconUrl;

		public ImageIcon smallIcon;

		public URL largeIconUrl;

		public ImageIcon largeIcon;
		
		public Integer ordinal;
		
		/** 
		 * This field is used when an enum value contains a annotation type name.
		 * This is used by the AnnotationsEnumMonitor to replace the name dynamically
		 * when the name changes.  When annotationTypeName is specified, the name property
		 * of the enum should contain a placeholder for the name to be filled in,
		 * for example: Incorrect {0} reserved
		 * 
		 * The value of this field will be null for enum values that don't contain an annotation type name.
		 * 
		 * @see com.onaro.sanscreen.client.view.annotations.AnnotationsEnumMonitor
		 */ 
		public String annotationTypeName;

		public Attributes(String value, String name) {
			this.name = name;
			this.value = value;
		}

		public Attributes(String value) {
			this(value, null);
		}

		/**
		 * Provide only the name because it used when comparing with other objects.
		 * 
		 * @return the name of this attribute
		 */
		@Override
		public String toString() {
			return name;
		}

		/**
		 * Compare an attribute with anything else by comparing the other value's string representation.
		 * 
		 * @return -1, 0, 1 if smaller, equals or greater than the other object
		 */
		public int compareTo(Attributes other) {
			if (other == null) {
				return 1;
			}
			
			if (ordinal != null && other.ordinal != null) {
			    return ordinal.compareTo(other.ordinal);
			}

            int cmp = 0;
			if (name != null) {
				cmp = name.compareToIgnoreCase(other.name);
			}
			else if (other.name != null) {
				cmp = -1;
			}
			return cmp != 0 ? cmp : value.compareTo(other.value);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (!(o instanceof Attributes)) return false;

			final Attributes attributes = (Attributes) o;

			if (ordinal != null ? !ordinal.equals(attributes.ordinal) : attributes.ordinal != null) return false;
			if (largeIcon != null ? !largeIcon.equals(attributes.largeIcon) : attributes.largeIcon != null)
				return false;
			if (enumType != null ? !enumType.equals(attributes.enumType) : attributes.enumType != null) return false;
			if (smallIcon != null ? !smallIcon.equals(attributes.smallIcon) : attributes.smallIcon != null)
				return false;
			if (value != null ? !value.equals(attributes.value) : attributes.value != null) return false;

			return true;
		}

		@Override
		public int hashCode() {
			int result;
			result = (value != null ? value.hashCode() : 0);
			result = 29 * result + (enumType != null ? enumType.hashCode() : 0);
			result = 29 * result + (ordinal != null ? ordinal.hashCode() : 0);
			result = 29 * result + (smallIcon != null ? smallIcon.hashCode() : 0);
			result = 29 * result + (largeIcon != null ? largeIcon.hashCode() : 0);
			return result;
		}
	}
}
