package com.onaro.util.sessions;

import com.netapp.oci.platform.common.ServerConstant;
import javax.ejb.CreateException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.rmi.UnmarshalException;
import java.util.HashMap;
import java.util.Map;

import com.onaro.sanscreen.server.sessions.SessionFactory;


public class SessionCache {

    private static final Logger LOG = LoggerFactory.getLogger(SessionCache.class);

    private boolean stopping = false;
    private static Context initialContext = null;

    private String user;
    private String pass;
    private int timeoutMillis;

    private ClassLoader classLoader;

    /**
     * Initializes the session cache.
     *
     * @param classLoader   The class loader used to load the proxy object. This should be the loader of the class using the session.
     * @param timeoutMillis Total tims to try to recover from communication problems before giving up. When 0, no retries will be attempted
     */
    public SessionCache(ClassLoader classLoader, String user, String pass, final int timeoutMillis) {
        this.timeoutMillis = timeoutMillis;
        this.user = user;
        this.pass = pass;
        this.classLoader = classLoader;
    }

    /**
     * Initializes the session cache.
     *
     * @param classLoader   The class loader used to load the proxy object. This should be the loader of the class using the session.
     * @param timeoutMillis Total tims to try to recover from communication problems before giving up. When 0, no retries will be attempted
     * @param logExceptions
     * @deprecated Use {@link #SessionCache(ClassLoader, ClientLogin, int).
     */
    @Deprecated
    public SessionCache(ClassLoader classLoader, String user, String pass, final int timeoutMillis, boolean logExceptions) {
        this(classLoader, user, pass, timeoutMillis);
    }

    public synchronized void stop() {
        stopping = true;

        // TODO: Interrupt all proxies
    }

    public synchronized boolean isStopping() {
        return stopping;
    }

    /**
     * Retrieves an EJB session.
     * This method will try to reuse previously created sessions and home interfaces (if possible).
     * The retrieved object is a Proxy that wrappes the Session object to detect network failures and try to recover
     * from then.
     *
     * @param jndiName JNDI name of EJB session to retrieve
     * @throws SessionCacheException
     */
    public synchronized <E> E getSession(final String jndiName) throws SessionCacheException {

        try {
            // Perform an additional lookup to check the type of object retrurned.
            return (E) lookupSession(jndiName);
        } catch (RemoteException e) {
            throw new SessionCacheException("Failed to lookup session " + jndiName, e);
        } catch (NamingException e) {
            throw new SessionCacheException("Failed to lookup session '" + jndiName +"'", e);
        } catch (MalformedURLException e) {
            throw new SessionCacheException("Failed to lookup session '" + jndiName +"'", e);
        }
    }

    public synchronized void cleanCache() {
        // do nothing for now. Check if we need to close RemoteAppSessionFactory.INSTANCE root context
        // but RemoteAppSessionFactory consumers...
    }

    /**
     * Retrieves an EJB session reference (after logging in, looking up, etc...). This method will try to reuse the
     * session cache if possible to save network calls.
     */
    public synchronized Object lookupSession(final String jndiName) throws RemoteException, NamingException, MalformedURLException {
        if (isStopping()) {
            throw new RemoteException("Communication Manager is currently stopping!");
        }

        if (SessionFactory.isRemoteApp()) {
            if(!RemoteAppSessionFactory.INSTANCE.isInited()) {
                try {
                    String jbossHttpPort = System.getProperty("jboss.http.port");
                    final URL codeBaseURL = new URL(System.getProperty("codebase"));
                    RemoteAppSessionFactory.INSTANCE.init(codeBaseURL.getHost(), jbossHttpPort, user, pass, timeoutMillis, true);
                } catch (MalformedURLException murle) {
                    LOG.error("Failed to parse codebase URL: did you specify a -Dcodebase= system property?");
                    throw murle;
                }
            }
            return RemoteAppSessionFactory.INSTANCE.lookup(jndiName);
        }

        if (initialContext == null) {
            initialContext = new InitialContext();
        }

        return initialContext.lookup(jndiName);
    }
}
